@@ -6,7 +6,7 @@
 
 #include <unistd.h>
 #include <pcp/pmapi.h>
-#include <pcp/impl.h>
+#include "libpcp.h"
 
 #define N_PMID_A sizeof(metrics_a)/sizeof(metrics_a[0])
 #define N_PMID_B sizeof(metrics_b)/sizeof(metrics_b[0])
@@ -59,14 +59,14 @@ main(int argc, char **argv)
     int		status = 0;
     int		done;
 
-    __pmSetProgname(argv[0]);
+    pmSetProgname(argv[0]);
 
     while ((c = getopt(argc, argv, "a:c:D:fl:n:s:t:VzZ:?")) != EOF) {
 	switch (c) {
 
 	case 'a':	/* archive name */
 	    if (type != 0) {
-		fprintf(stderr, "%s: at most one of -a and/or -h allowed\n", pmProgname);
+		fprintf(stderr, "%s: at most one of -a and/or -h allowed\n", pmGetProgname());
 		errflag++;
 	    }
 	    type = PM_CONTEXT_ARCHIVE;
@@ -75,25 +75,21 @@ main(int argc, char **argv)
 
 	case 'c':	/* configfile */
 	    if (configfile != (char *)0) {
-		fprintf(stderr, "%s: at most one -c option allowed\n", pmProgname);
+		fprintf(stderr, "%s: at most one -c option allowed\n", pmGetProgname());
 		errflag++;
 	    }
 	    configfile = optarg;
 	    break;	
 
-#ifdef PCP_DEBUG
 
-	case 'D':	/* debug flag */
-	    sts = __pmParseDebug(optarg);
+	case 'D':	/* debug options */
+	    sts = pmSetDebug(optarg);
 	    if (sts < 0) {
-		fprintf(stderr, "%s: unrecognized debug flag specification (%s)\n",
-		    pmProgname, optarg);
+		fprintf(stderr, "%s: unrecognized debug options specification (%s)\n",
+		    pmGetProgname(), optarg);
 		errflag++;
 	    }
-	    else
-		pmDebug |= sts;
 	    break;
-#endif
 
 	case 'f':	/* force */
 	    force++; 
@@ -101,7 +97,7 @@ main(int argc, char **argv)
 
 	case 'h':	/* contact PMCD on this hostname */
 	    if (type != 0) {
-		fprintf(stderr, "%s: at most one of -a and/or -h allowed\n", pmProgname);
+		fprintf(stderr, "%s: at most one of -a and/or -h allowed\n", pmGetProgname());
 		errflag++;
 	    }
 	    host = optarg;
@@ -119,7 +115,7 @@ main(int argc, char **argv)
 	case 's':	/* sample count */
 	    samples = (int)strtol(optarg, &endnum, 10);
 	    if (*endnum != '\0' || samples < 0) {
-		fprintf(stderr, "%s: -s requires numeric argument\n", pmProgname);
+		fprintf(stderr, "%s: -s requires numeric argument\n", pmGetProgname());
 		errflag++;
 	    }
 	    break;
@@ -127,7 +123,7 @@ main(int argc, char **argv)
 	case 't':	/* delta seconds (double) */
 	    delta = strtod(optarg, &endnum);
 	    if (*endnum != '\0' || delta <= 0.0) {
-		fprintf(stderr, "%s: -t requires floating point argument\n", pmProgname);
+		fprintf(stderr, "%s: -t requires floating point argument\n", pmGetProgname());
 		errflag++;
 	    }
 	    break;
@@ -138,7 +134,7 @@ main(int argc, char **argv)
 
 	case 'z':	/* timezone from host */
 	    if (tz != (char *)0) {
-		fprintf(stderr, "%s: at most one of -Z and/or -z allowed\n", pmProgname);
+		fprintf(stderr, "%s: at most one of -Z and/or -z allowed\n", pmGetProgname());
 		errflag++;
 	    }
 	    zflag++;
@@ -146,7 +142,7 @@ main(int argc, char **argv)
 
 	case 'Z':	/* $TZ timezone */
 	    if (zflag) {
-		fprintf(stderr, "%s: at most one of -Z and/or -z allowed\n", pmProgname);
+		fprintf(stderr, "%s: at most one of -Z and/or -z allowed\n", pmGetProgname());
 		errflag++;
 	    }
 	    tz = optarg;
@@ -160,7 +156,7 @@ main(int argc, char **argv)
     }
 
     if (zflag && type == 0) {
-	fprintf(stderr, "%s: -z requires an explicit -a or -h option\n", pmProgname);
+	fprintf(stderr, "%s: -z requires an explicit -a or -h option\n", pmGetProgname());
 	errflag++;
     }
 
@@ -171,7 +167,7 @@ main(int argc, char **argv)
 Options\n\
   -a   archive	  metrics source is an archive log\n\
   -c   configfile file to load configuration from\n\
-  -D   debug	  standard PCP debug flag\n\
+  -D   debugspec  standard PCP debugging options\n\
   -f		  force .. \n\
   -h   host	  metrics source is PMCD on host\n\
   -l   logfile	  redirect diagnostics and trace output\n\
@@ -181,19 +177,19 @@ Options\n\
   -V 	          verbose/diagnostic output\n\
   -z              set reporting timezone to local time for host from -a or -h\n\
   -Z   timezone   set reporting timezone\n",
-		pmProgname);
+		pmGetProgname());
 	exit(1);
     }
 
     if (logfile != (char *)0) {
-	__pmOpenLog(pmProgname, logfile, stderr, &sts);
+	pmOpenLog(pmGetProgname(), logfile, stderr, &sts);
 	if (sts < 0) {
-	    fprintf(stderr, "%s: Could not open logfile \"%s\"\n", pmProgname, logfile);
+	    fprintf(stderr, "%s: Could not open logfile \"%s\"\n", pmGetProgname(), logfile);
 	}
     }
 
-    if ((sts = pmLoadNameSpace(namespace)) < 0) {
-	printf("%s: Cannot load namespace from \"%s\": %s\n", pmProgname, namespace, pmErrStr(sts));
+    if (namespace != PM_NS_DEFAULT && (sts = pmLoadASCIINameSpace(namespace, 1)) < 0) {
+	printf("%s: Cannot load namespace from \"%s\": %s\n", pmGetProgname(), namespace, pmErrStr(sts));
 	exit(1);
     }
 
@@ -205,31 +201,31 @@ Options\n\
     if ((sts = pmNewContext(type, host)) < 0) {
 	if (type == PM_CONTEXT_HOST)
 	    fprintf(stderr, "%s: Cannot connect to PMCD on host \"%s\": %s\n",
-		pmProgname, host, pmErrStr(sts));
+		pmGetProgname(), host, pmErrStr(sts));
 	else
 	    fprintf(stderr, "%s: Cannot open archive \"%s\": %s\n",
-		pmProgname, host, pmErrStr(sts));
+		pmGetProgname(), host, pmErrStr(sts));
 	exit(1);
     }
 
     if (type == PM_CONTEXT_ARCHIVE) {
 	if ((sts = pmGetArchiveLabel(&label)) < 0) {
 	    fprintf(stderr, "%s: Cannot get archive label record: %s\n",
-		pmProgname, pmErrStr(sts));
+		pmGetProgname(), pmErrStr(sts));
 	    exit(1);
 	}
 	pmGetArchiveEnd(&eol);
 	eol.tv_sec -= 1;
     }
     else {
-	fprintf(stderr, "%s: must use an archive\n", pmProgname);
+	fprintf(stderr, "%s: must use an archive\n", pmGetProgname());
 	exit(1);
     }
 
     if (zflag) {
 	if ((tzh = pmNewContextZone()) < 0) {
 	    fprintf(stderr, "%s: Cannot set context timezone: %s\n",
-		pmProgname, pmErrStr(tzh));
+		pmGetProgname(), pmErrStr(tzh));
 	    exit(1);
 	}
 	if (type == PM_CONTEXT_ARCHIVE)
@@ -241,7 +237,7 @@ Options\n\
     else if (tz != (char *)0) {
 	if ((tzh = pmNewZone(tz)) < 0) {
 	    fprintf(stderr, "%s: Cannot set timezone to \"%s\": %s\n",
-		pmProgname, tz, pmErrStr(tzh));
+		pmGetProgname(), tz, pmErrStr(tzh));
 	    exit(1);
 	}
 	printf("Note: timezone set to \"TZ=%s\"\n\n", tz);
@@ -257,7 +253,16 @@ Options\n\
 
 
     if ((sts = pmLookupName(N_PMID_A, metrics_a, pmid_a)) != N_PMID_A) {
-	fprintf(stderr, "%s: pmLookupName: %s\n", pmProgname, pmErrStr(sts));
+	if (sts < 0)
+	    fprintf(stderr, "%s: pmLookupName: %s\n", pmGetProgname(), pmErrStr(sts));
+	else {
+	    int		i;
+	    for (i = 0; i < sts; i++) {
+		if (pmid_a[i] != PM_ID_NULL) continue;
+		sts = pmLookupName(1, &metrics_a[i], &pmid_a[i]);
+		fprintf(stderr, "%s: %s: lookup failed: %s\n", pmGetProgname(), metrics_a[i], pmErrStr(sts));
+	    }
+	}
 	exit(1);
     }
     for (i = 0; i < N_PMID_A; i++) {
@@ -265,7 +270,16 @@ Options\n\
     }
 
     if ((sts = pmLookupName(N_PMID_B, metrics_b, pmid_b)) != N_PMID_B) {
-	fprintf(stderr, "%s: pmLookupName: %s\n", pmProgname, pmErrStr(sts));
+	if (sts < 0)
+	    fprintf(stderr, "%s: pmLookupName: %s\n", pmGetProgname(), pmErrStr(sts));
+	else {
+	    int		i;
+	    for (i = 0; i < sts; i++) {
+		if (pmid_b[i] != PM_ID_NULL) continue;
+		sts = pmLookupName(1, &metrics_b[i], &pmid_b[i]);
+		fprintf(stderr, "%s: %s: lookup failed: %s\n", pmGetProgname(), metrics_b[i], pmErrStr(sts));
+	    }
+	}
 	exit(1);
     }
     for (i = 0; i < N_PMID_B; i++) {
@@ -273,7 +287,7 @@ Options\n\
     }
 
     if ((sts = pmLookupName(N_PMID_C, metrics_c, pmid_c)) != N_PMID_C) {
-	fprintf(stderr, "%s: pmLookupName: %s\n", pmProgname, pmErrStr(sts));
+	fprintf(stderr, "%s: pmLookupName: %s\n", pmGetProgname(), pmErrStr(sts));
 	exit(1);
     }
     for (i = 0; i < N_PMID_C; i++) {
@@ -285,12 +299,12 @@ Options\n\
     start.tv_sec += 2;
 
     printf("Start at: ");
-    __pmPrintStamp(stdout, &start);
+    pmPrintStamp(stdout, &start);
     printf("\n\n");
 
     printf("Pass One: rewind and fetch metrics_a until end of log\n");
     if ((sts = pmSetMode(PM_MODE_INTERP, &start, (int)(delta * 1000))) < 0) {
-	fprintf(stderr, "%s: pmSetMode: %s\n", pmProgname, pmErrStr(sts));
+	fprintf(stderr, "%s: pmSetMode: %s\n", pmGetProgname(), pmErrStr(sts));
 	exit(1);
     }
 
@@ -298,14 +312,14 @@ Options\n\
     for (sample=0; !done; sample++) {
 	if ((sts = pmFetch(N_PMID_A, pmid_a, &result)) < 0) {
 	    if (sts != PM_ERR_EOL) {
-		fprintf(stderr, "%s: pmFetch: %s\n", pmProgname, pmErrStr(sts));
+		fprintf(stderr, "%s: pmFetch: %s\n", pmGetProgname(), pmErrStr(sts));
 		status = 1;
 	    }
 	    break;
 	}
 
 	printf("sample %3d time=", sample);
-	__pmPrintStamp(stdout, &result->timestamp);
+	pmPrintStamp(stdout, &result->timestamp);
 	putchar(' ');
 	if (result->numpmid != N_PMID_A) {
 	    printf("Error: expected %d (got %d) value sets\n",
@@ -326,13 +340,12 @@ Options\n\
 	    result->timestamp.tv_usec > eol.tv_usec)
 		done = 1;
 
-	if (result != (pmResult *)0)
-	    pmFreeResult(result);
+	pmFreeResult(result);
     }
 
     printf("Pass Two: rewind and fetch metrics_b until end of log\n");
     if ((sts = pmSetMode(PM_MODE_INTERP, &start, (int)(delta * 1000))) < 0) {
-	fprintf(stderr, "%s: pmSetMode: %s\n", pmProgname, pmErrStr(sts));
+	fprintf(stderr, "%s: pmSetMode: %s\n", pmGetProgname(), pmErrStr(sts));
 	exit(1);
     }
 
@@ -340,14 +353,14 @@ Options\n\
     for (sample=0; !done; sample++) {
 	if ((sts = pmFetch(N_PMID_B, pmid_b, &result)) < 0) {
 	    if (sts != PM_ERR_EOL) {
-		fprintf(stderr, "%s: pmFetch: %s\n", pmProgname, pmErrStr(sts));
+		fprintf(stderr, "%s: pmFetch: %s\n", pmGetProgname(), pmErrStr(sts));
 		status = 1;
 	    }
 	    break;
 	}
 
 	printf("sample %3d time=", sample);
-	__pmPrintStamp(stdout, &result->timestamp);
+	pmPrintStamp(stdout, &result->timestamp);
 	putchar(' ');
 	if (result->numpmid != N_PMID_B) {
 	    printf("Error: expected %d (got %d) value sets\n",
@@ -356,7 +369,7 @@ Options\n\
 	}
 	else {
 	    if (result->vset[0]->numval != 1 ||
-		result->vset[0]->numval != 1) {
+		result->vset[1]->numval != 1) {
 		printf("Error: incorrect number of values\n");
 		status = 1;
 		__pmDumpResult(stdout, result);
@@ -369,13 +382,12 @@ Options\n\
 	    result->timestamp.tv_usec > eol.tv_usec)
 		done = 1;
 
-	if (result != (pmResult *)0)
-	    pmFreeResult(result);
+	pmFreeResult(result);
     }
 
     printf("Pass Three: rewind and fetch metrics_c until end of log\n");
     if ((sts = pmSetMode(PM_MODE_INTERP, &start, (int)(delta * 1000))) < 0) {
-	fprintf(stderr, "%s: pmSetMode: %s\n", pmProgname, pmErrStr(sts));
+	fprintf(stderr, "%s: pmSetMode: %s\n", pmGetProgname(), pmErrStr(sts));
 	exit(1);
     }
 
@@ -383,14 +395,14 @@ Options\n\
     for (sample=0; !done; sample++) {
 	if ((sts = pmFetch(N_PMID_C, pmid_c, &result)) < 0) {
 	    if (sts != PM_ERR_EOL) {
-		fprintf(stderr, "%s: pmFetch: %s\n", pmProgname, pmErrStr(sts));
+		fprintf(stderr, "%s: pmFetch: %s\n", pmGetProgname(), pmErrStr(sts));
 		status = 1;
 	    }
 	    break;
 	}
 
 	printf("sample %3d time=", sample);
-	__pmPrintStamp(stdout, &result->timestamp);
+	pmPrintStamp(stdout, &result->timestamp);
 	putchar(' ');
 	if (result->numpmid != N_PMID_C) {
 	    printf("Error: expected %d (got %d) value sets\n",
@@ -399,7 +411,7 @@ Options\n\
 	}
 	else {
 	    if (result->vset[0]->numval != 1 ||
-		result->vset[0]->numval != 1) {
+		result->vset[1]->numval != 1) {
 		printf("Error: incorrect number of values\n");
 		status = 1;
 		__pmDumpResult(stdout, result);
@@ -412,8 +424,7 @@ Options\n\
 	    result->timestamp.tv_usec > eol.tv_usec)
 		done = 1;
 
-	if (result != (pmResult *)0)
-	    pmFreeResult(result);
+	pmFreeResult(result);
     }
 
     exit(status);
