@@ -17,7 +17,6 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
 #include <rpc/rpc.h>
 
 #define AUTOFSD_PROGRAM 100099UL
@@ -34,10 +33,10 @@ main(int argc, char **argv)
     enum clnt_stat	stat;
     int			c;
     char		*p;
-    char		*host = "localhost";
+    char		*host = "local:";
     int			errflag = 0;
 
-    __pmSetProgname(argv[0]);
+    pmSetProgname(argv[0]);
 
     while ((c = getopt(argc, argv, "h:t:?")) != EOF) {
 	switch (c) {
@@ -48,7 +47,7 @@ main(int argc, char **argv)
 
 	case 't':	/* change timeout interval */
 	    if (pmParseInterval(optarg, &tv, &p) < 0) {
-		fprintf(stderr, "%s: illegal -t argument\n", pmProgname);
+		fprintf(stderr, "%s: illegal -t argument\n", pmGetProgname());
 		fputs(p, stderr);
 		free(p);
 		errflag++;
@@ -57,7 +56,7 @@ main(int argc, char **argv)
 
 	case '?':
 	default:
-	    fprintf(stderr, "Usage: %s [-h host] [-t timeout]\n", pmProgname);
+	    fprintf(stderr, "Usage: %s [-h host] [-t timeout]\n", pmGetProgname());
 	    errflag++;
 	    break;
 	}
