@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013 Red Hat.
+ * Copyright (c) 2013,2017 Red Hat.
  * Copyright (c) 1995,2003,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -15,7 +15,6 @@
 
 #include <sys/stat.h>
 #include "pmapi.h"
-#include "impl.h"
 
 #ifdef HAVE_VALUES_H
 #include <values.h>
@@ -32,11 +31,11 @@
 #include "./lex.h"
 #include "./gram.h"
 
-static __pmTimeval	now = { 0, 0 };
+static pmTimeval	now;
 
 int			infd;
 int			outfd;
-char			*myPmdaName = 0;
+char			*myPmdaName;
 
 extern int		_creds_timeout;
 
@@ -79,18 +78,14 @@ agent_creds(__pmPDU *pb)
 	return sts;
 
     for (i = 0; i < credcount; i++) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_CONTEXT)
+	if (pmDebugOptions.context)
 	    fprintf(stderr, "agent_creds: doing cred #%d from PID %d\n", i+1, sender);
-#endif
 	switch(credlist[i].c_type) {
 	case CVERSION:
 	    version = credlist[i].c_vala;
 	    vflag = 1;
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_CONTEXT)
+	    if (pmDebugOptions.context)
 		fprintf(stderr, "agent_creds: version cred (%u)\n", version);
-#endif
 	    break;
 	}
     }
@@ -175,9 +170,9 @@ openpmda(char *fname)
     }
 }
 
-#ifdef HAVE_SYS_UN_H
+#ifdef HAVE_STRUCT_SOCKADDR_UN
 void
-opensocket(char *fname)
+open_unix_socket(char *fname)
 {
     int			fd;
     struct stat		buf;
@@ -222,12 +217,72 @@ opensocket(char *fname)
 }
 #else
 void
-opensocket(char *fname)
+open_unix_socket(char *fname)
 {
-	__pmNotifyErr(LOG_CRIT, "UNIX domain sockets unsupported\n");
+    pmNotifyErr(LOG_CRIT, "UNIX domain sockets unsupported\n");
 }
 #endif
 
+#define MYSOCKETSZ 64
+
+static void
+open_socket(int port, int family, const char *protocol)
+{
+    __pmSockAddr	*addr;
+    int			fd, sts;
+    char		socket[MYSOCKETSZ];
+
+    fd = (family == AF_INET) ? __pmCreateSocket() : __pmCreateIPv6Socket();
+    if (fd < 0) {
+	fprintf(stderr, "opensocket: socket: %s\n", netstrerror());
+	return;
+    }
+
+    addr = __pmLoopBackAddress(family);
+    if (addr == NULL) {
+	fprintf(stderr, "opensocket: loopback: %s\n", netstrerror());
+	__pmCloseSocket(fd);
+	return;
+    }
+
+    closepmda();
+
+    __pmSockAddrSetPort(addr, port);
+    sts = __pmConnect(fd, addr, __pmSockAddrSize());
+    __pmSockAddrFree(addr);
+
+    if (sts < 0) {
+	fprintf(stderr, "opensocket: connect: %s\n", netstrerror());
+	__pmCloseSocket(fd);
+	return;
+    }
+
+    infd = fd;
+    outfd = fd;
+
+    pmsprintf(socket, MYSOCKETSZ, "%s port %d", protocol, port);
+    printf("Connect to PMDA on %s\n", socket);
+
+    connmode = CONN_DAEMON;
+    reset_profile();
+    if (myPmdaName != NULL)
+	free(myPmdaName);
+    myPmdaName = strdup(socket);
+    pmdaversion();
+}
+
+void
+open_inet_socket(int port)
+{
+    open_socket(port, AF_INET, "inet");
+}
+
+void
+open_ipv6_socket(int port)
+{
+    open_socket(port, AF_INET6, "ipv6");
+}
+
 void
 closepmda(void)
 {
@@ -258,11 +313,9 @@ dopmda_desc(pmID pmid, pmDesc *desc, int print)
 	if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_DESC) {
 	    if ((sts = __pmDecodeDesc(pb, desc)) >= 0) {
 		if (print)
-		    __pmPrintDesc(stdout, desc);
-#ifdef PCP_DEBUG
-		    else if (pmDebug & DBG_TRACE_PDU)
-			__pmPrintDesc(stdout, desc);
-#endif
+		    pmPrintDesc(stdout, desc);
+		    else if (pmDebugOptions.pdu)
+			pmPrintDesc(stdout, desc);
             }
 	    else
 		printf("Error: __pmDecodeDesc() failed: %s\n", pmErrStr(sts));
@@ -293,12 +346,15 @@ dopmda(int pdu)
     int			sts;
     pmDesc		desc;
     pmDesc		*desc_list = NULL;
-    pmResult		*result;
-    __pmInResult	*inresult;
+    pmResult		*result = NULL;
+    pmLabelSet		*labelset = NULL;
+    pmInResult	*inresult;
     __pmPDU		*pb;
     int			i;
     int			j;
+    int			type;
     int			ident;
+    int			numsets;
     int			length;
     char		*buffer;
     struct timeval	start;
@@ -311,7 +367,7 @@ dopmda(int pdu)
     int			pinpdu;
 
     if (timer != 0)
-	__pmtimevalNow(&start);
+	pmtimevalNow(&start);
   
     switch (pdu) {
 
@@ -398,6 +454,8 @@ dopmda(int pdu)
 		    else
 			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		}
+		else if (sts == 0)
+		    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
 
@@ -435,10 +493,8 @@ dopmda(int pdu)
 		    if ((sts = __pmDecodeResult(pb, &result)) < 0)
 			printf("Error: __pmDecodeResult() failed: %s\n", 
 			       pmErrStr(sts));
-#ifdef PCP_DEBUG
-		    else if (pmDebug & DBG_TRACE_FETCH)
+		    else if (pmDebugOptions.fetch)
 			__pmDumpResult(stdout, result);
-#endif
 		}
 		else if (sts == PDU_ERROR) {
 		    if ((i = __pmDecodeError(pb, &sts)) >= 0)
@@ -446,6 +502,8 @@ dopmda(int pdu)
 		    else
 			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		}
+		else if (sts == 0)
+		    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
 	    }
@@ -484,6 +542,8 @@ dopmda(int pdu)
 		    else
 			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		}
+		else if (sts == 0)
+		    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
 
@@ -538,6 +598,8 @@ dopmda(int pdu)
 			else
 			    printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		    }
+		    else if (sts == 0)
+			printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		    else
 			printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
 		    if (pinpdu > 0)
@@ -546,10 +608,70 @@ dopmda(int pdu)
 		}
 		else
 		    printf("Error: __pmSendTextReq() failed: %s\n", pmErrStr(sts));
-	    
 	    }
 	    break;
 
+	case PDU_LABEL_REQ:
+	    if (param.number & PM_LABEL_INDOM) {
+		printf("pmInDom: %s\n", pmInDomStr(param.indom));
+		ident = param.indom;
+	    }
+	    else if (param.number & PM_LABEL_CLUSTER) {
+		printf("Cluster: %s\n", strcluster(param.pmid));
+		ident = param.pmid;
+	    }
+	    else if (param.number & PM_LABEL_ITEM) {
+		printf("PMID: %s\n", pmIDStr(param.pmid));
+		ident = param.pmid;
+	    }
+	    else if (param.number & PM_LABEL_INSTANCES) {
+		printf("Instances of pmInDom: %s\n", pmInDomStr(param.indom));
+		ident = param.indom;
+	    }
+	    else /* param.number & (PM_LABEL_DOMAIN|PM_LABEL_CONTEXT) */
+		ident = PM_IN_NULL;
+
+	    if ((sts = __pmSendLabelReq(outfd, FROM_ANON, ident, param.number)) >= 0) {
+		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_LABEL) {
+		    if ((sts = __pmDecodeLabel(pb, &ident, &type, &labelset, &numsets)) >= 0) {
+			for (i = 0; i < numsets; i++) {
+			    if (labelset[i].inst != PM_IN_NULL)
+				printf("[%3d] Labels inst: %d\n", i, labelset[i].inst);
+			    else
+				printf("Labels:\n");
+			    for (j = 0; j < labelset[i].nlabels; j++) {
+				pmLabel	*lp = &labelset[i].labels[j];
+				char *name = labelset[i].json + lp->name;
+				char *value = labelset[i].json + lp->value;
+				printf("    %.*s=%.*s\n", lp->namelen, name, lp->valuelen, value);
+			    }
+			}
+			if (numsets > 0)
+			    pmFreeLabelSets(labelset, numsets);
+			else
+			    printf("Info: __pmDecodeLabel() returns 0 label sets\n");
+		    }
+		    else
+			printf("Error: __pmDecodeLabel() failed: %s\n", pmErrStr(sts));
+		}
+		else if (sts == PDU_ERROR) {
+		    if ((i = __pmDecodeError(pb, &sts)) >= 0)
+			printf("Error PDU: %s\n", pmErrStr(sts));
+		    else
+			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
+		}
+		else if (sts == 0)
+		    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
+		else
+		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
+	    }
+	    else
+		printf("Error: __pmSendLabelReq() failed: %s\n", pmErrStr(sts));
+
+	    break;
+
 	case PDU_PMNS_IDS:
             printf("PMID: %s\n", pmIDStr(param.pmid));
 	    if ((sts = __pmSendIDList(outfd, FROM_ANON, 1, &param.pmid, 0)) >= 0) {
@@ -569,6 +691,8 @@ dopmda(int pdu)
 		    else
 			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		}
+		else if (sts == 0)
+		    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
 
@@ -596,6 +720,8 @@ dopmda(int pdu)
 		    else
 			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		}
+		else if (sts == 0)
+		    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		else 
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
 
@@ -626,6 +752,8 @@ dopmda(int pdu)
 		    else
 			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		}
+		else if (sts == 0)
+		    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
 
@@ -640,7 +768,7 @@ dopmda(int pdu)
             printf("Metric: %s\n", param.name);
 	    if ((sts = __pmSendTraversePMNSReq(outfd, FROM_ANON, param.name)) >= 0) {
 		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_NAMES) {
-		    if ((sts = __pmDecodeNameList(pb, &numnames, &namelist, &statuslist)) >= 0) {
+		    if ((sts = __pmDecodeNameList(pb, &numnames, &namelist, NULL)) >= 0) {
 			for (i = 0; i < numnames; i++) {
 			    printf("   %s\n", namelist[i]);
 			}
@@ -655,6 +783,8 @@ dopmda(int pdu)
 		    else
 			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		}
+		else if (sts == 0)
+		    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
 
@@ -675,7 +805,7 @@ dopmda(int pdu)
 	    name[sizeof(name)-1] = '\0';
 
 	    printf("Attribute: %s=%s\n", name, buffer ? buffer : "''");
-	    if ((sts = __pmSendAuth(outfd, (int)getpid(), j, buffer, length)) >= 0)
+	    if ((sts = __pmSendAuth(outfd, 0 /* context */, j, buffer, length)) >= 0)
 		printf("Success\n");
 	    else
 		printf("Error: __pmSendAuth() failed: %s\n", pmErrStr(sts));
@@ -688,8 +818,8 @@ dopmda(int pdu)
     }
 
     if (sts >= 0 && timer != 0) {
-	__pmtimevalNow(&end);
-	printf("Timer: %f seconds\n", __pmtimevalSub(&end, &start));
+	pmtimevalNow(&end);
+	printf("Timer: %f seconds\n", pmtimevalSub(&end, &start));
     }
 }
 
@@ -755,12 +885,14 @@ fillResult(pmResult *result, int type)
 
 	if (vsp->numval == 0) {
 	    printf("Error: %s not available!\n", pmIDStr(param.pmid));
-	    return PM_ERR_VALUE;
+	    sts = PM_ERR_VALUE;
+	    goto done;
 	}
 
 	if (vsp->numval < 0) {
 	    printf("Error: %s: %s\n", pmIDStr(param.pmid), pmErrStr(vsp->numval));
-	    return vsp->numval;
+	    sts = vsp->numval;
+	    goto done;
 	}
 
 	for (i = 0; i < vsp->numval; i++) {
@@ -777,6 +909,10 @@ fillResult(pmResult *result, int type)
 	}
     }
 
+done:
+    if (type == PM_TYPE_STRING && atom.cp != NULL)
+	free(atom.cp);
+
     return sts;
 }
 
