@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013 Red Hat.
+ * Copyright (c) 2013,2017 Red Hat.
  * Copyright (c) 1995-2001 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -16,12 +16,22 @@
 #include "./dbpmda.h"
 #include "./lex.h"
 #include "./gram.h"
+#include <pcp/deprecated.h>
+
+/*
+ * This works because ...
+ * we get _another_ copy of debug_map[] from pmdbg.h, but this is a
+ * static array, and the options field points back into the real
+ * pmDebugOptions struct in libpcp, so we can see the internal state
+ * of the debugging options.
+ */
+#include "../../libpcp/src/pmdbg.h"
 
 extern pmdaInterface	dispatch;
 extern int		infd;
 extern int		outfd;
 
-__pmProfile		*profile;
+pmProfile		*profile;
 int			profile_changed;
 int			timer;
 int			get_desc;
@@ -33,27 +43,15 @@ static __pmContext	*ctxp;
 static char		**argv;
 static int		argc;
 
-/*
- * Warning: order of these strings _must_ match bit field sequence defined
- *	    in impl.h for DBG_TRACE_* macros
- */
-static char* debugFlags[] = {
-    "pdu", "fetch", "profile", "value", "context", "indom", "pdubuf", "log",
-    "logmeta", "optfetch", "af", "appl0", "appl1", "appl2", "pmns", "libpmda",
-    "timecontrol", "pmc", "derive", "lock", "interp", "config", "loop", "fault"
-};
-
-static int numFlags = sizeof(debugFlags)/sizeof(debugFlags[0]);
-
 void
 reset_profile(void)
 {
-    if ((profile = (__pmProfile *)realloc(profile, sizeof(__pmProfile))) == NULL) {
-	__pmNoMem("reset_profile", sizeof(__pmProfile), PM_FATAL_ERR);
+    if ((profile = (pmProfile *)realloc(profile, sizeof(pmProfile))) == NULL) {
+	pmNoMem("reset_profile", sizeof(pmProfile), PM_FATAL_ERR);
 	exit(1);
     }
     ctxp->c_instprof = profile;
-    memset(profile, 0, sizeof(__pmProfile));
+    memset(profile, 0, sizeof(pmProfile));
     profile->state = PM_PROFILE_INCLUDE;        /* default global state */
     profile_changed = 1;
 }
@@ -66,6 +64,12 @@ setup_context(void)
     pthread_mutex_t	save_c_lock;
 #endif
 
+    /* 
+     * we don't want or need any PMDAs to be loaded for this one
+     * PM_CONTEXT_LOCAL context
+     */
+    pmSpecLocalPMDA("clear");
+
     if ((sts = pmNewContext(PM_CONTEXT_LOCAL, NULL)) < 0) {
 	fprintf(stderr, "setup_context: creation failed: %s\n", pmErrStr(sts));
 	exit(1);
@@ -77,14 +81,13 @@ setup_context(void)
 	exit(1);
     }
     /*
-     * Note: ctxp->c_lock remains locked throughout ... setup_context()
-     *       is only called once, and a single context is used throughout
-     *       to "fake" out the connection to the current PMDA ... so
-     *       there is no PM_UNLOCK(ctxp->c_lock) anywhere in the dbpmda
-     *       code.
-     *       This works because ctxp->c_lock is a recursive lock and
-     *       dbpmda is single-threaded.
+     * Note: This application is single threaded, and once we have ctxp
+     *	     the associated __pmContext will not move and will only be
+     *	     accessed or modified synchronously either here or in libpcp.
+     *	     We unlock the context so that it can be locked as required
+     *	     within libpcp.
      */
+    PM_UNLOCK(ctxp->c_lock);
 
 #ifdef PM_MULTI_THREAD
     /* need to be careful about the initialized lock */
@@ -123,15 +126,29 @@ strnum(int n)
 {
     char	*buf;
 
-    buf = (char *)malloc(13);
+#define MYBUFSZ 13
+
+    buf = (char *)malloc(MYBUFSZ);
     if (buf == NULL) {
 	fprintf(stderr, "strnum: malloc failed: %s\n", osstrerror());
 	exit(1);
     }
-    sprintf(buf, "%d", n);
+    pmsprintf(buf, MYBUFSZ, "%d", n);
     return buf;
 }
 
+char *
+strcluster(pmID pmid)
+{
+    static char buffer[32];
+    char *p;
+
+    pmIDStr_r(pmid, buffer, sizeof(buffer));
+    p = rindex(buffer, '.');
+    *p = '\0';
+    return buffer;
+}
+
 void
 initmetriclist(void)
 {
@@ -177,7 +194,7 @@ addarglist(char *arg)
 
     if (param.argc >= argc) {
         argc = param.argc;
-	argv = (char **)realloc(argv, argc * sizeof(pmProgname));
+	argv = (char **)realloc(argv, argc * sizeof(char *));
 	if (argv == NULL) {
 	    fprintf(stderr, "addarglist: realloc failed: %s\n", osstrerror());
 	    exit(1);
@@ -191,12 +208,14 @@ addarglist(char *arg)
     param.argv = argv;
 }
 
+#define MYCMDSZ 200
+
 void
 watch(char *fname)
 {
-    char	cmd[200];
+    char	cmd[MYCMDSZ];
 
-    sprintf(cmd, "xterm -hold -title \"dbpmda watch %s\" -geom 80x16 -bg dodgerblue4 -e tail -f %s &",
+    pmsprintf(cmd, MYCMDSZ, "xterm -hold -title \"dbpmda watch %s\" -geom 80x16 -bg dodgerblue4 -e tail -f %s &",
 	fname, fname);
     
     if (system(cmd) != 0)
@@ -204,7 +223,7 @@ watch(char *fname)
 }
 
 void
-printindom(FILE *f, __pmInResult *irp)
+printindom(FILE *f, pmInResult *irp)
 {
     int		i;
 
@@ -231,6 +250,7 @@ dohelp(int command, int full)
 	dohelp(FETCH, HELP_USAGE);
 	dohelp(GETDESC, HELP_USAGE);
 	dohelp(INSTANCE, HELP_USAGE);
+	dohelp(LABEL, HELP_USAGE);
 	dohelp(PMNS_NAME, HELP_USAGE);
 	dohelp(NAMESPACE, HELP_USAGE);
 	dohelp(OPEN, HELP_USAGE);
@@ -260,7 +280,7 @@ dohelp(int command, int full)
 	    break;
 	case DBG:
 	    puts("debug all | none");
-	    puts("debug flag [ flag ... ] (flag is decimal or symbolic name)");
+	    puts("debug option [ option ... ] (option is a symbolic name or a decimal number)");
 	    break;
 	case DESC:
 	    puts("desc metric");
@@ -278,13 +298,23 @@ dohelp(int command, int full)
 	case INSTANCE:
 	    puts("instance indom# [ number | name | \"name\" ]");
 	    break;
+	case LABEL:
+	    puts("label context");
+	    puts("label domain");
+	    puts("label indom indom#");
+	    puts("label cluster cluster#");
+	    puts("label item metric");
+	    puts("label instances indom#");
+	    break;
 	case NAMESPACE:
 	    puts("namespace fname");
 	    break;
 	case OPEN:
 	    puts("open dso dsoname init_routine [ domain# ]");
 	    puts("open pipe execname [ arg ... ]");
-	    puts("open socket sockname");
+	    puts("open socket unix sockname");
+	    puts("open socket inet port#|service");
+	    puts("open socket ipv6 port#|service");
 	    break;
 	case PMNS_CHILDREN:
 	    puts("children metric-name");
@@ -339,10 +369,11 @@ dohelp(int command, int full)
 		break;
 	    case DBG:
 		puts(
-"Specify which debugging flags should be active (see pmdbg(1)).  Flags may\n"
-"be specified as integers or by name, with multiple flags separated by\n"
-"white space.  All flags may be selected or deselected if 'all' or 'none' is\n"
-"specified.  The current setting is displayed by the status command.\n\n");
+"Specify which debugging options should be active (see pmdbg(1)).  Options\n"
+"may be specified by name (or number for the old bit-field options), with\n"
+"multiple options separated by white space.  All options may be selected or\n"
+"deselected if 'all' or 'none' is specified.  The current setting is\n"
+"displayed by the status command.\n\n");
 		break;
 	    case DESC:
 		puts(
@@ -378,12 +409,14 @@ dohelp(int command, int full)
 		break;
 	    case OPEN:
 		puts(
-"Open a PMDA as either a DSO, via UNIX domain socket (named pipe), or as a\n"
+"Open a PMDA as either a DSO, via a network socket (unix/inet/ipv6), or as a\n"
 "daemon (connected with a pipe).  The 'dsoname' and 'execname' fields are\n"
-"the path to the PMDA shared object file or executable.  The 'sockname'\n"
-"field is the path of a named pipe where a PMDA is listening for connections.\n"
+"the path to the PMDA shared object file or executable.  The first socket PMDA\n"
+"field is the type - either unix (if supported), inet or ipv6.  The 'sockname'\n"
+"argument for unix sockets is a path of a named pipe where a PMDA is listening\n"
+"for connections.  The 'port' argument is a port number, 'serv' a service name\n"
+"typically defined in /etc/services (resolved to a port via getservent(3)).\n"
 "The arguments to this command are similar to a line in the pmcd.conf file.\n");
-
 		break;
 	    case PMNS_CHILDREN:
 	        puts(
@@ -422,7 +455,7 @@ dohelp(int command, int full)
 	    case STATUS:
 		puts(
 "Display the state of dbpmda, including which PMDA is connected, which\n"
-"pmDebug flags are set, and the current profile.\n");
+"debug options are set, and the current profile.\n");
 		break;
 	    case STORE:
 		puts(
@@ -439,7 +472,7 @@ dohelp(int command, int full)
 		break;
 	    case WATCH:
 		puts(
-"A xwsh window is opened which tails the specified log file.  This window\n"
+"An xterm window is opened which tails the specified log file.  This window\n"
 "must be closed by the user when no longer required.\n");
 		break;
 	    case WAIT:
@@ -454,6 +487,7 @@ void
 dostatus(void)
 {
     int		i = 0;
+    int		n;
 
     putchar('\n');
     printf("Namespace:              ");
@@ -492,23 +526,29 @@ dostatus(void)
 	}
     }
 
-    printf("pmDebug:                ");
-    if (pmDebug == (unsigned int) -1)
-	printf("-1 (all)\n");
-    else if (pmDebug == 0)
-	printf("0 (none)\n");
-    else {
-	printf("%u", pmDebug);
-	for (i = 0; i < numFlags; i++)
-	    if (pmDebug & (1 << i)) {
-		printf(" ( %s", debugFlags[i]);
-		break;
-	    }
-	for (i++; i < numFlags; i++)
-	    if (pmDebug & (1 << i))
-		printf(" + %s", debugFlags[i]);
-	printf(" )\n");
+    printf("Debug options:          ");
+    n = 0;
+    for (i = 0; i < num_debug; i++) {
+	if (*(debug_map[i].options))
+	    n++;
     }
+    if (n == num_debug)
+	printf("(all)");
+    else if (n == 0)
+	printf("(none)");
+    else {
+	n = 0;
+	for (i = 0; i < num_debug; i++) {
+	    if (*(debug_map[i].options)) {
+		if (n == 0)
+		    n = 1;
+		else
+		    putchar(' ');
+		printf("%s", debug_map[i].name);
+	    }
+	}
+    }
+    putchar('\n');
 
     printf("Timer:                  ");
     if (timer == 0)
@@ -541,20 +581,23 @@ _dbDumpResult(FILE *f, pmResult *resp, pmDesc *desc_list)
     int		i;
     int		j;
     int		n;
-    char	*p;
+    char	**names;
 
-    fprintf(f,"pmResult dump from " PRINTF_P_PFX "%p timestamp: %d.%06d ",
+    fprintf(f, "pmResult dump from " PRINTF_P_PFX "%p timestamp: %d.%06d ",
         resp, (int)resp->timestamp.tv_sec, (int)resp->timestamp.tv_usec);
-    __pmPrintStamp(f, &resp->timestamp);
+    pmPrintStamp(f, &resp->timestamp);
     fprintf(f, " numpmid: %d\n", resp->numpmid);
     for (i = 0; i < resp->numpmid; i++) {
 	pmValueSet	*vsp = resp->vset[i];
-	n = pmNameID(vsp->pmid, &p);
+	names = NULL; /* silence coverity */
+	n = pmNameAll(vsp->pmid, &names);
 	if (n < 0)
-	    fprintf(f,"  %s (%s):", pmIDStr(vsp->pmid), "<noname>");
+	    fprintf(f, "  %s (%s):", pmIDStr(vsp->pmid), "<noname>");
 	else {
-	    fprintf(f,"  %s (%s):", pmIDStr(vsp->pmid), p);
-	    free(p);
+	    fprintf(f, "  %s (", pmIDStr(vsp->pmid));
+	    __pmPrintMetricNames(f, n, names, " or ");
+	    fprintf(f, "):");
+	    free(names);
 	}
 	if (vsp->numval == 0) {
 	    fprintf(f, " No values returned!\n");
@@ -569,7 +612,7 @@ _dbDumpResult(FILE *f, pmResult *resp, pmDesc *desc_list)
 	for (j = 0; j < vsp->numval; j++) {
 	    pmValue	*vp = &vsp->vlist[j];
 	    if (vsp->numval > 1 || desc_list[i].indom != PM_INDOM_NULL) {
-		fprintf(f,"    inst [%d", vp->inst);
+		fprintf(f, "    inst [%d", vp->inst);
 		fprintf(f, " or ???]");
 		fputc(' ', f);
 	    }
