@@ -18,7 +18,8 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
 
 #define MIN_ITIMER_USEC 100
 
@@ -36,6 +37,29 @@ static int		afid = 0x8000;
 static int		block;
 static void		onalarm(int);
 
+#ifdef PM_MULTI_THREAD
+static pthread_mutex_t	AF_lock;
+#else
+void			*AF_lock;
+#endif
+
+#if defined(PM_MULTI_THREAD) && defined(PM_MULTI_THREAD_DEBUG)
+/*
+ * return true if lock == AF_lock
+ */
+int
+__pmIsAFLock(void *lock)
+{
+    return lock == (void *)&AF_lock;
+}
+#endif
+
+void
+init_AF_lock(void)
+{
+    __pmInitMutex(&AF_lock);
+}
+
 /*
  * Platform dependent routines follow, Windows is very different
  * to POSIX platforms in terms of signals and timers.  Note - we
@@ -54,17 +78,22 @@ static HANDLE	afblock;	/* mutex protecting callback */
 static HANDLE	aftimer;	/* equivalent to ITIMER_REAL */
 static int	afsetup;	/* one-time-setup: done flag */
 
-static void AFsetup(void)
+static void AFinit(void)
 {
-    if (afsetup)
+    PM_INIT_LOCKS();
+    PM_LOCK(AF_lock);
+    if (afsetup) {
+	PM_UNLOCK(AF_lock);
 	return;
+    }
     afsetup = 1;
     afblock = CreateMutex(NULL, FALSE, NULL);
     aftimer = CreateWaitableTimer(NULL, TRUE, NULL);
+    PM_UNLOCK(AF_lock);
 }
 static void AFhold(void)
 { 
-    AFsetup();
+    AFinit();
     WaitForSingleObject(afblock, INFINITE);
 }
 static void AFrelse(void)
@@ -82,7 +111,7 @@ static void AFsetitimer(struct timeval *interval)
     LARGE_INTEGER duetime;
     long long inc;
 
-    AFsetup();
+    AFinit();
 
     inc = interval->tv_sec * 10000000ULL;	/* sec -> 100 nsecs */
     inc += (interval->tv_usec * 10ULL);		/* used -> 100 nsecs */
@@ -95,9 +124,10 @@ static void AFsetitimer(struct timeval *interval)
 #else /* POSIX */
 static void AFsetitimer(struct timeval *interval)
 {
-    static struct itimerval val;
+    struct itimerval val;
 
     val.it_value = *interval;
+    val.it_interval.tv_sec = val.it_interval.tv_usec = 0;
     setitimer(ITIMER_REAL, &val, NULL);
 }
 
@@ -151,63 +181,24 @@ static void AFrearm(void) { signal(SIGALRM, onalarm); }
  * Platform independent code follows
  */
 
-#ifdef PCP_DEBUG
 static void
-printstamp(struct timeval *tp)
+printdelta(FILE *f, struct timeval *tp)
 {
-    static struct tm    *tmp;
-    time_t		tt =  (time_t)tp->tv_sec;
+    struct tm	*tmp;
+    struct tm	gmtbuf;
+    time_t	tt =  (time_t)tp->tv_sec;
 
-    tmp = localtime(&tt);
+    tmp = gmtime_r(&tt, &gmtbuf);
     fprintf(stderr, "%02d:%02d:%02d.%06ld", tmp->tm_hour, tmp->tm_min, tmp->tm_sec, (long)tp->tv_usec);
 }
 
-static void
-printdelta_pcp(struct timeval *tp)
-{
-    static struct tm    *tmp;
-    time_t		tt =  (time_t)tp->tv_sec;
-
-    tmp = gmtime(&tt);
-    fprintf(stderr, "%02d:%02d:%02d.%06ld", tmp->tm_hour, tmp->tm_min, tmp->tm_sec, (long)tp->tv_usec);
-}
-#endif
-
-/*
- * a := a + b for struct timevals
- */
-static void
-tadd(struct timeval *a, struct timeval *b)
-{
-    a->tv_usec += b->tv_usec;
-    if (a->tv_usec > 1000000) {
-	a->tv_usec -= 1000000;
-	a->tv_sec++;
-    }
-    a->tv_sec += b->tv_sec;
-}
-
-/*
- * a := a - b for struct timevals
- */
-static void
-tsub_real(struct timeval *a, struct timeval *b)
-{
-    a->tv_usec -= b->tv_usec;
-    if (a->tv_usec < 0) {
-	a->tv_usec += 1000000;
-	a->tv_sec--;
-    }
-    a->tv_sec -= b->tv_sec;
-}
-
 /*
  * a := a - b for struct timevals, but result is never less than zero
  */
 static void
 tsub(struct timeval *a, struct timeval *b)
 {
-    tsub_real(a, b);
+    pmtimevalDec(a, b);
     if (a->tv_sec < 0) {
 	/* clip negative values at zero */
 	a->tv_sec = 0;
@@ -234,18 +225,16 @@ enqueue(qelt *qp)
     qelt		*tqp;
     qelt		*priorp;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_AF) {
+    if (pmDebugOptions.af) {
 	struct timeval	now;
 
-	__pmtimevalNow(&now);
-	printstamp(&now);
+	pmtimevalNow(&now);
+	pmPrintStamp(stderr, &now);
 	fprintf(stderr, " AFenqueue " PRINTF_P_PFX "%p(%d, " PRINTF_P_PFX "%p) for ",
 		qp->q_func, qp->q_afid, qp->q_data);
-	printstamp(&qp->q_when);
+	pmPrintStamp(stderr, &qp->q_when);
 	fputc('\n', stderr);
     }
-#endif
 
     for (tqp = root, priorp = NULL;
 	 tqp != NULL && tcmp(&qp->q_when, &tqp->q_when) >= 0;
@@ -262,6 +251,26 @@ enqueue(qelt *qp)
     }
 }
 
+/*
+ * must be async-signal-safe
+ *
+ * called routines (for POSIX variant of libpcp code)
+ *
+ * AFhold
+ *   sighold		- problem, should use sigaction()
+ * pmtimevalNow
+ *  gettimeofday	- problem
+ *  qp->q_func		- potential problem if application func() does
+ *  			  not restrict itself to async-signal-safe routines
+ *  free		- problem
+ * pmtimevalInc	- ok
+ * pmtimevalDec	- ok
+ *
+ * in debug code
+ *   fprintf	- problem, but we are not going to rewrite all of debug code,
+ *		  so accept that if PCP debugging is enabled this
+ *		  code is no longer thread-safe
+ */
 static void
 onalarm(int dummy)
 {
@@ -273,18 +282,16 @@ onalarm(int dummy)
     if (!block)
 	AFhold();
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_AF) {
-	__pmtimevalNow(&now);
-	printstamp(&now);
+    if (pmDebugOptions.af) {
+	pmtimevalNow(&now);
+	pmPrintStamp(stderr, &now);
 	fprintf(stderr, " AFonalarm(%d)\n", dummy);
     }
-#endif
     if (root != NULL) {
 	/* something to do ... */
 	while (root != NULL) {
 	    /* compute difference between scheduled time and now */
-	    __pmtimevalNow(&now);
+	    pmtimevalNow(&now);
 	    tmp = root->q_when;
 	    tsub(&tmp, &now);
 	    if (tmp.tv_sec == 0 && tmp.tv_usec <= 10000) {
@@ -295,13 +302,11 @@ onalarm(int dummy)
 
 		qp = root;
 		root = root->q_next;
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_AF) {
-		    printstamp(&now);
+		if (pmDebugOptions.af) {
+		    pmPrintStamp(stderr, &now);
 		    fprintf(stderr, " AFcallback " PRINTF_P_PFX "%p(%d, " PRINTF_P_PFX "%p)\n",
 			    qp->q_func, qp->q_afid, qp->q_data);
 		}
-#endif
 		qp->q_func(qp->q_afid, qp->q_data);
              
 		if (qp->q_delta.tv_sec == 0 && qp->q_delta.tv_usec == 0) {
@@ -327,22 +332,20 @@ onalarm(int dummy)
 		     *      		   |               +-- now
 		     *      		   +-- now - q_delta
 		     */
-		    __pmtimevalNow(&now);
+		    pmtimevalNow(&now);
 		    for ( ; ; ) {
-			tadd(&qp->q_when, &qp->q_delta);
+			pmtimevalInc(&qp->q_when, &qp->q_delta);
 			tmp = qp->q_when;
-			tsub_real(&tmp, &now);
-			tadd(&tmp, &qp->q_delta);
+			pmtimevalDec(&tmp, &now);
+			pmtimevalInc(&tmp, &qp->q_delta);
 			if (tmp.tv_sec >= 0)
 			    break;
-#ifdef PCP_DEBUG
-			if (pmDebug & DBG_TRACE_AF) {
-			    printstamp(&now);
+			if (pmDebugOptions.af) {
+			    pmPrintStamp(stderr, &now);
 			    fprintf(stderr, " AFcallback event %d too slow, skip callback for ", qp->q_afid);
-			    printstamp(&qp->q_when);
+			    pmPrintStamp(stderr, &qp->q_when);
 			    fputc('\n', stderr);
 			}
-#endif
 		    }
 		    enqueue(qp);
 		}
@@ -356,25 +359,26 @@ onalarm(int dummy)
 	}
 
 	if (root == NULL) {
-	    pmprintf("Warning: AF event queue is empty, nothing more will be scheduled\n");
-	    pmflush();
+	    if (pmDebugOptions.af) {
+		pmPrintStamp(stderr, &now);
+		fprintf(stderr, "Warning: AF event queue is empty, nothing more will be scheduled\n");
+	    }
+	    ;
 	}
 	else {
 	    /* set itimer for head of queue */
 	    interval = root->q_when;
-	    __pmtimevalNow(&now);
+	    pmtimevalNow(&now);
 	    tsub(&interval, &now);
 	    if (interval.tv_sec == 0 && interval.tv_usec < MIN_ITIMER_USEC)
 		/* use minimal delay (platform dependent) */
 		interval.tv_usec = MIN_ITIMER_USEC;
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_AF) {
-		printstamp(&now);
+	    if (pmDebugOptions.af) {
+		pmPrintStamp(stderr, &now);
 		fprintf(stderr, " AFsetitimer for delta ");
-		printdelta_pcp(&interval);
+		printdelta(stderr, &interval);
 		fputc('\n', stderr);
 	    }
-#endif
 	    AFsetitimer(&interval);
 	}
     }
@@ -385,45 +389,49 @@ onalarm(int dummy)
 }
 
 int
-__pmAFregister(const struct timeval *delta, void *data, void (*func)(int, void *))
+__pmAFsetup(const struct timeval *start, const struct timeval *delta, void *data, void (*func)(int, void *))
 {
     qelt		*qp;
     struct timeval	now;
     struct timeval	interval;
 
+    __pmInitLocks();
+
+    if (PM_MULTIPLE_THREADS(PM_SCOPE_AF))
+	return PM_ERR_THREAD;
+
     if (!block)
 	AFhold();
     if (afid == 0x8000 && !block)	/* first time */
 	AFrearm();
     if ((qp = (qelt *)malloc(sizeof(qelt))) == NULL) {
-	return -errno;
+	return -oserror();
     }
     qp->q_afid = ++afid;
     qp->q_data = data;
     qp->q_delta = *delta;
     qp->q_func = func;
-    __pmtimevalNow(&qp->q_when);
-    tadd(&qp->q_when, &qp->q_delta);
+    pmtimevalNow(&qp->q_when);
+    if (start != NULL)
+	pmtimevalInc(&qp->q_when, start);
 
     enqueue(qp);
     if (root == qp) {
 	/* we ended up at the head of the list, set itimer */
 	interval = qp->q_when;
-	__pmtimevalNow(&now);
+	pmtimevalNow(&now);
 	tsub(&interval, &now);
 
 	if (interval.tv_sec == 0 && interval.tv_usec < MIN_ITIMER_USEC)
 	    /* use minimal delay (platform dependent) */
 	    interval.tv_usec = MIN_ITIMER_USEC;
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_AF) {
-	    printstamp(&now);
+	if (pmDebugOptions.af) {
+	    pmPrintStamp(stderr, &now);
 	    fprintf(stderr, " AFsetitimer for delta ");
-	    printdelta_pcp(&interval);
+	    printdelta(stderr, &interval);
 	    fputc('\n', stderr);
 	}
-#endif
 	AFsetitimer(&interval);
     }
 
@@ -432,6 +440,12 @@ __pmAFregister(const struct timeval *delta, void *data, void (*func)(int, void *
     return qp->q_afid;
 }
 
+int
+__pmAFregister(const struct timeval *delta, void *data, void (*func)(int, void *))
+{
+    return __pmAFsetup(delta, delta, data, func);
+}
+
 int
 __pmAFunregister(int afid)
 {
@@ -440,6 +454,11 @@ __pmAFunregister(int afid)
     struct timeval	now;
     struct timeval	interval;
 
+    __pmInitLocks();
+
+    if (PM_MULTIPLE_THREADS(PM_SCOPE_AF))
+	return PM_ERR_THREAD;
+
     if (!block)
 	AFhold();
     for (qp = root, priorp = NULL; qp != NULL && qp->q_afid != afid; qp = qp->q_next)
@@ -459,19 +478,17 @@ __pmAFunregister(int afid)
 	     * new head of queue
 	     */
 	    interval = root->q_when;
-	    __pmtimevalNow(&now);
+	    pmtimevalNow(&now);
 	    tsub(&interval, &now);
-	    if (interval.tv_sec == 0 && interval.tv_usec == 0)
-		/* arbitrary 0.1 msec as minimal delay */
-		interval.tv_usec = 100;
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_AF) {
-		printstamp(&now);
+	    if (interval.tv_sec == 0 && interval.tv_usec < MIN_ITIMER_USEC)
+		/* use minimal delay (platform dependent) */
+		interval.tv_usec = MIN_ITIMER_USEC;
+	    if (pmDebugOptions.af) {
+		pmPrintStamp(stderr, &now);
 		fprintf(stderr, " AFsetitimer for delta ");
-		printdelta_pcp(&interval);
+		printdelta(stderr, &interval);
 		fputc('\n', stderr);
 	    }
-#endif
 	    AFsetitimer(&interval);
 	}
     }
@@ -488,6 +505,10 @@ __pmAFunregister(int afid)
 void
 __pmAFblock(void)
 {
+    __pmInitLocks();
+
+    if (PM_MULTIPLE_THREADS(PM_SCOPE_AF))
+	return;
     block = 1;
     AFhold();
 }
@@ -495,6 +516,10 @@ __pmAFblock(void)
 void
 __pmAFunblock(void)
 {
+    __pmInitLocks();
+
+    if (PM_MULTIPLE_THREADS(PM_SCOPE_AF))
+	return;
     block = 0;
     AFrearm();
     AFrelse();
