@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2013 Red Hat.
+ * Copyright (c) 2012-2018 Red Hat.
  * Copyright (c) 2008-2009 Aconex.  All Rights Reserved.
  * Copyright (c) 2000-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
@@ -16,12 +16,30 @@
 
 #include <ctype.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
 #include "pmda.h"
 #ifdef HAVE_STRINGS_H
 #include <strings.h>
 #endif
 
+#ifdef PM_MULTI_THREAD
+static pthread_mutex_t	config_lock = PTHREAD_MUTEX_INITIALIZER;
+#else
+void			*config_lock;
+#endif
+
+#if defined(PM_MULTI_THREAD) && defined(PM_MULTI_THREAD_DEBUG)
+/*
+ * return true if lock == config_lock
+ */
+int
+__pmIsConfigLock(void *lock)
+{
+    return lock == (void *)&config_lock;
+}
+#endif
+
 #ifdef IS_MINGW
 /*
  * Fix up the Windows path separator quirkiness - PCP code deals
@@ -98,49 +116,55 @@ static int posix_style(void)
 {
     char	*s;
     int		sts;
-    PM_LOCK(__pmLock_libpcp);
-    s = getenv("SHELL");
+
+    PM_LOCK(__pmLock_extcall);
+    s = getenv("SHELL");		/* THREADSAFE */
     sts = (s && strncmp(s, "/bin/", 5) == 0);
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(__pmLock_extcall);
     return sts;
 }
 
+/*
+ * Called with __pmLock_extcall held, so setenv() is thread-safe.
+ */
 static void
 dos_formatter(char *var, char *prefix, char *val)
 {
-    char envbuf[MAXPATHLEN];
     int msys = posix_style();
 
     if (prefix && dos_rewrite_path(var, val, msys)) {
 	char *p = msys ? msys_native_path(prefix) : prefix;
-	snprintf(envbuf, sizeof(envbuf), "%s=%s%s", var, p, val);
+	char envbuf[MAXPATHLEN];
+
+	pmsprintf(envbuf, sizeof(envbuf), "%s%s", p, val);
+	setenv(var, envbuf, 1);	/* THREADSAFE */
     }
     else {
-	snprintf(envbuf, sizeof(envbuf), "%s=%s", var, val);
+	setenv(var, val, 1);	/* THREADSAFE */
     }
-    PM_LOCK(__pmLock_libpcp);
-    putenv(strdup(envbuf));
-    PM_UNLOCK(__pmLock_libpcp);
 }
 
-INTERN const __pmConfigCallback __pmNativeConfig = dos_formatter;
+PCP_DATA const __pmConfigCallback __pmNativeConfig = dos_formatter;
 char *__pmNativePath(char *path) { return dos_native_path(path); }
-int __pmPathSeparator() { return posix_style() ? '/' : '\\'; }
+int pmPathSeparator() { return posix_style() ? '/' : '\\'; }
 int __pmAbsolutePath(char *path) { return posix_style() ? path[0] == '/' : dos_absolute_path(path); }
 #else
 char *__pmNativePath(char *path) { return path; }
 int __pmAbsolutePath(char *path) { return path[0] == '/'; }
-int __pmPathSeparator() { return '/'; }
+int pmPathSeparator() { return '/'; }
 
+/*
+ * Called with __pmLock_extcall held, so setenv() is thread-safe.
+ */
 static void
 posix_formatter(char *var, char *prefix, char *val)
 {
-    /* +40 bytes for max PCP env variable name */
-    char	envbuf[MAXPATHLEN+40];
+    char	envbuf[MAXPATHLEN];
     char	*vp;
     char	*vend;
+    unsigned	length;
 
-    snprintf(envbuf, sizeof(envbuf), "%s=", var);
+    (void)prefix;
     vend = &val[strlen(val)-1];
     if (val[0] == *vend && (val[0] == '\'' || val[0] == '"')) {
 	/*
@@ -149,23 +173,21 @@ posix_formatter(char *var, char *prefix, char *val)
 	 */
 	vp = &val[1];
 	vend--;
-    }
-    else
+    } else {
 	vp = val;
-    strncat(envbuf, vp, vend-vp+1);
-    envbuf[strlen(var)+1+vend-vp+1+1] = '\0';
-
-    PM_LOCK(__pmLock_libpcp);
-    putenv(strdup(envbuf));
-    PM_UNLOCK(__pmLock_libpcp);
-    (void)prefix;
+    }
+    if ((length = vend - vp + 1) > sizeof(envbuf))
+	length = sizeof(envbuf);
+    pmsprintf(envbuf, sizeof(envbuf), "%.*s", length, vp);
+    setenv(var, envbuf, 1);		/* THREADSAFE */
 }
 
-INTERN const __pmConfigCallback __pmNativeConfig = posix_formatter;
+PCP_DATA const __pmConfigCallback __pmNativeConfig = posix_formatter;
 #endif
 
-void
-__pmConfig(__pmConfigCallback formatter)
+
+static void
+__pmconfig(__pmConfigCallback formatter, int fatal)
 {
     /*
      * Scan ${PCP_CONF-$PCP_DIR/etc/pcp.conf} and put all PCP config
@@ -173,6 +195,7 @@ __pmConfig(__pmConfigCallback formatter)
      */
     FILE *fp;
     char confpath[32];
+    char errmsg[PM_MAXERRMSGLEN];
     char dir[MAXPATHLEN];
     char var[MAXPATHLEN];
     char *prefix;
@@ -180,28 +203,45 @@ __pmConfig(__pmConfigCallback formatter)
     char *val;
     char *p;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
-    prefix = getenv("PCP_DIR");
-    if ((conf = getenv("PCP_CONF")) == NULL) {
+    PM_LOCK(__pmLock_extcall);
+    prefix = getenv("PCP_DIR");		/* THREADSAFE */
+    conf = getenv("PCP_CONF");		/* THREADSAFE */
+    if (conf == NULL) {
 	strncpy(confpath, "/etc/pcp.conf", sizeof(confpath));
-	if (prefix == NULL)
+	if (prefix == NULL) {
+	    /* THREADSAFE - no locks acquired in __pmNativePath() */
 	    conf = __pmNativePath(confpath);
+	}
 	else {
-	    snprintf(dir, sizeof(dir),
+	    pmsprintf(dir, sizeof(dir),
 			 "%s%s", prefix, __pmNativePath(confpath));
 	    conf = dir;
 	}
     }
+    conf = strdup(conf);
+    if (prefix != NULL) prefix = strdup(prefix);
+    PM_UNLOCK(__pmLock_extcall);
 
-    if (access((const char *)conf, R_OK) < 0 ||
-	(fp = fopen(conf, "r")) == (FILE *)NULL) {
-	char	errmsg[PM_MAXERRMSGLEN];
-	pmprintf("FATAL PCP ERROR: could not open config file \"%s\" : %s\n",
+    if ((fp = fopen(conf, "r")) == NULL) {
+	if (!fatal) {
+	    free(conf);
+	    if (prefix != NULL) free(prefix);
+	    return;
+	}
+	/*
+	 * we used to pmprintf() here to be sure the message
+	 * would be seen, given the seriousness of the situation
+	 * ... but that introduces recursion back into
+	 * pmGetOptionalConfig() to get the PCP settings that
+	 * control what how to dispose of output from pmprintf()
+	 * ... and kaboom.
+	 */
+	fprintf(stderr,
+	    "FATAL PCP ERROR: could not open config file \"%s\" : %s\n"
+	    "You may need to set PCP_CONF or PCP_DIR in your environment.\n",
 		conf, osstrerror_r(errmsg, sizeof(errmsg)));
-	pmprintf("You may need to set PCP_CONF or PCP_DIR in your environment.\n");
-	pmflush();
-	PM_UNLOCK(__pmLock_libpcp);
+	free(conf);
+	if (prefix != NULL) free(prefix);
 	exit(1);
     }
 
@@ -212,55 +252,91 @@ __pmConfig(__pmConfigCallback formatter)
 	val = p+1;
 	if ((p = strrchr(val, '\n')) != NULL)
 	    *p = '\0';
-	if ((p = getenv(var)) != NULL)
+	PM_LOCK(__pmLock_extcall);
+	p = getenv(var);		/* THREADSAFE */
+	if (p != NULL)
 	    val = p;
-	else
+	else {
+	    /*
+	     * THREADSAFE - no locks acquired in formatter() which is
+	     * really dos_formatter() or posix_formatter()
+	     */
 	    formatter(var, prefix, val);
-
-	if (pmDebug & DBG_TRACE_CONFIG)
-	    fprintf(stderr, "pmGetConfig: (init) %s=%s\n", var, val);
+	}
+	if (pmDebugOptions.config)
+	    fprintf(stderr, "pmgetconfig: (init) %s=%s\n", var, val);
+	PM_UNLOCK(__pmLock_extcall);
     }
     fclose(fp);
-    PM_UNLOCK(__pmLock_libpcp);
+    free(conf);
+    if (prefix != NULL) free(prefix);
+}
+
+void
+__pmConfig(__pmConfigCallback formatter)
+{
+    __pmconfig(formatter, PM_FATAL_ERR);
+}
+
+static char *
+pmgetconfig(const char *name, int fatal)
+{
+    /*
+     * state controls one-trip initialization
+     */
+    static int		state = 0;
+    char		*val;
+
+    PM_LOCK(config_lock);
+    if (state == 0) {
+	state = 1;
+	__pmconfig(__pmNativeConfig, fatal);
+    }
+    PM_UNLOCK(config_lock);
+
+    /*
+     * THREADSAFE TODO ... this is bad (and documented), returning a
+     * direct pointer into the env ... should strdup() here and fix all
+     * callers to free() as needed later
+     */
+    val = getenv(name);		/* THREAD-UNSAFE! */
+    if (val == NULL) {
+	if (pmDebugOptions.config) {
+	    fprintf(stderr, "pmgetconfig: getenv(%s) -> NULL\n", name);
+	}
+	if (!fatal)
+	    return NULL;
+	val = "";
+    }
+
+    if (pmDebugOptions.config)
+	fprintf(stderr, "pmgetconfig: %s=%s\n", name, val);
+
+    return val;
 }
 
 char *
 pmGetConfig(const char *name)
 {
-    /*
-     * state controls one-trip initialization, and recursion guard
-     * for pathological failures in initialization
-     */
-    static int		state = 0;
-    char		*val;
+    return pmgetconfig(name, PM_FATAL_ERR);
+}
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
-    if (state == 0) {
-	state = 1;
-	PM_UNLOCK(__pmLock_libpcp);
-	__pmConfig(__pmNativeConfig);
-	PM_LOCK(__pmLock_libpcp);
-	state = 2;
+char *
+pmGetOptionalConfig(const char *name)
+{
+    return pmgetconfig(name, PM_RECOV_ERR);
+}
+
+int
+pmGetUsername(char **username)
+{
+    char *user = pmGetOptionalConfig("PCP_USER");
+    if (user && user[0] != '\0') {
+	*username = user;
+	return 1;
     }
-    else if (state == 1) {
-	/* recursion from error in __pmConfig() ... no value is possible */
-	PM_UNLOCK(__pmLock_libpcp);
-	if (pmDebug & DBG_TRACE_CONFIG)
-	    fprintf(stderr, "pmGetConfig: %s= ... recursion error\n", name);
-	val = "";
-	return val;
-    }
-
-    if ((val = getenv(name)) == NULL) {
-	val = "";
-    }
-
-    if (pmDebug & DBG_TRACE_CONFIG)
-	fprintf(stderr, "pmGetConfig: %s=%s\n", name, val);
-
-    PM_UNLOCK(__pmLock_libpcp);
-    return val;
+    *username = "pcp";
+    return 0;
 }
 
 /*
@@ -290,12 +366,22 @@ static const char *
 ipv6_enabled(void)
 {
 #if defined(IS_LINUX)
-    return access("/proc/net/if_inet6", F_OK) == 0 ? enabled() : disabled();
+    int c;
+    FILE *fp = fopen("/proc/sys/net/ipv6/conf/all/disable_ipv6", "r");
+    if (fp == NULL)
+	return access("/proc/net/if_inet6", F_OK) == 0 ? enabled() : disabled();
+    c = fgetc(fp);
+    fclose(fp);
+    if (c == '1')
+	return disabled();
+    return enabled();
 #else
     return enabled();
 #endif
 }
 
+extern const char *compress_suffix_list(void);
+
 #ifdef PM_MULTI_THREAD
 #define MULTI_THREAD_ENABLED	enabled
 #else
@@ -313,6 +399,41 @@ ipv6_enabled(void)
 #define SECURE_SOCKETS_ENABLED	disabled
 #define AUTHENTICATION_ENABLED	disabled
 #endif
+#if defined(HAVE_STRUCT_SOCKADDR_UN)
+#define UNIX_DOMAIN_SOCKETS_ENABLED	enabled
+#else
+#define UNIX_DOMAIN_SOCKETS_ENABLED	disabled
+#endif
+#if defined(HAVE_STATIC_PROBES)
+#define STATIC_PROBES_ENABLED	enabled
+#else
+#define STATIC_PROBES_ENABLED	disabled
+#endif
+#if defined(HAVE_SERVICE_DISCOVERY)
+#define SERVICE_DISCOVERY_ENABLED	enabled
+#else
+#define SERVICE_DISCOVERY_ENABLED	disabled
+#endif
+#if defined(BUILD_WITH_LOCK_ASSERTS)
+#define LOCK_ASSERTS_ENABLED	enabled
+#else
+#define LOCK_ASSERTS_ENABLED	disabled
+#endif
+#if defined(PM_MULTI_THREAD_DEBUG)
+#define LOCK_DEBUG_ENABLED	enabled
+#else
+#define LOCK_DEBUG_ENABLED	disabled
+#endif
+#if defined(HAVE_LZMA_DECOMPRESSION)
+#define LZMA_DECOMPRESS		enabled
+#else
+#define LZMA_DECOMPRESS		disabled
+#endif
+#if defined(HAVE_TRANSPARENT_DECOMPRESSION)
+#define TRANSPARENT_DECOMPRESS	enabled
+#else
+#define TRANSPARENT_DECOMPRESS	disabled
+#endif
 
 typedef const char *(*feature_detector)(void);
 static struct {
@@ -331,6 +452,15 @@ static struct {
 	{ "secure_sockets",	SECURE_SOCKETS_ENABLED },	/* from pcp-3.7.x */
 	{ "ipv6",		ipv6_enabled },
 	{ "authentication",	AUTHENTICATION_ENABLED },	/* from pcp-3.8.x */
+	{ "unix_domain_sockets",UNIX_DOMAIN_SOCKETS_ENABLED },	/* from pcp-3.8.2 */
+	{ "static_probes",	STATIC_PROBES_ENABLED },	/* from pcp-3.8.3 */
+	{ "service_discovery",	SERVICE_DISCOVERY_ENABLED },	/* from pcp-3.8.6 */
+	{ "multi_archive_contexts", enabled },			/* from pcp-3.11.1 */
+	{ "lock_asserts",	LOCK_ASSERTS_ENABLED },		/* from pcp-3.11.10 */
+	{ "lock_debug",		LOCK_DEBUG_ENABLED },		/* from pcp-3.11.10 */
+	{ "lzma_decompress",	LZMA_DECOMPRESS },		/* from pcp-4.0.0 */
+	{ "transparent_decompress", TRANSPARENT_DECOMPRESS },	/* from pcp-4.0.0 */
+	{ "compress_suffixes",	compress_suffix_list },		/* from pcp-4.0.1 */
 };
 
 void
@@ -340,7 +470,7 @@ __pmAPIConfig(__pmAPIConfigCallback formatter)
 
     for (i = 0; i < sizeof(features)/sizeof(features[0]); i++) {
 	const char *value = features[i].detector();
-	if (pmDebug & DBG_TRACE_CONFIG)
+	if (pmDebugOptions.config)
 	    fprintf(stderr, "__pmAPIConfig: %s=%s\n",
 		  features[i].feature, value);
 	formatter(features[i].feature, value);
@@ -348,7 +478,7 @@ __pmAPIConfig(__pmAPIConfigCallback formatter)
 }
 
 const char *
-__pmGetAPIConfig(const char *name)
+pmGetAPIConfig(const char *name)
 {
     int i;
 
@@ -357,3 +487,12 @@ __pmGetAPIConfig(const char *name)
 	    return features[i].detector();
     return NULL;
 }
+
+/*
+ * binary encoding of current PCP version
+ */
+int
+pmGetVersion(void)
+{
+    return PM_VERSION_CURRENT;
+}
