@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013 Red Hat.
+ * Copyright (c) 2013-2017 Red Hat.
  * Copyright (c) 2010 Ken McDonell.  All Rights Reserved.
  * Copyright (c) 1995-2002,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
@@ -17,7 +17,7 @@
  *
  * atexit_installed is protected by the __pmLock_libpcp mutex.
  *
- * __pmSpecLocalPMDA() uses buffer[], but this routine is only called
+ * pmSpecLocalPMDA() uses buffer[], but this routine is only called
  * from main() in single-threaded apps like pminfo, pmprobe, pmval
  * and pmevent ... so we can ignore any multi-threading issues,
  * especially as buffer[] is only used on an error handling code path.
@@ -36,14 +36,15 @@
  * The same arguments apply to EndLocalContext() and __pmConnectLocal().
  *
  * __pmLocalPMDA() is a mixed bag, sharing some of the justification from
- * __pmSpecLocalPMDA() and some from __pmConnectLocal().
+ * pmSpecLocalPMDA() and some from __pmConnectLocal().
  *
  * Because __pmConnectLocal() is not going to be used in a multi-threaded
  * environment, the call to the thread-unsafe dlerror() is OK.
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
 #include "pmda.h"
 #include <ctype.h>
 #include <sys/stat.h>
@@ -67,8 +68,10 @@ build_dsotab(void)
      *
      */
     char	configFileName[MAXPATHLEN];
+    char	pathbuf[MAXPATHLEN];
     FILE	*configFile;
     char	*config;
+    char	*pmdas;
     char	*p;
     char	*q;
     struct stat	sbuf;
@@ -81,12 +84,15 @@ build_dsotab(void)
     numdso = 0;
     dsotab = NULL;
 
-    strcpy(configFileName, pmGetConfig("PCP_PMCDCONF_PATH"));
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_CONTEXT) {
+    if ((pmdas = pmGetOptionalConfig("PCP_PMDAS_DIR")) == NULL)
+	return PM_ERR_GENERIC;
+    if ((config = pmGetOptionalConfig("PCP_PMCDCONF_PATH")) == NULL)
+	return PM_ERR_GENERIC;
+    strncpy(configFileName, config, sizeof(configFileName));
+    configFileName[sizeof(configFileName) - 1] = '\0';
+    if (pmDebugOptions.context) {
 	fprintf(stderr, "build_dsotab: parsing %s\n", configFileName);
     }
-#endif
     if (stat(configFileName, &sbuf) < 0) {
 	return -oserror();
     }
@@ -95,7 +101,7 @@ build_dsotab(void)
 	return -oserror();
     }
     if ((config = malloc(sbuf.st_size+1)) == NULL) {
-	__pmNoMem("build_dsotbl:", sbuf.st_size+1, PM_RECOV_ERR);
+	pmNoMem("build_dsotbl:", sbuf.st_size+1, PM_RECOV_ERR);
 	fclose(configFile);
 	return -oserror();
     }
@@ -119,6 +125,19 @@ build_dsotab(void)
 	     */
 	    goto eatline;
 	}
+	if (strncmp(p, "linux", 5) == 0) {
+	    /*
+	     * the Linux PMDA is an exception now too ... we run it as root
+	     * (daemon) but we still want to make the DSO available for any
+	     * local context users.  We add this explicitly below.
+	     */
+	    domain = 60;
+	    init = "linux_init";
+	    pmsprintf(pathbuf, sizeof(pathbuf), "%s/linux/pmda_linux.so", pmdas);
+	    name = pathbuf;
+	    peekc = *p;
+	    goto dsoload;
+	}
 	/* skip the PMDA's name */
 	while (*p != '\0' && *p != '\n' && !isspace((int)*p))
 	    p++;
@@ -149,11 +168,10 @@ build_dsotab(void)
 	    p++;
 	peekc = *p;
 	*p = '\0';
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_CONTEXT) {
+dsoload:
+	if (pmDebugOptions.context) {
 	    fprintf(stderr, "[%d] domain=%d, name=%s, init=%s\n", lineno, domain, name, init);
 	}
-#endif
 	/*
 	 * a little bit recursive if we got here via __pmLocalPMDA(),
 	 * but numdso has been set correctly, so this is OK
@@ -183,22 +201,32 @@ build_dsoattrs(pmdaInterface *dispatch, __pmHashCtl *attrs)
     char *namep;
     int sts = 0;
 
-    snprintf(name, sizeof(name), "%u", getuid());
+#ifdef HAVE_GETUID
+    pmsprintf(name, sizeof(name), "%u", getuid());
     name[sizeof(name)-1] = '\0';
     if ((namep = strdup(name)) != NULL)
 	__pmHashAdd(PCP_ATTR_USERID, namep, attrs);
+#endif
 
-    snprintf(name, sizeof(name), "%u", getgid());
+#ifdef HAVE_GETGID
+    pmsprintf(name, sizeof(name), "%u", getgid());
     name[sizeof(name)-1] = '\0';
     if ((namep = strdup(name)) != NULL)
 	__pmHashAdd(PCP_ATTR_GROUPID, namep, attrs);
+#endif
+
+    pmsprintf(name, sizeof(name), "%" FMT_PID, (pid_t)getpid());
+    name[sizeof(name)-1] = '\0';
+    if ((namep = strdup(name)) != NULL)
+	__pmHashAdd(PCP_ATTR_PROCESSID, namep, attrs);
 
     if (dispatch->version.six.attribute != NULL) {
 	for (node = __pmHashWalk(attrs, PM_HASH_WALK_START);
 	     node != NULL;
 	     node = __pmHashWalk(attrs, PM_HASH_WALK_NEXT)) {
 	    if ((sts = dispatch->version.six.attribute(
-				0, node->key, node->data,
+				dispatch->version.six.ext->e_context,
+				node->key, node->data,
 				node->data ? strlen(node->data)+1 : 0,
 				dispatch->version.six.ext)) < 0)
 		break;
@@ -251,12 +279,10 @@ EndLocalContext(void)
 	if (dp->domain != -1 &&
 	    dp->dispatch.comm.pmda_interface >= PMDA_INTERFACE_5 &&
 	    dp->dispatch.version.four.ext->e_endCallBack != NULL) {
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_CONTEXT) {
+	    if (pmDebugOptions.context) {
 		fprintf(stderr, "NotifyEndLocalContext: DSO PMDA %s (%d) notified of context %d close\n", 
 		    dp->name, dp->domain, ctx);
 	    }
-#endif
 	    (*(dp->dispatch.version.four.ext->e_endCallBack))(ctx);
 	}
     }
@@ -284,6 +310,7 @@ __pmConnectLocal(__pmHashCtl *attrs)
     int			i;
     __pmDSO		*dp;
     char		pathbuf[MAXPATHLEN];
+    char		*pmdas;
     const char		*path;
 #if defined(HAVE_DLOPEN)
     unsigned int	challenge;
@@ -293,6 +320,9 @@ __pmConnectLocal(__pmHashCtl *attrs)
 #endif
 #endif
 
+    if ((pmdas = pmGetOptionalConfig("PCP_PMDAS_DIR")) == NULL)
+	return PM_ERR_GENERIC;
+
     if (numdso == -1) {
 	int	sts;
 	sts = build_dsotab();
@@ -301,21 +331,24 @@ __pmConnectLocal(__pmHashCtl *attrs)
 
     for (i = 0; i < numdso; i++) {
 	dp = &dsotab[i];
-	if (dp->domain == -1 || dp->handle != NULL)
+	if (dp->domain == -1)
 	    continue;
+	if (dp->handle != NULL)
+	    goto connected;
+
 	/*
 	 * __pmLocalPMDA() means the path to the DSO may be something
 	 * other than relative to $PCP_PMDAS_DIR ... need to try both
 	 * options and also with and without DSO_SUFFIX (so, dll, etc)
 	 */
-	snprintf(pathbuf, sizeof(pathbuf), "%s%c%s",
-		 pmGetConfig("PCP_PMDAS_DIR"), __pmPathSeparator(), dp->name);
-	if ((path = __pmFindPMDA(pathbuf)) == NULL) {
-	    snprintf(pathbuf, sizeof(pathbuf), "%s%c%s.%s",
-		 pmGetConfig("PCP_PMDAS_DIR"), __pmPathSeparator(), dp->name, DSO_SUFFIX);
+	if ((path = __pmFindPMDA(dp->name)) == NULL) {
+	    pmsprintf(pathbuf, sizeof(pathbuf), "%s%c%s",
+			pmdas, pmPathSeparator(), dp->name);
 	    if ((path = __pmFindPMDA(pathbuf)) == NULL) {
-		if ((path = __pmFindPMDA(dp->name)) == NULL) {
-		    snprintf(pathbuf, sizeof(pathbuf), "%s.%s", dp->name, DSO_SUFFIX);
+		pmsprintf(pathbuf, sizeof(pathbuf), "%s%c%s.%s",
+			    pmdas, pmPathSeparator(), dp->name, DSO_SUFFIX);
+		if ((path = __pmFindPMDA(pathbuf)) == NULL) {
+		    pmsprintf(pathbuf, sizeof(pathbuf), "%s.%s", dp->name, DSO_SUFFIX);
 		    if ((path = __pmFindPMDA(pathbuf)) == NULL) {
 			pmprintf("__pmConnectLocal: Warning: cannot find DSO at \"%s\" or \"%s\"\n", 
 			     pathbuf, dp->name);
@@ -330,8 +363,18 @@ __pmConnectLocal(__pmHashCtl *attrs)
 #if defined(HAVE_DLOPEN)
 	dp->handle = dlopen(path, RTLD_NOW);
 	if (dp->handle == NULL) {
+	    char	*errmsg;
+	    /*
+	     * Threadsafe note:
+	     *	dlerror() may not be threadsafe according to POSIX,
+	     *	but the glibc version is threadsafe.  Locking here
+	     *	is just being a little overly careful.
+	     */
+	    PM_LOCK(__pmLock_extcall);
+	    errmsg = dlerror();		/* THREADSAFE */
+	    PM_UNLOCK(__pmLock_extcall);
 	    pmprintf("__pmConnectLocal: Warning: error attaching DSO "
-		     "\"%s\"\n%s\n\n", path, dlerror());
+		     "\"%s\"\n%s\n\n", path, errmsg);
 	    pmflush();
 	    dp->domain = -1;
 	}
@@ -378,7 +421,7 @@ __pmConnectLocal(__pmHashCtl *attrs)
 	 */
 	challenge = 0xff;
 	dp->dispatch.comm.pmda_interface = challenge;
-	dp->dispatch.comm.pmapi_version = ~PMAPI_VERSION;
+	dp->dispatch.comm.pmapi_version = (~PMAPI_VERSION) & 0xff;
 	dp->dispatch.comm.flags = 0;
 	dp->dispatch.status = 0;
 
@@ -395,11 +438,12 @@ __pmConnectLocal(__pmHashCtl *attrs)
 	    dp->domain = -1;
 	}
 	else {
+connected:
 	    if (dp->dispatch.comm.pmda_interface < PMDA_INTERFACE_2 ||
 		dp->dispatch.comm.pmda_interface > PMDA_INTERFACE_LATEST) {
 		pmprintf("__pmConnectLocal: Error: Unknown PMDA interface "
 			 "version %d in \"%s\" DSO\n", 
-			 dp->dispatch.comm.pmda_interface, path);
+			 dp->dispatch.comm.pmda_interface, dp->name);
 		pmflush();
 		dlclose(dp->handle);
 		dp->domain = -1;
@@ -407,15 +451,19 @@ __pmConnectLocal(__pmHashCtl *attrs)
 	    else if (dp->dispatch.comm.pmapi_version != PMAPI_VERSION_2) {
 		pmprintf("__pmConnectLocal: Error: Unknown PMAPI version %d "
 			 "in \"%s\" DSO\n",
-			 dp->dispatch.comm.pmapi_version, path);
+			 dp->dispatch.comm.pmapi_version, dp->name);
 		pmflush();
 		dlclose(dp->handle);
 		dp->domain = -1;
 	    }
-	    else if (dp->dispatch.comm.pmda_interface >= PMDA_INTERFACE_6 &&
+	    else {
+		if (dp->dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
+		    dp->dispatch.version.four.ext->e_context = pmWhichContext();
+		if (dp->dispatch.comm.pmda_interface >= PMDA_INTERFACE_6 &&
 		    (dp->dispatch.comm.flags & PDU_FLAG_AUTH) != 0) {
-		/* Agent wants to know about connection attributes */
-		build_dsoattrs(&dp->dispatch, attrs);
+		    /* Agent wants to know about connection attributes */
+		    build_dsoattrs(&dp->dispatch, attrs);
+		}
 	    }
 	}
 #ifdef HAVE_ATEXIT
@@ -441,8 +489,7 @@ __pmLocalPMDA(int op, int domain, const char *name, const char *init)
     int		sts = 0;
     int		i;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_CONTEXT) {
+    if (pmDebugOptions.context) {
 	fprintf(stderr, "__pmLocalPMDA(op=");
 	if (op == PM_LOCAL_ADD) fprintf(stderr, "ADD");
 	else if (op == PM_LOCAL_DEL) fprintf(stderr, "DEL");
@@ -450,7 +497,6 @@ __pmLocalPMDA(int op, int domain, const char *name, const char *init)
 	else fprintf(stderr, "%d ???", op);
 	fprintf(stderr, ", domain=%d, name=%s, init=%s)\n", domain, name, init);
     }
-#endif
 
     if (numdso == -1) {
 	if (op != PM_LOCAL_CLEAR)
@@ -461,7 +507,7 @@ __pmLocalPMDA(int op, int domain, const char *name, const char *init)
     switch (op) {
 	case PM_LOCAL_ADD:
 	    if ((dsotab = (__pmDSO *)realloc(dsotab, (numdso+1)*sizeof(__pmDSO))) == NULL) {
-		__pmNoMem("__pmLocalPMDA realloc", (numdso+1)*sizeof(__pmDSO), PM_FATAL_ERR);
+		pmNoMem("__pmLocalPMDA realloc", (numdso+1)*sizeof(__pmDSO), PM_FATAL_ERR);
 		/*NOTREACHED*/
 	    }
 	    dsotab[numdso].domain = domain;
@@ -472,7 +518,7 @@ __pmLocalPMDA(int op, int domain, const char *name, const char *init)
 	    else {
 		if ((dsotab[numdso].name = strdup(name)) == NULL) {
 		    sts = -oserror();
-		    __pmNoMem("__pmLocalPMDA name", strlen(name)+1, PM_RECOV_ERR);
+		    pmNoMem("__pmLocalPMDA name", strlen(name)+1, PM_RECOV_ERR);
 		    return sts;
 		}
 	    }
@@ -483,7 +529,7 @@ __pmLocalPMDA(int op, int domain, const char *name, const char *init)
 	    else {
 		if ((dsotab[numdso].init = strdup(init)) == NULL) {
 		    sts = -oserror();
-		    __pmNoMem("__pmLocalPMDA init", strlen(init)+1, PM_RECOV_ERR);
+		    pmNoMem("__pmLocalPMDA init", strlen(init)+1, PM_RECOV_ERR);
 		    return sts;
 		}
 	    }
@@ -523,8 +569,7 @@ __pmLocalPMDA(int op, int domain, const char *name, const char *init)
 	    break;
     }
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_CONTEXT) {
+    if (pmDebugOptions.context) {
 	if (sts != 0) {
 	    char	errmsg[PM_MAXERRMSGLEN];
 	    fprintf(stderr, "__pmLocalPMDA -> %s\n", pmErrStr_r(sts, errmsg, sizeof(errmsg)));
@@ -538,7 +583,6 @@ __pmLocalPMDA(int op, int domain, const char *name, const char *init)
 		&dsotab[i], i, dsotab[i].domain, dsotab[i].name, dsotab[i].init, dsotab[i].handle);
 	}
     }
-#endif
 
     return sts;
 }
@@ -554,7 +598,7 @@ __pmLocalPMDA(int op, int domain, const char *name, const char *init)
  *	- init (name of DSO's initialization routine)
  */
 char *
-__pmSpecLocalPMDA(const char *spec)
+pmSpecLocalPMDA(const char *spec)
 {
     int		op;
     int		domain = -1;
@@ -567,7 +611,7 @@ __pmSpecLocalPMDA(const char *spec)
 
     if ((arg = sbuf = strdup(spec)) == NULL) {
 	sts = -oserror();
-	__pmNoMem("__pmSpecLocalPMDA dup spec", strlen(spec)+1, PM_RECOV_ERR);
+	pmNoMem("pmSpecLocalPMDA dup spec", strlen(spec)+1, PM_RECOV_ERR);
 	return "strdup failed";
     }
     if (strncmp(arg, "add", 3) == 0) {
@@ -673,7 +717,7 @@ doit:
 	/* see thread-safe note at the head of this file */
 	static char buffer[256];
 	char	errmsg[PM_MAXERRMSGLEN];
-	snprintf(buffer, sizeof(buffer), "__pmLocalPMDA: %s", pmErrStr_r(sts, errmsg, sizeof(errmsg)));
+	pmsprintf(buffer, sizeof(buffer), "__pmLocalPMDA: %s", pmErrStr_r(sts, errmsg, sizeof(errmsg)));
 	free(sbuf);
 	return buffer;
     }
