@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013-2017 Red Hat.
  * Copyright (c) 1995-2001,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -13,8 +14,9 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "fault.h"
+#include "internal.h"
 #include <ctype.h>
 #ifdef HAVE_SECURE_SOCKETS
 #include <prerror.h>
@@ -22,12 +24,41 @@
 #include <sslerr.h>
 #include <sasl.h>
 #endif
-#ifdef IS_MINGW
-extern const char *strerror_r(int, char *, size_t);
+
+#ifdef PM_MULTI_THREAD
+#ifndef IS_MINGW
+static pthread_mutex_t	err_lock = PTHREAD_MUTEX_INITIALIZER;
+#endif
+#else
+void			*err_lock;
+#endif
+
+#if defined(PM_MULTI_THREAD) && defined(PM_MULTI_THREAD_DEBUG)
+/*
+ * return true if lock == err_lock
+ */
+int
+__pmIsErrLock(void *lock)
+{
+    return lock == (void *)&err_lock;
+}
 #endif
 
 /*
- * if you modify this table at all, be sure to remake qa/006
+ * Note:
+ *     These error codes must match the errors defined in pmapi.h
+ *     although the ordering here may be different to try and group
+ *     related errors together.
+ *
+ *     All of the errors in pmapi.h and here are also defined for
+ *     Perl applications in src/perl/PMDA/PMDA.pm and for Python
+ *     applications in src/python/pmapi.c.  To ease maintenance
+ *     effort we aim to keep the _order_ of the error codes the same
+ *     here and in the Perl and Python definitions.
+ *
+ *     And finally if you modify this table at all, be sure to check
+ *     for remakes in the QA suite, e.g. 006
+ * 
  */
 static const struct {
     int  	err;
@@ -131,7 +162,29 @@ static const struct {
     { PM_ERR_FAULT,		"PM_ERR_FAULT",
 	"QA fault injected" },
     { PM_ERR_THREAD,		"PM_ERR_THREAD",
-        "Operation not supported for multi-threaded applications" },
+	"Operation not supported for multi-threaded applications" },
+    { PM_ERR_NOCONTAINER,	"PM_ERR_NOCONTAINER",
+	"Container not found" },
+    { PM_ERR_BADSTORE,		"PM_ERR_BADSTORE",
+	"Bad input to pmstore" },
+    { PM_ERR_LOGOVERLAP,	"PM_ERR_LOGOVERLAP",
+	"Archives overlap in time" },
+    { PM_ERR_LOGHOST,		"PM_ERR_LOGHOST",
+	"Archives differ by host" },
+    { PM_ERR_LOGCHANGETYPE,	"PM_ERR_LOGCHANGETYPE",
+	"The type of a metric has changed in an archive" },
+    { PM_ERR_LOGCHANGESEM,	"PM_ERR_LOGCHANGESEM",
+	"The semantics of a metric has changed in an archive" },
+    { PM_ERR_LOGCHANGEINDOM,	"PM_ERR_LOGCHANGEINDOM",
+	"The instance domain of a metric has changed in an archive" },
+    { PM_ERR_LOGCHANGEUNITS,	"PM_ERR_LOGCHANGEUNITS",
+	"The units of a metric have changed in an archive" },
+    { PM_ERR_NEEDCLIENTCERT,	"PM_ERR_NEEDCLIENTCERT",
+	"PMCD requires a client certificate" },
+    { PM_ERR_BADDERIVE,		"PM_ERR_BADDERIVE",
+	"Derived metric definition failed" },
+    { PM_ERR_NOLABELS,		"PM_ERR_NOLABELS",
+	"No support for metric label metadata" },
     /* insert new libpcp error codes here */
     { PM_ERR_NYI,		"PM_ERR_NYI",
 	"Functionality not yet implemented" },
@@ -152,8 +205,10 @@ strerror_x(int code, char *buf, int buflen)
 #ifdef HAVE_STRERROR_R_PTR
     char	*p;
     p = strerror_r(code, buf, buflen);
-    if (p != buf)
+    if (p != buf) {
 	strncpy(buf, p, buflen);
+	buf[buflen-1] = '\0';
+    }
 #else
     /*
      * the more normal POSIX and XSI compliant variants always fill buf[]
@@ -176,6 +231,7 @@ pmErrStr_r(int code, char *buf, int buflen)
 
     if (code == 0) {
 	strncpy(buf, "No error", buflen);
+	buf[buflen-1] = '\0';
 	return buf;
     }
 
@@ -190,8 +246,11 @@ pmErrStr_r(int code, char *buf, int buflen)
 #define DECODE_SASL_SPECIFIC_ERROR(c)	((c) < -1000 ? 0 : (c))
 
 	int error = DECODE_SECURE_SOCKETS_ERROR(code);
-	if (DECODE_SASL_SPECIFIC_ERROR(error))
-	    strncpy(buf, sasl_errstring(error, NULL, NULL), buflen);
+	if (DECODE_SASL_SPECIFIC_ERROR(error)) {
+	    PM_LOCK(__pmLock_extcall);
+	    pmsprintf(buf, buflen, "Authentication - %s", sasl_errstring(error, NULL, NULL));	/* THREADSAFE */
+	    PM_UNLOCK(__pmLock_extcall);
+	}
 	else
 	    strncpy(buf, PR_ErrorToString(error, PR_LANGUAGE_EN), buflen);
 	buf[buflen-1] = '\0';
@@ -200,6 +259,7 @@ pmErrStr_r(int code, char *buf, int buflen)
     }
 
 #ifndef IS_MINGW
+    PM_LOCK(err_lock);
     if (first) {
 	/*
 	 * reference message for an unrecognized error code.
@@ -234,6 +294,7 @@ PM_FAULT_POINT("libpcp/" __FILE__ ":1", PM_FAULT_ALLOC);
 	}
 	first = 0;
     }
+    PM_UNLOCK(err_lock);
     if (code < 0 && code > -PM_ERR_BASE) {
 	/* intro(2) / errno(3) errors, maybe */
 	strerror_x(-code, buf, buflen);
@@ -254,15 +315,18 @@ PM_FAULT_POINT("libpcp/" __FILE__ ":1", PM_FAULT_ALLOC);
 #else	/* WIN32 */
     if (code > -PM_ERR_BASE || code < -PM_ERR_NYI) {
 	const char	*bp;
-	if ((bp = wsastrerror(-code)) != NULL)
+	if ((bp = wsastrerror(-code)) != NULL) {
 	    strncpy(buf, bp, buflen);
+	    buf[buflen-1] = '\0';
+	}
 	else {
 	    /* No strerror_r in MinGW, so need to lock */
 	    char	*tbp;
-	    PM_LOCK(__pmLock_libpcp);
-	    tbp = strerror(-code);
+	    PM_LOCK(__pmLock_extcall);
+	    tbp = strerror(-code);		/* THREADSAFE */
 	    strncpy(buf, tbp, buflen);
-	    PM_UNLOCK(__pmLock_libpcp);
+	    buf[buflen-1] = '\0';
+	    PM_UNLOCK(__pmLock_extcall);
 	}
 
 	if (strncmp(buf, unknown, strlen(unknown)) != 0)
@@ -273,12 +337,13 @@ PM_FAULT_POINT("libpcp/" __FILE__ ":1", PM_FAULT_ALLOC);
     for (i = 0; errtab[i].err; i++) {
 	if (errtab[i].err == code) {
 	    strncpy(buf, errtab[i].errmess, buflen);
+	    buf[buflen-1] = '\0';
 	    return buf;
 	}
     }
 
     /* failure */
-    snprintf(buf, buflen, BADCODE,  code);
+    pmsprintf(buf, buflen, BADCODE,  code);
     return buf;
 }
 
@@ -295,8 +360,8 @@ __pmDumpErrTab(FILE *f)
 {
     int	i;
 
-    fprintf(f, "  Code  Symbolic Name        Message\n");
+    fprintf(f, "  Code  Symbolic Name          Message\n");
     for (i = 0; errtab[i].err; i++)
-	fprintf(f, "%6d  %-20s %s\n",
+	fprintf(f, "%6d  %-22s %s\n",
 	    errtab[i].err, errtab[i].symb, errtab[i].errmess);
 }
