@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2014-2015 Red Hat.
  * Copyright (c) 1995-2003 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -13,7 +14,7 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include <ctype.h>
 
 static __pmLogPort *logport;
@@ -71,7 +72,7 @@ is_portfile(const_dirent *dep)
 }
 
 /* The following function is used for selecting particular port files rather
- * than all valid files.  snprintf the pid of the pmlogger process or the
+ * than all valid files.  pmsprintf the pid of the pmlogger process or the
  * special constant PM_LOG_PRIMARY_LINK into the match array first.
  */
 #define PROCFS_ENTRY_SIZE 40	/* encompass any size of entry for pid */
@@ -110,8 +111,9 @@ __pmLogFindLocalPorts(int pid, __pmLogPort **result)
     if (result == NULL)
 	return -EINVAL;
 
-    lendir = snprintf(dir, sizeof(dir), "%s%cpmlogger",
-		pmGetConfig("PCP_TMP_DIR"), __pmPathSeparator());
+    if ((p = pmGetOptionalConfig("PCP_TMP_DIR")) == NULL)
+	return PM_ERR_GENERIC;
+    lendir = pmsprintf(dir, sizeof(dir), "%s%cpmlogger", p, pmPathSeparator());
 
     /* Set up the appropriate function to select files from the control port
      * directory.  Anticipate that this will usually be an exact match for
@@ -129,20 +131,22 @@ __pmLogFindLocalPorts(int pid, __pmLogPort **result)
 
 	default:			/* a specific pid (single) */
 	    if (!__pmProcessExists((pid_t)pid)) {
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_LOG) {
+		if (pmDebugOptions.log) {
 		    fprintf(stderr, "__pmLogFindLocalPorts() -> 0, "
 				"pid(%d) doesn't exist\n", pid);
 		}
-#endif
 		*result = NULL;
 		return 0;
 	    }
-	    snprintf(match, sizeof(match), "%d", pid);
+	    pmsprintf(match, sizeof(match), "%d", pid);
 	    break;
     }
 
     nf = scandir(dir, &files, scanfn, alphasort);
+    if (nf < 1 && pmDebugOptions.log) {
+	fprintf(stderr, "__pmLogFindLocalPorts: scandir() -> %d %s\n",
+		    nf, pmErrStr(oserror()));
+    }
     if (nf == -1 && oserror() == ENOENT)
 	nf = 0;
     else if (nf == -1) {
@@ -158,12 +162,10 @@ __pmLogFindLocalPorts(int pid, __pmLogPort **result)
 	return -oserror();
     }
     if (nf == 0) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LOG) {
+	if (pmDebugOptions.log) {
 	    fprintf(stderr, "__pmLogFindLocalPorts() -> 0, "
 			"num files = 0\n");
 	}
-#endif
 	*result = NULL;
 	free(files);
 	return 0;
@@ -182,7 +184,7 @@ __pmLogFindLocalPorts(int pid, __pmLogPort **result)
      */
     strcpy(namebuf, dir);
     p = namebuf + lendir;
-    *p++ = __pmPathSeparator();
+    *p++ = pmPathSeparator();
 
     /* open the file, try to read the port number and add the port to the
      * logport array if successful.
@@ -304,50 +306,44 @@ __pmIsLocalhost(const char *hostname)
 {
     int sts = 0;
 
-    if (strcasecmp(hostname, "localhost") == 0)
+    if (strcasecmp(hostname, "localhost") == 0 ||
+	strncmp(hostname, "local:", 6) == 0 ||
+	strncmp(hostname, "unix:", 5) == 0)
 	return 1;
     else {
 	char lhost[MAXHOSTNAMELEN+1];
-	struct hostent *he;
+	__pmHostEnt *servInfo1;
 
 	if (gethostname(lhost, MAXHOSTNAMELEN) < 0)
 	   return -oserror();
 
-	PM_INIT_LOCKS();
-	PM_LOCK(__pmLock_libpcp);
-        if ((he = gethostbyname(lhost)) != NULL ) {
-	    int i;
-	    unsigned int * laddrs;
-	    for (i=0; he->h_addr_list[i] != NULL; i++) ;
+	if ((servInfo1 = __pmGetAddrInfo(lhost)) != NULL) {
+	    __pmHostEnt		*servInfo2;
+	    __pmSockAddr	*addr1, *addr2;
+	    void		*enumIx1, *enumIx2;
 
-	    laddrs = (unsigned int *)calloc(i, sizeof (unsigned int));
-	    if (laddrs != NULL) {
-		int k;
-		for (k=0; k < i; k++) {
-		    laddrs[k] = ((struct in_addr *)he->h_addr_list[k])->s_addr;
-		}
-
-		if ((he = gethostbyname(hostname)) == NULL) {
-		    free(laddrs);
-		    PM_UNLOCK(__pmLock_libpcp);
-		    return -EHOSTUNREACH;
-		}
-
-		for (i--; i >= 0; i--) {
-		    for (k = 0; he->h_addr_list[k] != NULL; k++) {
-			struct in_addr *s=(struct in_addr *)he->h_addr_list[k];
-			if (s->s_addr == laddrs[i]) {
-			    free(laddrs);
-			    PM_UNLOCK(__pmLock_libpcp);
-			    return 1;
-			}
+	    if ((servInfo2 = __pmGetAddrInfo(hostname)) == NULL) {
+		__pmHostEntFree(servInfo1);
+		return -EHOSTUNREACH;
+	    }
+	    enumIx1 = NULL;
+	    for (addr1 = __pmHostEntGetSockAddr(servInfo1, &enumIx1);
+		 addr1 != NULL;
+		 addr1 = __pmHostEntGetSockAddr(servInfo1, &enumIx1)) {
+		enumIx2 = NULL;
+		for (addr2 = __pmHostEntGetSockAddr(servInfo2, &enumIx2);
+		     addr2 != NULL;
+		     addr2 = __pmHostEntGetSockAddr(servInfo2, &enumIx2)) {
+		    if (__pmSockAddrCompare(addr1, addr2) == 0) {
+			__pmHostEntFree(servInfo1);
+			__pmHostEntFree(servInfo2);
+			return 1;
 		    }
 		}
-
-		free(laddrs);
 	    }
+	    __pmHostEntFree(servInfo1);
+	    __pmHostEntFree(servInfo2);
 	}
-	PM_UNLOCK(__pmLock_libpcp);
     }
 
     return sts;
@@ -363,10 +359,12 @@ int
 __pmLogFindPort(const char *host, int pid, __pmLogPort **lpp)
 {
     int			ctx, oldctx;
+    char		*ctxhost;
     int			sts, numval;
     int			i, j;
     int			findone = pid != PM_LOG_ALL_PIDS;
     int			localcon = 0;	/* > 0 for local connection */
+    int			bytes;
     pmDesc		desc;
     pmResult		*res;
     char		*namelist[] = {"pmcd.pmlogger.port"};
@@ -384,16 +382,30 @@ __pmLogFindPort(const char *host, int pid, __pmLogPort **lpp)
 	return localcon;
 
     /* note: there may not be a current context */
+    ctx = 0;
     oldctx = pmWhichContext();
 
-    if ((ctx = pmNewContext(PM_CONTEXT_HOST, host)) < 0)
+    /*
+     * Enclose ctxhost in [] in case it is an ipv6 address. This prevents
+     * the first colon from being taken as a port separator by pmNewContext
+     * and does no harm otherwise.
+     */
+    bytes = strlen(host) + 2 + 1;
+    if ((ctxhost = malloc(bytes)) == NULL) {
+	sts = -ENOMEM;
+	goto ctxErr;
+    }
+    pmsprintf(ctxhost, bytes, "[%s]", host);
+    ctx = pmNewContext(PM_CONTEXT_HOST, ctxhost);
+    free(ctxhost);
+    if (ctx < 0)
 	return ctx;
     if ((sts = pmLookupName(1, namelist, &pmid)) < 0)
 	goto ctxErr;
 
     if ((sts = pmLookupDesc(pmid, &desc)) < 0)
 	goto ctxErr;
-    if ((sts = pmFetch(1, &pmid, &res) < 0))
+    if ((sts = pmFetch(1, &pmid, &res)) < 0)
 	goto ctxErr;
     if ((sts = numval = res->vset[0]->numval) < 0)
 	goto resErr;
@@ -432,6 +444,7 @@ resErr:
 ctxErr:
     if (oldctx >= 0)
 	pmUseContext(oldctx);
-    pmDestroyContext(ctx);
+    if (ctx >= 0)
+	pmDestroyContext(ctx);
     return sts;
 }
