@@ -10,25 +10,20 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  * License for more details.
- * 
- * You should have received a copy of the GNU Lesser General Public License
- * along with this library; if not, write to the Free Software Foundation,
- * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
- */
-
-/*
+ *
  * Generic routines to provide the "optimized" pmFetch bundling
  * services ... the optimization is driven by the crude heuristics
  * weights defined in optcost below.
+ *
+ * Thread-safe notes
+ *
+ * lrand48() is not thread-safe, but we don't really care here.
  */
 
-/* if DESPERATE, we need DEBUG */
-#if defined(DESPERATE) && !defined(PCP_DEBUG)
-#define DEBUG
-#endif
-
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
+#include <assert.h>
 
 /*
  * elements of optcost are
@@ -46,6 +41,23 @@
 /* default costs */
 static optcost_t	optcost = { 4, 1, 15, 10, 2, 0 };
 
+#ifdef PM_MULTI_THREAD
+static pthread_mutex_t	optfetch_lock = PTHREAD_MUTEX_INITIALIZER;
+#else
+void			*optfetch_lock;
+#endif
+
+#if defined(PM_MULTI_THREAD) && defined(PM_MULTI_THREAD_DEBUG)
+/*
+ * return true if lock == util_lock
+ */
+int
+__pmIsOptfetchLock(void *lock)
+{
+    return lock == (void *)&optfetch_lock;
+}
+#endif
+
 static int
 addpmid(fetchctl_t *fp, pmID pmid)
 {
@@ -61,7 +73,8 @@ addpmid(fetchctl_t *fp, pmID pmid)
     fp->f_numpmid++;
     fp->f_pmidlist = (pmID *)realloc(fp->f_pmidlist, fp->f_numpmid*sizeof(pmID));
     if (fp->f_pmidlist == NULL) {
-	__pmNoMem("addpmid", fp->f_numpmid*sizeof(pmID), PM_FATAL_ERR);
+	PM_UNLOCK(optfetch_lock);
+	pmNoMem("addpmid", fp->f_numpmid*sizeof(pmID), PM_FATAL_ERR);
     }
 
     for (j = fp->f_numpmid-1; j > i; j--)
@@ -97,7 +110,8 @@ addinst(int *numinst, int **instlist, optreq_t *new)
 
     ilist = (int *)realloc(*instlist, (numi + new->r_numinst)*sizeof(int));
     if (ilist == NULL) {
-	__pmNoMem("addinst.up", (numi + new->r_numinst)*sizeof(int), PM_FATAL_ERR);
+	PM_UNLOCK(optfetch_lock);
+	pmNoMem("addinst.up", (numi + new->r_numinst)*sizeof(int), PM_FATAL_ERR);
     }
 
     for (j = 0; j < new->r_numinst; j++) {
@@ -120,7 +134,8 @@ addinst(int *numinst, int **instlist, optreq_t *new)
 
     ilist = (int *)realloc(ilist, numi*sizeof(int));
     if (ilist == NULL) {
-	__pmNoMem("addinst.down", numi*sizeof(int), PM_FATAL_ERR);
+	PM_UNLOCK(optfetch_lock);
+	pmNoMem("addinst.down", numi*sizeof(int), PM_FATAL_ERR);
     }
     
     *numinst = numi;
@@ -140,6 +155,8 @@ missinst(int numa, int *lista, int numb, int *listb)
     int		i;
     int		j;
 
+    PM_ASSERT_IS_LOCKED(optfetch_lock);
+
     /* count in lista[] but _not_ in listb[] */
     if (numa == 0) {
 	/* special case for all instances in lista[] */
@@ -216,6 +233,8 @@ optCost(fetchctl_t *fp)
     int			cost = 0;
     int			done;
 
+    PM_ASSERT_IS_LOCKED(optfetch_lock);
+
     /*
      * cost per PMD for the pmids in this fetch
      */
@@ -270,11 +289,9 @@ optCost(fetchctl_t *fp)
     return cost;
 }
 
-#ifdef PCP_DEBUG
 static char *
-statestr(int state)
+statestr(int state, char *sbuf)
 {
-    static char		sbuf[100];
     sbuf[0] = '\0';
     if (state & OPT_STATE_NEW) strcat(sbuf, "NEW ");
     if (state & OPT_STATE_PMID) strcat(sbuf, "PMID ");
@@ -291,6 +308,8 @@ statestr(int state)
 static void
 dumplist(FILE *f, int style, char *tag, int numi, int *ilist)
 {
+    char	strbuf[20];
+
     fprintf(f, "%s: [%d]", tag, numi);
     if (ilist == NULL)
 	fprintf(f, " (nil)\n");
@@ -298,7 +317,7 @@ dumplist(FILE *f, int style, char *tag, int numi, int *ilist)
 	int		i;
 	for (i = 0; i < numi; i++) {
 	    if (style == 1)
-		fprintf(f, " %s", pmIDStr((pmID)ilist[i]));
+		fprintf(f, " %s", pmIDStr_r((pmID)ilist[i], strbuf, sizeof(strbuf)));
 	    else
 		fprintf(f, " %d", ilist[i]);
 	}
@@ -312,17 +331,18 @@ ___pmOptFetchDump(FILE *f, const fetchctl_t *fp)
     indomctl_t		*idp;
     pmidctl_t		*pmp;
     optreq_t		*rqp;
+    char		strbuf[100];
 
     fflush(stderr);
     fflush(stdout);
     fprintf(f, "Dump optfetch structures from " PRINTF_P_PFX "%p next=" PRINTF_P_PFX "%p\n", fp, fp->f_next);
-    fprintf(f, "Fetch Control @ " PRINTF_P_PFX "%p: cost=%d state=%s\n", fp, fp->f_cost, statestr(fp->f_state));
+    fprintf(f, "Fetch Control @ " PRINTF_P_PFX "%p: cost=%d state=%s\n", fp, fp->f_cost, statestr(fp->f_state, strbuf));
     dumplist(f, 1, "PMIDs", fp->f_numpmid, (int *)fp->f_pmidlist);
     for (idp = fp->f_idp; idp != NULL; idp = idp->i_next) {
-	fprintf(f, "  InDom %s Control @ " PRINTF_P_PFX "%p:\n", pmInDomStr(idp->i_indom), idp);
+	fprintf(f, "  InDom %s Control @ " PRINTF_P_PFX "%p:\n", pmInDomStr_r(idp->i_indom, strbuf, sizeof(strbuf)), idp);
 	dumplist(f, 0, "  instances", idp->i_numinst, idp->i_instlist);
 	for (pmp = idp->i_pmp; pmp != NULL; pmp = pmp->p_next) {
-	    fprintf(f, "    PMID %s Control @ " PRINTF_P_PFX "%p:\n", pmIDStr(pmp->p_pmid), pmp);
+	    fprintf(f, "    PMID %s Control @ " PRINTF_P_PFX "%p:\n", pmIDStr_r(pmp->p_pmid, strbuf, sizeof(strbuf)), pmp);
 	    dumplist(f, 0, "    instances", pmp->p_numinst, pmp->p_instlist);
 	    for (rqp = pmp->p_rqp; rqp != NULL; rqp = rqp->r_next) {
 		fprintf(f, "      Request @ " PRINTF_P_PFX "%p:\n", rqp);
@@ -342,26 +362,28 @@ __pmOptFetchDump(FILE *f, const fetchctl_t *root)
     for (fp = root; fp != NULL; fp = fp->f_next)
 	___pmOptFetchDump(f, fp);
 }
-#endif /* DEBUG */
 
 /*
- * add a new request into a group of fetches
+ * add a new request into a group of fetches ...
+ * only failure is from calloc() and this is fatal
  */
-int
+void
 __pmOptFetchAdd(fetchctl_t **root, optreq_t *new)
 {
     fetchctl_t		*fp;
     fetchctl_t		*tfp;
     indomctl_t		*idp;
-    pmidctl_t		*pmp;
+    pmidctl_t		*pmp = NULL;
     int			mincost;
     int			change;
     pmInDom		indom = new->r_desc->indom;
     pmID		pmid = new->r_desc->pmid;
 
+    PM_LOCK(optfetch_lock);
     /* add new fetch as first option ... will be reclaimed later if not used */
     if ((fp = (fetchctl_t *)calloc(1, sizeof(fetchctl_t))) == NULL) {
-	__pmNoMem("optAddFetch.fetch", sizeof(fetchctl_t), PM_FATAL_ERR);
+	PM_UNLOCK(optfetch_lock);
+	pmNoMem("optAddFetch.fetch", sizeof(fetchctl_t), PM_FATAL_ERR);
     }
     fp->f_next = *root;
     *root = fp;
@@ -388,7 +410,8 @@ __pmOptFetchAdd(fetchctl_t **root, optreq_t *new)
 
 	if (change & OPT_STATE_XINDOM) {
 	    if ((idp = (indomctl_t *)calloc(1, sizeof(indomctl_t))) == NULL) {
-		__pmNoMem("optAddFetch.indomctl", sizeof(indomctl_t), PM_FATAL_ERR);
+		PM_UNLOCK(optfetch_lock);
+		pmNoMem("optAddFetch.indomctl", sizeof(indomctl_t), PM_FATAL_ERR);
 	    }
 	    idp->i_indom = indom;
 	    idp->i_next = fp->f_idp;
@@ -397,7 +420,8 @@ __pmOptFetchAdd(fetchctl_t **root, optreq_t *new)
 	}
 	if (change & OPT_STATE_XPMID) {
 	    if ((pmp = (pmidctl_t *)calloc(1, sizeof(pmidctl_t))) == NULL) {
-		__pmNoMem("optAddFetch.pmidctl", sizeof(pmidctl_t), PM_FATAL_ERR);
+		PM_UNLOCK(optfetch_lock);
+		pmNoMem("optAddFetch.pmidctl", sizeof(pmidctl_t), PM_FATAL_ERR);
 	    }
 	    pmp->p_next = idp->i_pmp;
 	    idp->i_pmp = pmp;
@@ -411,9 +435,9 @@ __pmOptFetchAdd(fetchctl_t **root, optreq_t *new)
 	fp->f_newcost = optCost(fp);
 	if (fp == *root)
 	    fp->f_newcost += optcost.c_fetch;
-#ifdef DESPERATE
-	if (pmDebug & DBG_TRACE_OPTFETCH) {
-	    fprintf(stderr, "optFetch: cost=");
+	if (pmDebugOptions.optfetch && pmDebugOptions.desperate) {
+	    char	strbuf[100];
+	    fprintf(stderr, "__pmOptFetchAdd: fp=" PRINTF_P_PFX "%p cost=", fp);
 	    if (fp->f_cost == OPT_COST_INFINITY)
 		fprintf(stderr, "INFINITY");
 	    else
@@ -423,10 +447,11 @@ __pmOptFetchAdd(fetchctl_t **root, optreq_t *new)
 		fprintf(stderr, "INFINITY");
 	    else
 		fprintf(stderr, "%d", fp->f_newcost);
-	    fprintf(stderr, ", for %s @ grp 0x%x, state %s\n",
-		pmIDStr(pmid), fp, statestr(fp->f_state));
+	    fprintf(stderr, ", for %s @ grp " PRINTF_P_PFX "%p,",
+		pmIDStr_r(pmid, strbuf, sizeof(strbuf)), fp);
+	    fprintf(stderr, " state %s\n",
+		statestr(fp->f_state, strbuf));
 	}
-#endif
     }
 
     tfp = NULL;
@@ -444,13 +469,13 @@ __pmOptFetchAdd(fetchctl_t **root, optreq_t *new)
 	    tfp = fp;
 	}
     }
-#ifdef DESPERATE
-    if (pmDebug & DBG_TRACE_OPTFETCH) {
-	fprintf(stderr, "optFetch: chose %s cost=%d for %s @ grp 0x%x, change %s\n",
-		optcost.c_scope ? "global" : "incremental",
-		mincost, pmIDStr(pmid), tfp, statestr(tfp->f_state));
+    if (pmDebugOptions.optfetch && pmDebugOptions.desperate) {
+	char	strbuf[100];
+	fprintf(stderr, "__pmOptFetchAdd: fp=" PRINTF_P_PFX "%p chose %s cost=%d for %s @ grp " PRINTF_P_PFX "%p,",
+		tfp, optcost.c_scope ? "global" : "incremental",
+		mincost, pmIDStr_r(pmid, strbuf, sizeof(strbuf)), tfp);
+	fprintf(stderr, " change %s\n", statestr(tfp->f_state, strbuf));
     }
-#endif
 
     /*
      * Warning! Traversal of the list is a bit tricky, because the
@@ -467,6 +492,7 @@ __pmOptFetchAdd(fetchctl_t **root, optreq_t *new)
 	    }
 	    break;
 	}
+	assert(idp != NULL && pmp != NULL);
 	if (fp == tfp) {
 	    /*
 	     * The chosen one ...
@@ -511,7 +537,8 @@ __pmOptFetchAdd(fetchctl_t **root, optreq_t *new)
 	}
     }
 
-    return 0;
+    PM_UNLOCK(optfetch_lock);
+    return;
 }
 
 /*
@@ -529,6 +556,10 @@ __pmOptFetchDel(fetchctl_t **root, optreq_t *new)
     optreq_t		*rqp;
     optreq_t		*p_rqp;
 
+    if (pmDebugOptions.optfetch && pmDebugOptions.desperate) {
+	fprintf(stderr, "__pmOptFetchDel: " PRINTF_P_PFX "%p\n", new);
+    }
+
     p_fp = NULL;
     for (fp = *root; fp != NULL; fp = fp->f_next) {
 	p_idp = NULL;
@@ -580,7 +611,26 @@ __pmOptFetchDel(fetchctl_t **root, optreq_t *new)
 			    redoinst(fp);
 			    redopmid(fp);
 			    fp->f_state = OPT_STATE_PMID | OPT_STATE_PROFILE;
+			    if (pmDebugOptions.optfetch && pmDebugOptions.desperate) {
+
+
+				fprintf(stderr, "__pmOptFetchDel: redo " PRINTF_P_PFX "%p old cost=", fp);
+				if (fp->f_cost == OPT_COST_INFINITY)
+				    fprintf(stderr, "INFINITY");
+				else
+				    fprintf(stderr, "%d", fp->f_cost);
+			    }
+			    PM_LOCK(optfetch_lock);
 			    fp->f_cost = optCost(fp);
+			    PM_UNLOCK(optfetch_lock);
+			    if (pmDebugOptions.optfetch && pmDebugOptions.desperate) {
+				fprintf(stderr, " new cost=");
+				if (fp->f_cost == OPT_COST_INFINITY)
+				    fprintf(stderr, "INFINITY");
+				else
+				    fprintf(stderr, "%d", fp->f_cost);
+				fputc('\n', stderr);
+			    }
 			}
 			return 0;
 		    }
@@ -595,7 +645,7 @@ __pmOptFetchDel(fetchctl_t **root, optreq_t *new)
     return -1;
 }
 
-int
+void
 __pmOptFetchRedo(fetchctl_t **root)
 {
     fetchctl_t		*newroot = NULL;
@@ -641,7 +691,7 @@ __pmOptFetchRedo(fetchctl_t **root)
 
     if (numreq) {
 	/* something to do, randomly cut and splice the list of requests */
-	numreq = (int)lrand48() % numreq;
+	numreq = (int)lrand48() % numreq;	/* THREADSAFE - don't care */
 	t_rqp = rlist;
 	p_rqp = NULL;
 	for (rqp = rlist; rqp != NULL; rqp = rqp->r_next) {
@@ -670,19 +720,23 @@ __pmOptFetchRedo(fetchctl_t **root)
     }
 
     *root = newroot;
-    return 0;
+    return;
 }
 
-int
+void
 __pmOptFetchGetParams(optcost_t *ocp)
 {
+    PM_LOCK(optfetch_lock);
     *ocp = optcost;
-    return 0;
+    PM_UNLOCK(optfetch_lock);
+    return;
 }
 
-int
+void
 __pmOptFetchPutParams(optcost_t *ocp)
 {
+    PM_LOCK(optfetch_lock);
     optcost = *ocp;
-    return 0;
+    PM_UNLOCK(optfetch_lock);
+    return;
 }
