@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2012-2013 Red Hat.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -14,7 +15,8 @@
 
 #include <ctype.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
 
 /*
  * PDU for pm*InDom request (PDU_INSTANCE_REQ)
@@ -22,14 +24,14 @@
 typedef struct {
     __pmPDUHdr		hdr;
     pmInDom		indom;
-    __pmTimeval		when;			/* desired time */
+    pmTimeval		when;			/* desired time */
     int			inst;			/* may be PM_IN_NULL */
     int			namelen;		/* chars in name[], may be 0 */
     char		name[sizeof(int)];	/* may be missing */
 } instance_req_t;
 
 int
-__pmSendInstanceReq(int fd, int from, const __pmTimeval *when, pmInDom indom, 
+__pmSendInstanceReq(int fd, int from, const pmTimeval *when, pmInDom indom, 
 		    int inst, const char *name)
 {
     instance_req_t	*pp;
@@ -39,7 +41,7 @@ __pmSendInstanceReq(int fd, int from, const __pmTimeval *when, pmInDom indom,
     need = sizeof(instance_req_t) - sizeof(int);
     if (name != NULL)
 	need += PM_PDU_SIZE_BYTES(strlen(name));
-    if ((pp = (instance_req_t *)__pmFindPDUBuf(sizeof(need))) == NULL)
+    if ((pp = (instance_req_t *)__pmFindPDUBuf(need)) == NULL)
 	return -oserror();
     pp->hdr.len = need;
     pp->hdr.type = PDU_INSTANCE_REQ;
@@ -53,16 +55,14 @@ __pmSendInstanceReq(int fd, int from, const __pmTimeval *when, pmInDom indom,
     else {
 	pp->namelen = (int)strlen(name);
 	memcpy((void *)pp->name, (void *)name, pp->namelen);
-#ifdef PCP_DEBUG
 	if ((pp->namelen % sizeof(__pmPDU)) != 0) {
-	    /* for Purify */
+            /* clear the padding bytes, lest they contain garbage */
 	    int	pad;
 	    char	*padp = pp->name + pp->namelen;
 
 	    for (pad = sizeof(__pmPDU) - 1; pad >= (pp->namelen % sizeof(__pmPDU)); pad--)
 		*padp++ = '~';	/* buffer end */
 	}
-#endif
 	pp->namelen = htonl(pp->namelen);
     }
 
@@ -72,25 +72,38 @@ __pmSendInstanceReq(int fd, int from, const __pmTimeval *when, pmInDom indom,
 }
 
 int
-__pmDecodeInstanceReq(__pmPDU *pdubuf, __pmTimeval *when, pmInDom *indom, int *inst, char **name)
+__pmDecodeInstanceReq(__pmPDU *pdubuf, pmTimeval *when, pmInDom *indom, int *inst, char **name)
 {
     instance_req_t	*pp;
+    char		*pdu_end;
+    int			namelen;
 
     pp = (instance_req_t *)pdubuf;
+    pdu_end = (char *)pdubuf + pp->hdr.len;
+
+    if (pdu_end - (char *)pp < sizeof(instance_req_t) - sizeof(pp->name))
+	return PM_ERR_IPC;
+
     when->tv_sec = ntohl(pp->when.tv_sec);
     when->tv_usec = ntohl(pp->when.tv_usec);
     *indom = __ntohpmInDom(pp->indom);
     *inst = ntohl(pp->inst);
-    pp->namelen = ntohl(pp->namelen);
-    if (pp->namelen) {
-	if ((*name = (char *)malloc(pp->namelen+1)) == NULL)
+    namelen = ntohl(pp->namelen);
+    if (namelen > 0) {
+	if (namelen >= INT_MAX - 1 || namelen > pp->hdr.len)
+	    return PM_ERR_IPC;
+	if (pdu_end - (char *)pp < sizeof(instance_req_t) - sizeof(pp->name) + namelen)
+	    return PM_ERR_IPC;
+	if ((*name = (char *)malloc(namelen+1)) == NULL)
 	    return -oserror();
-	strncpy(*name, pp->name, pp->namelen);
-	(*name)[pp->namelen] = '\0';
+	strncpy(*name, pp->name, namelen);
+	(*name)[namelen] = '\0';
     }
-    else
+    else if (namelen < 0) {
+	return PM_ERR_IPC;
+    } else {
 	*name = NULL;
-
+    }
     return 0;
 }
 
@@ -111,19 +124,17 @@ typedef struct {
 } instance_t;
 
 int
-__pmSendInstance(int fd, int from, __pmInResult *result)
+__pmSendInstance(int fd, int from, pmInResult *result)
 {
-    instance_t	*rp;
+    instance_t		*rp;
     instlist_t		*ip;
     int			need;
     int			i;
     int			j;
     int			sts;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_INDOM)
+    if (pmDebugOptions.indom)
 	__pmDumpInResult(stderr, result);
-#endif
 
     need = sizeof(*rp) - sizeof(rp->rest);
     /* instlist_t + name rounded up to a __pmPDU boundary */
@@ -151,15 +162,13 @@ __pmSendInstance(int fd, int from, __pmInResult *result)
 	if (result->namelist != NULL) {
 	    ip->namelen = (int)strlen(result->namelist[i]);
 	    memcpy((void *)ip->name, (void *)result->namelist[i], ip->namelen);
-#ifdef PCP_DEBUG
 	    if ((ip->namelen % sizeof(__pmPDU)) != 0) {
-		/* for Purify */
+                /* clear the padding bytes, lest they contain garbage */
 		int	pad;
 		char	*padp = ip->name + ip->namelen;
 		for (pad = sizeof(__pmPDU) - 1; pad >= (ip->namelen % sizeof(__pmPDU)); pad--)
 		    *padp++ = '~';	/* buffer end */
 	    }
-#endif
 	    j += sizeof(*ip) - sizeof(ip->name) + PM_PDU_SIZE_BYTES(ip->namelen);
 	    ip->namelen = htonl(ip->namelen);
 	}
@@ -175,27 +184,38 @@ __pmSendInstance(int fd, int from, __pmInResult *result)
 }
 
 int
-__pmDecodeInstance(__pmPDU *pdubuf, __pmInResult **result)
+__pmDecodeInstance(__pmPDU *pdubuf, pmInResult **result)
 {
     int			i;
     int			j;
     instance_t		*rp;
     instlist_t		*ip;
-    __pmInResult	*res = NULL;
+    pmInResult	*res;
     int			sts;
     char		*p;
+    char		*pdu_end;
     int			keep_instlist;
     int			keep_namelist;
 
-    if ((res = (__pmInResult *)malloc(sizeof(*res))) == NULL)
+    rp = (instance_t *)pdubuf;
+    pdu_end = (char *)pdubuf + rp->hdr.len;
+
+    if (pdu_end - (char *)pdubuf < sizeof(instance_t) - sizeof(__pmPDU))
+	return PM_ERR_IPC;
+
+    if ((res = (pmInResult *)malloc(sizeof(*res))) == NULL)
 	return -oserror();
     res->instlist = NULL;
     res->namelist = NULL;
-
-    rp = (instance_t *)pdubuf;
     res->indom = __ntohpmInDom(rp->indom);
     res->numinst = ntohl(rp->numinst);
 
+    if (res->numinst >= (INT_MAX / sizeof(res->instlist[0])) ||
+	res->numinst >= (INT_MAX / sizeof(res->namelist[0])) ||
+	res->numinst >= rp->hdr.len) {
+	sts = PM_ERR_IPC;
+	goto badsts;
+    }
     if ((res->instlist = (int *)malloc(res->numinst * sizeof(res->instlist[0]))) == NULL) {
 	sts = -oserror();
 	goto badsts;
@@ -214,12 +234,25 @@ __pmDecodeInstance(__pmPDU *pdubuf, __pmInResult **result)
 
     for (i = j = 0; i < res->numinst; i++) {
 	ip = (instlist_t *)&rp->rest[j/sizeof(__pmPDU)];
+	if (sizeof(instlist_t) - sizeof(ip->name) > (size_t)(pdu_end - (char *)ip)) {
+	    sts = PM_ERR_IPC;
+	    goto badsts;
+	}
+
 	res->instlist[i] = ntohl(ip->inst);
 	if (res->instlist[i] != PM_IN_NULL)
 	    keep_instlist = 1;
 	ip->namelen = ntohl(ip->namelen);
 	if (ip->namelen > 0)
 	    keep_namelist = 1;
+	if (ip->namelen < 0) {
+	    sts = PM_ERR_IPC;
+	    goto badsts;
+	}
+	if (sizeof(instlist_t) - sizeof(int) + ip->namelen > (size_t)(pdu_end - (char *)ip)) {
+	    sts = PM_ERR_IPC;
+	    goto badsts;
+	}
 	if ((p = (char *)malloc(ip->namelen + 1)) == NULL) {
 	    sts = -oserror();
 	    goto badsts;
@@ -239,10 +272,8 @@ __pmDecodeInstance(__pmPDU *pdubuf, __pmInResult **result)
 	res->namelist = NULL;
     }
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_INDOM)
+    if (pmDebugOptions.indom)
 	__pmDumpInResult(stderr, res);
-#endif
     *result = res;
     return 0;
 
