@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2012-2014 Red Hat.
  * Copyright (c) 1995-2000 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -33,7 +34,8 @@
 
 #include <ctype.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
 
 /*
  * PDU for pmResult (PDU_RESULT)
@@ -48,7 +50,7 @@ typedef struct {
 
 typedef struct {
     __pmPDUHdr		hdr;
-    __pmTimeval		timestamp;	/* when returned */
+    pmTimeval		timestamp;	/* when returned */
     int			numpmid;	/* no. of PMIDs to follow */
     __pmPDU		data[1];	/* zero or more */
 } result_t;
@@ -65,7 +67,6 @@ __pmEncodeResult(int targetfd, const pmResult *result, __pmPDU **pdubuf)
     result_t	*pp;
     vlist_t	*vlp;
 
-    /* to start with, need space for result_t with no data (__pmPDU) */
     need = sizeof(result_t) - sizeof(__pmPDU);
     vneed = 0;
     /* now add space for each vlist_t (data in result_t) */
@@ -78,7 +79,7 @@ __pmEncodeResult(int targetfd, const pmResult *result, __pmPDU **pdubuf)
 	for (j = 0; j < vsp->numval; j++) {
 	    /* plus value, instance pair */
 	    need += sizeof(__pmValue_PDU);
-	    if (vsp->valfmt != PM_VAL_INSITU) {
+	    if (vsp->valfmt == PM_VAL_DPTR || vsp->valfmt == PM_VAL_SPTR) {
 		/* plus pmValueBlock */
 		vneed += PM_PDU_SIZE_BYTES(vsp->vlist[j].value.pval->vlen);
 	    }
@@ -87,7 +88,11 @@ __pmEncodeResult(int targetfd, const pmResult *result, __pmPDU **pdubuf)
 	    /* optional value format, if any values present */
 	    need += sizeof(int);
     }
-    if ((_pdubuf = __pmFindPDUBuf((int)(need+vneed))) == NULL)
+    /*
+     * Need to reserve additonal space for trailer (an int) in case the
+     * PDU buffer is used by __pmLogPutResult2()
+     */
+    if ((_pdubuf = __pmFindPDUBuf((int)(need+vneed+sizeof(int)))) == NULL)
 	return -oserror();
     pp = (result_t *)_pdubuf;
     pp->hdr.len = (int)(need+vneed);
@@ -107,9 +112,7 @@ __pmEncodeResult(int targetfd, const pmResult *result, __pmPDU **pdubuf)
 	    vlp->valfmt = htonl(vsp->valfmt);
 	for (j = 0; j < vsp->numval; j++) {
 	    vlp->vlist[j].inst = htonl(vsp->vlist[j].inst);
-	    if (vsp->valfmt == PM_VAL_INSITU)
-		vlp->vlist[j].value.lval = htonl(vsp->vlist[j].value.lval);
-	    else {
+	    if (vsp->valfmt == PM_VAL_DPTR || vsp->valfmt == PM_VAL_SPTR) {
 		/*
 		 * pmValueBlocks are harder!
 		 * -- need to copy the len field (len) + len bytes (vbuf)
@@ -117,20 +120,22 @@ __pmEncodeResult(int targetfd, const pmResult *result, __pmPDU **pdubuf)
 		int	nb;
 		nb = vsp->vlist[j].value.pval->vlen;
 		memcpy((void *)vbp, (void *)vsp->vlist[j].value.pval, nb);
-#ifdef PCP_DEBUG
 		if ((nb % sizeof(__pmPDU)) != 0) {
-		    /* for Purify */
+		    /* clear the padding bytes, lest they contain garbage */
 		    int	pad;
 		    char	*padp = (char *)vbp + nb;
 		    for (pad = sizeof(__pmPDU) - 1; pad >= (nb % sizeof(__pmPDU)); pad--)
 			*padp++ = '~';	/* buffer end */
 		}
-#endif
 		__htonpmValueBlock((pmValueBlock *)vbp);
 		/* point to the value block at the end of the PDU */
 		vlp->vlist[j].value.lval = htonl((int)(vbp - _pdubuf));
 		vbp += PM_PDU_SIZE(nb);
 	    }
+	    else {
+		/* punt on vsp->valfmt == PM_VAL_INSITU */
+		vlp->vlist[j].value.lval = htonl(vsp->vlist[j].value.lval);
+	    }
 	}
 	vlp->numval = htonl(vsp->numval);
 	if (j > 0)
@@ -144,17 +149,21 @@ __pmEncodeResult(int targetfd, const pmResult *result, __pmPDU **pdubuf)
     return 0;
 }
 
+/*
+ * Internal variant of __pmSendResult() with current context.
+ */
 int
-__pmSendResult(int fd, int from, const pmResult *result)
+__pmSendResult_ctx(__pmContext *ctxp, int fd, int from, const pmResult *result)
 {
     int		sts;
     __pmPDU	*pdubuf;
     result_t	*pp;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_PDU)
-	__pmDumpResult(stderr, result);
-#endif
+    if (ctxp != NULL)
+	PM_ASSERT_IS_LOCKED(ctxp->c_lock);
+
+    if (pmDebugOptions.pdu)
+	__pmDumpResult_ctx(ctxp, stderr, result);
     if ((sts = __pmEncodeResult(fd, result, &pdubuf)) < 0)
 	return sts;
     pp = (result_t *)pdubuf;
@@ -164,12 +173,22 @@ __pmSendResult(int fd, int from, const pmResult *result)
     return sts;
 }
 
+int
+__pmSendResult(int fd, int from, const pmResult *result)
+{
+    int	sts;
+    sts = __pmSendResult_ctx(NULL, fd, from, result);
+    return sts;
+}
+
 /*
- * enter here with pdubuf already pinned ... result may point into
+ * Internal variant of __pmDecodeResult() with current context.
+ *
+ * Enter here with pdubuf already pinned ... result may point into
  * _another_ pdu buffer that is pinned on exit
  */
 int
-__pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
+__pmDecodeResult_ctx(__pmContext *ctxp, __pmPDU *pdubuf, pmResult **result)
 {
     int		numpmid;	/* number of metrics */
     int		i;		/* range of metrics */
@@ -200,16 +219,31 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
     Bozo - unexpected sizeof pointer!!
 #endif
 
+    if (ctxp != NULL)
+	PM_ASSERT_IS_LOCKED(ctxp->c_lock);
+
     pp = (result_t *)pdubuf;
     pduend = (char *)pdubuf + pp->hdr.len;
-    if (pduend - (char *)pdubuf < sizeof(result_t) - sizeof(__pmPDU))
+    if (pduend - (char *)pdubuf < sizeof(result_t) - sizeof(__pmPDU)) {
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+	    fprintf(stderr, "__pmDecodeResult: Bad: len=%d smaller than min %d\n", pp->hdr.len, (int)(sizeof(result_t) - sizeof(__pmPDU)));
+	}
 	return PM_ERR_IPC;
+    }
 
     numpmid = ntohl(pp->numpmid);
-    if (numpmid < 0 || numpmid > pp->hdr.len)
+    if (numpmid < 0 || numpmid > pp->hdr.len) {
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+	    fprintf(stderr, "__pmDecodeResult: Bad: numpmid=%d negative or not smaller than PDU len %d\n", numpmid, pp->hdr.len);
+	}
 	return PM_ERR_IPC;
-    if (numpmid >= (INT_MAX - sizeof(pmResult)) / sizeof(pmValueSet *))
+    }
+    if (numpmid >= (INT_MAX - sizeof(pmResult)) / sizeof(pmValueSet *)) {
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+	    fprintf(stderr, "__pmDecodeResult: Bad: numpmid=%d larger than max %ld\n", numpmid, (long)(INT_MAX - sizeof(pmResult) / sizeof(pmValueSet *)));
+	}
 	return PM_ERR_IPC;
+    }
     if ((pr = (pmResult *)malloc(sizeof(pmResult) +
 			     (numpmid - 1) * sizeof(pmValueSet *))) == NULL) {
 	return -oserror();
@@ -224,80 +258,117 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
     for (i = 0; i < numpmid; i++) {
 	vlp = (vlist_t *)&pp->data[vsize/sizeof(__pmPDU)];
 
-	if (sizeof(*vlp) - sizeof(vlp->vlist) - sizeof(int) > (pduend - (char *)vlp))
+	if (sizeof(*vlp) - sizeof(vlp->vlist) - sizeof(int) > (pduend - (char *)vlp)) {
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+	    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] outer vlp past end of PDU buffer\n", i);
+	}
 	    goto corrupt;
+	}
 
 	vsize += sizeof(vlp->pmid) + sizeof(vlp->numval);
 	nvsize += sizeof(pmValueSet);
 	numval = ntohl(vlp->numval);
 	valfmt = ntohl(vlp->valfmt);
-#ifdef DESPERATE
-	{
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
 	    pmID	pmid = __ntohpmID(vlp->pmid);
 	    char	strbuf[20];
 	    fprintf(stderr, "vlist[%d] pmid: %s numval: %d",
 			i, pmIDStr_r(pmid, strbuf, sizeof(strbuf)), numval);
 	}
-#endif
 	/* numval may be negative - it holds an error code in that case */
-	if (numval > pp->hdr.len)
+	if (numval > pp->hdr.len) {
+	    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] numval=%d > len=%d\n", i, numval, pp->hdr.len);
+	    }
 	    goto corrupt;
+	}
 	if (numval > 0) {
-	    if (sizeof(*vlp) - sizeof(vlp->vlist) > (pduend - (char *)vlp))
+	    if (sizeof(*vlp) - sizeof(vlp->vlist) > (pduend - (char *)vlp)) {
+		if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] inner vlp past end of PDU buffer\n", i);
+		}
 		goto corrupt;
-	    if (numval >= (INT_MAX - sizeof(*vlp)) / sizeof(__pmValue_PDU))
+	    }
+	    if (numval >= (INT_MAX - sizeof(*vlp)) / sizeof(__pmValue_PDU)) {
+		if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] numval=%d > max=%ld\n", i, numval, (long)((INT_MAX - sizeof(*vlp)) / sizeof(__pmValue_PDU)));
+		}
 		goto corrupt;
+	    }
 	    vsize += sizeof(vlp->valfmt) + numval * sizeof(__pmValue_PDU);
 	    nvsize += (numval - 1) * sizeof(pmValue);
-#ifdef DESPERATE
-	    fprintf(stderr, " valfmt: %s",
+	    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		fprintf(stderr, " valfmt: %s",
 			valfmt == PM_VAL_INSITU ? "insitu" : "ptr");
-#endif
+	    }
 	    if (valfmt != PM_VAL_INSITU) {
 		for (j = 0; j < numval; j++) {
 		    __pmValue_PDU *pduvp;
 		    pmValueBlock *pduvbp;
 
 		    pduvp = &vlp->vlist[j];
-		    if (sizeof(__pmValue_PDU) > (size_t)(pduend - (char *)pduvp))
+		    if (sizeof(__pmValue_PDU) > (size_t)(pduend - (char *)pduvp)) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] initial pduvp past end of PDU buffer\n", i, j);
+			}
 			goto corrupt;
+		    }
 		    index = ntohl(pduvp->value.lval);
-		    if (index < 0 || index > pp->hdr.len)
+		    if (index < 0 || index > pp->hdr.len) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] index=%d\n", i, j, index);
+			}
 			goto corrupt;
+		    }
 		    pduvbp = (pmValueBlock *)&pdubuf[index];
 		    if (vsplit > (char *)pduvbp)
 			vsplit = (char *)pduvbp;
-		    if (sizeof(unsigned int) > (size_t)(pduend - (char *)pduvbp))
+		    if (sizeof(unsigned int) > (size_t)(pduend - (char *)pduvbp)) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] second pduvp past end of PDU buffer\n", i, j);
+			}
 			goto corrupt;
+		    }
 		    __ntohpmValueBlock(pduvbp);
-		    if (pduvbp->vlen < PM_VAL_HDR_SIZE || pduvbp->vlen > pp->hdr.len)
+		    if (pduvbp->vlen < PM_VAL_HDR_SIZE || pduvbp->vlen > pp->hdr.len) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] vlen=%d\n", i, j, pduvbp->vlen);
+			}
 			goto corrupt;
-		    if (pduvbp->vlen > (size_t)(pduend - (char *)pduvbp))
+		    }
+		    if (pduvbp->vlen > (size_t)(pduend - (char *)pduvbp)) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] pduvp past end of PDU buffer\n", i, j);
+			}
 			goto corrupt;
+		    }
 		    vbsize += PM_PDU_SIZE_BYTES(pduvbp->vlen);
-#ifdef DESPERATE
-		    fprintf(stderr, " len: %d type: %d",
+		    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			fprintf(stderr, " len: %d type: %d",
 			    pduvbp->vlen - PM_VAL_HDR_SIZE, pduvbp->vtype);
-#endif
+		    }
 		}
 	    }
 	}
-#ifdef DESPERATE
-	fputc('\n', stderr);
-#endif
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+	    fputc('\n', stderr);
+	}
     }
-#ifdef DESPERATE
-    fprintf(stderr, "vsize: %d nvsize: %d vbsize: %d\n", vsize, nvsize, vbsize);
-#endif
 
     need = nvsize + vbsize;
     offset = sizeof(result_t) - sizeof(__pmPDU) + vsize;
+
+    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+	fprintf(stderr, "need: %d vsize: %d nvsize: %d vbsize: %d offset: %d hdr.len: %d pduend: %p vsplit: %p (diff %d) pdubuf: %p (diff %d)\n", need, vsize, nvsize, vbsize, offset, pp->hdr.len, pduend, vsplit, (int)(pduend-vsplit), pdubuf, (int)(pduend-(char *)pdubuf));
+    }
+
     if (need < 0 ||
 	vsize > INT_MAX / sizeof(__pmPDU) ||
 	vbsize > INT_MAX / sizeof(pmValueBlock) ||
 	offset != pp->hdr.len - (pduend - vsplit) ||
-	offset + vbsize != pduend - (char *)pdubuf)
+	offset + vbsize != pduend - (char *)pdubuf) {
 	goto corrupt;
+    }
 
     /* the original pdubuf is already pinned so we won't allocate that again */
     if ((newbuf = (char *)__pmFindPDUBuf(need)) == NULL) {
@@ -311,7 +382,7 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
      *
      * From the original PDU buffer ...
      * :-----:---------:-----------:----------------:---------------------:
-     * : Hdr : timestamp : numpmid : ... vlists ... : .. pmValueBlocks .. :
+     * : Hdr :timestamp:  numpmid  : ... vlists ... : .. pmValueBlocks .. :
      * :-----:---------:-----------:----------------:---------------------:
      *                              <---  vsize ---> <----   vbsize  ---->
      *                                    bytes              bytes
@@ -345,22 +416,20 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
 	pr->vset[i] = nvsp;
 	nvsp->pmid = __ntohpmID(vlp->pmid);
 	nvsp->numval = ntohl(vlp->numval);
-#ifdef DESPERATE
-	{
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
 	    char	strbuf[20];
 	    fprintf(stderr, "new vlist[%d] pmid: %s numval: %d",
 			i, pmIDStr_r(nvsp->pmid, strbuf, sizeof(strbuf)), nvsp->numval);
 	}
-#endif
 
 	vsize += sizeof(nvsp->pmid) + sizeof(nvsp->numval);
 	nvsize += sizeof(pmValueSet);
 	if (nvsp->numval > 0) {
 	    nvsp->valfmt = ntohl(vlp->valfmt);
-#ifdef DESPERATE
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
 	    fprintf(stderr, " valfmt: %s",
 			    nvsp->valfmt == PM_VAL_INSITU ? "insitu" : "ptr");
-#endif
+	}
 	    vsize += sizeof(nvsp->valfmt) + nvsp->numval * sizeof(__pmValue_PDU);
 	    nvsize += (nvsp->numval - 1) * sizeof(pmValue);
 	    for (j = 0; j < nvsp->numval; j++) {
@@ -370,33 +439,37 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
 		nvp->inst = ntohl(vp->inst);
 		if (nvsp->valfmt == PM_VAL_INSITU) {
 		    nvp->value.lval = ntohl(vp->value.lval);
-#ifdef DESPERATE
-		    fprintf(stderr, " value: %d", nvp->value.lval);
-#endif
+		    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			fprintf(stderr, " value: %d", nvp->value.lval);
+		    }
 		}
-		else {
+		else if (nvsp->valfmt == PM_VAL_DPTR || nvsp->valfmt == PM_VAL_SPTR) {
 		    /*
-		     * in the input PDU buffer, pval is an index to the
+		     * in the input PDU buffer, lval is an index to the
 		     * start of the pmValueBlock, in units of __pmPDU
 		     */
-		    index = sizeof(__pmPDU) * ntohl(vp->value.pval) + offset;
+		    index = sizeof(__pmPDU) * ntohl(vp->value.lval) + offset;
 		    nvp->value.pval = (pmValueBlock *)&newbuf[index];
-#ifdef DESPERATE
-		    {
+		    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
 			int		k, len;
 			len = nvp->value.pval->vlen - PM_VAL_HDR_SIZE;
-			fprintf(stderr, " len: %d type: %dvalue: 0x", len,
-				nvp->value.pval->vtype;
+			fprintf(stderr, " len: %d type: %d value: 0x", len,
+				nvp->value.pval->vtype);
 			for (k = 0; k < len; k++)
 			    fprintf(stderr, "%02x", nvp->value.pval->vbuf[k]);
 		    }
-#endif
+		}
+		else {
+		    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			fprintf(stderr, " botch: valfmt=%d\n", nvsp->valfmt);
+		    }
+		    goto corrupt;
 		}
 	    }
 	}
-#ifdef DESPERATE
-	fputc('\n', stderr);
-#endif
+	if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+	    fputc('\n', stderr);
+	}
     }
     if (numpmid == 0)
 	__pmUnpinPDUBuf(newbuf);
@@ -414,22 +487,38 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
      * the PDU buffer) by adding the base address of the PDU buffer.
      */
     for (i = 0; i < numpmid; i++) {
-	if (sizeof(*vlp) - sizeof(vlp->vlist) - sizeof(int) > (pduend - (char *)vlp))
+	if (sizeof(*vlp) - sizeof(vlp->vlist) - sizeof(int) > (pduend - (char *)vlp)) {
+	    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] outer vlp past end of PDU buffer\n", i);
+	    }
 	    goto corrupt;
+	}
 
 	vsp = pr->vset[i] = (pmValueSet *)vlp;
 	vsp->pmid = __ntohpmID(vsp->pmid);
 	vsp->numval = ntohl(vsp->numval);
 	/* numval may be negative - it holds an error code in that case */
-	if (vsp->numval > pp->hdr.len)
+	if (vsp->numval > pp->hdr.len) {
+	    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] numval=%d > len=%d\n", i, vsp->numval, pp->hdr.len);
+	    }
 	    goto corrupt;
+	}
 
 	vsize += sizeof(vsp->pmid) + sizeof(vsp->numval);
 	if (vsp->numval > 0) {
-	    if (sizeof(*vlp) - sizeof(vlp->vlist) > (pduend - (char *)vlp))
+	    if (sizeof(*vlp) - sizeof(vlp->vlist) > (pduend - (char *)vlp)) {
+		if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] inner vlp past end of PDU buffer\n", i);
+		}
 		goto corrupt;
-	    if (vsp->numval >= (INT_MAX - sizeof(*vlp)) / sizeof(__pmValue_PDU))
+	    }
+	    if (vsp->numval >= (INT_MAX - sizeof(*vlp)) / sizeof(__pmValue_PDU)) {
+		if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] numval=%d > max=%ld\n", i, vsp->numval, (long)((INT_MAX - sizeof(*vlp)) / sizeof(__pmValue_PDU)));
+		}
 		goto corrupt;
+	    }
 	    vsp->valfmt = ntohl(vsp->valfmt);
 	    vsize += sizeof(vsp->valfmt) + vsp->numval * sizeof(__pmValue_PDU);
 	    for (j = 0; j < vsp->numval; j++) {
@@ -437,8 +526,12 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
 		pmValueBlock *pduvbp;
 
 		pduvp = &vsp->vlist[j];
-		if (sizeof(__pmValue_PDU) > (size_t)(pduend - (char *)pduvp))
+		if (sizeof(__pmValue_PDU) > (size_t)(pduend - (char *)pduvp)) {
+		    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] initial pduvp past end of PDU buffer\n", i, j);
+		    }
 		    goto corrupt;
+		}
 
 		pduvp->inst = ntohl(pduvp->inst);
 		if (vsp->valfmt == PM_VAL_INSITU) {
@@ -446,18 +539,34 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
 		} else {
 		    /* salvage pmValueBlocks from end of PDU */
 		    index = ntohl(pduvp->value.lval);
-		    if (index < 0 || index > pp->hdr.len)
+		    if (index < 0 || index > pp->hdr.len) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] index=%d\n", i, j, index);
+			}
 			goto corrupt;
+		    }
 		    pduvbp = (pmValueBlock *)&pdubuf[index];
 		    if (vsplit > (char *)pduvbp)
 			vsplit = (char *)pduvbp;
-		    if (sizeof(unsigned int) > (size_t)(pduend - (char *)pduvbp))
+		    if (sizeof(unsigned int) > (size_t)(pduend - (char *)pduvbp)) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] second pduvp past end of PDU buffer\n", i, j);
+			}
 			goto corrupt;
+		    }
 		    __ntohpmValueBlock(pduvbp);
-		    if (pduvbp->vlen < PM_VAL_HDR_SIZE || pduvbp->vlen > pp->hdr.len)
+		    if (pduvbp->vlen < PM_VAL_HDR_SIZE || pduvbp->vlen > pp->hdr.len) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] vlen=%d\n", i, j, pduvbp->vlen);
+			}
 			goto corrupt;
-		    if (pduvbp->vlen > (size_t)(pduend - (char *)pduvbp))
+		    }
+		    if (pduvbp->vlen > (size_t)(pduend - (char *)pduvbp)) {
+			if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+			    fprintf(stderr, "__pmDecodeResult: Bad: pmid[%d] value[%d] pduvp past end of PDU buffer\n", i, j);
+			}
 			goto corrupt;
+		    }
 		    pduvp->value.pval = pduvbp;
 		}
 	    }
@@ -468,8 +577,12 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
 	}
     }
     if (numpmid > 0) {
-	if (sizeof(result_t) - sizeof(__pmPDU) + vsize != pp->hdr.len - (pduend - vsplit))
+	if (sizeof(result_t) - sizeof(__pmPDU) + vsize != pp->hdr.len - (pduend - vsplit)) {
+	    if (pmDebugOptions.pdu && pmDebugOptions.desperate) {
+		fprintf(stderr, "__pmDecodeResult: Bad: vsplit past end of PDU buffer\n");
+	    }
 	    goto corrupt;
+	}
 	/*
 	 * PDU buffer already pinned on entry, pin again so that
 	 * the caller can safely call _both_ pmFreeResult() and
@@ -479,10 +592,8 @@ __pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
     }
 #endif
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_PDU)
-	__pmDumpResult(stderr, pr);
-#endif
+    if (pmDebugOptions.pdu)
+	__pmDumpResult_ctx(ctxp, stderr, pr);
 
     /*
      * Note we return with the input buffer (pdubuf) still pinned and
@@ -496,3 +607,11 @@ corrupt:
     free(pr);
     return PM_ERR_IPC;
 }
+
+int
+__pmDecodeResult(__pmPDU *pdubuf, pmResult **result)
+{
+    int	sts;
+    sts = __pmDecodeResult_ctx(NULL, pdubuf, result);
+    return sts;
+}
