@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2012-2013 Red Hat.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -13,7 +14,8 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
 
 /*
  * PDU for pmLookupText request (PDU_TEXT_REQ)
@@ -38,11 +40,10 @@ __pmSendTextReq(int fd, int from, int ident, int type)
 
     if (type & PM_TEXT_PMID)
 	pp->ident = __htonpmID((pmID)ident);
-    else if (type & PM_TEXT_INDOM)
-	pp->ident = __htonpmInDom((pmInDom)ident);
-    else
+    else /* (type & PM_TEXT_INDOM) */
 	pp->ident = __htonpmInDom((pmInDom)ident);
 
+    type &= ~PM_TEXT_DIRECT;
     pp->type = htonl(type);
 
     sts = __pmXmitPDU(fd, (__pmPDU *)pp);
@@ -65,9 +66,10 @@ __pmDecodeTextReq(__pmPDU *pdubuf, int *ident, int *type)
     *type = ntohl(pp->type);
     if ((*type) & PM_TEXT_PMID)
 	*ident = __ntohpmID(pp->ident);
-    else
-    if ((*type) & PM_TEXT_INDOM)
+    else if ((*type) & PM_TEXT_INDOM)
 	*ident = __ntohpmInDom(pp->ident);
+    else
+	*ident = PM_INDOM_NULL;
 
     return 0;
 }
@@ -87,9 +89,10 @@ __pmSendText(int fd, int ctx, int ident, const char *buffer)
 {
     text_t	*pp;
     size_t	need;
+    size_t	len = strlen(buffer);
     int		sts;
 
-    need = sizeof(text_t) - sizeof(pp->buffer) + PM_PDU_SIZE_BYTES(strlen(buffer));
+    need = sizeof(text_t) - sizeof(pp->buffer) + PM_PDU_SIZE_BYTES(len);
     if ((pp = (text_t *)__pmFindPDUBuf((int)need)) == NULL)
 	return -oserror();
     pp->hdr.len = (int)need;
@@ -100,13 +103,18 @@ __pmSendText(int fd, int ctx, int ident, const char *buffer)
      * The caller has to do this because the type of ident is not
      * part of the transmitted PDU_TEXT pdu; ident may be either
      * a pmID or pmInDom, and so the caller must use either
-     * __htonpmID or __htonpmInDom (respectfully).
+     * __htonpmID or __htonpmInDom (respectively).
      */
     pp->ident = ident;
-
-    pp->buflen = (int)strlen(buffer);
-    memcpy((void *)pp->buffer, (void *)buffer, pp->buflen);
-    pp->buflen = htonl(pp->buflen);
+    pp->buflen = htonl(len);
+    memcpy((void *)pp->buffer, (void *)buffer, len);
+    if (len % sizeof(__pmPDU) != 0) {
+	/* clear the padding bytes, lest they contain garbage */
+	int	pad;
+	char	*padp = pp->buffer + len;
+	for (pad = sizeof(__pmPDU) - 1; pad >= (len % sizeof(__pmPDU)); pad--)
+	    *padp++ = '~';	/* buffer end */
+    }
 
     sts = __pmXmitPDU(fd, (__pmPDU *)pp);
     __pmUnpinPDUBuf(pp);
