@@ -1,22 +1,22 @@
 /*
- * Copyright (c) 2012-2013 Red Hat.
+ * Copyright (c) 2012-2015 Red Hat.
  *
  * Server side security features - via Network Security Services (NSS) and
  * the Simple Authentication and Security Layer (SASL).
- *
- * This program is free software; you can redistribute it and/or modify it
- * under the terms of the GNU General Public License as published by the
- * Free Software Foundation; either version 2 of the License, or (at your
- * option) any later version.
- *
- * This program is distributed in the hope that it will be useful, but
+ * 
+ * This library is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as published
+ * by the Free Software Foundation; either version 2.1 of the License, or
+ * (at your option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
- * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
- * for more details.
+ * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+ * License for more details.
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #define SOCKET_INTERNAL
 #include "internal.h"
 #include <keyhi.h>
@@ -25,6 +25,7 @@
 #include <sys/stat.h>
 
 #define MAX_NSSDB_PASSWORD_LENGTH	256
+#define MAX_CERT_NAME_LENGTH		256
 
 static struct {
     /* NSS certificate management */
@@ -32,31 +33,75 @@ static struct {
     SECKEYPrivateKey	*private_key;
     const char		*password_file;
     SSLKEAType		certificate_KEA;
-    unsigned int	certificate_verified : 1;
-    unsigned int	ssl_session_cache_setup : 1;
     char		database_path[MAXPATHLEN];
 
-    /* SASL authentication fields */
+    unsigned int	server_features;
+    char		cert_nickname[MAX_CERT_NAME_LENGTH];
+
+    /* status flags (bitfields) */
+    unsigned int	initialized : 1;
+    unsigned int	init_failed : 1;
+    unsigned int	certificate_verified : 1;	/* NSS */
+    unsigned int	ssl_session_cache_setup : 1;	/* NSS */
 } secure_server;
 
+#ifdef PM_MULTI_THREAD
+static pthread_mutex_t	secureserver_lock;
+#else
+void			*secureserver_lock;
+#endif
+
+#if defined(PM_MULTI_THREAD) && defined(PM_MULTI_THREAD_DEBUG)
+/*
+ * return true if lock == secureserver_lock
+ */
 int
-__pmServerHasFeature(__pmServerFeature query)
+__pmIsSecureserverLock(void *lock)
+{
+    return lock == (void *)&secureserver_lock;
+}
+#endif
+
+void
+init_secureserver_lock(void)
+{
+    __pmInitMutex(&secureserver_lock);
+}
+
+int
+__pmSecureServerSetFeature(__pmServerFeature wanted)
+{
+    if (wanted == PM_SERVER_FEATURE_CERT_REQD){
+        secure_server.server_features |= (1 << wanted);
+        return 1;
+    }
+    return 0;
+}
+
+int
+__pmSecureServerClearFeature(__pmServerFeature clear)
+{
+    if (clear == PM_SERVER_FEATURE_CERT_REQD){
+    	secure_server.server_features &= ~(1<<clear);
+	return 1;
+    }
+    return 0;
+}
+
+int
+__pmSecureServerHasFeature(__pmServerFeature query)
 {
     int sts = 0;
 
     switch (query) {
     case PM_SERVER_FEATURE_SECURE:
-	PM_INIT_LOCKS();
-	PM_LOCK(__pmLock_libpcp);
-	sts = secure_server.certificate_verified;
-	PM_UNLOCK(__pmLock_libpcp);
-	break;
+	return ! secure_server.init_failed;
     case PM_SERVER_FEATURE_COMPRESS:
     case PM_SERVER_FEATURE_AUTH:
 	sts = 1;
 	break;
-    case PM_SERVER_FEATURE_IPV6:
-	sts = (strcmp(__pmGetAPIConfig("ipv6"), "true") == 0);
+    case PM_SERVER_FEATURE_CERT_REQD:
+	sts = ( (secure_server.server_features & (1 << PM_SERVER_FEATURE_CERT_REQD)) != 0 );
 	break;
     default:
 	break;
@@ -117,26 +162,24 @@ certificate_database_password(PK11SlotInfo *info, PRBool retry, void *arg)
     (void)arg;
     (void)info;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_ASSERT_IS_LOCKED(secureserver_lock);
     passfile[0] = '\0';
     if (secure_server.password_file)
 	strncpy(passfile, secure_server.password_file, MAXPATHLEN-1);
     passfile[MAXPATHLEN-1] = '\0';
-    PM_UNLOCK(__pmLock_libpcp);
 
     if (passfile[0] == '\0') {
-	__pmNotifyErr(LOG_ERR, "Password sought but no password file given");
+	pmNotifyErr(LOG_ERR, "Password sought but no password file given");
 	return NULL;
     }
     if (retry) {
-	__pmNotifyErr(LOG_ERR, "Retry attempted during password extraction");
+	pmNotifyErr(LOG_ERR, "Retry attempted during password extraction");
 	return NULL;	/* no soup^Wretries for you */
     }
 
     sts = secure_file_contents(passfile, &password, &length);
     if (sts < 0) {
-	__pmNotifyErr(LOG_ERR, "Cannot read password file \"%s\": %s",
+	pmNotifyErr(LOG_ERR, "Cannot read password file \"%s\": %s",
 			passfile, pmErrStr(sts));
 	return NULL;
     }
@@ -193,20 +236,20 @@ __pmValidCertificate(CERTCertDBHandle *db, CERTCertificate *cert, PRTime stamp)
 }
 
 static char *
-serverdb(char *path, size_t size)
+serverdb(char *path, size_t size, char *db_method)
 {
-    int sep = __pmPathSeparator();
+    int sep = pmPathSeparator();
     char *nss_method = getenv("PCP_SECURE_DB_METHOD");
 
     if (nss_method == NULL)
-	nss_method = "sql:";
+	nss_method = db_method;
 
     /*
      * Fill in a buffer with the server NSS database specification.
      * Return a pointer to the filesystem path component - without
      * the <method>:-prefix - for other routines to work with.
      */
-    snprintf(path, size, "%s" "%c" "etc" "%c" "pki" "%c" "nssdb",
+    pmsprintf(path, size, "%s" "%c" "etc" "%c" "pki" "%c" "nssdb",
 		nss_method, sep, sep, sep);
     return path + strlen(nss_method);
 }
@@ -214,23 +257,60 @@ serverdb(char *path, size_t size)
 int
 __pmSecureServerSetup(const char *db, const char *passwd)
 {
-    const char *nickname = SECURE_SERVER_CERTIFICATE;
-    SECStatus secsts;
-    int sts;
-
-    sts = __pmInitAuthServer();
-    if (sts < 0) {
-	__pmNotifyErr(LOG_ERR, "Failed to start authenticating server");
-	return sts;
-    }
-
-    PR_Init(PR_SYSTEM_THREAD, PR_PRIORITY_NORMAL, 1);
+    return __pmSecureServerCertificateSetup(db, passwd, SECURE_SERVER_CERTIFICATE);
+}
 
+int
+__pmSecureServerCertificateSetup(const char *db, const char *passwd, const char *cert_nickname)
+{
     PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(secureserver_lock);
 
     /* Configure optional (cmdline) password file in case DB locked */
     secure_server.password_file = passwd;
+
+    /*
+     * Configure location of the NSS database with a sane default.
+     * For servers, we default to the shared (sql) system-wide database.
+     * If command line db specified, pass it directly through - allowing
+     * any old database format, at the users discretion.
+     */
+    if (db) {
+	/* shortened-buffer-size (-2) guarantees null-termination */
+	strncpy(secure_server.database_path, db, MAXPATHLEN-2);
+    }
+
+    if (cert_nickname) {
+	strncpy(secure_server.cert_nickname, cert_nickname, MAX_CERT_NAME_LENGTH-2);
+    }
+    else {
+	strncpy(secure_server.cert_nickname, SECURE_SERVER_CERTIFICATE, MAX_CERT_NAME_LENGTH-2);
+    }
+
+    PM_UNLOCK(secureserver_lock);
+    return 0;
+}
+
+int
+__pmSecureServerInit(void)
+{
+    const PRUint16 *cipher;
+    SECStatus secsts;
+    int pathSpecified;
+    int sts = 0;
+
+    PM_INIT_LOCKS();
+    PM_LOCK(secureserver_lock);
+
+    /* Only attempt this once. */
+    if (secure_server.initialized)
+	goto done;
+    secure_server.initialized = 1;
+
+    if (PR_Initialized() != PR_TRUE)
+	PR_Init(PR_SYSTEM_THREAD, PR_PRIORITY_NORMAL, 1);
+
+    /* Configure optional (cmdline) password file in case DB locked */
     PK11_SetPasswordFunc(certificate_database_password);
 
     /*
@@ -239,44 +319,63 @@ __pmSecureServerSetup(const char *db, const char *passwd)
      * If command line db specified, pass it directly through - allowing
      * any old database format, at the users discretion.
      */
-    sts = -EINVAL;
-    if (!db) {
-	char *path = serverdb(secure_server.database_path, MAXPATHLEN);
+    if (!secure_server.database_path[0]) {
+	const char *path;
+	pathSpecified = 0;
+	path = serverdb(secure_server.database_path, MAXPATHLEN, "sql:");
 
 	/* this is the default case on some platforms, so no log spam */
 	if (access(path, R_OK|X_OK) < 0) {
-	    if (pmDebug & DBG_TRACE_CONTEXT)
-		__pmNotifyErr(LOG_INFO,
-			"Cannot access system security database: %s",
-			secure_server.database_path);
-	    sts = 0;	/* not fatal - just no secure connections */
+	    if (pmDebugOptions.context)
+		pmNotifyErr(LOG_INFO,
+			      "Cannot access system security database: %s",
+			      secure_server.database_path);
+	    sts = -EOPNOTSUPP;	/* not fatal - just no secure connections */
+	    secure_server.init_failed = 1;
 	    goto done;
 	}
-    } else {
-	/* shortened-buffer-size (-2) guarantees null-termination */
-	strncpy(secure_server.database_path, db, MAXPATHLEN-2);
+    }
+    else
+	pathSpecified = 1;
+
+    /*
+     * pmproxy acts as both a client and server. Since the
+     * server init path happens first, the db previously
+     * got opened readonly.  Instead try to open RW.
+     * Fallback if there is an error.
+     */
+
+    secsts = NSS_InitReadWrite(secure_server.database_path);
+
+    if( secsts != SECSuccess )
+    	secsts = NSS_Init(secure_server.database_path);
+
+    if (secsts != SECSuccess && !pathSpecified) {
+	/* fallback, older versions of NSS do not support sql: */
+	serverdb(secure_server.database_path, MAXPATHLEN, "");
+	secsts = NSS_Init(secure_server.database_path);
     }
 
-    secsts = NSS_Init(secure_server.database_path);
     if (secsts != SECSuccess) {
-	__pmNotifyErr(LOG_ERR, "Cannot setup certificate DB (%s): %s",
+	pmNotifyErr(LOG_ERR, "Cannot setup certificate DB (%s): %s",
 			secure_server.database_path,
 			pmErrStr(__pmSecureSocketsError(PR_GetError())));
+	sts = -EOPNOTSUPP;	/* not fatal - just no secure connections */
+	secure_server.init_failed = 1;
 	goto done;
     }
 
-    secsts = NSS_SetExportPolicy();
-    if (secsts != SECSuccess) {
-	__pmNotifyErr(LOG_ERR, "Unable to set NSS export policy: %s",
-		pmErrStr(__pmSecureSocketsError(PR_GetError())));
-	goto done;
-    }
+    /* Some NSS versions don't do this correctly in NSS_SetDomesticPolicy. */
+    for (cipher = SSL_GetImplementedCiphers(); *cipher != 0; ++cipher)
+	SSL_CipherPolicySet(*cipher, SSL_ALLOWED);
 
     /* Configure SSL session cache for multi-process server, using defaults */
-    secsts = SSL_ConfigMPServerSIDCache(0, 0, 0, NULL);
+    secsts = SSL_ConfigMPServerSIDCache(1, 0, 0, NULL);
     if (secsts != SECSuccess) {
-	__pmNotifyErr(LOG_ERR, "Unable to configure SSL session ID cache: %s",
+	pmNotifyErr(LOG_ERR, "Unable to configure SSL session ID cache: %s",
 		pmErrStr(__pmSecureSocketsError(PR_GetError())));
+	sts = -EOPNOTSUPP;	/* not fatal - just no secure connections */
+	secure_server.init_failed = 1;
 	goto done;
     } else {
 	secure_server.ssl_session_cache_setup = 1;
@@ -289,7 +388,7 @@ __pmSecureServerSetup(const char *db, const char *passwd)
      */
     CERTCertList *certlist;
     CERTCertDBHandle *nssdb = CERT_GetDefaultCertDB();
-    CERTCertificate *dbcert = PK11_FindCertFromNickname(nickname, NULL);
+    CERTCertificate *dbcert = PK11_FindCertFromNickname(secure_server.cert_nickname, NULL);
 
     if (dbcert) {
 	PRTime now = PR_Now();
@@ -301,8 +400,8 @@ __pmSecureServerSetup(const char *db, const char *passwd)
 	    for (node = CERT_LIST_HEAD(certlist);
 		 !CERT_LIST_END(node, certlist);
 		 node = CERT_LIST_NEXT (node)) {
-		if (pmDebug & DBG_TRACE_CONTEXT)
-		    __pmDumpCertificate(stderr, nickname, node->cert);
+		if (pmDebugOptions.context)
+		    __pmDumpCertificate(stderr, secure_server.cert_nickname, node->cert);
 		if (!__pmValidCertificate(nssdb, node->cert, now))
 		    continue;
 		secure_server.certificate_verified = 1;
@@ -315,29 +414,39 @@ __pmSecureServerSetup(const char *db, const char *passwd)
 	    secure_server.certificate_KEA = NSS_FindCertKEAType(dbcert);
 	    secure_server.private_key = PK11_FindKeyByAnyCert(dbcert, NULL);
 	    if (!secure_server.private_key) {
-		__pmNotifyErr(LOG_ERR, "Unable to extract %s private key",
-				nickname);
+		pmNotifyErr(LOG_ERR, "Unable to extract %s private key",
+				secure_server.cert_nickname);
 		CERT_DestroyCertificate(dbcert);
 		secure_server.certificate_verified = 0;
+		sts = -EOPNOTSUPP;	/* not fatal - just no secure connections */
+		secure_server.init_failed = 1;
 		goto done;
 	    }
 	} else {
-	    __pmNotifyErr(LOG_ERR, "Unable to find a valid %s", nickname);
+	    pmNotifyErr(LOG_ERR, "Unable to find a valid %s", secure_server.cert_nickname);
 	    CERT_DestroyCertificate(dbcert);
+	    sts = -EOPNOTSUPP;	/* not fatal - just no secure connections */
+	    secure_server.init_failed = 1;
 	    goto done;
 	}
     }
 
-    if (secure_server.certificate_verified) {
-	secure_server.certificate = dbcert;
-    } else if (pmDebug & DBG_TRACE_CONTEXT) {
-	__pmNotifyErr(LOG_INFO, "No valid %s in security database: %s",
-		nickname, secure_server.database_path);
+    if (! secure_server.certificate_verified) {
+	if (pmDebugOptions.context) {
+	    pmNotifyErr(LOG_INFO, "No valid %s in security database: %s",
+			  secure_server.cert_nickname, secure_server.database_path);
+	}
+	sts = -EOPNOTSUPP;	/* not fatal - just no secure connections */
+	secure_server.init_failed = 1;
+	goto done;
     }
+
+    secure_server.certificate = dbcert;
+    secure_server.init_failed = 0;
     sts = 0;
 
 done:
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(secureserver_lock);
     return sts;
 }
 
@@ -345,7 +454,7 @@ void
 __pmSecureServerShutdown(void)
 {
     PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(secureserver_lock);
     if (secure_server.certificate) {
 	CERT_DestroyCertificate(secure_server.certificate);
 	secure_server.certificate = NULL;
@@ -358,8 +467,11 @@ __pmSecureServerShutdown(void)
 	SSL_ShutdownServerSessionIDCache();
 	secure_server.ssl_session_cache_setup = 0;
     }    
-    PM_UNLOCK(__pmLock_libpcp);
-    NSS_Shutdown();
+    if (secure_server.initialized) {
+	NSS_Shutdown();
+	secure_server.initialized = 0;
+    }
+    PM_UNLOCK(secureserver_lock);
 }
 
 static int
@@ -376,22 +488,22 @@ __pmSecureServerNegotiation(int fd, int *strength)
 	return PM_ERR_IPC;
 
     PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(secureserver_lock);
     secsts = SSL_ConfigSecureServer(sslsocket,
 			secure_server.certificate,
 			secure_server.private_key,
 			secure_server.certificate_KEA);
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(secureserver_lock);
 
     if (secsts != SECSuccess) {
-	__pmNotifyErr(LOG_ERR, "Unable to configure secure server: %s",
+	pmNotifyErr(LOG_ERR, "Unable to configure secure server: %s",
 			    pmErrStr(__pmSecureSocketsError(PR_GetError())));
 	return PM_ERR_IPC;
     }
 
     secsts = SSL_ResetHandshake(sslsocket, PR_TRUE /*server*/);
     if (secsts != SECSuccess) {
-	__pmNotifyErr(LOG_ERR, "Unable to reset secure handshake: %s",
+	pmNotifyErr(LOG_ERR, "Unable to reset secure handshake: %s",
 			    pmErrStr(__pmSecureSocketsError(PR_GetError())));
 	return PM_ERR_IPC;
     }
@@ -401,7 +513,7 @@ __pmSecureServerNegotiation(int fd, int *strength)
     timer = PR_MillisecondsToInterval(msec);
     secsts = SSL_ForceHandshakeWithTimeout(sslsocket, timer);
     if (secsts != SECSuccess) {
-	__pmNotifyErr(LOG_ERR, "Unable to force secure handshake: %s",
+	pmNotifyErr(LOG_ERR, "Unable to force secure handshake: %s",
 			    pmErrStr(__pmSecureSocketsError(PR_GetError())));
 	return PM_ERR_IPC;
     }
@@ -414,7 +526,7 @@ __pmSecureServerNegotiation(int fd, int *strength)
     return 0;
 }
 
-static void
+static int
 __pmSetUserGroupAttributes(const char *username, __pmHashCtl *attrs)
 {
     char name[32];
@@ -422,17 +534,24 @@ __pmSetUserGroupAttributes(const char *username, __pmHashCtl *attrs)
     uid_t uid;
     gid_t gid;
 
-    if (__pmGetUserIdentity(username, &uid, &gid, PM_RECOV_ERR)) {
-	snprintf(name, sizeof(name), "%u", uid);
+    if (__pmGetUserIdentity(username, &uid, &gid, PM_RECOV_ERR) == 0) {
+	pmsprintf(name, sizeof(name), "%u", uid);
 	name[sizeof(name)-1] = '\0';
 	if ((namep = strdup(name)) != NULL)
 	    __pmHashAdd(PCP_ATTR_USERID, namep, attrs);
+	else
+	    return -ENOMEM;
 
-	snprintf(name, sizeof(name), "%u", gid);
+	pmsprintf(name, sizeof(name), "%u", gid);
 	name[sizeof(name)-1] = '\0';
 	if ((namep = strdup(name)) != NULL)
 	    __pmHashAdd(PCP_ATTR_GROUPID, namep, attrs);
+	else
+	    return -ENOMEM;
+	return 0;
     }
+    pmNotifyErr(LOG_ERR, "Authenticated user %s not found\n", username);
+    return -ESRCH;
 }
 
 static int
@@ -443,23 +562,27 @@ __pmAuthServerSetAttributes(sasl_conn_t *conn, __pmHashCtl *attrs)
     int sts;
 
     sts = sasl_getprop(conn, SASL_USERNAME, &property);
-    if (sts == SASL_OK && property) {
-	__pmNotifyErr(LOG_INFO, "Successful authentication for user \"%s\"\n",
-				(char *)property);
-	username = strdup((char *)property);
+    username = (char *)property;
+    if (sts == SASL_OK && username) {
+	int len = strlen(username);
+
+	pmNotifyErr(LOG_INFO,
+			"Successful authentication for user \"%s\"\n",
+			username);
+	if ((username = strdup(username)) == NULL) {
+	    pmNoMem("__pmAuthServerSetAttributes",
+			len, PM_RECOV_ERR);
+	    return -ENOMEM;
+	}
     } else {
-	username = NULL;
+	pmNotifyErr(LOG_ERR,
+			"Authentication complete, but no username\n");
+	return -ESRCH;
     }
 
-    if (!username) {
-	sts = property ? strlen(property) : 0;
-	__pmNoMem("__pmAuthServerSetAttributes", sts, PM_RECOV_ERR);
-	sts = -ENOMEM;
-    } else {
-	sts = __pmHashAdd(PCP_ATTR_USERNAME, username, attrs);
-	__pmSetUserGroupAttributes(username, attrs);
-    }
-    return sts;
+    if ((sts = __pmHashAdd(PCP_ATTR_USERNAME, username, attrs)) < 0)
+	return sts;
+    return __pmSetUserGroupAttributes(username, attrs);
 }
 
 static int
@@ -471,7 +594,7 @@ __pmAuthServerSetProperties(sasl_conn_t *conn, int ssf)
     /* set external security strength factor */
     saslsts = sasl_setprop(conn, SASL_SSF_EXTERNAL, &ssf);
     if (saslsts != SASL_OK && saslsts != SASL_CONTINUE) {
-	__pmNotifyErr(LOG_ERR, "SASL setting external SSF to %d: %s",
+	pmNotifyErr(LOG_ERR, "SASL setting external SSF to %d: %s",
 			ssf, sasl_errstring(saslsts, NULL, NULL));
 	return __pmSecureSocketsError(saslsts);
     }
@@ -482,7 +605,7 @@ __pmAuthServerSetProperties(sasl_conn_t *conn, int ssf)
     props.max_ssf = UINT_MAX;
     saslsts = sasl_setprop(conn, SASL_SEC_PROPS, &props);
     if (saslsts != SASL_OK && saslsts != SASL_CONTINUE) {
-	__pmNotifyErr(LOG_ERR, "SASL setting security properties: %s",
+	pmNotifyErr(LOG_ERR, "SASL setting security properties: %s",
 			sasl_errstring(saslsts, NULL, NULL));
 	return __pmSecureSocketsError(saslsts);
     }
@@ -499,7 +622,7 @@ __pmAuthServerNegotiation(int fd, int ssf, __pmHashCtl *attrs)
     sasl_conn_t *sasl_conn;
     __pmPDU *pb;
 
-    if (pmDebug & DBG_TRACE_AUTH)
+    if (pmDebugOptions.auth)
 	fprintf(stderr, "__pmAuthServerNegotiation(fd=%d, ssf=%d)\n",
 		fd, ssf);
 
@@ -516,18 +639,18 @@ __pmAuthServerNegotiation(int fd, int ssf, __pmHashCtl *attrs)
                             (unsigned int *)&length,
                             &count);
     if (saslsts != SASL_OK && saslsts != SASL_CONTINUE) {
-	__pmNotifyErr(LOG_ERR, "Generating client mechanism list: %s",
+	pmNotifyErr(LOG_ERR, "Generating client mechanism list: %s",
 			sasl_errstring(saslsts, NULL, NULL));
 	return __pmSecureSocketsError(saslsts);
     }
-    if (pmDebug & DBG_TRACE_AUTH)
+    if (pmDebugOptions.auth)
 	fprintf(stderr, "__pmAuthServerNegotiation - sending mechanism list "
 		"(%d items, %d bytes): \"%s\"\n", count, length, payload);
 
     if ((sts = __pmSendAuth(fd, FROM_ANON, 0, payload, length)) < 0)
 	return sts;
 
-    if (pmDebug & DBG_TRACE_AUTH)
+    if (pmDebugOptions.auth)
 	fprintf(stderr, "__pmAuthServerNegotiation - wait for mechanism\n");
 
     sts = pinned = __pmGetPDU(fd, ANY_SIZE, TIMEOUT_DEFAULT, &pb);
@@ -552,39 +675,39 @@ __pmAuthServerNegotiation(int fd, int ssf, __pmHashCtl *attrs)
 				(unsigned int *)&length);
 	    if (saslsts != SASL_OK && saslsts != SASL_CONTINUE) {
 		sts = __pmSecureSocketsError(saslsts);
-		if (pmDebug & DBG_TRACE_AUTH)
+		if (pmDebugOptions.auth)
 		    fprintf(stderr, "sasl_server_start failed: %d (%s)\n",
 				    saslsts, pmErrStr(sts));
 	    } else {
-		if (pmDebug & DBG_TRACE_AUTH)
-		    fprintf(stderr, "sasl_server_start success: sts=%s",
+		if (pmDebugOptions.auth)
+		    fprintf(stderr, "sasl_server_start success: sts=%s\n",
 			    saslsts == SASL_CONTINUE ? "continue" : "ok");
 	    }
 	}
-    } else if (sts == PDU_ERROR) {
-	__pmDecodeError(pb, &sts);
-    } else if (sts != PM_ERR_TIMEOUT) {
-	sts = PM_ERR_IPC;
     }
+    else if (sts == PDU_ERROR)
+	__pmDecodeError(pb, &sts);
+    else if (sts != PM_ERR_TIMEOUT)
+	sts = PM_ERR_IPC;
 
-    if (pinned)
+    if (pinned > 0)
 	__pmUnpinPDUBuf(pb);
     if (sts < 0)
 	return sts;
 
-    if (pmDebug & DBG_TRACE_AUTH)
+    if (pmDebugOptions.auth)
 	fprintf(stderr, "__pmAuthServerNegotiation method negotiated\n");
 
     while (saslsts == SASL_CONTINUE) {
 	if (!payload) {
-	    __pmNotifyErr(LOG_ERR, "No SASL data to send");
+	    pmNotifyErr(LOG_ERR, "No SASL data to send");
 	    sts = -EINVAL;
 	    break;
 	}
 	if ((sts = __pmSendAuth(fd, FROM_ANON, 0, payload, length)) < 0)
 	    break;
 
-	if (pmDebug & DBG_TRACE_AUTH)
+	if (pmDebugOptions.auth)
 	    fprintf(stderr, "__pmAuthServerNegotiation awaiting response\n");
 
 	sts = pinned = __pmGetPDU(fd, ANY_SIZE, TIMEOUT_DEFAULT, &pb);
@@ -598,25 +721,25 @@ __pmAuthServerNegotiation(int fd, int ssf, __pmHashCtl *attrs)
 		    sts = __pmSecureSocketsError(sts);
 		    break;
 		}
-		if (pmDebug & DBG_TRACE_AUTH) {
+		if (pmDebugOptions.auth) {
 		    fprintf(stderr, "__pmAuthServerNegotiation"
-				    " step recv (%d bytes)", length);
+				    " step recv (%d bytes)\n", length);
 		}
 	    }
-	} else if (sts == PDU_ERROR) {
-	    __pmDecodeError(pb, &sts);
-	} else if (sts != PM_ERR_TIMEOUT) {
-	    sts = PM_ERR_IPC;
 	}
+	else if (sts == PDU_ERROR)
+	    __pmDecodeError(pb, &sts);
+	else if (sts != PM_ERR_TIMEOUT)
+	    sts = PM_ERR_IPC;
 
-	if (pinned)
+	if (pinned > 0)
 	    __pmUnpinPDUBuf(pb);
 	if (sts < 0)
 	    break;
     }
 
     if (sts < 0) {
-	if (pmDebug & DBG_TRACE_AUTH)
+	if (pmDebugOptions.auth)
 	    fprintf(stderr, "__pmAuthServerNegotiation loop failed: %d\n", sts);
 	return sts;
     }
@@ -630,9 +753,18 @@ __pmSecureServerHandshake(int fd, int flags, __pmHashCtl *attrs)
     int sts, ssf = DEFAULT_SECURITY_STRENGTH;
 
     /* protect from unsupported requests from future/oddball clients */
-    if ((flags & ~(PDU_FLAG_SECURE|PDU_FLAG_COMPRESS|PDU_FLAG_AUTH)) != 0)
+    if (flags & ~(PDU_FLAG_SECURE | PDU_FLAG_SECURE_ACK | PDU_FLAG_COMPRESS |
+		  PDU_FLAG_AUTH | PDU_FLAG_CREDS_REQD | PDU_FLAG_CONTAINER |
+		  PDU_FLAG_CERT_REQD))
 	return PM_ERR_IPC;
 
+    if (flags & PDU_FLAG_CREDS_REQD) {
+	if (__pmHashSearch(PCP_ATTR_USERID, attrs) != NULL)
+	    return 0;	/* unix domain socket */
+	else
+	    flags |= PDU_FLAG_AUTH;	/* force authentication */
+    }
+
     if ((sts = __pmSecureServerIPCFlags(fd, flags)) < 0)
 	return sts;
     if (((flags & PDU_FLAG_SECURE) != 0) &&
