@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013 Red Hat.
+ * Copyright (c) 2013-2015 Red Hat.
  * Copyright (c) 2007 Aconex.  All Rights Reserved.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
@@ -20,7 +20,8 @@
 
 #include <ctype.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
 #ifdef HAVE_STRINGS_H
 #include <strings.h>
 #endif
@@ -31,7 +32,7 @@ parseAlloc(const char *func, size_t need)
     void    *tmp;
 
     if ((tmp = malloc(need)) == NULL)
-	__pmNoMem(func, need, PM_FATAL_ERR);
+	pmNoMem(func, need, PM_FATAL_ERR);
     return tmp;
 }
 
@@ -52,7 +53,7 @@ parseError(const char *func, const char *spec, const char *point, char *msg, cha
     *q++ = '\n';
     for (p = spec; p != point; p++)
 	*q++ = isgraph((int)*p) ? ' ' : *p;
-    sprintf(q, "^ -- %s\n", msg);	/* safe */
+    pmsprintf(q, need - (q - *rslt), "^ -- %s\n", msg);
 }
 
 static void *
@@ -213,8 +214,6 @@ pmParseMetricSpec(
 		for (;;) {
 		    if (scan >= i_end) {
 			metricError(spec, scan, "closing \" expected (pmParseMetricSpec botch?)", errmsg);
-			if (msp)
-			    pmFreeMetricSpec(msp);
 			if (i_str)
 			    free(i_str);
 			return PM_ERR_GENERIC;
@@ -398,6 +397,9 @@ __pmDropHostPort(pmHostSpec *specp)
  *	pcp -h app1.aconex.com:44321@firewall.aconex.com
  *	pcp -h app1.aconex.com@firewall.aconex.com
  *	pcp -h app1.aconex.com:44321
+ *      pcp -h 192.168.122.1:44321
+ *      pcp -h [fe80::5eff:35ff:fe07:55ca]:44321,4321@[fe80::5eff:35ff:fe07:55cc]:44322
+ *      pcp -h [fe80::5eff:35ff:fe07:55ca]:44321
  *
  * Basic algorithm:
  *	look for first colon, @ or null; preceding text is hostname
@@ -406,6 +408,12 @@ __pmDropHostPort(pmHostSpec *specp)
  *	if @, start following host specification at the following character,
  *	 by returning to the start and repeating the above for the next chunk.
  * Note:
+ *      IPv6 addresses contain colons and, so, must be separated from the
+ *      rest of the spec somehow. A common notation among ipv6-enabled
+ *      applications is to enclose the address within brackets, as in
+ *      [fe80::5eff:35ff:fe07:55ca]:44321. We keep it simple, however,
+ *      and allow any host spec to be enclosed in brackets.
+ * Note:
  *	Currently only two hosts are useful, but ability to handle more than
  *	one optional proxy host is there (i.e. proxy ->proxy ->... ->pmcd),
  *	in case someone implements the pmproxy->pmproxy protocol extension.
@@ -419,11 +427,30 @@ parseHostSpec(
     char **errmsg)              /* error message */
 {
     pmHostSpec *hsp = NULL;
-    const char *s, *start;
+    const char *s, *start, *next;
     int nhosts = 0, sts = 0;
 
     for (s = start = *position; s != NULL; s++) {
-	if (*s == ':' || *s == '@' || *s == '\0' || *s == '?') {
+	/* Allow the host spec to be enclosed in brackets. */
+	if (s == start && *s == '[') {
+	    for (s++; *s != ']' && *s != '\0'; s++)
+		;
+	    if (*s != ']') {
+		hostError(spec, s, "missing closing ']' for host spec", errmsg);
+		sts = PM_ERR_GENERIC;
+		goto fail;
+	    }
+	    next = s + 1; /* past the trailing ']' */
+	    if (*next != ':' && *next != '@' && *next != '\0' && *next != '/' && *next != '?') {
+		hostError(spec, next, "extra characters after host spec", errmsg);
+		sts = PM_ERR_GENERIC;
+		goto fail;
+	    }
+	    start++; /* past the initial '[' */
+	}
+	else
+	    next = s;
+	if (*next == ':' || *next == '@' || *next == '\0' || *next == '/' || *next == '?') {
 	    if (s == *position)
 		break;
 	    else if (s == start)
@@ -433,9 +460,10 @@ parseHostSpec(
 		sts = -ENOMEM;
 		goto fail;
 	    }
+	    s = next;
 	    if (*s == ':') {
 		for (++s, start = s; s != NULL; s++) {
-		    if (*s == ',' || *s == '@' || *s == '\0' || *s == '?') {
+		    if (*s == ',' || *s == '@' || *s == '\0' || *s == '/' || *s == '?') {
 			if (s - start < 1) {
 			    hostError(spec, s, "missing port", errmsg);
 			    sts = PM_ERR_GENERIC;
@@ -446,7 +474,7 @@ parseHostSpec(
 			if (sts < 0)
 			    goto fail;
 			start = s + 1;
-			if (*s == '@' || *s == '\0' || *s == '?')
+			if (*s == '@' || *s == '\0' || *s == '/' || *s == '?')
 			    break;
 			continue;
 		    }
@@ -476,6 +504,70 @@ fail:
     return sts;
 }
 
+/*
+ * Parse a socket path.
+ * Accept anything up to, but not including the first ':', '?' or the end of the spec.
+ */
+static int      /* 0 -> ok, PM_ERR_GENERIC -> error message is set */
+parseSocketPath(
+    const char *spec,
+    char **position,            /* parse this string, return end char */
+    pmHostSpec **rslt)          /* result allocated and returned here */
+{
+    pmHostSpec *hsp = NULL;
+    const char *s, *start, *path;
+    char absolute_path[MAXPATHLEN];
+    size_t len;
+    int nhosts = 0, delimited = 0;
+
+    /* Scan to the end of the string or to the first delimiter. */
+    for (s = start = *position; s != NULL; s++) {
+	if (*s == '\0')
+	    break;
+	if (*s == ':' || *s == '?') {
+	    delimited = 1;
+	    break;
+	}
+    }
+
+    /* If the path is empty, then provide the default. */
+    if (s == start || (delimited && s == start + 1)) {
+	path = __pmPMCDLocalSocketDefault();
+	len = strlen(path);
+    }
+    else {
+	path = start;
+	len = s - start;
+	if (len >= MAXPATHLEN)
+	    len = MAXPATHLEN - 1;
+    }
+
+    /*
+     * Make sure that the path is absolute. parseProtocolSpec() removes the
+     * (optional) "//" from "local://some/path".
+     */
+    if (*path != pmPathSeparator()) {
+	absolute_path[0] = pmPathSeparator();
+	strncpy(absolute_path + 1, path, len);
+	if (len < sizeof(absolute_path)-1)
+	    absolute_path[++len] = '\0';
+	absolute_path[sizeof(absolute_path)-1] = '\0';
+	path = absolute_path;
+    }
+
+    /* Add the path as the only member of the host list. */
+    hsp = hostAdd(hsp, &nhosts, path, len);
+    if (hsp == NULL) {
+	__pmFreeHostSpec(hsp, nhosts);
+	*rslt = NULL;
+	return -ENOMEM;
+    }
+
+    *position = (char *)s;
+    *rslt = hsp;
+    return 0;
+}
+
 int
 __pmParseHostSpec(
     const char *spec,           /* parse this string */
@@ -499,34 +591,72 @@ __pmParseHostSpec(
     return PM_ERR_GENERIC;
 }
 
-int
-__pmUnparseHostSpec(pmHostSpec *hostp, int count, char *string, size_t size)
+static int
+unparseHostSpec(pmHostSpec *hostp, int count, char *string, size_t size, int prefix)
 {
     int off = 0, len = size;	/* offset in string and space remaining */
     int i, j, sts;
 
     for (i = 0; i < count; i++) {
 	if (i > 0) {
-	    if ((sts = snprintf(string + off, len, "@")) >= size)
-		return -E2BIG;
+	    if ((sts = pmsprintf(string + off, len, "@")) >= size) {
+		off = -E2BIG;
+		goto done;
+	    }
 	    len -= sts; off += sts;
 	}
 
-	if ((sts = snprintf(string + off, len, "%s", hostp[i].name)) >= size)
-	    return -E2BIG;
+	if (prefix && hostp[i].nports == PM_HOST_SPEC_NPORTS_LOCAL) {
+	    if ((sts = pmsprintf(string + off, len, "local:/%s", hostp[i].name + 1)) >= size) {
+		off = -E2BIG;
+		goto done;
+	    }
+	}
+	else if (prefix && hostp[i].nports == PM_HOST_SPEC_NPORTS_UNIX) {
+	    if ((sts = pmsprintf(string + off, len, "unix:/%s", hostp[i].name + 1)) >= size) {
+		off = -E2BIG;
+		goto done;
+	    }
+	}
+	else {
+	    if ((sts = pmsprintf(string + off, len, "%s", hostp[i].name)) >= size) {
+		off = -E2BIG;
+		goto done;
+	    }
+	}
 	len -= sts; off += sts;
 
 	for (j = 0; j < hostp[i].nports; j++) {
-	    if ((sts = snprintf(string + off, len,
+	    if ((sts = pmsprintf(string + off, len,
 			    "%c%u", (j == 0) ? ':' : ',',
-			    hostp[i].ports[j])) >= size)
-		return -E2BIG;
+			    hostp[i].ports[j])) >= size) {
+		off = -E2BIG;
+		goto done;
+	    }
 	    len -= sts; off += sts;
 	}
     }
+
+done:
+    if (pmDebugOptions.context) {
+	fprintf(stderr, "__pmUnparseHostSpec([name=%s ports=%p nport=%d], count=%d, ...) -> ", hostp->name, hostp->ports, hostp->nports, count);
+	if (off < 0) {
+	    char	errmsg[PM_MAXERRMSGLEN];
+	    pmErrStr_r(off, errmsg, sizeof(errmsg));
+	    fprintf(stderr, "%s\n", errmsg);
+	}
+	else
+	    fprintf(stderr, "%d \"%s\"\n", off, string);
+    }
     return off;
 }
 
+int
+__pmUnparseHostSpec(pmHostSpec *hostp, int count, char *string, size_t size)
+{
+    return unparseHostSpec(hostp, count, string, size, 1);
+}
+
 void
 __pmFreeHostSpec(pmHostSpec *specp, int count)
 {
@@ -574,13 +704,22 @@ __pmFreeHostAttrsSpec(pmHostSpec *hosts, int count, __pmHashCtl *attrs)
 #define PCPS_PROTOCOL_PREFIX	PCPS_PROTOCOL_NAME ":"
 #define PCPS_PROTOCOL_SIZE	(sizeof(PCPS_PROTOCOL_NAME)-1)
 #define PCPS_PROTOCOL_PREFIXSZ	(sizeof(PCPS_PROTOCOL_PREFIX)-1)
+#define LOCAL_PROTOCOL_NAME	"local"
+#define LOCAL_PROTOCOL_PREFIX	LOCAL_PROTOCOL_NAME ":"
+#define LOCAL_PROTOCOL_SIZE	(sizeof(LOCAL_PROTOCOL_NAME)-1)
+#define LOCAL_PROTOCOL_PREFIXSZ	(sizeof(LOCAL_PROTOCOL_PREFIX)-1)
+#define UNIX_PROTOCOL_NAME	"unix"
+#define UNIX_PROTOCOL_PREFIX	UNIX_PROTOCOL_NAME ":"
+#define UNIX_PROTOCOL_SIZE	(sizeof(UNIX_PROTOCOL_NAME)-1)
+#define UNIX_PROTOCOL_PREFIXSZ	(sizeof(UNIX_PROTOCOL_PREFIX)-1)
 
 static int
 parseProtocolSpec(
     const char *spec,           /* the original, complete string to parse */
     char **position,
     int *attribute,
-    char **value)
+    char **value,
+    char **errmsg)
 {
     char *protocol = NULL;
     char *s = *position;
@@ -589,9 +728,19 @@ parseProtocolSpec(
     if (strncmp(s, PCP_PROTOCOL_PREFIX, PCP_PROTOCOL_PREFIXSZ) == 0) {
 	protocol = PCP_PROTOCOL_NAME;
 	s += PCP_PROTOCOL_PREFIXSZ;
+	*attribute = PCP_ATTR_PROTOCOL;
     } else if (strncmp(s, PCPS_PROTOCOL_PREFIX, PCPS_PROTOCOL_PREFIXSZ) == 0) {
 	protocol = PCPS_PROTOCOL_NAME;
 	s += PCPS_PROTOCOL_PREFIXSZ;
+	*attribute = PCP_ATTR_PROTOCOL;
+    } else if (strncmp(s, LOCAL_PROTOCOL_PREFIX, LOCAL_PROTOCOL_PREFIXSZ) == 0) {
+	protocol = LOCAL_PROTOCOL_NAME;
+	s += LOCAL_PROTOCOL_PREFIXSZ;
+	*attribute = PCP_ATTR_LOCAL;
+    } else if (strncmp(s, UNIX_PROTOCOL_PREFIX, UNIX_PROTOCOL_PREFIXSZ) == 0) {
+	protocol = UNIX_PROTOCOL_NAME;
+	s += UNIX_PROTOCOL_PREFIXSZ;
+	*attribute = PCP_ATTR_UNIXSOCK;
     }
 
     /* optionally skip over slash-delimiters */
@@ -600,7 +749,6 @@ parseProtocolSpec(
 	    s++;
 	if ((*value = strdup(protocol)) == NULL)
 	    return -ENOMEM;
-	*attribute = PCP_ATTR_PROTOCOL;
     } else {
 	*value = NULL;
 	*attribute = PCP_ATTR_NONE;
@@ -627,9 +775,6 @@ __pmLookupAttrKey(const char *attribute, size_t size)
 	(size == sizeof("username") &&
 	strncmp(attribute, "username", size) == 0))
 	return PCP_ATTR_USERNAME;
-    if (size == sizeof("authname") &&
-	strncmp(attribute, "authname", size) == 0)
-	return PCP_ATTR_AUTHNAME;
     if (size == sizeof("realm") &&
 	strncmp(attribute, "realm", size) == 0)
 	return PCP_ATTR_REALM;
@@ -648,6 +793,9 @@ __pmLookupAttrKey(const char *attribute, size_t size)
 	(size == sizeof("unixsock") &&
 	strncmp(attribute, "unixsock", size) == 0))
 	return PCP_ATTR_UNIXSOCK;
+    if ((size == sizeof("local") &&
+	 strncmp(attribute, "local", size) == 0))
+	return PCP_ATTR_LOCAL;
     if ((size == sizeof("uid") &&
 	strncmp(attribute, "uid", size) == 0) ||
 	(size == sizeof("userid") &&
@@ -658,9 +806,20 @@ __pmLookupAttrKey(const char *attribute, size_t size)
 	(size == sizeof("groupid") &&
 	strncmp(attribute, "groupid", size) == 0))
 	return PCP_ATTR_GROUPID;
+    if ((size == sizeof("pid") &&
+	strncmp(attribute, "pid", size) == 0) ||
+	(size == sizeof("processid") &&
+	strncmp(attribute, "processid", size) == 0))
+	return PCP_ATTR_PROCESSID;
     if (size == sizeof("secure") &&
 	strncmp(attribute, "secure", size) == 0)
 	return PCP_ATTR_SECURE;
+    if (size == sizeof("container") &&
+	strncmp(attribute, "container", size) == 0)
+	return PCP_ATTR_CONTAINER;
+    if (size == sizeof("exclusive") &&
+	strncmp(attribute, "exclusive", size) == 0)	/* deprecated */
+	return PCP_ATTR_EXCLUSIVE;
     return PCP_ATTR_NONE;
 }
 
@@ -688,8 +847,8 @@ parseAttributeSpec(
 
     for (s = start = *position; s != NULL; s++) {
 	/* parse: foo=bar&moo&goo=blah ... go! */
-	if (*s == '\0' || *s == '&') {
-	    if (*s == '\0' && s == start)
+	if (*s == '\0' || *s == '/' || *s == '&') {
+	    if ((*s == '\0' || *s == '/') && s == start)
 		break;
 	    len = v ? (v - start - 1) : (s - start);
 	    buflen = (len < sizeof(buffer)-1) ? len : sizeof(buffer)-1;
@@ -709,7 +868,7 @@ parseAttributeSpec(
 		}
 	    }
 	    v = NULL;
-	    if (*s == '\0')
+	    if (*s == '\0' || *s == '/')
 		break;
 	    start = s + 1;	/* start of attribute name */
 	    continue;
@@ -734,6 +893,8 @@ fail:
  *
  * pcp://oss.sgi.com:45892?user=otto&pass=blotto&compress=true
  * pcps://oss.sgi.com@proxy.org:45893?user=jimbo&pass=jones&compress=true
+ * local://path/to/socket:?user=jimbo&pass=jones
+ * unix://path/to/socket
  */
 int
 __pmParseHostAttrsSpec(
@@ -746,17 +907,30 @@ __pmParseHostAttrsSpec(
     char *value = NULL, *s = (char *)spec;
     int sts, attr;
 
+    *count = 0;			/* ensure this initialised for fail: code */
+
     /* parse optional protocol section */
-    if ((sts = parseProtocolSpec(spec, &s, &attr, &value)) < 0)
+    if ((sts = parseProtocolSpec(spec, &s, &attr, &value, errmsg)) < 0)
 	return sts;
 
-    if ((sts = parseHostSpec(spec, &s, host, count, errmsg)) < 0)
-	goto fail;
+    if (attr == PCP_ATTR_LOCAL || attr == PCP_ATTR_UNIXSOCK) {
+	/* We are looking for a socket path. */
+	if ((sts = parseSocketPath(spec, &s, host)) < 0)
+	    goto fail;
+	*count = 1;
+	host[0]->nports = (attr == PCP_ATTR_LOCAL) ?
+	    PM_HOST_SPEC_NPORTS_LOCAL : PM_HOST_SPEC_NPORTS_UNIX;
+    }
+    else {
+	/* We are looking for a host spec. */
+	if ((sts = parseHostSpec(spec, &s, host, count, errmsg)) < 0)
+	    goto fail;
+    }
 
     /* skip over an attributes delimiter */
     if (*s == '?') {
 	s++;	/* optionally skip over the question mark */
-    } else if (*s != '\0') {
+    } else if (*s != '\0' && *s != '/') {
 	hostError(spec, s, "unexpected terminal character", errmsg);
 	sts = PM_ERR_GENERIC;
 	goto fail;
@@ -789,29 +963,37 @@ __pmAttrKeyStr_r(__pmAttrKey key, char *string, size_t size)
 {
     switch (key) {
     case PCP_ATTR_PROTOCOL:
-	return snprintf(string, size, "protocol");
+	return pmsprintf(string, size, "protocol");
     case PCP_ATTR_COMPRESS:
-	return snprintf(string, size, "compress");
+	return pmsprintf(string, size, "compress");
     case PCP_ATTR_USERAUTH:
-	return snprintf(string, size, "userauth");
+	return pmsprintf(string, size, "userauth");
     case PCP_ATTR_USERNAME:
-	return snprintf(string, size, "username");
+	return pmsprintf(string, size, "username");
     case PCP_ATTR_AUTHNAME:
-	return snprintf(string, size, "authname");
+	return pmsprintf(string, size, "authname");
     case PCP_ATTR_PASSWORD:
-	return snprintf(string, size, "password");
+	return pmsprintf(string, size, "password");
     case PCP_ATTR_METHOD:
-	return snprintf(string, size, "method");
+	return pmsprintf(string, size, "method");
     case PCP_ATTR_REALM:
-	return snprintf(string, size, "realm");
+	return pmsprintf(string, size, "realm");
     case PCP_ATTR_SECURE:
-	return snprintf(string, size, "secure");
+	return pmsprintf(string, size, "secure");
     case PCP_ATTR_UNIXSOCK:
-	return snprintf(string, size, "unixsock");
+	return pmsprintf(string, size, "unixsock");
+    case PCP_ATTR_LOCAL:
+	return pmsprintf(string, size, "local");
     case PCP_ATTR_USERID:
-	return snprintf(string, size, "userid");
+	return pmsprintf(string, size, "userid");
     case PCP_ATTR_GROUPID:
-	return snprintf(string, size, "groupid");
+	return pmsprintf(string, size, "groupid");
+    case PCP_ATTR_PROCESSID:
+	return pmsprintf(string, size, "processid");
+    case PCP_ATTR_CONTAINER:
+	return pmsprintf(string, size, "container");
+    case PCP_ATTR_EXCLUSIVE:
+	return pmsprintf(string, size, "exclusive");	/* deprecated */
     case PCP_ATTR_NONE:
     default:
 	break;
@@ -831,19 +1013,22 @@ __pmAttrStr_r(__pmAttrKey key, const char *data, char *string, size_t size)
     switch (key) {
     case PCP_ATTR_PROTOCOL:
     case PCP_ATTR_USERNAME:
-    case PCP_ATTR_AUTHNAME:
     case PCP_ATTR_PASSWORD:
     case PCP_ATTR_METHOD:
     case PCP_ATTR_REALM:
     case PCP_ATTR_SECURE:
     case PCP_ATTR_USERID:
     case PCP_ATTR_GROUPID:
-	return snprintf(string, size, "%s=%s", name, data ? data : "");
+    case PCP_ATTR_PROCESSID:
+    case PCP_ATTR_CONTAINER:
+	return pmsprintf(string, size, "%s=%s", name, data ? data : "");
 
     case PCP_ATTR_UNIXSOCK:
+    case PCP_ATTR_LOCAL:
     case PCP_ATTR_COMPRESS:
     case PCP_ATTR_USERAUTH:
-	return snprintf(string, size, "%s", name);
+    case PCP_ATTR_EXCLUSIVE:	/* deprecated */
+	return pmsprintf(string, size, "%s", name);
 
     case PCP_ATTR_NONE:
     default:
@@ -865,12 +1050,22 @@ __pmUnparseHostAttrsSpec(
     int sts, first;
 
     if ((node = __pmHashSearch(PCP_ATTR_PROTOCOL, attrs)) != NULL) {
-	if ((sts = snprintf(string, len, "%s://", (char *)node->data)) >= size)
+	if ((sts = pmsprintf(string, len, "%s://", (char *)node->data)) >= len)
+	    return -E2BIG;
+	len -= sts; off += sts;
+    }
+    else if (__pmHashSearch(PCP_ATTR_UNIXSOCK, attrs) != NULL) {
+	if ((sts = pmsprintf(string, len, "unix:/")) >= len)
+	    return -E2BIG;
+	len -= sts; off += sts;
+    }
+    else if (__pmHashSearch(PCP_ATTR_LOCAL, attrs) != NULL) {
+	if ((sts = pmsprintf(string, len, "local:/")) >= len)
 	    return -E2BIG;
 	len -= sts; off += sts;
     }
 
-    if ((sts = __pmUnparseHostSpec(hosts, count, string + off, len)) >= size)
+    if ((sts = unparseHostSpec(hosts, count, string + off, len, 0)) >= len)
 	return sts;
     len -= sts; off += sts;
 
@@ -878,14 +1073,15 @@ __pmUnparseHostAttrsSpec(
     for (node = __pmHashWalk(attrs, PM_HASH_WALK_START);
 	 node != NULL;
 	 node = __pmHashWalk(attrs, PM_HASH_WALK_NEXT)) {
-	if (node->key == PCP_ATTR_PROTOCOL)
+	if (node->key == PCP_ATTR_PROTOCOL ||
+	    node->key == PCP_ATTR_UNIXSOCK || node->key == PCP_ATTR_LOCAL)
 	    continue;
-	if ((sts = snprintf(string + off, len, "%c", first ? '?' : '&')) >= size)
+	if ((sts = pmsprintf(string + off, len, "%c", first ? '?' : '&')) >= len)
 	    return -E2BIG;
 	len -= sts; off += sts;
 	first = 0;
 
-	if ((sts = unparseAttribute(node, string + off, len)) >= size)
+	if ((sts = unparseAttribute(node, string + off, len)) >= len)
 	    return -E2BIG;
 	len -= sts; off += sts;
     }
