@@ -14,7 +14,7 @@
  * Thread-safe notes
  *
  * These routines manipulate the environment and call lots of routines
- * like localtime(), asctime(), gmtime(), getenv(), putenv() ... all of
+ * like localtime(), asctime(), gmtime(), getenv(), setenv() ... all of
  * which are not thread-safe.
  *
  * We use the big lock to prevent concurrent execution.
@@ -25,13 +25,13 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
 
 static char	*envtz;			/* buffer in env */
 static int	envtzlen;
 
-static char	*savetz;		/* real $TZ from env */
-static char	**savetzp;
+static char	*savetz = NULL;		/* real $TZ from env */
 
 static int	nzone;			/* table of zones */
 static int	curzone = -1;
@@ -43,33 +43,50 @@ static char	**zone;
 
 extern char **_environ;
 
+/*
+ * THREADSAFE note - always called with __pmLock_extcall already acquired
+ */
 static void
 _pushTZ(void)
 {
-    char	**p;
+    char	*p;
 
-    savetzp = NULL;
-    for (p = _environ; *p != NULL; p++) {
-	if (strncmp(*p, "TZ=", 3) == 0) {
-	    savetz = *p;
-	    *p = envtz;
-	    savetzp = p;
-	    break;
-	}
+    PM_ASSERT_IS_LOCKED(__pmLock_extcall);
+
+    if (savetz) {
+	/* don't need previous saved value any more */
+	free(savetz);
+	savetz = NULL;
     }
-    if (*p == NULL)
-	putenv(envtz);
+
+    p = getenv("TZ");		/* THREADSAFE */
+    if (p != NULL)
+	/* save current value */
+	savetz = strdup(p);
+
+    setenv("TZ", envtz, 1);		/* THREADSAFE */
     tzset();
+
+    if (pmDebugOptions.context && pmDebugOptions.desperate)
+	fprintf(stderr, "_pushTZ() envtz=\"%s\" savetz=\"%s\" after TZ=\"%s\"\n", envtz, savetz, getenv("TZ"));		/* THREADSAFE */
 }
 
+/*
+ * THREADSAFE note - always called with __pmLock_extcall already acquired
+ */
 static void
 _popTZ(void)
 {
-    if (savetzp != NULL)
-	*savetzp = savetz;
+    PM_ASSERT_IS_LOCKED(__pmLock_extcall);
+
+    if (savetz != NULL)
+	setenv("TZ", savetz, 1);	/* THREADSAFE */
     else
-	putenv("TZ=");
+	unsetenv("TZ");			/* THREADSAFE */
     tzset();
+
+    if (pmDebugOptions.context && pmDebugOptions.desperate)
+	fprintf(stderr, "_popTZ() savetz=\"%s\" after TZ=\"%s\"\n", savetz, getenv("TZ"));		/* THREADSAFE */
 }
 
 /*
@@ -77,8 +94,10 @@ _popTZ(void)
  * timezone part is not more than PM_TZ_MAXLEN bytes
  * Assumes TZ= is in the start of tzbuffer and this is not touched.
  * And finally set TZ in the environment.
+ *
+ * THREADSAFE note - always called with __pmLock_extcall already acquired
  */
-void
+static void
 __pmSquashTZ(char *tzbuffer)
 {
     time_t	now = time(NULL);
@@ -87,11 +106,10 @@ __pmSquashTZ(char *tzbuffer)
 #ifndef IS_MINGW
     time_t	offset; 
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_ASSERT_IS_LOCKED(__pmLock_extcall);
 
     tzset();
-    t = localtime(&now);
+    t = localtime(&now);		/* THREADSAFE */
 
 #ifdef HAVE_ALTZONE
     offset = (t->tm_isdst > 0) ? altzone : timezone;
@@ -105,7 +123,7 @@ __pmSquashTZ(char *tzbuffer)
     }
 #else
     {
-	struct tm	*gmt = gmtime(&now);
+	struct tm	*gmt = gmtime(&now);		/* THREADSAFE */
 	offset = (gmt->tm_hour - t->tm_hour) * 3600 +
 		 (gmt->tm_min - t->tm_min) * 60;
     }
@@ -121,21 +139,20 @@ __pmSquashTZ(char *tzbuffer)
 	if (mins == 0) {
 	    /* -3 for +HH in worst case */
 	    if (len > PM_TZ_MAXLEN-3) len = PM_TZ_MAXLEN-3;
-	    snprintf(tzbuffer+3, PM_TZ_MAXLEN, "%*.*s%+d", len, len, tzn, hours);
+	    pmsprintf(tzbuffer, PM_TZ_MAXLEN, "%*.*s%+d", len, len, tzn, hours);
 	}
 	else {
 	    /* -6 for +HH:MM in worst case */
 	    if (len > PM_TZ_MAXLEN-6) len = PM_TZ_MAXLEN-6;
-	    snprintf(tzbuffer+3, PM_TZ_MAXLEN, "%*.*s%+d:%02d", len, len, tzn, hours, mins);
+	    pmsprintf(tzbuffer, PM_TZ_MAXLEN, "%*.*s%+d:%02d", len, len, tzn, hours, mins);
 	}
     }
     else {
-	strncpy(tzbuffer+3, tzn, PM_TZ_MAXLEN);
-	tzbuffer[PM_TZ_MAXLEN+4-1] = '\0';
+	strncpy(tzbuffer, tzn, PM_TZ_MAXLEN);
+	tzbuffer[PM_TZ_MAXLEN] = '\0';
     }
-    putenv(tzbuffer);
+    setenv("TZ", tzbuffer, 1);		/* THREADSAFE */
 
-    PM_UNLOCK(__pmLock_libpcp);
     return;
 
 #else	/* IS_MINGW */
@@ -160,9 +177,6 @@ __pmSquashTZ(char *tzbuffer)
     wchar_t *src;
     div_t d;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
-
     GetTimeZoneInformation(&tz);
     dst = cp = tzbuf;
     off = tzoff;
@@ -176,12 +190,14 @@ __pmSquashTZ(char *tzbuffer)
     }
     else
 	cp = dst;
-    d = div(tz.Bias+tz.StandardBias, 60);
-    sprintf(cp, "%d", d.quot);
-    sprintf(off, "%d", d.quot);
+    d = div(tz.Bias + tz.StandardBias, 60);
+    pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), "%d", d.quot);
+    pmsprintf(off, sizeof(tzoff) - (off - tzoff), "%d", d.quot);
     if (d.rem) {
-	sprintf(cp=strchr(cp, 0), ":%d", abs(d.rem));
-	sprintf(off=strchr(off, 0), ":%d", abs(d.rem));
+	cp = strchr(cp, 0);
+	pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), ":%d", abs(d.rem));
+	off = strchr(off, 0);
+	pmsprintf(off, sizeof(tzoff) - (off - tzoff), ":%d", abs(d.rem));
     }
     if (tz.StandardDate.wMonth) {
 	cp = strchr(cp, 0);
@@ -197,46 +213,55 @@ __pmSquashTZ(char *tzbuffer)
 	else
 	    cp = dst;
 	d = div(tz.Bias+tz.DaylightBias, 60);
-	sprintf(cp, "%d", d.quot);
-	if (d.rem)
-	    sprintf(cp=strchr(cp, 0), ":%d", abs(d.rem));
+	pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), "%d", d.quot);
+	if (d.rem) {
+	    cp = strchr(cp, 0);
+	    pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), ":%d", abs(d.rem));
+	}
 	cp = strchr(cp, 0);
-	sprintf(cp=strchr(cp, 0), ",M%d.%d.%d/%d",
+	cp = strchr(cp, 0);
+	pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), ",M%d.%d.%d/%d",
 		tz.DaylightDate.wMonth,
 		tz.DaylightDate.wDay,
 		tz.DaylightDate.wDayOfWeek,
 		tz.DaylightDate.wHour);
-	if (tz.DaylightDate.wMinute || tz.DaylightDate.wSecond)
-	    sprintf(cp=strchr(cp, 0), ":%d", tz.DaylightDate.wMinute);
-	if (tz.DaylightDate.wSecond)
-	    sprintf(cp=strchr(cp, 0), ":%d", tz.DaylightDate.wSecond);
+	if (tz.DaylightDate.wMinute || tz.DaylightDate.wSecond) {
+	    cp = strchr(cp, 0);
+	    pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), ":%d", tz.DaylightDate.wMinute);
+	}
+	if (tz.DaylightDate.wSecond) {
+	    cp = strchr(cp, 0);
+	    pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), ":%d", tz.DaylightDate.wSecond);
+	}
 	cp = strchr(cp, 0);
-	sprintf(cp=strchr(cp, 0), ",M%d.%d.%d/%d",
+	cp = strchr(cp, 0);
+	pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), ",M%d.%d.%d/%d",
 		tz.StandardDate.wMonth,
 		tz.StandardDate.wDay,
 		tz.StandardDate.wDayOfWeek,
 		tz.StandardDate.wHour);
-	if (tz.StandardDate.wMinute || tz.StandardDate.wSecond)
-	    sprintf(cp=strchr(cp, 0), ":%d", tz.StandardDate.wMinute);
-	if (tz.StandardDate.wSecond)
-	    sprintf(cp=strchr(cp, 0), ":%d", tz.StandardDate.wSecond);
+	if (tz.StandardDate.wMinute || tz.StandardDate.wSecond) {
+	    cp = strchr(cp, 0);
+	    pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), ":%d", tz.StandardDate.wMinute);
+	}
+	if (tz.StandardDate.wSecond) {
+	    cp = strchr(cp, 0);
+	    pmsprintf(cp, sizeof(tzbuf) - (cp - tzbuf), ":%d", tz.StandardDate.wSecond);
+	}
     }
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_TIMECONTROL)
+    if (pmDebugOptions.timecontrol)
 	fprintf(stderr, "Win32 TZ=%s\n", tzbuf);
-#endif
 
-    snprintf(tzbuffer+3, PM_TZ_MAXLEN, "%s", tzbuf);
-    putenv(tzbuffer);
+    pmsprintf(tzbuffer, PM_TZ_MAXLEN, "%s", tzbuf);
+    setenv("TZ", tzbuffer, 1);		/* THREADSAFE */
 
     tzset();
-    t = localtime(&now);
+    t = localtime(&now);		/* THREADSAFE */
     tzn = tzname[(t->tm_isdst > 0)];
 
-    snprintf(tzbuffer+3, PM_TZ_MAXLEN, "%s%s", tzn, tzoff);
-    putenv(tzbuffer);
+    pmsprintf(tzbuffer, PM_TZ_MAXLEN, "%s%s", tzn, tzoff);
+    setenv("TZ", tzbuffer, 1);		/* THREADSAFE */
 
-    PM_UNLOCK(__pmLock_libpcp);
     return;
 #endif
 }
@@ -250,22 +275,21 @@ __pmTimezone(void)
     static char *tzbuffer = NULL;
     char *tz;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(__pmLock_extcall);
 
-    tz = getenv("TZ");
+    tz = getenv("TZ");		/* THREADSAFE */
 
     if (tzbuffer == NULL) {
 	/*
-	 * size is PM_TZ_MAXLEN + length of "TZ=" + null byte terminator
+	 * size is PM_TZ_MAXLEN + null byte terminator
 	 */
-	tzbuffer = (char *)malloc(PM_TZ_MAXLEN+4);
+	tzbuffer = (char *)malloc(PM_TZ_MAXLEN+1);
 	if (tzbuffer == NULL) {
 	    /* not much we can do here ... */
-	    PM_UNLOCK(__pmLock_libpcp);
+	    PM_UNLOCK(__pmLock_extcall);
 	    return NULL;
 	}
-	strcpy(tzbuffer, "TZ=");
+	tzbuffer[0] = '\0';
     }
 
     if (tz == NULL || tz[0] == ':') {
@@ -273,7 +297,7 @@ __pmTimezone(void)
 	 * it's an Olson-style TZ and it does not supported on all IRIXes, so
 	 * squash it into a simple one (pv#788431). */
 	__pmSquashTZ(tzbuffer);
-	tz = &tzbuffer[3];
+	tz = tzbuffer;
     } else if (strlen(tz) > PM_TZ_MAXLEN) {
 	/* TZ is too long to fit into the internal PCP timezone structs
 	 * let's try to sqash it a bit */
@@ -282,7 +306,7 @@ __pmTimezone(void)
 	if ((tb = strdup(tz)) == NULL) {
 	    /* sorry state of affairs, go squash w/out copying buffer */
 	    __pmSquashTZ(tzbuffer);
-	    tz = &tzbuffer[3];
+	    tz = tzbuffer;
 	}
 	else {
 	    char *ptz = tz;
@@ -299,18 +323,18 @@ __pmTimezone(void)
 	    if (strlen(tb) > PM_TZ_MAXLEN) { 
 		/* Still too long - let's pretend it's Olson */
 		__pmSquashTZ(tzbuffer);
-		tz = &tzbuffer[3];
+		tz = tzbuffer;
 	    } else {
-		strcpy(tzbuffer+3, tb);
-		putenv(tzbuffer);
-		tz = tzbuffer+3;
+		strcpy(tzbuffer, tb);
+		setenv("TZ", tzbuffer, 1);		/* THREADSAFE */
+		tz = tzbuffer;
 	    }
 
 	    free(tb);
 	}
     }
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(__pmLock_extcall);
     return tz;
 }
 
@@ -320,33 +344,27 @@ __pmTimezone(void)
 char *
 __pmTimezone_r(char *buf, int buflen)
 {
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
     strcpy(buf, __pmTimezone());
-    PM_UNLOCK(__pmLock_libpcp);
     return buf;
 }
 
 int
 pmUseZone(const int tz_handle)
 {
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(__pmLock_extcall);
 
     if (tz_handle < 0 || tz_handle >= nzone) {
-	PM_UNLOCK(__pmLock_libpcp);
+	PM_UNLOCK(__pmLock_extcall);
 	return -1;
     }
 
     curzone = tz_handle;
-    strcpy(&envtz[3], zone[curzone]);
+    strcpy(envtz, zone[curzone]);
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_CONTEXT)
+    if (pmDebugOptions.context)
 	fprintf(stderr, "pmUseZone(%d) tz=%s\n", curzone, zone[curzone]);
-#endif
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(__pmLock_extcall);
     return 0;
 }
 
@@ -357,8 +375,7 @@ pmNewZone(const char *tz)
     int		hack = 0;
     int		sts;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(__pmLock_extcall);
 
     len = (int)strlen(tz);
     if (len == 3) {
@@ -380,25 +397,25 @@ pmNewZone(const char *tz)
 	envtz = (char *)malloc(envtzlen);
 	strcpy(envtz, "TZ=");
     }
-    strcpy(&envtz[3], tz);
+    strcpy(envtz, tz);
     if (hack)
 	/* see above */
-	strcpy(&envtz[6], "+0");
+	strcat(envtz, "+0");
 
     curzone = nzone++;
     zone = (char **)realloc(zone, nzone * sizeof(char *));
     if (zone == NULL) {
-	__pmNoMem("pmNewZone", nzone * sizeof(char *), PM_FATAL_ERR);
+	PM_UNLOCK(__pmLock_extcall);
+	pmNoMem("pmNewZone", nzone * sizeof(char *), PM_FATAL_ERR);
+	/* NOTREACHED */
     }
-    zone[curzone] = strdup(&envtz[3]);
+    zone[curzone] = strdup(envtz);
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_CONTEXT)
+    if (pmDebugOptions.context)
 	fprintf(stderr, "pmNewZone(%s) -> %d\n", zone[curzone], curzone);
-#endif
     sts = curzone;
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(__pmLock_extcall);
     return sts;
 }
 
@@ -408,7 +425,8 @@ pmNewContextZone(void)
     __pmContext	*ctxp;
     int		sts;
 
-    if ((ctxp = __pmHandleToPtr(pmWhichContext())) == NULL)
+    ctxp = __pmHandleToPtr(pmWhichContext());
+    if (ctxp == NULL) 
 	return PM_ERR_NOCONTEXT;
 
     if (ctxp->c_type == PM_CONTEXT_ARCHIVE) {
@@ -450,13 +468,12 @@ pmCtime(const time_t *clock, char *buf)
 #if !defined(IS_SOLARIS) && !defined(IS_MINGW)
     struct tm	tbuf;
 #endif
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
 
+    PM_LOCK(__pmLock_extcall);
     if (curzone >= 0) {
 	_pushTZ();
 #if defined(IS_SOLARIS) || defined(IS_MINGW)
-	strcpy(buf, asctime(localtime(clock)));
+	strcpy(buf, asctime(localtime(clock)));		/* THREADSAFE */
 #else
 	asctime_r(localtime_r(clock, &tbuf), buf);
 #endif
@@ -464,13 +481,13 @@ pmCtime(const time_t *clock, char *buf)
     }
     else {
 #if defined(IS_SOLARIS) || defined(IS_MINGW)
-	strcpy(buf, asctime(localtime(clock)));
+	strcpy(buf, asctime(localtime(clock)));		/* THREADSAFE */
 #else
 	asctime_r(localtime_r(clock, &tbuf), buf);
 #endif
     }
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(__pmLock_extcall);
     return buf;
 }
 
@@ -481,13 +498,12 @@ pmLocaltime(const time_t *clock, struct tm *result)
     struct tm	*tmp;
 #endif
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(__pmLock_extcall);
 
     if (curzone >= 0) {
 	_pushTZ();
 #if defined(IS_SOLARIS) || defined(IS_MINGW)
-	tmp = localtime(clock);
+	tmp = localtime(clock);		/* THREADSAFE */
         memcpy(result, tmp, sizeof(*result));
 #else
 	localtime_r(clock, result);
@@ -496,14 +512,14 @@ pmLocaltime(const time_t *clock, struct tm *result)
     }
     else {
 #if defined(IS_SOLARIS) || defined(IS_MINGW)
-	tmp = localtime(clock);
+	tmp = localtime(clock);		/* THREADSAFE */
         memcpy(result, tmp, sizeof(*result));
 #else
 	localtime_r(clock, result);
 #endif
     }
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(__pmLock_extcall);
     return result;
 }
 
@@ -512,8 +528,7 @@ __pmMktime(struct tm *timeptr)
 {
     time_t	ans;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(__pmLock_extcall);
 
     if (curzone >= 0) {
 	_pushTZ();
@@ -523,7 +538,7 @@ __pmMktime(struct tm *timeptr)
     else
 	ans = mktime(timeptr);
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(__pmLock_extcall);
     return ans;
 }
 
@@ -531,11 +546,13 @@ int
 pmWhichZone(char **tz)
 {
     int		sts;
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+
+    PM_LOCK(__pmLock_extcall);
+
     if (curzone >= 0)
 	*tz = zone[curzone];
     sts = curzone;
-    PM_UNLOCK(__pmLock_libpcp);
+
+    PM_UNLOCK(__pmLock_extcall);
     return sts;
 }
