@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013-2017 Red Hat.
  * Copyright (c) 2008-2010 Aconex.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -10,13 +11,18 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  * License for more details.
+ *
+ *
+ * Thread-safe notes
+ *
+ * rand() et al are not thread-safe, but we don't really care here.
  */
 
 /*
  * For the MinGW headers and library to work correctly, we need
  * something newer than the default Windows 95 versions of the Win32
- * APIs.  0x0500 is the magic sauce to select the Windows 2000 version
- * of the Win32 APIs, which is the minimal version needed for PCP.
+ * APIs - we select Windows7 below as that provides minimum version
+ * with modern IPv6, as well as symlink support.
  *
  * WINVER needs to be set before any of the MinGW headers are processed
  * and we include <windows.h> from pmapi.h via platform_defs.h.
@@ -24,10 +30,13 @@
  * Thanks to "Earnie" on the mingw-users@lists.sourceforge.net mailing
  * list for this tip.
  */
-#define WINVER 0x0500
+#define WINVER Windows7
+#define _WIN32_WINNT _WIN32_WINNT_WIN7
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
+#include "internal.h"
+#include "deprecated.h"
 #include <winbase.h>
 #include <psapi.h>
 
@@ -48,16 +57,17 @@ static struct {
 VOID CALLBACK
 SignalCallback(PVOID param, BOOLEAN timerorwait)
 {
-    int index = (int)param;
+    __psint_t index = (__psint_t)param;
 
     if (index >= 0 && index < MAX_SIGNALS)
 	signals[index].callback(signals[index].signal);
-    else
-	fprintf(stderr, "SignalCallback: bad signal index (%d)\n", index);
+    else {
+	fprintf(stderr, "SignalCallback: bad signal index (%" FMT_INT64 ")\n", (__int64_t)index);
+    }
 }
 
 static char *
-MapSignals(int sig, int *index)
+MapSignals(int sig, __psint_t *index)
 {
     static char name[8];
 
@@ -83,11 +93,12 @@ MapSignals(int sig, int *index)
 int
 __pmSetSignalHandler(int sig, __pmSignalHandler func)
 {
-    int sts, index;
+    int sts;
+    __psint_t index = 0;
     char *signame, evname[64];
     HANDLE eventhdl, waithdl;
 
-    if ((signame = MapSignals(sig, &index)) < 0)
+    if ((signame = MapSignals(sig, &index)) == NULL)
 	return index;
 
     if (signals[index].callback) {	/* remove old handler */
@@ -101,7 +112,7 @@ __pmSetSignalHandler(int sig, __pmSignalHandler func)
 	return 0;
 
     sts = 0;
-    snprintf(evname, sizeof(evname), "PCP/%" FMT_PID "/%s", getpid(), signame);
+    pmsprintf(evname, sizeof(evname), "PCP/%" FMT_PID "/%s", (pid_t)getpid(), signame);
     if (!(eventhdl = CreateEvent(NULL, FALSE, FALSE, TEXT(evname)))) {
 	sts = GetLastError();
 	fprintf(stderr, "CreateEvent::%s failed (%d)\n", signame, sts);
@@ -127,31 +138,42 @@ sigterm_callback(int sig)
 }
 
 int
-__pmSetProcessIdentity(const char *username)
+pmSetProcessIdentity(const char *username)
 {
     (void)username;
     return 0;	/* Not Yet Implemented */
 }
 
-int
-__pmSetProgname(const char *program)
+void
+pmSetProgname(const char *program)
 {
-    int	sts1, sts2;
-    char *p, *suffix = NULL;
-    WORD wVersionRequested = MAKEWORD(2, 2);
-    WSADATA wsaData;
+    char	*p, *suffix = NULL;
+    static int	setup;
+    WORD	wVersionRequested = MAKEWORD(2, 2);
+    WSADATA	wsaData;
 
-    /* Trim command name of leading directory components and ".exe" suffix */
-    if (program)
+    if (program == NULL) {
+	/* Restore the default application name */
+	pmProgname = "pcp";
+    } else {
+	/* Trim command name of leading directory components */
 	pmProgname = (char *)program;
-    for (p = pmProgname; pmProgname && *p; p++) {
-	if (*p == '\\' || *p == '/')
-	    pmProgname = p + 1;
-	if (*p == '.')
-	    suffix = p;
+	for (p = pmProgname; *p; p++) {
+	    if (*p == '\\' || *p == '/') {
+		pmProgname = p + 1;
+		suffix = NULL;
+	    }
+	    else if (*p == '.')
+		suffix = p;
+	}
+	/* Drop the .exe suffix from the name if we found it */
+	if (suffix && strcmp(suffix, ".exe") == 0)
+	    *suffix = '\0';
     }
-    if (suffix && strcmp(suffix, ".exe") == 0)
-	*suffix = '\0';
+
+    if (setup)
+	return;
+    setup = 1;
 
     /* Deal with all files in binary mode - no EOL futzing */
     _fmode = O_BINARY;
@@ -160,7 +182,7 @@ __pmSetProgname(const char *program)
      * If Windows networking is not setup, all networking calls fail;
      * this even includes gethostname(2), if you can believe that. :[
      */
-    sts1 = WSAStartup(wVersionRequested, &wsaData);
+    WSAStartup(wVersionRequested, &wsaData);
 
     /*
      * Here we are emulating POSIX signals using Event objects.
@@ -169,9 +191,52 @@ __pmSetProgname(const char *program)
      * get a look-in, IOW.  Other signals (HUP/USR1) are handled
      * in a similar way, but only by processes that need them.
      */
-    sts2 = __pmSetSignalHandler(SIGTERM, sigterm_callback);
+    __pmSetSignalHandler(SIGTERM, sigterm_callback);
+}
 
-    return sts1 | sts2;
+void
+__pmServerStart(int argc, char **argv, int flags)
+{
+    PROCESS_INFORMATION piProcInfo;
+    STARTUPINFO siStartInfo;
+    LPTSTR cmdline = NULL;
+    int i, sz, total = 3; /* -f\0 */
+
+    (void)flags;
+    fflush(stdout);
+    fflush(stderr);
+
+    for (i = 0; i < argc; i++)
+	total += strlen(argv[i]) + 1;
+    if ((cmdline = malloc(total)) == NULL) {
+	pmNotifyErr(LOG_ERR, "__pmServerStart: out-of-memory");
+	exit(1);
+    }
+    for (sz = i = 0; i < argc; i++)
+	sz += pmsprintf(cmdline + sz, total - sz, "%s ", argv[i]);
+    pmsprintf(cmdline + sz, total - sz, "-f");
+
+    ZeroMemory(&piProcInfo, sizeof(PROCESS_INFORMATION));
+    ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
+    siStartInfo.cb = sizeof(STARTUPINFO);
+
+    if (0 == CreateProcess(
+		NULL, cmdline,
+		NULL, NULL,	/* process and thread attributes */
+		FALSE,		/* inherit handles */
+		CREATE_NEW_PROCESS_GROUP | CREATE_NO_WINDOW | DETACHED_PROCESS,
+		NULL,		/* environment (from parent) */
+		NULL,		/* current directory */
+		&siStartInfo,	/* STARTUPINFO pointer */
+		&piProcInfo)) {	/* receives PROCESS_INFORMATION */
+	pmNotifyErr(LOG_ERR, "__pmServerStart: CreateProcess");
+	/* but keep going */
+    }
+    else {
+	/* parent, let her exit, but avoid ugly "Log finished" messages */
+	fclose(stderr);
+	exit(0);
+    }
 }
 
 void *
@@ -269,7 +334,11 @@ __pmProcessCreate(char **argv, int *infd, int *outfd)
  
     for (command = argv[0], i = 0; command && *command; command = argv[++i]) {
 	int length = strlen(command);
-	cmdline = realloc(cmdline, sz + length + 1); /* 1space or 1null */
+	/* add 1space or 1null */
+	if ((cmdline = realloc(cmdline, sz + length + 1)) == NULL) {
+	    pmNoMem("__pmProcessCreate", sz + length + 1, PM_FATAL_ERR);
+	    /* NOTREACHED */
+	}
 	strcpy(&cmdline[sz], command);
 	cmdline[sz + length] = ' ';
 	sz += length + 1;
@@ -375,7 +444,7 @@ __pmProcessRunTimes(double *usr, double *sys)
 }
 
 void
-__pmtimevalNow(struct timeval *tv)
+pmtimevalNow(struct timeval *tv)
 {
     struct timespec ts;
     union {
@@ -415,11 +484,42 @@ sleep(unsigned int seconds)
     return 0;
 }
 
-void setlinebuf(FILE *stream)
+void
+setlinebuf(FILE *stream)
 {
     setvbuf(stream, NULL, _IONBF, 0);	/* no line buffering in Win32 */
 }
 
+int
+fsync(int fd)
+{
+    return FlushFileBuffers((HANDLE)_get_osfhandle(fd)) ? 0 : -1;
+}
+
+int
+symlink(const char *oldpath, const char *newpath)
+{
+    return CreateSymbolicLink(newpath, oldpath, 0) ? 0 : -1;
+}
+
+int
+readlink(const char *path, char *buf, size_t bufsiz)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+long int
+random(void)
+{
+    return rand();
+}
+
+void
+srandom(unsigned int seed)
+{
+    srandom(seed);
+}
+
 long int
 lrand48(void)
 {
@@ -456,6 +556,51 @@ rindex(const char *string, int marker)
     return NULL;
 }
 
+char *
+strcasestr(const char *string, const char *substr)
+{
+    int i, j;
+    int sublen = strlen(substr);
+    int length = strlen(string) - sublen + 1;
+
+    for (i = 0; i < length; i++) {
+	for (j = 0; j < sublen; j++)
+	    if (toupper(string[i+j]) != toupper(substr[j]))
+		goto outerloop;
+	return (char *) substr + i;
+    outerloop:
+	continue;
+    }
+    return NULL;
+}
+
+char *
+strsep(char **stringp, const char *delim)
+{
+    char	*ss, *se;
+    const char	*dp;
+
+    if ((ss = *stringp) == NULL)
+	return NULL;
+
+    for (se = ss; *se; se++) {
+	for (dp = delim; *dp; dp++) {
+	    if (*se == *dp)
+		break;
+	}
+    }
+
+    if (*se != '\0') {
+	/* match: terminate and update stringp to point past match */
+	*se++ = '\0';
+	*stringp = se;
+    }
+    else
+	*stringp = NULL;
+
+    return ss;
+}
+
 void *
 dlopen(const char *filename, int flag)
 {
@@ -486,11 +631,13 @@ static char *eventlogPrefix;
 void
 openlog(const char *ident, int option, int facility)
 {
+    PM_LOCK(__pmLock_extcall);
     if (eventlog)
 	closelog();
-    eventlog = RegisterEventSource(NULL, "Application");
+    eventlog = RegisterEventSource(NULL, "Application");	/* THREADSAFE */
     if (ident)
 	eventlogPrefix = strdup(ident);
+    PM_UNLOCK(__pmLock_extcall);
 }
 
 void
@@ -509,7 +656,7 @@ syslog(int priority, const char *format, ...)
 	openlog(NULL, 0, 0);
 
     if (eventlogPrefix)
-	offset = snprintf(p, sizeof(logmsg), "%s: ", eventlogPrefix);
+	offset = pmsprintf(p, sizeof(logmsg), "%s: ", eventlogPrefix);
 
     switch (priority) {
     case LOG_EMERG:
@@ -529,7 +676,7 @@ syslog(int priority, const char *format, ...)
 	break;
     }
     msgptr = logmsg;
-    snprintf(p + offset, sizeof(logmsg) - offset, format, arg);
+    pmsprintf(p + offset, sizeof(logmsg) - offset, format, arg);
     ReportEvent(eventlog, eventlogPriority, 0, 0, NULL, 1, 0, &msgptr, NULL);
     va_end(arg);
 }
@@ -537,22 +684,22 @@ syslog(int priority, const char *format, ...)
 void
 closelog(void)
 {
+    PM_LOCK(__pmLock_extcall);
     if (eventlog) {
-	DeregisterEventSource(eventlog);
+	DeregisterEventSource(eventlog);		/* THREADSAFE */
 	if (eventlogPrefix)
 	    free(eventlogPrefix);
 	eventlogPrefix = NULL;
     }
     eventlog = NULL;
+    PM_UNLOCK(__pmLock_extcall);
 }
 
 const char *
 strerror_r(int errnum, char *buf, size_t buflen)
 {
-    /* strerror_s is missing from the MinGW string.h */
-    /* we need to wait for it until we can do this:  */
-    /* return strerror_s(buf, buflen, errnum); */
-    return strerror(errnum);
+    strerror_s(buf, buflen, errnum);
+    return (const char *)buf;
 }
 
 /* Windows socket error codes - what a nightmare! */
@@ -632,3 +779,152 @@ wsastrerror(int code)
 	    return wsatab[i].errmess;
     return NULL;
 }
+
+/*
+ * User and group account management using Security IDs (SIDs)
+ */
+int
+__pmValidUserID(__pmUserID sid)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmValidGroupID(__pmGroupID sid)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmEqualUserIDs(__pmUserID sid1, __pmUserID sid2)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmEqualGroupIDs(__pmGroupID sid1, __pmGroupID sid2)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+void
+__pmUserIDFromString(const char *username, __pmUserID *sid)
+{
+    /* NYI */
+}
+
+void
+__pmGroupIDFromString(const char *groupname, __pmGroupID *sid)
+{
+    /* NYI */
+}
+
+char *
+__pmUserIDToString(__pmUserID sid, char *buffer, size_t size)
+{
+    return NULL;	/* NYI */
+}
+
+char *
+__pmGroupIDToString(__pmGroupID gid, char *buffer, size_t size)
+{
+    return NULL;	/* NYI */
+}
+
+int
+__pmUsernameToID(const char *username, __pmUserID *uidp)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmGroupnameToID(const char *groupname, __pmGroupID *gidp)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+char *
+__pmGroupnameFromID(__pmGroupID gid, char *buf, size_t size)
+{
+    return NULL;	/* NYI */
+}
+
+char *
+__pmUsernameFromID(__pmUserID uid, char *buf, size_t size)
+{
+    return NULL;	/* NYI */
+}
+
+int
+__pmUsersGroupIDs(const char *username, __pmGroupID **groupids, unsigned int *ngroups)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmGroupsUserIDs(const char *groupname, __pmUserID **userids, unsigned int *nusers)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmGetUserIdentity(const char *username, __pmUserID *uid, __pmGroupID *gid, int mode)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+/*
+ * always called with __pmLock_extcall already held, so THREADSAFE
+ */
+int
+setenv(const char *name, const char *value, int overwrite)
+{
+    char	*ebuf;
+
+    if (getenv(name) != NULL) {		/* THREADSAFE */
+	/* already in the environment */
+	if (!overwrite)
+	    return(0);
+    }
+
+    if ((ebuf = (char *)malloc(strlen(name) + strlen(value) + 2)) == NULL)
+	return -1;
+
+    strncpy(ebuf, name, strlen(name)+1);
+    strncat(ebuf, "=", 1);
+    strncat(ebuf, value, strlen(value));
+
+    return _putenv(ebuf);		/* THREADSAFE */
+}
+
+/*
+ * always called with __pmLock_extcall already held, so THREADSAFE
+ */
+int
+unsetenv(const char *name)
+{
+    char	*ebuf;
+    int		sts;
+
+    if ((ebuf = (char *)malloc(strlen(name) + 2)) == NULL)
+	return -1;
+
+    /* strange but true */
+    strncpy(ebuf, name, strlen(name)+1);
+    strncat(ebuf, "=", 1);
+    sts = _putenv(ebuf);		/* THREADSAFE */
+    free(ebuf);
+    return sts;
+}
+
+int
+win32_inet_pton(int af, const char *src, void *dst)
+{
+    return InetPton(af, src, dst);
+}
+
+const char *
+win32_inet_ntop(int af, void *src, char *dst, socklen_t size)
+{
+    return InetNtop(af, src, dst, size);
+}
