@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2017 Red Hat.
  * Copyright (c) 2010 Ken McDonell.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -13,11 +14,11 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "import.h"
 #include "private.h"
 
-static __pmTimeval	stamp;
+static pmTimeval	stamp;
 
 int
 _pmi_put_result(pmi_context *current, pmResult *result)
@@ -27,6 +28,7 @@ _pmi_put_result(pmi_context *current, pmResult *result)
     char	myname[MAXHOSTNAMELEN];
     __pmPDU	*pb;
     __pmLogCtl	*lcp = &current->logctl;
+    __pmArchCtl	*acp = &current->archctl;
     int		k;
     int		i;
     int		m;
@@ -51,12 +53,15 @@ _pmi_put_result(pmi_context *current, pmResult *result)
 	else
 	    host = current->hostname;
 
-	sts = __pmLogCreate(host, current->archive, PM_LOG_VERS02, lcp);
+	acp->ac_log = &current->logctl;
+	sts = __pmLogCreate(host, current->archive, PM_LOG_VERS02, acp);
 	if (sts < 0)
 	    return sts;
 
-	if (current->timezone == NULL)
-	    strcpy(lcp->l_label.ill_tz, __pmTimezone());
+	if (current->timezone == NULL) {
+	    char	tzbuf[PM_TZ_MAXLEN];
+	    strcpy(lcp->l_label.ill_tz, __pmTimezone_r(tzbuf, sizeof(tzbuf)));
+	}
 	else
 	    strcpy(lcp->l_label.ill_tz, current->timezone);
 	pmNewZone(lcp->l_label.ill_tz);
@@ -64,8 +69,9 @@ _pmi_put_result(pmi_context *current, pmResult *result)
 
 	/*
 	 * do the label records (it is too late when __pmLogPutResult
-	 * is called as we've already output some metadata) ... this
-	 * code is stolen from __pmLogPutResult
+	 * or __pmLogPutResult2 is called as we've already output some
+	 * metadata) ... this code is stolen from logputresult() in
+	 * libpcp
 	 */
 	lcp->l_label.ill_start.tv_sec = stamp.tv_sec;
 	lcp->l_label.ill_start.tv_usec = stamp.tv_usec;
@@ -74,13 +80,13 @@ _pmi_put_result(pmi_context *current, pmResult *result)
 	lcp->l_label.ill_vol = PM_LOG_VOL_META;
 	__pmLogWriteLabel(lcp->l_mdfp, &lcp->l_label);
 	lcp->l_label.ill_vol = 0;
-	__pmLogWriteLabel(lcp->l_mfp, &lcp->l_label);
+	__pmLogWriteLabel(acp->ac_mfp, &lcp->l_label);
 	lcp->l_state = PM_LOG_STATE_INIT;
-	__pmLogPutIndex(&current->logctl, &stamp);
+	__pmLogPutIndex(&current->archctl, &stamp);
     }
 
-    __pmOverrideLastFd(fileno(lcp->l_mfp));
-    if ((sts = __pmEncodeResult(fileno(lcp->l_mfp), result, &pb)) < 0)
+    __pmOverrideLastFd(__pmFileno(acp->ac_mfp));
+    if ((sts = __pmEncodeResult(__pmFileno(acp->ac_mfp), result, &pb)) < 0)
 	return sts;
 
     needti = 0;
@@ -91,8 +97,10 @@ _pmi_put_result(pmi_context *current, pmResult *result)
 	    if (current->metric[m].meta_done == 0) {
 		char	**namelist = &current->metric[m].name;
 
-		if ((sts = __pmLogPutDesc(lcp, &current->metric[m].desc, 1, namelist)) < 0)
+		if ((sts = __pmLogPutDesc(acp, &current->metric[m].desc, 1, namelist)) < 0) {
+		    __pmUnpinPDUBuf(pb);
 		    return sts;
+		}
 		current->metric[m].meta_done = 1;
 		needti = 1;
 	    }
@@ -100,8 +108,10 @@ _pmi_put_result(pmi_context *current, pmResult *result)
 		for (i = 0; i < current->nindom; i++) {
 		    if (current->metric[m].desc.indom == current->indom[i].indom) {
 			if (current->indom[i].meta_done == 0) {
-			    if ((sts = __pmLogPutInDom(lcp, current->indom[i].indom, &stamp, current->indom[i].ninstance, current->indom[i].inst, current->indom[i].name)) < 0)
+			    if ((sts = __pmLogPutInDom(acp, current->indom[i].indom, &stamp, current->indom[i].ninstance, current->indom[i].inst, current->indom[i].name)) < 0) {
+				__pmUnpinPDUBuf(pb);
 				return sts;
+			    }
 			    current->indom[i].meta_done = 1;
 			    needti = 1;
 			}
@@ -112,12 +122,15 @@ _pmi_put_result(pmi_context *current, pmResult *result)
 	}
     }
     if (needti) {
-	__pmLogPutIndex(lcp, &stamp);
+	__pmLogPutIndex(acp, &stamp);
     }
 
-    if ((sts = __pmLogPutResult(lcp, pb)) < 0)
+    if ((sts = __pmLogPutResult2(acp, pb)) < 0) {
+	__pmUnpinPDUBuf(pb);
 	return sts;
+    }
 
+    __pmUnpinPDUBuf(pb);
     return 0;
 }
 
@@ -127,7 +140,9 @@ _pmi_end(pmi_context *current)
     /* Final temporal index update to finish the archive
      * ... same logic here as in run_done() for pmlogger
      */
-    __pmLogPutIndex(&current->logctl, &stamp);
+    __pmLogPutIndex(&current->archctl, &stamp);
+
+    __pmLogClose(&current->archctl);
 
     current->state = CONTEXT_END;
     return 0;
