@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013-2017 Red Hat.
  * Copyright (c) 2010 Ken McDonell.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -13,7 +14,7 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "import.h"
 #include "domain.h"
 #include "private.h"
@@ -23,6 +24,18 @@ static pmi_context *context_tab;
 static int ncontext;
 static pmi_context *current;
 
+static void
+printstamp(FILE *f, const struct timeval *tp)
+{
+    struct tm	tmp;
+    time_t	now;
+
+    now = (time_t)tp->tv_sec;
+    pmLocaltime(&now, &tmp);
+    fprintf(f, "%4d-%02d-%02d %02d:%02d:%02d.%06d", 1900+tmp.tm_year, tmp.tm_mon, tmp.tm_mday, tmp.tm_hour, tmp.tm_min, tmp.tm_sec, (int)(tp->tv_usec));
+}
+ 
+
 void
 pmiDump(void)
 {
@@ -65,7 +78,7 @@ pmiDump(void)
 	    fprintf(f, "  metric[%d] name=%s pmid=%s\n",
 		m, current->metric[m].name,
 		pmIDStr_r(current->metric[m].pmid, strbuf, sizeof(strbuf)));
-	    __pmPrintDesc(f, &current->metric[m].desc);
+	    pmPrintDesc(f, &current->metric[m].desc);
 	}
     }
     if (current->nindom == 0)
@@ -125,7 +138,7 @@ pmiUnits(int dimSpace, int dimTime, int dimCount, int scaleSpace, int scaleTime,
 pmID
 pmiID(int domain, int cluster, int item)
 {
-    return pmid_build(domain, cluster, item);
+    return pmID_build(domain, cluster, item);
 }
 
 pmInDom
@@ -185,6 +198,9 @@ pmiErrStr_r(int code, char *buf, int buflen)
 	case PMI_ERR_BADMETRICNAME:
 	    msg = "Illegal metric name";
 	    break;
+	case PMI_ERR_BADTIMESTAMP:
+	    msg = "Illegal result timestamp";
+	    break;
 	default:
 	    return pmErrStr_r(code, buf, buflen);
     }
@@ -203,7 +219,7 @@ pmiStart(const char *archive, int inherit)
     ncontext++;
     context_tab = (pmi_context *)realloc(context_tab, ncontext*sizeof(context_tab[0]));
     if (context_tab == NULL) {
-	__pmNoMem("pmiStart: context_tab", ncontext*sizeof(context_tab[0]), PM_FATAL_ERR);
+	pmNoMem("pmiStart: context_tab", ncontext*sizeof(context_tab[0]), PM_FATAL_ERR);
     }
     old_current = &context_tab[c];
     current = &context_tab[ncontext-1];
@@ -211,19 +227,21 @@ pmiStart(const char *archive, int inherit)
     current->state = CONTEXT_START;
     current->archive = strdup(archive);
     if (current->archive == NULL) {
-	__pmNoMem("pmiStart", strlen(archive)+1, PM_FATAL_ERR);
+	pmNoMem("pmiStart", strlen(archive)+1, PM_FATAL_ERR);
     }
     current->hostname = NULL;
     current->timezone = NULL;
     current->result = NULL;
     memset((void *)&current->logctl, 0, sizeof(current->logctl));
+    memset((void *)&current->archctl, 0, sizeof(current->archctl));
+    current->archctl.ac_log = &current->logctl;
     if (inherit && old_current != NULL) {
 	current->nmetric = old_current->nmetric;
 	if (old_current->metric != NULL) {
 	    int		m;
 	    current->metric = (pmi_metric *)malloc(current->nmetric*sizeof(pmi_metric));
 	    if (current->metric == NULL) {
-		__pmNoMem("pmiStart: pmi_metric", current->nmetric*sizeof(pmi_metric), PM_FATAL_ERR);
+		pmNoMem("pmiStart: pmi_metric", current->nmetric*sizeof(pmi_metric), PM_FATAL_ERR);
 	    }
 	    for (m = 0; m < current->nmetric; m++) {
 		current->metric[m].name = old_current->metric[m].name;
@@ -239,7 +257,7 @@ pmiStart(const char *archive, int inherit)
 	    int		i;
 	    current->indom = (pmi_indom *)malloc(current->nindom*sizeof(pmi_indom));
 	    if (current->indom == NULL) {
-		__pmNoMem("pmiStart: pmi_indom", current->nindom*sizeof(pmi_indom), PM_FATAL_ERR);
+		pmNoMem("pmiStart: pmi_indom", current->nindom*sizeof(pmi_indom), PM_FATAL_ERR);
 	    }
 	    for (i = 0; i < current->nindom; i++) {
 		int		j;
@@ -249,16 +267,16 @@ pmiStart(const char *archive, int inherit)
 		if (old_current->indom[i].ninstance > 0) {
 		    current->indom[i].name = (char **)malloc(current->indom[i].ninstance*sizeof(char *));
 		    if (current->indom[i].name == NULL) {
-			__pmNoMem("pmiStart: name", current->indom[i].ninstance*sizeof(char *), PM_FATAL_ERR);
+			pmNoMem("pmiStart: name", current->indom[i].ninstance*sizeof(char *), PM_FATAL_ERR);
 		    }
 		    current->indom[i].inst = (int *)malloc(current->indom[i].ninstance*sizeof(int));
 		    if (current->indom[i].inst == NULL) {
-			__pmNoMem("pmiStart: inst", current->indom[i].ninstance*sizeof(int), PM_FATAL_ERR);
+			pmNoMem("pmiStart: inst", current->indom[i].ninstance*sizeof(int), PM_FATAL_ERR);
 		    }
 		    current->indom[i].namebuflen = old_current->indom[i].namebuflen;
 		    current->indom[i].namebuf = (char *)malloc(old_current->indom[i].namebuflen);
 		    if (current->indom[i].namebuf == NULL) {
-			__pmNoMem("pmiStart: namebuf", old_current->indom[i].namebuflen, PM_FATAL_ERR);
+			pmNoMem("pmiStart: namebuf", old_current->indom[i].namebuflen, PM_FATAL_ERR);
 		    }
 		    np = current->indom[i].namebuf;
 		    for (j = 0; j < current->indom[i].ninstance; j++) {
@@ -283,7 +301,7 @@ pmiStart(const char *archive, int inherit)
 	    int		h;
 	    current->handle = (pmi_handle *)malloc(current->nhandle*sizeof(pmi_handle));
 	    if (current->handle == NULL) {
-		__pmNoMem("pmiStart: pmi_handle", current->nhandle*sizeof(pmi_handle), PM_FATAL_ERR);
+		pmNoMem("pmiStart: pmi_handle", current->nhandle*sizeof(pmi_handle), PM_FATAL_ERR);
 	    }
 	    for (h = 0; h < current->nhandle; h++) {
 		current->handle[h].midx = old_current->handle[h].midx;
@@ -292,6 +310,7 @@ pmiStart(const char *archive, int inherit)
 	}
 	else
 	    current->handle = NULL;
+	current->last_stamp = old_current->last_stamp;
     }
     else {
 	current->nmetric = 0;
@@ -300,6 +319,7 @@ pmiStart(const char *archive, int inherit)
 	current->indom = NULL;
 	current->nhandle = 0;
 	current->handle = NULL;
+	current->last_stamp.tv_sec = current->last_stamp.tv_usec = 0;
     }
     return ncontext;
 }
@@ -331,7 +351,7 @@ pmiSetHostname(const char *value)
 	return PM_ERR_NOCONTEXT;
     current->hostname = strdup(value);
     if (current->hostname == NULL) {
-	__pmNoMem("pmiSetHostname", strlen(value)+1, PM_FATAL_ERR);
+	pmNoMem("pmiSetHostname", strlen(value)+1, PM_FATAL_ERR);
     }
     return current->last_sts = 0;
 }
@@ -343,7 +363,7 @@ pmiSetTimezone(const char *value)
 	return PM_ERR_NOCONTEXT;
     current->timezone = strdup(value);
     if (current->timezone == NULL) {
-	__pmNoMem("pmiSetTimezone", strlen(value)+1, PM_FATAL_ERR);
+	pmNoMem("pmiSetTimezone", strlen(value)+1, PM_FATAL_ERR);
     }
     return current->last_sts = 0;
 }
@@ -383,6 +403,9 @@ int
 pmiAddMetric(const char *name, pmID pmid, int type, pmInDom indom, int sem, pmUnits units)
 {
     int		m;
+    int		item;
+    int		cluster;
+    size_t	size;
     pmi_metric	*mp;
 
     if (current == NULL)
@@ -428,19 +451,29 @@ pmiAddMetric(const char *name, pmID pmid, int type, pmInDom indom, int sem, pmUn
     }
 
     current->nmetric++;
-    current->metric = (pmi_metric *)realloc(current->metric, current->nmetric*sizeof(pmi_metric));
+    size = current->nmetric * sizeof(pmi_metric);
+    current->metric = (pmi_metric *)realloc(current->metric, size);
     if (current->metric == NULL) {
-	__pmNoMem("pmiAddMetric: pmi_metric", current->nmetric*sizeof(pmi_metric), PM_FATAL_ERR);
+	pmNoMem("pmiAddMetric: pmi_metric", size, PM_FATAL_ERR);
     }
     mp = &current->metric[current->nmetric-1];
+    if (pmid != PM_ID_NULL) {
+	mp->pmid = pmid;
+    } else {
+	/* choose a PMID on behalf of the caller - check boundaries first */
+	item = cluster = current->nmetric;
+	if (item >= (1<<22)) {	/* enough room for unique item:cluster? */
+	    current->nmetric--;
+	    return current->last_sts = PMI_ERR_DUPMETRICID;	/* wrap */
+	}
+	item %= (1<<10);
+	cluster >>= 10;
+	mp->pmid = pmID_build(PMI_DOMAIN, cluster, item);
+    }
     mp->name = strdup(name);
     if (mp->name == NULL) {
-	__pmNoMem("pmiAddMetric: name", strlen(name)+1, PM_FATAL_ERR);
+	pmNoMem("pmiAddMetric: name", strlen(name)+1, PM_FATAL_ERR);
     }
-    if (pmid == PM_ID_NULL)
-	mp->pmid = pmid_build(PMI_DOMAIN, 0, current->nmetric);
-    else
-	mp->pmid = pmid;
     mp->desc.pmid = mp->pmid;
     mp->desc.type = type;
     mp->desc.indom = indom;
@@ -473,7 +506,7 @@ pmiAddInstance(pmInDom indom, const char *instance, int inst)
 	current->nindom++;
 	current->indom = (pmi_indom *)realloc(current->indom, current->nindom*sizeof(pmi_indom));
 	if (current->indom == NULL) {
-	    __pmNoMem("pmiAddInstance: pmi_indom", current->nindom*sizeof(pmi_indom), PM_FATAL_ERR);
+	    pmNoMem("pmiAddInstance: pmi_indom", current->nindom*sizeof(pmi_indom), PM_FATAL_ERR);
 	}
 	current->indom[i].indom = indom;
 	current->indom[i].ninstance = 0;
@@ -508,15 +541,15 @@ pmiAddInstance(pmInDom indom, const char *instance, int inst)
     idp->ninstance++;
     idp->name = (char **)realloc(idp->name, idp->ninstance*sizeof(char *));
     if (idp->name == NULL) {
-	__pmNoMem("pmiAddInstance: name", idp->ninstance*sizeof(char *), PM_FATAL_ERR);
+	pmNoMem("pmiAddInstance: name", idp->ninstance*sizeof(char *), PM_FATAL_ERR);
     }
     idp->inst = (int *)realloc(idp->inst, idp->ninstance*sizeof(int));
     if (idp->inst == NULL) {
-	__pmNoMem("pmiAddInstance: inst", idp->ninstance*sizeof(int), PM_FATAL_ERR);
+	pmNoMem("pmiAddInstance: inst", idp->ninstance*sizeof(int), PM_FATAL_ERR);
     }
     idp->namebuf = (char *)realloc(idp->namebuf, idp->namebuflen+strlen(instance)+1);
     if (idp->namebuf == NULL) {
-	__pmNoMem("pmiAddInstance: namebuf", idp->namebuflen+strlen(instance)+1, PM_FATAL_ERR);
+	pmNoMem("pmiAddInstance: namebuf", idp->namebuflen+strlen(instance)+1, PM_FATAL_ERR);
     }
     strcpy(&idp->namebuf[idp->namebuflen], instance);
     idp->namebuflen += strlen(instance)+1;
@@ -627,7 +660,7 @@ pmiGetHandle(const char *name, const char *instance)
     current->nhandle++;
     current->handle = (pmi_handle *)realloc(current->handle, current->nhandle*sizeof(pmi_handle));
     if (current->handle == NULL) {
-	__pmNoMem("pmiGetHandle: pmi_handle", current->nhandle*sizeof(pmi_handle), PM_FATAL_ERR);
+	pmNoMem("pmiGetHandle: pmi_handle", current->nhandle*sizeof(pmi_handle), PM_FATAL_ERR);
     }
     hp = &current->handle[current->nhandle-1];
     hp->midx = tmp.midx;
@@ -647,6 +680,22 @@ pmiPutValueHandle(int handle, const char *value)
     return current->last_sts = _pmi_stuff_value(current, &current->handle[handle-1], value);
 }
 
+static int
+check_timestamp(void)
+{
+    if (current->result->timestamp.tv_sec < current->last_stamp.tv_sec ||
+        (current->result->timestamp.tv_sec == current->last_stamp.tv_sec &&
+	 current->result->timestamp.tv_usec < current->last_stamp.tv_usec)) {
+	fprintf(stderr, "Fatal Error: timestamp ");
+	printstamp(stderr, &current->result->timestamp);
+	fprintf(stderr, " not greater than previous valid timestamp ");
+	printstamp(stderr, &current->last_stamp);
+	fputc('\n', stderr);
+	return PMI_ERR_BADTIMESTAMP;
+    }
+    return 0;
+}
+
 int
 pmiWrite(int sec, int usec)
 {
@@ -658,13 +707,16 @@ pmiWrite(int sec, int usec)
 	return current->last_sts = PMI_ERR_NODATA;
 
     if (sec < 0) {
-	__pmtimevalNow(&current->result->timestamp);
+	pmtimevalNow(&current->result->timestamp);
     }
     else {
 	current->result->timestamp.tv_sec = sec;
 	current->result->timestamp.tv_usec = usec;
     }
-    sts = _pmi_put_result(current, current->result);
+    if ((sts = check_timestamp()) == 0) {
+	sts = _pmi_put_result(current, current->result);
+	current->last_stamp = current->result->timestamp;
+    }
 
     pmFreeResult(current->result);
     current->result = NULL;
@@ -675,8 +727,54 @@ pmiWrite(int sec, int usec)
 int
 pmiPutResult(const pmResult *result)
 {
+    int		sts;
+
     if (current == NULL)
 	return PM_ERR_NOCONTEXT;
 
-    return  current->last_sts = _pmi_put_result(current, current->result);
+    current->result = (pmResult *)result;
+    if ((sts = check_timestamp()) == 0) {
+	sts = _pmi_put_result(current, current->result);
+	current->last_stamp = current->result->timestamp;
+    }
+    current->result = NULL;
+
+    return current->last_sts = sts;
+}
+
+int
+pmiPutMark(void)
+{
+    __pmArchCtl *acp;
+    struct {
+	__pmPDU		hdr;
+	pmTimeval	timestamp;	/* when returned */
+	int		numpmid;	/* zero PMIDs to follow */
+	__pmPDU		tail;
+    } mark;
+
+    if (current == NULL)
+	return PM_ERR_NOCONTEXT;
+
+    if (current->last_stamp.tv_sec == 0 && current->last_stamp.tv_usec == 0)
+	/* no earlier pmResult, no point adding a mark record */
+	return 0;
+    acp = &current->archctl;
+
+    mark.hdr = htonl((int)sizeof(mark));
+    mark.tail = mark.hdr;
+    mark.timestamp.tv_sec = current->last_stamp.tv_sec;
+    mark.timestamp.tv_usec = current->last_stamp.tv_usec + 1000;	/* + 1msec */
+    if (mark.timestamp.tv_usec > 1000000) {
+	mark.timestamp.tv_usec -= 1000000;
+	mark.timestamp.tv_sec++;
+    }
+    mark.timestamp.tv_sec = htonl(mark.timestamp.tv_sec);
+    mark.timestamp.tv_usec = htonl(mark.timestamp.tv_usec);
+    mark.numpmid = htonl(0);
+
+    if (__pmFwrite(&mark, 1, sizeof(mark), acp->ac_mfp) != sizeof(mark))
+	return -oserror();
+
+    return 0;
 }
