@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013,2015 Red Hat.
  * Copyright (c) 1995-2000,2003 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -10,20 +11,16 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
+#include <time.h>
 #include "pmcd.h"
-
-#ifdef IS_SOLARIS
-#define _REENTRANT
-#include <time.h>
-#undef _REENTRANT
+#include "config.h"
+#if HAVE_STATIC_PROBES
+#include "probes.h"
 #else
-#include <time.h>
+#define PCP_PROBE_PMCD_PDU(type,who,p1,p2)
+#define PCP_PROBE_PMCD(type,who,p1,p2)
 #endif
 
 /*
@@ -48,8 +45,8 @@ static unsigned int	next;
  * to 1, pmcd.control.traceconn (trace connections) and/or
  * pmcd.control.tracepdu (trace PDUs)
  */
-PMCD_INTERN int		_pmcd_trace_nbufs = 20;
-PMCD_INTERN int		_pmcd_trace_mask;
+PMCD_DATA int		pmcd_trace_nbufs = 20;
+PMCD_DATA int		pmcd_trace_mask;
 
 void
 pmcd_init_trace(int n)
@@ -57,10 +54,10 @@ pmcd_init_trace(int n)
     if (trace != NULL)
 	free(trace);
     if ((trace = (tracebuf *)malloc(n * sizeof(tracebuf))) == NULL) {
-	 __pmNoMem("pmcd_init_trace", n * sizeof(tracebuf), PM_RECOV_ERR);
+	 pmNoMem("pmcd_init_trace", n * sizeof(tracebuf), PM_RECOV_ERR);
 	 return;
     }
-    _pmcd_trace_nbufs = n;
+    pmcd_trace_nbufs = n;
     next = 0;
 }
 
@@ -69,23 +66,27 @@ pmcd_trace(int type, int who, int p1, int p2)
 {
     int		p;
 
+    switch (type) {
+	case TR_XMIT_PDU:
+	case TR_RECV_PDU:
+	    PCP_PROBE_PMCD_PDU(type, who, p1, p2);
+	    if ((pmcd_trace_mask & TR_MASK_PDU) == 0)
+		return;
+	    break;
+	default:
+	    PCP_PROBE_PMCD(type, who, p1, p2);
+	    if ((pmcd_trace_mask & TR_MASK_CONN) == 0)
+		return;
+	    break;
+    }
+
     if (trace == NULL) {
-	pmcd_init_trace(_pmcd_trace_nbufs);
+	pmcd_init_trace(pmcd_trace_nbufs);
 	if (trace == NULL)
 	    return;
     }
 
-    switch (type) {
-	case TR_XMIT_PDU:
-	case TR_RECV_PDU:
-	    if ((_pmcd_trace_mask & TR_MASK_PDU) == 0)
-		return;
-	default:
-	    if ((_pmcd_trace_mask & TR_MASK_CONN) == 0)
-		return;
-    }
-
-    p = (next++) % _pmcd_trace_nbufs;
+    p = (next++) % pmcd_trace_nbufs;
 
     time(&trace[p].t_stamp);
     trace[p].t_type = type;
@@ -93,7 +94,7 @@ pmcd_trace(int type, int who, int p1, int p2)
     trace[p].t_p1 = p1;
     trace[p].t_p2 = p2;
 
-    if (_pmcd_trace_mask & TR_MASK_NOBUF)
+    if (pmcd_trace_mask & TR_MASK_NOBUF)
 	/* unbuffered, dump it now */
 	pmcd_dump_trace(stderr);
 }
@@ -107,15 +108,15 @@ pmcd_dump_trace(FILE *f)
     struct tm		*this;
     char		strbuf[20];
 
-    if ((_pmcd_trace_mask & TR_MASK_NOBUF) == 0)
+    if ((pmcd_trace_mask & TR_MASK_NOBUF) == 0)
 	fprintf(f, "\n->PMCD event trace: ");
     if (trace != NULL && next != 0) {
-	if (next < _pmcd_trace_nbufs)
+	if (next < pmcd_trace_nbufs)
 	    i = 0;
 	else
-	    i = next - _pmcd_trace_nbufs;
-	if ((_pmcd_trace_mask & TR_MASK_NOBUF) == 0) {
-	    fprintf(f, "starting at %s", ctime(&trace[i % _pmcd_trace_nbufs].t_stamp));
+	    i = next - pmcd_trace_nbufs;
+	if ((pmcd_trace_mask & TR_MASK_NOBUF) == 0) {
+	    fprintf(f, "starting at %s", ctime(&trace[i % pmcd_trace_nbufs].t_stamp));
 	    last.tm_hour = -1;
 	}
 	else
@@ -123,7 +124,7 @@ pmcd_dump_trace(FILE *f)
 
 	for ( ; i < next; i++) {
 	    fprintf(f, "->");
-	    p = i % _pmcd_trace_nbufs;
+	    p = i % pmcd_trace_nbufs;
 	    this = localtime(&trace[p].t_stamp);
 	    if (this->tm_hour != last.tm_hour ||
 		this->tm_min != last.tm_min ||
@@ -146,7 +147,7 @@ pmcd_dump_trace(FILE *f)
 			fprintf(f, "New client: [%d] ", trace[p].t_who);
 			cip = GetClient(trace[p].t_who);
 			if (cip == NULL) {
-			    fprintf(f, "-- unknown\n?");
+			    fprintf(f, "-- unknown?\n");
 			}
 			else {
 			    __pmSockAddr	*saddr = (__pmSockAddr *)cip->addr;
@@ -165,9 +166,9 @@ pmcd_dump_trace(FILE *f)
 		    break;
 
 		case TR_DEL_CLIENT:
-		    fprintf(f, "End client: fd=%d", trace[p].t_who);
-		    if (trace[p].t_p1 != 0)
-			fprintf(f, ", err=%d: %s", trace[p].t_p1, pmErrStr(trace[p].t_p1));
+		    fprintf(f, "End client: [%d] fd=%d", trace[p].t_who, trace[p].t_p1);
+		    if (trace[p].t_p2 != 0)
+			fprintf(f, ", err=%d: %s", trace[p].t_p2, pmErrStr(trace[p].t_p2));
 		    fputc('\n', f);
 		    break;
 
@@ -282,7 +283,7 @@ pmcd_dump_trace(FILE *f)
     else
 	fprintf(f, "<empty>\n");
 
-    if ((_pmcd_trace_mask & TR_MASK_NOBUF) == 0)
+    if ((pmcd_trace_mask & TR_MASK_NOBUF) == 0)
 	fputc('\n', f);
     next = 0;			/* empty the circular buffer */
 }
