@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013,2015-2017 Red Hat.
  * Copyright (c) 2005 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -13,14 +14,13 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "pmda.h"
+#include "libdefs.h"
 #include <fcntl.h>
 #include <ctype.h>
 #include <sys/stat.h>
 
-static __uint32_t hash(const char *, int, __uint32_t);
-
 /*
  * simple linked list for each cache at this stage
  */
@@ -38,7 +38,9 @@ typedef struct entry {
     time_t		stamp;
 } entry_t;
 
-#define VERSION 1	/* version of external file format */
+#define CACHE_VERSION1	1
+#define CACHE_VERSION2	2
+#define CACHE_VERSION	CACHE_VERSION2	/* version of external file format */
 #define MAX_HASH_TRY	10
 
 /*
@@ -56,13 +58,17 @@ typedef struct hdr {
     int			hbits;
     int			nentry;		/* number of entries */
     int			ins_mode;	/* see insert_cache() */
-    int			hstate;		/* dirty/clean state for save_cache() */
+    int			hstate;		/* dirty/clean/string state */
     int			keyhash_cnt[MAX_HASH_TRY];
+    int			maxinst;	/* maximum inst */
 } hdr_t;
 
+#define DEFAULT_MAXINST 0x7fffffff
+
 /* bitfields for hstate */
-#define DIRTY_INSTANCE	1
-#define DIRTY_STAMP	2
+#define DIRTY_INSTANCE	0x1
+#define DIRTY_STAMP	0x2
+#define CACHE_STRINGS	0x4
 
 static hdr_t	*base;		/* start of cache headers */
 static char 	filename[MAXPATHLEN];
@@ -70,21 +76,21 @@ static char 	filename[MAXPATHLEN];
 static char	*vdp;		/* first trip mkdir for load/save */
 
 /*
- * count character to end of string or first space, whichever comes
- * first
+ * Count character to end of string or first space, whichever comes
+ * first.  In the special case of string caches, spaces are allowed.
  */
 static int
-get_hashlen(const char *str)
+get_hashlen(hdr_t *h, const char *str)
 {
     const char	*q = str;
 
-    while (*q && *q != ' ')
+    while (*q && (*q != ' ' || (h->hstate & CACHE_STRINGS) != 0))
 	q++;
     return (int)(q-str);
 }
 
 static unsigned int
-hash_str(const char *str, int len)
+hash_str(const signed char *str, int len)
 {
     return hash(str, len, 0);
 }
@@ -167,7 +173,7 @@ find_cache(pmInDom indom, int *sts)
 
     if ((h = (hdr_t *)malloc(sizeof(hdr_t))) == NULL) {
 	char	strbuf[20];
-	__pmNotifyErr(LOG_ERR, 
+	pmNotifyErr(LOG_ERR, 
 	     "find_cache: indom %s: unable to allocate memory for hdr_t",
 	     pmInDomStr_r(indom, strbuf, sizeof(strbuf)));
 	*sts = PM_ERR_GENERIC;
@@ -187,6 +193,7 @@ find_cache(pmInDom indom, int *sts)
     h->hstate = 0;
     for (i = 0; i < MAX_HASH_TRY; i++)
 	h->keyhash_cnt[i] = 0;
+    h->maxinst = DEFAULT_MAXINST;
     return h;
 }
 
@@ -306,7 +313,7 @@ static entry_t *
 find_name(hdr_t *h, const char *name, int *sts)
 {
     entry_t	*e;
-    int		hashlen = get_hashlen(name);
+    int		hashlen = get_hashlen(h, name);
 
     *sts = 0;
     for (e = h->first; e != NULL; e = e->next) {
@@ -330,7 +337,6 @@ find_inst(hdr_t *h, int inst)
     return e;
 }
 
-
 /*
  * supports find by instance identifier (name == NULL) else
  * find by instance name
@@ -357,12 +363,12 @@ find_entry(hdr_t *h, const char *name, int inst, int *sts)
 	/*
 	 * search by instance name
 	 */
-	int	hashlen = get_hashlen(name);
+	int	hashlen = get_hashlen(h, name);
 
 	if (h->ctl_name == NULL)
 	    /* no hash, use linear search */
 	    return find_name(h, name, sts);
-	for (e = h->ctl_name[hash_str(name, hashlen) & h->hbits]; e != NULL; e = e->h_name) {
+	for (e = h->ctl_name[hash_str((const signed char *)name, hashlen) & h->hbits]; e != NULL; e = e->h_name) {
 	    if (e->state != PMDA_CACHE_EMPTY) {
 		if ((*sts = name_eq(e, name, hashlen)))
 		    return e;
@@ -430,7 +436,7 @@ redo_hash(hdr_t *h, int resize)
 	    for (e = old_name[oldi]; e != NULL; ) {
 		t = e;
 		e = e->h_name;
-		i = hash_str(t->name, t->hashlen) & h->hbits;
+		i = hash_str((const signed char *)t->name, t->hashlen) & h->hbits;
 		t->h_name = h->ctl_name[i];
 		h->ctl_name[i] = t;
 	    }
@@ -552,7 +558,7 @@ insert_cache(hdr_t *h, const char *name, int inst, int *sts)
     entry_t	*last_e = NULL;
     char	*dup;
     int		i;
-    int		hashlen = get_hashlen(name);
+    int		hashlen = get_hashlen(h, name);
 
     *sts = 0;
 
@@ -572,24 +578,20 @@ insert_cache(hdr_t *h, const char *name, int inst, int *sts)
 	if (e != NULL) {
 	    if (name_eq(e, name, hashlen) != 1) {
 		/* instance id the same, different name */
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_INDOM) {
+		if (pmDebugOptions.indom) {
 		    fprintf(stderr, "pmdaCache: store: indom %s: instance %d ", pmInDomStr(h->indom), e->inst);
 		    fprintf(stderr, " in cache, name \"%s\" does not match new entry \"%s\"\n", e->name, name);
 		}
-#endif
 		*sts = PM_ERR_INST;
 	    }
 	    return e;
 	}
 	e = find_entry(h, name, PM_IN_NULL, sts);
 	if (e != NULL) {
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_INDOM) {
+	    if (pmDebugOptions.indom) {
 		fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\"", pmInDomStr(h->indom), e->name);
 		fprintf(stderr, " in cache, id %d does not match new entry %d\n", e->inst, inst);
 	    }
-#endif
 	    *sts = PM_ERR_INST;
 	    return e;
 	}
@@ -603,7 +605,7 @@ insert_cache(hdr_t *h, const char *name, int inst, int *sts)
 
     if ((dup = strdup(name)) == NULL) {
 	char	strbuf[20];
-	__pmNotifyErr(LOG_ERR, 
+	pmNotifyErr(LOG_ERR, 
 	     "insert_cache: indom %s: unable to allocate %d bytes for name: %s\n",
 	     pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)), (int)strlen(name), name);
 	*sts = PM_ERR_GENERIC;
@@ -616,7 +618,7 @@ insert_cache(hdr_t *h, const char *name, int inst, int *sts)
 	    if (last_e == NULL)
 		inst = 0;
 	    else {
-		if (last_e->inst == 0x7fffffff) {
+		if (last_e->inst == h->maxinst) {
 		    /*
 		     * overflowed inst identifier, need to shift to 
 		     * ins_mode == 1
@@ -634,12 +636,12 @@ retry:
 	    for (e = h->first; e != NULL; e = e->next) {
 		if (inst < e->inst)
 		    break;
-		if (inst == 0x7fffffff) {
+		if (inst == h->maxinst) {
 		    /*
 		     * 2^32-1 is the maximum number of instances we can have
 		     */
 		    char	strbuf[20];
-		    __pmNotifyErr(LOG_ERR, 
+		    pmNotifyErr(LOG_ERR, 
 			 "insert_cache: indom %s: too many instances",
 			 pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)));
 		    *sts = PM_ERR_GENERIC;
@@ -654,7 +656,7 @@ retry:
 
     if ((e = (entry_t *)malloc(sizeof(entry_t))) == NULL) {
 	char	strbuf[20];
-	__pmNotifyErr(LOG_ERR, 
+	pmNotifyErr(LOG_ERR, 
 	     "insert_cache: indom %s: unable to allocate memory for entry_t",
 	     pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)));
 	*sts = PM_ERR_GENERIC;
@@ -674,7 +676,7 @@ retry:
     }
     e->inst = inst;
     e->name = dup;
-    e->hashlen = get_hashlen(dup);
+    e->hashlen = get_hashlen(h, dup);
     e->key = NULL;
     e->state = PMDA_CACHE_INACTIVE;
     e->private = NULL;
@@ -697,7 +699,7 @@ retry:
 
     /* link into the name hash list, if any */
     if (h->ctl_name != NULL) {
-	i = hash_str(e->name, e->hashlen) & h->hbits;
+	i = hash_str((const signed char *)e->name, e->hashlen) & h->hbits;
 	e->h_name = h->ctl_name[i];
 	h->ctl_name[i] = e;
     }
@@ -721,30 +723,55 @@ load_cache(hdr_t *h)
     char	buf[1024];	/* input line buffer, is this big enough? */
     char	*p;
     int		sts;
-    int		sep = __pmPathSeparator();
+    int		sep = pmPathSeparator();
     char	strbuf[20];
 
     if (vdp == NULL) {
-	vdp = pmGetConfig("PCP_VAR_DIR");
-	snprintf(filename, sizeof(filename),
+	if ((vdp = pmGetOptionalConfig("PCP_VAR_DIR")) == NULL)
+	    return PM_ERR_GENERIC;
+	pmsprintf(filename, sizeof(filename),
 		"%s%c" "config" "%c" "pmda", vdp, sep, sep);
 	mkdir2(filename, 0755);
     }
 
-    snprintf(filename, sizeof(filename), "%s%cconfig%cpmda%c%s",
-		vdp, sep, sep, sep, pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)));
+    pmsprintf(filename, sizeof(filename), "%s%cconfig%cpmda%c%s",
+		vdp, sep, sep, sep,
+		pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)));
     if ((fp = fopen(filename, "r")) == NULL)
 	return -oserror();
     if (fgets(buf, sizeof(buf), fp) == NULL) {
-	__pmNotifyErr(LOG_ERR, 
+	pmNotifyErr(LOG_ERR, 
 	     "pmdaCacheOp: %s: empty file?", filename);
 	fclose(fp);
 	return PM_ERR_GENERIC;
     }
-    s = sscanf(buf, "%d %d", &x, &h->ins_mode);
-    if (s != 2 || x != 1 || h->ins_mode < 0 || h->ins_mode > 1) {
-	__pmNotifyErr(LOG_ERR, 
-	     "pmdaCacheOp: %s: illegal first record: %s",
+    /* First grab the file version. */
+    s = sscanf(buf, "%d ", &x);
+    if (s != 1 || x <= 0 || x > CACHE_VERSION) {
+	pmNotifyErr(LOG_ERR, 
+	     "pmdaCacheOp: %s: illegal cache header record: %s",
+	     filename, buf);
+	fclose(fp);
+	return PM_ERR_GENERIC;
+    }	
+
+    /* Based on the file version, grab the entire line. */
+    switch (x) {
+	case CACHE_VERSION1:
+	    h->maxinst = DEFAULT_MAXINST;
+	    s = sscanf(buf, "%d %d", &x, &h->ins_mode);
+	    if (s != 2)
+		s = 0;
+	    break;
+	default:
+	    s = sscanf(buf, "%d %d %d", &x, &h->ins_mode, &h->maxinst);
+	    if (s != 3)
+		s = 0;
+	    break;
+    }
+    if (s == 0 || h->ins_mode < 0 || h->ins_mode > 1 || h->maxinst < 0) {
+	pmNotifyErr(LOG_ERR, 
+	     "pmdaCacheOp: %s: illegal cache header record: %s",
 	     filename, buf);
 	fclose(fp);
 	return PM_ERR_GENERIC;
@@ -790,9 +817,10 @@ load_cache(hdr_t *h)
 	     */
 	    keylen = (pend - p) / 2;
 	    if ((key = malloc(keylen)) == NULL) {
-		__pmNotifyErr(LOG_ERR, 
+		pmNotifyErr(LOG_ERR, 
 		     "load_cache: indom %s: unable to allocate memory for keylen=%d",
 		     pmInDomStr(h->indom), keylen);
+		fclose(fp);
 		return PM_ERR_GENERIC;
 	    }
 	    q = key;
@@ -811,7 +839,7 @@ load_cache(hdr_t *h)
 	}
 	if (*p == '\0') {
 bad:
-	    __pmNotifyErr(LOG_ERR, 
+	    pmNotifyErr(LOG_ERR, 
 		 "pmdaCacheOp: %s: illegal record: %s",
 		 filename, buf);
 	    if (key) free(key);
@@ -825,7 +853,7 @@ bad:
 	    return sts;
 	}
 	if (sts != 0) {
-	    __pmNotifyErr(LOG_WARNING,
+	    pmNotifyErr(LOG_WARNING,
 		"pmdaCacheOp: %s: loading instance %d (\"%s\") ignored, already in cache as %d (\"%s\")",
 		filename, inst, p, e->inst, e->name);
 	}
@@ -835,12 +863,10 @@ bad:
     }
     fclose(fp);
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_INDOM) {
+    if (pmDebugOptions.indom) {
 	fprintf(stderr, "After PMDA_CACHE_LOAD\n");
 	dump(stderr, h, 0);
     }
-#endif
 
     return cnt;
 }
@@ -852,26 +878,29 @@ save_cache(hdr_t *h, int hstate)
     entry_t	*e;
     int		cnt;
     time_t	now;
-    int		sep = __pmPathSeparator();
+    int		sep = pmPathSeparator();
+    int		state = h->hstate & ~CACHE_STRINGS;
     char	strbuf[20];
 
-    if ((h->hstate & hstate) == 0) {
+    if ((state & hstate) == 0) {
 	/* nothing to be done */
 	return 0;
     }
 
     if (vdp == NULL) {
-	vdp = pmGetConfig("PCP_VAR_DIR");
-	snprintf(filename, sizeof(filename),
+	if ((vdp = pmGetOptionalConfig("PCP_VAR_DIR")) == NULL)
+	    return PM_ERR_GENERIC;
+	pmsprintf(filename, sizeof(filename),
 		"%s%c" "config" "%c" "pmda", vdp, sep, sep);
 	mkdir2(filename, 0755);
     }
 
-    snprintf(filename, sizeof(filename), "%s%cconfig%cpmda%c%s",
-		vdp, sep, sep, sep, pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)));
+    pmsprintf(filename, sizeof(filename), "%s%cconfig%cpmda%c%s",
+		vdp, sep, sep, sep,
+		pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)));
     if ((fp = fopen(filename, "w")) == NULL)
 	return -oserror();
-    fprintf(fp, "%d %d\n", VERSION, h->ins_mode);
+    fprintf(fp, "%d %d %d\n", CACHE_VERSION, h->ins_mode, h->maxinst);
 
     now = time(NULL);
     cnt = 0;
@@ -893,17 +922,15 @@ save_cache(hdr_t *h, int hstate)
 	cnt++;
     }
     fclose(fp);
-    h->hstate = 0;
+    h->hstate &= ~(DIRTY_INSTANCE | DIRTY_STAMP);
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_INDOM) {
+    if (pmDebugOptions.indom) {
 	fprintf(stderr, "After cache_save hstate={");
 	if (hstate & DIRTY_INSTANCE) fprintf(stderr, "DIRTY_INSTANCE");
 	if (hstate & DIRTY_STAMP) fprintf(stderr, "DIRTY_STAMP");
 	fprintf(stderr, "}\n");
 	dump(stderr, h, 0);
     }
-#endif
 
     return cnt;
 }
@@ -943,14 +970,12 @@ store(pmInDom indom, int flags, const char *name, pmInDom inst, int keylen, cons
     if ((e = find_entry(h, name, inst, &sts)) == NULL) {
 
 	if (flags != PMDA_CACHE_ADD) {
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_INDOM) {
+	    if (pmDebugOptions.indom) {
 		fprintf(stderr, "pmdaCache store: indom %s: instance \"%s\"", pmInDomStr(indom), name);
 		if (inst != PM_IN_NULL)
 		    fprintf(stderr, " (%d)", inst);
 		fprintf(stderr, " not in cache: flags=%d not allowed\n", flags);
 	    }
-#endif
 	    return PM_ERR_INST;
 	}
 
@@ -974,7 +999,7 @@ store(pmInDom indom, int flags, const char *name, pmInDom inst, int keylen, cons
 	    e->keylen = keylen;
 	    if (keylen > 0) {
 		if ((e->key = malloc(keylen)) == NULL) {
-		    __pmNotifyErr(LOG_ERR, 
+		    pmNotifyErr(LOG_ERR, 
 			 "store: indom %s: unable to allocate memory for keylen=%d",
 			 pmInDomStr(indom), keylen);
 		    return PM_ERR_GENERIC;
@@ -1066,15 +1091,13 @@ pmdaCacheStoreKey(pmInDom indom, int flags, const char *name, int keylen, const
 	 * if keys are not equal => failure
 	 */
 	if (key_eq(e, mykeylen, mykey) == 0) {
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_INDOM) {
+	    if (pmDebugOptions.indom) {
 		fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\" (%d) in cache ", pmInDomStr(indom), e->name, e->inst);
 		KeyStr(stderr, e->keylen, (const char *)e->key);
 		fprintf(stderr, " does not match new entry ");
 		KeyStr(stderr, mykeylen, mykey);
 		fputc('\n', stderr);
 	    }
-#endif
 	    return PM_ERR_INST;
 	}
 	/* keys the same, use inst from existing entry */
@@ -1083,7 +1106,7 @@ pmdaCacheStoreKey(pmInDom indom, int flags, const char *name, int keylen, const
     else {
 	/* we're in the inst guessing game ... */
 	for (i = 0; i < MAX_HASH_TRY; i++) {
-	    try = hash(mykey, mykeylen, try);
+	    try = hash((const signed char *)mykey, mykeylen, try);
 	    /* strip top bit ... instance id must be positive */
 	    inst = try & ~(1 << (8*sizeof(__uint32_t)-1));
 	    e = find_entry(h, NULL, inst, &sts);
@@ -1110,31 +1133,27 @@ pmdaCacheStoreKey(pmInDom indom, int flags, const char *name, int keylen, const
 	    if (strcmp(e->name, name) == 0) {
 		if (key_eq(e, mykeylen, mykey) == 1)
 		    break;
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_INDOM) {
+		if (pmDebugOptions.indom) {
 		    fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\" (%d) in cache, ", pmInDomStr(indom), e->name, e->inst);
 		    KeyStr(stderr, e->keylen, (const char *)e->key);
 		    fprintf(stderr, " does not match new entry ");
 		    KeyStr(stderr, mykeylen, mykey);
 		    fputc('\n', stderr);
 		}
-#endif
 		return PM_ERR_INST;
 	    }
 	    else if (key_eq(e, mykeylen, mykey) == 1) {
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_INDOM) {
+		if (pmDebugOptions.indom) {
 		    fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance %d ", pmInDomStr(indom), e->inst);
 		    KeyStr(stderr, e->keylen, (const char *)e->key);
 		    fprintf(stderr, " in cache, name \"%s\" does not match new entry \"%s\"\n", e->name, name);
 		}
-#endif
 		return PM_ERR_INST;
 	    }
 	}
 	if (i == MAX_HASH_TRY) {
 	    /* failed after MAX_HASH_TRY rehash attempts ... */
-	    __pmNotifyErr(LOG_ERR, 
+	    pmNotifyErr(LOG_ERR, 
 		 "pmdaCacheStoreKey: indom %s: unable allocate a new id for instance \"%s\" based on a key of %d bytes\n",
 		 pmInDomStr(h->indom), name, keylen);
 	    return PM_ERR_GENERIC;
@@ -1181,6 +1200,13 @@ int pmdaCacheOp(pmInDom indom, int op)
 	case PMDA_CACHE_SYNC:
 	    return save_cache(h, DIRTY_INSTANCE|DIRTY_STAMP);
 
+	case PMDA_CACHE_STRINGS:
+	    /* must be set before any cache entries are added */
+	    if (h->nentry > 0)
+		return -E2BIG;
+	    h->hstate |= CACHE_STRINGS;
+	    return 0;
+
 	case PMDA_CACHE_ACTIVE:
 	    sts = 0;
 	    for (e = h->first; e != NULL; e = e->next) {
@@ -1361,13 +1387,11 @@ int pmdaCacheLookupKey(pmInDom indom, const char *name, int keylen, const void *
 	}
     }
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_INDOM) {
+    if (pmDebugOptions.indom) {
 	    fprintf(stderr, "pmdaCacheLookupKey: indom %s: ", pmInDomStr(h->indom));
 	    KeyStr(stderr, mykeylen, mykey);
 	    fprintf(stderr, ": no matching key in cache\n");
 	}
-#endif
     return PM_ERR_INST;
 }
 
@@ -1402,133 +1426,30 @@ int pmdaCachePurge(pmInDom indom, time_t recent)
     return cnt;
 }
 
-/*
---------------------------------------------------------------------
-lookup2.c, by Bob Jenkins, December 1996, Public Domain.
-hash(), hash2(), hash3, and mix() are externally useful functions.
-Routines to test the hash are included if SELF_TEST is defined.
-You can use this free for any purpose.  It has no warranty.
---------------------------------------------------------------------
-*/
-
-#include <stdio.h>
-#include <stddef.h>
-#include <stdlib.h>
-typedef  unsigned long  int  ub4;   /* unsigned 4-byte quantities */
-typedef  unsigned       char ub1;
-
-#define hashsize(n) ((ub4)1<<(n))
-#define hashmask(n) (hashsize(n)-1)
-
-/*
---------------------------------------------------------------------
-mix -- mix 3 32-bit values reversibly.
-For every delta with one or two bit set, and the deltas of all three
-  high bits or all three low bits, whether the original value of a,b,c
-  is almost all zero or is uniformly distributed,
-* If mix() is run forward or backward, at least 32 bits in a,b,c
-  have at least 1/4 probability of changing.
-* If mix() is run forward, every bit of c will change between 1/3 and
-  2/3 of the time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
-mix() was built out of 36 single-cycle latency instructions in a 
-  structure that could supported 2x parallelism, like so:
-      a -= b; 
-      a -= c; x = (c>>13);
-      b -= c; a ^= x;
-      b -= a; x = (a<<8);
-      c -= a; b ^= x;
-      c -= b; x = (b>>13);
-      ...
-  Unfortunately, superscalar Pentiums and Sparcs can't take advantage 
-  of that parallelism.  They've also turned some of those single-cycle
-  latency instructions into multi-cycle latency instructions.  Still,
-  this is the fastest good hash I could find.  There were about 2^^68
-  to choose from.  I only looked at a billion or so.
---------------------------------------------------------------------
-*/
-#define mix(a,b,c) \
-{ \
-  a -= b; a -= c; a ^= (c>>13); \
-  b -= c; b -= a; b ^= (a<<8); \
-  c -= a; c -= b; c ^= (b>>13); \
-  a -= b; a -= c; a ^= (c>>12);  \
-  b -= c; b -= a; b ^= (a<<16); \
-  c -= a; c -= b; c ^= (b>>5); \
-  a -= b; a -= c; a ^= (c>>3);  \
-  b -= c; b -= a; b ^= (a<<10); \
-  c -= a; c -= b; c ^= (b>>15); \
-}
-
-/*
---------------------------------------------------------------------
-hash() -- hash a variable-length key into a 32-bit value
-  k     : the key (the unaligned variable-length array of bytes)
-  len   : the length of the key, counting by bytes
-  level : can be any 4-byte value
-Returns a 32-bit value.  Every bit of the key affects every bit of
-the return value.  Every 1-bit and 2-bit delta achieves avalanche.
-About 36+6len instructions.
-
-The best hash table sizes are powers of 2.  There is no need to do
-mod a prime (mod is sooo slow!).  If you need less than 32 bits,
-use a bitmask.  For example, if you need only 10 bits, do
-  h = (h & hashmask(10));
-In which case, the hash table should have hashsize(10) elements.
-
-If you are hashing n strings (ub1 **)k, do it like this:
-  for (i=0, h=0; i<n; ++i) h = hash( k[i], len[i], h);
-
-By Bob Jenkins, 1996.  bob_jenkins@burtleburtle.net.  You may use this
-code any way you wish, private, educational, or commercial.  It's free.
-
-See http://burlteburtle.net/bob/hash/evahash.html
-Use for hash table lookup, or anything where one collision in 2^32 is
-acceptable.  Do NOT use for cryptographic purposes.
---------------------------------------------------------------------
-*/
-
-static __uint32_t
-hash(const char *k, int length, __uint32_t initval)
+int pmdaCacheResize(pmInDom indom, int maximum)
 {
-   __uint32_t a,b,c,len;
+    hdr_t	*h;
+    int		sts;
+    entry_t	*e;
 
-   /* Set up the internal state */
-   len = length;
-   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
-   c = initval;           /* the previous hash value */
+    if (indom == PM_INDOM_NULL)
+	return PM_ERR_INDOM;
 
-   /*---------------------------------------- handle most of the key */
-   while (len >= 12)
-   {
-      a += (k[0] +((__uint32_t)k[1]<<8) +((__uint32_t)k[2]<<16)
-      +((__uint32_t)k[3]<<24));
-      b += (k[4] +((__uint32_t)k[5]<<8) +((__uint32_t)k[6]<<16)
-      +((__uint32_t)k[7]<<24));
-      c += (k[8] +((__uint32_t)k[9]<<8)
-      +((__uint32_t)k[10]<<16)+((__uint32_t)k[11]<<24));
-      mix(a,b,c);
-      k += 12; len -= 12;
-   }
+    if ((h = find_cache(indom, &sts)) == NULL)
+	return sts;
 
-   /*------------------------------------- handle the last 11 bytes */
-   c += length;
-   switch(len)              /* all the case statements fall through */
-   {
-   case 11: c+=((__uint32_t)k[10]<<24);
-   case 10: c+=((__uint32_t)k[9]<<16);
-   case 9 : c+=((__uint32_t)k[8]<<8);
-      /* the first byte of c is reserved for the length */
-   case 8 : b+=((__uint32_t)k[7]<<24);
-   case 7 : b+=((__uint32_t)k[6]<<16);
-   case 6 : b+=((__uint32_t)k[5]<<8);
-   case 5 : b+=k[4];
-   case 4 : a+=((__uint32_t)k[3]<<24);
-   case 3 : a+=((__uint32_t)k[2]<<16);
-   case 2 : a+=((__uint32_t)k[1]<<8);
-   case 1 : a+=k[0];
-     /* case 0: nothing left to add */
-   }
-   mix(a,b,c);
-   /*-------------------------------------------- report the result */
-   return c;
+    if (maximum < 0)
+	return PM_ERR_SIGN;
+
+    /* Find the largest inst in the queue. */
+    for (e = h->first; e != NULL; e = e->next) {
+	/* If the new maximum is smaller than an existing inst, error. */
+	if (maximum < e->inst)
+	    return PM_ERR_TOOBIG;
+    }
+    h->maxinst = maximum;
+    /* The timestamp doesn't really need updating, but the cache
+       header does now that we've modified the maximum value. */
+    h->hstate |= DIRTY_STAMP;
+    return 0;
 }
