@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013 Red Hat.
+ * Copyright (c) 2013-2014,2017-2018 Red Hat.
  * Copyright (c) 1995-2000 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -14,12 +14,12 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "pmda.h"
 #include "libdefs.h"
 
 /*
- * count the number of instances in an instance domain
+ * Count the number of instances in an instance domain
  */
 
 int
@@ -42,22 +42,20 @@ __pmdaCntInst(pmInDom indom, pmdaExt *pmda)
 	}
 	if (i == pmda->e_nindoms) {
 	    char	strbuf[20];
-	    __pmNotifyErr(LOG_WARNING, "cntinst: unknown indom %s", pmInDomStr_r(indom, strbuf, sizeof(strbuf)));
+	    pmNotifyErr(LOG_WARNING, "__pmdaCntInst: unknown indom %s", pmInDomStr_r(indom, strbuf, sizeof(strbuf)));
 	}
     }
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_INDOM) {
+    if (pmDebugOptions.indom) {
 	char	strbuf[20];
 	fprintf(stderr, "__pmdaCntInst(indom=%s) -> %d\n", pmInDomStr_r(indom, strbuf, sizeof(strbuf)), sts);
     }
-#endif
 
     return sts;
 }
 
 /*
- * commence a new round of instance selection
+ * Commence a new round of instance selection
  */
 
 static pmdaIndom	last;
@@ -113,18 +111,16 @@ __pmdaStartInst(pmInDom indom, pmdaExt *pmda)
 	}
     }
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_INDOM) {
+    if (pmDebugOptions.indom) {
 	char	strbuf[20];
 	fprintf(stderr, "__pmdaStartInst(indom=%s) e_ordinal=%d\n",
 	    pmInDomStr_r(indom, strbuf, sizeof(strbuf)), pmda->e_ordinal);
     }
-#endif
     return;
 }
 
 /* 
- * select the next instance
+ * Select the next instance
  */
 
 int
@@ -147,13 +143,11 @@ __pmdaNextInst(int *inst, pmdaExt *pmda)
 		pmda->e_ordinal++;
 		if (__pmInProfile(pmda->e_idp->it_indom, pmda->e_prof, myinst)) {
 		    *inst = myinst;
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_INDOM) {
+		    if (pmDebugOptions.indom) {
 			char	strbuf[20];
 			fprintf(stderr, "__pmdaNextInst(indom=%s) -> %d e_ordinal=%d (cache)\n",
 			    pmInDomStr_r(pmda->e_idp->it_indom, strbuf, sizeof(strbuf)), myinst, pmda->e_ordinal);
 		    }
-#endif
 		    return 1;
 		}
 	    }
@@ -166,13 +160,11 @@ __pmdaNextInst(int *inst, pmdaExt *pmda)
 				 pmda->e_idp->it_set[j].i_inst)) {
 		    *inst = pmda->e_idp->it_set[j].i_inst;
 		    pmda->e_ordinal = j+1;
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_INDOM) {
+		    if (pmDebugOptions.indom) {
 			char	strbuf[20];
 			fprintf(stderr, "__pmdaNextInst(indom=%s) -> %d e_ordinal=%d\n",
 			    pmInDomStr_r(pmda->e_idp->it_indom, strbuf, sizeof(strbuf)), *inst, pmda->e_ordinal);
 		    }
-#endif
 		    return 1;
 		}
 	    }
@@ -182,6 +174,101 @@ __pmdaNextInst(int *inst, pmdaExt *pmda)
     return 0;
 }
 
+static int
+__pmdaCountInst(pmDesc *dp, pmdaExt *pmda)
+{
+    int		inst, numval = 0;
+
+    if (dp->indom != PM_INDOM_NULL) {
+	/* count instances in indom */
+	__pmdaStartInst(dp->indom, pmda);
+	while (__pmdaNextInst(&inst, pmda))
+	    numval++;
+    }
+    else {
+	/* singular instance domains */
+	numval = 1;
+    }
+    return numval;
+}
+
+
+/*
+ * Helper routines for performing metric table searches.
+ *
+ * There are currently three ways - using the PMID hash that hangs off
+ * of the e_ext structure, using the direct mapping mechanism (require
+ * PMIDs be allocated one after-the-other, all in one cluster), or via
+ * a linear search of the metric table array.
+ */
+
+static pmdaMetric *
+__pmdaHashedSearch(pmID pmid, __pmHashCtl *hash)
+{
+    __pmHashNode *node;
+
+    if ((node = __pmHashSearch(pmid, hash)) == NULL)
+	return NULL;
+    return (pmdaMetric *)node->data;
+}
+
+static pmdaMetric *
+__pmdaDirectSearch(pmID pmid, pmdaExt *pmda)
+{
+    __pmID_int	*pmidp = (__pmID_int *)&pmid;
+
+    /*
+     * pmidp->domain is correct ... PMCD guarantees this, but
+     * pmda->e_direct only works for a single cluster
+     */
+    if (pmidp->item < pmda->e_nmetrics && 
+	pmidp->cluster == 
+	((__pmID_int *)&pmda->e_metrics[pmidp->item].m_desc.pmid)->cluster) {
+	/* pmidp->item is unsigned, so must be >= 0 */
+	return &pmda->e_metrics[pmidp->item];
+    }
+    return NULL;
+}
+
+static pmdaMetric *
+__pmdaLinearSearch(pmID pmid, pmdaExt *pmda)
+{
+    int		i;
+
+    for (i = 0; i < pmda->e_nmetrics; i++) {
+	if (pmda->e_metrics[i].m_desc.pmid == pmid) {
+	    /* found the hard way */
+	    return &pmda->e_metrics[i];
+	}
+    }
+    return NULL;
+}
+
+static pmdaMetric *
+__pmdaMetricSearch(pmdaExt *pmda, pmID pmid, pmdaMetric *mbuf, e_ext_t *extp)
+{
+    pmdaMetric	*metap;
+
+    if (pmda->e_flags & PMDA_EXT_FLAG_HASHED)
+	metap = __pmdaHashedSearch(pmid, &extp->hashpmids);
+    else if (pmda->e_direct)
+	metap = __pmdaDirectSearch(pmid, pmda);
+    else
+	metap = __pmdaLinearSearch(pmid, pmda);
+
+    /* possibly a highly dynamic metric with null metrictab[] */
+    if (!metap) {
+	metap = mbuf;
+	memset(metap, 0, sizeof(pmdaMetric));
+
+	if (extp->dispatch->version.any.desc != NULL) {
+	    /* may need a temporary pmdaMetric for callback interfaces */
+	    (*(extp->dispatch->version.any.desc))(pmid, &mbuf->m_desc, pmda);
+        }
+    }
+    return metap;
+}
+
 /*
  * Save the profile away for use in __pmdaNextInst() during subsequent
  * fetches ... it is the _caller_ of pmdaProfile()'s responsibility to
@@ -195,23 +282,23 @@ __pmdaNextInst(int *inst, pmdaExt *pmda)
  */
 
 int
-pmdaProfile(__pmProfile *prof, pmdaExt *pmda)
+pmdaProfile(pmProfile *prof, pmdaExt *pmda)
 {
     pmda->e_prof = prof;
     return 0;
 }
 
 /*
- * return desciption of an instance or instance domain
+ * Return description of an instance or instance domain
  */
 
 int
-pmdaInstance(pmInDom indom, int inst, char *name, __pmInResult **result, pmdaExt *pmda)
+pmdaInstance(pmInDom indom, int inst, char *name, pmInResult **result, pmdaExt *pmda)
 {
     int			i;
     int			namelen;
     int			err = 0;
-    __pmInResult  	*res;
+    pmInResult  	*res;
     pmdaIndom		*idp = NULL;	/* initialize to pander to gcc */
     int			have_cache = 0;
     int			myinst;
@@ -234,7 +321,7 @@ pmdaInstance(pmInDom indom, int inst, char *name, __pmInResult **result, pmdaExt
 	idp = &pmda->e_indoms[i];
     }
 
-    if ((res = (__pmInResult *)malloc(sizeof(*res))) == NULL)
+    if ((res = (pmInResult *)malloc(sizeof(*res))) == NULL)
         return -oserror();
     res->indom = indom;
 
@@ -337,13 +424,11 @@ pmdaInstance(pmInDom indom, int inst, char *name, __pmInResult **result, pmdaExt
 		char *instname = idp->it_set[i].i_name;
 		if (strcmp(name, instname) == 0) {
 		    /* accept an exact match */
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_LIBPMDA) {
+		    if (pmDebugOptions.libpmda) {
 			fprintf(stderr, 
 				"pmdaInstance: exact match name=%s id=%d\n",
 				name, idp->it_set[i].i_inst);
 		    }
-#endif
 		    res->instlist[0] = idp->it_set[i].i_inst;
 		    break;
 		}
@@ -354,12 +439,10 @@ pmdaInstance(pmInDom indom, int inst, char *name, __pmInResult **result, pmdaExt
 		    if (p != NULL) {
 			int len = (int)(p - instname);
 			if (namelen == len && strncmp(name, instname, len) == 0) {
-#ifdef PCP_DEBUG
-			    if (pmDebug & DBG_TRACE_LIBPMDA) {
+			    if (pmDebugOptions.libpmda) {
 				fprintf(stderr, "pmdaInstance: matched argument name=\"%s\" with indom id=%d name=\"%s\" len=%d\n",
 				    name, idp->it_set[i].i_inst, instname, len);
 			    }
-#endif
 			    res->instlist[0] = idp->it_set[i].i_inst;
 			    break;
 			}
@@ -383,7 +466,28 @@ pmdaInstance(pmInDom indom, int inst, char *name, __pmInResult **result, pmdaExt
 }
 
 /*
- * resize the pmResult and call the e_callback for each metric instance
+ * The first byte of the (unused) timestamp field has been
+ * co-opted as a flags byte - now indicating state changes
+ * that have happened within a PMDA and that need to later
+ * be propogated through to any connected clients.
+ */
+
+static void
+__pmdaEncodeStatus(pmResult *result, unsigned char byte)
+{
+    unsigned char	*flags;
+
+    memset(&result->timestamp, 0, sizeof(result->timestamp));
+    if (byte) {
+	flags = (unsigned char *)&result->timestamp;
+	*flags |= byte;
+    }
+}
+
+#define PMDA_STATUS_CHANGE (PMDA_EXT_LABEL_CHANGE|PMDA_EXT_NAMES_CHANGE)
+
+/*
+ * Resize the pmResult and call the e_callback for each metric instance
  * required in the profile.
  */
 
@@ -396,15 +500,33 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
     int			need;
     int			inst;
     int			numval;
+    int			version;
+    unsigned char	flags;
     pmValueSet		*vset;
+    pmValueSet		*tmp_vset;
     pmDesc		*dp;
-    __pmID_int		*pmidp;
+    pmdaMetric          metabuf;
     pmdaMetric		*metap;
     pmAtomValue		atom;
     int			type;
+    int			lsts;
+    char		idbuf[20];
+    char		strbuf[20];
     e_ext_t		*extp = (e_ext_t *)pmda->e_ext;
 
-    if (extp->pmda_interface >= PMDA_INTERFACE_5)
+    if ((pmDebugOptions.libpmda) && (pmDebugOptions.desperate)) {
+	fprintf(stderr, "pmdaFetch(%d, pmid[0] %s", numpmid, pmIDStr_r(pmidlist[0], idbuf, sizeof(idbuf)));
+	if (numpmid > 1)
+	    fprintf(stderr, "... pmid[%d] %s", numpmid-1, pmIDStr_r(pmidlist[numpmid-1], idbuf, sizeof(idbuf)));
+	fprintf(stderr, ", ...) called\n");
+    }
+
+    if (extp->dispatch->version.any.ext != pmda)
+	fprintf(stderr, "Botch: pmdaFetch: PMDA domain=%d pmda=%p extp=%p backpointer=%p pmda-via-backpointer %p NOT EQUAL to pmda\n",
+	    pmda->e_domain, pmda, extp, extp->dispatch, extp->dispatch->version.any.ext);
+
+    version = extp->dispatch->comm.pmda_interface;
+    if (version >= PMDA_INTERFACE_5)
 	__pmdaSetContext(pmda->e_context);
 
     if (numpmid > extp->maxnpmids) {
@@ -416,70 +538,41 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 	    return -oserror();
 	extp->maxnpmids = numpmid;
     }
-
-    extp->res->timestamp.tv_sec = 0;
-    extp->res->timestamp.tv_usec = 0;
     extp->res->numpmid = numpmid;
 
+    flags = 0;
+    if (version >= PMDA_INTERFACE_7 && (pmda->e_flags & PMDA_STATUS_CHANGE)) {
+	if (pmda->e_flags & PMDA_EXT_LABEL_CHANGE)
+	    flags |= PMCD_LABEL_CHANGE;
+	if (pmda->e_flags & PMDA_EXT_NAMES_CHANGE)
+	    flags |= PMCD_NAMES_CHANGE;
+	if (pmDebugOptions.libpmda)
+	    fprintf(stderr, "pmdaFetch flags pmda=0x%x to pmcd=0x%x\n",
+			    pmda->e_flags, (int)flags);
+    }
+    __pmdaEncodeStatus(extp->res, flags);
+
+    /* Look up the pmDesc for the incoming pmids in our pmdaMetrics tables,
+       if present.  Fall back to .desc callback if not found (for highly
+       dynamic pmdas). */
     for (i = 0; i < numpmid; i++) {
-
-    	dp = NULL;
-	metap = NULL;
-	pmidp = (__pmID_int *)&pmidlist[i];
-
-	if (pmda->e_direct) {
-/*
- * pmidp->domain is correct ... PMCD gurantees this, but
- * next to check the cluster
- */
-	    if (pmidp->item < pmda->e_nmetrics &&
-		pmidlist[i] == pmda->e_metrics[pmidp->item].m_desc.pmid) {
-/* 
- * pmidp->item is unsigned, so must be >= 0 
- */
-		metap = &pmda->e_metrics[pmidp->item];
-		dp = &(metap->m_desc);
-	    }
-	}
+	metap = __pmdaMetricSearch(pmda, pmidlist[i], &metabuf, extp);
+	/*
+	 * if search failed, then metap == metabuf, and metabuf.m_desc.pmid
+	 * will be zero
+	 */
+	dp = &(metap->m_desc);
+	if (dp->pmid != 0)
+	    numval = __pmdaCountInst(dp, pmda);
 	else {
-	    for (j = 0; j < pmda->e_nmetrics; j++) {
-		if (pmidlist[i] == pmda->e_metrics[j].m_desc.pmid) {
-/*
- * found the hard way 
- */
-		    metap = &pmda->e_metrics[j];
-		    dp = &(metap->m_desc);
-		    break;
-		}
-	    }
-	}
-	
-	if (dp == NULL) {
 	    /* dynamic name metrics may often vanish, avoid log spam */
-	    if (extp->pmda_interface < PMDA_INTERFACE_4) {
-		char	strbuf[20];
-		__pmNotifyErr(LOG_ERR,
+	    if (version < PMDA_INTERFACE_4) {
+		pmNotifyErr(LOG_ERR,
 			"pmdaFetch: Requested metric %s is not defined",
 			 pmIDStr_r(pmidlist[i], strbuf, sizeof(strbuf)));
 	    }
 	    numval = PM_ERR_PMID;
 	}
-	else {
-
-	    if (dp->indom != PM_INDOM_NULL) {
-		/* count instances in the profile */
-		numval = 0;
-		/* count instances in indom */
-		__pmdaStartInst(dp->indom, pmda);
-		while (__pmdaNextInst(&inst, pmda)) {
-		    numval++;
-		}
-	    }
-	    else {
-		/* singular instance domains */
-		numval = 1;
-	    }
-	}
 
 	/* Must use individual malloc()s because of pmFreeResult() */
 	if (numval >= 1)
@@ -510,49 +603,47 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 	    if (j == numval) {
 		/* more instances than expected! */
 		numval++;
-		extp->res->vset[i] = vset = (pmValueSet *)realloc(vset,
+		extp->res->vset[i] = tmp_vset = (pmValueSet *)realloc(vset,
 			    sizeof(pmValueSet) + (numval - 1)*sizeof(pmValue));
-		if (vset == NULL) {
+		if (tmp_vset == NULL) {
+		    free(vset);
+		    vset = NULL;
 		    sts = -oserror();
 		    goto error;
 		}
+		vset = tmp_vset;
 	    }
 	    vset->vlist[j].inst = inst;
 
 	    if ((sts = (*(pmda->e_fetchCallBack))(metap, inst, &atom)) < 0) {
-
+		pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf));
 		if (sts == PM_ERR_PMID) {
-		    char	strbuf[20];
-		    __pmNotifyErr(LOG_ERR, 
+		    pmNotifyErr(LOG_ERR, 
 		        "pmdaFetch: PMID %s not handled by fetch callback\n",
-				 pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf)));
+				strbuf);
 		}
 		else if (sts == PM_ERR_INST) {
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_LIBPMDA) {
-			char	strbuf[20];
-			__pmNotifyErr(LOG_ERR,
+		    if (pmDebugOptions.libpmda) {
+			pmNotifyErr(LOG_ERR,
 			    "pmdaFetch: Instance %d of PMID %s not handled by fetch callback\n",
-				     inst, pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf)));
+				    inst, strbuf);
 		    }
-#endif
 		}
 		else if (sts == PM_ERR_APPVERSION ||
-			(sts == PM_ERR_AGAIN) ||
-			(sts == PM_ERR_PERMISSION)) {
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_LIBPMDA) {
-			char	strbuf[20];
-			__pmNotifyErr(LOG_ERR,
-				 "pmdaFetch: Unavailable metric PMID %s\n",
-				 pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf)));
+			 sts == PM_ERR_PERMISSION ||
+			 sts == PM_ERR_AGAIN ||
+			 sts == PM_ERR_NYI) {
+		    if (pmDebugOptions.libpmda) {
+			pmNotifyErr(LOG_ERR,
+			     "pmdaFetch: Unavailable metric PMID %s[%d]\n",
+				    strbuf, inst);
 		    }
-#endif
 		}
-		else
-		    __pmNotifyErr(LOG_ERR,
-				 "pmdaFetch: Fetch callback error: %s\n",
-				 pmErrStr(sts));
+		else {
+		    pmNotifyErr(LOG_ERR,
+			"pmdaFetch: Fetch callback error from metric PMID %s[%d]: %s\n",
+				strbuf, inst, pmErrStr(sts));
+		}
 	    }
 	    else {
 		/*
@@ -567,34 +658,26 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 		 *	== 2 (PMDA_FETCH_DYNAMIC) => OK and free(atom.vp)
 		 *	     after __pmStuffValue() called
 		 */
-		if (extp->pmda_interface == PMDA_INTERFACE_2 ||
-		    (extp->pmda_interface >= PMDA_INTERFACE_3 && sts > 0)) {
-		    int		lsts;
+		if ((version == PMDA_INTERFACE_2) || (version >= PMDA_INTERFACE_3 && sts > 0)) {
 
 		    if ((lsts = __pmStuffValue(&atom, &vset->vlist[j], type)) == PM_ERR_TYPE) {
-			char	strbuf[20];
-			char	st2buf[20];
-			__pmNotifyErr(LOG_ERR, 
-				     "pmdaFetch: Descriptor type (%s) for metric %s is bad",
-				     pmTypeStr_r(type, strbuf, sizeof(strbuf)),
-				     pmIDStr_r(dp->pmid, st2buf, sizeof(st2buf)));
+			pmNotifyErr(LOG_ERR, "pmdaFetch: Descriptor type (%s) for metric %s is bad",
+				    pmTypeStr_r(type, strbuf, sizeof(strbuf)),
+				    pmIDStr_r(dp->pmid, idbuf, sizeof(idbuf)));
 		    }
 		    else if (lsts >= 0) {
 			vset->valfmt = lsts;
 			j++;
 		    }
-		    if (extp->pmda_interface >= PMDA_INTERFACE_5 && sts == PMDA_FETCH_DYNAMIC) {
+		    if (version >= PMDA_INTERFACE_5 && sts == PMDA_FETCH_DYNAMIC) {
 			if (type == PM_TYPE_STRING)
 			    free(atom.cp);
 			else if (type == PM_TYPE_AGGREGATE)
 			    free(atom.vbp);
 			else {
-			    char	strbuf[20];
-			    char	st2buf[20];
-			    __pmNotifyErr(LOG_WARNING,
-					  "pmdaFetch: Attempt to free value for metric %s of wrong type %s\n",
-					  pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf)),
-					  pmTypeStr_r(type, st2buf, sizeof(st2buf)));
+			    pmNotifyErr(LOG_WARNING, "pmdaFetch: Attempt to free value for metric %s of wrong type %s\n",
+					pmIDStr_r(dp->pmid, idbuf, sizeof(idbuf)),
+					pmTypeStr_r(type, strbuf, sizeof(strbuf)));
 			}
 		    }
 		    if (lsts < 0)
@@ -607,12 +690,14 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 	    vset->numval = sts;
 	else
 	    vset->numval = j;
-
     }
+
+    /* success, we will send this PDU - safe to clear flags */
+    pmda->e_flags &= ~PMDA_STATUS_CHANGE;
     *resp = extp->res;
     return 0;
 
- error:
+error:
 
     if (i) {
 	extp->res->numpmid = i;
@@ -628,48 +713,37 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 int
 pmdaDesc(pmID pmid, pmDesc *desc, pmdaExt *pmda)
 {
-    int			j;
-    int			sts = 0;
     e_ext_t		*extp = (e_ext_t *)pmda->e_ext;
+    pmdaMetric		*metric;
+    char		strbuf[32];
+    int			version;
 
-    if (extp->pmda_interface >= PMDA_INTERFACE_5)
+    if ((pmDebugOptions.libpmda) && (pmDebugOptions.desperate)) {
+	char	dbgbuf[20];
+	fprintf(stderr, "pmdaDesc(%s, ...) called\n", pmIDStr_r(pmid, dbgbuf, sizeof(dbgbuf)));
+    }
+
+    version = extp->dispatch->comm.pmda_interface;
+    if (version >= PMDA_INTERFACE_5)
 	__pmdaSetContext(pmda->e_context);
 
-    if (pmda->e_direct) {
-	__pmID_int	*pmidp = (__pmID_int *)&pmid;
-	/*
-	 * pmidp->domain is correct ... PMCD gurantees this, but
-	 * pmda->e_direct only works for a single cluster
-	 */
-	if (pmidp->item < pmda->e_nmetrics && 
-	    pmidp->cluster == 
-	     	((__pmID_int *)&pmda->e_metrics[pmidp->item].m_desc.pmid)->cluster) {
-	    /* pmidp->item is unsigned, so must be >= 0 */
-	    *desc = pmda->e_metrics[pmidp->item].m_desc;
-	}
-	else {
-	    char	strbuf[20];
-	    __pmNotifyErr(LOG_ERR, "Requested metric %s is not defined",
-			 pmIDStr_r(pmid, strbuf, sizeof(strbuf)));
-	    sts = PM_ERR_PMID;
-	}
+    if (pmda->e_flags & PMDA_EXT_FLAG_HASHED)
+	metric = __pmdaHashedSearch(pmid, &extp->hashpmids);
+    else if (pmda->e_direct)
+	metric = __pmdaDirectSearch(pmid, pmda);
+    else
+	metric = __pmdaLinearSearch(pmid, pmda);
+
+    if (metric) {
+	*desc = metric->m_desc;
+	return 0;
     }
-    else {
-	for (j = 0; j < pmda->e_nmetrics; j++) {
-	    if (pmda->e_metrics[j].m_desc.pmid == pmid) {
-		/* found the hard way */
-		*desc = pmda->e_metrics[j].m_desc;
-		break;
-	    }
-	}
-	if (j == pmda->e_nmetrics) {
-	    char	strbuf[20];
-	    __pmNotifyErr(LOG_ERR, "Requested metric %s is not defined",
-			 pmIDStr_r(pmid, strbuf, sizeof(strbuf)));
-	    sts = PM_ERR_PMID;
-	}
-    }
-    return sts;
+
+    /* dynamic name metrics may often vanish, avoid log spam */
+    if (version < PMDA_INTERFACE_4)
+	pmNotifyErr(LOG_ERR, "pmdaDesc: Requested metric %s is not defined",
+			pmIDStr_r(pmid, strbuf, sizeof(strbuf)));
+    return PM_ERR_PMID;
 }
 
 /*
@@ -681,7 +755,7 @@ pmdaText(int ident, int type, char **buffer, pmdaExt *pmda)
 {
     e_ext_t		*extp = (e_ext_t *)pmda->e_ext;
 
-    if (extp->pmda_interface >= PMDA_INTERFACE_5)
+    if (extp->dispatch->comm.pmda_interface >= PMDA_INTERFACE_5)
 	__pmdaSetContext(pmda->e_context);
 
     if (pmda->e_help >= 0) {
@@ -696,6 +770,204 @@ pmdaText(int ident, int type, char **buffer, pmdaExt *pmda)
     return (*buffer == NULL) ? PM_ERR_TEXT : 0;
 }
 
+/*
+ * Provide default handlers to fill set(s) of labels for the context
+ * or requested ID of type domain, indom, cluster, item or instances.
+ */
+
+int
+pmdaLabel(int ident, int type, pmLabelSet **lpp, pmdaExt *pmda)
+{
+    e_ext_t		*extp = (e_ext_t *)pmda->e_ext;
+    pmLabelSet		*rlp, *lp = NULL;
+    size_t		size;
+    char		idbuf[32], *idp;
+    char		errbuf[PM_MAXERRMSGLEN];
+    int			sts = 0, count, inst, numinst;
+
+    if (extp->dispatch->comm.pmda_interface >= PMDA_INTERFACE_5)
+	__pmdaSetContext(pmda->e_context);
+
+    switch (type) {
+    case PM_LABEL_CONTEXT:
+	if (pmDebugOptions.labels)
+	    fprintf(stderr, "pmdaLabel: context %d labels request\n",
+			pmda->e_context);
+	if ((lp = *lpp) == NULL)	/* use default handler */
+	    return __pmGetContextLabels(lpp);
+	return pmdaAddLabelFlags(lp, type);
+
+    case PM_LABEL_DOMAIN:
+	if (pmDebugOptions.labels)
+	    fprintf(stderr, "pmdaLabel: domain %d (%s) labels request\n",
+			pmda->e_domain, pmda->e_name);
+	if ((lp = *lpp) == NULL)	/* use default handler */
+	    return __pmGetDomainLabels(pmda->e_domain, pmda->e_name, lpp);
+	return pmdaAddLabelFlags(lp, type);
+
+    case PM_LABEL_INDOM:
+	if (pmDebugOptions.labels)
+	    fprintf(stderr, "pmdaLabel: InDom %s labels request\n",
+			    pmInDomStr_r(ident, idbuf, sizeof(idbuf)));
+	if ((lp = *lpp) == NULL)	/* no default handler */
+	    return 0;
+	return pmdaAddLabelFlags(lp, type);
+
+    case PM_LABEL_CLUSTER:
+	if (pmDebugOptions.labels) {
+	    pmIDStr_r(ident, idbuf, sizeof(idbuf));
+	    idp = rindex(idbuf, '.');
+	    *idp = '\0';	/* drop the final (item) part */
+	    fprintf(stderr, "pmdaLabel: cluster %s labels request\n", idbuf);
+	}
+	if ((lp = *lpp) == NULL)	/* no default handler */
+	    return 0;
+	return pmdaAddLabelFlags(lp, type);
+
+    case PM_LABEL_ITEM:
+	if (pmDebugOptions.labels)
+	    fprintf(stderr, "pmdaLabel: cluster %s labels request\n",
+			    pmIDStr_r(ident, idbuf, sizeof(idbuf)));
+	if ((lp = *lpp) == NULL)	/* no default handler */
+	    return 0;
+	return pmdaAddLabelFlags(lp, type);
+
+    case PM_LABEL_INSTANCES:
+	if (extp->dispatch->comm.pmda_interface < PMDA_INTERFACE_7)
+	    return 0;
+
+	if (ident == PM_INDOM_NULL)
+	    numinst = 1;
+	else
+	    numinst = __pmdaCntInst(ident, pmda);
+
+	if (pmDebugOptions.labels)
+	    fprintf(stderr, "pmdaLabel: InDom %s %d instance labels request\n",
+			    pmInDomStr_r(ident, idbuf, sizeof(idbuf)), numinst);
+
+	if (numinst == 0)
+	    return 0;
+
+	/* allocate minimally-sized chunk of contiguous memory upfront */
+	size = numinst * sizeof(pmLabelSet);
+	if ((lp = (pmLabelSet *)malloc(size)) == NULL)
+	    return -oserror();
+	*lpp = lp;
+
+	inst = PM_IN_NULL;
+	if (ident != PM_INDOM_NULL) {
+	    __pmdaStartInst(ident, pmda);
+	    __pmdaNextInst(&inst, pmda);
+	}
+
+	count = 0;
+	do {
+	    if (count == numinst) {
+		/* more instances than expected! */
+		numinst++;
+		size = numinst * sizeof(pmLabelSet);
+		if ((rlp = (pmLabelSet *)realloc(*lpp, size)) == NULL)
+		    return -oserror();
+		*lpp = rlp;
+		lp = rlp + count;
+	    }
+	    memset(lp, 0, sizeof(*lp));
+
+	    if ((sts = (*(pmda->e_labelCallBack))(ident, inst, &lp)) < 0) {
+		pmInDomStr_r(ident, idbuf, sizeof(idbuf));
+		pmErrStr_r(sts, errbuf, sizeof(errbuf));
+		pmNotifyErr(LOG_DEBUG, "pmdaLabel: "
+				"InDom %s[%d]: %s\n", idbuf, inst, errbuf);
+	    }
+	    if ((lp->nlabels = sts) > 0)
+		pmdaAddLabelFlags(lp, type);
+	    lp->inst = inst;
+	    count++;
+	    lp++;
+
+	} while (ident != PM_INDOM_NULL && __pmdaNextInst(&inst, pmda));
+
+	return count;
+
+    default:
+	break;
+    }
+
+    return PM_ERR_TYPE;
+}
+
+int
+pmdaAddLabelFlags(pmLabelSet *lsp, int flags)
+{
+    int		i;
+
+    if (lsp == NULL)
+	return 0;
+    for (i = 0; i < lsp->nlabels; i++)
+	lsp->labels[i].flags |= flags;
+    return 1;
+}
+
+/*
+ * Add labels (name:value pairs) to a labelset, varargs style.
+ */
+
+int
+pmdaAddLabels(pmLabelSet **lsp, const char *fmt, ...)
+{
+    char		errbuf[PM_MAXERRMSGLEN];
+    char		buf[PM_MAXLABELJSONLEN];
+    va_list		arg;
+    int			sts;
+
+    va_start(arg, fmt);
+    sts = vsnprintf(buf, sizeof(buf), fmt, arg);
+    va_end(arg);
+    if (sts < 0)
+	return sts;
+    if (sts >= sizeof(buf))
+	buf[sizeof(buf)-1] = '\0';
+
+    if (pmDebugOptions.labels)
+	fprintf(stderr, "pmdaAddLabels: %s\n", buf);
+
+    if ((sts = __pmAddLabels(lsp, buf, 0)) < 0) {
+	pmNotifyErr(LOG_ERR, "pmdaAddLabels: %s (%s)\n", buf,
+		pmErrStr_r(sts, errbuf, sizeof(errbuf)));
+    }
+    return sts;
+}
+
+/*
+ * Add notes (optional name:value pairs) to a labelset, varargs style.
+ */
+
+int
+pmdaAddNotes(pmLabelSet **lsp, const char *fmt, ...)
+{
+    char		errbuf[PM_MAXERRMSGLEN];
+    char		buf[PM_MAXLABELJSONLEN];
+    va_list		arg;
+    int			sts;
+
+    va_start(arg, fmt);
+    sts = vsnprintf(buf, sizeof(buf), fmt, arg);
+    va_end(arg);
+    if (sts < 0)
+	return sts;
+    if (sts >= sizeof(buf))
+	buf[sizeof(buf)-1] = '\0';
+
+    if (pmDebugOptions.labels)
+	fprintf(stderr, "pmdaAddNotes: %s\n", buf);
+
+    if ((sts = __pmAddLabels(lsp, buf, PM_LABEL_OPTIONAL)) < 0) {
+	pmNotifyErr(LOG_ERR, "pmdaAddNotes: %s (%s)\n", buf,
+		pmErrStr_r(sts, errbuf, sizeof(errbuf)));
+    }
+    return sts;
+}
+
 /*
  * Tell PMCD there is nothing to store
  */
@@ -738,7 +1010,16 @@ pmdaChildren(const char *name, int traverse, char ***offspring, int **status, pm
 }
 
 int
-pmdaAttribute(int context, int attribute, const char *value, int size, pmdaExt *pmda)
+pmdaAttribute(int ctx, int attr, const char *value, int size, pmdaExt *pmda)
 {
-    return 0;	/* simply ignore everything by default */
+    if (pmDebugOptions.attr || pmDebugOptions.auth) {
+	char buffer[256];
+	if (!__pmAttrStr_r(attr, value, buffer, sizeof(buffer))) {
+	    pmNotifyErr(LOG_ERR, "Bad attr: ctx=%d, attr=%d\n", ctx, attr);
+	} else {
+	    buffer[sizeof(buffer)-1] = '\0';
+	    pmNotifyErr(LOG_INFO, "Attribute: ctx=%d %s", ctx, buffer);
+	}
+    }
+    return 0;
 }
