@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013 Red Hat.
+ * Copyright (c) 2013,2016,2018 Red Hat.
  * Copyright (c) 1999-2000 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -19,7 +19,10 @@
 #define HAVE_V_FOUR(interface)	((interface) >= PMDA_INTERFACE_4)
 #define HAVE_V_FIVE(interface)	((interface) >= PMDA_INTERFACE_5)
 #define HAVE_V_SIX(interface)	((interface) >= PMDA_INTERFACE_6)
-#define HAVE_ANY(interface)	((interface) <= PMDA_INTERFACE_6 && HAVE_V_TWO(interface))
+#define HAVE_V_SEVEN(interface)	((interface) >= PMDA_INTERFACE_7)
+#define HAVE_ANY(interface)	((interface) <= PMDA_INTERFACE_7 && HAVE_V_TWO(interface))
+
+struct dynamic;
 
 /*
  * Auxilliary structure used to save data from pmdaDSO or pmdaDaemon and
@@ -27,9 +30,23 @@
  * when multiple DSO PMDAs are in use
  */
 typedef struct {
-    int		pmda_interface;
-    pmResult	*res;			/* high-water allocation for */
-    int		maxnpmids;		/* pmResult for each PMDA */
+    pmdaInterface	*dispatch;	/* back pointer to our pmdaInterface */
+    pmResult		*res;		/* high-water allocation for */
+    int			maxnpmids;	/* pmResult for each PMDA */
+    __pmHashCtl		hashpmids;	/* hashed metrictab lookups */
+    int			ndynamics;	/* number of dynamics entries, below */
+    struct dynamic	*dynamics;	/* dynamic metric manipulation table */
 } e_ext_t;
 
+/*
+ * Local hash function
+ */
+extern __uint32_t hash(const signed char *, int, __uint32_t);
+
+/*
+ * These ones escaped via the exports file, but are only used within
+ * the libpcp_pmda library, so pull the definitions back from <pcp/pmda.h>
+ */
+PMDA_CALL extern __pmnsNode * pmdaNodeLookup(__pmnsNode *, const char *);
+
 #endif /* LIBDEFS_H */
