@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013,2017-2018 Red Hat.
  * Copyright (c) 1995-2000 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -13,25 +14,19 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "pmda.h"
 #include "libdefs.h"
 
-extern int __pmdaSetupPDU(int, int, char *);
-
-
 int
 __pmdaInFd(pmdaInterface *dispatch)
 {
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	return dispatch->version.four.ext->e_infd;
-    else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	return dispatch->version.two.ext->e_infd;
-    else {
-	__pmNotifyErr(LOG_CRIT, "PMDA interface version %d not supported",
-		     dispatch->comm.pmda_interface);
-        return -1;
-    }
+    if (HAVE_ANY(dispatch->comm.pmda_interface))
+	return dispatch->version.any.ext->e_infd;
+
+    pmNotifyErr(LOG_CRIT, "PMDA interface version %d not supported",
+		    dispatch->comm.pmda_interface);
+    return -1;
 }
 
 int
@@ -51,16 +46,18 @@ __pmdaMainPDU(pmdaInterface *dispatch)
     int			subtype;
     pmResult		*result;
     int			ctxnum;
-    __pmTimeval		when;
+    int			length;
+    pmTimeval		when;
     int			ident;
     int			type;
     pmInDom		indom;
     int			inst;
     char		*iname;
-    __pmInResult	*inres;
+    pmInResult		*inres;
+    pmLabelSet		*labels = NULL;
     char		*buffer;
-    __pmProfile  	*new_profile;
-    static __pmProfile	*profile = NULL;
+    pmProfile  		*new_profile;
+    static pmProfile	*profile = NULL;
     static int		first_time = 1;
     static pmdaExt	*pmda = NULL;
     int			pinpdu;
@@ -68,29 +65,24 @@ __pmdaMainPDU(pmdaInterface *dispatch)
     /* Initial version checks */
     if (first_time) {
 	if (dispatch->status != 0) {
-	    __pmNotifyErr(LOG_ERR, "PMDA Initialisation Failed");
+	    pmNotifyErr(LOG_ERR, "PMDA Initialisation Failed");
 	    return -1;
 	}
-	if (HAVE_V_FOUR(dispatch->comm.pmda_interface)) {
-	    pmda = dispatch->version.four.ext; /* PMDA Interface V.4 dynamic PMNS */
-	}
-	else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	    pmda = dispatch->version.two.ext; /* PMDA Interface V.2 or V.3 Extensions */
-	else {
-	    __pmNotifyErr(LOG_CRIT, "PMDA interface version %d not supported",
+	if (!HAVE_ANY(dispatch->comm.pmda_interface)) {
+	    pmNotifyErr(LOG_CRIT, "PMDA interface version %d not supported",
 			 dispatch->comm.pmda_interface);
 	    return -1;
 	}
+	pmda = dispatch->version.any.ext;
 	dispatch->comm.pmapi_version = PMAPI_VERSION;
 	first_time = 0;
     }
 
     pinpdu = sts = __pmGetPDU(pmda->e_infd, ANY_SIZE, TIMEOUT_NEVER, &pb);
     if (sts == 0)
-	/* End of File */
 	return PM_ERR_EOF;
     if (sts < 0) {
-	__pmNotifyErr(LOG_ERR, "IPC Error: %s\n", pmErrStr(sts));
+	pmNotifyErr(LOG_ERR, "IPC Error: %s\n", pmErrStr(sts));
 	return sts;
     }
 
@@ -105,44 +97,38 @@ __pmdaMainPDU(pmdaInterface *dispatch)
      * if defined, callback once per PDU to check availability, etc.
      */
     if (pmda->e_checkCallBack) {
-	int		i;
-	i = (*(pmda->e_checkCallBack))();
-	if (i < 0) {
-	    if (sts != PDU_PROFILE)
+	op_sts = (*(pmda->e_checkCallBack))();
+	if (op_sts < 0) {
+	    if (sts != PDU_PROFILE && sts != PDU_ATTR)
 		/* all other PDUs expect an ACK */
-		__pmSendError(pmda->e_outfd, FROM_ANON, i);
+		__pmSendError(pmda->e_outfd, FROM_ANON, op_sts);
 	    __pmUnpinPDUBuf(pb);
 	    return 0;
 	}
     }
 
     switch (sts) {
-	int	endsts;
-	char	strbuf[20];
-
     case PDU_ERROR:
 	/*
 	 * If __pmDecodeError() fails, just ignore it as no response PDU
 	 * is required nor expected.
 	 * Expect PM_ERR_NOTCONN to mark client context being closed.
 	 */
-	if (__pmDecodeError(pb, &endsts) >= 0) {
-	    if (endsts == PM_ERR_NOTCONN) {
+	if (__pmDecodeError(pb, &op_sts) >= 0) {
+	    if (op_sts == PM_ERR_NOTCONN) {
 		if (HAVE_V_FIVE(dispatch->comm.pmda_interface)) {
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_CONTEXT) {
-			__pmNotifyErr(LOG_DEBUG, "Received PDU_ERROR (end context %d)\n", dispatch->version.four.ext->e_context);
+		    if (pmDebugOptions.context) {
+			pmNotifyErr(LOG_DEBUG, "Received PDU_ERROR (end context %d)\n", dispatch->version.four.ext->e_context);
 		    }
-#endif
 		    if (pmda->e_endCallBack != NULL) {
 			(*(pmda->e_endCallBack))(dispatch->version.four.ext->e_context);
 		    }
 		}
 	    }
 	    else {
-		__pmNotifyErr(LOG_ERR,
+		pmNotifyErr(LOG_ERR,
 		      "%s: unexpected error pdu from pmcd: %s?\n",
-		      pmda->e_name, pmErrStr(endsts));
+		      pmda->e_name, pmErrStr(op_sts));
 	    }
 	}
 	break;
@@ -153,11 +139,9 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 	 * the correct profile, if required
 	 */
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_PROFILE\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_PROFILE\n");
 	}
-#endif
 
 	/*
 	 * free last profile received (if any)
@@ -166,10 +150,7 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 	if (__pmDecodeProfile(pb, &ctxnum, &new_profile) < 0) 
 	   break;
 
-	if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	    sts = dispatch->version.four.profile(new_profile, pmda);
-	else
-	    sts = dispatch->version.two.profile(new_profile, pmda);
+	sts = dispatch->version.any.profile(new_profile, pmda);
 	if (sts < 0) {
 	    __pmFreeProfile(new_profile);
 	} else {
@@ -184,41 +165,29 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 	 * the correct profile, if required
 	 */
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_FETCH\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_FETCH\n");
 	}
-#endif
 
 	sts = __pmDecodeFetch(pb, &ctxnum, &when, &npmids, &pmidlist);
-
-	/* Ignore "when"; pmcd should intercept archive log requests */
 	if (sts >= 0) {
-	    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-		sts = dispatch->version.four.fetch(npmids, pmidlist, &result, pmda);
-	    else
-		sts = dispatch->version.two.fetch(npmids, pmidlist, &result, pmda);
-
+	    sts = dispatch->version.any.fetch(npmids, pmidlist, &result, pmda);
 	    __pmUnpinPDUBuf(pmidlist);
 	}
 	if (sts < 0)
 	    __pmSendError(pmda->e_outfd, FROM_ANON, sts);
 	else {
-	    /* this is for PURIFY to prevent a UMR in __pmXmitPDU */
-	    result->timestamp.tv_sec = 0;
-	    result->timestamp.tv_usec = 0;
 	    __pmSendResult(pmda->e_outfd, FROM_ANON, result);
-	    (pmda->e_resultCallBack)(result);
+	    if (pmda->e_resultCallBack)
+		pmda->e_resultCallBack(result);
 	}
 	break;
 
     case PDU_PMNS_NAMES:
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_PMNS_NAMES\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_PMNS_NAMES\n");
 	}
-#endif
 
 	if ((sts = __pmDecodeNameList(pb, &npmids, &namelist, NULL)) >= 0) {
 	    if (HAVE_V_FOUR(dispatch->comm.pmda_interface)) {
@@ -232,10 +201,10 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 		    sts = dispatch->version.four.pmid(namelist[0], &pmid, pmda);
 	    }
 	    else {
-		/* Not INTERFACE_4 */
+		/* Not INTERFACE_4 or later */
 		sts = PM_ERR_NAME;
 	    }
-	    __pmUnpinPDUBuf(namelist);
+	    free(namelist);
 	}
 	if (sts < 0)
 	    __pmSendError(pmda->e_outfd, FROM_ANON, sts);
@@ -245,11 +214,9 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 
     case PDU_PMNS_CHILD:
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_PMNS_CHILD\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_PMNS_CHILD\n");
 	}
-#endif
 
 	if ((sts = __pmDecodeChildReq(pb, &name, &subtype)) >= 0) {
 	    if (HAVE_V_FOUR(dispatch->comm.pmda_interface)) {
@@ -265,6 +232,7 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 		/* Not INTERFACE_4 */
 		sts = PM_ERR_NAME;
 	    }
+	    free(name);
 	}
 	if (sts < 0)
 	    __pmSendError(pmda->e_outfd, FROM_ANON, sts);
@@ -276,11 +244,9 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 
     case PDU_PMNS_TRAVERSE:
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_PMNS_TRAVERSE\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_PMNS_TRAVERSE\n");
 	}
-#endif
 
 	if ((sts = __pmDecodeTraversePMNSReq(pb, &name)) >= 0) {
 	    if (HAVE_V_FOUR(dispatch->comm.pmda_interface)) {
@@ -305,11 +271,9 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 
     case PDU_PMNS_IDS:
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_PMNS_IDS\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_PMNS_IDS\n");
 	}
-#endif
 
 	sts = __pmDecodeIDList(pb, 1, &pmid, &op_sts);
 	if (sts >= 0)
@@ -332,46 +296,47 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 
     case PDU_DESC_REQ:
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_DESC_REQ\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_DESC_REQ\n");
 	}
-#endif
 
-	if ((sts = __pmDecodeDescReq(pb, &pmid)) >= 0) {
-	    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-		sts = dispatch->version.four.desc(pmid, &desc, pmda);
-	    else
-		sts = dispatch->version.two.desc(pmid, &desc, pmda);
-	}
+	if ((sts = __pmDecodeDescReq(pb, &pmid)) >= 0)
+	    sts = dispatch->version.any.desc(pmid, &desc, pmda);
 	if (sts < 0)
 	    __pmSendError(pmda->e_outfd, FROM_ANON, sts);
 	else
 	    __pmSendDesc(pmda->e_outfd, FROM_ANON, &desc);
 	break;
 
+    case PDU_LABEL_REQ:
+
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_LABEL_REQ\n");
+	}
+
+	if ((sts = __pmDecodeLabelReq(pb, &ident, &type)) >= 0 &&
+	    HAVE_V_SEVEN(dispatch->comm.pmda_interface)) {
+	    ctxnum = dispatch->version.seven.ext->e_context;
+	    sts = dispatch->version.seven.label(ident, type, &labels, pmda);
+	}
+	if (sts < 0)
+	    __pmSendError(pmda->e_outfd, FROM_ANON, sts);
+	else {
+	    if (sts > 0 && !(type & PM_LABEL_INSTANCES))
+		sts = 1;
+	    __pmSendLabel(pmda->e_outfd, FROM_ANON, ident, type, labels, sts);
+	    pmFreeLabelSets(labels, sts);
+	}
+	break;
+
     case PDU_INSTANCE_REQ:
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_INSTANCE_REQ\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_INSTANCE_REQ\n");
 	}
-#endif
 
-	if ((sts = __pmDecodeInstanceReq(pb, &when, &indom, &inst, 
-					 &iname)) >= 0) {
-	    /*
-	     * Note: when is ignored.
-	     *		If we get this far, we are _only_ dealing
-	     *		with current data (pmcd handles the other
-	     *		cases).
-	     */
-
-	    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-		sts = dispatch->version.four.instance(indom, inst, iname, &inres, pmda);
-	    else
-		sts = dispatch->version.two.instance(indom, inst, iname, &inres, pmda);
-	}
+	if ((sts = __pmDecodeInstanceReq(pb, &when, &indom, &inst, &iname)) >= 0)
+	    sts = dispatch->version.any.instance(indom, inst, iname, &inres, pmda);
 	if (sts < 0)
 	    __pmSendError(pmda->e_outfd, FROM_ANON, sts);
 	else {
@@ -384,18 +349,12 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 
     case PDU_TEXT_REQ:
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_TEXT_REQ\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_TEXT_REQ\n");
 	}
-#endif
 
-	if ((sts = __pmDecodeTextReq(pb, &ident, &type)) >= 0) {
-	    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-		sts = dispatch->version.four.text(ident, type, &buffer, pmda);
-	    else
-		sts = dispatch->version.two.text(ident, type, &buffer, pmda);
-	}
+	if ((sts = __pmDecodeTextReq(pb, &ident, &type)) >= 0)
+	    sts = dispatch->version.any.text(ident, type, &buffer, pmda);
 	if (sts < 0)
 	    __pmSendError(pmda->e_outfd, FROM_ANON, sts);
 	else {
@@ -405,50 +364,41 @@ __pmdaMainPDU(pmdaInterface *dispatch)
 
     case PDU_RESULT:
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_RESULT\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_RESULT\n");
 	}
-#endif
 
-	if ((sts = __pmDecodeResult(pb, &result)) >= 0) {
-	    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-		sts = dispatch->version.four.store(result, pmda);
-	    else
-		sts = dispatch->version.two.store(result, pmda);
-	}
+	if ((sts = __pmDecodeResult(pb, &result)) >= 0)
+	    sts = dispatch->version.any.store(result, pmda);
 	__pmSendError(pmda->e_outfd, FROM_ANON, sts);
 	pmFreeResult(result);
 	break;
 
     case PDU_CONTROL_REQ:
-	/*
-	 * The protocol followed by the PMCD is to send these as
-	 * advisory notices, but not to expect any response, and
-	 * in particular no response PDU is required.
-	 */
-
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    __pmNotifyErr(LOG_DEBUG, "Received PDU_CONTROL_REQ\n");
+	if (pmDebugOptions.libpmda) {
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_CONTROL_REQ\n");
 	}
-#endif
-
-#if 0
-	/*
-	 * this function moved to libpcp_dev.a - this isn't being used
-	 * by any PMDA, so its out of 2.0 libpcp.
-	 */
-	if ((sts = __pmDecodeControlReq(pb, &result, &control, &state, &delta)) >= 0) {
-	    __pmNotifyErr(LOG_ERR, "PDU_CONTROL_REQ not supported");
-	}
-#endif
 	break;
 
-    default:
-	__pmNotifyErr(LOG_ERR,
+    case PDU_ATTR:
+	if (pmDebugOptions.libpmda)
+	    pmNotifyErr(LOG_DEBUG, "Received PDU_ATTR\n");
+	if (__pmDecodeAttr(pb, &subtype, &buffer, &length) < 0) 
+	    break;
+	if (HAVE_V_SIX(dispatch->comm.pmda_interface)) {
+	    ctxnum = dispatch->version.six.ext->e_context;
+	    sts = dispatch->version.six.attribute(ctxnum, subtype, buffer, length, pmda);
+	} else {
+	    sts = PM_ERR_GENERIC;
+	}
+	break;
+
+    default: {
+	char	strbuf[20];
+	pmNotifyErr(LOG_ERR,
 		      "%s: Unrecognised pdu type: %s?\n",
 		      pmda->e_name, __pmPDUTypeStr_r(sts, strbuf, sizeof(strbuf)));
+	}
 	break;
     }
 
@@ -478,12 +428,10 @@ pmdaMain(pmdaInterface *dispatch)
 void
 pmdaSetResultCallBack(pmdaInterface *dispatch, pmdaResultCallBack callback)
 {
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	dispatch->version.four.ext->e_resultCallBack = callback;
-    else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	dispatch->version.two.ext->e_resultCallBack = callback;
+    if (HAVE_ANY(dispatch->comm.pmda_interface))
+	dispatch->version.any.ext->e_resultCallBack = callback;
     else {
-	__pmNotifyErr(LOG_CRIT, "Unable to set result callback for PMDA interface version %d.",
+	pmNotifyErr(LOG_CRIT, "Unable to set result callback for PMDA interface version %d.",
 		     dispatch->comm.pmda_interface);
 	dispatch->status = PM_ERR_GENERIC;
     }
@@ -495,7 +443,7 @@ pmdaSetEndContextCallBack(pmdaInterface *dispatch, pmdaEndContextCallBack callba
     if (HAVE_V_FIVE(dispatch->comm.pmda_interface) || callback == NULL)
 	dispatch->version.four.ext->e_endCallBack = callback;
     else {
-	__pmNotifyErr(LOG_CRIT, "Unable to set end context callback for PMDA interface version %d.",
+	pmNotifyErr(LOG_CRIT, "Unable to set end context callback for PMDA interface version %d.",
 		     dispatch->comm.pmda_interface);
 	dispatch->status = PM_ERR_GENERIC;
     }
@@ -504,12 +452,10 @@ pmdaSetEndContextCallBack(pmdaInterface *dispatch, pmdaEndContextCallBack callba
 void
 pmdaSetFetchCallBack(pmdaInterface *dispatch, pmdaFetchCallBack callback)
 {
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	dispatch->version.four.ext->e_fetchCallBack = callback;
-    else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	dispatch->version.two.ext->e_fetchCallBack = callback;
+    if (HAVE_ANY(dispatch->comm.pmda_interface))
+	dispatch->version.any.ext->e_fetchCallBack = callback;
     else {
-	__pmNotifyErr(LOG_CRIT, "Unable to set fetch callback for PMDA interface version %d.",
+	pmNotifyErr(LOG_CRIT, "Unable to set fetch callback for PMDA interface version %d.",
 		     dispatch->comm.pmda_interface);
 	dispatch->status = PM_ERR_GENERIC;
     }
@@ -518,12 +464,10 @@ pmdaSetFetchCallBack(pmdaInterface *dispatch, pmdaFetchCallBack callback)
 void
 pmdaSetCheckCallBack(pmdaInterface *dispatch, pmdaCheckCallBack callback)
 {
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	dispatch->version.four.ext->e_checkCallBack = callback;
-    else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	dispatch->version.two.ext->e_checkCallBack = callback;
+    if (HAVE_ANY(dispatch->comm.pmda_interface))
+	dispatch->version.any.ext->e_checkCallBack = callback;
     else {
-	__pmNotifyErr(LOG_CRIT, "Unable to set check callback for PMDA interface version %d.",
+	pmNotifyErr(LOG_CRIT, "Unable to set check callback for PMDA interface version %d.",
 		     dispatch->comm.pmda_interface);
 	dispatch->status = PM_ERR_GENERIC;
     }
@@ -532,13 +476,32 @@ pmdaSetCheckCallBack(pmdaInterface *dispatch, pmdaCheckCallBack callback)
 void
 pmdaSetDoneCallBack(pmdaInterface *dispatch, pmdaDoneCallBack callback)
 {
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	dispatch->version.four.ext->e_doneCallBack = callback;
-    else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	dispatch->version.two.ext->e_doneCallBack = callback;
+    if (HAVE_ANY(dispatch->comm.pmda_interface))
+	dispatch->version.any.ext->e_doneCallBack = callback;
     else {
-	__pmNotifyErr(LOG_CRIT, "Unable to set done callback for PMDA interface version %d.",
+	pmNotifyErr(LOG_CRIT, "Unable to set done callback for PMDA interface version %d.",
 		     dispatch->comm.pmda_interface);
 	dispatch->status = PM_ERR_GENERIC;
     }
 }
+
+void
+pmdaSetLabelCallBack(pmdaInterface *dispatch, pmdaLabelCallBack callback)
+{
+    if (HAVE_V_SEVEN(dispatch->comm.pmda_interface))
+	dispatch->version.any.ext->e_labelCallBack = callback;
+    else {
+	pmNotifyErr(LOG_CRIT, "Unable to set label callback for PMDA interface version %d.",
+		     dispatch->comm.pmda_interface);
+	dispatch->status = PM_ERR_GENERIC;
+    }
+}
+
+void
+pmdaSendError(pmdaInterface *dispatch, int err)
+{
+    pmdaExt     *pmda = dispatch->version.any.ext;
+
+    /* Usually err is PM_ERR_PMDAREADY or PM_ERR_PMDANOTREADY */
+    __pmSendError(pmda->e_outfd, FROM_ANON, err);
+}
