@@ -18,7 +18,6 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
 #include "trace.h"
 #include "trace_dev.h"
 
@@ -51,6 +50,25 @@ typedef struct {
      */
 } tracedata_t;
 
+#ifdef HAVE_NETWORK_BYTEORDER
+#define trace_htonll(a) do { } while (0) /* noop */
+#define trace_ntohll(a) do { } while (0) /* noop */
+#else
+static void
+trace_htonll(char *p)
+{   
+    char        c;
+    int         i;
+
+    for (i = 0; i < 4; i++) {
+        c = p[i];
+        p[i] = p[7-i];
+        p[7-i] = c; 
+    }
+}
+#define trace_ntohll(v) trace_htonll(v)
+#endif
+
 int
 __pmtracesenddata(int fd, char *tag, int taglen, int tagtype, double data)
 {
@@ -89,10 +107,9 @@ __pmtracesenddata(int fd, char *tag, int taglen, int tagtype, double data)
     cp = (char *)pp;
     cp += sizeof(tracedata_t);
     memcpy((void *)cp, (void *)&data, sizeof(double));
-    __htonll((char *)cp);	/* send in network byte order */
+    trace_htonll(cp);	/* send in network byte order */
     cp += sizeof(double);
     strcpy(cp, tag);
-#ifdef PCP_DEBUG
     if ((taglen % sizeof(__pmTracePDU)) != 0) {
 	/* for Purify */
 	int	pad;
@@ -100,7 +117,6 @@ __pmtracesenddata(int fd, char *tag, int taglen, int tagtype, double data)
 	for (pad = sizeof(__pmTracePDU) - 1; pad >= (taglen % sizeof(__pmTracePDU)); pad--)
 	    *padp++ = '~';	/* buffer end */
     }
-#endif
 
 #ifdef PMTRACE_DEBUG
     if (__pmstate & PMTRACE_STATE_PDU)
@@ -111,33 +127,46 @@ __pmtracesenddata(int fd, char *tag, int taglen, int tagtype, double data)
 }
 
 int
-__pmtracedecodedata(__pmTracePDU *pdubuf, char **tag, int *taglen,
+__pmtracedecodedata(__pmTracePDU *pdubuf, char **tag, int *taglenp,
 				int *tagtype, int *protocol, double *data)
 {
     tracedata_t	*pp;
     int		*ip;
     char	*cp;
+    char	*pduend;
+    int		taglen;
 
     if (pdubuf == NULL)
 	return PMTRACE_ERR_IPC;
 
     pp = (tracedata_t *)pdubuf;
+    pduend = (char *)pdubuf + pp->hdr.len;
+
+    if (pduend - (char*)pp < sizeof(tracedata_t) + sizeof(double))
+	return PMTRACE_ERR_IPC;
+
     ip = (int *)&pp->bits;
     *ip = ntohl(*ip);
-    if (pp->bits.taglen == 0)
-	return PMTRACE_ERR_IPC;
+    taglen = pp->bits.taglen;
     if (pp->bits.version != TRACE_PDU_VERSION)
 	return PMTRACE_ERR_VERSION;
-    *taglen = pp->bits.taglen;
+    if (taglen <= 0 || taglen >= CHAR_MAX - 1 || taglen > pp->hdr.len)
+	return PMTRACE_ERR_IPC;
+    if (pduend - (char *)pp < sizeof(tracedata_t) + sizeof(double) + taglen)
+	return PMTRACE_ERR_IPC;
+    *taglenp = taglen;
     *tagtype = pp->bits.tagtype;
     *protocol = pp->bits.protocol;
 
     cp = (char *)pp;
     cp += sizeof(tracedata_t);
     memcpy((void *)data, (void *)cp, sizeof(double));
-    __ntohll((char *)data);	/* receive in network byte order */
+    trace_ntohll((char *)data);	/* receive in network byte order */
     cp += sizeof(double);
-    *tag = cp;
+    if ((*tag = (char *)malloc(taglen+1)) == NULL)
+	return -oserror();
+    strncpy(*tag, cp, taglen);
+    (*tag)[taglen] = '\0';
 
 #ifdef PMTRACE_DEBUG
     if (__pmstate & PMTRACE_STATE_PDU)
