@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 1997-2000,2003 Silicon Graphics, Inc.  All Rights Reserved.
+ * Copyright (c) 2012 Red Hat.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
@@ -14,7 +15,7 @@
 
 #include <signal.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "trace.h"
 #include "trace_dev.h"
 
@@ -25,8 +26,6 @@ typedef struct {
 static more_ctl		*more;
 static int		maxfd = -1;
 
-extern int	__pmfd;
-
 static char *
 pdutypestr(int type)
 {
@@ -34,7 +33,7 @@ pdutypestr(int type)
     else if (type == TRACE_PDU_DATA) return "DATA";
     else {
 	static char     buf[20];
-	snprintf(buf, sizeof(buf), "TYPE-%d?", type);
+	pmsprintf(buf, sizeof(buf), "TYPE-%d?", type);
 	return buf;
     }
 }
@@ -119,7 +118,7 @@ pduread(int fd, char *buf, int len, int mode, int timeout)
      */
     int				status = 0;
     int				have = 0;
-    fd_set			onefd;
+    __pmFdSet			onefd;
     static int			done_default = 0;
     static struct timeval	def_wait = { 10, 0 };
 
@@ -136,12 +135,7 @@ pduread(int fd, char *buf, int len, int mode, int timeout)
 		    return status;
 		}
 		else {
-		    def_wait.tv_sec = (int)def_timeout;	/* truncate -> secs */
-		    if (def_timeout > (double)def_wait.tv_sec)
-			def_wait.tv_usec = (long)((def_timeout -
-					(double)def_wait.tv_sec) * 1000000);
-		    else
-			def_wait.tv_usec = 0;
+		    pmtimevalFromReal(def_timeout, &def_wait);
 		}
 	    }
 	    done_default = 1;
@@ -160,9 +154,9 @@ pduread(int fd, char *buf, int len, int mode, int timeout)
 	    }
 	    else
 		wait = def_wait;
-	    FD_ZERO(&onefd);
-	    FD_SET(fd, &onefd);
-	    status = select(fd+1, &onefd, NULL, NULL, &wait);
+	    __pmFD_ZERO(&onefd);
+	    __pmFD_SET(fd, &onefd);
+	    status = __pmSelectRead(fd+1, &onefd, &wait);
 	    if (status == 0)
 		return PMTRACE_ERR_TIMEOUT;
 	    else if (status < 0) {
@@ -170,7 +164,7 @@ pduread(int fd, char *buf, int len, int mode, int timeout)
 		return status;
 	    }
 	}
-	status = (int)read(fd, buf, len);
+	status = (int)__pmRead(fd, buf, len);
 	if (status <= 0) {	/* EOF or error */
 	    setoserror(neterror());
 	    return status;
@@ -211,9 +205,6 @@ __pmtracexmitPDU(int fd, __pmTracePDU *pdubuf)
 	signal(SIGPIPE, user_onpipe);
 #endif
 
-    if (__pmfd == -1)
-	return PMTRACE_ERR_IPC;
-
     php->from = (__int32_t)getpid();
 #ifdef PMTRACE_DEBUG
     if (__pmstate & PMTRACE_STATE_PDU) {
@@ -241,7 +232,7 @@ __pmtracexmitPDU(int fd, __pmTracePDU *pdubuf)
     php->len = htonl(php->len);
     php->from = htonl(php->from);
     php->type = htonl(php->type);
-    n = (int)write(fd, pdubuf, len);
+    n = (int)__pmWrite(fd, pdubuf, len);
     php->len = ntohl(php->len);
     php->from = ntohl(php->from);
     php->type = ntohl(php->type);
@@ -263,9 +254,6 @@ __pmtracegetPDU(int fd, int timeout, __pmTracePDU **result)
     __pmTracePDU	*pdubuf_prev;
     __pmTracePDUHdr	*php;
 
-    if (__pmfd == -1)
-	return PMTRACE_ERR_IPC;
-
     /*
      *	This stuff is a little tricky.  What we try to do is read()
      *	an amount of data equal to the largest PDU we have (or are
@@ -411,7 +399,7 @@ __pmtracegetPDU(int fd, int timeout, __pmTracePDU **result)
 	    *p++ = '~';	/* buffer end */
 
 	fprintf(stderr, "[%" FMT_PID "]__pmtracegetPDU: %s fd=%d len=%d from=%d",
-		getpid(), pdutypestr(php->type), fd, php->len, php->from);
+		(pid_t)getpid(), pdutypestr(php->type), fd, php->len, php->from);
 
 	for (j = 0; j < jend; j++) {
 	    if ((j % 8) == 0)
