@@ -1,6 +1,7 @@
 /*
  * trace.c - client-side interface for trace PMDA
  *
+ * Copyright (c) 2014 Red Hat.
  * Copyright (c) 1997-2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -16,7 +17,7 @@
 
 #include <inttypes.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 
 #if defined(HAVE_PTHREAD_H)
 #include <pthread.h>
@@ -26,9 +27,9 @@
 #error !bozo!
 #endif
 
+#include "hash.h"
 #include "trace.h"
 #include "trace_dev.h"
-#include "trace_hash.h"
 
 static int	_pmtimedout = 1;
 static time_t	_pmttimeout = 0;
@@ -43,13 +44,6 @@ static __uint64_t _pmtraceid(void);
 
 int	__pmstate = PMTRACE_STATE_NONE;
 
-
-double
-__pmtracetvsub(const struct timeval *a, const struct timeval *b)
-{
-    return (double)(a->tv_sec - b->tv_sec + (double)(a->tv_usec - b->tv_usec)/1000000.0);
-}
-
 /* transaction data unit */
 typedef struct {
     __uint64_t		id;
@@ -88,7 +82,7 @@ _pmlibdel(void *entry)
 	free(data);
 }
 
-int			__pmfd = 0;
+static int		__pmfd;
 static __pmHashTable	_pmtable;
 
 #if defined(HAVE_PTHREAD_MUTEX_T)
@@ -168,7 +162,7 @@ pmtracebegin(const char *tag)
 	hash.pad = 0;
 	if ((hash.tag = strdup(tag)) == NULL)
 	    b_sts = -oserror();
-	__pmtimevalNow(&hash.start);
+	pmtimevalNow(&hash.start);
 	if (b_sts >= 0) {
 	    hash.inprogress = 1;
 	    b_sts = __pmhashinsert(&_pmtable, tag, &hash);
@@ -186,7 +180,7 @@ pmtracebegin(const char *tag)
 	fprintf(stderr, "pmtracebegin: updating transaction '%s' "
 			"(id=0x%" PRIx64 ")\n", tag, hash.id);
 #endif
-	__pmtimevalNow(&hptr->start);
+	pmtimevalNow(&hptr->start);
 	hptr->inprogress = 1;
     }
 
@@ -212,7 +206,7 @@ pmtraceend(const char *tag)
     if ((len = strlen(tag)+1) >= MAXTAGNAMELEN)
 	return PMTRACE_ERR_TAGLENGTH;
 
-    __pmtimevalNow(&now);
+    pmtimevalNow(&now);
 
     /* give just enough info for comparison routine */
     hash.tag = (char *)tag;
@@ -236,7 +230,7 @@ pmtraceend(const char *tag)
 			"(id=0x%" PRIx64 ")\n", tag, hash.id);
 #endif
 	hptr->inprogress = 0;
-	hptr->data = __pmtracetvsub(&now, &hptr->start);
+	hptr->data = pmtimevalSub(&now, &hptr->start);
 
 	if (sts >= 0 && _pmtimedout) {
 	    sts = _pmtracereconnect();
@@ -445,15 +439,27 @@ pmtraceerrstr(int code)
 static int
 _pmtraceremaperr(int sts)
 {
+    int save_oserror;
+    int socket_closed;
+
+    /*
+     * sts is negative.
+     * Use __pmSocketClosed() to decode it, since it may have come from
+     * __pmSecureSocketsError(). __pmSocketClosed uses oserror() and expects it to
+     * be non-negative.
+     */
+    save_oserror = oserror();
+    setoserror(-sts);
+    socket_closed = __pmSocketClosed();
+    setoserror(save_oserror);
+
 #ifdef PMTRACE_DEBUG
     if (__pmstate & PMTRACE_STATE_COMMS)
 	fprintf(stderr, "_pmtraceremaperr: status %d remapped to %d\n", sts,
-		(sts == -EBADF || sts == -EPIPE || sts == -ETIMEDOUT ||
-		 sts == -ECONNRESET || sts == -ECONNREFUSED)?
-		PMTRACE_ERR_IPC : sts);
+		socket_closed ? PMTRACE_ERR_IPC : sts);
 #endif
-    if (sts == -EBADF || sts == -EPIPE || sts == -ETIMEDOUT ||
-		sts == -ECONNRESET || sts == -ECONNREFUSED) {
+
+    if (socket_closed) {
 	_pmtimedout = 1;
 	return PMTRACE_ERR_IPC;
     }
@@ -603,7 +609,7 @@ _pmtracereconnect(void)
     }
     if (__pmfd >= 0) {
 	__pmtracenomoreinput(__pmfd);
-	close(__pmfd);
+	__pmCloseSocket(__pmfd);
 	__pmfd = -1;
     }
     if (_pmtraceconnect(1) < 0) {
@@ -674,17 +680,20 @@ _pmauxtraceconnect(void)
     int			port = TRACE_PORT;
     char		hostname[MAXHOSTNAMELEN];
     struct timeval	timeout = { 3, 0 };     /* default 3 secs */
-    struct sockaddr_in	myaddr;
-    struct hostent	*servinfo;
-    struct linger	nolinger = {1, 0};
+    __pmSockAddr	*myaddr;
+    __pmHostEnt		*servinfo;
+    void		*enumIx;
 #ifndef IS_MINGW
     struct itimerval	_pmolditimer;
     void		(*old_handler)(int foo);
 #endif
-    int			rc, sts = 0;
-    int			flags;
-    int			nodelay=1;
+    int			rc, sts;
+    int			flags = 0;
     char		*sptr, *endptr, *endnum;
+    struct timeval	canwait = { 5, 000000 };
+    struct timeval	stv;
+    struct timeval	*pstv;
+    __pmFdSet		wfds;
 
 #ifdef PMTRACE_DEBUG
     if (__pmstate & PMTRACE_STATE_NOAGENT) {
@@ -703,8 +712,8 @@ _pmauxtraceconnect(void)
     if ((sptr = getenv(TRACE_ENV_HOST)) != NULL)
 	strcpy(hostname, sptr);
     else {
-	(void)gethostname(hostname, MAXHOSTNAMELEN);
-	hostname[MAXHOSTNAMELEN-1] = '\0';
+       (void)gethostname(hostname, MAXHOSTNAMELEN);
+       hostname[MAXHOSTNAMELEN-1] = '\0';
     }
     if ((sptr = getenv(TRACE_ENV_PORT)) != NULL) {
 	port = (int)strtol(sptr, &endnum, 0);
@@ -718,117 +727,155 @@ _pmauxtraceconnect(void)
 	if (*endptr != '\0' || timesec < 0.0)
 	    fprintf(stderr, "trace warning: bogus PCP_TRACE_TIMEOUT.");
 	else {
-	    timeout.tv_sec = (time_t)timesec;
-	    timeout.tv_usec = (int)((timesec - (double)timeout.tv_sec)*1000000);
+	    pmtimevalFromReal(timesec, &timeout);
 	}
     }
     if (getenv(TRACE_ENV_NOAGENT) != NULL)
 	__pmstate |= PMTRACE_STATE_NOAGENT;
 
-    if ((servinfo = gethostbyname(hostname)) == NULL) {
+    if ((servinfo = __pmGetAddrInfo(hostname)) == NULL) {
 #ifdef PMTRACE_DEBUG
 	if (__pmstate & PMTRACE_STATE_COMMS)
-	    fprintf(stderr, "_pmtraceconnect(gethostbyname(hostname=%s): "
+	    fprintf(stderr, "_pmtraceconnect(__pmGetAddrInfo(hostname=%s): "
 		    "hosterror=%d, ``%s''\n", hostname, hosterror(),
 		    hoststrerror());
 #endif
 	return -EHOSTUNREACH;
     }
 
-    /* Create socket and attempt to connect to the local PMDA */
-    if ((__pmfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
+    /* Try each address in turn until one connects. */
+    sts = EHOSTUNREACH;
+    __pmfd = -1;
+    enumIx = NULL;
+    for (myaddr = __pmHostEntGetSockAddr(servinfo, &enumIx);
+	 myaddr != NULL;
+	 myaddr = __pmHostEntGetSockAddr(servinfo, &enumIx)) {
+	/* Create a socket */
+	if (__pmSockAddrIsInet(myaddr))
+	    __pmfd = __pmCreateSocket();
+	else if (__pmSockAddrIsIPv6(myaddr))
+	    __pmfd = __pmCreateIPv6Socket();
+	else {
+	    fprintf(stderr, "_pmtraceconnect(invalid address family): %d\n",
+		    __pmSockAddrGetFamily(myaddr));
+	}
+	if (__pmfd < 0) {
+	    sts = neterror();
+	    __pmSockAddrFree(myaddr);
+	    continue; /* Try the next address */
+	}
+
+	/* Set the port. */
+	__pmSockAddrSetPort(myaddr, port);
+
+#ifndef IS_MINGW
+	/* arm interval timer */
+	_pmmyitimer.it_value.tv_sec = timeout.tv_sec;
+	_pmmyitimer.it_value.tv_usec = timeout.tv_usec;
+	_pmmyitimer.it_interval.tv_sec = 0;
+	_pmmyitimer.it_interval.tv_usec = 0;
+	old_handler = signal(SIGALRM, _pmtracealarm);
+	setitimer(ITIMER_REAL, &_pmmyitimer, &_pmolditimer);
+#endif
+
+#ifdef PMTRACE_DEBUG
+	if (__pmstate & PMTRACE_STATE_COMMS) {
+	    char *name = __pmHostEntGetName (servinfo);
+	    fprintf(stderr, "_pmtraceconnect: PMDA host=%s port=%d timeout=%d"
+		    "secs\n", name == NULL ? "unknown" : name, port, (int)timeout.tv_sec);
+	    if (name != NULL)
+		free(name);
+	}
+#endif
+
+	/* Attempt to connect */
+	flags = __pmConnectTo(__pmfd, myaddr, port);
+	__pmSockAddrFree(myaddr);
+
+	if (flags < 0) {
+	    /*
+	     * Mark failure in case we fall out the end of the loop
+	     * and try next address. __pmfd has been closed in __pmConnectTo().
+	     */
+	    sts = -flags;
+	    __pmfd = -1;
+	    continue;
+	}
+
+	/* FNDELAY and we're in progress - wait on select */
+	stv = canwait;
+	pstv = (stv.tv_sec || stv.tv_usec) ? &stv : NULL;
+	__pmFD_ZERO(&wfds);
+	__pmFD_SET(__pmfd, &wfds);
+	if ((rc = __pmSelectWrite(__pmfd+1, &wfds, pstv)) == 1) {
+	    sts = __pmConnectCheckError(__pmfd);
+	}
+	else if (rc == 0) {
+	    sts = ETIMEDOUT;
+	}
+	else {
+	    sts = (rc < 0) ? neterror() : EINVAL;
+	}
+ 
+#ifndef IS_MINGW
+	/* re-arm interval timer */
+	setitimer(ITIMER_REAL, &off_itimer, &_pmmyitimer);
+	signal(SIGALRM, old_handler);
+	if (_pmolditimer.it_value.tv_sec != 0 && _pmolditimer.it_value.tv_usec != 0) {
+	    _pmolditimer.it_value.tv_usec -= timeout.tv_usec - _pmmyitimer.it_value.tv_usec;
+	    while (_pmolditimer.it_value.tv_usec < 0) {
+		_pmolditimer.it_value.tv_usec += 1000000;
+		_pmolditimer.it_value.tv_sec--;
+	    }
+	    while (_pmolditimer.it_value.tv_usec > 1000000) {
+		_pmolditimer.it_value.tv_usec -= 1000000;
+		_pmolditimer.it_value.tv_sec++;
+	    }
+	    _pmolditimer.it_value.tv_sec -= timeout.tv_sec - _pmmyitimer.it_value.tv_sec;
+	    if (_pmolditimer.it_value.tv_sec < 0) {
+		/* missed the user's itimer, pretend there is 1 msec to go! */
+		_pmolditimer.it_value.tv_sec = 0;
+		_pmolditimer.it_value.tv_usec = 1000;
+	    }
+	    setitimer(ITIMER_REAL, &_pmolditimer, &_pmmyitimer);
+	}
+#endif
+
+	/* Was the connection successful? */
+	if (sts == 0)
+	    break;
+
+	/* Unsuccessful connection. */
+	__pmCloseSocket(__pmfd);
+	__pmfd = -1;
+    } /* loop over addresses */
+
+    __pmHostEntFree(servinfo);
+
+    /* Was the connection successful? */
+    if (__pmfd < 0) {
 #ifdef PMTRACE_DEBUG
 	if (__pmstate & PMTRACE_STATE_COMMS)
 	    fprintf(stderr, "_pmtraceconnect(socket failed): %s\n",
 		    netstrerror());
 #endif
-	return -neterror();
-    }
-
-    /* avoid 200 ms delay */
-    if (setsockopt(__pmfd, IPPROTO_TCP, TCP_NODELAY, (char *)&nodelay,
-				    (__pmSockLen)sizeof(nodelay)) < 0) {
-#ifdef PMTRACE_DEBUG
-	if (__pmstate & PMTRACE_STATE_COMMS)
-	    fprintf(stderr, "_pmtraceconnect(setsockopt1 failed): %s\n",
-		    netstrerror());
-#endif
-	return -neterror();
-    }
-
-    /* don't linger on close */
-    if (setsockopt(__pmfd, SOL_SOCKET, SO_LINGER, (char *)&nolinger,
-				    (__pmSockLen)sizeof(nolinger)) < 0) {
-#ifdef PMTRACE_DEBUG
-	if (__pmstate & PMTRACE_STATE_COMMS)
-	    fprintf(stderr, "_pmtraceconnect(setsockopt2 failed): %s\n",
-		    netstrerror());
-#endif
-	return -neterror();
-    }
-
-    memset(&myaddr, 0, sizeof(myaddr));
-    myaddr.sin_family = AF_INET;
-    memcpy(&myaddr.sin_addr, servinfo->h_addr, servinfo->h_length);
-    myaddr.sin_port = htons(port);
-
-#ifndef IS_MINGW
-    /* arm interval timer */
-    _pmmyitimer.it_value.tv_sec = timeout.tv_sec;
-    _pmmyitimer.it_value.tv_usec = timeout.tv_usec;
-    _pmmyitimer.it_interval.tv_sec = 0;
-    _pmmyitimer.it_interval.tv_usec = 0;
-    old_handler = signal(SIGALRM, _pmtracealarm);
-    setitimer(ITIMER_REAL, &_pmmyitimer, &_pmolditimer);
-#endif
-
-#ifdef PMTRACE_DEBUG
-    if (__pmstate & PMTRACE_STATE_COMMS) {
-	fprintf(stderr, "_pmtraceconnect: PMDA host=%s port=%d timeout=%d"
-		"secs\n", servinfo->h_name, port, (int)timeout.tv_sec);
-    }
-#endif
-
-    if ((rc = connect(__pmfd, (struct sockaddr*) &myaddr, sizeof(myaddr))) < 0)
-	return -neterror();
-
-#ifndef IS_MINGW
-    /* re-arm interval timer */
-    setitimer(ITIMER_REAL, &off_itimer, &_pmmyitimer);
-    signal(SIGALRM, old_handler);
-    if (_pmolditimer.it_value.tv_sec != 0 && _pmolditimer.it_value.tv_usec != 0) {
-	_pmolditimer.it_value.tv_usec -= timeout.tv_usec - _pmmyitimer.it_value.tv_usec;
-	while (_pmolditimer.it_value.tv_usec < 0) {
-	    _pmolditimer.it_value.tv_usec += 1000000;
-	    _pmolditimer.it_value.tv_sec--;
-	}
-	while (_pmolditimer.it_value.tv_usec > 1000000) {
-	    _pmolditimer.it_value.tv_usec -= 1000000;
-	    _pmolditimer.it_value.tv_sec++;
-	}
-	_pmolditimer.it_value.tv_sec -= timeout.tv_sec - _pmmyitimer.it_value.tv_sec;
-	if (_pmolditimer.it_value.tv_sec < 0) {
-	    /* missed the user's itimer, pretend there is 1 msec to go! */
-	    _pmolditimer.it_value.tv_sec = 0;
-	    _pmolditimer.it_value.tv_usec = 1000;
-	}
-	setitimer(ITIMER_REAL, &_pmolditimer, &_pmmyitimer);
-    }
-#endif
-
-    if (rc < 0) {
-	if (sts == EINTR)
-	    sts = -ETIMEDOUT;
-	close(__pmfd);	/* safe for tracemoreinput(), as no PDUs yet */
-	__pmfd = -1;
-	return sts;
+	return -sts;
     }
 
     _pmtimedout = 0;
 
+    /* Restore the original file status flags. */
+    if (__pmSetFileStatusFlags(__pmfd, flags) < 0) {
+#ifdef PMTRACE_DEBUG
+	if (__pmstate & PMTRACE_STATE_COMMS)
+	    fprintf(stderr, ":_pmtraceconnect: cannot restore file status flags\n");
+#endif
+	return -oserror();
+    }
+
     /* make sure this file descriptor is closed if exec() is called */
-    if ((flags = fcntl(__pmfd, F_GETFD)) != -1)
-	sts = fcntl(__pmfd, F_SETFD, flags | FD_CLOEXEC);
+    if ((flags = __pmGetFileDescriptorFlags(__pmfd)) != -1)
+	sts = __pmSetFileDescriptorFlags(__pmfd, flags | FD_CLOEXEC);
     else
 	sts = -1;
     if (sts == -1)
@@ -836,8 +883,8 @@ _pmauxtraceconnect(void)
 
     if (__pmtraceprotocol(TRACE_PROTOCOL_QUERY) == TRACE_PROTOCOL_ASYNC) {
 	/* in the asynchronoous protocol - ensure no delay after close */
-	if ((flags = fcntl(__pmfd, F_GETFL)) != -1)
-	    sts = fcntl(__pmfd, F_SETFL, flags | FNDELAY);
+	if ((flags = __pmGetFileStatusFlags(__pmfd)) != -1)
+	    sts = __pmSetFileStatusFlags(__pmfd, flags | FNDELAY);
 	else
 	    sts = -1;
 	if (sts == -1)
