@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2013 Red Hat.
+ * Copyright (c) 2012-2016 Red Hat.
  * Copyright (c) 1995-2005 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -17,7 +17,7 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "pmcd.h"
 #include <ctype.h>
 #include <sys/stat.h>
@@ -52,6 +52,7 @@ static time_t   	configFileTime;
 int		szAgents;		/* Number currently allocated */
 int		mapdom[MAXDOMID+2];	/* The DomainId-to-AgentIndex map */
 					/* Don't use it during parsing! */
+int		pmdarootfd;		/* fd for pmdaroot(1) services */
 
 static FILE	*inputStream;		/* Input stream for scanner */
 static int	scanInit;
@@ -117,16 +118,16 @@ GetNextLine(void)
     int		still_to_read;
     int		atEOF = 0;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL2)
+    if (pmDebugOptions.appl2)
 	fprintf(stderr, "%d: GetNextLine()\n", nLines);
-#endif
 
     if (szLineBuf == 0) {
 	szLineBuf = LINEBUF_SIZE;
 	linebuf = (char *)malloc(szLineBuf);
-	if (linebuf == NULL)
-	    __pmNoMem("pmcd config: GetNextLine init", szLineBuf, PM_FATAL_ERR);
+	if (linebuf == NULL) {
+	    pmNoMem("pmcd config: GetNextLine init", szLineBuf, PM_FATAL_ERR);
+	    /*NOTREACHED*/
+	}
     }
 
     linebuf[0] = '\0';
@@ -192,7 +193,7 @@ GetNextLine(void)
 	    if ((linebuf = realloc(linebuf, szLineBuf)) == NULL) {
 		static char	fallback[2];
 
-		__pmNoMem("pmcd config: GetNextLine", szLineBuf, PM_RECOV_ERR);
+		pmNoMem("pmcd config: GetNextLine", szLineBuf, PM_RECOV_ERR);
 		linebuf = fallback;
 		linebuf[0] = '\0';
 		scanError = 1;
@@ -215,12 +216,10 @@ GetNextLine(void)
     } while (more);
     nLines++;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL2) {
+    if (pmDebugOptions.appl2) {
 	fprintf(stderr, "\n===================NEWLINE=================\n\n");
 	fprintf(stderr, "len = %d\nline = \"%s\"\n", (int)strlen(linebuf), linebuf);
     }
-#endif
 }
 
 /* Advance through the input stream until either a non-whitespace character, a
@@ -281,10 +280,10 @@ FindNextToken(void)
     static char	*rawToken;		/* Used in pathological quoting case */
     char	ch;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL2)
-	fprintf(stderr, "FindNextToken()\n");
-#endif
+    if (pmDebugOptions.appl2) {
+	fprintf(stderr, "FindNextToken() ");
+	fprintf(stderr, "scanInit=%d scanError=%d scanReadOnly=%d doingAccess=%d tokenQuoted=%d token=%p tokenend=%p\n", scanInit, scanError, scanReadOnly, doingAccess, tokenQuoted, token, tokenend);
+    }
 
     do {
 	if (scanInit) {
@@ -296,8 +295,10 @@ FindNextToken(void)
 	    if (*token == '\n')		/* If at end of line, get next line */
 		GetNextLine();
 	    else {			/* Otherwise move past "old" token */
-		if (tokenQuoted)	/* Move past last quote */
+		if (tokenQuoted) {	/* Move past last quote */ 
 		    tokenend++;
+		    tokenQuoted = 0;
+		}
 		token = tokenend;
 	    }
 	    SkipWhitespace();		/* Find null, newline or non-space  */
@@ -356,22 +357,10 @@ FindNextToken(void)
 			if (*(tokenend - 1) == '\\')
 			    fixToken = 1;
 			else {
-			    int	err = 1;
-
 			    if (inQuotes) {
 				inQuotes = 0;
-				err = tokenend[1] && !isspace((int)tokenend[1]);
 				gotEnd = 1;
 			    }
-			    if (err) {
-				scanError = 1;
-				*token = '\0';
-				tokenend = token;
-				fprintf(stderr,
-					     "pmcd config[line %d]: Error: quoted string must have whitespace either side\n",
-					     nLines);
-				return;
-			    }
 			}
 			break;
 
@@ -422,13 +411,12 @@ FindNextToken(void)
     else
 	tokenQuoted = 0;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL2) {
+    if (pmDebugOptions.appl2) {
 	fputs("TOKEN = '", stderr);
 	PrintToken(stderr);
-	fputs("'\n", stderr);
+	fputs("' ", stderr);
+	fprintf(stderr, "scanInit=%d scanError=%d scanReadOnly=%d doingAccess=%d tokenQuoted=%d token=%p tokenend=%p\n", scanInit, scanError, scanReadOnly, doingAccess, tokenQuoted, token, tokenend);
     }
-#endif
 }
 
 /* Move to the next line of the input stream. */
@@ -466,7 +454,7 @@ BuildCmdLine(char **argv)
     if ((cmdLine = (char *)malloc(cmdLen)) == NULL) {
 	fprintf(stderr, "pmcd config[line %d]: Error: failed to build command line\n",
 		nLines);
-	__pmNoMem("pmcd config: BuildCmdLine", cmdLen, PM_RECOV_ERR);
+	pmNoMem("pmcd config: BuildCmdLine", cmdLen, PM_RECOV_ERR);
 	return NULL;
     }
     for (i = 0, p = cmdLine; argv[i] != NULL; i++) {
@@ -498,22 +486,24 @@ BuildArgv(void)
 {
     int		nArgs;
     char	**result;
+    char	**result_new;
 
     nArgs = 0;
     result = NULL;
     do {
 	/* Make result big enough for new arg and terminating NULL pointer */
-	result = (char **)realloc(result, (nArgs + 2) * sizeof(char *));
-	if (result != NULL) {
+	result_new = (char **)realloc(result, (nArgs + 2) * sizeof(char *));
+	if (result_new != NULL) {
+	    result = result_new;
 	    if (*token != '/')
 		result[nArgs] = CopyToken();
 	    else if ((result[nArgs] = CopyToken()) != NULL)
 		__pmNativePath(result[nArgs]);
 	}
-	if (result == NULL || result[nArgs] == NULL) {
+	if (result_new == NULL || result[nArgs] == NULL) {
 	    fprintf(stderr, "pmcd config[line %d]: Error: failed to build argument list\n",
 		    nLines);
-	    __pmNoMem("pmcd config: build argv", nArgs * sizeof(char *),
+	    pmNoMem("pmcd config: build argv", nArgs * sizeof(char *),
 		     PM_RECOV_ERR);
 	    if (result != NULL) {
 		while (nArgs >= 0) {
@@ -525,10 +515,8 @@ BuildArgv(void)
 	    }
 	    return NULL;
 	}
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL2)
+	if (pmDebugOptions.appl2)
 	    fprintf(stderr, "argv[%d] = '%s'\n", nArgs, result[nArgs]);
-#endif
 
 	nArgs++;
 	FindNextToken();
@@ -548,18 +536,21 @@ GetNewAgent(void)
     if (agent == NULL) {
 	agent = (AgentInfo*)malloc(sizeof(AgentInfo) * MIN_AGENTS_ALLOC);
 	if (agent == NULL) {
-	    perror("GetNewAgentIndex: malloc");
-	    exit(1);
+	    pmNoMem("GetNewAgentIndex: malloc", sizeof(AgentInfo) * MIN_AGENTS_ALLOC, PM_FATAL_ERR);
+	    /*NOTREACHED*/
 	}
 	szAgents = MIN_AGENTS_ALLOC;
     }
     else if (nAgents >= szAgents) {
+	int	i;
 	agent = (AgentInfo*)
 	    realloc(agent, sizeof(AgentInfo) * 2 * szAgents);
 	if (agent == NULL) {
-	    perror("GetNewAgentIndex: realloc");
-	    exit(1);
+	    pmNoMem("GetNewAgentIndex: realloc", sizeof(AgentInfo) * 2 * szAgents, PM_FATAL_ERR);
+	    /*NOTREACHED*/
 	}
+	for (i = 0; i < nAgents; i++)
+	    pmdaInterfaceMoved(&agent[i].ipc.dso.dispatch);
 	szAgents *= 2;
     }
 
@@ -620,7 +611,8 @@ ParseDso(char *pmDomainLabel, int pmDomainId)
     if ((entryPoint = CopyToken()) == NULL) {
 	fprintf(stderr, "pmcd config[line %d]: Error: couldn't copy DSO entry point\n",
 			 nLines);
-	__pmNoMem("pmcd config", tokenend - token + 1, PM_FATAL_ERR);
+	pmNoMem("pmcd config", tokenend - token + 1, PM_FATAL_ERR);
+	/*NOTREACHED*/
     }
 
     FindNextToken();
@@ -640,7 +632,8 @@ ParseDso(char *pmDomainLabel, int pmDomainId)
     if ((pathName = CopyToken()) == NULL) {
 	fprintf(stderr, "pmcd config[line %d]: Error: couldn't copy DSO pathname\n",
 			nLines);
-	__pmNoMem("pmcd config", tokenend - token + 1, PM_FATAL_ERR);
+	pmNoMem("pmcd config", tokenend - token + 1, PM_FATAL_ERR);
+	/*NOTREACHED*/
     }
     __pmNativePath(pathName);
 
@@ -682,11 +675,13 @@ ParseSocket(char *pmDomainLabel, int pmDomainId)
     FindNextToken();
     if (TokenIs("inet"))
 	addrDomain = AF_INET;
+    else if (TokenIs("ipv6"))
+	addrDomain = AF_INET6;
     else if (TokenIs("unix"))
 	addrDomain = AF_UNIX;
     else {
 	fprintf(stderr,
-		     "pmcd config[line %d]: Error: expected socket address domain (`inet' or `unix')\n",
+		     "pmcd config[line %d]: Error: expected socket address domain (`inet', `ipv6', or `unix')\n",
 		     nLines);
 	return -1;
     }
@@ -703,14 +698,15 @@ ParseSocket(char *pmDomainLabel, int pmDomainId)
 	if ((socketName = CopyToken()) == NULL) {
 	    fprintf(stderr, "pmcd config[line %d]: Error: couldn't copy port name\n",
 			 nLines);
-	    __pmNoMem("pmcd config", tokenend - token + 1, PM_FATAL_ERR);
+	    pmNoMem("pmcd config", tokenend - token + 1, PM_FATAL_ERR);
+	    /*NOTREACHED*/
 	}
     FindNextToken();
 
     /* If an internet domain port name was specified, find the corresponding
      port number. */
 
-    if (addrDomain == AF_INET && socketName) {
+    if ((addrDomain == AF_INET || addrDomain == AF_INET6) && socketName) {
 	struct servent *service;
 
 	service = getservbyname(socketName, NULL);
@@ -815,8 +811,7 @@ ParseAccessSpec(int allow, int *specOps, int *denyOps, int *maxCons, int recursi
     int		haveComma = 0;
 
     if (*token == ';') {
-	fprintf(stderr,
-		     "pmcd config[line %d]: Error: empty or incomplete permissions list\n",
+	fprintf(stderr, "pmcd config[line %d]: Error: empty or incomplete permissions list\n",
 		     nLines);
 	return -1;
     }
@@ -861,7 +856,7 @@ ParseAccessSpec(int allow, int *specOps, int *denyOps, int *maxCons, int recursi
 	    *specOps = PMCD_OP_ALL;		/* Do this AFTER any recursive call */
 	}
 	else if (TokenIs("maximum") || TokenIs("unlimited")) {
-	    int	unlimited = *token == 'u' || *token == 'U';
+	    int	unlimited = (*token == 'u' || *token == 'U');
 
 	    if (*maxCons) {
 		fprintf(stderr,
@@ -957,36 +952,36 @@ ParseAccessSpec(int allow, int *specOps, int *denyOps, int *maxCons, int recursi
 }
 
 static int
-ParseHosts(int allow)
+ParseNames(char ***namesp, const char *nametype)
 {
-    int		sts = 0;
-    int		i = 0;
-    int		j;
-    int		ok = 0;
+    static char	**names;
+    static int	szNames;
+    int		nnames = 0;
     int		another = 1;
-    static char	**name = NULL;
-    static int	szNames = 0;
-    int		specOps = 0;
-    int		denyOps = 0;
-    int		maxCons = 0;		/* Zero=>unspecified, -1=>unlimited */
 
-    while (*token && another && *token != ':' && *token != ';') {
-	if (i == szNames) {
+    /* Beware of quoted tokens of length longer than 1. e.g. ":*" */
+    while (*token && another &&
+	   ((tokenend - token > 1) || (*token != ':' && *token != ';'))) {
+	if (nnames == szNames) {
 	    int		need;
 
 	    szNames += 8;
 	    need = szNames * (int)sizeof(char**);
-	    if ((name = (char **)realloc(name, need)) == NULL)
-		__pmNoMem("pmcd ParseHosts name list", need, PM_FATAL_ERR);
+	    if ((names = (char **)realloc(names, need)) == NULL) {
+		pmNoMem("pmcd ParseNames name list", need, PM_FATAL_ERR);
+		/*NOTREACHED*/
+	    }
+	}
+	if ((names[nnames++] = CopyToken()) == NULL) {
+	    pmNoMem("pmcd ParseNames name", tokenend - token, PM_FATAL_ERR);
+	    /*NOTREACHED*/
 	}
-	if ((name[i] = CopyToken()) == NULL)
-	    __pmNoMem("pmcd ParseHosts name", tokenend - token, PM_FATAL_ERR);
 	FindNextToken();
 	if (*token != ',' && *token != ':') {
 	    fprintf(stderr,
 			 "pmcd config[line %d]: Error: ',' or ':' expected after \"%s\"\n",
-			 nLines, name[i]);
-	    goto error;
+			 nLines, names[nnames-1]);
+	    return -1;
 	}
 	if (*token == ',') {
 	    FindNextToken();
@@ -994,46 +989,61 @@ ParseHosts(int allow)
 	}
 	else
 	    another = 0;
-	i++;
     }
-    if (i == 0) {
+    if (nnames == 0) {
 	fprintf(stderr,
-		     "pmcd config[line %d]: Error: no hosts in allow/disallow statement\n",
-		     nLines);
-	goto error;
+		     "pmcd config[line %d]: Error: no %ss in allow/disallow statement\n",
+		     nLines, nametype);
+	return -1;
     }
     if (another) {
-	fprintf(stderr, "pmcd config[line %d]: Error: host expected after ','\n",
-		     nLines);
-	goto error;
+	fprintf(stderr, "pmcd config[line %d]: Error: %s expected after ','\n",
+		     nLines, nametype);
+	return -1;
     }
     if (*token != ':') {
 	fprintf(stderr, "pmcd config[line %d]: Error: ':' expected after \"%s\"\n",
-		nLines, name[i]);
-	goto error;
+		nLines, names[nnames-1]);
+	return -1;
     }
+    *namesp = names;
+    return nnames;
+}
+
+static int
+ParseHosts(int allow)
+{
+    int		sts;
+    int		nhosts;
+    int		i;
+    int		ok = 0;
+    int		specOps = 0;
+    int		denyOps = 0;
+    int		maxCons = 0;		/* Zero=>unspecified, -1=>unlimited */
+    char	**hostnames;
+
+    if ((nhosts = ParseNames(&hostnames, "host")) < 0)
+	goto error;
+
     FindNextToken();
     if (ParseAccessSpec(allow, &specOps, &denyOps, &maxCons, 0) < 0)
 	goto error;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL1) {
-	for (j = 0; j < i; j++)
-	    fprintf(stderr, "ACCESS: %s specOps=%02x denyOps=%02x maxCons=%d\n",
-		    name[j], specOps, denyOps, maxCons);
+    if (pmDebugOptions.appl1) {
+	for (i = 0; i < nhosts; i++)
+	    fprintf(stderr, "HOST ACCESS: %s specOps=%02x denyOps=%02x maxCons=%d\n",
+		    hostnames[i], specOps, denyOps, maxCons);
     }
-#endif
-    
-    /* Make new entries for hosts in host access list */
 
-    for (j = 0; j < i; j++) {
-	if ((sts = __pmAccAddHost(name[j], specOps, denyOps, maxCons)) < 0) {
-	    if (sts == -EHOSTUNREACH || -EHOSTDOWN)
+    /* Make new entries for hosts in host access list */
+    for (i = 0; i < nhosts; i++) {
+	if ((sts = __pmAccAddHost(hostnames[i], specOps, denyOps, maxCons)) < 0) {
+	    if (sts == -EHOSTUNREACH || sts == -EHOSTDOWN)
 		fprintf(stderr, "Warning: the following access control specification will be ignored\n");
 	    fprintf(stderr,
 			 "pmcd config[line %d]: Warning: access control error for host '%s': %s\n",
-			 nLines, name[j], pmErrStr(sts));
-	    if (sts == -EHOSTUNREACH || -EHOSTDOWN)
+			 nLines, hostnames[i], pmErrStr(sts));
+	    if (sts == -EHOSTUNREACH || sts == -EHOSTDOWN)
 		;
 	    else
 		goto error;
@@ -1044,8 +1054,94 @@ ParseHosts(int allow)
     return ok;
 
 error:
-    for (j = 0; j < i; j++)
-	free(name[j]);
+    for (i = 0; i < nhosts; i++)
+	free(hostnames[i]);
+    return -1;
+}
+
+static int
+ParseUsers(int allow)
+{
+    int		sts;
+    int		nusers;
+    int		i;
+    int		ok = 0;
+    int		specOps = 0;
+    int		denyOps = 0;
+    int		maxCons = 0;		/* Zero=>unspecified, -1=>unlimited */
+    char	**usernames;
+
+    if ((nusers = ParseNames(&usernames, "user")) < 0)
+	goto error;
+
+    FindNextToken();
+    if (ParseAccessSpec(allow, &specOps, &denyOps, &maxCons, 0) < 0)
+	goto error;
+
+    if (pmDebugOptions.appl1) {
+	for (i = 0; i < nusers; i++)
+	    fprintf(stderr, "USER ACCESS: %s specOps=%02x denyOps=%02x maxCons=%d\n",
+		    usernames[i], specOps, denyOps, maxCons);
+    }
+
+    /* Make new entries for users in user access list */
+    for (i = 0; i < nusers; i++) {
+	if ((sts = __pmAccAddUser(usernames[i], specOps, denyOps, maxCons)) < 0) {
+	    fprintf(stderr,
+			 "pmcd config[line %d]: Warning: access control error for user '%s': %s\n",
+			 nLines, usernames[i], pmErrStr(sts));
+	    goto error;
+	}
+	ok = 1;
+    }
+    return ok;
+
+error:
+    for (i = 0; i < nusers; i++)
+	free(usernames[i]);
+    return -1;
+}
+
+static int
+ParseGroups(int allow)
+{
+    int		sts;
+    int		ngroups;
+    int		i;
+    int		ok = 0;
+    int		specOps = 0;
+    int		denyOps = 0;
+    int		maxCons = 0;		/* Zero=>unspecified, -1=>unlimited */
+    char	**groupnames;
+
+    if ((ngroups = ParseNames(&groupnames, "group")) < 0)
+	goto error;
+
+    FindNextToken();
+    if (ParseAccessSpec(allow, &specOps, &denyOps, &maxCons, 0) < 0)
+	goto error;
+
+    if (pmDebugOptions.appl1) {
+	for (i = 0; i < ngroups; i++)
+	    fprintf(stderr, "GROUP ACCESS: %s specOps=%02x denyOps=%02x maxCons=%d\n",
+		    groupnames[i], specOps, denyOps, maxCons);
+    }
+
+    /* Make new entries for groups in group access list */
+    for (i = 0; i < ngroups; i++) {
+	if ((sts = __pmAccAddGroup(groupnames[i], specOps, denyOps, maxCons)) < 0) {
+	    fprintf(stderr,
+			 "pmcd config[line %d]: Warning: access control error for group '%s': %s\n",
+			 nLines, groupnames[i], pmErrStr(sts));
+	    goto error;
+	}
+	ok = 1;
+    }
+    return ok;
+
+error:
+    for (i = 0; i < ngroups; i++)
+	free(groupnames[i]);
     return -1;
 }
 
@@ -1055,7 +1151,8 @@ ParseAccessControls(void)
     int		sts = 0;
     int		tmp;
     int		allow;
-    int		nHosts = 0;
+    int		naccess = 0;
+    int		need_creds = 0;
 
     doingAccess = 1;
     /* This gets a little tricky, because the token may be "[access]", or
@@ -1107,10 +1204,28 @@ ParseAccessControls(void)
 	    return -1;
 	}
 	FindNextToken();
-	if ((tmp = ParseHosts(allow)) < 0)
-	    sts = -1;
-	else if (tmp > 0)
-	    nHosts++;
+	if (TokenIs("user") || TokenIs("users")) {
+	    FindNextToken();
+	    if ((tmp = ParseUsers(allow)) < 0)
+		sts = -1;
+	    else
+		need_creds = 1;
+	} else if (TokenIs("group") || TokenIs("groups")) {
+	    FindNextToken();
+	    if ((tmp = ParseGroups(allow)) < 0)
+		sts = -1;
+	    else
+		need_creds = 1;
+	} else if (TokenIs("host") || TokenIs("hosts")) {
+	    FindNextToken();
+	    if ((tmp = ParseHosts(allow)) < 0)
+		sts = -1;
+	} else {
+	    if ((tmp = ParseHosts(allow)) < 0)
+		sts = -1;
+	}
+	if (tmp > 0)
+	    naccess++;
 	while (*token && !scanError && *token != ';')
 	    FindNextToken();
 	if (!*token || scanError)
@@ -1120,17 +1235,18 @@ ParseAccessControls(void)
     if (sts != 0)
 	return sts;
 
-    if (nHosts == 0) {
+    if (naccess == 0) {
 	fprintf(stderr,
 		     "pmcd config[line %d]: Error: no valid statements in [access] section\n",
 		     nLines);
 	return -1;
     }
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL1)
-	__pmAccDumpHosts(stderr);
-#endif
+    if (need_creds)
+	__pmServerSetFeature(PM_SERVER_FEATURE_CREDS_REQD);
+
+    if (pmDebugOptions.appl1)
+	__pmAccDumpLists(stderr);
 
     return 0;
 }
@@ -1156,8 +1272,10 @@ ReadConfigFile(FILE *configFile)
 	if (*token == '[')		/* Start of access control specs */
 	    break;
 
-	if ((pmDomainLabel = CopyToken()) == NULL)
-	    __pmNoMem("pmcd config: domain label", tokenend - token + 1, PM_FATAL_ERR);
+	if ((pmDomainLabel = CopyToken()) == NULL) {
+	    pmNoMem("pmcd config: domain label", tokenend - token + 1, PM_FATAL_ERR);
+	    /*NOTREACHED*/
+	}
 
 	FindNextToken();
 	if (TokenIsNumber()) {
@@ -1227,13 +1345,13 @@ doneLine:
 }
 
 static int
-DoAuthentication(AgentInfo *ap, int clientID)
+DoAttributes(AgentInfo *ap, int clientID)
 {
     int sts = 0;
     __pmHashCtl *attrs = &client[clientID].attrs;
     __pmHashNode *node;
 
-    if ((ap->status.flags & PDU_FLAG_AUTH) == 0)
+    if ((ap->status.flags & (PDU_FLAG_AUTH|PDU_FLAG_CONTAINER)) == 0)
 	return 0;
 
     if (ap->ipcType == AGENT_DSO) {
@@ -1243,6 +1361,12 @@ DoAuthentication(AgentInfo *ap, int clientID)
 	for (node = __pmHashWalk(attrs, PM_HASH_WALK_START);
 	     node != NULL;
 	     node = __pmHashWalk(attrs, PM_HASH_WALK_NEXT)) {
+	    if (pmDebugOptions.attr) {
+		char buffer[64];
+		__pmAttrStr_r(node->key, node->data, buffer, sizeof(buffer));
+		fprintf(stderr, "pmcd: send client[%d] attr %s to dso agent[%d]",
+			clientID, buffer, (int)(ap - agent));
+	    }
 	    if ((sts = ap->ipc.dso.dispatch.version.six.attribute(
 				clientID, node->key, node->data,
 				node->data ? strlen(node->data)+1 : 0,
@@ -1256,45 +1380,62 @@ DoAuthentication(AgentInfo *ap, int clientID)
 	for (node = __pmHashWalk(attrs, PM_HASH_WALK_START);
 	     node != NULL;
 	     node = __pmHashWalk(attrs, PM_HASH_WALK_NEXT)) {
-	    if ((sts = __pmSendAuth(ap->inFd,
+	    if (pmDebugOptions.attr) {
+		char buffer[64];
+		__pmAttrStr_r(node->key, node->data, buffer, sizeof(buffer));
+		fprintf(stderr, "pmcd: send client[%d] attr %s to daemon agent[%d]",
+			clientID, buffer, (int)(ap - agent));
+	    }
+	    if ((sts = __pmSendAttr(ap->inFd,
 				clientID, node->key, node->data,
 				node->data ? strlen(node->data)+1 : 0)) < 0)
 		break;
 	}
     }
-    return sts;
-}
-
-/*
- * Once a secure client arrives, we need to inform any interested PMDAs.
- * Iterate over the authenticating agents and send connection attributes.
- */
-int
-AgentsAuthentication(int clientID)
-{
-    int agentID, sts = 0;
-
-    for (agentID = 0; agentID < nAgents; agentID++) {
-	if (agent[agentID].status.connected &&
-	   (sts = DoAuthentication(&agent[agentID], clientID)) < 0)
-	    break;
+    if (sts < 0) {
+	if (pmDebugOptions.appl0)
+	    fprintf(stderr, "ATTR error: \"%s\" agent : %s\n",
+		    ap->pmDomainLabel, pmErrStr(sts));
+	CleanupAgent(ap, AT_COMM, ap->inFd);
+	return PM_ERR_AGAIN;
     }
     return sts;
 }
 
 /*
- * Once a PMDA has started, we need to inform it about secure clients.
- * Iterate over the authenticated clients and send connection attributes
+ * Once a new client arrives, we'll need to inform any interested PMDAs -
+ * iterate over all of the active agents and send connection attributes.
  */
 int
-ClientsAuthentication(AgentInfo *ap)
+AgentsAttributes(int clientID)
+{
+    int agentID, sts = 0;
+
+    for (agentID = 0; agentID < nAgents; agentID++) {
+	if (agent[agentID].status.connected &&
+	   (sts = DoAttributes(&agent[agentID], clientID)) < 0)
+	    break;
+    }
+    if (sts == 0 && agentID == nAgents) {
+	client[clientID].status.attributes = 0;
+    }
+    return sts;
+}
+
+/*
+ * Once a PMDA has started, we may need to inform it about the clients -
+ * iterate over the authenticated clients and send connection attributes.
+ */
+int
+ClientsAttributes(AgentInfo *ap)
 {
     int clientID, sts = 0;
 
     for (clientID = 0; clientID < nClients; clientID++) {
 	if (client[clientID].status.connected &&
-	   (sts = DoAuthentication(ap, clientID)) < 0)
+	   (sts = DoAttributes(ap, clientID)) < 0)
 	    break;
+	client[clientID].status.attributes = 0;
     }
     return sts;
 }
@@ -1313,8 +1454,7 @@ DoAgentCreds(AgentInfo* aPtr, __pmPDU *pb)
 
     if ((sts = __pmDecodeCreds(pb, &sender, &credcount, &credlist)) < 0)
 	return sts;
-    else if (_pmcd_trace_mask)
-	pmcd_trace(TR_RECV_PDU, aPtr->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
+    pmcd_trace(TR_RECV_PDU, aPtr->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 
     for (i = 0; i < credcount; i++) {
 	switch (credlist[i].c_type) {
@@ -1322,11 +1462,9 @@ DoAgentCreds(AgentInfo* aPtr, __pmPDU *pb)
 	    vcp = (__pmVersionCred *)&credlist[i];
 	    aPtr->pduVersion = version = vcp->c_version;
 	    aPtr->status.flags = flags = vcp->c_flags;
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_CONTEXT)
+	    if (pmDebugOptions.context)
 		fprintf(stderr, "pmcd: version creds (version=%u,flags=%x)\n",
 			aPtr->pduVersion, aPtr->status.flags);
-#endif
 	    break;
 	}
     }
@@ -1346,14 +1484,13 @@ DoAgentCreds(AgentInfo* aPtr, __pmPDU *pb)
 	handshake.c_type = CVERSION;
 	handshake.c_version = PDU_VERSION;
 	handshake.c_flags = (flags & PDU_FLAG_AUTH);
-	if ((sts = __pmSendCreds(aPtr->inFd, (int)getpid(), 1, cp)) < 0)
+	if ((sts = __pmSendCreds(aPtr->inFd, (int)pmcd_pid, 1, cp)) < 0)
 	    return sts;
-	else if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, aPtr->inFd, PDU_CREDS, credcount);
+	pmcd_trace(TR_XMIT_PDU, aPtr->inFd, PDU_CREDS, credcount);
 
-	/* send auth attributes for existing connected clients */
-	if ((flags & PDU_FLAG_AUTH) != 0 &&
-	    (sts = ClientsAuthentication(aPtr)) < 0)
+	/* send connection attributes for existing connected clients */
+	if ((flags & (PDU_FLAG_AUTH|PDU_FLAG_CONTAINER)) != 0 &&
+	    (sts = ClientsAttributes(aPtr)) < 0)
 	    return sts;
     }
 
@@ -1398,7 +1535,7 @@ ConnectSocketAgent(AgentInfo *aPtr)
     int		sts = 0;
     int		fd = -1;	/* pander to gcc */
 
-    if (aPtr->ipc.socket.addrDomain == AF_INET) {
+    if (aPtr->ipc.socket.addrDomain == AF_INET || aPtr->ipc.socket.addrDomain == AF_INET6) {
 	__pmSockAddr	*addr;
 	__pmHostEnt	*host;
 	void		*enumIx;
@@ -1411,10 +1548,18 @@ ConnectSocketAgent(AgentInfo *aPtr)
 	for (addr = __pmHostEntGetSockAddr(host, &enumIx);
 	     addr != NULL;
 	     addr = __pmHostEntGetSockAddr(host, &enumIx)) {
-	    if (__pmSockAddrIsInet(addr))
+	    if (__pmSockAddrIsInet(addr)) {
+		/* Only consider addresses of the chosen family. */
+		if (aPtr->ipc.socket.addrDomain != AF_INET)
+		    continue;
 	        fd = __pmCreateSocket();
-	    else if (__pmSockAddrIsIPv6(addr))
+	    }
+	    else if (__pmSockAddrIsIPv6(addr)) {
+		/* Only consider addresses of the chosen family. */
+		if (aPtr->ipc.socket.addrDomain != AF_INET6)
+		    continue;
 	        fd = __pmCreateIPv6Socket();
+	    }
 	    else {
 	        fprintf(stderr,
 			"pmcd: Error creating socket for \"%s\" agent : invalid address family %d\n",
@@ -1440,7 +1585,7 @@ ConnectSocketAgent(AgentInfo *aPtr)
 	__pmHostEntFree(host);
     }
     else {
-#if defined(HAVE_SYS_UN_H)
+#if defined(HAVE_STRUCT_SOCKADDR_UN)
 	struct sockaddr_un	addr;
 	int			len;
 
@@ -1477,7 +1622,7 @@ ConnectSocketAgent(AgentInfo *aPtr)
 
 error:
     if (fd != -1) {
-        if (aPtr->ipc.socket.addrDomain == AF_INET)
+        if (aPtr->ipc.socket.addrDomain == AF_INET || aPtr->ipc.socket.addrDomain == AF_INET6)
 	    __pmCloseSocket(fd);
 	else
 	    close(fd);
@@ -1486,6 +1631,50 @@ error:
 }
 
 #ifndef IS_MINGW
+static pid_t
+CreateRootAgentPOSIX(AgentInfo *aPtr)
+{
+    int		sts, inFd, outFd = -1;
+    char	*args;
+    pid_t	pid;
+    static int	enabled;
+
+    if (!enabled) {
+	/* once-off initialisation */
+	enabled = -EOPNOTSUPP;
+	if ((args = getenv("PMCD_ROOT_AGENT")) == NULL)
+	    enabled = 1;	/* unset, default to enabled */
+	else if (strcmp(args, "0") != 0)
+	    enabled = 1;
+	/* enabled is non-zero now */
+    }
+    if (enabled < 0)
+	return enabled;
+
+    if (aPtr->ipcType == AGENT_PIPE)
+	args = aPtr->ipc.pipe.commandLine;
+    else if (aPtr->ipcType == AGENT_SOCKET)
+	args = aPtr->ipc.socket.commandLine;
+    else
+	return -EINVAL;
+
+    if (pmdarootfd <= 0) {
+	if ((pmdarootfd = pmdaRootConnect(NULL)) <= 0)
+	    return pmdarootfd;
+    }
+
+    if ((sts = pmdaRootProcessStart(pmdarootfd, aPtr->ipcType,
+			aPtr->pmDomainLabel, strlen(aPtr->pmDomainLabel),
+			args, strlen(args), &pid, &inFd, &outFd)) < 0) {
+	return sts;
+    }
+
+    pmcd_openfds_sethi(outFd);
+    aPtr->outFd = outFd;
+    aPtr->inFd = inFd;
+    return pid;
+}
+
 static pid_t
 CreateAgentPOSIX(AgentInfo *aPtr)
 {
@@ -1676,14 +1865,19 @@ CreateAgent(AgentInfo *aPtr)
     fflush(stdout);
 
 #ifdef IS_MINGW
-    childPid = CreateAgentWin32(aPtr);
-#else
-    childPid = CreateAgentPOSIX(aPtr);
-#endif
-    if (childPid < 0)
+    if ((childPid = CreateAgentWin32(aPtr)) < 0)
 	return (int)childPid;
-
     aPtr->status.isChild = 1;
+#else
+    if ((childPid = CreateRootAgentPOSIX(aPtr)) < 0) {
+	if ((childPid = CreateAgentPOSIX(aPtr)) < 0)
+	    return (int)childPid;
+	aPtr->status.isChild = 1;
+    } else {
+	aPtr->status.isRootChild = 1;
+    }
+#endif
+
     if (aPtr->ipcType == AGENT_PIPE) {
 	aPtr->ipc.pipe.agentPid = childPid;
 	/* ready for version negotiation */
@@ -1696,11 +1890,9 @@ CreateAgent(AgentInfo *aPtr)
     else if (aPtr->ipcType == AGENT_SOCKET)
 	aPtr->ipc.socket.agentPid = childPid;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "pmcd: started PMDA %s (%d), pid=%" FMT_PID "\n",
 	        aPtr->pmDomainLabel, aPtr->pmDomainId, childPid);
-#endif
     return 0;
 }
 
@@ -1745,6 +1937,13 @@ PrintAgentInfo(FILE *stream)
 		else
 		    fprintf(stream, "dom=inet port=%d", aPtr->ipc.socket.port);
 	    }
+	    else if (aPtr->ipc.socket.addrDomain == AF_INET6) {
+		if (aPtr->ipc.socket.name)
+		    fprintf(stream, "dom=ipv6 port=%s (%d)",
+		        aPtr->ipc.socket.name, aPtr->ipc.socket.port);
+		else
+		    fprintf(stream, "dom=ipv6 port=%d", aPtr->ipc.socket.port);
+	    }
 	    else {
 		fputs("dom=???", stream);
 	    }
@@ -1782,6 +1981,9 @@ GetAgentDso(AgentInfo *aPtr)
     DsoInfo		*dso = &aPtr->ipc.dso;
     const char		*name;
     unsigned int	challenge;
+#if defined(HAVE_DLOPEN)
+    char		*dlerrstr;
+#endif
 
     aPtr->status.connected = 0;
     aPtr->reason = REASON_NOSTART;
@@ -1799,7 +2001,8 @@ GetAgentDso(AgentInfo *aPtr)
 	free(dso->pathName);
 	dso->pathName = strdup(name);
 	if (dso->pathName == NULL) {
-	    __pmNoMem("pmcd config: pathName", strlen(name), PM_FATAL_ERR);
+	    pmNoMem("pmcd config: pathName", strlen(name), PM_FATAL_ERR);
+	    /*NOTREACHED*/
 	}
 	dso->xlatePath = 1;
     }
@@ -1836,10 +2039,12 @@ GetAgentDso(AgentInfo *aPtr)
      dispatch table for the DSO. */
 
 #if defined(HAVE_DLOPEN)
+    dlerror();
     dso->initFn = (void (*)(pmdaInterface*))dlsym(dso->dlHandle, dso->entryPoint);
     if (dso->initFn == NULL) {
-	fprintf(stderr, "Couldn't find init function `%s' in %s DSO\n",
-		     dso->entryPoint, aPtr->pmDomainLabel);
+        dlerrstr = dlerror();
+	fprintf(stderr, "Couldn't find init function `%s' in %s DSO: %s\n",
+		     dso->entryPoint, aPtr->pmDomainLabel, dlerrstr);
 	dlclose(dso->dlHandle);
 	return -1;
     }
@@ -1859,7 +2064,9 @@ GetAgentDso(AgentInfo *aPtr)
      */
     challenge = 0xff;
     dso->dispatch.comm.pmda_interface = challenge;
-    dso->dispatch.comm.pmapi_version = ~PMAPI_VERSION;
+    /* set in 2 steps to avoid int to bitfield truncation warnings */
+    dso->dispatch.comm.pmapi_version = PMAPI_VERSION;
+    dso->dispatch.comm.pmapi_version = ~dso->dispatch.comm.pmapi_version;
 
     dso->dispatch.comm.flags = 0;
     dso->dispatch.status = 0;
@@ -1880,7 +2087,7 @@ GetAgentDso(AgentInfo *aPtr)
 
     if (dso->dispatch.comm.pmda_interface < PMDA_INTERFACE_2 ||
 	dso->dispatch.comm.pmda_interface > PMDA_INTERFACE_LATEST) {
-	__pmNotifyErr(LOG_ERR,
+	pmNotifyErr(LOG_ERR,
 		 "Unknown PMDA interface version (%d) used by DSO %s\n",
 		 dso->dispatch.comm.pmda_interface, aPtr->pmDomainLabel);
 #if defined(HAVE_DLOPEN)
@@ -1892,7 +2099,7 @@ GetAgentDso(AgentInfo *aPtr)
     if (dso->dispatch.comm.pmapi_version == PMAPI_VERSION_2)
 	aPtr->pduVersion = PDU_VERSION2;
     else {
-	__pmNotifyErr(LOG_ERR,
+	pmNotifyErr(LOG_ERR,
 		 "Unsupported PMAPI version (%d) used by DSO %s\n",
 		 dso->dispatch.comm.pmapi_version, aPtr->pmDomainLabel);
 #if defined(HAVE_DLOPEN)
@@ -1905,14 +2112,12 @@ GetAgentDso(AgentInfo *aPtr)
     aPtr->status.connected = 1;
     aPtr->status.flags = dso->dispatch.comm.flags;
     if (dso->dispatch.comm.flags & PDU_FLAG_AUTH)
-	ClientsAuthentication(aPtr);
+	ClientsAttributes(aPtr);
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "pmcd: started DSO PMDA %s (%d) using pmPMDA version=%d, "
 		"PDU version=%d\n", aPtr->pmDomainLabel, aPtr->pmDomainId,
 		dso->dispatch.comm.pmda_interface, aPtr->pduVersion);
-#endif
 
     return 0;
 }
@@ -1960,6 +2165,7 @@ ContactAgents(void)
 	    else
 		pmcd_trace(TR_ADD_AGENT, aPtr->pmDomainId, aPtr->inFd, aPtr->outFd);
 	    MarkStateChanges(PMCD_ADD_AGENT);
+	    pmcd_seqnum++;
 	    aPtr->status.notReady = aPtr->status.startNotReady;
 	}
 	else
@@ -1984,6 +2190,24 @@ ContactAgents(void)
     }
 }
 
+int
+VerifyConfig(char *fileName)
+{
+    FILE	*configFile;
+    int		sts = -1;
+
+    if ((configFile = fopen(fileName, "r")) == NULL) {
+	fprintf(stderr, "pmcd verify %s: %s\n", fileName, osstrerror());
+	return sts;
+    }
+    if (__pmAccAddOp(PMCD_OP_FETCH) < 0 || __pmAccAddOp(PMCD_OP_STORE) < 0)
+	fprintf(stderr, "pmcd verify cannot create access ops structures\n");
+    else
+	sts = ReadConfigFile(configFile);
+    fclose(configFile);
+    return sts;
+}
+
 int
 ParseInitAgents(char *fileName)
 {
@@ -2003,25 +2227,19 @@ ParseInitAgents(char *fileName)
     else {
 #if defined(HAVE_ST_MTIME_WITH_E) && defined(HAVE_STAT_TIME_T)
 	configFileTime = statBuf.st_mtime;
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
+	if (pmDebugOptions.appl0)
 	    fprintf(stderr, "ParseInitAgents: configFileTime=%ld sec\n",
 	        (long)configFileTime);
-#endif
 #elif defined(HAVE_ST_MTIME_WITH_SPEC)
 	configFileTime = statBuf.st_mtimespec; /* struct assignment */
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
+	if (pmDebugOptions.appl0)
 	    fprintf(stderr, "ParseInitAgents: configFileTime=%ld.%09ld sec\n",
 	        (long)configFileTime.tv_sec, (long)configFileTime.tv_nsec);
-#endif
 #elif defined(HAVE_STAT_TIMESTRUC) || defined(HAVE_STAT_TIMESPEC) || defined(HAVE_STAT_TIMESPEC_T)
 	configFileTime = statBuf.st_mtim; /* struct assignment */
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
+	if (pmDebugOptions.appl0)
 	    fprintf(stderr, "ParseInitAgents: configFileTime=%ld.%09ld sec\n",
 	        (long)configFileTime.tv_sec, (long)configFileTime.tv_nsec);
-#endif
 #else
 !bozo!
 #endif
@@ -2144,10 +2362,14 @@ DupAgent(AgentInfo *dest, AgentInfo *src)
     memcpy(&dest->status, &src->status, sizeof(dest->status));
     if (src->ipcType == AGENT_DSO) {
 	dest->ipc.dso.dlHandle = src->ipc.dso.dlHandle;
+	/*
+	 * initFn is never needed again (DSO PMDA initialization has
+	 * already been done), but copy it across so that PrintAgentInfo()
+	 * reports the init routine's address
+	 */
+	dest->ipc.dso.initFn = src->ipc.dso.initFn;
 	memcpy(&dest->ipc.dso.dispatch, &src->ipc.dso.dispatch,
 	       sizeof(dest->ipc.dso.dispatch));
-	/* initFn should never be needed */
-	dest->ipc.dso.initFn = (DsoInitPtr)0;
     }
     else if (src->ipcType == AGENT_SOCKET)
 	dest->ipc.socket.agentPid = src->ipc.socket.agentPid;
@@ -2199,7 +2421,7 @@ ParseRestartAgents(char *fileName)
 		    /* try to discover more ... */
 		    __pmPDU	*pb;
 		    sts = __pmGetPDU(ap->outFd, ANY_SIZE, TIMEOUT_NEVER, &pb);
-		    if (sts > 0 && _pmcd_trace_mask)
+		    if (sts > 0)
 			pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		    if (sts == 0)
 			pmcd_trace(TR_EOF, ap->outFd, -1, -1);
@@ -2232,31 +2454,26 @@ ParseRestartAgents(char *fileName)
      * restart any deceased agents
      */
 #if defined(HAVE_ST_MTIME_WITH_SPEC)
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "ParseRestartAgents: new configFileTime=%ld.%09ld sec\n",
 	    (long)statBuf.st_mtimespec.tv_sec, (long)statBuf.st_mtimespec.tv_nsec);
-#endif
     if (statBuf.st_mtimespec.tv_sec == configFileTime.tv_sec &&
-        statBuf.st_mtimespec.tv_nsec == configFileTime.tv_nsec) {
+        statBuf.st_mtimespec.tv_nsec == configFileTime.tv_nsec)
 #elif defined(HAVE_STAT_TIMESPEC_T) || defined(HAVE_STAT_TIMESTRUC) || defined(HAVE_STAT_TIMESPEC)
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "ParseRestartAgents: new configFileTime=%ld.%09ld sec\n",
 	    (long)statBuf.st_mtim.tv_sec, (long)statBuf.st_mtim.tv_nsec);
-#endif
     if (statBuf.st_mtim.tv_sec == configFileTime.tv_sec &&
-        statBuf.st_mtim.tv_nsec == configFileTime.tv_nsec) {
+        statBuf.st_mtim.tv_nsec == configFileTime.tv_nsec)
 #elif defined(HAVE_STAT_TIME_T)
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "ParseRestartAgents: new configFileTime=%ld sec\n",
 	    (long)configFileTime);
-#endif
-    if (statBuf.st_mtime == configFileTime) {
+    if (statBuf.st_mtime == configFileTime)
 #else
 !bozo!
 #endif
+    {
 	fprintf(stderr, "Configuration file '%s' unchanged\n", fileName);
 	fprintf(stderr, "Restarting any deceased agents:\n");
 	j = 0;
@@ -2277,12 +2494,13 @@ ParseRestartAgents(char *fileName)
 	    }
 
 	    MarkStateChanges(PMCD_RESTART_AGENT);
+	    pmcd_seqnum++;
 	}
 	PrintAgentInfo(stderr);
-	__pmAccDumpHosts(stderr);
+	__pmAccDumpLists(stderr);
 	return;
     }
-    
+
     /* Save the current agent[] and host access tables, Reset the internal
      * state of the config file parser and re-parse the config file.
      */
@@ -2293,7 +2511,7 @@ ParseRestartAgents(char *fileName)
     szAgents = 0;
     scanInit = 0;
     scanError = 0;
-    if (__pmAccSaveHosts() < 0) {
+    if (__pmAccSaveLists() < 0) {
 	fprintf(stderr, "Error saving access controls\n");
 	sts = -2;
     }
@@ -2314,12 +2532,12 @@ ParseRestartAgents(char *fileName)
 	fprintf(stderr, "Configuration left unchanged\n");
 	agent = oldAgent;
 	nAgents = oldNAgents;
-	if (sts != -2 && __pmAccRestoreHosts() < 0) {
+	if (sts != -2 && __pmAccRestoreLists() < 0) {
 	    fprintf(stderr, "Error restoring access controls!\n");
 	    exit(1);
 	}
 	PrintAgentInfo(stderr);
-	__pmAccDumpHosts(stderr);
+	__pmAccDumpLists(stderr);
 	return;
     }
 
@@ -2334,6 +2552,7 @@ ParseRestartAgents(char *fileName)
 	    if (!AgentsDiffer(&agent[i], &oldAgent[j]) &&
 		oldAgent[j].status.connected) {
 		DupAgent(&agent[i], &oldAgent[j]);
+		pmdaInterfaceMoved(&agent[i].ipc.dso.dispatch);
 		oldAgent[j].status.restartKeep = 1;
 	    }
 
@@ -2343,7 +2562,7 @@ ParseRestartAgents(char *fileName)
 	FreeAgent(&oldAgent[j]);
     }
     free(oldAgent);
-    __pmAccFreeSavedHosts();
+    __pmAccFreeSavedLists();
 
     /* Start the new agents */
     ContactAgents();
@@ -2353,32 +2572,32 @@ ParseRestartAgents(char *fileName)
 	if (agent[i].status.connected)
 	    mapdom[agent[i].pmDomainId] = i;
 
-    /* Now recaluculate the access controls for each client and update the
-     * current connection count in the hostList entries matchine the client.
+    /* Now recalculate the access controls for each client and update the
+     * connection count in the ACL entries matching the client (and account).
      * If the client is no longer permitted the connection because of a change
      * in permissions or connection limit, the client's connection is closed.
      */
     for (i = 0; i < nClients; i++) {
 	ClientInfo	*cp = &client[i];
-	int		s;
 
-	if ((s = __pmAccAddClient(cp->addr, &cp->denyOps)) < 0) {
+	if (!cp->status.connected)
+	    continue;
+	if ((sts = CheckClientAccess(cp)) >= 0)
+	    sts = CheckAccountAccess(cp);
+	if (sts < 0) {
 	    /* ignore errors, the client is being terminated in any case */
-	    if (_pmcd_trace_mask)
-		pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, s);
-	    __pmSendError(cp->fd, FROM_ANON, s);
-	    CleanupClient(cp, s);
+	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
+	    __pmSendError(cp->fd, FROM_ANON, sts);
+	    CleanupClient(cp, sts);
 	}
     }
 
     PrintAgentInfo(stderr);
-    __pmAccDumpHosts(stderr);
+    __pmAccDumpLists(stderr);
 
-    /*
-     * gather any deceased children, some may be PMDAs that were
+    /* Gather any deceased children, some may be PMDAs that were
      * terminated by CleanupAgent or killed and had not exited
      * when the previous harvest() was done
      */
     HarvestAgents(0);
 }
-
