@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2013 Red Hat.
+ * Copyright (c) 2012-2018 Red Hat.
  * Copyright (c) 1995 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -14,7 +14,7 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "pmcd.h"
 
 /* Freq. histogram: pmids for each agent in current fetch request */
@@ -64,7 +64,7 @@ SplitPmidList(int nPmids, pmID *pmidList)
 	resIndex = (int *)malloc((nAgents + 1) * sizeof(int));
 	aFreq = (int *)malloc((nAgents + 1) * sizeof(int));
 	if (resIndex == NULL || aFreq == NULL) {
-	    __pmNoMem("SplitPmidList.resIndex", 2 * (nAgents + 1) * sizeof(int), PM_FATAL_ERR);
+	    pmNoMem("SplitPmidList.resIndex", 2 * (nAgents + 1) * sizeof(int), PM_FATAL_ERR);
 	}
     }
 
@@ -119,7 +119,7 @@ doit:
 	    free(result);
 	result = (DomPmidList *)malloc(resultSize);
 	if (result == NULL) {
-	    __pmNoMem("SplitPmidList.result", resultSize, PM_FATAL_ERR);
+	    pmNoMem("SplitPmidList.result", resultSize, PM_FATAL_ERR);
 	}
 	currentSize = resultSize;
     }
@@ -176,13 +176,15 @@ MakeBadResult(int npmids, pmID *list, int sts)
 	/* npmids - 1 because there is already 1 pmValueSet* in a pmResult */
     result = (pmResult *)malloc(need);
     if (result == NULL) {
-	__pmNoMem("MakeBadResult.result", need, PM_FATAL_ERR);
+	pmNoMem("MakeBadResult.result", need, PM_FATAL_ERR);
     }
+    result->timestamp.tv_sec = 0;
+    result->timestamp.tv_usec = 0;
     result->numpmid = npmids;
     for (i = 0; i < npmids; i++) {
 	vSet = (pmValueSet *)malloc(sizeof(pmValueSet));
 	if (vSet == NULL) {
-	    __pmNoMem("MakeBadResult.vSet", sizeof(pmValueSet), PM_FATAL_ERR);
+	    pmNoMem("MakeBadResult.vSet", sizeof(pmValueSet), PM_FATAL_ERR);
 	}
 	result->vset[i] = vSet;
 	vSet->pmid = list[i];
@@ -194,14 +196,16 @@ MakeBadResult(int npmids, pmID *list, int sts)
 static pmResult *
 SendFetch(DomPmidList *dpList, AgentInfo *aPtr, ClientInfo *cPtr, int ctxnum)
 {
+    __pmHashCtl		*hcp;
+    __pmHashNode	*hp;
+    pmProfile		*profile;
     pmResult		*result = NULL;
     int			sts = 0;
-    static __pmTimeval	when = {0, 0};	/* Agents never see archive requests */
+    static pmTimeval	when = {0, 0};	/* Agents never see archive requests */
     int			bad = 0;
     int			i;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0) {
+    if (pmDebugOptions.appl0) {
 	fprintf(stderr, "SendFetch %d metrics to PMDA domain %d ",
 	    dpList->listSize, dpList->domain);
 	switch (aPtr->ipcType) {
@@ -224,7 +228,6 @@ SendFetch(DomPmidList *dpList, AgentInfo *aPtr, ClientInfo *cPtr, int ctxnum)
 	for (i = 0; i < dpList->listSize; i++)
 	    fprintf(stderr, "  pmid[%d] %s\n", i, pmIDStr(dpList->list[i]));
     }
-#endif
 
     /* status.madeDsoResult is only used for DSO agents so don't waste time by
      * checking that the agent is a DSO first.
@@ -232,18 +235,23 @@ SendFetch(DomPmidList *dpList, AgentInfo *aPtr, ClientInfo *cPtr, int ctxnum)
     aPtr->status.madeDsoResult = 0;
 
     if (aPtr->profClient != cPtr || ctxnum != aPtr->profIndex) {
+	hcp = &cPtr->profile;
+	hp = __pmHashSearch(ctxnum, hcp);
+	if (hp != NULL)
+	    profile = (pmProfile *)hp->data;
+	else
+	    profile = NULL;
 	if (aPtr->ipcType == AGENT_DSO) {
 	    if (aPtr->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 		aPtr->ipc.dso.dispatch.version.four.ext->e_context = cPtr - client;
-	    sts = aPtr->ipc.dso.dispatch.version.any.profile(cPtr->profile[ctxnum],
+	    sts = aPtr->ipc.dso.dispatch.version.any.profile(profile,
 				     aPtr->ipc.dso.dispatch.version.any.ext);
 	}
 	else {
 	    if (aPtr->status.notReady == 0) {
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, aPtr->inFd, PDU_PROFILE, ctxnum);
+		pmcd_trace(TR_XMIT_PDU, aPtr->inFd, PDU_PROFILE, ctxnum);
 		if ((sts = __pmSendProfile(aPtr->inFd, cPtr - client,
-					   ctxnum, cPtr->profile[ctxnum])) < 0) {
+					   ctxnum, profile)) < 0) {
 		    pmcd_trace(TR_XMIT_ERR, aPtr->inFd, PDU_PROFILE, sts);
 		}
 	    } else {
@@ -266,29 +274,26 @@ SendFetch(DomPmidList *dpList, AgentInfo *aPtr, ClientInfo *cPtr, int ctxnum)
 				   aPtr->ipc.dso.dispatch.version.any.ext);
 	    if (sts >= 0) {
 		if (result == NULL) {
-		    __pmNotifyErr(LOG_WARNING,
-				 "\"%s\" agent (DSO) returned a null result\n",
-				 aPtr->pmDomainLabel);
+		    pmNotifyErr(LOG_WARNING,
+				"\"%s\" agent (DSO) returned a null result\n",
+				aPtr->pmDomainLabel);
 		    sts = PM_ERR_PMID;
 		    bad = 1;
 		}
-		else {
-		    if (result->numpmid != dpList->listSize) {
-			__pmNotifyErr(LOG_WARNING,
-				     "\"%s\" agent (DSO) returned %d pmIDs (%d expected)\n",
-				     aPtr->pmDomainLabel,
-				     result->numpmid,dpList->listSize);
-			sts = PM_ERR_PMID;
-			bad = 2;
-		    }
+		else if (result->numpmid != dpList->listSize) {
+		    pmNotifyErr(LOG_WARNING,
+				"\"%s\" agent (DSO) returned %d pmIDs (%d expected)\n",
+				aPtr->pmDomainLabel,
+				result->numpmid,dpList->listSize);
+		    sts = PM_ERR_PMID;
+		    bad = 2;
 		}
 	    }
 	}
 	else {
 	    if (aPtr->status.notReady == 0) {
 		/* agent is ready for PDUs */
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, aPtr->inFd, PDU_FETCH, dpList->listSize);
+		pmcd_trace(TR_XMIT_PDU, aPtr->inFd, PDU_FETCH, dpList->listSize);
 		if ((sts = __pmSendFetch(aPtr->inFd, cPtr - client, ctxnum, &when,
 				   dpList->listSize, dpList->list)) < 0)
 		    pmcd_trace(TR_XMIT_ERR, aPtr->inFd, PDU_FETCH, sts);
@@ -301,8 +306,7 @@ SendFetch(DomPmidList *dpList, AgentInfo *aPtr, ClientInfo *cPtr, int ctxnum)
     }
 
     if (sts < 0) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
+	if (pmDebugOptions.appl0)
 	    switch (bad) {
 		case 0:
 		    fprintf(stderr, "FETCH error: \"%s\" agent : %s\n",
@@ -318,7 +322,6 @@ SendFetch(DomPmidList *dpList, AgentInfo *aPtr, ClientInfo *cPtr, int ctxnum)
 			    result->numpmid, dpList->listSize);
 		    break;
 	    }
-#endif
 	if (aPtr->ipcType == AGENT_DSO) {
 	    aPtr->status.madeDsoResult = 1;
 	    sts = 0;
@@ -332,13 +335,29 @@ SendFetch(DomPmidList *dpList, AgentInfo *aPtr, ClientInfo *cPtr, int ctxnum)
     return result;
 }
 
+/*
+ * pmResults coming back from PMDAs have their timestamp field
+ * overloaded to contain out-of-band information such as state
+ * changes that may need to be communicated back to clients.
+ * Extract the flags that indicate those state changes here.
+ */
+static int
+ExtractState(pmResult *result)
+{
+    unsigned char	byte;
+
+    memcpy(&byte, &result->timestamp, sizeof(unsigned char));
+    return (int)byte;
+}
+
 int
 DoFetch(ClientInfo *cip, __pmPDU* pb)
 {
     int			i, j;
     int 		sts;
     int			ctxnum;
-    __pmTimeval		when;
+    unsigned int	changes = 0;
+    pmTimeval		when;
     int			nPmids;
     pmID		*pmidList;
     static pmResult	*endResult = NULL;
@@ -352,6 +371,9 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
     int			nWait;
     int			maxFd;
     struct timeval	timeout;
+    __pmHashCtl		*hcp;
+    __pmHashNode	*hp;
+    pmProfile		*profile;
 
     if (nAgents > nDoms) {
 	if (results != NULL)
@@ -361,7 +383,7 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
 	results = (pmResult **)malloc((nAgents + 1) * sizeof (pmResult *));
 	resIndex = (int *)malloc((nAgents + 1) * sizeof(int));
 	if (results == NULL || resIndex == NULL) {
-	    __pmNoMem("DoFetch.results", (nAgents + 1) * sizeof (pmResult *) + (nAgents + 1) * sizeof(int), PM_FATAL_ERR);
+	    pmNoMem("DoFetch.results", (nAgents + 1) * sizeof (pmResult *) + (nAgents + 1) * sizeof(int), PM_FATAL_ERR);
 	}
 	nDoms = nAgents;
     }
@@ -372,13 +394,19 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
 	return sts;
 
     /* Check that a profile has been received from the specified context */
-    if (ctxnum < 0 || ctxnum >= cip->szProfile ||
-	cip->profile[ctxnum] == NULL) {
+    profile = NULL;
+    if (ctxnum >= 0) {
+	hcp = &cip->profile;
+	hp = __pmHashSearch(ctxnum, hcp);
+	if (hp != NULL)
+	    profile = (pmProfile *)hp->data;
+    }
+    if (ctxnum < 0 || profile == NULL) {
 	__pmUnpinPDUBuf(pb);
-	if (ctxnum < 0 || ctxnum >= cip->szProfile)
-	    __pmNotifyErr(LOG_ERR, "DoFetch: bad ctxnum=%d\n", ctxnum);
+	if (ctxnum < 0)
+	    pmNotifyErr(LOG_ERR, "DoFetch: bad ctxnum=%d\n", ctxnum);
 	else
-	    __pmNotifyErr(LOG_ERR, "DoFetch: no profile for ctxnum=%d\n", ctxnum);
+	    pmNotifyErr(LOG_ERR, "DoFetch: no profile for ctxnum=%d\n", ctxnum);
 	return PM_ERR_NOPROFILE;
     }
 
@@ -388,7 +416,7 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
 	    free(endResult);
 	need = (int)sizeof(pmResult) + (nPmids - 1) * (int)sizeof(pmValueSet *);
 	if ((endResult = (pmResult *)malloc(need)) == NULL) {
-	    __pmNoMem("DoFetch.endResult", need, PM_FATAL_ERR);
+	    pmNoMem("DoFetch.endResult", need, PM_FATAL_ERR);
 	}
 	maxnpmids = nPmids;
     }
@@ -414,6 +442,8 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
 	    if (fd > maxFd)
 		maxFd = fd;
 	    nWait++;
+	} else {
+	    changes |= ExtractState(results[j]);
 	}
     }
     /* Construct pmResult for bad-pmID list */
@@ -424,13 +454,15 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
     while (nWait > 0) {
         __pmFD_COPY(&readyFds, &waitFds);
 	if (nWait > 1) {
-	    timeout.tv_sec = _pmcd_timeout;
+	    timeout.tv_sec = pmcd_timeout;
 	    timeout.tv_usec = 0;
 
+            retry:
+	    setoserror(0);
 	    sts = __pmSelectRead(maxFd+1, &readyFds, &timeout);
 
 	    if (sts == 0) {
-		__pmNotifyErr(LOG_INFO, "DoFetch: select timeout");
+		pmNotifyErr(LOG_INFO, "DoFetch: select timeout");
 
 		/* Timeout, terminate agents with undelivered results */
 		for (i = 0; i < nAgents; i++) {
@@ -449,8 +481,10 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
 		break;
 	    }
 	    else if (sts < 0) {
+		if (neterror() == EINTR)
+		    goto retry;
 		/* this is not expected to happen! */
-		__pmNotifyErr(LOG_ERR, "DoFetch: fatal select failure: %s\n",
+		pmNotifyErr(LOG_ERR, "DoFetch: fatal select failure: %s\n",
 			netstrerror());
 		Shutdown();
 		exit(1);
@@ -466,20 +500,21 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
 	    ap->status.busy = 0;
 	    __pmFD_CLR(ap->outFd, &waitFds);
 	    nWait--;
-	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-	    if (sts > 0 && _pmcd_trace_mask)
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_RESULT) {
-		if ((sts = __pmDecodeResult(pb, &results[i])) >= 0)
-		    if (results[i]->numpmid != aFreq[i]) {
+		if ((sts = __pmDecodeResult(pb, &results[i])) >= 0) {
+		    if (results[i]->numpmid == aFreq[i]) {
+			changes |= ExtractState(results[i]);
+		    } else {
+			if (pmDebugOptions.appl0)
+			    pmNotifyErr(LOG_ERR, "DoFetch: \"%s\" agent given %d pmIDs, returned %d\n",
+					 ap->pmDomainLabel, aFreq[i], results[i]->numpmid);
 			pmFreeResult(results[i]);
 			sts = PM_ERR_IPC;
-#ifdef PCP_DEBUG
-			if (pmDebug & DBG_TRACE_APPL0)
-			    __pmNotifyErr(LOG_ERR, "DoFetch: \"%s\" agent given %d pmIDs, returned %d\n",
-					 ap->pmDomainLabel, aFreq[i], results[i]->numpmid);
-#endif
 		    }
+		}
 	    }
 	    else {
 		if (sts == PDU_ERROR) {
@@ -516,23 +551,24 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
 		    sts = CheckError(&agent[i], sts);
 		}
 
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_APPL0) {
+		if (pmDebugOptions.appl0) {
 		    fprintf(stderr, "RESULT error from \"%s\" agent : %s\n",
 			    ap->pmDomainLabel, pmErrStr(sts));
 		}
-#endif
 		if (sts == PM_ERR_IPC || sts == PM_ERR_TIMEOUT)
 		    CleanupAgent(ap, AT_COMM, ap->outFd);
 	    }
 	}
     }
 
+    if (changes)
+	MarkStateChanges(changes);
+
     endResult->numpmid = nPmids;
-    __pmtimevalNow(&endResult->timestamp);
+    pmtimevalNow(&endResult->timestamp);
     /* The order of the pmIDs in the per-domain results is the same as in the
      * original request, but on a per-domain basis.  resIndex is an array of
-     * indeces (one per agent) of the next metric to be retrieved from each
+     * indices (one per agent) of the next metric to be retrieved from each
      * per-domain result's vset.
      */
     memset(resIndex, 0, (nAgents + 1) * sizeof(resIndex[0]));
@@ -541,8 +577,7 @@ DoFetch(ClientInfo *cip, __pmPDU* pb)
 	j = mapdom[((__pmID_int *)&pmidList[i])->domain];
 	endResult->vset[i] = results[j]->vset[resIndex[j]++];
     }
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cip->fd, PDU_RESULT, endResult->numpmid);
+    pmcd_trace(TR_XMIT_PDU, cip->fd, PDU_RESULT, endResult->numpmid);
 
     sts = 0;
     if (cip->status.changes) {
