@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2013 Red Hat.
+ * Copyright (c) 2012-2014,2017-2018 Red Hat.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -29,12 +29,10 @@ CheckError(AgentInfo *ap, int sts)
     int		retSts;
 
     if (sts == PM_ERR_PMDANOTREADY) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
-	    __pmNotifyErr(LOG_INFO, "%s agent (%s) sent NOT READY\n",
+	if (pmDebugOptions.appl0)
+	    pmNotifyErr(LOG_INFO, "%s agent (%s) sent NOT READY\n",
 			 ap->pmDomainLabel,
 			 ap->status.notReady ? "not ready" : "ready");
-#endif
 	if (ap->status.notReady == 0) {
 	    ap->status.notReady = 1;
 	    retSts = PM_ERR_AGAIN;
@@ -43,12 +41,10 @@ CheckError(AgentInfo *ap, int sts)
 	    retSts = PM_ERR_IPC;
     }
     else if (sts == PM_ERR_PMDAREADY) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
-	    __pmNotifyErr(LOG_INFO, "%s agent (%s) sent unexpected READY\n",
+	if (pmDebugOptions.appl0)
+	    pmNotifyErr(LOG_INFO, "%s agent (%s) sent unexpected READY\n",
 			 ap->pmDomainLabel,
 			 ap->status.notReady ? "not ready" : "ready");
-#endif
 	retSts = PM_ERR_IPC;
     }
     else
@@ -64,7 +60,7 @@ DoText(ClientInfo *cp, __pmPDU* pb)
     int		ident;
     int		type;
     AgentInfo	*ap;
-    char	*buffer;
+    char	*buffer = NULL;
 
     if ((sts = __pmDecodeTextReq(pb, &ident, &type)) < 0)
 	return sts;
@@ -83,13 +79,12 @@ DoText(ClientInfo *cp, __pmPDU* pb)
     else {
 	if (ap->status.notReady)
 	    return PM_ERR_AGAIN;
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_TEXT_REQ, ident);
+	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_TEXT_REQ, ident);
 	sts = __pmSendTextReq(ap->inFd, cp - client, ident, type);
 	if (sts >= 0) {
 	    int		pinpdu;
-	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-	    if (sts > 0 && _pmcd_trace_mask)
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_TEXT)
 		sts = __pmDecodeText(pb, &ident, &buffer);
@@ -117,8 +112,7 @@ DoText(ClientInfo *cp, __pmPDU* pb)
 	CleanupAgent(ap, AT_COMM, ap->inFd);	
 
     if (sts >= 0) {
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_TEXT, ident);
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_TEXT, ident);
 	sts = __pmSendText(cp->fd, FROM_ANON, ident, buffer);
 	if (sts < 0 && ap->ipcType != AGENT_DSO) {
 	    pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_TEXT, sts);
@@ -135,41 +129,28 @@ DoText(ClientInfo *cp, __pmPDU* pb)
 int
 DoProfile(ClientInfo *cp, __pmPDU *pb)
 {
-    __pmProfile	*newProf;
+    __pmHashCtl	*hcp;
+    pmProfile	*newProf;
     int		ctxnum, sts, i;
 
     sts = __pmDecodeProfile(pb, &ctxnum, &newProf);
     if (sts >= 0) {
-	/* Allocate more profile pointers if required */
-	if (ctxnum >= cp->szProfile) {
-	    __pmProfile	**newProfPtrs;
-	    int		need, oldSize = cp->szProfile;
-
-	    if (ctxnum - cp->szProfile < 4)
-		cp->szProfile += 4;
-	    else
-		cp->szProfile = ctxnum + 1;
-	    need = cp->szProfile * (int)sizeof(__pmProfile *);
-	    if ((newProfPtrs = (__pmProfile **)malloc(need)) == NULL) {
-		cp->szProfile = oldSize;
-		__pmNoMem("DoProfile.newProfPtrs", need, PM_RECOV_ERR);
-		__pmFreeProfile(newProf);
-		return -oserror();
+	__pmHashNode	*hp;
+	hcp = &cp->profile;
+	if ((hp = __pmHashSearch(ctxnum, hcp)) != NULL) {
+	    /* seen this context slot before for this client */
+	    pmProfile	*profile = (pmProfile *)hp->data;
+	    if (profile != NULL)
+		__pmFreeProfile(profile);
+	    hp->data = (void *)newProf;
+	}
+	else {
+	    /* first time for this context slot for this client */
+	    if ((sts = __pmHashAdd(ctxnum, newProf, hcp)) > 0) {
+		/* __pmHashAdd returns 1 for success, but we want zero. */
+		sts = 0;
 	    }
-
-	    /* Copy any old pointers and zero the newly allocated ones */
-	    if ((need = oldSize * (int)sizeof(__pmProfile *))) {
-		memcpy(newProfPtrs, cp->profile, need);
-		free(cp->profile);	/* But not the __pmProfile ptrs! */
-	    }
-	    need = (cp->szProfile - oldSize) * (int)sizeof(__pmProfile *);
-	    memset(&newProfPtrs[oldSize], 0, need);
-	    cp->profile = newProfPtrs;
 	}
-	else				/* cp->profile is big enough */
-	    if (cp->profile[ctxnum] != NULL)
-		__pmFreeProfile(cp->profile[ctxnum]);
-	cp->profile[ctxnum] = newProf;
 
 	/* "Invalidate" any references to the client context's profile in the
 	 * agents to which the old profile was last sent
@@ -190,7 +171,7 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
     int		sts, s;
     pmID	pmid;
     AgentInfo	*ap;
-    pmDesc	desc;
+    pmDesc	desc = {0};
     int		fdfail = -1;
 
     if ((sts = __pmDecodeDescReq(pb, &pmid)) < 0)
@@ -210,13 +191,12 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
     else {
 	if (ap->status.notReady)
 	    return PM_ERR_AGAIN;
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_DESC_REQ, (int)pmid);
+	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_DESC_REQ, (int)pmid);
 	sts = __pmSendDescReq(ap->inFd, cp - client, pmid);
 	if (sts >= 0) {
 	    int		pinpdu;
-	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-	    if (sts > 0 && _pmcd_trace_mask)
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_DESC)
 		sts = __pmDecodeDesc(pb, &desc);
@@ -243,8 +223,7 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
     }
 
     if (sts >= 0) {
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_DESC, (int)desc.pmid);
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_DESC, (int)desc.pmid);
 	sts = __pmSendDesc(cp->fd, FROM_ANON, &desc);
 	if (sts < 0) {
 	    pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_DESC, sts);
@@ -261,14 +240,14 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
 }
 
 int
-DoInstance(ClientInfo *cp, __pmPDU* pb)
+DoInstance(ClientInfo *cp, __pmPDU *pb)
 {
     int			sts, s;
-    __pmTimeval		when;
+    pmTimeval		when;
     pmInDom		indom;
     int			inst;
     char		*name;
-    __pmInResult		*inresult = NULL;
+    pmInResult	*inresult = NULL;
     AgentInfo		*ap;
     int			fdfail = -1;
 
@@ -276,15 +255,8 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
     if (sts < 0)
 	return sts;
     if (when.tv_sec != 0 || when.tv_usec != 0) {
-	/*
-	 * we have no idea how to do anything but current, yet!
-	 *
-	 * TODO EXCEPTION PCP 2.0 ...
-	 * this may be left over from the pmvcr days, and can be tossed?
-	 * ... leaving it here is benign
-	 */
 	if (name != NULL) free(name);
-	return PM_ERR_NYI;
+	return PM_ERR_IPC;
     }
     if ((ap = FindDomainAgent(((__pmInDom_int *)&indom)->domain)) == NULL) {
 	if (name != NULL) free(name);
@@ -307,13 +279,12 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
 	    if (name != NULL) free(name);
 	    return PM_ERR_AGAIN;
 	}
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_INSTANCE_REQ, (int)indom);
+	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_INSTANCE_REQ, (int)indom);
 	sts = __pmSendInstanceReq(ap->inFd, cp - client, &when, indom, inst, name);
 	if (sts >= 0) {
 	    int		pinpdu;
-	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-	    if (sts > 0 && _pmcd_trace_mask)
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_INSTANCE)
 		sts = __pmDecodeInstance(pb, &inresult);
@@ -342,15 +313,13 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
     if (name != NULL) free(name);
 
     if (sts >= 0) {
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_INSTANCE, (int)(inresult->indom));
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_INSTANCE, (int)(inresult->indom));
 	sts = __pmSendInstance(cp->fd, FROM_ANON, inresult);
 	if (sts < 0) {
 	    pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_INSTANCE, sts);
 	    CleanupClient(cp, sts);
 	}
-	if (inresult != NULL)
-	    __pmFreeInResult(inresult);
+	__pmFreeInResult(inresult);
     }
     else
 	if (ap->ipcType != AGENT_DSO &&
@@ -361,6 +330,230 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
     return sts;
 }
 
+static int
+GetChangedContextLabels(pmLabelSet **sets, int *changed)
+{
+    int		sts;
+
+    *changed = 0;
+    if ((sts = __pmGetContextLabels(sets)) == 1) {
+	if (!pmcd_labels) {
+	    *changed = 1;
+	} else {
+	    if (strcmp(pmcd_labels, sets[0]->json) != 0)
+		*changed = 1;
+	    free(pmcd_labels);
+	}
+	pmcd_labels = strndup(sets[0]->json, sets[0]->jsonlen);
+    }
+    else if (pmcd_labels) {
+	free(pmcd_labels);
+	pmcd_labels = NULL;
+	*changed = 1;
+    }
+    return sts;
+}
+
+void
+CheckLabelChange(void)
+{
+    pmLabelSet	*sets = NULL;
+    int		nsets;
+
+    if ((nsets = GetChangedContextLabels(&sets, &labelChanged)) > 0)
+	pmFreeLabelSets(sets, nsets);
+}
+
+static int
+GetContextLabels(ClientInfo *cp, pmLabelSet **sets)
+{
+    __pmHashNode	*node;
+    const char		*userid;
+    const char		*groupid;
+    const char		*container;
+    static const char	func[] = "GetContextLabels";
+    static char		host[MAXHOSTNAMELEN];
+    static char		domain[MAXDOMAINNAMELEN];
+    static char		machineid[MAXMACHINEIDLEN];
+    char		buf[PM_MAXLABELJSONLEN];
+    char		*hostname;
+    int			sts;
+
+    if ((sts = GetChangedContextLabels(sets, &labelChanged)) >= 0) {
+	if ((hostname = pmcd_hostname) == NULL) {
+	    if ((sts = gethostname(host, MAXHOSTNAMELEN)) < 0) {
+		if (pmDebugOptions.labels)
+		    fprintf(stderr, "%s: gethostname() -> %d (%s)\n",
+			    func, sts, pmErrStr(sts));
+		host[0] = '\0';
+	    }
+	    if (host[0] == '\0')
+		pmsprintf(host, sizeof(host), "localhost");
+	    hostname = pmcd_hostname = host;
+	}
+	if (domain[0] == '\0') {
+	    if ((sts = getdomainname(domain, MAXDOMAINNAMELEN)) < 0) {
+		if (pmDebugOptions.labels)
+		    fprintf(stderr, "%s: getdomainname() -> %d (%s)\n",
+			    func, sts, pmErrStr(sts));
+		domain[0] = '\0';
+	    }
+	    if (domain[0] == '\0' || strcmp(domain, "(none)") == 0)
+		pmsprintf(domain, sizeof(domain), "localdomain");
+	}
+	if (machineid[0] == '\0') {
+	    if ((sts = getmachineid(machineid, MAXMACHINEIDLEN)) < 0) {
+		if (pmDebugOptions.labels)
+		    fprintf(stderr, "%s: getmachineid() -> %d (%s)\n",
+			    func, sts, pmErrStr(sts));
+		machineid[0] = '\0';
+	    }
+	    if (machineid[0] == '\0')
+		pmsprintf(machineid, sizeof(machineid), "localmachine");
+	}
+	userid = ((node = __pmHashSearch(PCP_ATTR_USERID, &cp->attrs)) ?
+			(const char *)node->data : NULL);
+	groupid = ((node = __pmHashSearch(PCP_ATTR_GROUPID, &cp->attrs)) ?
+			(const char *)node->data : NULL);
+	container = ((node = __pmHashSearch(PCP_ATTR_CONTAINER, &cp->attrs)) ?
+			(const char *)node->data : NULL);
+
+	sts = pmsprintf(buf, sizeof(buf), "{\"hostname\":\"%s\"", hostname);
+	if (domain[0] != '\0')
+	    sts += pmsprintf(buf+sts, sizeof(buf)-sts, ",\"domainname\":\"%s\"",
+			    domain);
+	if (machineid[0] != '\0')
+	    sts += pmsprintf(buf+sts, sizeof(buf)-sts, ",\"machineid\":\"%s\"",
+			    machineid);
+	if (userid)
+	    sts += pmsprintf(buf+sts, sizeof(buf)-sts, ",\"userid\":%s",
+			    userid);
+	if (groupid)
+	    sts += pmsprintf(buf+sts, sizeof(buf)-sts, ",\"groupid\":%s",
+			    groupid);
+	if (container)
+	    sts += pmsprintf(buf+sts, sizeof(buf)-sts, ",\"container\":%s",
+			    container);
+	pmsprintf(buf+sts, sizeof(buf)-sts, "}");
+	if ((sts = __pmAddLabels(sets, buf, PM_LABEL_CONTEXT)) > 0)
+	    return 1;
+    }
+    return sts;
+}
+
+int
+DoLabel(ClientInfo *cp, __pmPDU *pb)
+{
+    int			sts, s;
+    int			ident, type, nsets = 0;
+    pmLabelSet		*sets = NULL;
+    AgentInfo		*ap = NULL;
+    int			fdfail = -1;
+
+    sts = __pmDecodeLabelReq(pb, &ident, &type);
+    if (sts < 0)
+	return sts;
+
+    switch (type) {
+	case PM_LABEL_CONTEXT:
+	    nsets = sts = GetContextLabels(cp, &sets);
+	    goto response;
+	case PM_LABEL_DOMAIN:
+	    if (!(ap = FindDomainAgent(ident)))
+		return PM_ERR_NOAGENT;
+	    break;
+	case PM_LABEL_INDOM:
+	    if (!(ap = FindDomainAgent(((__pmInDom_int *)&ident)->domain)))
+		return PM_ERR_INDOM;
+	    break;
+	case PM_LABEL_CLUSTER:
+	case PM_LABEL_ITEM:
+	case PM_LABEL_INSTANCES:
+	    if (!(ap = FindDomainAgent(((__pmID_int *)&ident)->domain)))
+		return PM_ERR_PMID;
+	    break;
+	default:
+	    return PM_ERR_TYPE;
+    }
+
+    if (!ap->status.connected)
+	return PM_ERR_NOAGENT;
+
+    if (ap->ipcType == AGENT_DSO) {
+	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
+	    ap->ipc.dso.dispatch.version.seven.ext->e_context = cp - client;
+	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_7) {
+	    sts = ap->ipc.dso.dispatch.version.seven.label(ident, type, &sets,
+					ap->ipc.dso.dispatch.version.any.ext);
+	} else if (type & PM_LABEL_DOMAIN) {
+	    sts = __pmGetDomainLabels(ap->pmDomainId, ap->pmDomainLabel, &sets);
+	    nsets = sts;
+	    goto response;
+	} else {
+	    sts = 0;
+	}
+	if (sts >= 0)
+	    nsets = sts;
+    }
+    else {
+	if (ap->status.notReady)
+	    return PM_ERR_AGAIN;
+
+	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_LABEL_REQ, ident);
+	sts = __pmSendLabelReq(ap->inFd, cp - client, ident, type);
+	if (sts >= 0) {
+	    int		pinpdu;
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+	    if (sts > 0)
+		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
+	    if (sts == PDU_LABEL)
+		sts = __pmDecodeLabel(pb, &ident, &type, &sets, &nsets);
+	    else if (sts == PDU_ERROR) {
+		nsets = 0;
+		sets = NULL;
+		s = __pmDecodeError(pb, &sts);
+		if (s < 0)
+		    sts = s;
+		else
+		    sts = CheckError(ap, sts);
+		pmcd_trace(TR_RECV_ERR, ap->outFd, PDU_LABEL, sts);
+	    }
+	    else {
+		pmcd_trace(TR_WRONG_PDU, ap->outFd, PDU_LABEL, sts);
+		sts = PM_ERR_IPC;	/* Wrong PDU type */
+		fdfail = ap->outFd;
+	    }
+	    if (pinpdu > 0)
+		__pmUnpinPDUBuf(pb);
+	}
+	else {
+	    pmcd_trace(TR_XMIT_ERR, ap->inFd, PDU_LABEL_REQ, sts);
+	    fdfail = ap->inFd;
+	}
+    }
+
+response:
+    if (sts >= 0) {
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_LABEL, (int)ident);
+	if (nsets > 1 && !(type & PM_LABEL_INSTANCES))
+	    nsets = 1;
+	sts = __pmSendLabel(cp->fd, FROM_ANON, ident, type, sets, nsets);
+	if (sts < 0) {
+	    pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_LABEL, sts);
+	    CleanupClient(cp, sts);
+	}
+	pmFreeLabelSets(sets, nsets);
+    }
+    else {
+	if (ap && ap->ipcType != AGENT_DSO &&
+	    (sts == PM_ERR_IPC || sts == PM_ERR_TIMEOUT || sts == -EPIPE) &&
+	    fdfail != -1)
+	    CleanupAgent(ap, AT_COMM, fdfail);
+    }
+
+    return sts;
+}
+
 /*
  * This handler is for remote versions of pmNameAll or pmNameID.
  * Note: only one pmid for the list should be sent.
@@ -408,13 +601,12 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
 	    /* daemon PMDA ... ship request on */
 	    if (ap->status.notReady)
 		return PM_ERR_AGAIN;
-	    if (_pmcd_trace_mask)
-		pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_IDS, 1);
+	    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_IDS, 1);
 	    sts = __pmSendIDList(ap->inFd, cp - client, 1, &idlist[0], 0);
 	    if (sts >= 0) {
 		int		pinpdu;
-		pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-		if (sts > 0 && _pmcd_trace_mask)
+		pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+		if (sts > 0)
 		    pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		if (sts == PDU_PMNS_NAMES) {
 		    sts = __pmDecodeNameList(pb, &numnames, &namelist, NULL);
@@ -442,8 +634,7 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
 
     numnames = sts;
 
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
     if ((sts = __pmSendNameList(cp->fd, FROM_ANON, numnames, namelist, NULL)) < 0){
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
@@ -466,8 +657,11 @@ fail:
 int
 DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 {
-    int		sts = 0;
+    int		sts;
     int		numids = 0;
+    int		numok;
+    int		lsts;
+    int		domain;
     pmID	*idlist = NULL;
     char	**namelist = NULL;
     int		i;
@@ -476,31 +670,35 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
     if ((sts = __pmDecodeNameList(pb, &numids, &namelist, NULL)) < 0)
 	goto done;
 
-    if ((idlist = (pmID*)malloc(sizeof(int)*numids)) == NULL) {
+    if ((idlist = (pmID *)calloc(numids, sizeof(int))) == NULL) {
         sts = -oserror();
 	goto done;
     }
 
     sts = pmLookupName(numids, namelist, idlist);
+    /*
+     * even if this fails, or looks up fewer than numids, we have to
+     * check each PMID looking for dynamic metrics and process them
+     * separately with the help of the PMDA, if possible
+     */
     for (i = 0; i < numids; i++) {
-	if (idlist[i] == PM_ID_NULL) continue;
-	if (pmid_domain(idlist[i]) == DYNAMIC_PMID && pmid_item(idlist[i]) == 0) {
-	    int		lsts;
-	    int		domain = pmid_cluster(idlist[i]);
-	    /*
-	     * don't return <domain>.*.* ... all return paths from here
-	     * must either set a valid PMID in idlist[i] or indicate
-	     * the first error in the return from pmLookupName
-	     */
-	    idlist[i] = PM_ID_NULL;	/* default case if cannot translate */
-	    if ((ap = FindDomainAgent(domain)) == NULL) {
-		if (sts > 0) sts = PM_ERR_NOAGENT;
-		continue;
-	    }
-	    if (!ap->status.connected) {
-		if (sts > 0) sts = PM_ERR_NOAGENT;
-		continue;
-	    }
+	if (idlist[i] == PM_ID_NULL || !IS_DYNAMIC_ROOT(idlist[i]))
+	    continue;
+	lsts = 0;
+	domain = pmID_cluster(idlist[i]);
+	/*
+	 * don't return <domain>.*.* ... all return paths from here
+	 * must either set a valid PMID in idlist[i] or indicate
+	 * an error via lsts
+	 */
+	idlist[i] = PM_ID_NULL;	/* default case if cannot translate */
+	if ((ap = FindDomainAgent(domain)) == NULL) {
+	    lsts = PM_ERR_NOAGENT;
+	}
+	else if (!ap->status.connected) {
+	    lsts = PM_ERR_NOAGENT;
+	}
+	else {
 	    if (ap->ipcType == AGENT_DSO) {
 		if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 		    ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
@@ -510,7 +708,7 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		}
 		else {
 		    /* Not PMDA_INTERFACE_4 or later */
-		    lsts = PM_ERR_NAME;
+		    lsts = PM_ERR_NONLEAF;
 		}
 	    }
 	    else {
@@ -519,13 +717,12 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		if (ap->status.notReady)
 		    lsts = PM_ERR_AGAIN;
 		else {
-		    if (_pmcd_trace_mask)
-			pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_NAMES, 1);
+		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_NAMES, 1);
 		    lsts = __pmSendNameList(ap->inFd, cp - client, 1, &namelist[i], NULL);
 		    if (lsts >= 0) {
 			int		pinpdu;
-			pinpdu = lsts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-			if (lsts > 0 && _pmcd_trace_mask)
+			pinpdu = lsts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+			if (lsts > 0)
 			    pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 			if (lsts == PDU_PMNS_IDS) {
 			    int		xsts;
@@ -557,29 +754,33 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		    fdfail != -1)
 		    CleanupAgent(ap, AT_COMM, fdfail);
 	    }
-	    /*
-	     * only set error status to the current error status
-	     * if this is the first error for this set of metrics
-	     */
-	    if (lsts < 0 && sts > 0) sts = lsts;
+	}
+	/*
+	 * only set error status to the current error status
+	 * if this is the first error for this set of metrics,
+	 * and if either it is a fatal error, or numids is 1
+	 */
+	if (lsts < 0 && sts > 0) {
+	    if ((lsts != PM_ERR_NAME && lsts != PM_ERR_NOAGENT &&
+		 lsts != PM_ERR_NONLEAF) || numids == 1)
+	    sts = lsts;
 	}
     }
 
-    if (sts < 0) {
-        /* If get an error which should be passed back along
-         * with valid data to the client
-         * then do NOT fail -> return status with the id-list.
-         */
-    	if (sts != PM_ERR_NAME && sts != PM_ERR_NONLEAF && sts != PM_ERR_NOAGENT)
-	  goto done;
+    if (sts < 0)
+	/* fatal error or explicit error in the numids == 1 case */
+	goto done;
+
+    numok = numids;
+    for (i = 0; i < numids; i++) {
+	if (idlist[i] == PM_ID_NULL)
+	    numok--;
     }
 
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_IDS, numids);
-    if ((sts = __pmSendIDList(cp->fd, FROM_ANON, numids, idlist, sts)) < 0) {
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_IDS, numok);
+    if ((sts = __pmSendIDList(cp->fd, FROM_ANON, numids, idlist, numok)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_IDS, sts);
 	CleanupClient(cp, sts);
-    	goto done;
     }
 
 done:
@@ -609,8 +810,8 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 
     namelist[0] = name;
     sts = pmLookupName(1, namelist, idlist);
-    if (sts == 1 && pmid_domain(idlist[0]) == DYNAMIC_PMID && pmid_item(idlist[0]) == 0) {
-	int		domain = pmid_cluster(idlist[0]);
+    if (sts == 1 && IS_DYNAMIC_ROOT(idlist[0])) {
+	int		domain = pmID_cluster(idlist[0]);
 	AgentInfo	*ap = NULL;
 	if ((ap = FindDomainAgent(domain)) == NULL) {
 	    sts = PM_ERR_NOAGENT;
@@ -645,13 +846,12 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 	    if (ap->status.notReady)
 		sts = PM_ERR_AGAIN;
 	    else {
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_CHILD, 1);
+		pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_CHILD, 1);
 		sts = __pmSendChildReq(ap->inFd, cp - client, name, subtype);
 		if (sts >= 0) {
 		    int		pinpdu;
-		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-		    if (sts > 0 && _pmcd_trace_mask)
+		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+		    if (sts > 0)
 			pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		    if (sts == PDU_PMNS_NAMES) {
 			sts = __pmDecodeNameList(pb, &numnames,
@@ -700,8 +900,7 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
     }
 
     numnames = sts;
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
     if ((sts = __pmSendNameList(cp->fd, FROM_ANON, numnames, offspring, statuslist)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
@@ -725,16 +924,16 @@ static int travNL_i;      /* array index */
 static void
 AddLengths(const char *name)
 {
-  travNL_strlen += strlen(name) + 1;
-  travNL_num++;
+    travNL_strlen += strlen(name) + 1;
+    travNL_num++;
 }
 
 static void
 BuildNameList(const char *name)
 {
-  travNL[travNL_i++] = travNL_ptr;
-  strcpy(travNL_ptr, name);
-  travNL_ptr += strlen(name) + 1;
+    travNL[travNL_i++] = travNL_ptr;
+    strcpy(travNL_ptr, name);
+    travNL_ptr += strlen(name) + 1;
 }
 
 /*
@@ -788,8 +987,8 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 	sts = pmLookupName(1, namelist, idlist);
 	if (sts < 1)
 	    continue;
-	if (pmid_domain(idlist[0]) == DYNAMIC_PMID && pmid_item(idlist[0]) == 0) {
-	    int		domain = pmid_cluster(idlist[0]);
+	if (IS_DYNAMIC_ROOT(idlist[0])) {
+	    int		domain = pmID_cluster(idlist[0]);
 	    AgentInfo	*ap;
 	    if ((ap = FindDomainAgent(domain)) == NULL)
 		continue;
@@ -801,8 +1000,7 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 		if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_4) {
 		    sts = ap->ipc.dso.dispatch.version.four.children(namelist[0], 1, &offspring, &statuslist,
 				      ap->ipc.dso.dispatch.version.four.ext);
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_PMNS) {
+		    if (pmDebugOptions.pmns) {
 			fprintf(stderr, "traverse_dynamic: DSO PMDA: expand dynamic PMNS entry %s (%s) -> ", namelist[0], pmIDStr(idlist[0]));
 			if (sts < 0)
 			    fprintf(stderr, "%s\n", pmErrStr(sts));
@@ -814,7 +1012,6 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 			    }
 			}
 		    }
-#endif
 		    if (sts < 0)
 			continue;
 		    if (statuslist) free(statuslist);
@@ -829,21 +1026,19 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 		int		fdfail = -1;
 		if (ap->status.notReady)
 		    continue;
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_TRAVERSE, 1);
+		pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_TRAVERSE, 1);
 		sts = __pmSendTraversePMNSReq(ap->inFd, cp - client, namelist[0]);
 		if (sts >= 0) {
 		    int		numnames;
 		    __pmPDU	*pb;
 		    int		pinpdu;
-		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-		    if (sts > 0 && _pmcd_trace_mask)
+		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+		    if (sts > 0)
 			pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		    if (sts == PDU_PMNS_NAMES) {
 			sts = __pmDecodeNameList(pb, &numnames,
 						       &offspring, &statuslist);
-#ifdef PCP_DEBUG
-			if (pmDebug & DBG_TRACE_PMNS) {
+			if (pmDebugOptions.pmns) {
 			    fprintf(stderr, "traverse_dynamic: daemon PMDA: expand dynamic PMNS entry %s (%s) -> ", namelist[0], pmIDStr(idlist[0]));
 			    if (sts < 0)
 				fprintf(stderr, "%s\n", pmErrStr(sts));
@@ -855,7 +1050,6 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 				}
 			    }
 			}
-#endif
 			if (statuslist) {
 			    free(statuslist);
 			    statuslist = NULL;
@@ -970,11 +1164,9 @@ DoPMNSTraverse(ClientInfo *cp, __pmPDU *pb)
     travNL_num = 0;
     if ((sts = pmTraversePMNS(name, AddLengths)) < 0)
     	goto check;
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_PMNS) {
+    if (pmDebugOptions.pmns) {
 	fprintf(stderr, "DoPMNSTraverse: %d names below %s after pmTraversePMNS\n", travNL_num, name);
     }
-#endif
 
     /* for each ptr, string bytes, and string terminators */
     travNL_need = travNL_num * (int)sizeof(char*) + travNL_strlen;
@@ -999,10 +1191,8 @@ check:
     if (travNL_num < 1)
 	goto done;
 
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, travNL_num);
-    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, 
-                  travNL_num, travNL, NULL)) < 0) {
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, travNL_num);
+    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, travNL_num, travNL, NULL)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
 	goto done;
@@ -1016,6 +1206,69 @@ done:
 
 /*************************************************************************/
 
+static int
+GetAttribute(ClientInfo *cp, int code)
+{
+    __pmPDU	*pb;
+    char	*value;
+    int		vlen, attr;
+    int		sts, pinpdu;
+
+    /* Expecting an attribute (code) PDU from the client */
+    pinpdu = sts = __pmGetPDU(cp->fd, LIMIT_SIZE, pmcd_timeout, &pb);
+    if (sts > 0)
+	pmcd_trace(TR_RECV_PDU, cp->fd, sts, (int)((__psint_t)pb & 0xffffffff));
+    if (sts == PDU_ATTR) {
+	if ((sts = __pmDecodeAttr(pb, &attr, &value, &vlen)) == 0) {
+	    if (code != attr) {	/* unanticipated attribute */
+		sts = PM_ERR_IPC;
+	    } else if ((value = strndup(value, vlen)) == NULL) {
+		sts = -ENOMEM;
+	    } else {	/* stash the attribute for this client */
+		sts = __pmHashAdd(attr, (void *)value, &cp->attrs);
+	    }
+	}
+    } else if (sts > 0) {	/* unexpected PDU type */
+	sts = PM_ERR_IPC;
+    }
+    if (pinpdu > 0)
+	__pmUnpinPDUBuf(pb);
+    return sts;
+}
+
+static int
+ConnectionAttributes(ClientInfo *cp, int flags)
+{
+    int sts;
+
+    if ((sts = __pmSecureServerHandshake(cp->fd, flags, &cp->attrs)) < 0) {
+	if (pmDebugOptions.auth)
+	    fprintf(stderr, "DoCreds: __pmSecureServerHandshake gave %d: %s\n",
+		    sts, pmErrStr(sts));
+	return sts;
+    }
+
+    if ((flags & PDU_FLAG_CONTAINER) &&
+	(sts = GetAttribute(cp, PCP_ATTR_CONTAINER)) < 0) {
+	if (pmDebugOptions.attr)
+	    fprintf(stderr, "DoCreds: failed GetAttribute container %d: %s\n",
+		    sts, pmErrStr(sts));
+	return sts;
+    }
+
+    return 0;
+}
+
+static int
+CheckCertRequired(ClientInfo *cp)
+{
+    if(__pmServerHasFeature(PM_SERVER_FEATURE_CERT_REQD))
+	if ( !__pmSockAddrIsLoopBack(cp->addr) && !__pmSockAddrIsUnix(cp->addr) )
+	    return 1;
+
+    return 0;
+}
+
 int
 DoCreds(ClientInfo *cp, __pmPDU *pb)
 {
@@ -1025,9 +1278,7 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 
     if ((sts = __pmDecodeCreds(pb, &sender, &credcount, &credlist)) < 0)
 	return sts;
-
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_RECV_PDU, cp->fd, PDU_CREDS, credcount);
+    pmcd_trace(TR_RECV_PDU, cp->fd, PDU_CREDS, credcount);
 
     for (i = 0; i < credcount; i++) {
 	switch(credlist[i].c_type) {
@@ -1035,17 +1286,39 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 		vcp = (__pmVersionCred *)&credlist[i];
 		flags = vcp->c_flags;
 		version = vcp->c_version;
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_CONTEXT)
-		    fprintf(stderr, "pmcd: version cred (%u) flags=%x\n", vcp->c_version, vcp->c_flags);
-#endif
+		if (pmDebugOptions.attr) {
+		    static const struct {
+			int	flag;
+			char	*name;
+		    } flag_dbg[] = {
+			{ PDU_FLAG_SECURE, 	"SECURE" },
+			{ PDU_FLAG_COMPRESS,	"COMPRESS" },
+			{ PDU_FLAG_AUTH,	"AUTH" },
+			{ PDU_FLAG_CREDS_REQD,	"CREDS_REQD" },
+			{ PDU_FLAG_SECURE_ACK,	"SECURE_ACK" },
+			{ PDU_FLAG_NO_NSS_INIT,	"NO_NSS_INIT" },
+			{ PDU_FLAG_CONTAINER,	"CONTAINER" },
+			{ PDU_FLAG_LABEL,	"LABEL" },
+		    };
+		    int	i;
+		    int	first = 1;
+		    fprintf(stderr, "DoCreds: version cred (%u) flags=%x (", vcp->c_version, vcp->c_flags);
+		    for (i = 0; i < sizeof(flag_dbg)/sizeof(flag_dbg[0]); i++) {
+			if (flags & flag_dbg[i].flag) {
+			    if (first)
+				first = 0;
+			    else
+				fputc('|', stderr);
+			    fprintf(stderr, "%s", flag_dbg[i].name);
+			}
+		    }
+		    fprintf(stderr, ")\n");
+		}
 		break;
 
 	    default:
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_CONTEXT)
-		    fprintf(stderr, "pmcd: Error: bogus cred type %d\n", credlist[i].c_type);
-#endif
+		if (pmDebugOptions.auth)
+		    fprintf(stderr, "DoCreds: Error: bogus cred type %d\n", credlist[i].c_type);
 		sts = PM_ERR_IPC;
 		break;
 	}
@@ -1055,12 +1328,44 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 
     if (sts >= 0 && version)
 	sts = __pmSetVersionIPC(cp->fd, version);
+
+    /*
+     * In normal operation, some of this code is redundant. A 
+     * remote client should error out during initial handshake
+     * if it does not support client certs.
+     *
+     * We still need to check local connections and allow those through
+     * in all cases.
+     */
+    if (CheckCertRequired(cp) && (flags & PDU_FLAG_SECURE) == 0)
+	return PM_ERR_NEEDCLIENTCERT;
+
     if (sts >= 0 && flags) {
-	/* new client has arrived; may want encryption, authentication, etc */
-	if ((sts = __pmSecureServerHandshake(cp->fd, flags, &cp->attrs)) < 0)
+	/*
+	 * new client has arrived; may want encryption, authentication, etc
+	 * complete the handshake (depends on features requested), continue
+	 * on to check access is allowed for the authenticated persona, and
+	 * finally notify any interested PMDAs
+	 */
+	if ((sts = ConnectionAttributes(cp, flags)) < 0)
 	    return sts;
-	/* client arrived and may be authenticated, notify interested PMDAs */
-	sts = AgentsAuthentication(cp - client);
     }
+    if ((sts = CheckAccountAccess(cp)) < 0) {	/* host access done already */
+	if (pmDebugOptions.auth)
+	    fprintf(stderr, "DoCreds: CheckAccountAccess returns %d: %s\n",
+		    sts, pmErrStr(sts));
+	return sts;
+    }
+    /*
+     * account authentication successful (if needed) and/or other attributes
+     * have been given - in these cases, we need to inform interested PMDAs.
+     */
+    else if (sts > 0 || (flags & PDU_FLAG_CONTAINER)) {
+	sts = AgentsAttributes(cp - client);
+	if (sts < 0 && (pmDebugOptions.auth || pmDebugOptions.attr))
+	    fprintf(stderr, "DoCreds: AgentsAttributes returns %d: %s\n",
+		    sts, pmErrStr(sts));
+    }
+
     return sts;
 }
