@@ -14,7 +14,7 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "pmcd.h"
 #include <assert.h>
 
@@ -52,7 +52,7 @@ SplitResult(pmResult *res)
 	aFreq = (int *)malloc((nAgents + 1) * sizeof(int));
 	resIndex = (int *)malloc((nAgents + 1) * sizeof(int));
 	if (aFreq == NULL || resIndex == NULL) {
-	    __pmNoMem("SplitResult.freq", 2 * (nAgents + 1) * sizeof(int), PM_FATAL_ERR);
+	    pmNoMem("SplitResult.freq", 2 * (nAgents + 1) * sizeof(int), PM_FATAL_ERR);
 	}
     }
 
@@ -81,7 +81,7 @@ SplitResult(pmResult *res)
     need = nGood + 1 + ((aFreq[nAgents]) ? 1 : 0);
     need *= sizeof(pmResult *);
     if ((results = (pmResult **) malloc(need)) == NULL) {
-	__pmNoMem("SplitResult.results", need, PM_FATAL_ERR);
+	pmNoMem("SplitResult.results", need, PM_FATAL_ERR);
     }
     j = 0;
     for (i = 0; i <= nAgents; i++)
@@ -89,7 +89,7 @@ SplitResult(pmResult *res)
 	    need = (int)sizeof(pmResult) + (aFreq[i] - 1) * (int)sizeof(pmValueSet *);
 	    results[j] = (pmResult *) malloc(need);
 	    if (results[j] == NULL) {
-		__pmNoMem("SplitResult.domain", need, PM_FATAL_ERR);
+		pmNoMem("SplitResult.domain", need, PM_FATAL_ERR);
 	    }
 	    results[j]->numpmid = aFreq[i];
 	    j++;
@@ -97,7 +97,7 @@ SplitResult(pmResult *res)
 
     /* Make the "end of list" pmResult */
     if ((results[j] = (pmResult *) malloc(sizeof(pmResult))) == NULL) {
-	__pmNoMem("SplitResult.domain", sizeof(pmResult), PM_FATAL_ERR);
+	pmNoMem("SplitResult.domain", sizeof(pmResult), PM_FATAL_ERR);
     }
     results[j]->numpmid = 0;
 
@@ -129,7 +129,7 @@ int
 DoStore(ClientInfo *cp, __pmPDU* pb)
 {
     int		sts;
-    int		s;
+    int		s = 0;
     AgentInfo	*ap;
     pmResult	*result;
     pmResult	**dResult;
@@ -166,8 +166,7 @@ DoStore(ClientInfo *cp, __pmPDU* pb)
 	else {
 	    if (ap->status.notReady == 0) {
 		/* agent is ready for PDUs */
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_RESULT, dResult[i]->numpmid);
+		pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_RESULT, dResult[i]->numpmid);
 		s = __pmSendResult(ap->inFd, cp - client, dResult[i]);
 		if (s >= 0) {
 		    ap->status.busy = 1;
@@ -203,13 +202,15 @@ DoStore(ClientInfo *cp, __pmPDU* pb)
     while (nWait > 0) {
         __pmFD_COPY(&readyFds, &waitFds);
 	if (nWait > 1) {
-	    timeout.tv_sec = _pmcd_timeout;
+	    timeout.tv_sec = pmcd_timeout;
 	    timeout.tv_usec = 0;
 
+	    retry:
+	    setoserror(0);
 	    s = __pmSelectRead(maxFd+1, &readyFds, &timeout);
 
 	    if (s == 0) {
-		__pmNotifyErr(LOG_INFO, "DoStore: select timeout");
+		pmNotifyErr(LOG_INFO, "DoStore: select timeout");
 
 		/* Timeout, terminate agents that haven't responded */
 		for (i = 0; i < nAgents; i++) {
@@ -221,9 +222,11 @@ DoStore(ClientInfo *cp, __pmPDU* pb)
 		sts = PM_ERR_IPC;
 		break;
 	    }
-	    else if (sts < 0) {
+	    else if (s < 0) {
+		if (neterror() == EINTR)
+		    goto retry;
 		/* this is not expected to happen! */
-		__pmNotifyErr(LOG_ERR, "DoStore: fatal select failure: %s\n",
+		pmNotifyErr(LOG_ERR, "DoStore: fatal select failure: %s\n",
 			netstrerror());
 		Shutdown();
 		exit(1);
@@ -238,8 +241,8 @@ DoStore(ClientInfo *cp, __pmPDU* pb)
 	    ap->status.busy = 0;
 	    __pmFD_CLR(ap->outFd, &waitFds);
 	    nWait--;
-	    pinpdu = s = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-	    if (s > 0 && _pmcd_trace_mask)
+	    pinpdu = s = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+	    if (s > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, s, (int)((__psint_t)pb & 0xffffffff));
 	    if (s == PDU_ERROR) {
 		int ss;
@@ -294,8 +297,7 @@ DoStore(ClientInfo *cp, __pmPDU* pb)
     if (sts >= 0) {
 	/* send PDU_ERROR, even if result was 0 */
 	int s;
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, 0);
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, 0);
 	s = __pmSendError(cp->fd, FROM_ANON, 0);
 	if (s < 0)
 	    CleanupClient(cp, s);
