@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2013 Red Hat.
+ * Copyright (c) 2012-2017 Red Hat.
  * Copyright (c) 1995-2001,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -14,21 +14,23 @@
  */
 
 #include "pmcd.h"
-#include "impl.h"
+#include "libpcp.h"
 #include <sys/stat.h>
 #include <assert.h>
 
-#define SHUTDOWNWAIT	12	/* < PMDAs wait previously used in rc_pcp */
+#define PMDAROOT	1	/* domain identifier for pmdaroot(1) */
+#define SHUTDOWNWAIT	15	/* PMDAs wait time, in 10msec increments */
 #define MAXPENDING	5	/* maximum number of pending connections */
-#define FDNAMELEN	40	/* maximum length of a fd description */
+#define FDNAMELEN	80	/* maximum length of a fd description */
 #define STRINGIFY(s)	#s
 #define TO_STRING(s)	STRINGIFY(s)
 
-#ifdef PCP_DEBUG
 static char	*FdToString(int);
-#endif
+static void	ResetBadHosts(void);
 
 int		AgentDied;		/* for updating mapdom[] */
+int		AgentPendingRestart;	/* for automatic restart */
+int		labelChanged;		/* For SIGHUP labels check */
 static int	timeToDie;		/* For SIGINT handling */
 static int	restart;		/* For SIGHUP restart */
 static int	maxReqPortFd;		/* Largest request port fd */
@@ -41,7 +43,9 @@ static char	*pmnsfile = PM_NS_DEFAULT;
 static char	*username;
 static char	*certdb;		/* certificate database path (NSS) */
 static char	*dbpassfile;		/* certificate database password file */
-static int	dupok;			/* set to 1 for -N pmnsfile */
+static char	*cert_nickname;		/* Alternate nickname to use for server certificate */
+static int	dupok = 1;		/* set to 0 for -N pmnsfile */
+static char	sockpath[MAXPATHLEN];	/* local unix domain socket path */
 
 #ifdef HAVE_SA_SIGINFO
 static pid_t	killer_pid;
@@ -55,7 +59,7 @@ DontStart(void)
     FILE	*tty;
     FILE	*log;
 
-    __pmNotifyErr(LOG_ERR, "pmcd not started due to errors!\n");
+    pmNotifyErr(LOG_ERR, "pmcd not started due to errors!\n");
 
     if ((tty = fopen(fatalfile, "w")) != NULL) {
 	fflush(stderr);
@@ -71,63 +75,88 @@ DontStart(void)
 	    fprintf(tty, "Log file \"%s\" has vanished!\n", logfile);
 	fclose(tty);
     }
+    /*
+     * We are often called after the request ports have been opened. If we don't
+     * explicitely close them, then the unix domain socket file (if any) will be
+     * left in the file system, causing "address already in use" the next time
+     * pmcd starts.
+     */
+    __pmServerCloseRequestPorts();
+
     exit(1);
 }
 
-static int
-CreatePIDfile(void)
-{
-    char	pidpath[MAXPATHLEN];
-    FILE	*pidfile;
+static pmLongOptions longopts[] = {
+    PMAPI_OPTIONS_HEADER("General options"),
+    PMOPT_DEBUG,
+    PMOPT_NAMESPACE,
+    PMOPT_UNIQNAMES,
+    PMOPT_HELP,
+    PMAPI_OPTIONS_HEADER("Service options"),
+    { "", 0, 'A', 0, "disable service advertisement" },
+    { "foreground", 0, 'f', 0, "run in the foreground" },
+    { "hostname", 1, 'H', "HOST", "set the hostname to be used for pmcd.hostname metric" },
+    { "username", 1, 'U', "USER", "in daemon mode, run as named user [default pcp]" },
+    PMAPI_OPTIONS_HEADER("Configuration options"),
+    { "config", 1, 'c', "PATH", "path to configuration file" },
+    { "certdb", 1, 'C', "PATH", "path to NSS certificate database" },
+    { "passfile", 1, 'P', "PATH", "password file for certificate database access" },
+    { "", 1, 'L', "BYTES", "maximum size for PDUs from clients [default 65536]" },
+    { "", 1, 'q', "TIME", "PMDA initial negotiation timeout (seconds) [default 3]" },
+    { "", 1, 't', "TIME", "PMDA response timeout (seconds) [default 5]" },
+    { "verify", 0, 'v', 0, "check validity of pmcd configuration, then exit" },
+    PMAPI_OPTIONS_HEADER("Connection options"),
+    { "interface", 1, 'i', "ADDR", "accept connections on this IP address" },
+    { "port", 1, 'p', "N", "accept connections on this port" },
+    { "socket", 1, 's', "PATH", "Unix domain socket file [default $PCP_RUN_DIR/pmcd.socket]" },
+    PMAPI_OPTIONS_HEADER("Diagnostic options"),
+    { "trace", 1, 'T', "FLAG", "Event trace control" },
+    { "log", 1, 'l', "PATH", "redirect diagnostics and trace output" },
+    { "", 1, 'x', "PATH", "fatal messages at startup sent to file [default /dev/tty]" },
+    PMAPI_OPTIONS_END
+};
 
-    snprintf(pidpath, sizeof(pidpath), "%s%c" "pmcd.pid",
-		pmGetConfig("PCP_RUN_DIR"), __pmPathSeparator());
-    if ((pidfile = fopen(pidpath, "w")) == NULL) {
-	fprintf(stderr, "Error: Cant open PID file %s\n", pidpath);
-	return -1;
-    }
-    fprintf(pidfile, "%" FMT_PID, getpid());
-    fflush(pidfile);
-    fclose(pidfile);
-    return 0;
-}
+static pmOptions opts = {
+    .flags = PM_OPTFLAG_POSIX,
+    .short_options = "Ac:C:D:fH:i:l:L:M:N:n:p:P:q:Qs:St:T:U:vx:?",
+    .long_options = longopts,
+};
 
 static void
 ParseOptions(int argc, char *argv[], int *nports)
 {
     int		c;
     int		sts;
-    int		errflag = 0;
-    int		usage = 0;
     char	*endptr;
+    int		verify = 0;
+    int		usage = 0;
     int		val;
 
-    strcpy(configFileName, pmGetConfig("PCP_PMCDCONF_PATH"));
+    endptr = pmGetConfig("PCP_PMCDCONF_PATH");
+    strncpy(configFileName, endptr, sizeof(configFileName)-1);
 
-#ifdef HAVE_GETOPT_NEEDS_POSIXLY_CORRECT
-    /*
-     * pmcd does not really need this for its own options because the
-     * arguments like "arg -x" are not valid.  But the PMDA's launched
-     * by pmcd from pmcd.conf may not be so lucky.
-     */
-    putenv("POSIXLY_CORRECT=");
-#endif
-
-    while ((c = getopt(argc, argv, "C:D:fi:l:L:N:n:p:P:q:t:T:U:x:?")) != EOF)
+    while ((c = pmgetopt_r(argc, argv, &opts)) != EOF) {
 	switch (c) {
 
-	    case 'C':	/* path to NSS certificate database */
-		certdb = optarg;
+	    case 'A':	/* disable pmcd service advertising */
+		__pmServerClearFeature(PM_SERVER_FEATURE_DISCOVERY);
 		break;
 
-	    case 'D':	/* debug flag */
-		sts = __pmParseDebug(optarg);
+	    case 'c':	/* configuration file */
+		strncpy(configFileName, opts.optarg, sizeof(configFileName)-1);
+		break;
+
+	    case 'C':	/* path to NSS certificate database */
+		certdb = opts.optarg;
+		break;
+
+	    case 'D':	/* debug options */
+		sts = pmSetDebug(opts.optarg);
 		if (sts < 0) {
-		    fprintf(stderr, "%s: unrecognized debug flag specification (%s)\n",
-			pmProgname, optarg);
-		    errflag++;
+		    pmprintf("%s: unrecognized debug options specification (%s)\n",
+			pmGetProgname(), opts.optarg);
+		    opts.errors++;
 		}
-		pmDebug |= sts;
 		break;
 
 	    case 'f':
@@ -137,86 +166,110 @@ ParseOptions(int argc, char *argv[], int *nports)
 
 	    case 'i':
 		/* one (of possibly several) interfaces for client requests */
-		__pmServerAddInterface(optarg);
+		__pmServerAddInterface(opts.optarg);
+		break;
+
+	    case 'H':
+		/* use the given name as the pmcd.hostname for this host */
+		pmcd_hostname = opts.optarg;
 		break;
 
 	    case 'l':
 		/* log file name */
-		logfile = optarg;
+		logfile = opts.optarg;
 		break;
 
 	    case 'L': /* Maximum size for PDUs from clients */
-		val = (int)strtol (optarg, NULL, 0);
-		if ( val <= 0 ) {
-		    fputs("pmcd: -L requires a positive value\n", stderr);
-		    errflag++;
+		val = (int)strtol(opts.optarg, NULL, 0);
+		if (val <= 0) {
+		    pmprintf("%s: -L requires a positive value\n", pmGetProgname());
+		    opts.errors++;
 		} else {
-		    __pmSetPDUCeiling (val);
+		    __pmSetPDUCeiling(val);
 		}
 		break;
 
+	    case 'M':	/* nickname for the server cert. Use to query the nssdb */
+		cert_nickname = opts.optarg;
+		break;
+
 	    case 'N':
-		dupok = 1;
+		dupok = 0;
 		/*FALLTHROUGH*/
 	    case 'n':
 	    	/* name space file name */
-		pmnsfile = optarg;
+		pmnsfile = opts.optarg;
 		break;
 
 	    case 'p':
-		if (__pmServerAddPorts(optarg) < 0) {
-		    fprintf(stderr,
-			"pmcd: -p requires a positive numeric argument (%s)\n",
-			optarg);
-		    errflag++;
+		if (__pmServerAddPorts(opts.optarg) < 0) {
+		    pmprintf("%s: -p requires a positive numeric argument (%s)\n",
+			pmGetProgname(), opts.optarg);
+		    opts.errors++;
 		} else {
 		    *nports += 1;
 		}
 		break;
 		    
 	    case 'P':	/* password file for certificate database access */
-		dbpassfile = optarg;
+		dbpassfile = opts.optarg;
 		break;
 
 	    case 'q':
-		val = (int)strtol(optarg, &endptr, 10);
+		val = (int)strtol(opts.optarg, &endptr, 10);
 		if (*endptr != '\0' || val <= 0.0) {
-		    fprintf(stderr,
-			    "pmcd: -q requires a positive numeric argument\n");
-		    errflag++;
-		}
-		else
+		    pmprintf("%s: -q requires a positive numeric argument\n",
+			pmGetProgname());
+		    opts.errors++;
+		} else {
 		    _creds_timeout = val;
+		}
+		break;
+
+	    case 'Q':	/* require clients to provide a trusted cert */
+		__pmServerSetFeature(PM_SERVER_FEATURE_CERT_REQD);
+		break;
+
+	    case 's':	/* path to local unix domain socket */
+		pmsprintf(sockpath, sizeof(sockpath), "%s", opts.optarg);
+		break;
+
+	    case 'S':	/* only allow authenticated clients */
+		__pmServerSetFeature(PM_SERVER_FEATURE_CREDS_REQD);
 		break;
 
 	    case 't':
-		val = (int)strtol(optarg, &endptr, 10);
+		val = (int)strtol(opts.optarg, &endptr, 10);
 		if (*endptr != '\0' || val < 0.0) {
-		    fprintf(stderr,
-			    "pmcd: -t requires a positive numeric argument\n");
-		    errflag++;
+		    pmprintf("%s: -t requires a positive numeric argument\n",
+			pmGetProgname());
+		    opts.errors++;
+		} else {
+		    pmcd_timeout = val;
 		}
-		else
-		    _pmcd_timeout = val;
 		break;
 
 	    case 'T':
-		val = (int)strtol(optarg, &endptr, 10);
+		val = (int)strtol(opts.optarg, &endptr, 10);
 		if (*endptr != '\0' || val < 0) {
-		    fprintf(stderr,
-			    "pmcd: -T requires a positive numeric argument\n");
-		    errflag++;
+		    pmprintf("%s: -T requires a positive numeric argument\n",
+			pmGetProgname());
+		    opts.errors++;
+		} else {
+		    pmcd_trace_mask = val;
 		}
-		else
-		    _pmcd_trace_mask = val;
 		break;
 
 	    case 'U':
-		username = optarg;
+		username = opts.optarg;
+		break;
+
+	    case 'v':
+		verify = 1;
 		break;
 
 	    case 'x':
-		fatalfile = optarg;
+		fatalfile = opts.optarg;
 		break;
 
 	    case '?':
@@ -224,33 +277,21 @@ ParseOptions(int argc, char *argv[], int *nports)
 		break;
 
 	    default:
-		errflag++;
+		opts.errors++;
 		break;
 	}
+    }
 
-    if (usage ||errflag || optind < argc) {
-	fprintf(stderr,
-"Usage: %s [options]\n\n"
-"Options:\n"
-"  -C dirname      path to NSS certificate database\n"
-"  -f              run in the foreground\n" 
-"  -i ipaddress    accept connections on this IP address\n"
-"  -l logfile      redirect diagnostics and trace output\n"
-"  -L bytes        maximum size for PDUs from clients [default 65536]\n"
-"  -n pmnsfile     use an alternative PMNS\n"
-"  -N pmnsfile     use an alternative PMNS (duplicate PMIDs are allowed)\n"
-"  -p port         accept connections on this port\n"
-"  -P passfile     password file for certificate database access\n"
-"  -q timeout      PMDA initial negotiation timeout (seconds) [default 3]\n"
-"  -T traceflag    Event trace control\n"
-"  -t timeout      PMDA response timeout (seconds) [default 5]\n"
-"  -U username     in daemon mode, run as named user [default pcp]\n"
-"  -x file         fatal messages at startup sent to file [default /dev/tty]\n",
-			pmProgname);
+    if (usage || opts.errors || opts.optind < argc) {
+	pmUsageMessage(&opts);
 	if (usage)
 	    exit(0);
-	else
-	    DontStart();
+	DontStart();
+    }
+
+    if (verify) {
+	sts = VerifyConfig(configFileName);
+	exit(sts < 0);
     }
 }
 
@@ -269,16 +310,17 @@ HandleClientInput(__pmFdSet *fdsPtr)
 
     for (i = 0; i < nClients; i++) {
 	int		pinpdu;
+
 	if (!client[i].status.connected || !__pmFD_ISSET(client[i].fd, fdsPtr))
 	    continue;
 
 	cp = &client[i];
 	this_client_id = i;
 
-	pinpdu = sts = __pmGetPDU(cp->fd, LIMIT_SIZE, _pmcd_timeout, &pb);
-	if (sts > 0 && _pmcd_trace_mask)
+	pinpdu = sts = __pmGetPDU(cp->fd, LIMIT_SIZE, pmcd_timeout, &pb);
+	if (sts > 0) {
 	    pmcd_trace(TR_RECV_PDU, cp->fd, sts, (int)((__psint_t)pb & 0xffffffff));
-	if (sts <= 0) {
+	} else {
 	    CleanupClient(cp, sts);
 	    continue;
 	}
@@ -292,14 +334,13 @@ HandleClientInput(__pmFdSet *fdsPtr)
 	    continue;
 	}
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
+	if (pmDebugOptions.appl0)
 	    ShowClients(stderr);
-#endif
 
 	switch (php->type) {
 	    case PDU_PROFILE:
-		sts = DoProfile(cp, pb);
+		sts = (cp->denyOps & PMCD_OP_FETCH) ?
+		      PM_ERR_PERMISSION : DoProfile(cp, pb);
 		break;
 
 	    case PDU_FETCH:
@@ -312,6 +353,11 @@ HandleClientInput(__pmFdSet *fdsPtr)
 		      PM_ERR_PERMISSION : DoInstance(cp, pb);
 		break;
 
+	    case PDU_LABEL_REQ:
+		sts = (cp->denyOps & PMCD_OP_FETCH) ?
+		      PM_ERR_PERMISSION : DoLabel(cp, pb);
+		break;
+
 	    case PDU_DESC_REQ:
 		sts = (cp->denyOps & PMCD_OP_FETCH) ?
 		      PM_ERR_PERMISSION : DoDesc(cp, pb);
@@ -355,73 +401,118 @@ HandleClientInput(__pmFdSet *fdsPtr)
 		sts = PM_ERR_IPC;
 	}
 	if (sts < 0) {
-
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL0)
+	    if (pmDebugOptions.appl0)
 		fprintf(stderr, "PDU:  %s client[%d]: %s\n",
 		    __pmPDUTypeStr(php->type), i, pmErrStr(sts));
-#endif
 	    /* Make sure client still alive before sending. */
 	    if (cp->status.connected) {
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
+		pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
 		sts = __pmSendError(cp->fd, FROM_ANON, sts);
 		if (sts < 0)
-		    __pmNotifyErr(LOG_ERR, "HandleClientInput: "
+		    pmNotifyErr(LOG_ERR, "HandleClientInput: "
 			"error sending Error PDU to client[%d] %s\n", i, pmErrStr(sts));
 	    }
 	}
 	if (pinpdu > 0)
 	    __pmUnpinPDUBuf(pb);
+
+	/*
+	 * May need to send connection attributes to interested PMDAs, if
+	 * something changed for this client during this PDU exchange.
+	 */
+	if (client[i].status.attributes) {
+	    if (pmDebugOptions.appl1)
+		pmNotifyErr(LOG_INFO, "Client idx=%d,seq=%d attrs reset\n",
+				i, client[i].seq);
+	    AgentsAttributes(i);
+	}
     }
 }
 
-/* Called to shutdown pmcd in an orderly manner */
+/*
+ * Shutdown (and ShutdownAgent helper) shut pmcd down in an orderly manner
+ */
+
+static void
+ShutdownAgent(AgentInfo *ap)
+{
+    if (!ap->status.connected)
+	return;
+    if (ap->inFd != -1) {
+	if (__pmSocketIPC(ap->inFd))
+	    __pmCloseSocket(ap->inFd);
+	else
+	    close(ap->inFd);
+    }
+    if (ap->outFd != -1) {
+	if (__pmSocketIPC(ap->outFd))
+	    __pmCloseSocket(ap->outFd);
+	else
+	    close(ap->outFd);
+    }
+    if (ap->ipcType == AGENT_SOCKET &&
+	ap->ipc.socket.addrDomain == AF_UNIX) {
+	/* remove the Unix domain socket */
+	unlink(ap->ipc.socket.name);
+    }
+}
+
+void
+TerminateAgent(AgentInfo *ap)
+{
+    pid_t pid;
+
+    if (!ap->status.connected)
+	return;
+    pid = (ap->ipcType == AGENT_SOCKET) ?
+	   ap->ipc.socket.agentPid : ap->ipc.pipe.agentPid;
+    if (ap->status.isRootChild && pmdarootfd > 0)
+	pmdaRootProcessTerminate(pmdarootfd, pid);
+    else
+	__pmProcessTerminate(pid, 1);
+}
 
 void
 Shutdown(void)
 {
-    int	i;
+    AgentInfo	*ap, *root = NULL;
+    int		i;
 
     for (i = 0; i < nAgents; i++) {
-	AgentInfo *ap = &agent[i];
-	if (!ap->status.connected)
+	ap = &agent[i];
+	if (ap->pmDomainId == PMDAROOT)
+	    root = ap;
+	if (ap->status.isRootChild)
+	    ShutdownAgent(ap);
+    }
+    for (i = 0; i < nAgents; i++) {
+	ap = &agent[i];
+	if (ap->pmDomainId == PMDAROOT)
 	    continue;
-	if (ap->inFd != -1) {
-	    if (__pmSocketIPC(ap->inFd))
-		__pmCloseSocket(ap->inFd);
-	    else
-		close(ap->inFd);
-	}
-	if (ap->outFd != -1) {
-	    if (__pmSocketIPC(ap->outFd))
-		__pmCloseSocket(ap->outFd);
-	    else
-		close(ap->outFd);
-	}
-	if (ap->ipcType == AGENT_SOCKET &&
-	    ap->ipc.socket.addrDomain == AF_UNIX) {
-	    /* remove the Unix domain socket */
-	    unlink(ap->ipc.socket.name);
-	}
+	if (!ap->status.isRootChild)
+	    ShutdownAgent(ap);
     }
     if (HarvestAgents(SHUTDOWNWAIT) < 0) {
-	/* terminate with prejudice any still remaining */
+	/* terminate with prejudice any still remaining non-root PMDAs */
 	for (i = 0; i < nAgents; i++) {
-	    AgentInfo *ap = &agent[i];
-	    if (ap->status.connected) {
-		pid_t pid = ap->ipcType == AGENT_SOCKET ?
-			    ap->ipc.socket.agentPid : ap->ipc.pipe.agentPid;
-		__pmProcessTerminate(pid, 1);
-	    }
+	    ap = &agent[i];
+	    if (ap == root)
+		continue;
+	    TerminateAgent(ap);
 	}
     }
-    for (i = 0; i < nClients; i++)
+    if (root) {
+	TerminateAgent(root);
+	ShutdownAgent(root);
+	HarvestAgents(0);
+    }
+    for (i = 0; i < nClients; i++) {
 	if (client[i].status.connected)
 	    __pmCloseSocket(client[i].fd);
+    }
     __pmServerCloseRequestPorts();
     __pmSecureServerShutdown();
-    __pmNotifyErr(LOG_INFO, "pmcd Shutdown\n");
+    pmNotifyErr(LOG_INFO, "pmcd Shutdown\n");
     fflush(stderr);
 }
 
@@ -433,40 +524,48 @@ SignalShutdown(void)
     char	buf[256];
 #endif
     if (killer_pid != 0) {
-	__pmNotifyErr(LOG_INFO, "pmcd caught %s from pid=%" FMT_PID " uid=%d\n",
+	pmNotifyErr(LOG_INFO, "pmcd caught %s from pid=%" FMT_PID " uid=%d\n",
 	    killer_sig == SIGINT ? "SIGINT" : "SIGTERM", killer_pid, killer_uid);
 #if DESPERATE
-	__pmNotifyErr(LOG_INFO, "Try to find process in ps output ...\n");
-	sprintf(buf, "sh -c \". \\$PCP_DIR/etc/pcp.env; ( \\$PCP_PS_PROG \\$PCP_PS_ALL_FLAGS | \\$PCP_AWK_PROG 'NR==1 {print} \\$2==%" FMT_PID " {print}' )\"", killer_pid);
+	pmNotifyErr(LOG_INFO, "Try to find process in ps output ...\n");
+	pmsprintf(buf, sizeof(buf), "sh -c \". \\$PCP_DIR/etc/pcp.env; ( \\$PCP_PS_PROG \\$PCP_PS_ALL_FLAGS | \\$PCP_AWK_PROG 'NR==1 {print} \\$2==%" FMT_PID " {print}' )\"", killer_pid);
 	system(buf);
 #endif
     }
     else {
-	__pmNotifyErr(LOG_INFO, "pmcd caught %s from unknown process\n",
+	pmNotifyErr(LOG_INFO, "pmcd caught %s from unknown process\n",
 			killer_sig == SIGINT ? "SIGINT" : "SIGTERM");
     }
 #else
-    __pmNotifyErr(LOG_INFO, "pmcd caught %s\n",
+    pmNotifyErr(LOG_INFO, "pmcd caught %s\n",
 		    killer_sig == SIGINT ? "SIGINT" : "SIGTERM");
 #endif
     Shutdown();
     exit(0);
 }
 
-void
+static void
 SignalRestart(void)
 {
     time_t	now;
 
     time(&now);
-    __pmNotifyErr(LOG_INFO, "\n\npmcd RESTARTED at %s", ctime(&now));
+    pmNotifyErr(LOG_INFO, "\n\npmcd RESTARTED at %s", ctime(&now));
     fprintf(stderr, "\nCurrent PMCD clients ...\n");
     ShowClients(stderr);
     ResetBadHosts();
+    CheckLabelChange();
     ParseRestartAgents(configFileName);
 }
 
-void
+static void
+SignalReloadLabels(void)
+{
+    /* Inform clients there's been a change in context label state */
+    MarkStateChanges(PMCD_LABEL_CHANGE);
+}
+
+static void
 SignalReloadPMNS(void)
 {
     int sts;
@@ -480,21 +579,17 @@ SignalReloadPMNS(void)
      * This caveat was allowed to make the code a lot simpler. 
      */
     if (__pmHasPMNSFileChanged(pmnsfile)) {
-	__pmNotifyErr(LOG_INFO, "Reloading PMNS \"%s\"",
+	pmNotifyErr(LOG_INFO, "Reloading PMNS \"%s\"",
 	   (pmnsfile==PM_NS_DEFAULT)?"DEFAULT":pmnsfile);
 	pmUnloadNameSpace();
-	if (dupok)
-	    sts = pmLoadASCIINameSpace(pmnsfile, 1);
-	else
-	    sts = pmLoadNameSpace(pmnsfile);
+	sts = pmLoadASCIINameSpace(pmnsfile, dupok);
 	if (sts < 0) {
-	    __pmNotifyErr(LOG_ERR, "PMNS \"%s\" load failed: %s",
-		(pmnsfile == PM_NS_DEFAULT) ? "DEFAULT" : pmnsfile,
-		pmErrStr(sts));
+	    pmNotifyErr(LOG_ERR, "pmLoadASCIINameSpace(%s, %d): %s\n",
+		(pmnsfile == PM_NS_DEFAULT) ? "DEFAULT" : pmnsfile, dupok, pmErrStr(sts));
 	}
     }
     else {
-	__pmNotifyErr(LOG_INFO, "PMNS file \"%s\" is unchanged",
+	pmNotifyErr(LOG_INFO, "PMNS file \"%s\" is unchanged",
 		(pmnsfile == PM_NS_DEFAULT) ? "DEFAULT" : pmnsfile);
     }
 }
@@ -521,8 +616,8 @@ HandleReadyAgents(__pmFdSet *readyFds)
 
 		/* Expect an error PDU containing PM_ERR_PMDAREADY */
 		reason = AT_COMM;	/* most errors are protocol failures */
-		pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-		if (sts > 0 && _pmcd_trace_mask)
+		pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+		if (sts > 0)
 		    pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		if (sts == PDU_ERROR) {
 		    s = __pmDecodeError(pb, &sts);
@@ -532,14 +627,12 @@ HandleReadyAgents(__pmFdSet *readyFds)
 		    }
 		    else {
 			/* sts is the status code from the error PDU */
-#ifdef PCP_DEBUG
-			if (pmDebug && DBG_TRACE_APPL0)
-			    __pmNotifyErr(LOG_INFO,
+			if (pmDebugOptions.appl0)
+			    pmNotifyErr(LOG_INFO,
 				 "%s agent (not ready) sent %s status(%d)\n",
 				 ap->pmDomainLabel,
 				 sts == PM_ERR_PMDAREADY ?
 					     "ready" : "unknown", sts);
-#endif
 			if (sts == PM_ERR_PMDAREADY) {
 			    ap->status.notReady = 0;
 			    sts = 1;
@@ -556,14 +649,13 @@ HandleReadyAgents(__pmFdSet *readyFds)
 			pmcd_trace(TR_RECV_ERR, ap->outFd, PDU_RESULT, sts);
 		    else
 			pmcd_trace(TR_WRONG_PDU, ap->outFd, PDU_ERROR, sts);
-		    sts = PM_ERR_IPC; /* Wrong PDU type */
+ 		    sts = PM_ERR_IPC; /* Wrong PDU type */
 		}
 		if (pinpdu > 0)
 		    __pmUnpinPDUBuf(pb);
 
 		if (ap->ipcType != AGENT_DSO && sts <= 0)
 		    CleanupAgent(ap, reason, fd);
-
 	    }
 	}
     }
@@ -571,49 +663,68 @@ HandleReadyAgents(__pmFdSet *readyFds)
 }
 
 static void
-CheckNewClient(__pmFdSet * fdset, int rfd)
+CheckNewClient(__pmFdSet * fdset, int rfd, int family)
 {
-    int		s, sts, challenge, accepted = 1;
+    int		s, sts, accepted = 1;
+    __uint32_t	challenge;
     ClientInfo	*cp;
-    __pmPDUInfo	xchallenge;
 
     if (__pmFD_ISSET(rfd, fdset)) {
 	if ((cp = AcceptNewClient(rfd)) == NULL)
 	    return;	/* Accept failed and no client added */
 
 	sts = __pmAccAddClient(cp->addr, &cp->denyOps);
+#if defined(HAVE_STRUCT_SOCKADDR_UN)
+	if (sts >= 0 && family == AF_UNIX) {
+	    if ((sts = __pmServerSetLocalCreds(cp->fd, &cp->attrs)) < 0) {
+		pmNotifyErr(LOG_ERR,
+			"ClientLoop: error extracting local credentials: %s",
+			pmErrStr(sts));
+	    }
+	}
+#endif
 	if (sts >= 0) {
 	    memset(&cp->pduInfo, 0, sizeof(cp->pduInfo));
 	    cp->pduInfo.version = PDU_VERSION;
 	    cp->pduInfo.licensed = 1;
+	    cp->pduInfo.features = PDU_FLAG_LABEL;
 	    if (__pmServerHasFeature(PM_SERVER_FEATURE_SECURE))
-		cp->pduInfo.features |= PDU_FLAG_SECURE;
+		cp->pduInfo.features |= (PDU_FLAG_SECURE | PDU_FLAG_SECURE_ACK);
 	    if (__pmServerHasFeature(PM_SERVER_FEATURE_COMPRESS))
 		cp->pduInfo.features |= PDU_FLAG_COMPRESS;
-	    if (__pmServerHasFeature(PM_SERVER_FEATURE_AUTH))
+	    if (__pmServerHasFeature(PM_SERVER_FEATURE_AUTH))       /*optional*/
 		cp->pduInfo.features |= PDU_FLAG_AUTH;
-	    challenge = *(int*)(&cp->pduInfo);
+            if (__pmServerHasFeature(PM_SERVER_FEATURE_CERT_REQD))  /* Required for remote connections only */
+		cp->pduInfo.features |= PDU_FLAG_CERT_REQD;	    /* Enforced in connect.c:check_feature_flags */
+	    if (__pmServerHasFeature(PM_SERVER_FEATURE_CREDS_REQD)) /*required*/
+		cp->pduInfo.features |= PDU_FLAG_CREDS_REQD;
+	    if (__pmServerHasFeature(PM_SERVER_FEATURE_CONTAINERS))
+		cp->pduInfo.features |= PDU_FLAG_CONTAINER;
+	    challenge = *(__uint32_t *)(&cp->pduInfo);
 	    sts = 0;
 	}
 	else {
-	    /* __pmAccAddClient failed, this is grim! */
 	    challenge = 0;
 	    accepted = 0;
 	}
 
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
-	xchallenge = *(__pmPDUInfo *)&challenge;
-	xchallenge = __htonpmPDUInfo(xchallenge);
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
 
 	/* reset (no meaning, use fd table to version) */
 	cp->pduInfo.version = UNKNOWN_VERSION;
 
-	s = __pmSendXtendError(cp->fd, FROM_ANON, sts, *(unsigned int *)&xchallenge);
+	s = __pmSendXtendError(cp->fd, FROM_ANON, sts, htonl(challenge));
 	if (s < 0) {
-	    __pmNotifyErr(LOG_ERR,
-		"ClientLoop: error sending Conn ACK PDU to new client %s\n",
-		pmErrStr(s));
+	    /*
+	     * Port-probe style connections frequently drop just before
+	     * reaching here, as this is the first PDU we send.  Rather
+	     * than being chatty in pmcd.log write this diagnostic only
+	     * under debugging conditions.
+	     */
+	    if (pmDebugOptions.appl0)
+		pmNotifyErr(LOG_INFO, "ClientLoop: "
+			"error sending Conn ACK PDU to new client %s\n",
+			pmErrStr(s));
 	    if (sts >= 0)
 	        /*
 		 * prefer earlier failure status if any, else
@@ -635,7 +746,8 @@ ClientLoop(void)
     int		i, fd, sts;
     int		maxFd;
     int		checkAgents;
-    int		reload_ns = 0;
+    int		reload_namespace = 0;
+    int		restartAgents = -1;	/* initial state unknown */
     __pmFdSet	readableFds;
 
     for (;;) {
@@ -659,41 +771,66 @@ ClientLoop(void)
 		if (fd > maxFd)
 		    maxFd = fd + 1;
 		checkAgents = 1;
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_APPL0)
-		    __pmNotifyErr(LOG_INFO,
+		if (pmDebugOptions.appl0)
+		    pmNotifyErr(LOG_INFO,
 				 "not ready: check %s agent on fd %d (max = %d)\n",
 				 ap->pmDomainLabel, fd, maxFd);
-#endif
 	    }
 	}
 
 	sts = __pmSelectRead(maxFd, &readableFds, NULL);
 	if (sts > 0) {
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL0)
+	    if (pmDebugOptions.appl0)
 		for (i = 0; i <= maxClientFd; i++)
 		    if (__pmFD_ISSET(i, &readableFds))
-			fprintf(stderr, "DATA: from %s (fd %d)\n", FdToString(i), i);
-#endif
+			fprintf(stderr, "DATA: from %s (fd %d)\n",
+				FdToString(i), i);
 	    __pmServerAddNewClients(&readableFds, CheckNewClient);
 	    if (checkAgents)
-		reload_ns = HandleReadyAgents(&readableFds);
+		reload_namespace = HandleReadyAgents(&readableFds);
 	    HandleClientInput(&readableFds);
 	}
 	else if (sts == -1 && neterror() != EINTR) {
-	    __pmNotifyErr(LOG_ERR, "ClientLoop select: %s\n", netstrerror());
+	    pmNotifyErr(LOG_ERR, "ClientLoop select: %s\n", netstrerror());
 	    break;
 	}
+	if (AgentDied) {
+	    if (restartAgents == -1) {
+		char *args;
+
+		if ((args = getenv("PMCD_RESTART_AGENTS")) == NULL)
+		    restartAgents = 1;	/* unset, default to enabled */
+		else {
+		    restartAgents = (strcmp(args, "0") != 0);
+		    fprintf(stderr, "Warning: restartAgents=%d from PMCD_RESTART_AGENTS=%s in environment\n", restartAgents, args);
+		}
+	    }
+	    AgentPendingRestart = restartAgents;
+	}
+	if (AgentPendingRestart) {
+	    static time_t last_restart;
+	    time_t now = time(NULL);
+
+	    if ((now - last_restart) >= 60) {
+		AgentPendingRestart = 0;
+		last_restart = now;
+		pmNotifyErr(LOG_INFO, "Auto-restarting agents.\n");
+		restart = 1;
+	    }
+	}
 	if (restart) {
 	    restart = 0;
-	    reload_ns = 1;
+	    reload_namespace = 1;
 	    SignalRestart();
 	}
-	if (reload_ns) {
-	    reload_ns = 0;
+	if (reload_namespace) {
+	    reload_namespace = 0;
 	    SignalReloadPMNS();
 	}
+	if (labelChanged) {
+	    labelChanged = 0;
+	    SignalReloadLabels();
+	}
 	if (timeToDie) {
 	    SignalShutdown();
 	    break;
@@ -709,7 +846,7 @@ ClientLoop(void)
 }
 
 #ifdef HAVE_SA_SIGINFO
-void
+static void
 SigIntProc(int sig, siginfo_t *sip, void *x)
 {
     killer_sig = sig;
@@ -719,100 +856,118 @@ SigIntProc(int sig, siginfo_t *sip, void *x)
     }
     timeToDie = 1;
 }
+#elif IS_MINGW
+static void
+SigIntProc(int sig)
+{
+    SignalShutdown();
+}
 #else
-void SigIntProc(int sig)
+static void
+SigIntProc(int sig)
 {
     killer_sig = sig;
-#ifndef IS_MINGW
     signal(SIGINT, SigIntProc);
     signal(SIGTERM, SigIntProc);
     timeToDie = 1;
-#else
-    SignalShutdown();
-#endif
 }
 #endif
 
-void SigHupProc(int s)
+#ifdef IS_MINGW
+static void
+SigHupProc(int sig)
 {
-#ifndef IS_MINGW
-    signal(SIGHUP, SigHupProc);
-    restart = 1;
-#else
+    pmcd_sighups++;
     SignalRestart();
     SignalReloadPMNS();
-#endif
+    SignalReloadLabels();
 }
-
-#if HAVE_TRACE_BACK_STACK
-/*
- * max callback procedure depth (MAX_PCS) and max function name length
- * (MAX_SIZE)
- */
-#define MAX_PCS 30
-#define MAX_SIZE 48
-
-#include <libexc.h>
+#else
+static void
+SigHupProc(int sig)
+{
+    signal(SIGHUP, SigHupProc);
+    restart = 1;
+    pmcd_sighups++;
+}
+#endif
 
 static void
-do_traceback(FILE *f)
+SigBad(int sig)
 {
-    __uint64_t	call_addr[MAX_PCS];
-    char	*call_fn[MAX_PCS];
-    char	names[MAX_PCS][MAX_SIZE];
-    int		res;
-    int		i;
+    if (pmDebugOptions.desperate) {
+	pmNotifyErr(LOG_ERR, "Unexpected signal %d ...\n", sig);
 
-    for (i = 0; i < MAX_PCS; i++)
-	call_fn[i] = names[i];
-    res = trace_back_stack(MAX_PCS, call_addr, call_fn, MAX_PCS, MAX_SIZE);
-    for (i = 1; i < res; i++)
-#if defined(HAVE_64BIT_PTR)
-	fprintf(f, "  0x%016llx [%s]\n", call_addr[i], call_fn[i]);
-#else
-	fprintf(f, "  0x%08lx [%s]\n", (__uint32_t)call_addr[i], call_fn[i]);
-#endif
-    return;
+	/* -D desperate on the command line to enable traceback,
+	 * if we have platform support for it
+	 */
+	fprintf(stderr, "\nProcedure call traceback ...\n");
+	__pmDumpStack(stderr);
+	fflush(stderr);
+    }
+    _exit(sig);
 }
-#endif
 
-void SigBad(int sig)
-{
-    __pmNotifyErr(LOG_ERR, "Unexpected signal %d ...\n", sig);
-#if HAVE_TRACE_BACK_STACK
-    fprintf(stderr, "\nProcedure call traceback ...\n");
-    do_traceback(stderr);
-#endif
-    fprintf(stderr, "\nDumping to core ...\n");
-    fflush(stderr);
-    abort();
-}
+#define ENV_WARN_PORT	1
+#define ENV_WARN_LOCAL	2
+#define ENV_WARN_MAXPENDING	4
 
 int
 main(int argc, char *argv[])
 {
     int		sts;
     int		nport = 0;
+    int		localhost = 0;
+    int		maxpending = MAXPENDING;
+    int		env_warn = 0;
     char	*envstr;
 #ifdef HAVE_SA_SIGINFO
     static struct sigaction act;
 #endif
 
-    umask(022);
-    __pmSetProgname(argv[0]);
-    __pmProcessDataSize(NULL);
-    __pmGetUsername(&username);
-    __pmSetInternalState(PM_STATE_PMCS);
+    pmcd_pid = getpid();
 
-    if ((envstr = getenv("PMCD_PORT")) != NULL)
+    umask(022);
+    __pmProcessDataSize(NULL);
+    pmGetUsername(&username);
+    __pmSetInternalState(PM_STATE_PMCS);
+    __pmServerSetFeature(PM_SERVER_FEATURE_DISCOVERY);
+    __pmServerSetFeature(PM_SERVER_FEATURE_CONTAINERS);
+
+    if ((envstr = getenv("PMCD_PORT")) != NULL) {
 	nport = __pmServerAddPorts(envstr);
+	env_warn |= ENV_WARN_PORT;
+    }
+    if ((envstr = getenv("PMCD_LOCAL")) != NULL) {
+	if ((localhost = atoi(envstr)) != 0) {
+	    __pmServerSetFeature(PM_SERVER_FEATURE_LOCAL);
+	    env_warn |= ENV_WARN_LOCAL;
+	}
+    }
+    if ((envstr = getenv("PMCD_MAXPENDING")) != NULL) {
+	maxpending = atoi(envstr);
+	env_warn |= ENV_WARN_MAXPENDING;
+    }
     ParseOptions(argc, argv, &nport);
+    if (localhost)
+	__pmServerAddInterface("INADDR_LOOPBACK");
     if (nport == 0)
 	__pmServerAddPorts(TO_STRING(SERVER_PORT));
 
+    /* Set the local socket path. A message will be generated into the log
+     * if this fails, but it is not fatal, since other connection options
+     * may exist. 
+     */
+    __pmServerSetLocalSocket(sockpath);
+
+    /* Set the service spec. This will cause our service to be advertised on
+     * the network if that is supported.
+     */
+    __pmServerSetServiceSpec(PM_SERVER_SERVICE_SPEC);
+
     if (run_daemon) {
-	fflush(stderr);
-	StartDaemon(argc, argv);
+	__pmServerStart(argc, argv, 1);
+	pmcd_pid = getpid();
     }
 
 #ifdef HAVE_SA_SIGINFO
@@ -828,11 +983,19 @@ main(int argc, char *argv[])
     __pmSetSignalHandler(SIGBUS, SigBad);
     __pmSetSignalHandler(SIGSEGV, SigBad);
 
-    if ((sts = __pmServerOpenRequestPorts(&clientFds, MAXPENDING)) < 0)
+    if ((sts = __pmServerOpenRequestPorts(&clientFds, maxpending)) < 0)
 	DontStart();
     maxReqPortFd = maxClientFd = sts;
 
-    __pmOpenLog(pmProgname, logfile, stderr, &sts);
+    /*
+     * would prefer open log earlier so any messages up to this point
+     * are not lost, but that's not possible ... it has to be after the
+     * command line arguments have been parsed and after the request
+     * port has been successfully established.
+     * Note that if this fails don't worry as messages will still
+     * go to stderr.
+     */
+    pmOpenLog(pmGetProgname(), logfile, stderr, &sts);
     /* close old stdout, and force stdout into same stream as stderr */
     fflush(stdout);
     close(fileno(stdout));
@@ -840,12 +1003,17 @@ main(int argc, char *argv[])
     /* if this fails beware of the sky falling in */
     assert(sts >= 0);
 
-    if (dupok)
-	sts = pmLoadASCIINameSpace(pmnsfile, 1);
-    else
-	sts = pmLoadNameSpace(pmnsfile);
+    if (env_warn & ENV_WARN_PORT)
+	fprintf(stderr, "Warning: nports=%d from PMCD_PORT=%s in environment\n", nport, getenv("PMCD_PORT"));
+    if (env_warn & ENV_WARN_LOCAL)
+	fprintf(stderr, "Warning: localhost only from PMCD_LOCAL=%s in environment\n", getenv("PMCD_LOCAL"));
+    if (env_warn & ENV_WARN_MAXPENDING)
+	fprintf(stderr, "Warning: maxpending=%d from PMCD_MAXPENDING=%s from environment\n", maxpending, getenv("PMCD_MAXPENDING"));
+
+    sts = pmLoadASCIINameSpace(pmnsfile, dupok);
     if (sts < 0) {
-	fprintf(stderr, "Error: pmLoadNameSpace: %s\n", pmErrStr(sts));
+	fprintf(stderr, "Error: pmLoadASCIINameSpace(%s, %d): %s\n",
+	    (pmnsfile == PM_NS_DEFAULT) ? "DEFAULT" : pmnsfile, dupok, pmErrStr(sts));
 	DontStart();
     }
 
@@ -861,18 +1029,18 @@ main(int argc, char *argv[])
     }
 
     if (run_daemon) {
-	if (CreatePIDfile() < 0)
+	if (__pmServerCreatePIDFile(PM_SERVER_SERVICE_SPEC, PM_FATAL_ERR) < 0)
 	    DontStart();
-	if (__pmSetProcessIdentity(username) < 0)
+	if (pmSetProcessIdentity(username) < 0)
 	    DontStart();
     }
 
-    if (__pmSecureServerSetup(certdb, dbpassfile) < 0)
+    if (__pmSecureServerCertificateSetup(certdb, dbpassfile, cert_nickname) < 0)
 	DontStart();
 
     PrintAgentInfo(stderr);
-    __pmAccDumpHosts(stderr);
-    fprintf(stderr, "\npmcd: PID = %" FMT_PID, getpid());
+    __pmAccDumpLists(stderr);
+    fprintf(stderr, "\npmcd: PID = %" FMT_PID, pmcd_pid);
     fprintf(stderr, ", PDU version = %u\n", PDU_VERSION);
     __pmServerDumpRequestPorts(stderr);
     fflush(stderr);
@@ -911,14 +1079,15 @@ AddBadHost(struct __pmSockAddr *hostId)
 	szBadHosts += 8;
 	need = szBadHosts * (int)sizeof(badHost[0]);
 	if ((badHost = (__pmSockAddr **)realloc(badHost, need)) == NULL) {
-	    __pmNoMem("pmcd.AddBadHost", need, PM_FATAL_ERR);
+	    pmNoMem("pmcd.AddBadHost", need, PM_FATAL_ERR);
+	    /*NOTREACHED*/
 	}
     }
     badHost[nBadHosts++] = __pmSockAddrDup(hostId);
     return 1;
 }
 
-void
+static void
 ResetBadHosts(void)
 {
     if (szBadHosts) {
@@ -930,6 +1099,7 @@ ResetBadHosts(void)
     }
     nBadHosts = 0;
     szBadHosts = 0;
+    badHost = NULL;
 }
 
 void
@@ -939,11 +1109,7 @@ CleanupClient(ClientInfo *cp, int sts)
     int		i, msg;
     int		force;
 
-#ifdef PCP_DEBUG
-    force = pmDebug & DBG_TRACE_APPL0;
-#else
-    force = 0;
-#endif
+    force = pmDebugOptions.appl0;
 
     if (sts != 0 || force) {
 	/* for access violations, only print the message if this host hasn't
@@ -952,7 +1118,7 @@ CleanupClient(ClientInfo *cp, int sts)
 	if (sts == PM_ERR_PERMISSION || sts == PM_ERR_CONNLIMIT) {
 	    if ( (msg = AddBadHost(cp->addr)) ) {
 		caddr = __pmSockAddrToString(cp->addr);
-		fprintf(stderr, "access violation from host %s:\n", caddr);
+		fprintf(stderr, "access violation from host %s\n", caddr);
 		free(caddr);
 	    }
 	}
@@ -976,7 +1142,7 @@ CleanupClient(ClientInfo *cp, int sts)
     if (sts != PM_ERR_PERMISSION && sts != PM_ERR_CONNLIMIT)
         __pmAccDelClient(cp->addr);
 
-    pmcd_trace(TR_DEL_CLIENT, cp->fd, sts, 0);
+    pmcd_trace(TR_DEL_CLIENT, cp-client, cp->fd, sts);
     DeleteClient(cp);
 
     if (maxClientFd < maxReqPortFd)
@@ -987,7 +1153,6 @@ CleanupClient(ClientInfo *cp, int sts)
 	    agent[i].profClient = NULL;
 }
 
-#ifdef PCP_DEBUG
 /* Convert a file descriptor to a string describing what it is for. */
 static char *
 FdToString(int fd)
@@ -1003,21 +1168,20 @@ FdToString(int fd)
     for (i = 0; i < nClients; i++)
         if (client[i].status.connected) {
 	    if (fd == client[i].fd) {
-	        sprintf(fdStr, "client[%d] input socket", i);
+	        pmsprintf(fdStr, sizeof(fdStr), "client[%d] input socket", i);
 		return fdStr;
 	    }
 	}
     for (i = 0; i < nAgents; i++)
 	if (agent[i].status.connected) {
 	    if (fd == agent[i].inFd) {
-		sprintf(fdStr, "agent[%d] input", i);
+		pmsprintf(fdStr, sizeof(fdStr), "agent[%d] input", i);
 		return fdStr;
 	    }
 	    else if (fd  == agent[i].outFd) {
-		sprintf(fdStr, "agent[%d] output", i);
+		pmsprintf(fdStr, sizeof(fdStr), "agent[%d] output", i);
 		return fdStr;
 	    }
 	}
     return stdFds[0];
 }
-#endif
