@@ -1,6 +1,7 @@
 /*
  * Event support for the bash tracing PMDA
  *
+ * Copyright (c) 2016 Red Hat.
  * Copyright (c) 2012 Nathan Scott.  All rights reversed.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -19,8 +20,6 @@
 #include "event.h"
 #include "pmda.h"
 
-static char *prefix = "pmdabash";
-static char *pcptmpdir;			/* probably /var/tmp */
 static char pidpath[MAXPATHLEN];
 
 /*
@@ -85,10 +84,10 @@ process_head_parser(bash_process_t *verify, const char *buffer, size_t size)
 
     size = 16 + strlen(script);		/* pid and script name */
     verify->instance = malloc(size);
-    snprintf(verify->instance, size, "%u %s", verify->pid, script);
+    pmsprintf(verify->instance, size, "%u %s", verify->pid, script);
 
-    if (pmDebug & DBG_TRACE_APPL0)
-	__pmNotifyErr(LOG_DEBUG, "process header v%d: inst='%s' ppid=%d",
+    if (pmDebugOptions.appl0)
+	pmNotifyErr(LOG_DEBUG, "process header v%d: inst='%s' ppid=%d",
 			verify->version, verify->instance, verify->parent);
 }
 
@@ -137,11 +136,11 @@ process_verify(const char *bashname, bash_process_t *verify)
     if (*endnum != '\0' || verify->pid < 1)
 	return -1;
 
-    snprintf(path, sizeof(path), "%s%c.%s", pidpath, __pmPathSeparator(), bashname);
+    pmsprintf(path, sizeof(path), "%s%c.%s", pidpath, pmPathSeparator(), bashname);
     if (process_head_verify(path, verify) < 0)
 	return -1;
 
-    snprintf(path, sizeof(path), "%s%c%s", pidpath, __pmPathSeparator(), bashname);
+    pmsprintf(path, sizeof(path), "%s%c%s", pidpath, pmPathSeparator(), bashname);
     if ((fd = open(path, O_RDONLY | O_NONBLOCK)) < 0)
 	return -1;
     if (fstat(fd, &stat) < 0 || !S_ISFIFO(stat.st_mode)) {
@@ -162,15 +161,15 @@ process_alloc(const char *bashname, bash_process_t *init, int numclients)
     int queueid = pmdaEventNewActiveQueue(bashname, bash_maxmem, numclients);
     bash_process_t *bashful = malloc(sizeof(bash_process_t));
 
-    if (pmDebug & DBG_TRACE_APPL1)
-	__pmNotifyErr(LOG_DEBUG, "process_alloc: %s, queueid=%d", bashname, queueid);
+    if (pmDebugOptions.appl1)
+	pmNotifyErr(LOG_DEBUG, "process_alloc: %s, queueid=%d", bashname, queueid);
 
     if (!bashful) {
-	__pmNotifyErr(LOG_ERR, "process allocation out of memory");
+	pmNotifyErr(LOG_ERR, "process allocation out of memory");
 	return NULL;
     }
     if (queueid < 0) {
-	__pmNotifyErr(LOG_ERR, "attempt to dup queue for %s", bashname);
+	pmNotifyErr(LOG_ERR, "attempt to dup queue for %s", bashname);
 	free(bashful);
 	return NULL;
     }
@@ -184,7 +183,7 @@ process_alloc(const char *bashname, bash_process_t *init, int numclients)
     bashful->queueid = queueid;
     bashful->exited = 0;
     bashful->finished = 0;
-    bashful->restrict = 0;
+    bashful->noaccess = 0;
     bashful->version = init->version;
     bashful->padding = 0;
 
@@ -195,8 +194,8 @@ process_alloc(const char *bashname, bash_process_t *init, int numclients)
     /* copy of pointer to dynamically allocated memory */
     bashful->instance = init->instance;
 
-    if (pmDebug & DBG_TRACE_APPL0)
-	__pmNotifyErr(LOG_DEBUG, "process_alloc: %s", bashful->instance);
+    if (pmDebugOptions.appl0)
+	pmNotifyErr(LOG_DEBUG, "process_alloc: %s", bashful->instance);
 
     return bashful;
 }
@@ -206,9 +205,9 @@ event_start(bash_process_t *bp, struct timeval *timestamp)
 {
     int	start = memcmp(timestamp, &bp->startstat, sizeof(*timestamp));
 
-    if (pmDebug & DBG_TRACE_APPL0)
-	__pmNotifyErr(LOG_DEBUG, "check start event for %s (%d), %ld vs %ld",
-		bp->instance, start, bp->startstat.tv_sec, timestamp->tv_sec);
+    if (pmDebugOptions.appl0)
+	pmNotifyErr(LOG_DEBUG, "check start event for %s (%d), %ld vs %ld",
+		bp->instance, start, (long int)bp->startstat.tv_sec, (long int)timestamp->tv_sec);
 
     return start == 0;
 }
@@ -229,8 +228,8 @@ process_init(const char *bashname, bash_process_t **bp)
     pmAtomValue atom;
     pmdaEventClients(&atom);
 
-    if (pmDebug & DBG_TRACE_APPL0)
-	__pmNotifyErr(LOG_DEBUG, "process_init: %s (%d clients)",
+    if (pmDebugOptions.appl0)
+	pmNotifyErr(LOG_DEBUG, "process_init: %s (%d clients)",
 			bashname, atom.ul);
 
     if (process_verify(bashname, &init) < 0)
@@ -273,7 +272,7 @@ process_read(bash_process_t *process)
 #endif
 #endif
 	if (sts != 0) {
-	    __pmNotifyErr(LOG_ERR, "event buffer allocation failure");
+	    pmNotifyErr(LOG_ERR, "event buffer allocation failure");
 	    return -1;
 	}
     }
@@ -292,21 +291,21 @@ multiread:
      */
     if (bytes == 0)
 	return 0;
-    if (bytes < 0 && (errno == EBADF))
+    if (bytes < 0 && (errno == EBADF || errno == EISDIR || errno == EINVAL))
 	return 0;
     if (bytes < 0 && (errno == EAGAIN || errno == EWOULDBLOCK))
 	return 0;
     if (bytes > bash_maxmem)
 	return 0;
     if (bytes < 0) {
-	__pmNotifyErr(LOG_ERR, "read failure on process %s: %s",
+	pmNotifyErr(LOG_ERR, "read failure on process %s: %s",
 		      process->instance, strerror(errno));
 	return -1;
     }
 
     process_stat_timestamp(process, &timestamp);
 
-    buffer[bufsize-1] = '\0';
+    buffer[offset+bytes] = '\0';
     for (s = p = buffer, j = 0; *s != '\0' && j < bufsize-1; s++, j++) {
 	if (*s != '\n')
 	    continue;
@@ -318,7 +317,7 @@ multiread:
     /* did we just do a full buffer read? */
     if (p == buffer) {
 	char msg[64];
-	__pmNotifyErr(LOG_ERR, "Ignoring long (%d bytes) line: \"%s\"", (int)
+	pmNotifyErr(LOG_ERR, "Ignoring long (%d bytes) line: \"%s\"", (int)
 			bytes, __pmdaEventPrint(p, bytes, msg, sizeof(msg)));
     } else if (j == bufsize - 1) {
 	offset = bufsize-1 - (p - buffer);
@@ -333,13 +332,13 @@ process_unlink(bash_process_t *process, const char *bashname)
 {
     char path[MAXPATHLEN];
 
-    snprintf(path, sizeof(path), "%s%c%s", pidpath, __pmPathSeparator(), bashname);
+    pmsprintf(path, sizeof(path), "%s%c%s", pidpath, pmPathSeparator(), bashname);
     unlink(path);
-    snprintf(path, sizeof(path), "%s%c.%s", pidpath, __pmPathSeparator(), bashname);
+    pmsprintf(path, sizeof(path), "%s%c.%s", pidpath, pmPathSeparator(), bashname);
     unlink(path);
 
-    if (pmDebug & DBG_TRACE_APPL0)
-	__pmNotifyErr(LOG_DEBUG, "process_unlink: removed %s", bashname);
+    if (pmDebugOptions.appl0)
+	pmNotifyErr(LOG_DEBUG, "process_unlink: removed %s", bashname);
 }
 
 static int
@@ -347,13 +346,13 @@ process_drained(bash_process_t *process)
 {
     pmAtomValue value = { 0 };
 
-    if (pmDebug & DBG_TRACE_APPL0)
-	__pmNotifyErr(LOG_DEBUG, "process_queue_drained check on queue %d (pid %d)",
+    if (pmDebugOptions.appl0)
+	pmNotifyErr(LOG_DEBUG, "process_queue_drained check on queue %d (pid %d)",
 		      process->queueid, process->pid);
     if (pmdaEventQueueMemory(process->queueid, &value) < 0)
 	return 1;	/* error, consider it drained and cleanup */
-    if (pmDebug & DBG_TRACE_APPL0)
-	__pmNotifyErr(LOG_DEBUG, "process_queue_drained: %s (%llu)", value.ll?"n":"y", (long long)value.ull);
+    if (pmDebugOptions.appl0)
+	pmNotifyErr(LOG_DEBUG, "process_queue_drained: %s (%llu)", value.ll?"n":"y", (long long)value.ull);
     return value.ull == 0;
 }
 
@@ -373,8 +372,8 @@ process_done(bash_process_t *process, const char *bashname)
 	    process_stat_timestamp(process, &timestamp);
 	    pmdaEventQueueAppend(process->queueid, NULL, 0, &timestamp);
 
-	    if (pmDebug & DBG_TRACE_APPL0)
-		__pmNotifyErr(LOG_DEBUG, "process_done: marked queueid %d (pid %d) done",
+	    if (pmDebugOptions.appl0)
+		pmNotifyErr(LOG_DEBUG, "process_done: marked queueid %d (pid %d) done",
 					process->queueid, process->pid);
 	}
     }
@@ -393,8 +392,8 @@ event_refresh(pmInDom bash_indom)
     bash_process_t *bp;
     int i, id, sts, num = scandir(pidpath, &files, NULL, NULL);
 
-    if (pmDebug & DBG_TRACE_APPL0 && num > 2)
-	__pmNotifyErr(LOG_DEBUG, "event_refresh: phase1: %d files", num - 2);
+    if (pmDebugOptions.appl0 && num > 2)
+	pmNotifyErr(LOG_DEBUG, "event_refresh: phase1: %d files", num - 2);
 
     pmdaCacheOp(bash_indom, PMDA_CACHE_INACTIVE);
 
@@ -429,6 +428,6 @@ event_refresh(pmInDom bash_indom)
 void
 event_init(void)
 {
-    pcptmpdir = pmGetConfig("PCP_TMP_DIR");
-    sprintf(pidpath, "%s%c%s", pcptmpdir, __pmPathSeparator(), prefix);
+    char *tmpdir = pmGetConfig("PCP_TMP_DIR");
+    pmsprintf(pidpath, sizeof(pidpath), "%s%cbash", tmpdir, pmPathSeparator());
 }
