@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2011-2012 Aconex.  All Rights Reserved.
+# Copyright (c) 2011-2013 Aconex.  All Rights Reserved.
 # 
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -12,6 +12,17 @@
 # for more details.
 # 
 
+# NOTE: This PMDA attempts to support all versions of ElasticSearch. The
+# ElasticSearch REST API changed from <=0.20 to >=0.90. We internally version
+# the ES REST API, since the upstream developers did not version it.
+#
+# | Internal version | ElasticSearch version range |
+# |         1        |         [0, 0.20.6]         |
+# |         2        |        [0.90.0, 2.x]        |
+#
+# The ElasticSearch version can be mapped to an internal version by running
+# es_rest_version_internal().
+
 use strict;
 use warnings;
 use JSON;
@@ -20,8 +31,7 @@ use LWP::UserAgent;
 
 my $es_port = 9200;
 my $es_instance = 'localhost';
-my $es_user = 'nobody';
-use vars qw($pmda $http $es_cluster $es_nodes $es_nodestats $es_root $es_searchstats);
+use vars qw($pmda $http $es_cluster $es_nodes $es_nodestats $es_root $es_searchstats $es_cluster_state);
 
 my $nodes_indom = 0;
 my @nodes_instances;
@@ -29,10 +39,14 @@ my @nodes_instance_ids;
 my $search_indom = 1;
 my @search_instances;
 my @search_instance_ids;
+my $index_indom = 2;
+my @index_instances;
+my @index_instance_ids;
 
 my @cluster_cache;		# time of last refresh for each cluster
 my $cache_interval = 2;		# min secs between refreshes for clusters
 my $http_timeout = 1;		# max secs for a request (*must* be small).
+my $all_node_stats = 0;		# fetch all node stats if true, local node only if false
 
 # Configuration files for overriding the above settings
 for my $file (pmda_config('PCP_PMDAS_DIR') . '/elasticsearch/es.conf', 'es.conf') {
@@ -47,8 +61,14 @@ $http->timeout($http_timeout);	# if elasticsearch not timely, no soup for you
 # http GET of elasticsearch json from a given url
 sub es_agent_get
 {
-    my $response = $http->get(shift);
-    return undef unless $response->is_success;
+    my $request = shift;
+
+    # $pmda->log("es_agent_get request: $request");
+    my $response = $http->get($request);
+    my $success = $response->is_success;
+    # $pmda->log("es_agent_get success: $success");
+
+    return undef unless $success;
     return $response->decoded_content;
 }
 
@@ -75,6 +95,24 @@ sub es_data_node_instances
     $pmda->replace_indom($nodes_indom, \@nodes_instances);
 }
 
+sub es_data_index_instances
+{
+    my $indexIDs = shift;
+    my $i = 0;
+
+    @index_instances = ();
+    @index_instance_ids = ();
+    foreach my $index (keys %$indexIDs){
+	$index_instances[$i*2] = $i;
+	$index_instances[($i*2)+1] = $index;
+	$index_instance_ids[$i*2] = $i;
+	$index_instance_ids[($i*2)+1] = $index;
+	$i++;
+    }
+    $pmda->replace_indom($index_indom, \@index_instances);
+
+}
+
 # crack json data structure, extract index names
 sub es_search_instances
 {
@@ -100,9 +138,60 @@ sub es_refresh_cluster_health
     $es_cluster = defined($content) ? decode_json($content) : undef;
 }
 
+# Update the JSON hash of ES indices so we can later map the metric names
+# much more easily back to the PMID (during the fetch callback routine).
+#
+sub es_rewrite_cluster_state
+{
+    my $indices = $es_cluster_state->{'metadata'}->{'indices'};
+    foreach my $index_key (keys %$indices) {
+	# Go over each setting key and transpose what the key name is called
+	my $settings = $indices->{$index_key}->{'settings'};
+
+        if (es_rest_version_internal() == 1) {
+	    foreach my $settings_key (keys %$settings) {
+	        # Convert keys like "index.version.created" to "version_created"
+	        my $transformed_key = $settings_key;
+	        $transformed_key =~ s/index\.//;
+	        $transformed_key =~ s/\./_/g;
+	        $settings->{$transformed_key} = $settings->{$settings_key};
+	    }
+        } else {
+            $indices->{$index_key}->{'settings'} = $settings->{'index'};
+	    $settings = $indices->{$index_key}->{'settings'};
+            $settings->{'version_created'} = $settings->{'version'}->{'created'};
+        }
+    }
+}
+
+sub es_refresh_cluster_state
+{
+    my $content = es_agent_get($baseurl . "_cluster/state");
+    if (defined($content)) {
+	$es_cluster_state = decode_json($content);
+	es_rewrite_cluster_state();
+	es_data_index_instances($es_cluster_state->{'metadata'}->{'indices'});
+    } else {
+	$es_cluster_state = undef;
+    }
+}
+
 sub es_refresh_cluster_nodes_stats_all
 {
-    my $content = es_agent_get($baseurl . "_cluster/nodes/stats?all");
+    my $url = undef;
+    if (es_rest_version_internal() == 1) {
+        $url = "_cluster/nodes/";
+    } else {
+        $url = "_nodes/";
+    }
+
+    if (not $all_node_stats) {
+        $url .= "_local/";
+    }
+
+    $url .= "stats";
+
+    my $content = es_agent_get($baseurl . $url);
     if (defined($content)) {
 	$es_nodestats = decode_json($content);
 	es_data_node_instances($es_nodestats->{'nodes'});
@@ -113,7 +202,14 @@ sub es_refresh_cluster_nodes_stats_all
 
 sub es_refresh_cluster_nodes_all
 {
-    my $content = es_agent_get($baseurl . "_cluster/nodes?all");
+    my $url = undef;
+    if (es_rest_version_internal() == 1) {
+        $url = "_cluster/nodes?all";
+    } else {
+        $url = "_nodes?all";
+    }
+
+    my $content = es_agent_get($baseurl . $url);
     if (defined($content)) {
 	$es_nodes = decode_json($content);
 	es_data_node_instances($es_nodes->{'nodes'});
@@ -133,7 +229,11 @@ sub es_refresh_stats_search
     my $content = es_agent_get($baseurl . "_stats/search");
     if (defined($content)) {
 	$es_searchstats = decode_json($content);
-	es_search_instances($es_searchstats->{'_all'}->{'indices'});
+        if (es_rest_version_internal() == 1) {
+	    es_search_instances($es_searchstats->{'_all'}->{'indices'});
+        } else {
+	    es_search_instances($es_searchstats->{'indices'});
+        }
     } else {
 	$es_searchstats = undef;
     }
@@ -162,6 +262,8 @@ sub es_refresh
 	es_refresh_stats_search();
 	# avoid 2nd refresh call on metrics in other cluster
 	$cluster_cache[4] = $cluster_cache[5] = $now;
+    } elsif ($cluster == 6 || $cluster == 7) { # Update the cluster state
+	es_refresh_cluster_state();
     }
     $cluster_cache[$cluster] = $now;
 }
@@ -177,7 +279,35 @@ sub es_lookup_search
 {
     my ($json, $inst) = @_;
     my $searchID = $search_instance_ids[($inst*2)+1];
-    return $json->{'_all'}->{'indices'}->{$searchID};
+
+    if (es_rest_version_internal() == 1) {
+        return $json->{'_all'}->{'indices'}->{$searchID};
+    } else {
+        return $json->{'indices'}->{$searchID};
+    }
+}
+
+sub es_lookup_index
+{
+    my ($json, $inst) = @_;
+    my $indexID = $index_instance_ids[($inst*2)+1];
+    return $json->{'metadata'}->{'indices'}->{$indexID};
+}
+
+# Returns our internal REST version number
+sub es_rest_version_internal
+{
+    my $es_version = es_value($es_root, ['version', 'number']);
+
+    if ($es_version =~ /^0\.[0-9]+\.[0-9]+$/ and
+            (not $es_version =~ /^0\.90/)) {
+        return 1;
+    } else {
+        return 2;
+    }
+
+    # Assume a recent version of ES
+    return 2;
 }
 
 # iterate over metric-name components, performing hash lookups as we go.
@@ -288,10 +418,28 @@ sub es_fetch_callback
 	if (!defined($search))		{ return (PM_ERR_AGAIN, 0); }
 	return es_value($search, \@metric_subnames);
     }
+    elsif ($cluster == 6) {
+	if (!defined($es_cluster_state)){ return (PM_ERR_AGAIN, 0); }
+	if ($inst != PM_IN_NULL)	{ return (PM_ERR_INST, 0); }
+	# remove first couple (i.e. elasticsearch.cluster.)
+	splice(@metric_subnames, 0, 2);
+	return es_value($es_cluster_state, \@metric_subnames);
+    }
+    elsif ($cluster == 7) {
+	# Remove elasticsearch.index
+	splice(@metric_subnames, 0, 2);
+	if (!defined($es_cluster_state)){ return (PM_ERR_AGAIN, 0); }
+	if ($inst == PM_IN_NULL)    { return (PM_ERR_INST, 0); }
+	$search = es_lookup_index($es_cluster_state, $inst);
+	if (!defined($search))      { return (PM_ERR_AGAIN, 0); }
+	return es_value($search, \@metric_subnames);
+
+    }
     return (PM_ERR_PMID, 0);
 }
 
 $pmda = PCP::PMDA->new('elasticsearch', 108);
+$pmda->connect_pmcd;
 
 # cluster stats
 $pmda->add_metric(pmda_pmid(0,0), PM_TYPE_STRING, PM_INDOM_NULL,
@@ -708,72 +856,38 @@ $pmda->add_metric(pmda_pmid(5,7), PM_TYPE_U64, $search_indom,
 		  'elasticsearch.search.perindex.total.search.fetch_time_in_millis',
 		  'Time spent in search fetches to all elasticsearch primaries', '');
 
+# cluster state
+$pmda->add_metric(pmda_pmid(6,0), PM_TYPE_STRING, PM_INDOM_NULL,
+		  PM_SEM_INSTANT, pmda_units(0,0,0,0,0,0),
+		  'elasticsearch.cluster.master_node',
+		  'Internal identifier of the master node of the cluster', '');
+
+# index state
+$pmda->add_metric(pmda_pmid(7,0), PM_TYPE_64, $index_indom,
+		  PM_SEM_INSTANT, pmda_units(0,1,0,0,PM_TIME_MSEC,0),
+		  'elasticsearch.index.settings.gateway_snapshot_interval',
+		  'Interval between gateway snapshots', '');
+$pmda->add_metric(pmda_pmid(7,1), PM_TYPE_U64, $index_indom,
+		  PM_SEM_INSTANT, pmda_units(0,0,0,0,0,0),
+		  'elasticsearch.index.settings.number_of_replicas',
+		  'Number of replicas of shards index setting', '');
+$pmda->add_metric(pmda_pmid(7,2), PM_TYPE_U64, $index_indom,
+		  PM_SEM_INSTANT, pmda_units(0,0,0,0,0,0),
+		  'elasticsearch.index.settings.number_of_shards',
+		  'Number of shards index setting', '');
+$pmda->add_metric(pmda_pmid(7,3), PM_TYPE_U64, $index_indom,
+		  PM_SEM_INSTANT, pmda_units(0,0,0,0,0,0),
+		  'elasticsearch.index.settings.version_created',
+		  'The version of elasticsearch the index was created with', '');
+
 $pmda->add_indom($nodes_indom, \@nodes_instances,
 		'Instance domain exporting each elasticsearch node', '');
 $pmda->add_indom($search_indom, \@search_instances,
 		'Instance domain exporting each elasticsearch index', '');
+$pmda->add_indom($index_indom, \@index_instances,
+		'Instance domain exporting each elasticsearch index metadata', '');
 
 $pmda->set_fetch_callback(\&es_fetch_callback);
 $pmda->set_refresh(\&es_refresh);
-$pmda->set_user($es_user);
+$pmda->set_user('pcp');
 $pmda->run;
-
-=pod
-
-=head1 NAME
-
-pmdaelasticsearch - elasticsearch performance metrics domain agent (PMDA)
-
-=head1 DESCRIPTION
-
-B<pmdaelasticsearch> is a Performance Metrics Domain Agent (PMDA) which exports
-metric values from elasticsearch.
-
-
-=head1 INSTALLATION
-
-This PMDA requires that elasticsearch is running on the local host and
-is accepting queries on TCP port 9200
-
-
-Install the elasticsearch PMDA by using the B<Install> script as root:
-
-	# cd $PCP_PMDAS_DIR/elasticsearch
-	# ./Install
-
-To uninstall, do the following as root:
-
-	# cd $PCP_PMDAS_DIR/elasticsearch
-	# ./Remove
-
-B<pmdaelasticsearch> is launched by pmcd(1) and should never be executed
-directly.  The Install and Remove scripts notify pmcd(1) when
-the agent is installed or removed.
-
-=head1 FILES
-
-=over
-
-=item $PCP_PMDAS_DIR/elasticsearch/es.conf
-
-optional configuration file for B<pmdaelasticsearch>
-
-elasticsearch PMDA for PCP
-
-=item $PCP_PMDAS_DIR/elasticsearch/Install
-
-installation script for the B<pmdaelasticsearch> agent
-
-=item $PCP_PMDAS_DIR/elasticsearch/Remove
-
-undo installation script for the B<pmdaelasticsearch> agent
-
-=item $PCP_LOG_DIR/pmcd/elasticsearch.log
-
-default log file for error messages from B<pmdaelasticsearch>
-
-=back
-
-=head1 SEE ALSO
-
-pmcd(1).
