@@ -16,7 +16,6 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
 #include "pmda.h"
 
 #include <limits.h>
@@ -73,7 +72,7 @@ static pmdaMetric metrictab[] = {
 	PMDA_PMUNITS(0,0,0,0,0,0) } },
     { (void *)"hinv.physmem",
       { PMDA_PMID(CL_SYSCTL,1), PM_TYPE_U64, PM_INDOM_NULL, PM_SEM_DISCRETE,
-	PMDA_PMUNITS(1,0,0,PM_SPACE_BYTE,0,0) } },
+	PMDA_PMUNITS(1,0,0,PM_SPACE_MBYTE,0,0) } },
     { (void *)"kernel.all.load",
       { PMDA_PMID(CL_SYSCTL,2), PM_TYPE_FLOAT, LOADAV_INDOM, PM_SEM_INSTANT,
 	PMDA_PMUNITS(0,0,0,0,0,0) } },
@@ -384,7 +383,9 @@ do_sysctl(mib_t *mp, size_t xpect)
     for ( ; mp->m_fetched == 0; ) {
 	int	sts;
 	sts = sysctl(mp->m_mib, (u_int)mp->m_miblen, mp->m_data, &mp->m_datalen, NULL, 0);
-	fprintf(stderr, "sysctl(%s%s) -> %d (datalen=%d)\n", mp->m_name, mp->m_data == NULL ? " firstcall" : "", sts, (int)mp->m_datalen);
+	if (pmDebugOptions.appl0) {
+	    fprintf(stderr, "sysctl(%s%s) -> %d (datalen=%d)\n", mp->m_name, mp->m_data == NULL ? " firstcall" : "", sts, (int)mp->m_datalen);
+	}
 	if (sts == 0 && mp->m_data != NULL) {
 	    mp->m_fetched = 1;
 	    break;
@@ -394,7 +395,7 @@ do_sysctl(mib_t *mp, size_t xpect)
 	    mp->m_data = realloc(mp->m_data, mp->m_datalen);
 	    if (mp->m_data == NULL) {
 		fprintf(stderr, "Error: %s: buffer alloc failed for sysctl metric \"%s\"\n", mp->m_pcpname, mp->m_name);
-		__pmNoMem("do_sysctl", mp->m_datalen, PM_FATAL_ERR);
+		pmNoMem("do_sysctl", mp->m_datalen, PM_FATAL_ERR);
 		/*NOTREACHED*/
 	    }
 	}
@@ -408,53 +409,6 @@ do_sysctl(mib_t *mp, size_t xpect)
     return mp->m_datalen;
 }
 
-/*
- * kernel memory reader setup
- */
-struct nlist	symbols[] = {
-	{ .n_name = "_ifnet" },
-	{ .n_name = NULL }
-};
-kvm_t	*kvmp;
-
-static void
-kmemread_init(void)
-{
-    int		sts;
-    int		i;
-    char	errmsg[_POSIX2_LINE_MAX];
-
-    /*
-     * If we're running as a daemon PMDA, assume we're setgid kmem,
-     * so we can open /dev/kmem, and downgrade privileges after the
-     * kvm_open().
-     * For a DSO PMDA, we have to assume pmcd has the required
-     * privileges and don't dink with them.
-     */
-    kvmp = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errmsg);
-    if (!isDSO)
-	setgid(getgid());
-    if (kvmp == NULL) {
-	fprintf(stderr, "kmemread_init: kvm_openfiles failed: %s\n", errmsg);
-	return;
-    }
-
-    sts = kvm_nlist(kvmp, symbols);
-    if (sts < 0) {
-	fprintf(stderr, "kmemread_init: kvm_nlist failed: %s\n", pmErrStr(-errno));
-	for (i = 0; i < sizeof(symbols)/sizeof(symbols[0])-1; i++)
-	    symbols[i].n_value = 0;
-	return;
-    }
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0) {
-	for (i = 0; i < sizeof(symbols)/sizeof(symbols[0])-1; i++) {
-	    fprintf(stderr, "Info: kernel symbol %s found at 0x%08lx\n", symbols[i].n_name, symbols[i].n_value);
-	}
-    }
-#endif
-
-}
 
 /*
  * Callback provided to pmdaFetch ... come here once per metric-instance
@@ -464,13 +418,14 @@ static int
 freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 {
     int			sts = PM_ERR_PMID;
-    __pmID_int		*idp = (__pmID_int *)&(mdesc->m_desc.pmid);
+    unsigned int	cluster = pmID_cluster(mdesc->m_desc.pmid);
+    unsigned int	item = pmID_item(mdesc->m_desc.pmid);
     mib_t		*mp;
 
     mp = (mib_t *)mdesc->m_user;
-    if (idp->cluster == CL_SYSCTL) {
+    if (cluster == CL_SYSCTL) {
 	/* sysctl() simple cases */
-	switch (idp->item) {
+	switch (item) {
 	    /* 32-bit integer values */
 	    case 0:		/* hinv.ncpu */
 	    case 18:		/* swap.pagesin */
@@ -488,7 +443,6 @@ freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 		break;
 
 	    /* 64-bit integer values */
-	    case 1:		/* hinv.physmem */
 	    case 25:		/* swap.length */
 	    case 26:		/* swap.used */
 		sts = do_sysctl(mp, sizeof(atom->ull));
@@ -498,6 +452,16 @@ freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 		}
 		break;
 
+	    /* long integer value */
+	    case 1:		/* hinv.physmem */
+		sts = do_sysctl(mp, sizeof(long));
+		if (sts > 0) {
+		    /* stsctl() returns bytes, convert to MBYTES */
+		    atom->ull = (*((long *)mp->m_data))/(1024*1024);
+		    sts = 1;
+		}
+		break;
+
 	    /* string values */
 	    case 15:		/* hinv.cpu.vendor */
 	    case 16:		/* hinv.cpu.model */
@@ -511,7 +475,7 @@ freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 
 	    /* structs and aggregates */
 	    case 2:		/* kernel.all.load */
-		sts = do_sysctl(mp, 3*sizeof(atom->d));
+		sts = do_sysctl(mp, sizeof(struct loadavg));
 		if (sts > 0) {
 		    int			i;
 		    struct loadavg	*lp = (struct loadavg *)mp->m_data;
@@ -533,7 +497,12 @@ freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 	    case 5:		/* kernel.all.cpu.sys */
 	    case 6:		/* kernel.all.cpu.intr */
 	    case 7:		/* kernel.all.cpu.idle */
-		sts = do_sysctl(mp, CPUSTATES*sizeof(atom->ull));
+		/*
+		 * assume this declaration is correct ...
+		 * long pc_cp_time[CPUSTATES];	...
+		 * from /usr/include/sys/pcpu.h
+		 */
+		sts = do_sysctl(mp, CPUSTATES*sizeof(long));
 		if (sts > 0) {
 		    /*
 		     * PMID assignment is important in the "-3" below so
@@ -542,7 +511,7 @@ freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 		     * i.e. CP_USER, CP_NICE, CP_SYS, CP_INTR and
 		     * CP_IDLE
 		     */
-		    atom->ull = 1000*((__uint64_t *)mp->m_data)[idp->item-3]/cpuhz;
+		    atom->ull = 1000*((__uint64_t)((long *)mp->m_data)[item-3])/cpuhz;
 		    sts = 1;
 		}
 		break;
@@ -563,7 +532,7 @@ freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 		     * CPU up to the maximum number of CPUs installed in
 		     * the system.
 		     */
-		    atom->ull = 1000*((__uint64_t *)mp->m_data)[inst * CPUSTATES + idp->item-8]/cpuhz;
+		    atom->ull = 1000*((__uint64_t *)mp->m_data)[inst * CPUSTATES + item-8]/cpuhz;
 		    sts = 1;
 		}
 		break;
@@ -579,9 +548,9 @@ freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 
 	}
     }
-    else if (idp->cluster == CL_SPECIAL) {
+    else if (cluster == CL_SPECIAL) {
 	/* special cases */
-	switch (idp->item) {
+	switch (item) {
 	    case 0:	/* hinv.ndisk */
 		refresh_disk_metrics();
 		atom->ul = pmdaCacheOp(indomtab[DISK_INDOM].it_indom, PMDA_CACHE_SIZE_ACTIVE);
@@ -684,11 +653,11 @@ freebsd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 
 	}
     }
-    else if (idp->cluster == CL_DISK) {
+    else if (cluster == CL_DISK) {
 	/* disk metrics */
 	sts = do_disk_metrics(mdesc, inst, atom);
     }
-    else if (idp->cluster == CL_NETIF) {
+    else if (cluster == CL_NETIF) {
 	/* network interface metrics */
 	sts = do_netif_metrics(mdesc, inst, atom);
     }
@@ -716,11 +685,11 @@ freebsd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
      * they have changed
      */
     for (i = 0; !done_disk && !done_netif && i < numpmid; i++) {
-	if (pmid_cluster(pmidlist[i]) == CL_DISK) {
+	if (pmID_cluster(pmidlist[i]) == CL_DISK) {
 	    refresh_disk_metrics();
 	    done_disk = 1;
 	}
-	else if (pmid_cluster(pmidlist[i]) == CL_NETIF) {
+	else if (pmID_cluster(pmidlist[i]) == CL_NETIF) {
 	    refresh_netif_metrics();
 	    done_netif = 1;
 	}
@@ -733,7 +702,7 @@ freebsd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
  * wrapper for pmdaInstance ... refresh required instance domain first
  */
 static int
-freebsd_instance(pmInDom indom, int inst, char *name, __pmInResult **result, pmdaExt *pmda)
+freebsd_instance(pmInDom indom, int inst, char *name, pmInResult **result, pmdaExt *pmda)
 {
     /*
      * indomtab[] instance names and ids are not used for some indoms,
@@ -793,12 +762,12 @@ freebsd_init(pmdaInterface *dp)
 
     if (isDSO) {
 	char	mypath[MAXPATHLEN];
-	int sep = __pmPathSeparator();
-	snprintf(mypath, sizeof(mypath), "%s%c" "freebsd" "%c" "help",
+	int sep = pmPathSeparator();
+	pmsprintf(mypath, sizeof(mypath), "%s%c" "freebsd" "%c" "help",
 		pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
 	pmdaDSO(dp, PMDA_INTERFACE_5, "freebsd DSO", mypath);
     } else {
-	__pmSetProcessIdentity(username);
+	pmSetProcessIdentity(username);
     }
 
     if (dp->status != 0)
@@ -836,7 +805,7 @@ freebsd_init(pmdaInterface *dp)
 			map[i].m_mib = (int *)malloc(map[i].m_miblen*sizeof(map[i].m_mib[0]));
 			if (map[i].m_mib == NULL) {
 			    fprintf(stderr, "Error: %s (%s): failed mib alloc for sysctl metric \"%s\"\n", map[i].m_pcpname, pmIDStr(metrictab[m].m_desc.pmid), map[i].m_name);
-			    __pmNoMem("freebsd_init: mib", map[i].m_miblen*sizeof(map[i].m_mib[0]), PM_FATAL_ERR);
+			    pmNoMem("freebsd_init: mib", map[i].m_miblen*sizeof(map[i].m_mib[0]), PM_FATAL_ERR);
 			    /*NOTREACHED*/
 			}
 			memcpy(map[i].m_mib, mib, map[i].m_miblen*sizeof(map[i].m_mib[0]));
@@ -846,8 +815,7 @@ freebsd_init(pmdaInterface *dp)
 			metrictab[m].m_user = (void *)&bad_mib;
 		    }
 		}
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_APPL0) {
+		if (pmDebugOptions.appl0) {
 		    int	p;
 		    fprintf(stderr, "Info: %s (%s): sysctl metric \"%s\" -> ", (char *)metrictab[m].m_user, pmIDStr(metrictab[m].m_desc.pmid), map[i].m_name);
 		    for (p = 0; p < map[i].m_miblen; p++) {
@@ -856,7 +824,6 @@ freebsd_init(pmdaInterface *dp)
 		    }
 		    fputc('\n', stderr);
 		}
-#endif
 		metrictab[m].m_user = (void *)&map[i];
 		break;
 	    }
@@ -877,10 +844,8 @@ freebsd_init(pmdaInterface *dp)
 	exit(1);
     }
     cpuhz = clockrates.stathz;
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "Info: CPU time \"hz\" = %d\n", cpuhz);
-#endif
 
     sts = sysctlbyname("kern.cp_times", NULL, &sz, NULL, 0);
     if (sts < 0) {
@@ -892,10 +857,8 @@ freebsd_init(pmdaInterface *dp)
      * explain this
      */
     ncpu = sz / (CPUSTATES * sizeof(__uint64_t));
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "Info: ncpu = %d\n", ncpu);
-#endif
 
     sz = sizeof(pagesize);
     sts = sysctlbyname("hw.pagesize", &pagesize, &sz, NULL, 0);
@@ -903,10 +866,8 @@ freebsd_init(pmdaInterface *dp)
 	fprintf(stderr, "Fatal Error: sysctlbyname(\"hw.pagesize\", ...) failed: %s\n", pmErrStr(-errno));
 	exit(1);
     }
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "Info: VM pagesize = %d\n", pagesize);
-#endif
 
     /*
      * Build some instance domains ...
@@ -914,26 +875,24 @@ freebsd_init(pmdaInterface *dp)
     indomtab[CPU_INDOM].it_numinst = ncpu;
     indomtab[CPU_INDOM].it_set = (pmdaInstid *)malloc(ncpu * sizeof(pmdaInstid));
     if (indomtab[CPU_INDOM].it_set == NULL) {
-	__pmNoMem("freebsd_init: CPU_INDOM it_set", ncpu * sizeof(pmdaInstid), PM_FATAL_ERR);
+	pmNoMem("freebsd_init: CPU_INDOM it_set", ncpu * sizeof(pmdaInstid), PM_FATAL_ERR);
 	/*NOTREACHED*/
     }
     for (i = 0; i < ncpu; i++) {
 	indomtab[CPU_INDOM].it_set[i].i_inst = i;
-	snprintf(iname, sizeof(iname), "cpu%d", i);
+	pmsprintf(iname, sizeof(iname), "cpu%d", i);
 	indomtab[CPU_INDOM].it_set[i].i_name = strdup(iname);
 	if (indomtab[CPU_INDOM].it_set[i].i_name == NULL) {
-	    __pmNoMem("freebsd_init: CPU_INDOM strdup iname", strlen(iname), PM_FATAL_ERR);
+	    pmNoMem("freebsd_init: CPU_INDOM strdup iname", strlen(iname), PM_FATAL_ERR);
 	    /*NOTREACHED*/
 	}
     }
-
-    kmemread_init();
 }
 
 static void
 usage(void)
 {
-    fprintf(stderr, "Usage: %s [options]\n\n", pmProgname);
+    fprintf(stderr, "Usage: %s [options]\n\n", pmGetProgname());
     fputs("Options:\n"
 	  "  -d domain    use domain (numeric) for metrics domain of PMDA\n"
 	  "  -l logfile   write log into logfile rather than using default log name\n"
@@ -941,7 +900,8 @@ usage(void)
 	  "\nExactly one of the following options may appear:\n"
 	  "  -i port      expect PMCD to connect on given inet port (number or name)\n"
 	  "  -p           expect PMCD to supply stdin/stdout (pipe)\n"
-	  "  -u socket    expect PMCD to connect on given unix domain socket\n",
+	  "  -u socket    expect PMCD to connect on given unix domain socket\n"
+	  "  -6 port      expect PMCD to connect on given ipv6 port (number or name)\n",
 	  stderr);		
     exit(1);
 }
@@ -953,20 +913,20 @@ int
 main(int argc, char **argv)
 {
     int			c, err = 0;
-    int			sep = __pmPathSeparator();
+    int			sep = pmPathSeparator();
     pmdaInterface	dispatch;
     char		mypath[MAXPATHLEN];
 
     isDSO = 0;
-    __pmSetProgname(argv[0]);
-    __pmGetUsername(&username);
+    pmSetProgname(argv[0]);
+    pmGetUsername(&username);
 
-    snprintf(mypath, sizeof(mypath), "%s%c" "freebsd" "%c" "help",
+    pmsprintf(mypath, sizeof(mypath), "%s%c" "freebsd" "%c" "help",
 		pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
-    pmdaDaemon(&dispatch, PMDA_INTERFACE_5, pmProgname, FREEBSD,
+    pmdaDaemon(&dispatch, PMDA_INTERFACE_5, pmGetProgname(), FREEBSD,
 		"freebsd.log", mypath);
 
-    while ((c = pmdaGetOpt(argc, argv, "D:d:i:l:pu:U:?", &dispatch, &err)) != EOF) {
+    while ((c = pmdaGetOpt(argc, argv, "D:d:i:l:pu:U:6:?", &dispatch, &err)) != EOF) {
 	switch(c) {
 	case 'U':
 	    username = optarg;
