@@ -1,4 +1,5 @@
 #
+# Copyright (c) 2012 Red Hat.
 # Copyright (c) 2008 Aconex.  All Rights Reserved.
 #
 # This program is free software; you can redistribute it and/or modify it
@@ -40,11 +41,15 @@ sub kvm_fetch_callback
 }
 
 $pmda = PCP::PMDA->new('kvm', 95);
+$pmda->connect_pmcd;
 
+# May need to be root to read directory $kvm_path (/sys/kernel/debug/kvm)
+# - hence we cannot default to using $pmda->set_user('pcp') below.
+#
 my $pmid = 0;
 opendir(DIR, $kvm_path) || $pmda->err("pmdakvm failed to open $kvm_path: $!");
 my @metrics = grep {
-    unless (/^\./) {
+    unless (/^(\.|[0-9])/) {
 	$pmda->add_metric(pmda_pmid(0,$pmid++), PM_TYPE_U64, PM_INDOM_NULL,
 		PM_SEM_COUNTER, pmda_units(0,0,1,0,0,PM_COUNT_ONE),
 		"kvm.$_", '', '');
@@ -54,60 +59,7 @@ my @metrics = grep {
 closedir DIR;
 
 $pmda->set_fetch_callback(\&kvm_fetch_callback);
-# Don't lower permissions ... need to be root to read /sys/kernel/debug/kvm
-#
-# $pmda->set_user('nobody');
+# Careful with permissions - may need to be root to read /sys/kernel/debug/kvm
+# see note above.
+#$pmda->set_user('pcp') if -r $kvm_path;
 $pmda->run;
-
-=pod
-
-=head1 NAME
-
-pmdakvm - Linux virtualisation performance metrics domain agent (PMDA)
-
-=head1 DESCRIPTION
-
-B<pmdakvm> is a Performance Metrics Domain Agent (PMDA) which exports
-metric values from the Linux KVM virtualisation subsystem.
-
-Unlike many PMDAs it dynamically enumerates its metric hierarchy,
-based entirely on the contents of /sys/kernel/debug/kvm.
-
-=head1 INSTALLATION
-
-If you want access to the names and values for the kvm performance
-metrics, do the following as root:
-
-	# cd $PCP_PMDAS_DIR/kvm
-	# ./Install
-
-If you want to undo the installation, do the following as root:
-
-	# cd $PCP_PMDAS_DIR/kvm
-	# ./Remove
-
-B<pmdakvm> is launched by pmcd(1) and should never be executed
-directly.  The Install and Remove scripts notify pmcd(1) when
-the agent is installed or removed.
-
-=head1 FILES
-
-=over
-
-=item $PCP_PMDAS_DIR/kvm/Install
-
-installation script for the B<pmdakvm> agent
-
-=item $PCP_PMDAS_DIR/kvm/Remove
-
-undo installation script for the B<pmdakvm> agent
-
-=item $PCP_LOG_DIR/pmcd/kvm.log
-
-default log file for error messages from B<pmdakvm>
-
-=back
-
-=head1 SEE ALSO
-
-pmcd(1) and kvm(1).
