@@ -1,6 +1,7 @@
 /*
  * Linux Filesystem Cluster
  *
+ * Copyright (c) 2014-2016 Red Hat.
  * Copyright (c) 2000,2004,2007-2008 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -13,89 +14,9 @@
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
  */
-
-#include "pmapi.h"
-#include "impl.h"
-#include "pmda.h"
+#include "linux.h"
 #include "filesys.h"
 
-static void 
-refresh_filesys_projects(pmInDom qindom, filesys_t *fs)
-{
-    char		buffer[MAXPATHLEN];
-    project_t		*qp;
-    fs_quota_stat_t	s;
-    fs_disk_quota_t	d;
-    size_t		idsz, devsz;
-    FILE		*projects;
-    char		*p, *idend;
-    uint32_t		prid;
-    int			qcmd, sts;
-
-    qcmd = QCMD(Q_XGETQSTAT, XQM_PRJQUOTA);
-    if (quotactl(qcmd, fs->device, 0, (void*)&s) < 0)
-	return;
-
-    if (s.qs_flags & XFS_QUOTA_PDQ_ENFD)
-	fs->flags |= FSF_QUOT_PROJ_ENF;
-    if (s.qs_flags & XFS_QUOTA_PDQ_ACCT) 
-	fs->flags |= FSF_QUOT_PROJ_ACC;
-    else
-	return;
-
-    projects = fopen("/etc/projects", "r");
-    if (projects == NULL)
-	return; 
-
-    qcmd = QCMD(Q_XQUOTASYNC, XQM_PRJQUOTA);
-    quotactl(qcmd, fs->device, 0, NULL);
-
-    while (fgets(buffer, sizeof(buffer), projects)) {
-	if (buffer[0] == '#')
-	    continue;
-
-	prid = strtol(buffer, &idend, 10);
-	idsz = idend - buffer;
-	qcmd = QCMD(Q_XGETQUOTA, XQM_PRJQUOTA);
-	if (!idsz || quotactl(qcmd, fs->device, prid, (void *)&d) < 0)
-	    continue;
-
-	devsz = strlen(fs->device);
-	p = malloc(idsz+devsz+2);
-	if (!p)
-	    continue;
-	memcpy(p, buffer, idsz);
-	p[idsz] = ':'; 
-	memcpy(&p[idsz+1], fs->device, devsz+1);
-
-	qp = NULL;
-	sts = pmdaCacheLookupName(qindom, p, NULL, (void **)&qp);
-	if (sts == PMDA_CACHE_ACTIVE)	/* repeated line in /etc/projects? */
-	    goto next;
-	if (sts != PMDA_CACHE_INACTIVE) {
-	    qp = (project_t *)malloc(sizeof(project_t));
-	    if (!qp)
-		goto next;
-#if PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_LIBPMDA)
-		fprintf(stderr, "refresh_filesys_projects: add \"%s\"\n", p);
-#endif
-	}
-	qp->space_hard = d.d_blk_hardlimit;
-	qp->space_soft = d.d_blk_softlimit;
-	qp->space_used = d.d_bcount;
-	qp->space_time_left = d.d_btimer;
-	qp->files_hard = d.d_ino_hardlimit;
-	qp->files_soft = d.d_ino_softlimit;
-	qp->files_used = d.d_icount;
-	qp->files_time_left = d.d_itimer;
-	pmdaCacheStore(qindom, PMDA_CACHE_ADD, p, (void *)qp);
-next:
-	free(p);
-    }
-    fclose(projects);
-}
-
 char *
 scan_filesys_options(const char *options, const char *option)
 {
@@ -103,6 +24,7 @@ scan_filesys_options(const char *options, const char *option)
     char *s;
 
     strncpy(buffer, options, sizeof(buffer));
+    buffer[sizeof(buffer)-1] = '\0';
 
     s = strtok(buffer, ",");
     while (s) {
@@ -114,22 +36,27 @@ scan_filesys_options(const char *options, const char *option)
 }
 
 int
-refresh_filesys(pmInDom filesys_indom, pmInDom quota_indom,
-		pmInDom tmpfs_indom)
+refresh_filesys(pmInDom filesys_indom, pmInDom tmpfs_indom,
+		struct linux_container *cp)
 {
     char buf[MAXPATHLEN];
-    char realdevice[MAXPATHLEN];
+    char src[MAXPATHLEN];
     filesys_t *fs;
     pmInDom indom;
     FILE *fp;
     char *path, *device, *type, *options;
     int sts;
 
-    pmdaCacheOp(quota_indom, PMDA_CACHE_INACTIVE);
     pmdaCacheOp(tmpfs_indom, PMDA_CACHE_INACTIVE);
     pmdaCacheOp(filesys_indom, PMDA_CACHE_INACTIVE);
 
-    if ((fp = fopen("/proc/mounts", "r")) == (FILE *)NULL)
+    /*
+     * When operating within a container namespace, cannot refer
+     * to "self" due to it being a symlinked pid from the host.
+     */
+    pmsprintf(src, sizeof(src), "%s/proc/%s/mounts",
+				linux_statspath, cp ? "1" : "self");
+    if ((fp = fopen(src, "r")) == NULL)
 	return -oserror();
 
     while (fgets(buf, sizeof(buf), fp) != NULL) {
@@ -143,7 +70,12 @@ refresh_filesys(pmInDom filesys_indom, pmInDom quota_indom,
 	    strcmp(type, "nfs") == 0 ||
 	    strcmp(type, "devfs") == 0 ||
 	    strcmp(type, "devpts") == 0 ||
+	    strcmp(type, "devtmpfs") == 0 ||
+	    strcmp(type, "selinuxfs") == 0 ||
+	    strcmp(type, "securityfs") == 0 ||
+	    strcmp(type, "configfs") == 0 ||
 	    strcmp(type, "cgroup") == 0 ||
+	    strcmp(type, "sysfs") == 0 ||
 	    strncmp(type, "auto", 4) == 0)
 	    continue;
 
@@ -154,8 +86,12 @@ refresh_filesys(pmInDom filesys_indom, pmInDom quota_indom,
 	}
 	else if (strncmp(device, "/dev", 4) != 0)
 	    continue;
-	if (realpath(device, realdevice) != NULL)
-	    device = realdevice;
+
+	/* keep dm and md persistent names, RHBZ#1349932 */
+	if (strncmp(device, "/dev/mapper", 11) != 0 && strncmp(device, "/dev/md", 7) != 0) {
+	    if (realpath(device, src) != NULL)
+		device = src;
+	}
 
 	sts = pmdaCacheLookupName(indom, device, NULL, (void **)&fs);
 	if (sts == PMDA_CACHE_ACTIVE)	/* repeated line in /proc/mounts? */
@@ -177,18 +113,13 @@ refresh_filesys(pmInDom filesys_indom, pmInDom quota_indom,
 	    fs->device = strdup(device);
 	    fs->path = strdup(path);
 	    fs->options = strdup(options);
-#if PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_LIBPMDA) {
+	    if (pmDebugOptions.libpmda) {
 		fprintf(stderr, "refresh_filesys: add \"%s\" \"%s\"\n",
 		    fs->path, device);
 	    }
-#endif
 	    pmdaCacheStore(indom, PMDA_CACHE_ADD, device, fs);
 	}
 	fs->flags = 0;
-
-	if (strcmp(type, "xfs") == 0)
-	    refresh_filesys_projects(quota_indom, fs);
     }
 
     /*
