@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2016 Red Hat.
  * Copyright (c) 2011 Aconex.  All Rights Reserved.
  *
  * This program is free software; you can redistribute it and/or modify it
@@ -14,6 +15,8 @@
 
 extern unsigned int irq_err_count;
 
-extern void interrupts_init();
+extern void interrupts_init(pmdaExt *, pmdaMetric *, int);
 extern int refresh_interrupt_values(void);
+extern int refresh_softirqs_values(void);
 extern int interrupts_fetch(int, int, unsigned int, pmAtomValue *);
+
