@@ -1,6 +1,7 @@
 /*
  * Linux /proc/sys/fs metrics cluster
  *
+ * Copyright (c) 2014,2016 Red Hat.
  * Copyright (c) 2003,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -12,53 +13,50 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
  */
-
-#include "pmapi.h"
-#include "impl.h"
-#include "pmda.h"
+#include "linux.h"
 #include "proc_sys_fs.h"
 
-static int err_reported = 0;
-
 int
 refresh_proc_sys_fs(proc_sys_fs_t *proc_sys_fs)
 {
+    static int err_reported;
+    char buf[MAXPATHLEN];
     FILE *filesp = NULL;
     FILE *inodep = NULL;
     FILE *dentryp = NULL;
 
     memset(proc_sys_fs, 0, sizeof(proc_sys_fs_t));
 
-    if ( (filesp  = fopen("/proc/sys/fs/file-nr", "r")) == (FILE *)NULL ||
-	 (inodep  = fopen("/proc/sys/fs/inode-state", "r")) == (FILE *)NULL ||
-	 (dentryp = fopen("/proc/sys/fs/dentry-state", "r")) == (FILE *)NULL) {
-	proc_sys_fs->errcode = -errno;
+    if ((filesp  = linux_statsfile("/proc/sys/fs/file-nr", buf, sizeof(buf))) == NULL ||
+	(inodep  = linux_statsfile("/proc/sys/fs/inode-state", buf, sizeof(buf))) == NULL ||
+	(dentryp = linux_statsfile("/proc/sys/fs/dentry-state", buf, sizeof(buf))) == NULL) {
+	proc_sys_fs->errcode = -oserror();
 	if (err_reported == 0)
 	    fprintf(stderr, "Warning: vfs metrics are not available : %s\n",
-		    strerror(errno));
+		    osstrerror());
     }
     else {
 	proc_sys_fs->errcode = 0;
-	fscanf(filesp,  "%d %d %d",
+	if (fscanf(filesp,  "%d %d %d",
 			&proc_sys_fs->fs_files_count,
 			&proc_sys_fs->fs_files_free,
-			&proc_sys_fs->fs_files_max);
-	fscanf(inodep,  "%d %d",
+			&proc_sys_fs->fs_files_max) != 3)
+	    proc_sys_fs->errcode = PM_ERR_VALUE;
+	if (fscanf(inodep,  "%d %d",
 			&proc_sys_fs->fs_inodes_count,
-			&proc_sys_fs->fs_inodes_free);
-	fscanf(dentryp, "%d %d",
+			&proc_sys_fs->fs_inodes_free) != 2)
+	    proc_sys_fs->errcode = PM_ERR_VALUE;
+	if (fscanf(dentryp, "%d %d",
 			&proc_sys_fs->fs_dentry_count,
-			&proc_sys_fs->fs_dentry_free);
-#if PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LIBPMDA) {
-	    fprintf(stderr, "refresh_proc_sys_fs: found vfs metrics\n");
+			&proc_sys_fs->fs_dentry_free) != 2)
+	    proc_sys_fs->errcode = PM_ERR_VALUE;
+	if (pmDebugOptions.libpmda) {
+	    if (proc_sys_fs->errcode == 0)
+		fprintf(stderr, "refresh_proc_sys_fs: found vfs metrics\n");
+	    else
+		fprintf(stderr, "refresh_proc_sys_fs: botch! missing vfs metrics\n");
 	}
-#endif
     }
     if (filesp)
 	fclose(filesp);
@@ -67,7 +65,7 @@ refresh_proc_sys_fs(proc_sys_fs_t *proc_sys_fs)
     if (dentryp)
 	fclose(dentryp);
 
-    if (err_reported == 0)
+    if (!err_reported)
 	err_reported = 1;
 
     if (proc_sys_fs->errcode == 0)
