@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) Red Hat 2014,2016.
  * Copyright (c) International Business Machines Corp., 2002
  * This code contributed by Mike Mason <mmlnx@us.ibm.com>
  *
@@ -12,27 +13,28 @@
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
  */
-
 #include <fcntl.h>
-#include "pmapi.h"
+#include "linux.h"
 #include "proc_uptime.h"
 
 int
 refresh_proc_uptime(proc_uptime_t *proc_uptime)
 {
-    char buf[80];
+    char buf[MAXPATHLEN];
     int fd, n;
-    float uptime, idletime;
+    float uptime = 0.0, idletime = 0.0;
 
     memset(proc_uptime, 0, sizeof(proc_uptime_t));
-
-    if ((fd = open("/proc/uptime", O_RDONLY)) < 0)
+    pmsprintf(buf, sizeof(buf), "%s/proc/uptime", linux_statspath);
+    if ((fd = open(buf, O_RDONLY)) < 0)
 	return -oserror();
 
     n = read(fd, buf, sizeof(buf));
     close(fd);
     if (n < 0)
 	return -oserror();
+    else if (n > 0)
+	n--;
     buf[n] = '\0';
 
     sscanf((const char *)buf, "%f %f", &uptime, &idletime);
