@@ -3,7 +3,7 @@
  *  
  * Original implementation by Troy Dawson (dawson@fnal.gov)
  *
- * Copyright (c) 2012 Red Hat.
+ * Copyright (c) 2012,2014 Red Hat.
  * Copyright (c) 2001,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -24,7 +24,6 @@
 #include <fcntl.h>
 
 #include "pmapi.h"
-#include "impl.h"
 #include "pmda.h"
 #include "domain.h"
 #include "lmsensors.h"
@@ -645,253 +644,248 @@ lmsensors_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
    lm87 sensor87;
    lm79 sensor79;
    lm75 sensor75;
-    __pmID_int		*idp = (__pmID_int *)&(mdesc->m_desc.pmid);
+   unsigned int cluster = pmID_cluster(mdesc->m_desc.pmid);
+   unsigned int item = pmID_item(mdesc->m_desc.pmid);
 
-    if ((idp->cluster < 0 || idp->cluster > 5) || (idp->item < 0 || idp->item > 16))
+    if (cluster > 5)
 	return PM_ERR_PMID;
     else if (inst != PM_IN_NULL)
 	return PM_ERR_INST;
 
-	if (idp->cluster == 0) {	/*lmsensors*/
-	    switch (idp->item) {
-		case 0:
-			atom->l = schips.total;
-			break ;
-		case 1:
-			atom->l = schips.n_lm75;
-			break ;
-		case 2:
-			atom->l = schips.n_lm79;
-			break ;
-		case 3:
-			atom->l = schips.n_lm87;
-			break ;
-		case 4:
-			atom->l = schips.n_w83781d;
-			break ;
-		case 5:
-			atom->l = schips.n_mtp008;
-			break ;
-		default:
-			printf ("No chips for you today\n");
-			break ;
-	   }
-	}
-	if (idp->cluster == 1) {	/*lmsensors.lm75*/
-		if (schips.n_lm75 > 0) {
-		    sensor75=get_lm75();
-		    switch (idp->item) {
-			case 0:
-				atom->f = sensor75.temp;
-				break ;
-			default:
-				printf ("The magic 8 ball says: Not likely \n");
-				break ;
-			}
-		} else atom->f=9999;
-	}
-	if (idp->cluster == 2) {	/*lmsensors.lm79*/
-		if (schips.n_lm79 > 0) {
-		    sensor79=get_lm79();
-		    switch (idp->item) {
-			case 0:
-				atom->l = sensor79.fan1;
-				break ;
-			case 1:
-				atom->l = sensor79.fan2;
-				break ;
-			case 2:
-				atom->l = sensor79.fan3;
-				break ;
-			case 3:
-				atom->l = sensor79.fan_div;
-				break ;
-			case 4:
-				atom->f = sensor79.temp;
-				break ;
-			case 5:
-				atom->l = sensor79.alarms;
-				break ;
-			case 6:
-				atom->f = sensor79.VCore1;
-				break ;
-			case 7:
-				atom->f = sensor79.VCore2;
-				break ;
-			case 8:
-				atom->f = sensor79.p33V;
-				break ;
-			case 9:
-				atom->f = sensor79.p5V;
-				break ;
-			case 10:
-				atom->f = sensor79.p12V;
-				break ;
-			case 11:
-				atom->f = sensor79.n12V;
-				break ;
-			case 12:
-				atom->f = sensor79.n5V;
-				break ;
-			case 13:
-				atom->f = sensor79.vid;
-				break ;
-			default:
-				printf ("Zis one ... e be tricky\n");
-				break ;
-			}
-		} else atom->f=9999;
-	}
-	if (idp->cluster == 3) {	/*lmsensors.lm87*/
-		if (schips.n_lm87 > 0) {
-		    sensor87=get_lm87();
-		    switch (idp->item) {
-			case 0:
-				atom->l = sensor87.fan1;
-				break ;
-			case 1:
-				atom->l = sensor87.fan2;
-				break ;
-			case 2:
-				atom->f = sensor87.temp1;
-				break ;
-			case 3:
-				atom->f = sensor87.CPUtemp;
-				break ;
-			case 4:
-				atom->f = sensor87.Vccp1;
-				break ;
-			case 5:
-				atom->f = sensor87.Vccp2;
-				break ;
-			case 6:
-				atom->f = sensor87.p25V;
-				break ;
-			case 7:
-				atom->f = sensor87.p33V;
-				break ;
-			case 8:
-				atom->f = sensor87.p5V;
-				break ;
-			case 9:
-				atom->f = sensor87.p12V;
-				break ;
-			case 10:
-				atom->f = sensor87.vid;
-				break ;
-			default:
-				printf ("Zis one ... e be tricky\n");
-				break ;
-			}
-		} else atom->f=9999;
-	}
-	if (idp->cluster == 4) {	/*lmsensors.w83781d*/
-		if (schips.n_w83781d > 0) {
-		    sensorw83781d=get_w83781d();
-		    switch (idp->item) {
-			case 0:
-				atom->l = sensorw83781d.fan1;
-				break ;
-			case 1:
-				atom->l = sensorw83781d.fan2;
-				break ;
-			case 2:
-				atom->l = sensorw83781d.fan3;
-				break ;
-			case 3:
-				atom->l = sensorw83781d.fan_div;
-				break ;
-			case 4:
-				atom->f = sensorw83781d.temp1;
-				break ;
-			case 5:
-				atom->f = sensorw83781d.temp2;
-				break ;
-			case 6:
-				atom->f = sensorw83781d.temp3;
-				break ;
-			case 7:
-				atom->l = sensorw83781d.alarms;
-				break ;
-			case 8:
-				atom->l = sensorw83781d.beep;
-				break ;
-			case 9:
-				atom->f = sensorw83781d.VCore1;
-				break ;
-			case 10:
-				atom->f = sensorw83781d.VCore2;
-				break ;
-			case 11:
-				atom->f = sensorw83781d.p33V;
-				break ;
-			case 12:
-				atom->f = sensorw83781d.p5V;
-				break ;
-			case 13:
-				atom->f = sensorw83781d.p12V;
-				break ;
-			case 14:
-				atom->f = sensorw83781d.n12V;
-				break ;
-			case 15:
-				atom->f = sensorw83781d.n5V;
-				break ;
-			case 16:
-				atom->f = sensorw83781d.vid;
-				break ;
-			default:
-				printf ("Zis one ... e be tricky\n");
-				break ;
-			}
-		} else atom->f=9999;
-	}
-	if (idp->cluster == 5) {	/*lmsensors.mtp008*/
-		if (schips.n_mtp008 > 0) {
-		    sensormtp008=get_mtp008();
-		    switch (idp->item) {
-			case 0:
-				atom->l = sensormtp008.fan1;
-				break ;
-			case 1:
-				atom->l = sensormtp008.fan2;
-				break ;
-			case 2:
-				atom->l = sensormtp008.fan3;
-				break ;
-			case 3:
-				atom->f = sensormtp008.temp1;
-				break ;
-			case 4:
-				atom->f = sensormtp008.temp2;
-				break ;
-			case 5:
-				atom->f = sensormtp008.VCore1;
-				break ;
-			case 6:
-				atom->f = sensormtp008.VCore2;
-				break ;
-			case 7:
-				atom->f = sensormtp008.p33V;
-				break ;
-			case 8:
-				atom->f = sensormtp008.p12V;
-				break ;
-			case 9:
-				atom->f = sensormtp008.n12V;
-				break ;
-			case 10:
-				atom->f = sensormtp008.vid;
-				break ;
-			case 11:
-				atom->f = sensormtp008.vtt;
-				break ;
-			default:
-				printf ("Zis one ... e be tricky\n");
-				break ;
-			}
-		} else atom->f=9999;
-	}
+    if (cluster == 0) {	/*lmsensors*/
+	switch (item) {
+	    case 0:
+		    atom->l = schips.total;
+		    break ;
+	    case 1:
+		    atom->l = schips.n_lm75;
+		    break ;
+	    case 2:
+		    atom->l = schips.n_lm79;
+		    break ;
+	    case 3:
+		    atom->l = schips.n_lm87;
+		    break ;
+	    case 4:
+		    atom->l = schips.n_w83781d;
+		    break ;
+	    case 5:
+		    atom->l = schips.n_mtp008;
+		    break ;
+	    default:
+		    return PM_ERR_PMID;
+       }
+    }
+    if (cluster == 1) {	/*lmsensors.lm75*/
+	    if (schips.n_lm75 > 0) {
+		sensor75=get_lm75();
+		switch (item) {
+		    case 0:
+			    atom->f = sensor75.temp;
+			    break ;
+		    default:
+			    return PM_ERR_PMID;
+		    }
+	    } else atom->f=9999;
+    }
+    if (cluster == 2) {	/*lmsensors.lm79*/
+	    if (schips.n_lm79 > 0) {
+		sensor79=get_lm79();
+		switch (item) {
+		    case 0:
+			    atom->l = sensor79.fan1;
+			    break ;
+		    case 1:
+			    atom->l = sensor79.fan2;
+			    break ;
+		    case 2:
+			    atom->l = sensor79.fan3;
+			    break ;
+		    case 3:
+			    atom->l = sensor79.fan_div;
+			    break ;
+		    case 4:
+			    atom->f = sensor79.temp;
+			    break ;
+		    case 5:
+			    atom->l = sensor79.alarms;
+			    break ;
+		    case 6:
+			    atom->f = sensor79.VCore1;
+			    break ;
+		    case 7:
+			    atom->f = sensor79.VCore2;
+			    break ;
+		    case 8:
+			    atom->f = sensor79.p33V;
+			    break ;
+		    case 9:
+			    atom->f = sensor79.p5V;
+			    break ;
+		    case 10:
+			    atom->f = sensor79.p12V;
+			    break ;
+		    case 11:
+			    atom->f = sensor79.n12V;
+			    break ;
+		    case 12:
+			    atom->f = sensor79.n5V;
+			    break ;
+		    case 13:
+			    atom->f = sensor79.vid;
+			    break ;
+		    default:
+			    return PM_ERR_PMID;
+		    }
+	    } else atom->f=9999;
+    }
+    if (cluster == 3) {	/*lmsensors.lm87*/
+	    if (schips.n_lm87 > 0) {
+		sensor87=get_lm87();
+		switch (item) {
+		    case 0:
+			    atom->l = sensor87.fan1;
+			    break ;
+		    case 1:
+			    atom->l = sensor87.fan2;
+			    break ;
+		    case 2:
+			    atom->f = sensor87.temp1;
+			    break ;
+		    case 3:
+			    atom->f = sensor87.CPUtemp;
+			    break ;
+		    case 4:
+			    atom->f = sensor87.Vccp1;
+			    break ;
+		    case 5:
+			    atom->f = sensor87.Vccp2;
+			    break ;
+		    case 6:
+			    atom->f = sensor87.p25V;
+			    break ;
+		    case 7:
+			    atom->f = sensor87.p33V;
+			    break ;
+		    case 8:
+			    atom->f = sensor87.p5V;
+			    break ;
+		    case 9:
+			    atom->f = sensor87.p12V;
+			    break ;
+		    case 10:
+			    atom->f = sensor87.vid;
+			    break ;
+		    default:
+			    return PM_ERR_PMID;
+		    }
+	    } else atom->f=9999;
+    }
+    if (cluster == 4) {	/*lmsensors.w83781d*/
+	    if (schips.n_w83781d > 0) {
+		sensorw83781d=get_w83781d();
+		switch (item) {
+		    case 0:
+			    atom->l = sensorw83781d.fan1;
+			    break ;
+		    case 1:
+			    atom->l = sensorw83781d.fan2;
+			    break ;
+		    case 2:
+			    atom->l = sensorw83781d.fan3;
+			    break ;
+		    case 3:
+			    atom->l = sensorw83781d.fan_div;
+			    break ;
+		    case 4:
+			    atom->f = sensorw83781d.temp1;
+			    break ;
+		    case 5:
+			    atom->f = sensorw83781d.temp2;
+			    break ;
+		    case 6:
+			    atom->f = sensorw83781d.temp3;
+			    break ;
+		    case 7:
+			    atom->l = sensorw83781d.alarms;
+			    break ;
+		    case 8:
+			    atom->l = sensorw83781d.beep;
+			    break ;
+		    case 9:
+			    atom->f = sensorw83781d.VCore1;
+			    break ;
+		    case 10:
+			    atom->f = sensorw83781d.VCore2;
+			    break ;
+		    case 11:
+			    atom->f = sensorw83781d.p33V;
+			    break ;
+		    case 12:
+			    atom->f = sensorw83781d.p5V;
+			    break ;
+		    case 13:
+			    atom->f = sensorw83781d.p12V;
+			    break ;
+		    case 14:
+			    atom->f = sensorw83781d.n12V;
+			    break ;
+		    case 15:
+			    atom->f = sensorw83781d.n5V;
+			    break ;
+		    case 16:
+			    atom->f = sensorw83781d.vid;
+			    break ;
+		    default:
+			    return PM_ERR_PMID;
+		    }
+	    } else atom->f=9999;
+    }
+    if (cluster == 5) {	/*lmsensors.mtp008*/
+	    if (schips.n_mtp008 > 0) {
+		sensormtp008=get_mtp008();
+		switch (item) {
+		    case 0:
+			    atom->l = sensormtp008.fan1;
+			    break ;
+		    case 1:
+			    atom->l = sensormtp008.fan2;
+			    break ;
+		    case 2:
+			    atom->l = sensormtp008.fan3;
+			    break ;
+		    case 3:
+			    atom->f = sensormtp008.temp1;
+			    break ;
+		    case 4:
+			    atom->f = sensormtp008.temp2;
+			    break ;
+		    case 5:
+			    atom->f = sensormtp008.VCore1;
+			    break ;
+		    case 6:
+			    atom->f = sensormtp008.VCore2;
+			    break ;
+		    case 7:
+			    atom->f = sensormtp008.p33V;
+			    break ;
+		    case 8:
+			    atom->f = sensormtp008.p12V;
+			    break ;
+		    case 9:
+			    atom->f = sensormtp008.n12V;
+			    break ;
+		    case 10:
+			    atom->f = sensormtp008.vid;
+			    break ;
+		    case 11:
+			    atom->f = sensormtp008.vtt;
+			    break ;
+		    default:
+			    return PM_ERR_PMID;
+		    }
+	    } else atom->f=9999;
+    }
 
     return 0;
 }
@@ -904,23 +898,26 @@ lmsensors_init(pmdaInterface *dp)
 {
     get_chips();
 
-    __pmSetProcessIdentity(username);
+    pmSetProcessIdentity(username);
     pmdaSetFetchCallBack(dp, lmsensors_fetchCallBack);
     pmdaInit(dp, NULL, 0, 
 	     metrictab, sizeof(metrictab)/sizeof(metrictab[0]));
 }
 
-static void
-usage(void)
-{
-    fprintf(stderr, "Usage: %s [options]\n\n", pmProgname);
-    fputs("Options:\n"
-	  "  -d domain    use domain (numeric) for metrics domain of PMDA\n"
-	  "  -l logfile   write log into logfile rather than using default log name\n"
-	  "  -U username  user account to run under (default \"pcp\")\n",
-	      stderr);		
-    exit(1);
-}
+pmLongOptions   longopts[] = {
+    PMDA_OPTIONS_HEADER("Options"),
+    PMOPT_DEBUG,
+    PMDAOPT_DOMAIN,
+    PMDAOPT_LOGFILE,
+    PMDAOPT_USERNAME,
+    PMOPT_HELP,
+    PMDA_OPTIONS_END
+};
+
+pmdaOptions     opts = {
+    .short_options = "D:d:l:U:?",
+    .long_options = longopts,
+};
 
 /*
  * Set up the agent if running as a daemon.
@@ -928,30 +925,25 @@ usage(void)
 int
 main(int argc, char **argv)
 {
-    int			c, sep = __pmPathSeparator();
-    int			err = 0;
+    int			sep = pmPathSeparator();
     pmdaInterface	desc;
     char		mypath[MAXPATHLEN];
 
-    __pmSetProgname(argv[0]);
-    __pmGetUsername(&username);
+    pmSetProgname(argv[0]);
+    pmGetUsername(&username);
 
-    snprintf(mypath, sizeof(mypath), "%s%c" "lmsensors" "%c" "help",
+    pmsprintf(mypath, sizeof(mypath), "%s%c" "lmsensors" "%c" "help",
 		pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
-    pmdaDaemon(&desc, PMDA_INTERFACE_2, pmProgname, LMSENSORS,
+    pmdaDaemon(&desc, PMDA_INTERFACE_2, pmGetProgname(), LMSENSORS,
 		"lmsensors.log", mypath);
 
-    while ((c = pmdaGetOpt(argc, argv, "D:d:l:U:?", &desc, &err)) != EOF) {
-	switch(c) {
-	case 'U':
-	    username = optarg;
-	    break;
-	default:
-	    err++;
-	}
+    pmdaGetOptions(argc, argv, &opts, &desc);
+    if (opts.errors) {
+	pmdaUsageMessage(&opts);
+	exit(1);
     }
-    if (err)
-	usage();
+    if (opts.username)
+	username = opts.username;
 
     pmdaOpenLog(&desc);
     lmsensors_init(&desc);
