@@ -1,4 +1,5 @@
 #
+# Copyright (c) 2012 Red Hat.
 # Copyright (c) 2004 Silicon Graphics, Inc.  All Rights Reserved.
 # 
 # This program is free software; you can redistribute it and/or modify it
@@ -11,10 +12,6 @@
 # or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 # for more details.
 # 
-# You should have received a copy of the GNU General Public License along
-# with this program; if not, write to the Free Software Foundation, Inc.,
-# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
-# 
 
 use strict;
 use warnings;
@@ -32,6 +29,7 @@ use vars qw( $total $news_regex %news_hash @news_count @news_last );
 my ($nnrpd_count, $rn_count, $trn_count, $xrn_count, $vn_count) = (0,0,0,0,0);
 my $news_file = pmda_config('PCP_PMDAS_DIR') . '/news/active';
 my $news_indom = 0;
+my $ps_command = pmda_config('PCP_PS_PROG') . ' ' . pmda_config('PCP_PS_ALL_FLAGS') . ' |';
 my $pmda;
 
 sub news_fetch		# called once per ``fetch'' pdu, before callbacks
@@ -48,7 +46,7 @@ sub news_fetch		# called once per ``fetch'' pdu, before callbacks
     }
 
     ($nnrpd_count, $rn_count, $trn_count, $xrn_count, $vn_count) = (0,0,0,0,0);
-    if (open(READERS, 'ps -ef |')) {
+    if (open(READERS, $ps_command)) {
 	while (<READERS>) {
 	    s/\b(:?\d\d:){2}\d\d\b/ Mmm DD /;	# replace times with dates
 	    s/\s*?(\S+?\s+?){8}//;		# nuke the first eight fields
@@ -145,7 +143,7 @@ $pmda->set_fetch_callback(\&news_fetch_callback);
 
 &news_init;
 
-$pmda->set_user('nobody');
+$pmda->set_user('pcp');
 $pmda->run;
 
 =pod
