@@ -1,7 +1,7 @@
 /*
  * systemd support for the systemd PMDA
  *
- * Copyright (c) 2012 Red Hat.
+ * Copyright (c) 2012-2014 Red Hat.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -19,7 +19,7 @@
 #define _POSIX_C_SOURCE 200112L  /* for strtoull */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "pmda.h"
 #include "domain.h"
 
@@ -32,24 +32,31 @@
 #include <stdlib.h>
 #include <assert.h>
 #include <ctype.h>
-
-
-static int _isDSO = 1;
-static char *username = "adm";
+#include <grp.h>
 
 #define DEFAULT_MAXMEM  (2 * 1024 * 1024)       /* 2 megabytes */
 long maxmem;
 int maxfd;
 fd_set fds;
 static int interval_expired;
-static struct timeval interval = { 2, 0 };
-
-static sd_journal* journald_context = NULL; /* Used for monitoring only. */
-static sd_journal* journald_context_seeky = NULL; /* Used for event detail extraction,
-                                                     involving seeks. */
+static struct timeval interval = { 60, 0 };
+static sd_journal *journald_context; /* Used for monitoring only. */
+static sd_journal *journald_context_seeky; /* Used for event detail extraction,
+                                              involving seeks. */
 static int queue_entries = -1;
+static char *username = "adm";
+
+
+/* Track per-context PCP_ATTR_USERID | _GROUPID, so we
+   can filter event records for that context. */
+static int uid_gid_filter_p = 1;
+struct uid_gid_tuple {
+    char wildcard_p; /* do not filter for this context. */
+    char uid_p; char gid_p; /* uid/gid received flags. */
+    int uid; int gid; }; /* uid/gid received from PCP_ATTR_* */
+static struct uid_gid_tuple *ctxtab = NULL;
+int ctxtab_size = 0;
 
-static __pmnsTree *pmns;
 
 static pmdaMetric metrictab[] = {
 /* numclients */
@@ -61,7 +68,7 @@ static pmdaMetric metrictab[] = {
 #define METRICTAB_MAXMEM_PMID metrictab[1].m_desc.pmid
     { NULL,
       { PMDA_PMID(0,1), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT,
-        PMDA_PMUNITS(0,0,1,0,0,0) }, },
+        PMDA_PMUNITS(1,0,0,PM_SPACE_BYTE,0,0) }, },
 /* journal.field.cursor */
 #define METRICTAB_JOURNAL_CURSOR_PMID metrictab[2].m_desc.pmid
     { NULL,
@@ -87,31 +94,62 @@ static pmdaMetric metrictab[] = {
     { NULL,
       { PMDA_PMID(2,1), PM_TYPE_EVENT, PM_INDOM_NULL, PM_SEM_INSTANT,
         PMDA_PMUNITS(0,0,0,0,0,0) }, },
+/* journal.count */
+#define METRICTAB_JOURNAL_COUNT_PMID metrictab[7].m_desc.pmid
+    { NULL,
+      { PMDA_PMID(2,2), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER,
+        PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) }, },
+/* journal.bytes */
+#define METRICTAB_JOURNAL_BYTES_PMID metrictab[8].m_desc.pmid
+    { NULL,
+      { PMDA_PMID(2,3), PM_TYPE_U64, PM_INDOM_NULL, PM_SEM_COUNTER,
+        PMDA_PMUNITS(1,0,0,PM_SPACE_BYTE,0,0) }, },
 };
 
 
-
-
-
-
 void systemd_shutdown(void)
 {
-    if (journald_context >= 0)
+    if (journald_context != 0)
         sd_journal_close (journald_context);
 
-    if (journald_context_seeky >= 0)
+    if (journald_context_seeky != 0)
         sd_journal_close (journald_context_seeky);
 
     /* XXX: pmdaEvent zap queues? */
 }
 
 
+/* Return a strndup (or NULL) of a field of the current journal entry,
+   since sd_journal_get_data returns data that is not
+   \0-terminated. */
+char *
+my_sd_journal_get_data(sd_journal *j, const char *field)
+{
+    int rc;
+    const char* str;
+    size_t str_len;
+
+    assert (j != NULL);
+    assert (field != NULL);
+
+    rc = sd_journal_get_data(j, field,
+                             (const void**) & str, & str_len);
+    if (rc < 0)
+        return NULL;
+
+    return strndup (str, str_len);
+}
+
+
 void systemd_refresh(void)
 {
+    /* Absorb any changes such as inotify() messages. */
+    (void) sd_journal_process(journald_context);
+    (void) sd_journal_process(journald_context_seeky);
+
     while (1) {
         char *cursor = NULL;
         char *timestamp_str = NULL;
-        size_t timestamp_len = 0;
         struct timeval timestamp;
 
         int rc = sd_journal_next(journald_context);
@@ -120,7 +158,7 @@ void systemd_refresh(void)
             break;
 
         if (rc < 0) {
-            __pmNotifyErr(LOG_ERR, "sd_journal_next failure: %s", strerror(-rc));
+            pmNotifyErr(LOG_ERR, "sd_journal_next failure: %s", strerror(-rc));
             break;
         }
 
@@ -128,32 +166,34 @@ void systemd_refresh(void)
            actual journal records. */
         rc = sd_journal_get_cursor(journald_context, &cursor);
         if (rc < 0) {
-            __pmNotifyErr(LOG_ERR, "sd_journal_get_cursor failure: %s",
+            pmNotifyErr(LOG_ERR, "sd_journal_get_cursor failure: %s",
                           strerror(-rc));
             break;
         }
 
         /* Extract a timestamp from the journald event fields. */
-        rc = sd_journal_get_data(journald_context, "_SOURCE_REALTIME_TIMESTAMP",
-                                 (const void**) & timestamp_str, & timestamp_len);
-        if (rc < 0)
-            rc = sd_journal_get_data(journald_context, "__REALTIME_TIMESTAMP",
-                                     (const void**) & timestamp_str, & timestamp_len);
-        if (rc == 0) {
+        timestamp_str = my_sd_journal_get_data(journald_context,
+                                               "_SOURCE_REALTIME_TIMESTAMP");
+        if (timestamp_str == NULL)
+            timestamp_str = my_sd_journal_get_data(journald_context,
+                                                   "__REALTIME_TIMESTAMP");
+        if (timestamp_str == NULL)
+            rc = -ENOMEM;
+        else {
+            const char* curse;
             unsigned long long epoch_us;
-            assert (timestamp_str != NULL);
             /* defined in systemd.journal-fields(7) as
                FIELD_NAME=NNNN, where NNNN is decimal us since epoch. */
-            timestamp_str = strchr (timestamp_str, '=');
-            if (timestamp_str == NULL)
-                rc = -1;
+            curse = strchr (timestamp_str, '=');
+            if (curse == NULL)
+                rc = -EINVAL;
             else {
-                assert (timestamp_str != NULL);
-                timestamp_str ++;
-                epoch_us = strtoull (timestamp_str, NULL, 10);
+                curse ++;
+                epoch_us = strtoull (curse, NULL, 10);
                 timestamp.tv_sec  = epoch_us / 1000000;
                 timestamp.tv_usec = epoch_us % 1000000;
             }
+            free (timestamp_str);
         }
         /* Improvise. */
         if (rc < 0)
@@ -164,7 +204,7 @@ void systemd_refresh(void)
                                   cursor, strlen(cursor)+1 /* \0 */, &timestamp);
         free(cursor); /* Already copied. */
         if (rc < 0) {
-            __pmNotifyErr(LOG_ERR, "pmdaEventQueueAppend failure: %s", pmErrStr(rc));
+            pmNotifyErr(LOG_ERR, "pmdaEventQueueAppend failure: %s", pmErrStr(rc));
             break;
         }
     }
@@ -179,6 +219,104 @@ enum journald_field_encoding {
 
 
 
+int
+systemd_journal_event_filter (void *rp, void *data, size_t size)
+{
+    int rc;
+    struct uid_gid_tuple* ugt = rp;
+
+    assert (ugt == & ctxtab[pmdaGetContext()]);
+    if (pmDebugOptions.appl0)
+        pmNotifyErr(LOG_DEBUG, "filter (%d) uid=%d gid=%d data=%p bytes=%u\n",
+                      pmdaGetContext(), ugt->uid, ugt->gid, data, (unsigned)size);
+
+    /* The data/size pair gives the object in the event queue, i.e.,
+       the systemd journal cursor string.  It has not yet been turned
+       into a PM_TYPE_EVENT tuple yet, and if we have our way, it won't
+       be (for non-participating clients). */
+
+    /* The general filtering idea is to only feed journal records to clients
+       if their uid matches the _UID=NNN field -or- gid matches _GID=MMM
+       -or- the client is highly authenticated (wildcard_p) -or- per-uid filtering
+       was turned off at the pmda level. */
+
+    /* Reminder: function rc == 0 passes the filter. */
+
+    /* Unfiltered?  Everyone gets egg soup! */
+    if (! uid_gid_filter_p)
+        return 0;
+
+    if (pmDebugOptions.appl0)
+        pmNotifyErr(LOG_DEBUG, "filter (%d) uid%s%d gid%s%d wildcard=%d\n",
+                      pmdaGetContext(),
+                      ugt->uid_p?"=":"?", ugt->uid,
+                      ugt->gid_p?"=":"?", ugt->gid,
+                      ugt->wildcard_p);
+
+    /* Superuser?  May we offer some goulash? */
+    if (ugt->wildcard_p)
+        return 0;
+
+    /* Unauthenticated context?  No soup for you! */
+    if (! ugt->uid_p && ! ugt->gid_p)
+        return 1;
+
+    /* OK, we need to take a look at the journal record in question. */
+
+    if (pmDebugOptions.appl0)
+        pmNotifyErr(LOG_DEBUG, "filter cursor=%s\n", (const char*) data);
+
+    (void) size; /* already known \0-terminated */
+    rc = sd_journal_seek_cursor(journald_context_seeky, (char*) data);
+    if (rc < 0) {
+        pmNotifyErr(LOG_ERR, "filter cannot seek to cursor=%s\n",
+                      (const char*) data);
+        return 1; /* No point trying again in systemd_journal_decoder. */
+    }
+
+    rc = sd_journal_next(journald_context_seeky);
+    if (rc < 0) {
+        pmNotifyErr(LOG_ERR, "filter cannot advance to next\n");
+        return 1; /* No point trying again in systemd_journal_decoder. */
+    }
+
+    if (ugt->uid_p) {
+        char *uid_str = my_sd_journal_get_data(journald_context_seeky, "_UID");
+        if (uid_str) {
+            int uid = atoi (& uid_str[5]); /* skip over _UID= */
+            free (uid_str);
+            if (uid == ugt->uid)
+                return 0; /* You're a somebody.  Here's a bowl of stew. */
+        }
+    }
+
+    if (ugt->gid_p) {
+        char *gid_str = my_sd_journal_get_data(journald_context_seeky, "_GID");
+        if (gid_str) {
+            int gid = atoi (& gid_str[5]); /* skip over _GID= */
+            free (gid_str);
+            if (gid == ugt->gid)
+                return 0; /* You're with pals.  Here's a bowl of miso. */
+        }
+    }
+
+    /* No soup for you! */
+    return 1;
+}
+
+
+void
+systemd_journal_event_filter_release (void *rp)
+{
+    /* NB: We have nothing to release, as we don't do memory allocation
+       for the filter per se - we clean up during end-context time.
+       We can't send a NULL to pmdaEventSetFilter for release purposes
+       (since it'll blindly call it), so need this dummy function. */
+
+    (void) rp;
+}
+
+
 int
 systemd_journal_decoder(int eventarray, void *buffer, size_t size,
                         struct timeval *timestamp, void *data)
@@ -257,25 +395,37 @@ systemd_journal_decoder(int eventarray, void *buffer, size_t size,
 }
 
 
-
-static int
-systemd_profile(__pmProfile *prof, pmdaExt *pmda)
+void enlarge_ctxtab(int context)
 {
-    pmdaEventNewClient(pmda->e_context);
-    return 0;
+    /* Grow the context table if necessary. */
+    if (ctxtab_size /* cardinal */ <= context /* ordinal */) {
+        size_t need = (context + 1) * sizeof(struct uid_gid_tuple);
+        ctxtab = realloc (ctxtab, need);
+        if (ctxtab == NULL)
+            pmNoMem("systemd ctx table", need, PM_FATAL_ERR);
+        /* Blank out new entries. */
+        while (ctxtab_size <= context)
+            memset (& ctxtab[ctxtab_size++], 0, sizeof(struct uid_gid_tuple));
+    }
 }
 
+
 static int
 systemd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 {
-    if (_isDSO) { /* poll, in lieu of systemdMain(). */
-        systemd_refresh();
-    }
-
-    pmdaEventNewClient(pmda->e_context);
+    int sts;
+    (void) pmdaEventNewClient(pmda->e_context);
+    enlarge_ctxtab(pmda->e_context);
+    sts = pmdaEventSetFilter(pmda->e_context, queue_entries,
+                             & ctxtab[pmda->e_context], /* any non-NULL value */
+                             systemd_journal_event_filter,
+                             systemd_journal_event_filter_release /* NULL */);
+    if (sts < 0)
+        return sts;
     return pmdaFetch(numpmid, pmidlist, resp, pmda);
 }
 
+
 static int
 systemd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 {
@@ -285,7 +435,7 @@ systemd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
     if (id == METRICTAB_NUMCLIENTS_PMID) {
         sts = pmdaEventClients(atom);
     } else if (id == METRICTAB_MAXMEM_PMID) {
-        atom->ull = (unsigned long long)maxmem;
+        atom->ul = (unsigned long)maxmem;
         sts = PMDA_FETCH_STATIC;
     } else if (id == METRICTAB_JOURNAL_CURSOR_PMID) {
         sts = PMDA_FETCH_NOVALUES;
@@ -293,6 +443,10 @@ systemd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
         sts = PMDA_FETCH_NOVALUES;
     } else if (id == METRICTAB_JOURNAL_BLOB_PMID) {
         sts = PMDA_FETCH_NOVALUES;
+    } else if (id == METRICTAB_JOURNAL_COUNT_PMID) {
+	sts = pmdaEventQueueCounter(queue_entries, atom);
+    } else if (id == METRICTAB_JOURNAL_BYTES_PMID) {
+	sts = pmdaEventQueueBytes(queue_entries, atom);
     } else if (id == METRICTAB_JOURNAL_RECORDS_PMID) {
         enum journald_field_encoding jfe = JFE_STRING_BLOB_AUTO;
         sts = pmdaEventSetAccess(pmdaGetContext(), queue_entries, 1);
@@ -311,131 +465,151 @@ systemd_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
     return sts;
 }
 
+
 static int
-systemd_store(pmResult *result, pmdaExt *pmda)
+systemd_contextAttributeCallBack(int context,
+                                 int attr, const char *value, int length, pmdaExt *pmda)
 {
-    int i;
+    static int rootlike_gids_found = 0;
+    static int adm_gid = -1;
+    static int wheel_gid = -1;
+    static int systemd_journal_gid = -1;
+    int id;
 
-    pmdaEventNewClient(pmda->e_context); /* since there is no storeCallback */
-
-    for (i = 0; i < result->numpmid; i++) {
-        pmValueSet *vsp = result->vset[i];
-        pmID id = vsp->pmid;
-        int sts;
-
-        (void) id;
-        /* NB: nothing writeable at the moment. */
-        sts = PM_ERR_PERMISSION;
-        if (sts < 0 )
-            return sts;
+    /* Look up root-like gids if needed.  A later PCP client that
+       matches any of these group-id's is treated as if root/adm,
+       i.e., journal records are not filtered for them (wildcard_p).
+       XXX: we could  examine group-membership lists and check against
+       uid to also set wildcard_p. */
+    if (! rootlike_gids_found) {
+        struct group *grp;
+        grp = getgrnam("adm");
+        if (grp) adm_gid = grp->gr_gid;
+        grp = getgrnam("wheel");
+        if (grp) wheel_gid = grp->gr_gid;
+        grp = getgrnam("systemd-journal");
+        if (grp) systemd_journal_gid = grp->gr_gid;
+        rootlike_gids_found = 1;
     }
+
+    enlarge_ctxtab(context);
+    assert (ctxtab != NULL && context < ctxtab_size);
+
+    /* NB: we maintain separate uid_p and gid_p for filtering
+       purposes; it's possible that a pcp client might send only
+       PCP_ATTR_USERID, leaving gid=0, possibly leading us to
+       misinterpret that as GROUPID=0 (root) and sending back _GID=0
+       records. */
+    switch (attr) {
+    case PCP_ATTR_USERID:
+        ctxtab[context].uid_p = 1;
+        id = atoi(value);
+        ctxtab[context].uid = id;
+        if (id == 0) /* root */
+            ctxtab[context].wildcard_p = 1;
+        break;
+
+    case PCP_ATTR_GROUPID:
+        ctxtab[context].gid_p = 1;
+        id = atoi(value);
+        ctxtab[context].gid = id;
+        if (id == adm_gid ||
+            id == wheel_gid ||
+            id == systemd_journal_gid)
+            ctxtab[context].wildcard_p = 1;
+        break;
+    }
+
+    if (pmDebugOptions.appl0)
+        pmNotifyErr(LOG_DEBUG, "attrib (%d) uid%s%d gid%s%d wildcard=%d\n",
+                      context,
+                      ctxtab[context].uid_p?"=":"?", ctxtab[context].uid,
+                      ctxtab[context].gid_p?"=":"?", ctxtab[context].gid,
+                      ctxtab[context].wildcard_p);
+
     return 0;
 }
 
+
 static void
 systemd_end_contextCallBack(int context)
 {
     pmdaEventEndClient(context);
+
+    /* assert (ctxtab != NULL && context < ctxtab_size); */
+
+    /* NB: don't do that; this callback may be hit without any fetch
+       calls having been performed, this ctxtab not stretching all the
+       way to [context]. */
+
+    if (context < ctxtab_size)
+        memset (& ctxtab[context], 0, sizeof(struct uid_gid_tuple));
 }
 
-static int
-systemd_pmid(const char *name, pmID *pmid, pmdaExt *pmda)
-{
-    pmdaEventNewClient(pmda->e_context);
-    return pmdaTreePMID(pmns, name, pmid);
-}
 
 static int
-systemd_name(pmID pmid, char ***nameset, pmdaExt *pmda)
+systemd_desc(pmID pmid, pmDesc *desc, pmdaExt *pmda)
 {
-    pmdaEventNewClient(pmda->e_context);
-    return pmdaTreeName(pmns, pmid, nameset);
+    return pmdaDesc(pmid, desc, pmda);
 }
 
-static int
-systemd_children(const char *name, int traverse, char ***kids, int **sts,
-                pmdaExt *pmda)
-{
-    pmdaEventNewClient(pmda->e_context);
-    return pmdaTreeChildren(pmns, name, traverse, kids, sts);
-}
 
 static int
 systemd_text(int ident, int type, char **buffer, pmdaExt *pmda)
 {
-    pmdaEventNewClient(pmda->e_context);
     return pmdaText(ident, type, buffer, pmda);
 }
 
 
-
 void
 systemd_init(pmdaInterface *dp)
 {
     int sts;
     int journal_fd;
 
-    if (_isDSO) {
-        char helppath[MAXPATHLEN];
-        int sep = __pmPathSeparator();
-        snprintf(helppath, sizeof(helppath), "%s%c" "systemd" "%c" "help",
-                 pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
-        pmdaDSO(dp, PMDA_INTERFACE_5, "systemd DSO", helppath);
-        /* A user's own journal may be accessed without process
-           identity changes. */
-    } else {
-        /* The systemwide journal may be accessed by the adm user (group);
-           root access is not necessary. */
-        __pmSetProcessIdentity(username);
-    }
-
-    dp->version.four.fetch = systemd_fetch;
-    dp->version.four.store = systemd_store;
-    dp->version.four.profile = systemd_profile;
-    dp->version.four.pmid = systemd_pmid;
-    dp->version.four.name = systemd_name;
-    dp->version.four.children = systemd_children;
-    dp->version.four.text = systemd_text;
-
+    dp->version.six.desc = systemd_desc;
+    dp->version.six.fetch = systemd_fetch;
+    dp->version.six.text = systemd_text;
+    dp->version.six.attribute = systemd_contextAttributeCallBack;
     pmdaSetFetchCallBack(dp, systemd_fetchCallBack);
     pmdaSetEndContextCallBack(dp, systemd_end_contextCallBack);
-
     pmdaInit(dp, NULL, 0, metrictab, sizeof(metrictab)/sizeof(metrictab[0]));
 
-    /* Create the dynamic PMNS tree and populate it. */
-    if ((sts = __pmNewPMNS(&pmns)) < 0) {
-        __pmNotifyErr(LOG_ERR, "%s: failed to create new pmns: %s\n",
-                        pmProgname, pmErrStr(sts));
-        pmns = NULL;
-        return;
-    }
-
     /* Initialize the systemd side.  This is failure-tolerant.  */
     /* XXX: SD_JOURNAL_{LOCAL|RUNTIME|SYSTEM}_ONLY */
-    int rc = sd_journal_open(& journald_context, 0);
-    if (rc < 0) {
-        __pmNotifyErr(LOG_ERR, "sd_journal_open failure: %s",
-                      strerror(-rc));
+    sts = sd_journal_open(& journald_context, 0);
+    if (sts < 0) {
+        pmNotifyErr(LOG_ERR, "sd_journal_open failure: %s",
+                      strerror(-sts));
+        dp->status = sts;
         return;
     }
 
-    rc = sd_journal_open(& journald_context_seeky, 0);
-    if (rc < 0) {
-        __pmNotifyErr(LOG_ERR, "sd_journal_open #2 failure: %s",
-                      strerror(-rc));
+    sts = sd_journal_open(& journald_context_seeky, 0);
+    if (sts < 0) {
+        pmNotifyErr(LOG_ERR, "sd_journal_open #2 failure: %s",
+                      strerror(-sts));
+        dp->status = sts;
         return;
     }
 
-    rc = sd_journal_seek_tail(journald_context);
-    if (rc < 0) {
-        __pmNotifyErr(LOG_ERR, "sd_journal_seek_tail failure: %s",
-                      strerror(-rc));
+    sts = sd_journal_seek_tail(journald_context);
+    if (sts < 0) {
+        pmNotifyErr(LOG_ERR, "sd_journal_seek_tail failure: %s",
+                      strerror(-sts));
+    }
+
+    /* Work around RHBZ979487. */
+    sts = sd_journal_previous_skip(journald_context, 1);
+    if (sts < 0) {
+        pmNotifyErr(LOG_ERR, "sd_journal_previous_skip failure: %s",
+                      strerror(-sts));
     }
 
     /* Arrange to wake up for journal events. */
     journal_fd = sd_journal_get_fd(journald_context);
     if (journal_fd < 0) {
-        __pmNotifyErr(LOG_ERR, "sd_journal_get_fd failure: %s",
+        pmNotifyErr(LOG_ERR, "sd_journal_get_fd failure: %s",
                       strerror(-journal_fd));
         /* NB: not a fatal error; the select() loop will still time out and
            periodically poll.  This makes it ok for sd_journal_reliable_fd()
@@ -449,7 +623,7 @@ systemd_init(pmdaInterface *dp)
        just use different decoder callbacks. */
     queue_entries = pmdaEventNewQueue("systemd", maxmem);
     if (queue_entries < 0)
-        __pmNotifyErr(LOG_ERR, "pmdaEventNewQueue failure: %s",
+        pmNotifyErr(LOG_ERR, "pmdaEventNewQueue failure: %s",
                       pmErrStr(queue_entries));
 }
 
@@ -457,26 +631,27 @@ systemd_init(pmdaInterface *dp)
 void
 systemdMain(pmdaInterface *dispatch)
 {
-    fd_set              readyfds;
-    int                 nready, pmcdfd;
+    int pmcdfd;
 
     pmcdfd = __pmdaInFd(dispatch);
     if (pmcdfd > maxfd)
         maxfd = pmcdfd;
 
-    FD_ZERO(&fds);
     FD_SET(pmcdfd, &fds);
 
     for (;;) {
+        fd_set readyfds;
+        int nready;
         struct timeval select_timeout = interval;
+
         memcpy(&readyfds, &fds, sizeof(readyfds));
         nready = select(maxfd+1, &readyfds, NULL, NULL, & select_timeout);
-        if (pmDebug & DBG_TRACE_APPL2)
-            __pmNotifyErr(LOG_DEBUG, "select: nready=%d interval=%d",
+        if (pmDebugOptions.appl2)
+            pmNotifyErr(LOG_DEBUG, "select: nready=%d interval=%d",
                           nready, interval_expired);
         if (nready < 0) {
             if (neterror() != EINTR) {
-                __pmNotifyErr(LOG_ERR, "select failure: %s", netstrerror());
+                pmNotifyErr(LOG_ERR, "select failure: %s", netstrerror());
                 exit(1);
             } else if (!interval_expired) {
                 continue;
@@ -484,18 +659,19 @@ systemdMain(pmdaInterface *dispatch)
         }
 
         if (nready > 0 && FD_ISSET(pmcdfd, &readyfds)) {
-            if (pmDebug & DBG_TRACE_APPL0)
-                __pmNotifyErr(LOG_DEBUG, "processing pmcd PDU [fd=%d]", pmcdfd);
+            if (pmDebugOptions.appl0)
+                pmNotifyErr(LOG_DEBUG, "processing pmcd PDU [fd=%d]", pmcdfd);
             if (__pmdaMainPDU(dispatch) < 0) {
                 exit(1);        /* fatal if we lose pmcd */
             }
-            if (pmDebug & DBG_TRACE_APPL0)
-                __pmNotifyErr(LOG_DEBUG, "completed pmcd PDU [fd=%d]", pmcdfd);
+            if (pmDebugOptions.appl0)
+                pmNotifyErr(LOG_DEBUG, "completed pmcd PDU [fd=%d]", pmcdfd);
         }
         systemd_refresh();
     }
 }
 
+
 static void
 convertUnits(char **endnum, long *maxmem)
 {
@@ -519,6 +695,7 @@ convertUnits(char **endnum, long *maxmem)
     (*endnum)++;
 }
 
+
 static void
 usage(void)
 {
@@ -527,10 +704,11 @@ usage(void)
             "Options:\n"
             "  -d domain    use domain (numeric) for metrics domain of PMDA\n"
             "  -l logfile   write log into logfile rather than using default log name\n"
-            "  -m memory    maximum memory used per logfile (default %ld bytes)\n"
+            "  -m memory    maximum memory used per queue (default %ld bytes)\n"
             "  -s interval  default delay between iterations (default %d sec)\n"
-            "  -U username  user account to run under (default \"adm\")\n",
-            pmProgname, maxmem, (int)interval.tv_sec);
+            "  -U username  user account to run under (default \"adm\")\n"
+            "  -f           disable per-uid/gid record filtering (default on)\n",
+            pmGetProgname(), maxmem, (int)interval.tv_sec);
     exit(1);
 }
 
@@ -542,19 +720,17 @@ main(int argc, char **argv)
     char                *endnum;
     pmdaInterface       desc;
     long                minmem;
-    int                 c, err = 0, sep = __pmPathSeparator();
-
-    _isDSO = 0;
+    int                 c, err = 0, sep = pmPathSeparator();
 
     minmem = getpagesize();
     maxmem = (minmem > DEFAULT_MAXMEM) ? minmem : DEFAULT_MAXMEM;
-    __pmSetProgname(argv[0]);
-    snprintf(helppath, sizeof(helppath), "%s%c" "systemd" "%c" "help",
+    pmSetProgname(argv[0]);
+    pmsprintf(helppath, sizeof(helppath), "%s%c" "systemd" "%c" "help",
                 pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
-    pmdaDaemon(&desc, PMDA_INTERFACE_5, pmProgname, SYSTEMD,
+    pmdaDaemon(&desc, PMDA_INTERFACE_6, pmGetProgname(), SYSTEMD,
                 "systemd.log", helppath);
 
-    while ((c = pmdaGetOpt(argc, argv, "D:d:l:m:s:U:?", &desc, &err)) != EOF) {
+    while ((c = pmdaGetOpt(argc, argv, "D:d:l:m:s:U:f?", &desc, &err)) != EOF) {
         switch (c) {
             case 'm':
                 maxmem = strtol(optarg, &endnum, 10);
@@ -562,7 +738,7 @@ main(int argc, char **argv)
                     convertUnits(&endnum, &maxmem);
                 if (*endnum != '\0' || maxmem < minmem) {
                     fprintf(stderr, "%s: invalid max memory '%s' (min=%ld)\n",
-                            pmProgname, optarg, minmem);
+                            pmGetProgname(), optarg, minmem);
                     err++;
                 }
                 break;
@@ -570,7 +746,7 @@ main(int argc, char **argv)
             case 's':
                 if (pmParseInterval(optarg, &interval, &endnum) < 0) {
                     fprintf(stderr, "%s: -s requires a time interval: %s\n",
-                            pmProgname, endnum);
+                            pmGetProgname(), endnum);
                     free(endnum);
                     err++;
                 }
@@ -580,6 +756,10 @@ main(int argc, char **argv)
                 username = optarg;
                 break;
 
+            case 'f':
+                uid_gid_filter_p = 0;
+                break;
+
             default:
                 err++;
                 break;
@@ -589,19 +769,23 @@ main(int argc, char **argv)
     if (err)
         usage();
 
+    FD_ZERO (&fds);
     pmdaOpenLog(&desc);
-    systemd_init(&desc);
+
+    /* The systemwide journal may be accessed by the adm user (group);
+       root access is not necessary. */
+    pmSetProcessIdentity(username);
+    desc.comm.flags |= PDU_FLAG_AUTH;
     pmdaConnect(&desc);
-    systemdMain(&desc);
+    // After this point, systemd_init is allowed to take some extra time.
+    systemd_init(&desc); // sets some fds
+    systemdMain(&desc); // sets some more fds
     systemd_shutdown();
     exit(0);
 }
 
-
-
 /*
   Local Variables:
   c-basic-offset: 4
   End:
 */
-
