@@ -1,6 +1,7 @@
 /*
  * Trivial, configurable PMDA
  *
+ * Copyright (c) 2014 Red Hat.
  * Copyright (c) 1995,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -12,14 +13,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include <pcp/pmapi.h>
-#include <pcp/impl.h>
 #include <pcp/pmda.h>
 #include "domain.h"
 
@@ -51,15 +47,27 @@ static char	*username;
 static char	mypath[MAXPATHLEN];
 static int	isDSO = 1;		/* ==0 if I am a daemon */
 
+static pmLongOptions longopts[] = {
+    PMDA_OPTIONS_HEADER("Options"),
+    PMOPT_DEBUG,
+    PMDAOPT_DOMAIN,
+    PMDAOPT_LOGFILE,
+    PMDAOPT_USERNAME,
+    PMOPT_HELP,
+    PMDA_OPTIONS_END
+};
+static pmdaOptions opts = {
+    .short_options = "D:d:l:U:?",
+    .long_options = longopts,
+};
+
 /*
  * callback provided to pmdaFetch
  */
 static int
 trivial_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 {
-    __pmID_int		*idp = (__pmID_int *)&(mdesc->m_desc.pmid);
-
-    if (idp->cluster != 0 || idp->item != 0)
+    if (pmID_cluster(mdesc->m_desc.pmid) != 0 || pmID_item(mdesc->m_desc.pmid) != 0)
 	return PM_ERR_PMID;
     else if (inst != PM_IN_NULL)
 	return PM_ERR_INST;
@@ -75,12 +83,12 @@ void
 trivial_init(pmdaInterface *dp)
 {
     if (isDSO) {
-	int sep = __pmPathSeparator();
-	snprintf(mypath, sizeof(mypath), "%s%c" "trivial" "%c" "help",
+	int sep = pmPathSeparator();
+	pmsprintf(mypath, sizeof(mypath), "%s%c" "trivial" "%c" "help",
 		pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
 	pmdaDSO(dp, PMDA_INTERFACE_2, "trivial DSO", mypath);
     } else {
-	__pmSetProcessIdentity(username);
+	pmSetProcessIdentity(username);
     }
 
     if (dp->status != 0)
@@ -92,48 +100,31 @@ trivial_init(pmdaInterface *dp)
 	     metrictab, sizeof(metrictab)/sizeof(metrictab[0]));
 }
 
-static void
-usage(void)
-{
-    fprintf(stderr, "Usage: %s [options]\n\n", pmProgname);
-    fputs("Options:\n"
-	  "  -d domain    use domain (numeric) for metrics domain of PMDA\n"
-	  "  -l logfile   write log into logfile rather than using default log name\n"
-	  "  -U username  user account to run under (default \"pcp\")\n",
-	      stderr);		
-    exit(1);
-}
-
 /*
  * Set up the agent if running as a daemon.
  */
 int
 main(int argc, char **argv)
 {
-    int			c, err = 0;
-    int			sep = __pmPathSeparator();
+    int			sep = pmPathSeparator();
     pmdaInterface	desc;
 
     isDSO = 0;
-    __pmSetProgname(argv[0]);
-    __pmGetUsername(&username);
+    pmSetProgname(argv[0]);
+    pmGetUsername(&username);
 
-    snprintf(mypath, sizeof(mypath), "%s%c" "trivial" "%c" "help",
+    pmsprintf(mypath, sizeof(mypath), "%s%c" "trivial" "%c" "help",
 		pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
-    pmdaDaemon(&desc, PMDA_INTERFACE_2, pmProgname, TRIVIAL,
+    pmdaDaemon(&desc, PMDA_INTERFACE_2, pmGetProgname(), TRIVIAL,
 		"trivial.log", mypath);
 
-    while ((c = pmdaGetOpt(argc, argv, "D:d:l:U:?", &desc, &err)) != EOF) {
-	switch(c) {
-	case 'U':
-	    username = optarg;
-	    break;
-	default:
-	    err++;
-	}
+    pmdaGetOptions(argc, argv, &opts, &desc);
+    if (opts.errors) {
+	pmdaUsageMessage(&opts);
+	exit(1);
     }
-    if (err)
-	usage();
+    if (opts.username)
+	username = opts.username;
 
     pmdaOpenLog(&desc);
     trivial_init(&desc);
