@@ -1,6 +1,7 @@
 /*
  * txmon PMDA
  *
+ * Copyright (c) 2012 Red Hat.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -15,19 +16,12 @@
  */
 
 #include <pcp/pmapi.h>
-#include <pcp/impl.h>
 #include <pcp/pmda.h>
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #include "domain.h"
 
-/*
- * there is no obvious way to deduce or determine this ...
- * <sys/sbd.h> gives MAX values for R4K and R10K as 128 ...
- * I know that for real IRIX, 32 is cool!
- */
 #define CACHELINE	32
-
 #define RND_TO_CACHE_LINE(x) (((x + CACHELINE - 1) / CACHELINE) * CACHELINE)
 
 /*
@@ -88,6 +82,7 @@ static pmdaMetric metrictab[] = {
 static int	shmid = -1;
 
 static char	mypath[MAXPATHLEN];
+static char	*username;
 
 /*
  * callback provided to pmdaFetch
@@ -96,22 +91,23 @@ static int
 txmon_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 {
     stat_t		*sp;
-    __pmID_int		*idp = (__pmID_int *)&(mdesc->m_desc.pmid);
+    unsigned int	cluster = pmID_cluster(mdesc->m_desc.pmid);
+    unsigned int	item = pmID_item(mdesc->m_desc.pmid);
     unsigned int	real_count;
 
     if (inst != PM_IN_NULL && mdesc->m_desc.indom == PM_INDOM_NULL)
 	return PM_ERR_INST;
 
-    if (idp->cluster != 0)
+    if (cluster != 0)
 	return PM_ERR_PMID;
 
-    if (idp->item <= 3) {
+    if (item <= 3) {
 	if (inst >= control->n_tx)
 	    return PM_ERR_INST;
 
 	sp = (stat_t *)((__psint_t)control + control->index[inst]);
 
-	switch (idp->item) {
+	switch (item) {
 	    case 0:				/* txmon.count */
 		if (control->level < 1)
 		    return PM_ERR_AGAIN;
@@ -136,7 +132,7 @@ txmon_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 	}
     }
     else {
-	switch (idp->item) {
+	switch (item) {
 
 	    case 4:				/* txmon.control.level */
 		atom->ul = control->level;
@@ -163,16 +159,14 @@ txmon_store(pmResult *result, pmdaExt *pmda)
     int		val;
     int		sts = 0;
     pmValueSet	*vsp = NULL;
-    __pmID_int	*pmidp = NULL;
     stat_t	*sp;
 
     for (i = 0; i < result->numpmid; i++) {
 	vsp = result->vset[i];
-	pmidp = (__pmID_int *)&vsp->pmid;
 
-	if (pmidp->cluster == 0) {	/* all storable metrics are cluster 0 */
+	if (pmID_cluster(vsp->pmid) == 0) {	/* all storable metrics are cluster 0 */
 
-	    switch (pmidp->item) {
+	    switch (pmID_item(vsp->pmid)) {
 		case 0:				/* no store for these ones */
 		case 1:
 		case 2:
@@ -219,33 +213,20 @@ txmon_init(pmdaInterface *dp)
     if (dp->status != 0)
 	return;
 
+    pmSetProcessIdentity(username);
+
     dp->version.two.store = txmon_store;
 
     pmdaSetFetchCallBack(dp, txmon_fetchCallBack);
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0) {
+    if (pmDebugOptions.appl0) {
 	fprintf(stderr, "after pmdaSetFetchCallBack() control @ %p\n", control);
     }
-#endif
 
     pmdaInit(dp, indomtab, 1, metrictab,
 	     sizeof(metrictab)/sizeof(metrictab[0]));
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0) {
+    if (pmDebugOptions.appl0) {
 	fprintf(stderr, "after pmdaInit() control @ %p\n", control);
     }
-#endif
-}
-static void
-usage(void)
-{
-    fprintf(stderr, "Usage: %s [options] tx_type [...]\n\n", pmProgname);
-    fputs(
-"Options:\n"
-"  -d domain    use domain (numeric) for metrics domain of PMDA\n"
-"  -l logfile   write log into logfile rather than using default log name\n",
-	stderr);		
-    exit(1);
 }
 
 /*
@@ -259,38 +240,58 @@ done(void)
 	shmctl(shmid, IPC_RMID, NULL);
 }
 
+pmLongOptions	longopts[] = {
+    PMDA_OPTIONS_HEADER("Options"),
+    PMOPT_DEBUG,
+    PMDAOPT_DOMAIN,
+    PMDAOPT_LOGFILE,
+    PMDAOPT_USERNAME,
+    PMOPT_HELP,
+    PMDA_OPTIONS_END
+};
+
+pmdaOptions	opts = {
+    .short_options = "D:d:l:U:?",
+    .long_options = longopts,
+    .short_usage = "[options] tx_type [...]",
+};
+
 /*
  * Set up the agent.
  */
 int
 main(int argc, char **argv)
 {
-    int			err = 0;
-    int			sep = __pmPathSeparator();
-    pmdaInterface	dispatch;
+    int			n, sep = pmPathSeparator();
     char		*p;
-    int			n;
+    pmdaInterface	dispatch;
     size_t		index_size;
     size_t		shm_size;
     stat_t		*sp;
 
-    __pmSetProgname(argv[0]);
+    pmSetProgname(argv[0]);
+    pmGetUsername(&username);
 
-    snprintf(mypath, sizeof(mypath), "%s%c" "txmon" "%c" "help",
+    pmsprintf(mypath, sizeof(mypath), "%s%c" "txmon" "%c" "help",
 		pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
-    pmdaDaemon(&dispatch, PMDA_INTERFACE_2, pmProgname, TXMON,
+    pmdaDaemon(&dispatch, PMDA_INTERFACE_2, pmGetProgname(), TXMON,
 		"txmon.log", mypath);
 
-    if (pmdaGetOpt(argc, argv, "D:d:l:?", &dispatch, &err) != EOF)
-	err++;
-    if (err)
-	usage();
-
-    n = argc - optind;
-    if (n < 1) {
-	fprintf(stderr, "No transaction types specified?\n");
-	usage();
+    pmdaGetOptions(argc, argv, &opts, &dispatch);
+    if (opts.errors) {
+	pmdaUsageMessage(&opts);
+	exit(1);
     }
+    n = argc - opts.optind;
+    if (n < 1) {
+	pmprintf("No transaction types specified\n\n");
+	pmdaUsageMessage(&opts);
+	exit(1);
+    }
+    if (opts.username)
+	username = opts.username;
+
+    pmdaOpenLog(&dispatch);
 
     /*
      * create the instance domain table ... one entry per transaction type
@@ -316,7 +317,7 @@ main(int argc, char **argv)
      * create the shm segment, and install exit() handler to remove it
      */
     if ((shmid = shmget(KEY, shm_size, IPC_CREAT|0666)) < 0) {
-	fprintf(stderr, "shmid: %s\n", osstrerror());
+	fprintf(stderr, "shmget: %s\n", osstrerror());
 	exit(1);
     }
     atexit(done);
@@ -324,11 +325,9 @@ main(int argc, char **argv)
 	fprintf(stderr, "shmat: %s\n", osstrerror());
 	exit(1);
     }
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0) {
+    if (pmDebugOptions.appl0) {
 	fprintf(stderr, "shmat -> control @ %p\n", control);
     }
-#endif
 
     /*
      * set up the shm control info and directory
@@ -346,17 +345,15 @@ main(int argc, char **argv)
 	 */
 	control->index[n] = p - (char *)control;
 	sp = (stat_t *)p;
-	strncpy(sp->type, argv[optind++], MAXNAMESIZE);
+	strncpy(sp->type, argv[opts.optind++], MAXNAMESIZE);
 	sp->type[MAXNAMESIZE-1] = '\0';		/* ensure null terminated */
 	sp->reset_count = 0;
 	sp->count = 0;
 	sp->max_time = -1.0;
 	sp->sum_time = 0.0;
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0) {
+	if (pmDebugOptions.appl0) {
 	    fprintf(stderr, "index[%d]=%d @ %p name=\"%s\"\n", n, control->index[n], p, sp->type);
 	}
-#endif
 	p += RND_TO_CACHE_LINE(sizeof(stat_t));
 
 	/*
@@ -366,14 +363,11 @@ main(int argc, char **argv)
 	tx_indom[n].i_name = sp->type;
     }
 
-
     /*
      * the real work is done below here ...
      */
-    pmdaOpenLog(&dispatch);
     txmon_init(&dispatch);
     pmdaConnect(&dispatch);
     pmdaMain(&dispatch);
-
     exit(0);
 }
