@@ -111,9 +111,10 @@ Options\n\
   -t delay	maximum number of seconds between reading weblog files\n\
   -u socket	expect PMCD to connect on given unix domain socket\n\
   -U username   user account to run under (default \"pcp\")\n\
+  -6 port	expect PMCD to connect on given ipv6 port (number or name)\n\
 \n\
 If none of the -i, -p or -u options are given, the configuration file is\n\
-checked and then %s terminates.\n", pmProgname, pmProgname);
+checked and then %s terminates.\n", pmGetProgname(), pmGetProgname());
     exit(1);
 }
 
@@ -125,6 +126,7 @@ logmessage(int priority, const char *format, ...)
     char	*level;
     char	*p;
     time_t	now;
+    int		bytes;
 
     buffer[0] = '\0';
     time(&now);
@@ -159,12 +161,17 @@ logmessage(int priority, const char *format, ...)
             break;
     }
 
-    va_start (arglist, format);
-    vsnprintf (buffer, sizeof(buffer), format, arglist);
+    va_start(arglist, format);
+    bytes = vsnprintf(buffer, sizeof(buffer), format, arglist);
+    va_end(arglist);
+    if (bytes >= sizeof(buffer))
+	buffer[sizeof(buffer)-1] = '\0';
+    if (bytes < 0)
+	buffer[0] = '\0';
     for (p = buffer; *p; p++);
     if (*(--p) == '\n') *p = '\0';
-    fprintf (stderr, "[%.19s] %s(%" FMT_PID ") %s: %s\n", ctime(&now), pmProgname, getpid(), level, buffer) ;
-    va_end (arglist) ;
+
+    fprintf(stderr, "[%.19s] %s(%" FMT_PID ") %s: %s\n", ctime(&now), pmGetProgname(), (pid_t)getpid(), level, buffer) ;
 }
 
 /*
@@ -321,16 +328,14 @@ receivePDUs(pmdaInterface *dispatch)
     for (;;) {
 
 	FD_SET(fileno(stdin), &rfds);
-	__pmtimevalNow(&timeout);
+	pmtimevalNow(&timeout);
 	timeout.tv_usec = 0;
 	interval = (time_t)wl_refreshDelay - (timeout.tv_sec % (time_t)wl_refreshDelay);
 	timeout.tv_sec = interval;
 
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL1)
+	if (pmDebugOptions.appl1)
 	    logmessage(LOG_DEBUG, "Select set for %d seconds\n", 
 			 interval);
-#endif
 
 	sts = select(nfds, &rfds, (fd_set*)0, (fd_set*)0, &timeout);
 	if (sts < 0) {
@@ -340,10 +345,8 @@ receivePDUs(pmdaInterface *dispatch)
 
 	if (sts == 0) {
 
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL1)
+	    if (pmDebugOptions.appl1)
 	    	logmessage(LOG_DEBUG, "Select timed out\n");
-#endif
 
 	    refreshAll();
 	    continue;
@@ -447,7 +450,7 @@ main(int argc, char **argv)
     int			argCount = 0;
     int			checkOnly = 0;
     int			sts = 0;
-    int			sep = __pmPathSeparator();
+    int			sep = pmPathSeparator();
     int			n = 0;
     int			serverTableSize = 0;
     int			regexTableSize = 0;
@@ -463,27 +466,23 @@ main(int argc, char **argv)
 	char		*argString;
     } regexargs[2];
 
-#ifdef PCP_DEBUG
     struct timeval	start;
     struct timeval	end;
     double		startTime;
-#endif
 
-    __pmSetProgname(argv[0]);
-    __pmGetUsername(&wl_username);
+    pmSetProgname(argv[0]);
+    pmGetUsername(&wl_username);
 
-#ifdef PCP_DEBUG
-    __pmtimevalNow(&start);
-#endif
+    pmtimevalNow(&start);
 
     wl_isDSO = 0;
 
-    snprintf(wl_helpFile, sizeof(wl_helpFile), "%s%c" "weblog" "%c" "help",
+    pmsprintf(wl_helpFile, sizeof(wl_helpFile), "%s%c" "weblog" "%c" "help",
 		pmGetConfig("PCP_PMDAS_DIR"), sep, sep);
-    pmdaDaemon(&desc, PMDA_INTERFACE_2, pmProgname, WEBSERVER,
+    pmdaDaemon(&desc, PMDA_INTERFACE_2, pmGetProgname(), WEBSERVER,
 		wl_logFile, wl_helpFile);
 
-    while ((n = pmdaGetOpt(argc, argv, "CD:d:h:i:l:n:pS:t:u:U:?", 
+    while ((n = pmdaGetOpt(argc, argv, "CD:d:h:i:l:n:pS:t:u:U:6:?", 
 			   &desc, &err)) != EOF) {
 	switch (n) {
 
@@ -495,7 +494,7 @@ main(int argc, char **argv)
 	    wl_sprocThresh = (int)strtol(optarg, &endnum, 10);
 	    if (*endnum != '\0') {
 		fprintf(stderr, "%s: -S requires numeric argument\n",
-			pmProgname);
+			pmGetProgname());
 		err++;
 	    }
 	    break;
@@ -504,7 +503,7 @@ main(int argc, char **argv)
 	    if (pmParseInterval(optarg, &delta, &err_msg) < 0) {
                     (void)fprintf(stderr,
                             "%s: -n requires a time interval: %s\n",
-                            err_msg, pmProgname);
+                            err_msg, pmGetProgname());
                     free(err_msg);
                     err++;
                 }
@@ -517,7 +516,7 @@ main(int argc, char **argv)
 	    if (pmParseInterval(optarg, &delta, &err_msg) < 0) {
                     (void)fprintf(stderr,
                             "%s: -t requires a time interval: %s\n",
-                            err_msg, pmProgname);
+                            err_msg, pmGetProgname());
                     free(err_msg);
                     err++;
                 }
@@ -531,7 +530,7 @@ main(int argc, char **argv)
 	    break;
 
 	default:
-	    fprintf(stderr, "%s: Unknown option \"-%c\"", pmProgname, (char)n);
+	    fprintf(stderr, "%s: Unknown option \"-%c\"", pmGetProgname(), (char)n);
 	    err++;
 	    break;
 	}
@@ -556,7 +555,7 @@ main(int argc, char **argv)
 	 * on into the logfile
 	 */
 	pmdaOpenLog(&desc);
-	__pmSetProcessIdentity(wl_username);
+	pmSetProcessIdentity(wl_username);
     }
 
     /*
@@ -600,7 +599,7 @@ main(int argc, char **argv)
 		wl_regexTable = (WebRegex*)realloc(wl_regexTable,
 					   regexTableSize * sizeof(WebRegex));
 		if (wl_regexTable == (WebRegex*)0) {
-		    __pmNoMem("main.wl_regexInst", 
+		    pmNoMem("main.wl_regexInst", 
 			     (wl_numRegex + 1) * sizeof(WebRegex),
 			     PM_FATAL_ERR);
 		}
@@ -622,7 +621,7 @@ main(int argc, char **argv)
 		    if (strcmp(wl_regexTable[n].name, 
 			       wl_regexTable[wl_numRegex].name) == 0) {
 
-			snprintf(emess, sizeof(emess), "duplicate regex name (%s)",
+			pmsprintf(emess, sizeof(emess), "duplicate regex name (%s)",
 				wl_regexTable[wl_numRegex].name);
 			yyerror(emess);
 			break;
@@ -720,7 +719,7 @@ main(int argc, char **argv)
 
 	    wl_regexTable[wl_numRegex].regex = malloc(sizeof(*wl_regexTable[wl_numRegex].regex));
 	    if(wl_regexTable[wl_numRegex].regex == NULL) {
-		__pmNoMem("main.wl_regex", 
+		pmNoMem("main.wl_regex", 
 			  sizeof(*wl_regexTable[wl_numRegex].regex),
 			  PM_FATAL_ERR);
 	    }
@@ -730,11 +729,9 @@ main(int argc, char **argv)
 		continue;
 	    }
 
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL0)
+	    if (pmDebugOptions.appl0)
 	    	logmessage(LOG_DEBUG, "%d regex %s: %s\n", 
 			wl_numRegex, wl_regexTable[wl_numRegex].name, buf1);
-#endif
 
 	    wl_regexTable[wl_numRegex].posix_regexp = 1;
 	    wl_numRegex++;
@@ -750,7 +747,7 @@ main(int argc, char **argv)
 		wl_regexTable = (WebRegex*)realloc(wl_regexTable,
 					   regexTableSize * sizeof(WebRegex));
 		if (wl_regexTable == (WebRegex*)0) {
-		    __pmNoMem("main.wl_regexInst", 
+		    pmNoMem("main.wl_regexInst", 
 			     (wl_numRegex + 1) * sizeof(WebRegex),
 			     PM_FATAL_ERR);
 		}
@@ -772,7 +769,7 @@ main(int argc, char **argv)
 		    if (strcmp(wl_regexTable[n].name, 
 			       wl_regexTable[wl_numRegex].name) == 0) {
 
-			snprintf(emess, sizeof(emess), "duplicate regex name (%s)",
+			pmsprintf(emess, sizeof(emess), "duplicate regex name (%s)",
 				wl_regexTable[wl_numRegex].name);
 			yyerror(emess);
 			break;
@@ -813,11 +810,9 @@ main(int argc, char **argv)
 		continue;
 	    }
 
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL0)
+	    if (pmDebugOptions.appl0)
 	    	logmessage(LOG_DEBUG, "%d NON POSIX regex %s: %s\n", 
 			wl_numRegex, wl_regexTable[wl_numRegex].name, buf1);
-#endif
 
 	    wl_regexTable[wl_numRegex].posix_regexp = 0;
 	    wl_numRegex++;
@@ -833,7 +828,7 @@ main(int argc, char **argv)
 		wl_serverInst = (pmdaInstid*)realloc(wl_serverInst,
 					 serverTableSize * sizeof(pmdaInstid));
 		if (wl_serverInst == (pmdaInstid*)0) {
-		    __pmNoMem("main.wl_serverInst", 
+		    pmNoMem("main.wl_serverInst", 
 			     (wl_numServers + 1) * sizeof(pmdaInstid),
 			     PM_FATAL_ERR);
 		}
@@ -841,7 +836,7 @@ main(int argc, char **argv)
 		wl_servers = (WebServer*)realloc(wl_servers,
 					 serverTableSize * sizeof(WebServer));
 		if (wl_servers == (WebServer*)0) {
-		    __pmNoMem("main.wl_servers", 
+		    pmNoMem("main.wl_servers", 
 			     (wl_numServers + 1) * sizeof(WebServer),
 			     PM_FATAL_ERR);
 		}
@@ -861,7 +856,7 @@ main(int argc, char **argv)
 	    if (wl_numServers) {
 		for (n = 0; n < wl_numServers; n++) {
 		    if (strcmp(buf1, wl_serverInst[n].i_name) == 0) {
-			snprintf(emess, sizeof(emess), "duplicate server name (%s)", buf1);
+			pmsprintf(emess, sizeof(emess), "duplicate server name (%s)", buf1);
 			yyerror(emess);
 			break;
 		    }
@@ -929,7 +924,7 @@ main(int argc, char **argv)
 		    break;
 
 	    if (n == wl_numRegex) {
-	    	snprintf(emess, sizeof(emess), "access log regex \"%s\" not defined", buf1);
+	    	pmsprintf(emess, sizeof(emess), "access log regex \"%s\" not defined", buf1);
 		yyerror(emess);
 		skip_to_eol(configFile);
 		continue;
@@ -952,7 +947,7 @@ main(int argc, char **argv)
 	    if (server->counts.active) {
 		tmpFp = fopen(server->access.fileName, "r");
 		if (tmpFp == (FILE*)0) {
-		    snprintf(emess, sizeof(emess), "cannot open access log \"%s\"", buf2);
+		    pmsprintf(emess, sizeof(emess), "cannot open access log \"%s\"", buf2);
 		    yywarn(emess);
 		    server->access.filePtr = -1;
 		}
@@ -985,7 +980,7 @@ main(int argc, char **argv)
 		    break;
 
 	    if (n == wl_numRegex) {
-	    	snprintf(emess, sizeof(emess), "error log regex \"%s\" not defined", buf1);
+	    	pmsprintf(emess, sizeof(emess), "error log regex \"%s\" not defined", buf1);
 		yyerror(emess);
 		skip_to_eol(configFile);
 		continue;
@@ -997,7 +992,7 @@ main(int argc, char **argv)
 	    if (server->counts.active) {
 		tmpFp = fopen(server->error.fileName, "r");
 		if (tmpFp == (FILE*)0) {
-		    snprintf(emess, sizeof(emess), "cannot open error log \"%s\"", buf2);
+		    pmsprintf(emess, sizeof(emess), "cannot open error log \"%s\"", buf2);
 		    yywarn(emess);
 		    server->error.filePtr = -1;
 		}
@@ -1007,8 +1002,7 @@ main(int argc, char **argv)
 
 	    check_to_eol(configFile);
 
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL0) {
+	    if (pmDebugOptions.appl0) {
 		logmessage(LOG_DEBUG, "%d Server %s, %d, %d, %s, %d, %s\n",
 			wl_numServers,
 			wl_serverInst[wl_numServers].i_name,
@@ -1018,7 +1012,6 @@ main(int argc, char **argv)
 			server->error.format,
 			server->error.fileName);
 	    }
-#endif
 
 	    if (server->counts.active)
 		wl_numActive++;
@@ -1026,7 +1019,7 @@ main(int argc, char **argv)
 	    wl_numServers++;
 	}
 	else {
-	    snprintf(emess, sizeof(emess), "illegal keyword \"%s\"", buf1);
+	    pmsprintf(emess, sizeof(emess), "illegal keyword \"%s\"", buf1);
 	    yyerror(emess);
 	    skip_to_eol(configFile);
 	    continue;
@@ -1063,7 +1056,7 @@ main(int argc, char **argv)
 		   "wl_numServers = %d, wl_sprocThresh = %d",
 		   wl_numServers,
 		   wl_sprocThresh);
-	__pmNoMem("main.wl_sproc", 
+	pmNoMem("main.wl_sproc", 
 		  (wl_numSprocs+1) * sizeof(WebSproc),
 		  PM_FATAL_ERR);
     }
@@ -1093,14 +1086,12 @@ main(int argc, char **argv)
 		exit(1);
 	    }
 
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL2)
+	    if (pmDebugOptions.appl2)
 		logmessage(LOG_DEBUG,
 			      "Creating in pipe (in=%d, out=%d) for sproc %d\n",
 			      proc->inFD[0],
 			      proc->inFD[1],
 			      n);
-#endif
 
 	    sts = pipe1(proc->outFD);
 	    if (sts) {
@@ -1110,14 +1101,12 @@ main(int argc, char **argv)
 		exit(1);
 	    }
 
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL2)
+	    if (pmDebugOptions.appl2)
 		logmessage(LOG_DEBUG,
 			      "Creating out pipe (in=%d, out=%d) for sproc %d\n",
 			      proc->outFD[0],
 			      proc->outFD[1],
 			      n);
-#endif
 
 	    proc->firstServer = (n)*wl_sprocThresh;
 	    if (n != wl_numSprocs)
@@ -1144,14 +1133,12 @@ main(int argc, char **argv)
 		exit(1);
 	    }
 
-#ifdef PCP_DEBUG
-	    if(pmDebug & DBG_TRACE_APPL0) {
+	    if(pmDebugOptions.appl0) {
 		logmessage(LOG_INFO,
 			   "main: created sproc %d: pid %" FMT_PID "\n",
 			   n,
 			   proc->pid);
 	    }
-#endif
 
 	    /* close off unwanted pipes */
 
@@ -1172,12 +1159,10 @@ main(int argc, char **argv)
     wl_sproc[0].lastServer  = (wl_numServers <= wl_sprocThresh) ?
 	wl_numServers - 1 : wl_sprocThresh - 1;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	logmessage(LOG_DEBUG,
 		      "Main process will monitor servers 0 to %d\n",
 		      wl_sproc[0].lastServer);
-#endif
 
     for (n=0; n <= wl_sproc[0].lastServer; n++) {
     	if (wl_servers[n].counts.active) {
@@ -1186,13 +1171,10 @@ main(int argc, char **argv)
 	}
     }
 
-#ifdef PCP_DEBUG
-    __pmtimevalNow(&end);
-    startTime = (end.tv_sec - start.tv_sec) + 
-	        ((end.tv_usec - start.tv_usec) / 1000000.0);
-    if (pmDebug & DBG_TRACE_APPL0)
+    pmtimevalNow(&end);
+    startTime = pmtimevalSub(&end, &start);
+    if (pmDebugOptions.appl0)
 	logmessage(LOG_DEBUG, "Agent started in %f seconds", startTime);
-#endif
 
     receivePDUs(&desc);
 
