@@ -2,6 +2,7 @@
  * dstruct.c - central data structures and associated operations
  ***********************************************************************
  *
+ * Copyright (c) 2013-2015 Red Hat.
  * Copyright (c) 1995-2003 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -15,7 +16,6 @@
  * for more details.
  */
 #include "pmapi.h"
-#include "impl.h"
 #include <math.h>
 #include <ctype.h>
 #include <limits.h>
@@ -40,7 +40,6 @@
  * constants
  ***********************************************************************/
 
-char	localHost[MAXHOSTNAMELEN+1];	/* "official" name of localhost */
 double	mynan;				/* not-a-number run time initialized */
 
 
@@ -52,7 +51,7 @@ char		*pmnsfile = PM_NS_DEFAULT;	/* alternate name space */
 Archive		*archives;			/* list of open archives */
 RealTime	first = -1;			/* archive starting point */
 RealTime	last = 0.0;			/* archive end point */
-char		*dfltHost;			/* default host */
+char		*dfltHostConn;			/* default host connection string */
 RealTime	dfltDelta = DELTA_DFLT;		/* default sample interval */
 char		*startFlag;			/* start time specified? */
 char		*stopFlag;			/* end time specified? */
@@ -61,13 +60,16 @@ char		*offsetFlag;			/* offset time specified? */
 RealTime	runTime;			/* run time interval */
 int		hostZone;			/* timezone from host? */
 char		*timeZone;			/* timezone from command line */
+int		quiet;				/* suppress default diagnostics */
 int		verbose;			/* verbosity 0, 1 or 2 */
 int		interactive;			/* interactive mode, -d */
 int		isdaemon;			/* run as a daemon */
 int		agent;				/* secret agent mode? */
 int		applet;				/* applet mode? */
 int		dowrap;				/* counter wrap? default no */
-int		noDnsFlag;			/* do a default name lookup? */
+int		doexit;				/* time to exit stage left? */
+int		dorotate;			/* is a log rotation pending? */
+int		inrun;				/* parsing done, in run() */
 pmiestats_t	*perf;				/* live performance data */
 pmiestats_t	instrument;			/* used if no mmap (archive) */
 
@@ -121,8 +123,8 @@ getReal(void)
 {
     struct timeval t;
 
-    __pmtimevalNow(&t);
-    return realize(t);
+    pmtimevalNow(&t);
+    return pmtimevalToReal(&t);
 }
 
 
@@ -168,30 +170,17 @@ reflectTime(RealTime d)
     weekday = (double) tm.tm_wday;
 }
 
-
-/* convert RealTime to timeval */
-void
-unrealize(RealTime rt, struct timeval *tv)
-{
-    tv->tv_sec = (time_t)rt;
-    tv->tv_usec = (int)(1000000 * (rt - tv->tv_sec));
-}
-
-
 /* convert RealTime to timespec */
-void
+static void
 unrealizenano(RealTime rt, struct timespec *ts)
 {
     ts->tv_sec = (time_t)rt;
     ts->tv_nsec = (int)(1000000000 * (rt - ts->tv_sec));
 }
 
-#define SLEEP_EVAL	0
-#define SLEEP_RETRY	1
-
-/* sleep until eval or retry RealTime */
+/* sleep until eval RealTime */
 void
-sleepTight(Task *t, int type)
+sleepTight(Task *t)
 {
     RealTime	sched;
     RealTime	delay;	/* interval to sleep */
@@ -203,8 +192,7 @@ sleepTight(Task *t, int type)
     if (need_wait) {
 	/* harvest terminated children */
 	while ((pid = waitpid(-1, &sts, WNOHANG)) > (pid_t)0) {
-#if PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL2) {
+	    if (pmDebugOptions.appl2) {
 		fprintf(stderr, "sleepTight: wait: pid=%" FMT_PID " done status=0x%x", pid, sts);
 		if (WIFEXITED(sts))
 		    fprintf(stderr, " exit=%d", WEXITSTATUS(sts));
@@ -212,7 +200,6 @@ sleepTight(Task *t, int type)
 		    fprintf(stderr, " signal=%d", WTERMSIG(sts));
 		fprintf(stderr, "\n");
 	    }
-#endif
 	    ;
 	}
 	need_wait = 0;
@@ -223,31 +210,29 @@ sleepTight(Task *t, int type)
 	struct timespec ts, tleft;
 	static RealTime	last_sched = -1;
 	static Task *last_t;
-	static int last_type;
 	RealTime cur = getReal();
 
-	sched = type == SLEEP_EVAL ? t->eval : t->retry;
+	sched = t->eval;
 
 	delay = sched - cur;
 	if (delay < 0) {
 	    int		show_detail = 0;
-	    if (delay <= -1) {
+	    if (delay <= -1 && !quiet) {
 		fprintf(stderr, "sleepTight: negative delay (%f). sched=%f, cur=%f\n",
 			    delay, sched, cur);
 		show_detail = 1;
 	    }
-#if PCP_DEBUG
 	    else {
-		if (pmDebug & DBG_TRACE_APPL2) {
+		if (pmDebugOptions.appl2) {
 		    fprintf(stderr, "sleepTight: small negative delay (%f). sched=%f, cur=%f\n",
 			    delay, sched, cur);
 		    show_detail = 1;
 		}
 	    }
-#endif
+
 	    if (show_detail) {
 		if (last_sched > 0) {
-		    fprintf(stderr, "Last sleepTight (%s) until: ", last_type == SLEEP_EVAL ? "eval" : "retry");
+		    fprintf(stderr, "Last sleepTight until: ");
 		    showFullTime(stderr, last_sched);
 		    fputc('\n', stderr);
 		    fprintf(stderr, "Last ");
@@ -259,7 +244,7 @@ sleepTight(Task *t, int type)
 		fprintf(stderr, "Harvest children done: ");
 		showFullTime(stderr, cur);
 		fputc('\n', stderr);
-		fprintf(stderr, "Want sleepTight (%s) until: ", type == SLEEP_EVAL ? "eval" : "retry");
+		fprintf(stderr, "Want sleepTight until: ");
 		showFullTime(stderr, sched);
 		fputc('\n', stderr);
 		fprintf(stderr, "This ");
@@ -270,18 +255,24 @@ sleepTight(Task *t, int type)
 	    unrealizenano(delay, &ts);
 	    for (;;) {	/* loop to catch early wakeup from nanosleep */
 		if (ts.tv_sec < 0 || ts.tv_nsec > 999999999) {
-		    fprintf(stderr, "sleepTight: invalid args: %ld %ld\n",
-			    ts.tv_sec, ts.tv_nsec);
+		    fprintf(stderr, "sleepTight: invalid args: %ld.%09ld\n",
+			    (long int)ts.tv_sec, (long int)ts.tv_nsec);
 		    break;
 		}
 		sts = nanosleep(&ts, &tleft);
+		/* deferred signal handling done immediately */
+		if (doexit)
+		    exit(doexit);
+		if (dorotate) {
+		    logRotate();
+		    dorotate = 0;
+		}
 		if (sts == 0 || (sts < 0 && oserror() != EINTR))
 		    break;
 		ts = tleft;
 	    }
 	}
 	last_t = t;
-	last_type = type;
 	last_sched = sched;
     }
 }
@@ -289,6 +280,9 @@ sleepTight(Task *t, int type)
 
 /***********************************************************************
  * ring buffer management
+ *
+ * each element in the ring buffer holds a value or a pointer to a string
+ * value
  ***********************************************************************/
 
 void
@@ -298,14 +292,27 @@ newRingBfr(Expr *x)
     char    *p;
     int     i;
 
-    sz = ((x->sem == SEM_TRUTH) || (x->sem == SEM_CHAR)) ?
-	    sizeof(char) * x->tspan :
-	    sizeof(double) * x->tspan;
+    switch (x->sem) {
+
+    case SEM_BOOLEAN:
+    case SEM_CHAR:
+	    sz = sizeof(char);
+	    break;
+
+    default:
+	    if (x->metrics != NULL && x->metrics->desc.type == PM_TYPE_STRING)
+		sz = sizeof(char *);
+	    else
+		sz = sizeof(double);
+	    break;
+    }
+
+    sz *= x->tspan;
     if (x->ring) free(x->ring);
-    x->ring = alloc(x->nsmpls * sz);
-    p = (char *) x->ring;
+    x->ring = zalloc(x->nsmpls * sz);
+    p = (char *)x->ring;
     for (i = 0; i < x->nsmpls; i++) {
-	x->smpls[i].ptr = (void *) p;
+	x->smpls[i].ptr = (void *)p;
 	p += sz;
     }
 }
@@ -352,7 +359,7 @@ alloc(size_t size)
     void *p;
 
     if ((p = malloc(size)) == NULL) {
-	__pmNoMem("pmie.alloc", size, PM_FATAL_ERR);
+	pmNoMem("pmie.alloc", size, PM_FATAL_ERR);
     }
     return p;
 }
@@ -364,7 +371,7 @@ zalloc(size_t size)
     void *p;
 
     if ((p = calloc(1, size)) == NULL) {
-	__pmNoMem("pmie.zalloc", size, PM_FATAL_ERR);
+	pmNoMem("pmie.zalloc", size, PM_FATAL_ERR);
     }
     return p;
 }
@@ -373,7 +380,7 @@ zalloc(size_t size)
 void *
 aalloc(size_t align, size_t size)
 {
-    void	*p;
+    void	*p = NULL;
     int		sts = 0;
 #ifdef HAVE_POSIX_MEMALIGN
     sts = posix_memalign(&p, align, size);
@@ -387,7 +394,7 @@ aalloc(size_t align, size_t size)
 #endif
 #endif
     if (sts != 0) {
-	__pmNoMem("pmie.aalloc", size, PM_FATAL_ERR);
+	pmNoMem("pmie.aalloc", size, PM_FATAL_ERR);
     }
     return p;
 }
@@ -399,18 +406,18 @@ ralloc(void *p, size_t size)
     void *q;
 
     if ((q = realloc(p, size)) == NULL) {
-	__pmNoMem("pmie.ralloc", size, PM_FATAL_ERR);
+	pmNoMem("pmie.ralloc", size, PM_FATAL_ERR);
     }
     return q;
 }
 
 char *
-sdup(char *p)
+sdup(const char *p)
 {
     char *q;
 
     if ((q = strdup(p)) == NULL) {
-	__pmNoMem("pmie.sdup", strlen(p), PM_FATAL_ERR);
+	pmNoMem("pmie.sdup", strlen(p), PM_FATAL_ERR);
     }
     return q;
 }
@@ -444,7 +451,7 @@ newExpr(int op, Expr *arg1, Expr *arg2,
 	arg = primary(arg1, arg2);
 	x->metrics = arg->metrics;
     }
-    if (sem == SEM_NUMVAR || sem == SEM_NUMCONST || sem == SEM_TRUTH ||
+    if (sem == SEM_NUMVAR || sem == SEM_NUMCONST || sem == SEM_BOOLEAN ||
         sem == SEM_CHAR || sem == SEM_REGEX)
 	x->units = noUnits;
     else {
@@ -476,11 +483,12 @@ newFetch(Host *owner)
 
 
 Host *
-newHost(Task *owner, Symbol name)
+newHost(Task *owner, Symbol name, Symbol conn)
 {
     Host *h = (Host *) zalloc(sizeof(Host));
 
     h->name = symCopy(name);
+    h->conn = symCopy(conn);
     h->task = owner;
     return h;
 }
@@ -504,7 +512,7 @@ agentId(char *name)
 
     if ((sts = pmLookupName(1, &name, &pmid)) < 0) {
 	fprintf(stderr, "%s: agentId: metric %s not found in namespace: %s\n",
-		pmProgname, name, pmErrStr(sts));
+		pmGetProgname(), name, pmErrStr(sts));
 	exit(1);
     }
     return pmid;
@@ -729,17 +737,28 @@ changeSmpls(Expr **p, int nsmpls)
     int    i;
 
     if (nsmpls == x->nsmpls) return;
-    *p = x = (Expr *) ralloc(x, sizeof(Expr) + (nsmpls - 1) * sizeof(Sample));
+    x = (Expr *) ralloc(x, sizeof(Expr) + (nsmpls - 1) * sizeof(Sample));
     x->nsmpls = nsmpls;
     x->nvals = x->tspan * nsmpls;
     x->valid = 0;
     if (x->op == CND_FETCH) {
+	/* node relocated, fix pointers from associated Metric structures */
 	m = x->metrics;
 	for (i = 0; i < x->hdom; i++) {
 	    m->expr = x;
 	    m++;
 	}
     }
+    /*
+     * we have relocated node from *p to x ... need to fix the
+     * parent pointers in any descendent nodes
+     */
+    if (x->arg1 != NULL)
+	x->arg1->parent = x;
+    if (x->arg2 != NULL)
+	x->arg2->parent = x;
+
+    *p = x;
     newRingBfr(x);
 }
 
@@ -782,6 +801,14 @@ instExpr(Expr *x)
 		up = 1;
 		x->units = arg1->units;
 	    }
+	    if (x->op == CND_INSTANT && x->metrics->desc.sem == PM_SEM_COUNTER) {
+		up = 1;
+		x->units.dimTime++;
+	    }
+	    if (x->op == CND_RATE) {
+		up = 1;
+		x->units.dimTime--;
+	    }
 	}
 	else if (!UNITS_UNKNOWN(arg1->units) &&
 		 !UNITS_UNKNOWN(arg2->units)) {
@@ -897,11 +924,6 @@ void dstructInit(void)
     /* not-a-number initialization */
     mynan = zero / zero;
 
-    /* initialize default host */
-    gethostname(localHost, MAXHOSTNAMELEN);
-    localHost[MAXHOSTNAMELEN-1] = '\0';
-    dfltHost = localHost;
-
     /* set up symbol tables */
     symSetTable(&hosts);
     symSetTable(&metrics);
@@ -960,7 +982,7 @@ agentInit(void)
     /* Only load PMNS if it's default and hence not already loaded */
     if (pmnsfile == PM_NS_DEFAULT && (sts = pmLoadNameSpace(pmnsfile)) < 0) {
 	fprintf(stderr, "%s: agentInit: cannot load metric namespace: %s\n",
-		pmProgname, pmErrStr(sts));
+		pmGetProgname(), pmErrStr(sts));
 	exit(1);
     }
 
@@ -1017,6 +1039,10 @@ static struct {
     { cndEq_1_n,	"cndEq_1_n" },
     { cndEq_n_1,	"cndEq_n_1" },
     { cndEq_n_n,	"cndEq_n_n" },
+    { cndEqStr_1_1,	"cndEqStr_1_1" },
+    { cndEqStr_1_n,	"cndEqStr_1_n" },
+    { cndEqStr_n_1,	"cndEqStr_n_1" },
+    { cndEqStr_n_n,	"cndEqStr_n_n" },
     { cndFall_1,	"cndFall_1" },
     { cndFall_n,	"cndFall_n" },
     { cndFetch_1,	"cndFetch_1" },
@@ -1024,20 +1050,22 @@ static struct {
     { cndFetch_n,	"cndFetch_n" },
     { cndGt_1_1,	"cndGt_1_1" },
     { cndGt_1_n,	"cndGt_1_n" },
-    { cndGt_n_1,	"cndGt_n_1" },
-    { cndGt_n_n,	"cndGt_n_n" },
     { cndGte_1_1,	"cndGte_1_1" },
     { cndGte_1_n,	"cndGte_1_n" },
     { cndGte_n_1,	"cndGte_n_1" },
     { cndGte_n_n,	"cndGte_n_n" },
+    { cndGt_n_1,	"cndGt_n_1" },
+    { cndGt_n_n,	"cndGt_n_n" },
+    { cndInstant_1,	"cndInstant_1" },
+    { cndInstant_n,	"cndInstant_n" },
     { cndLt_1_1,	"cndLt_1_1" },
     { cndLt_1_n,	"cndLt_1_n" },
-    { cndLt_n_1,	"cndLt_n_1" },
-    { cndLt_n_n,	"cndLt_n_n" },
     { cndLte_1_1,	"cndLte_1_1" },
     { cndLte_1_n,	"cndLte_1_n" },
     { cndLte_n_1,	"cndLte_n_1" },
     { cndLte_n_n,	"cndLte_n_n" },
+    { cndLt_n_1,	"cndLt_n_1" },
+    { cndLt_n_n,	"cndLt_n_n" },
     { cndMax_host,	"cndMax_host" },
     { cndMax_inst,	"cndMax_inst" },
     { cndMax_time,	"cndMax_time" },
@@ -1054,6 +1082,10 @@ static struct {
     { cndNeq_1_n,	"cndNeq_1_n" },
     { cndNeq_n_1,	"cndNeq_n_1" },
     { cndNeq_n_n,	"cndNeq_n_n" },
+    { cndNeqStr_1_1,	"cndNeqStr_1_1" },
+    { cndNeqStr_1_n,	"cndNeqStr_1_n" },
+    { cndNeqStr_n_1,	"cndNeqStr_n_1" },
+    { cndNeqStr_n_n,	"cndNeqStr_n_n" },
     { cndNot_1,		"cndNot_1" },
     { cndNot_n,		"cndNot_n" },
     { cndOr_1_1,	"cndOr_1_1" },
@@ -1088,7 +1120,7 @@ static struct {
     { SEM_UNKNOWN,	"UNKNOWN" },
     { SEM_NUMVAR,	"NUMVAR" },
     { SEM_NUMCONST,	"NUMCONST" },
-    { SEM_TRUTH,	"TRUTH" },
+    { SEM_BOOLEAN,	"TRUTH" },
     { SEM_CHAR,		"CHAR" },
     { SEM_REGEX,	"REGEX" },
     { PM_SEM_COUNTER,	"COUNTER" },
@@ -1139,7 +1171,7 @@ __dumpExpr(int level, Expr *x)
     else
 	fprintf(stderr, " units=%s\n", pmUnitsStr(&x->units));
     if (x->valid > 0) {
-	if (x->sem == SEM_TRUTH || x->sem == SEM_CHAR ||
+	if (x->sem == SEM_BOOLEAN || x->sem == SEM_CHAR ||
 	    x->sem == SEM_NUMVAR || x->sem == SEM_NUMCONST ||
 	    x->sem == PM_SEM_COUNTER || x->sem == PM_SEM_INSTANT ||
 	    x->sem == PM_SEM_DISCRETE) {
@@ -1152,21 +1184,26 @@ __dumpExpr(int level, Expr *x)
 			    fprintf(stderr, ", ");
 			fprintf(stderr, "{%d} ", k);
 		    }
-		    if (x->sem == SEM_TRUTH) {
+		    if (x->sem == SEM_BOOLEAN) {
 			char 	c = *((char *)x->smpls[j].ptr+k);
-			if ((int)c == TRUE)
+			if ((int)c == B_TRUE)
 			    fprintf(stderr, "true");
-			else if ((int)c == FALSE)
+			else if ((int)c == B_FALSE)
 			    fprintf(stderr, "false");
-			else if ((int)c == DUNNO)
+			else if ((int)c == B_UNKNOWN)
 			    fprintf(stderr, "unknown");
 			else
 			    fprintf(stderr, "bogus (0x%x)", c & 0xff);
 		    }
 		    else if (x->sem == SEM_CHAR) {
+			/* string constant */
 			if (k == 0)
 			    fprintf(stderr, "\"%s\"", (char *)x->smpls[j].ptr);
 		    }
+		    else if (x->metrics != NULL && x->metrics->desc.type == PM_TYPE_STRING) {
+			/* string-valued metric */
+			fprintf(stderr, "\"%s\"", getStringValue(x, k));
+		    }
 		    else {
 			double	v = *((double *)x->smpls[j].ptr+k);
 			int		fp_bad = 0;
@@ -1206,8 +1243,8 @@ __dumpMetric(int level, Metric *m)
     fprintf(stderr, "expr=" PRINTF_P_PFX "%p profile=" PRINTF_P_PFX "%p host=" PRINTF_P_PFX "%p next=" PRINTF_P_PFX "%p prev=" PRINTF_P_PFX "%p\n",
 	m->expr, m->profile, m->host, m->next, m->prev);
     for (i = 0; i < level; i++) fprintf(stderr, ".. ");
-    fprintf(stderr, "metric=%s host=%s conv=%g specinst=%d m_indom=%d\n",
-	symName(m->mname), symName(m->hname), m->conv, m->specinst, m->m_idom);
+    fprintf(stderr, "metric=%s host=%s via=%s conv=%g specinst=%d m_indom=%d\n",
+            symName(m->mname), symName(m->hname), symName(m->hconn), m->conv, m->specinst, m->m_idom);
     if (m->desc.indom != PM_INDOM_NULL) {
 	numinst =  m->specinst == 0 ? m->m_idom : m->specinst;
 	for (j = 0; j < numinst; j++) {
@@ -1287,15 +1324,44 @@ dumpTask(Task *t)
     fprintf(stderr, "  eval time: ");
     showFullTime(stderr, t->eval);
     fputc('\n', stderr);
-    fprintf(stderr, "  retry time: ");
-    showFullTime(stderr, t->retry);
+    fprintf(stderr, "  retry delta: ");
+    if (t->retry == 0)
+	fprintf(stderr, "N/A");
+    else
+	fprintf(stderr, "%d", t->retry);
     fputc('\n', stderr);
     if (t->hosts == NULL)
 	fprintf(stderr, "  host=<null>\n");
     else
-	fprintf(stderr, "  host=%s (%s)\n", symName(t->hosts->name), t->hosts->down ? "down" : "up");
+	fprintf(stderr, "  host=%s via=%s (%s)\n", symName(t->hosts->name), symName(t->hosts->conn), t->hosts->down ? "down" : "up");
     fprintf(stderr, "  rules:\n");
     for (i = 0; i < t->nrules; i++) {
 	fprintf(stderr, "    %s\n", symName(t->rules[i]));
     }
 }
+
+/*
+ * return a pointer to the nth string-value for smpls[0] of an expression
+ */
+char *getStringValue(Expr *x, int nth)
+{
+    if (x->sem == SEM_CHAR) {
+	/* string constant */
+	if (nth == 0)
+	    return (char *)x->smpls[0].ptr;
+	fprintf(stderr, "getStringValue: botch: nth=%d, not 0 for string constant\n", nth);
+	dumpExpr(x);
+	return NULL;
+    }
+    else if ((x->sem == PM_SEM_INSTANT || x->sem == PM_SEM_DISCRETE) &&
+	     x->metrics != NULL && x->metrics->desc.type == PM_TYPE_STRING) {
+	if (0 <= nth && nth < x->nvals)
+	    return *((char **)x->smpls[0].ptr + nth);
+	fprintf(stderr, "getStringValue: botch: nth=%d, not in the range 0..%d for string-valued metric\n", nth, x->nvals);
+	dumpExpr(x);
+	return NULL;
+    }
+    fprintf(stderr, "getStringValue(" PRINTF_P_PFX "%p, %d): botch: no clue what to do here!\n", x, nth);
+    dumpExpr(x);
+    return NULL;
+}
