@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013-2015 Red Hat.
  * Copyright (c) 1995 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -10,10 +11,6 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 /***********************************************************************
@@ -26,9 +23,9 @@
 #include <stddef.h>
 #include <sys/types.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "symbol.h"
-#include "pmiestats.h"
+#include "stats.h"
 
 
 /***********************************************************************
@@ -44,13 +41,13 @@ struct task;
 
 
 /***********************************************************************
- * numeric and truth value constants
+ * (Kleene) 3-valued boolean values
  ***********************************************************************/
 
-typedef char Truth;
-#define FALSE	0
-#define TRUE	1
-#define DUNNO	2
+typedef char Boolean;
+#define B_FALSE 0
+#define B_TRUE 1
+#define B_UNKNOWN 2
 
 extern double	mynan;	/* definitely not-a-number */
 
@@ -65,7 +62,7 @@ typedef unsigned int	TickTime;	/* time counted in deltas */
 typedef double		RealTime;	/* wall clock time or interval */
 #define MINUTE		60		/* one minute of real time */
 #define DELAY_MAX	32		/* maximum initial evaluation delay */
-#define RETRY		5		/* retry interval */
+#define RETRY		5		/* initial retry interval */
 #define DELTA_DFLT	10		/* default sample interval */
 #define DELTA_MIN	0.1		/* minimum sample interval */
 
@@ -140,7 +137,8 @@ typedef struct metric {
     struct metric   *next;	/* fetch/wait list forward pointer */
     struct metric   *prev;	/* fetch/wait list backward pointer */
     Symbol	    mname;	/* metric name */
-    Symbol	    hname;	/* host name */
+    Symbol	    hconn;	/* host connection string (hostspec or archive filename) */
+    Symbol	    hname;	/* (derived) host name */
     pmDesc          desc;       /* pmAPI metric description */
     double	    conv;	/* conversion factor into canonical units */
     int		    specinst;	/* count of specific instances in rule and */
@@ -197,6 +195,7 @@ typedef struct host {
     struct host	    *next;	/* Host list forward pointer */
     struct host	    *prev;	/* Host list backward pointer */
     Symbol          name;       /* host machine */
+    Symbol          conn;       /* host machine connection */
     int	    	    down;	/* host is not delivering metrics */
     Metric	    *waits;	/* wait list of Metrics */
     Metric          *duds;	/* bad Metrics discovered during evaluation */
@@ -211,7 +210,7 @@ typedef struct task {
     RealTime      delta;	/* sample interval */
     TickTime      tick;		/* count up deltas */
     RealTime      eval;		/* scheduled evaluation time */
-    RealTime	  retry;	/* scheduled retry down Hosts and Metrics */
+    int		  retry;	/* delta for retry down Hosts and Metrics */
     int		  nrules;	/* number of rules in this task */
     Symbol	  *rules;	/* array of rules to be evaluated */
     Host          *hosts;	/* fetches to be executed and waiting */
@@ -221,7 +220,7 @@ typedef struct task {
 /* value semantics - as in pmDesc plus following */
 #define SEM_UNKNOWN	0	/* semantics not yet available */
 #define SEM_NUMVAR	10	/* numeric variable value */
-#define SEM_TRUTH	11	/* truth value */
+#define SEM_BOOLEAN	11	/* boolean (3-state) value */
 #define SEM_CHAR	12	/* character (string) */
 #define SEM_NUMCONST	13	/* numeric constant value */
 #define SEM_REGEX	14	/* compiled regular expression */
@@ -233,6 +232,7 @@ typedef int Op;
 #define CND_FETCH	1
 #define CND_DELAY	2
 #define CND_RATE	3
+#define CND_INSTANT	9
 /* arithmetic */
 #define CND_NEG		4
 #define CND_ADD		5
@@ -254,7 +254,9 @@ typedef int Op;
 #define CND_MIN_TIME	21
 /* relational */
 #define CND_EQ		30
+#define CND_EQ_STR	62
 #define CND_NEQ		31
+#define CND_NEQ_STR	63
 #define CND_LT		32
 #define CND_LTE		33
 #define CND_GT		34
@@ -267,6 +269,8 @@ typedef int Op;
 #define CND_OR		44
 #define CND_MATCH	45
 #define CND_NOMATCH	46
+#define CND_RULESET	47
+#define CND_OTHER	48
 /* quantification */
 #define CND_ALL_HOST	50
 #define CND_ALL_INST	51
@@ -318,12 +322,12 @@ void *alloc(size_t);
 void *zalloc(size_t);
 void *ralloc(void *, size_t);
 void *aalloc(size_t, size_t);
-char *sdup(char *);
+char *sdup(const char *);
 
 Expr *newExpr(int, Expr *, Expr *, int, int, int, int, int);
 Profile *newProfile(Fetch *, pmInDom);
 Fetch *newFetch(Host *);
-Host *newHost(Task *, Symbol);
+Host *newHost(Task *, Symbol, Symbol);
 Task *newTask(RealTime, int);
 void newResult(Task *);
 
@@ -352,20 +356,16 @@ void rotate(Expr *);
 Expr *primary(Expr *, Expr *);
 void changeSmpls(Expr **, int);
 void instFetchExpr(Expr *);
+char *getStringValue(Expr *, int);
 
 /***********************************************************************
  * time methods
  ***********************************************************************/
 
-/* convert timeval to RealTime */
-#define realize(t) (1.0e-6 * (RealTime)(t).tv_usec + (RealTime)(t).tv_sec)
-/* convert RealTime to timeval */
-void unrealize(RealTime, struct timeval *);
 RealTime getReal(void);			/* return current time */
 void reflectTime(RealTime);		/* update time vars to reflect now */
-#define SLEEP_EVAL	0
-#define SLEEP_RETRY	1
-void sleepTight(Task *, int);		/* sleep until retry or eval time */
+void sleepTight(Task *);		/* sleep until eval time */
+void logRotate(void);			/* close current, start a new log */
 
 /*
  * diagnostic tracing
@@ -398,23 +398,21 @@ extern char        *pmnsfile;	/* alternate namespace */
 extern Archive	   *archives;	/* archives given on command line */
 extern RealTime	   first;	/* archive starting point */
 extern RealTime	   last;	/* archive end point */
-extern char	   localHost[];	/* local host name */
-extern char	   *dfltHost;	/* default host name */
+extern char	   *dfltHostConn;  /* host connspec or archive path  */
 extern RealTime	   dfltDelta;	/* default sample interval */
-extern char        *startFlag;  /* start time specified? */
-extern char        *stopFlag;   /* end time specified? */
-extern char        *alignFlag;  /* align time specified? */
-extern char        *offsetFlag; /* offset time specified? */
 extern RealTime    runTime;	/* run time interval */
 extern int	   hostZone;	/* timezone from host? */
 extern char	   *timeZone;	/* timezone from command line */
+extern int	   quiet;	/* suppress default diagnostics */
 extern int	   verbose;	/* verbosity 0, 1 or 2 */
 extern int	   interactive;	/* interactive mode, -d */
 extern int	   isdaemon;	/* run as a daemon */
 extern int         agent;	/* secret agent mode? */
 extern int         applet;	/* applet mode? */
 extern int	   dowrap;	/* counter wrap? default no */
-extern int         noDnsFlag;	/* hostname DNS lookup needed? */
+extern int	   doexit;	/* signalled its time to exit */
+extern int	   dorotate;	/* log rotation was requested */
+extern int	   inrun;	/* parsing done, in run() */
 extern pmiestats_t *perf;	/* pmie performance data ptr */
 extern pmiestats_t instrument;	/* pmie performance data struct */
 
@@ -462,5 +460,3 @@ void agentInit(void);		/* initialize evaluation parameters */
 #define UNITS_UNKNOWN(u) (u.dimCount == -1 && u.scaleCount == 7)
 
 #endif /* DSTRUCT_H */
-
-
