@@ -2,6 +2,7 @@
  * pmie.c - performance inference engine
  ***********************************************************************
  *
+ * Copyright (c) 2013-2015,2017 Red Hat.
  * Copyright (c) 1995-2003 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -24,9 +25,9 @@
 
 #include <ctype.h>
 #include <limits.h>
-#include "pmapi.h"
-#include "impl.h"
 #include <sys/stat.h>
+#include "pmapi.h"
+#include "libpcp.h"
 
 #include "dstruct.h"
 #include "stomp.h"
@@ -35,12 +36,6 @@
 #include "eval.h"
 #include "show.h"
 
-#if HAVE_TRACE_BACK_STACK
-#define MAX_PCS 30	/* max callback procedure depth */
-#define MAX_SIZE 48	/* max function name length     */
-#include <libexc.h>
-#endif
-
 
 /***********************************************************************
  * constants
@@ -55,8 +50,8 @@ static char *intro  = "Performance Co-Pilot Inference Engine (pmie), "
 char	*clientid;
 
 static FILE *logfp;
-static char logfile[MAXPATHLEN+1];
-static char perffile[MAXPATHLEN+1];	/* /var/tmp/<pid> file name */
+static char logfile[MAXPATHLEN];
+static char perffile[MAXPATHLEN];	/* /var/tmp/<pid> file name */
 static char *username;
 
 static char menu[] =
@@ -71,46 +66,61 @@ static char menu[] =
 "  h or ?             - print this menu of commands\n"
 "  q                  - quit\n\n";
 
-static char usage[] =
-    "Usage: %s [options] [filename ...]\n\n"
-    "Options:\n"
-    "  -A align     align sample times on natural boundaries\n"
-    "  -a archive   metrics source is a PCP log archive\n"
-    "  -b           one line buffered output stream, stdout on stderr\n"
-    "  -C           parse configuration and exit\n"
-    "  -c filename  configuration file\n"
-    "  -d           interactive debugging mode\n"
-    "  -e           force timestamps to be reported when used with -V, -v or -W\n"
-    "  -f           run in foreground\n"
-    "  -H           do not do a name lookup on the default hostname\n"
-    "  -h host      metrics source is PMCD on host\n"
-    "  -j stompfile stomp protocol (JMS) file [default %s%cconfig%cpmie%cstomp]\n"
-    "  -l logfile   send status and error messages to logfile\n"
-    "  -n pmnsfile  use an alternative PMNS\n"
-    "  -O offset    initial offset into the time window\n"
-    "  -S starttime start of the time window\n"
-    "  -T endtime   end of the time window\n"
-    "  -t interval  sample interval [default 10 seconds]\n"
-    "  -U username  in daemon mode, run as named user [default pcp]\n"
-    "  -V           verbose mode, annotated expression values printed\n"
-    "  -v           verbose mode, expression values printed\n"
-    "  -W           verbose mode, satisfying expression values printed\n"
-    "  -x           run in domain agent mode (summary PMDA)\n"
-    "  -Z timezone  set reporting timezone\n"
-    "  -z           set reporting timezone to local time of metrics source\n";
-
 /***********************************************************************
- * usage message
+ * command line usage
  ***********************************************************************/
 
-static void
-usageMessage(void)
+static int
+override(int opt, pmOptions *opts)
 {
-    int sep = __pmPathSeparator();
-    fprintf(stderr, usage, pmProgname, pmGetConfig("PCP_VAR_DIR"), sep,sep,sep);
-    exit(1);
+    if (opt == 'a' || opt == 'h' || opt == 'H' || opt == 'V')
+	return 1;
+    return 0;
 }
 
+static pmLongOptions longopts[] = {
+    PMAPI_OPTIONS_HEADER("General options"),
+    PMOPT_ALIGN,
+    PMOPT_ARCHIVE,
+    PMOPT_DEBUG,
+    PMOPT_HOST,
+    PMOPT_NAMESPACE,
+    PMOPT_ORIGIN,
+    PMOPT_START,
+    PMOPT_FINISH,
+    PMOPT_INTERVAL,
+    PMOPT_TIMEZONE,
+    PMOPT_HOSTZONE,
+    PMOPT_HELP,
+    PMAPI_OPTIONS_HEADER("Runtime options"),
+    { "check", 0, 'C', 0, "parse configuration and exit" },
+    { "config", 1, 'c', "FILE", "configuration file" },
+    { "interact", 0, 'd', 0, "interactive debugging mode" },
+    { "foreground", 0, 'f', 0, "run in the foreground, not as a daemon" },
+    { "", 0, 'H', NULL }, /* was: no DNS lookup on the default hostname */
+    { "", 1, 'j', "FILE", "stomp protocol (JMS) file" },
+    { "logfile", 1, 'l', "FILE", "send status and error messages to FILE" },
+    { "username", 1, 'U', "USER", "run as named USER in daemon mode [default pcp]" },
+    PMAPI_OPTIONS_HEADER("Reporting options"),
+    { "buffer", 0, 'b', 0, "one line buffered output stream, stdout on stderr" },
+    { "timestamp", 0, 'e', 0, "force timestamps to be reported with -V, -v or -W" },
+    { "quiet", 0, 'q', 0, "quiet mode, default diagnostics suppressed" },
+    { "", 0, 'v', 0, "verbose mode, expression values printed" },
+    { "verbose", 0, 'V', 0, "verbose mode, annotated expression values printed" },
+    { "", 0, 'W', 0, "verbose mode, satisfying expression values printed" },
+    { "secret-applet", 0, 'X', 0, "run in secret applet mode (thin client)" },
+    { "secret-agent", 0, 'x', 0, "run in secret agent mode (summary PMDA)" },
+    PMAPI_OPTIONS_END
+};
+
+static pmOptions opts = {
+    .flags = PM_OPTFLAG_STDOUT_TZ,
+    .short_options = "a:A:bc:CdD:efHh:j:l:n:O:qS:t:T:U:vVWXxzZ:?",
+    .long_options = longopts,
+    .short_usage = "[options] [filename ...]",
+    .override = override,
+};
+
 
 /***********************************************************************
  * interactive commands
@@ -191,49 +201,43 @@ load(char *fname)
     Symbol	s;
     Expr	*d;
     int		sts = 0;
-    int		sep = __pmPathSeparator();
+    int		sep = pmPathSeparator();
     char	config[MAXPATHLEN+1];
 
     /* search for configfile on configuration file path */
     if (fname && access(fname, F_OK) != 0) {
 	sts = oserror();	/* always report the first error */
 	if (__pmAbsolutePath(fname)) {
-	    fprintf(stderr, "%s: cannot access config file %s: %s\n", pmProgname, 
+	    fprintf(stderr, "%s: cannot access config file %s: %s\n", pmGetProgname(), 
 		    fname, strerror(sts));
 	    exit(1);
 	}
-#if PCP_DEBUG
-	else if (pmDebug & DBG_TRACE_APPL0) {
+	else if (pmDebugOptions.appl0) {
 	    fprintf(stderr, "load: cannot access config file %s: %s\n", fname, strerror(sts));
 	}
-#endif
-	snprintf(config, sizeof(config), "%s%c" "config" "%c" "pmie" "%c%s",
+	pmsprintf(config, sizeof(config)-1, "%s%c" "config%c" "pmie%c" "%s",
 		pmGetConfig("PCP_VAR_DIR"), sep, sep, sep, fname);
 	if (access(config, F_OK) != 0) {
 	    fprintf(stderr, "%s: cannot access config file as either %s or %s: %s\n",
-		    pmProgname, fname, config, strerror(sts));
+		    pmGetProgname(), fname, config, strerror(sts));
 	    exit(1);
 	}
-#if PCP_DEBUG
-	else if (pmDebug & DBG_TRACE_APPL0) {
+	else if (pmDebugOptions.appl0) {
 	    fprintf(stderr, "load: using standard config file %s\n", config);
 	}
-#endif
 	fname = config;
     }
-#if PCP_DEBUG
-    else if (pmDebug & DBG_TRACE_APPL0) {
+    else if (pmDebugOptions.appl0) {
 	fprintf(stderr, "load: using config file %s\n",
 		fname == NULL? "<stdin>":fname);
     }
-#endif
 
     if (perf->config[0] == '\0') {	/* keep record of first config */
 	if (fname == NULL)
 	    strcpy(perf->config, "<stdin>");
 	else if (realpath(fname, perf->config) == NULL) {
 	    fprintf(stderr, "%s: failed to resolve realpath for %s: %s\n",
-		    pmProgname, fname, osstrerror());
+		    pmGetProgname(), fname, osstrerror());
 	    exit(1);
 	}
     }
@@ -260,7 +264,7 @@ list(char *name)
 	if ( (s = symLookup(&rules, name)) )
 	    showSyntax(stdout, s);
 	else
-	    printf("%s: error - rule \"%s\" not defined\n", pmProgname, name);
+	    printf("%s: error - rule \"%s\" not defined\n", pmGetProgname(), name);
     }
     else {	/* all rules */
 	t = taskq;
@@ -289,7 +293,7 @@ sublist(char *name)
 	if ( (s = symLookup(&rules, name)) )
 	    showSubsyntax(stdout, s);
 	else
-	    printf("%s: error - rule '%s' not defined\n", pmProgname, name);
+	    printf("%s: error - rule '%s' not defined\n", pmGetProgname(), name);
     }
     else {	/* all rules */
 	t = taskq;
@@ -319,52 +323,58 @@ stopmonitor(void)
 static void
 startmonitor(void)
 {
-    struct hostent	*hep = noDnsFlag ? NULL : gethostbyname(dfltHost);
     void		*ptr;
+    char		*path;
     int			fd;
     char		zero = '\0';
     char		pmie_dir[MAXPATHLEN];
 
     /* try to create the port file directory. OK if it already exists */
-    snprintf(pmie_dir, sizeof(pmie_dir), "%s%c%s",
-	     pmGetConfig("PCP_TMP_DIR"), __pmPathSeparator(), PMIE_SUBDIR);
-    if ( (mkdir2(pmie_dir, S_IRWXU | S_IRWXG | S_IRWXO) < 0) &&
-	 (oserror() != EEXIST) ) {
-	fprintf(stderr, "%s: error creating stats file dir %s: %s\n",
-		pmProgname, pmie_dir, osstrerror());
-	exit(1);
+    pmsprintf(pmie_dir, sizeof(pmie_dir), "%s%c%s",
+	     pmGetConfig("PCP_TMP_DIR"), pmPathSeparator(), PMIE_SUBDIR);
+    if (mkdir2(pmie_dir, S_IRWXU | S_IRWXG | S_IRWXO) < 0) {
+	if (oserror() != EEXIST) {
+	    fprintf(stderr, "%s: warning cannot create stats file dir %s: %s\n",
+		    pmGetProgname(), pmie_dir, osstrerror());
+	}
     }
-
-    chmod(pmie_dir, S_IRWXU | S_IRWXG | S_IRWXO | S_ISVTX);
     atexit(stopmonitor);
 
     /* create and initialize memory mapped performance data file */
-    sprintf(perffile, "%s%c%" FMT_PID, pmie_dir, __pmPathSeparator(), getpid());
+    pmsprintf(perffile, sizeof(perffile),
+		"%s%c%" FMT_PID, pmie_dir, pmPathSeparator(), (pid_t)getpid());
     unlink(perffile);
     if ((fd = open(perffile, O_RDWR | O_CREAT | O_EXCL | O_TRUNC,
 			     S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {
-	fprintf(stderr, "%s: cannot create stats file %s: %s\n",
-		pmProgname, perffile, osstrerror());
-	exit(1);
+	/* cannot create stats file; too bad, so sad, continue on without it */
+	perf = &instrument;
+	return;
     }
     /* seek to struct size and write one zero */
     lseek(fd, sizeof(pmiestats_t)-1, SEEK_SET);
     if (write(fd, &zero, 1) != 1) {
 	fprintf(stderr, "%s: Warning: write failed for stats file %s: %s\n",
-		pmProgname, perffile, osstrerror());
+		pmGetProgname(), perffile, osstrerror());
     }
 
     /* map perffile & associate the instrumentation struct with it */
     if ((ptr = __pmMemoryMap(fd, sizeof(pmiestats_t), 1)) == NULL) {
 	fprintf(stderr, "%s: memory map failed for stats file %s: %s\n",
-		pmProgname, perffile, osstrerror());
-	exit(1);
+		pmGetProgname(), perffile, osstrerror());
+	perf = &instrument;
+    } else {
+	perf = (pmiestats_t *)ptr;
     }
     close(fd);
 
-    perf = (pmiestats_t *)ptr;
-    strcpy(perf->logfile, logfile[0] == '\0'? "<none>" : logfile);
-    strcpy(perf->defaultfqdn, hep == NULL? dfltHost : hep->h_name);
+    path = (logfile[0] == '\0') ? "<none>" : logfile;
+    strncpy(perf->logfile, path, sizeof(perf->logfile));
+    perf->logfile[sizeof(perf->logfile)-1] = '\0';
+    /* Don't try to improvise a current fdqn for "the" pmcd. 
+       It'll be filled in periodically by newContext(). */
+    strncpy(perf->defaultfqdn, "(uninitialized)", sizeof(perf->defaultfqdn));
+    perf->defaultfqdn[sizeof(perf->defaultfqdn)-1] = '\0';
+
     perf->version = 1;
 }
 
@@ -378,14 +388,15 @@ sigintproc(int sig)
 {
     __pmSetSignalHandler(SIGINT, SIG_IGN);
     __pmSetSignalHandler(SIGTERM, SIG_IGN);
-    __pmNotifyErr(LOG_INFO, "%s caught SIGINT or SIGTERM\n", pmProgname);
-    exit(1);
-}
-
-static void
-sigbye(int sig)
-{
-    exit(0);
+    if (pmDebugOptions.desperate)
+	pmNotifyErr(LOG_INFO, "%s caught SIGINT or SIGTERM\n", pmGetProgname());
+    if (inrun)
+	doexit = sig;
+    else {
+	/* for RH BZ 1327226 */
+	fprintf(stderr, "\nInterrupted!\n");
+	exit(1);
+    }
 }
 
 static void
@@ -402,56 +413,44 @@ remap_stdout_stderr(void)
     if ((j = dup(fileno(stderr))) != i)
 	fprintf(stderr, "%s: Warning: failed to link stdout ... "
 			"dup() returns %d, expected %d (stderr=%d)\n",
-			pmProgname, j, i, fileno(stderr));
+			pmGetProgname(), j, i, fileno(stderr));
+}
+
+void
+logRotate(void)
+{
+    FILE *fp;
+    int sts;
+
+    fp = __pmRotateLog(pmGetProgname(), logfile, logfp, &sts);
+    if (sts != 0) {
+	fprintf(stderr, "pmie: PID = %" FMT_PID ", via %s\n\n",
+                (pid_t)getpid(), dfltHostConn);
+	remap_stdout_stderr();
+	logfp = fp;
+    } else {
+	pmNotifyErr(LOG_ERR, "pmie: log rotation failed\n");
+    }
 }
 
 static void
 sighupproc(int sig)
 {
-    FILE *fp;
-    int sts;
-
-    fp = __pmRotateLog(pmProgname, logfile, logfp, &sts);
-    if (sts != 0) {
-	fprintf(stderr, "pmie: PID = %" FMT_PID ", default host = %s\n\n",
-			getpid(), dfltHost);
-	remap_stdout_stderr();
-	logfp = fp;
-    } else {
-	__pmNotifyErr(LOG_ERR, "pmie: log rotation failed\n");
-    }
-}
-
-
-static void
-dotraceback(void)
-{
-#if HAVE_TRACE_BACK_STACK
-    __uint64_t	call_addr[MAX_PCS];
-    char	*call_fn[MAX_PCS];
-    char	names[MAX_PCS][MAX_SIZE];
-    int		res;
-    int		i;
-
-    fprintf(stderr, "\nProcedure call traceback ...\n");
-    for (i = 0; i < MAX_PCS; i++)
-        call_fn[i] = names[i];
-    res = trace_back_stack(MAX_PCS, call_addr, call_fn, MAX_PCS, MAX_SIZE);
-    for (i = 1; i < res; i++)
-    fprintf(stderr, "  " PRINTF_P_PFX "%p [%s]\n", (void *)call_addr[i], call_fn[i]);
-#endif
-    return;
+    __pmSetSignalHandler(SIGHUP, sighupproc);
+   dorotate = 1;
 }
 
 static void
 sigbadproc(int sig)
 {
-    __pmNotifyErr(LOG_ERR, "Unexpected signal %d ...\n", sig);
-    dotraceback();
-    fprintf(stderr, "\nDumping to core ...\n");
-    fflush(stderr);
+    if (pmDebugOptions.desperate) {
+	pmNotifyErr(LOG_ERR, "Unexpected signal %d ...\n", sig);
+	fprintf(stderr, "\nProcedure call traceback ...\n");
+	__pmDumpStack(stderr);
+	fflush(stderr);
+    }
     stopmonitor();
-    abort();
+    _exit(sig);
 }
 
 
@@ -469,12 +468,13 @@ getargs(int argc, char *argv[])
     char		*msg;
     int			checkFlag = 0;
     int			foreground = 0;
-    int			err = 0;
     int			sts;
     int			c;
     int			bflag = 0;
+    int			dfltConn = 0;	/* default context type */
     Archive		*a;
     struct timeval	tv, tv1, tv2;
+
     extern int		showTimeFlag;
     extern int		errs;		/* syntax errors from syntax.c */
 
@@ -483,18 +483,18 @@ getargs(int argc, char *argv[])
     memset(&tv2, 0, sizeof(tv2));
     dstructInit();
 
-    while ((c=getopt(argc, argv, "a:A:bc:CdD:efHh:j:l:n:O:S:t:T:U:vVWXxzZ:?")) != EOF) {
+    while ((c = pmGetOptions(argc, argv, &opts)) != EOF) {
         switch (c) {
 
 	case 'a':			/* archives */
 	    if (dfltConn && dfltConn != PM_CONTEXT_ARCHIVE) {
-		fprintf(stderr, "%s: at most one of -a or -h allowed\n",
-			pmProgname);
-		err++;
+                /* (technically, multiple -a's are allowed.) */
+		pmprintf("%s: at most one of -a or -h allowed\n", pmGetProgname());
+		opts.errors++;
 		break;
 	    }
-	    dfltConn = PM_CONTEXT_ARCHIVE;
-	    subopts = optarg;
+	    dfltConn = opts.context = PM_CONTEXT_ARCHIVE;
+	    subopts = opts.optarg;
 	    for ( ; ; ) {
 		subopt = subopts;
 		subopts = strchr(subopts, ',');
@@ -511,22 +511,17 @@ getargs(int argc, char *argv[])
 	    foreground = 1;
 	    break;
 
-	case 'A': 			/* sample alignment */
-	    alignFlag = optarg;
-	    break;
-
 	case 'b':			/* line buffered, stdout on stderr */
 	    bflag++;
 	    break;
 
 	case 'c': 			/* configuration file */
 	    if (interactive) {
-		fprintf(stderr, "%s: at most one of -c and -d allowed\n",
-			pmProgname);
-		err++;
+		pmprintf("%s: at most one of -c and -d allowed\n", pmGetProgname());
+		opts.errors++;
 		break;
 	    }
-	    configfile = optarg;
+	    configfile = opts.optarg;
 	    break;
 
 	case 'C': 			/* check config and exit */
@@ -535,25 +530,13 @@ getargs(int argc, char *argv[])
 
 	case 'd': 			/* interactive mode */
 	    if (configfile) {
-		fprintf(stderr, "%s: at most one of -c and -d allowed\n",
-			pmProgname);
-		err++;
+		pmprintf("%s: at most one of -c and -d allowed\n", pmGetProgname());
+		opts.errors++;
 		break;
 	    }
 	    interactive = 1;
 	    break;
 
-	case 'D':	/* debug flag */
-	    sts = __pmParseDebug(optarg);
-	    if (sts < 0) {
-		fprintf(stderr, "%s: unrecognized debug flag specification "
-			"(%s)\n", pmProgname, optarg);
-		err++;
-	    }
-	    else
-		pmDebug |= sts;
-	    break;
-
 	case 'e':	/* force timestamps */
 	    showTimeFlag = 1;
 	    break;
@@ -562,68 +545,42 @@ getargs(int argc, char *argv[])
 	    foreground = 1;
 	    break;
 
-	case 'H': 			/* no name lookup on exported host */
-	    noDnsFlag = 1;
+	case 'H': 			/* deprecated: no DNS lookups */
 	    break;
 
 	case 'h': 			/* default host name */
 	    if (dfltConn) {
-		fprintf(stderr, "%s: at most one of -a or -h allowed\n",
-			pmProgname);
-		err++;
+		pmprintf("%s: at most one of -a or -h allowed\n", pmGetProgname());
+		opts.errors++;
 		break;
 	    }
-	    dfltConn = PM_CONTEXT_HOST;
-	    dfltHost = optarg;
+	    dfltConn = opts.context = PM_CONTEXT_HOST;
+	    dfltHostConn = opts.optarg;
 	    break;
 
         case 'j':			/* stomp protocol (JMS) config */
-	    stompfile = optarg;
+	    stompfile = opts.optarg;
 	    break;
 
 	case 'l':			/* alternate log file */
 	    if (commandlog != NULL) {
-		fprintf(stderr, "%s: at most one -l option is allowed\n",
-			pmProgname);
-		err++;
+		pmprintf("%s: at most one -l option is allowed\n", pmGetProgname());
+		opts.errors++;
 		break;
 	    }
-	    commandlog = optarg;
+	    commandlog = opts.optarg;
 	    isdaemon = 1;
 	    break;
 
-        case 'n':			/* alternate namespace file */
-	    pmnsfile = optarg;
-	    break;
-
-	case 'O':			/* position within time window */
-	    offsetFlag = optarg;
-	    break;
-
-	case 'S':			/* start run time */
-	    startFlag = optarg;
-	    break;
-
-	case 't':			/* sample interval */
-	    if (pmParseInterval(optarg, &tv1, &msg) == 0)
-		dfltDelta = realize(tv1);
-	    else {
-		fprintf(stderr, "%s: could not parse -t argument (%s)\n", pmProgname, optarg);
-		fputs(msg, stderr);
-		free(msg);
-		err++;
-	    }
-	    break;
-
-	case 'T':			/* evaluation period */
-	    stopFlag = optarg;
-	    break;
-
 	case 'U': 			/* run as named user */
-	    username = optarg;
+	    username = opts.optarg;
 	    isdaemon = 1;
 	    break;
 
+	case 'q': 			/* suppress default diagnostics */
+	    quiet = 1;
+	    break;
+
 	case 'v': 			/* print values */
 	    verbose = 1;
 	    break;
@@ -647,52 +604,53 @@ getargs(int argc, char *argv[])
 	    verbose = 1;
 	    isdaemon = 1;
 	    break;
-
-	case 'z':			/* timezone from host */
-	    hostZone = 1;
-	    if (timeZone) {
-		fprintf(stderr, "%s: only one of -Z and -z allowed\n",
-			pmProgname);
-		err++;
-	    }
-	    break;
-
-	case 'Z':			/* explicit TZ string */
-	    timeZone = optarg;
-	    if (hostZone) {
-		fprintf(stderr, "%s: only one of -Z and -z allowed\n",
-			pmProgname);
-		err++;
-	    }
-	    break;
-
-	case '?':
-            err++;
 	}
     }
 
-    if (configfile && optind != argc) {
-	fprintf(stderr, "%s: extra filenames cannot be given after using -c\n",
-		pmProgname);
-	err++;
+    if (!opts.errors && configfile && opts.optind != argc) {
+	pmprintf("%s: extra filenames cannot be given after using -c\n",
+		pmGetProgname());
+	opts.errors++;
     }
-    if (bflag && agent) {
-	fprintf(stderr, "%s: the -b and -x options are incompatible\n",
-		pmProgname);
-	err++;
+    if (!opts.errors && bflag && agent) {
+	pmprintf("%s: the -b and -x options are incompatible\n",
+		pmGetProgname());
+	opts.errors++;
+    }
+    if (opts.errors) {
+    	pmUsageMessage(&opts);
+	exit(1);
+    }
+    /* check if archives/hosts available in the environment */
+    if (!dfltConn && opts.narchives) {
+	dfltConn = opts.context = PM_CONTEXT_ARCHIVE;
+	for (c = 0; c < opts.narchives; c++) {
+	    a = (Archive *)zalloc(sizeof(Archive));
+	    a->fname = opts.archives[c];
+	    if (!initArchive(a))
+		exit(1);
+	}
+	foreground = 1;
+    }
+    if (!dfltConn && opts.nhosts) {
+	dfltConn = opts.context = PM_CONTEXT_HOST;
+	dfltHostConn = opts.hosts[c];
     }
-    if (err)
-    	usageMessage();
 
     if (foreground)
 	isdaemon = 0;
 
+    hostZone = opts.tzflag;
+    timeZone = opts.timezone;
+    if (opts.interval.tv_sec || opts.interval.tv_usec)
+	dfltDelta = pmtimevalToReal(&opts.interval);
+
     if (archives || interactive)
 	perf = &instrument;
 
     if (isdaemon) {			/* daemon mode */
 	/* done before opening log to get permissions right */
-	__pmSetProcessIdentity(username);
+	pmSetProcessIdentity(username);
 
 #if defined(HAVE_TERMIO_SIGNALS)
 	signal(SIGTTOU, SIG_IGN);
@@ -706,15 +664,15 @@ getargs(int argc, char *argv[])
     }
     else {
 	/* need to catch these so the atexit() processing is done */
-	__pmSetSignalHandler(SIGINT, sigbye);
-	__pmSetSignalHandler(SIGTERM, sigbye);
+	__pmSetSignalHandler(SIGINT, sigintproc);
+	__pmSetSignalHandler(SIGTERM, sigintproc);
     }
 
     if (commandlog != NULL) {
-	logfp = __pmOpenLog(pmProgname, commandlog, stderr, &sts);
+	logfp = pmOpenLog(pmGetProgname(), commandlog, stderr, &sts);
 	if (realpath(commandlog, logfile) == NULL) {
 	    fprintf(stderr, "%s: cannot find realpath for log %s: %s\n",
-		    pmProgname, commandlog, osstrerror());
+		    pmGetProgname(), commandlog, osstrerror());
 	    exit(1);
 	}
 	__pmSetSignalHandler(SIGHUP, (isdaemon && !agent) ? sighupproc : SIG_IGN);
@@ -728,73 +686,72 @@ getargs(int argc, char *argv[])
     if ((bflag || isdaemon) && !agent)
 	remap_stdout_stderr();
 
-    if (dfltConn == 0) {
-	/* default case, no -a or -h */
-	dfltConn = PM_CONTEXT_HOST;
-	dfltHost = localHost;
+    /* default host from leftmost archive on command line, or from
+       discovery after a brief connection */
+    if (archives) {
+	a = archives;
+	while (a->next)
+	    a = a->next;
+        dfltHostConn = a->fname;
+    } else if (!dfltConn || dfltConn == PM_CONTEXT_HOST) {
+	if (dfltConn == 0)	/* default case, no -a or -h */
+	    dfltHostConn = "local:";
     }
 
     if (!archives && !interactive) {
 	if (commandlog != NULL)
-	    fprintf(stderr, "pmie: PID = %" FMT_PID ", default host = %s\n\n", getpid(), dfltHost);
+            fprintf(stderr, "pmie: PID = %" FMT_PID ", via %s\n\n",
+                    (pid_t)getpid(), dfltHostConn);
 	startmonitor();
     }
 
-    /* default host from leftmost archive on command line */
-    if (archives && dfltHost == localHost) {
-	a = archives;
-	while (a->next)
-	    a = a->next;
-	dfltHost = a->hname;
-    }
-
     /* initialize time */
-    now = archives ? first : getReal() + 1.0;
+    now = archives ? first : getReal();
     zoneInit();
     reflectTime(dfltDelta);
 
     /* parse time window - just to check argument syntax */
-    unrealize(now, &tv1);
-    if (archives)
-	unrealize(last, &tv2);
-    else
+    pmtimevalFromReal(now, &tv1);
+    if (archives) {
+	pmtimevalFromReal(last, &tv2);
+    } else {
 	tv2.tv_sec = INT_MAX;		/* sizeof(time_t) == sizeof(int) */
-    if (pmParseTimeWindow(startFlag, stopFlag, alignFlag, offsetFlag,
+	tv2.tv_usec = 0;
+    }
+    if (pmParseTimeWindow(opts.start_optarg, opts.finish_optarg,
+			  opts.align_optarg, opts.origin_optarg,
                           &tv1, &tv2,
                           &tv, &tv2, &tv1,
 		          &msg) < 0) {
 	fputs(msg, stderr);
         exit(1);
     }
-    start = realize(tv1);
-    stop = realize(tv2);
+    start = pmtimevalToReal(&tv1);
+    stop = pmtimevalToReal(&tv2);
     runTime = stop - start;
 
-    /* initialize PMAPI */
-    if (pmnsfile != PM_NS_DEFAULT && (sts = pmLoadNameSpace(pmnsfile)) < 0) {
-	fprintf(stderr, "%s: pmLoadNameSpace failed: %s\n", pmProgname,
-		pmErrStr(sts));
-	exit(1);
-    }
-
     /* when not in secret agent mode, register client id with pmcd */
     if (!agent)
 	clientid = __pmGetClientId(argc, argv);
 
-    if (!interactive && optind == argc) {	/* stdin or config file */
+    if (!interactive && opts.optind == argc) {	/* stdin or config file */
 	load(configfile);
     }
     else {					/* list of 1/more filenames */
-	while (optind < argc) {
-	    load(argv[optind]);
-	    optind++;
+	while (opts.optind < argc) {
+	    load(argv[opts.optind]);
+	    opts.optind++;
 	}
     }
 
-#if PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL1)
+    if (pmDebugOptions.appl1)
 	dumpRules();
-#endif
+
+    /* really parse time window */
+    if (!archives) {
+	now = getReal();
+	reflectTime(dfltDelta);
+    }
 
     if (checkFlag)
 	exit(errs == 0 ? 0 : 1);	/* exit 1 for syntax errors ...
@@ -819,17 +776,15 @@ getargs(int argc, char *argv[])
     if (agent)
 	agentInit();			/* initialize secret agent stuff */
 
-    /* really parse time window */
-    if (!archives) {
-	now = getReal() + 1.0;
-	reflectTime(dfltDelta);
-    }
-    unrealize(now, &tv1);
-    if (archives)
-	unrealize(last, &tv2);
-    else
+    pmtimevalFromReal(now, &tv1);
+    if (archives) {
+	pmtimevalFromReal(last, &tv2);
+    } else {
 	tv2.tv_sec = INT_MAX;
-    if (pmParseTimeWindow(startFlag, stopFlag, alignFlag, offsetFlag,
+	tv2.tv_usec = 0;
+    }
+    if (pmParseTimeWindow(opts.start_optarg, opts.finish_optarg,
+			  opts.align_optarg, opts.origin_optarg,
 		          &tv1, &tv2,
                           &tv, &tv2, &tv1,
 		          &msg) < 0) {
@@ -838,8 +793,8 @@ getargs(int argc, char *argv[])
     }
 
     /* set run timing window */
-    start = realize(tv1);
-    stop = realize(tv2);
+    start = pmtimevalToReal(&tv1);
+    stop = pmtimevalToReal(&tv2);
     runTime = stop - start;
 }
 
@@ -869,10 +824,8 @@ interact(void)
 	    case 'f':
 		token = scanArg(finger);
 		load(token);
-#if PCP_DEBUG
-		if (pmDebug & DBG_TRACE_APPL1)
+		if (pmDebugOptions.appl1)
 		    dumpRules();
-#endif
 		break;
 
 	    case 'l':
@@ -884,7 +837,7 @@ interact(void)
 		token = scanArg(finger);
 		if (token) {
 		    if (pmParseInterval(token, &tv1, &msg) == 0)
-			runTime = realize(tv1);
+			runTime = pmtimevalToReal(&tv1);
 		    else {
 			fputs(msg, stderr);
 			free(msg);
@@ -905,20 +858,22 @@ interact(void)
 	    case 'S':
 		token = scanArg(finger);
 		if (token == NULL) {
-		    fprintf(stderr, "%s: error - argument required\n", pmProgname);
+		    fprintf(stderr, "%s: error - argument required\n", pmGetProgname());
 		    break;
 		}
-		unrealize(start, &tv1);
-		if (archives)
-		    unrealize(last, &tv2);
-		else
+		pmtimevalFromReal(start, &tv1);
+		if (archives) {
+		    pmtimevalFromReal(last, &tv2);
+		} else {
 		    tv2.tv_sec = INT_MAX;
+		    tv2.tv_usec = 0;
+		}
 		if (__pmParseTime(token, &tv1, &tv2, &tv1, &msg) < 0) {
 		    fputs(msg, stderr);
 		    free(msg);
 		    break;
 		}
-		start = realize(tv1);
+		start = pmtimevalToReal(&tv1);
 		if (archives)
 		    invalidate();
 		break;
@@ -926,7 +881,7 @@ interact(void)
 	    case 'T':
 		token = scanArg(finger);
 		if (token == NULL) {
-		    fprintf(stderr, "%s: error - argument required\n", pmProgname);
+		    fprintf(stderr, "%s: error - argument required\n", pmGetProgname());
 		    break;
 		}
 		if (pmParseInterval(token, &tv1, &msg) < 0) {
@@ -934,7 +889,7 @@ interact(void)
 		    free(msg);
 		    break;
 		}
-		runTime = realize(tv1);
+		runTime = pmtimevalToReal(&tv1);
 		break;
 	    case 'q':
 		quit = 1;
@@ -966,8 +921,7 @@ interact(void)
 int
 main(int argc, char **argv)
 {
-    __pmSetProgname(argv[0]);
-    __pmGetUsername(&username);
+    pmGetUsername(&username);
     setlinebuf(stdout);
 
     /* PCP_COUNTER_WRAP in environment enables "counter wrap" logic */
