@@ -2,6 +2,7 @@
  * pragmatics.c - inference engine pragmatics analysis
  * 
  * Copyright (c) 1995-2003 Silicon Graphics, Inc.  All Rights Reserved.
+ * Copyright (c) 2013-2015 Red Hat, Inc.
  * 
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -25,7 +26,7 @@
 #include <math.h>
 #include <ctype.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "dstruct.h"
 #include "eval.h"
 #include "pragmatics.h"
@@ -35,88 +36,87 @@
 
 extern char	*clientid;
 
-int	dfltConn;	/* default context type */
-
 /* for initialization of pmUnits struct */
 pmUnits noUnits;
 pmUnits countUnits = { .dimCount = 1 };
 
 char *
-findsource(char *host)
+findsource(const char *host, const char *hconn)
 {
     static char	buf[MAXPATHLEN+MAXHOSTNAMELEN+30];
 
-    if (archives) {
-	Archive	*a = archives;
-	while (a) {	/* find archive for host */
-	    if (strcmp(host, a->hname) == 0)
-		break;
-	    a = a->next;
-	}
-	if (a)
-	    snprintf(buf, sizeof(buf), "archive %s (host %s)", a->fname, host);
-	else
-	    snprintf(buf, sizeof(buf), "host %s in unknown archive!", host);
-    }
+    if (archives)
+        pmsprintf(buf, sizeof(buf), "archive %s (host %s)", hconn, host);
     else
-	snprintf(buf, sizeof(buf), "host %s", host);
+	pmsprintf(buf, sizeof(buf), "pmcd %s (host %s)", hconn, host);
 
     return buf;
 }
 
 /***********************************************************************
  * PMAPI context creation & destruction
- ***********************************************************************/
+ ***********************************************************************
+ *
+ * Also, compute the host name from the hconn string,
+ * by replacing given interned symbol if necessary.
+ */
 
-int	/* > 0: context handle,  -1: retry later */
-newContext(char *host)
+int	/* >= 0: context handle,  -1: retry later */
+newContext(Symbol *host, const char *hconn, int is_temp)
 {
-    Archive	*a;
     int		sts = -1;
 
     if (archives) {
-	a = archives;
-	while (a) {	/* find archive for host */
-	    if (strcmp(host, a->hname) == 0)
-		break;
-	    a = a->next;
-	}
-	if (a) {	/* archive found */
-	    if ((sts = pmNewContext(PM_CONTEXT_ARCHIVE, a->fname)) < 0) {
-		fprintf(stderr, "%s: cannot open archive %s\n",
-			pmProgname, a->fname);
-		fprintf(stderr, "pmNewContext: %s\n", pmErrStr(sts));
-		exit(1);
-	    }
-	}
-	else {		/* no archive for host */
-	    fprintf(stderr, "%s: no archive for host %s\n", pmProgname, host);
-	    exit(1);
-	}
+        if ((sts = pmNewContext(PM_CONTEXT_ARCHIVE, hconn)) < 0) {
+            fprintf(stderr, "%s: cannot open archive %s\n",
+                    pmGetProgname(), hconn);
+            fprintf(stderr, "pmNewContext: %s\n", pmErrStr(sts));
+            exit(1);
+        }
     }
-    else if ((sts = pmNewContext(PM_CONTEXT_HOST, host)) < 0) {
-	if (host_state_changed(host, STATE_FAILINIT) == 1) {
+    else if ((sts = pmNewContext(PM_CONTEXT_HOST, hconn)) < 0) {
+	if (host_state_changed(hconn, STATE_FAILINIT) == 1) {
 	    if (sts == -ECONNREFUSED)
 		fprintf(stderr, "%s: warning - pmcd "
-			"on host %s does not respond\n",
-			pmProgname, host);
+			"via %s does not respond\n",
+			pmGetProgname(), hconn);
 	    else if (sts == PM_ERR_PERMISSION)
-		fprintf(stderr, "%s: warning - host %s does not "
+		fprintf(stderr, "%s: warning - pmcd via %s does not "
 			"permit delivery of metrics to the local host\n",
-			pmProgname, host);
+			pmGetProgname(), hconn);
 	    else if (sts == PM_ERR_CONNLIMIT)
 		fprintf(stderr, "%s: warning - pmcd "
-			"on host %s has exceeded its connection limit\n",
-			pmProgname, host);
+			"via %s has exceeded its connection limit\n",
+			pmGetProgname(), hconn);
 	    else
-		fprintf(stderr, "%s: warning - host %s is unreachable\n", 
-			pmProgname, host);
+		fprintf(stderr, "%s: warning - pmcd via %s is unreachable\n", 
+			pmGetProgname(), hconn);
 	}
 	sts = -1;
     }
-    else if (clientid != NULL)
+    else if (is_temp == 0 && clientid != NULL)
 	/* register client id with pmcd */
 	__pmSetClientId(clientid);
+
+    /* Update the host name if not already known; for archives it'll
+       already be known, since it'll have been interred back in fetchExpr() */
+    if (sts >= 0 && host) {
+        const char* host_name = pmGetContextHostName(sts);
+        if (strcmp(symName(*host), host_name)) { /* new or changed? */
+            symFree (*host);
+            *host = symIntern(&hosts, host_name);
+        }
+    }
+
+    /* Update the mmv, to identify the pmcd host we've just connected
+       to.  Note that this can change, as different pmie rules can
+       have different host-domain metrics evaluated at different
+       times. */ 
+    if (*host && perf) {
+        strncpy(perf->defaultfqdn, symName(*host), sizeof(perf->defaultfqdn));
+        perf->defaultfqdn[sizeof(perf->defaultfqdn)-1] = '\0';
+    }
+
     return sts;
 }
 
@@ -234,7 +234,7 @@ findHost(Task *t, Metric *m)
 	h = h->next;
     }
 
-    h = newHost(t, m->hname);	/* add new host */
+    h = newHost(t, m->hname, m->hconn);	/* add new host */
     if (t->hosts) {
 	h->next = t->hosts;
 	t->hosts->prev = h;
@@ -278,7 +278,8 @@ findFetch(Host *h, Metric *m)
     /* create new Fetch bundle */
     if (! f) {
 	f = newFetch(h);
-	if ((f->handle = newContext(symName(h->name))) < 0) {
+        f->handle = newContext(&h->name,symName(h->conn), 0);
+	if (f->handle < 0) {
 	    free(f);
 	    h->down = 1;
 	    return NULL;
@@ -288,18 +289,15 @@ findFetch(Host *h, Metric *m)
 	    int tmp_mode = PM_MODE_INTERP;
 	    getDoubleAsXTB(&h->task->delta, &tmp_ival, &tmp_mode);
 
-	    tv.tv_sec = (time_t)start;
-	    tv.tv_usec = (int)((start - tv.tv_sec) * 1000000.0);
+	    pmtimevalFromReal(start, &tv);
 	    if ((sts = pmSetMode(tmp_mode, &tv, tmp_ival)) < 0) {
-		fprintf(stderr, "%s: pmSetMode failed: %s\n", pmProgname,
+		fprintf(stderr, "%s: pmSetMode failed: %s\n", pmGetProgname(),
 			pmErrStr(sts));
 		exit(1);
 	    }
-#if PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL1) {
+	    if (pmDebugOptions.appl1) {
 		fprintf(stderr, "findFetch: fetch=0x%p host=0x%p delta=%.6f handle=%d\n", f, h, h->task->delta, f->handle);
 	    }
-#endif
 	}
 	f->next = NULL;
 	f->prev = NULL;
@@ -358,7 +356,7 @@ findProfile(Fetch *f, Metric *m)
 
     /* add instances required by Metric to Profile */
     if ((sts = pmUseContext(f->handle)) < 0) {
-	fprintf(stderr, "%s: pmUseContext failed: %s\n", pmProgname,
+	fprintf(stderr, "%s: pmUseContext failed: %s\n", pmGetProgname(),
 		pmErrStr(sts));
 	exit(1);
     }
@@ -370,7 +368,7 @@ findProfile(Fetch *f, Metric *m)
     if (m->specinst == 0 && p->need_all == 0) {
 	sts = pmDelProfile(p->indom, 0, (int *)0);
 	if (sts < 0) {
-	    fprintf(stderr, "%s: pmDelProfile failed: %s\n", pmProgname,
+	    fprintf(stderr, "%s: pmDelProfile failed: %s\n", pmGetProgname(),
 		    pmErrStr(sts));
 	    exit(1);
 	}
@@ -383,7 +381,7 @@ findProfile(Fetch *f, Metric *m)
 	sts = 0;
 
     if (sts < 0) {
-	fprintf(stderr, "%s: pmAddProfile failed: %s\n", pmProgname,
+	fprintf(stderr, "%s: pmAddProfile failed: %s\n", pmGetProgname(),
 		pmErrStr(sts));
 	exit(1);
     }
@@ -412,8 +410,7 @@ bundle(Task *t, Expr *x)
 		waitMetric(m);
 	    m++;
 	}
-#if PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL1) {
+	if (pmDebugOptions.appl1) {
  	    fprintf(stderr, "bundle: task " PRINTF_P_PFX "%p nth=%d prev=" PRINTF_P_PFX "%p next=" PRINTF_P_PFX "%p delta=%.3f nrules=%d\n",
  		t, t->nth, t->prev, t->next, t->delta, t->nrules+1);
 	    __dumpExpr(1, x);
@@ -423,7 +420,6 @@ bundle(Task *t, Expr *x)
 		m++;
 	    }
 	}
-#endif
     }
     else {
 	if (x->arg1) {
@@ -459,7 +455,7 @@ sendDesc(Expr *x, pmValueSet *vset)
 
 	case SEM_NUMVAR:
 	case SEM_NUMCONST:
-	case SEM_TRUTH:
+	case SEM_BOOLEAN:
 	    /* map to a numeric value */
 	    d.type = PM_TYPE_DOUBLE;
 	    d.sem = PM_SEM_INSTANT;
@@ -494,45 +490,51 @@ initArchive(Archive *a)
     struct timeval  tv;
     int		    sts;
     int		    handle;
-    int		    n;
     Archive	    *b;
+    const char	    *tmp;
 
     /* setup temorary context for the archive */
     if ((sts = pmNewContext(PM_CONTEXT_ARCHIVE, a->fname)) < 0) {
 	fprintf(stderr, "%s: cannot open archive %s\n"
 		"pmNewContext failed: %s\n",
-		pmProgname, a->fname, pmErrStr(sts));
+		pmGetProgname(), a->fname, pmErrStr(sts));
 	return 0;
     }
     handle = sts;
 
+    tmp = pmGetContextHostName(handle);
+    if (strlen(tmp) == 0) {
+	fprintf(stderr, "%s: pmGetContextHostName(%d) failed\n",
+	    pmGetProgname(), handle);
+	return 0;
+    }
+    if ((a->hname = strdup(tmp)) == NULL)
+	pmNoMem("host name copy", strlen(tmp)+1, PM_FATAL_ERR);
+
     /* get the goodies from archive label */
     if ((sts = pmGetArchiveLabel(&label)) < 0) {
 	fprintf(stderr, "%s: cannot read label from archive %s\n"
 			"pmGetArchiveLabel failed: %s\n", 
-			pmProgname, a->fname, pmErrStr(sts));
+			pmGetProgname(), a->fname, pmErrStr(sts));
 	pmDestroyContext(handle);
 	return 0;
     }
-    n = (int) strlen(label.ll_hostname);
-    a->hname = (char *) alloc(n + 1);
-    strcpy(a->hname, label.ll_hostname);
-    a->first = realize(label.ll_start);
+    a->first = pmtimevalToReal(&label.ll_start);
     if ((sts = pmGetArchiveEnd(&tv)) < 0) {
 	fprintf(stderr, "%s: archive %s is corrupted\n"
 		"pmGetArchiveEnd failed: %s\n",
-		pmProgname, a->fname, pmErrStr(sts));
+		pmGetProgname(), a->fname, pmErrStr(sts));
 	pmDestroyContext(handle);
 	return 0;
     }
-    a->last = realize(tv);
+    a->last = pmtimevalToReal(&tv);
 
     /* check for duplicate host */
     b = archives;
     while (b) {
 	if (strcmp(a->hname, b->hname) == 0) {
 	    fprintf(stderr, "%s: Error: archive %s not legal - archive %s is already open "
-		    "for host %s\n", pmProgname, a->fname, b->fname, b->hname);
+		    "for host %s\n", pmGetProgname(), a->fname, b->fname, b->hname);
 	    pmDestroyContext(handle);
 	    return 0;
 	}
@@ -560,49 +562,25 @@ zoneInit(void)
 {
     int		sts;
     int		handle = -1;
-    Archive	*a;
 
     if (timeZone) {			/* TZ from timezone string */
 	if ((sts = pmNewZone(timeZone)) < 0)
 	    fprintf(stderr, "%s: cannot set timezone to %s\n"
-		    "pmNewZone failed: %s\n", pmProgname, timeZone,
+		    "pmNewZone failed: %s\n", pmGetProgname(), timeZone,
 		    pmErrStr(sts));
-    }
-    else if (! archives && hostZone) {	/* TZ from live host */
-	if ((handle = pmNewContext(PM_CONTEXT_HOST, dfltHost)) < 0)
+    } 
+    else if (hostZone) { /* TZ from live host or archive */
+	if ((handle = pmNewContext(archives ? PM_CONTEXT_ARCHIVE : PM_CONTEXT_HOST, dfltHostConn)) < 0)
 	    fprintf(stderr, "%s: cannot set timezone from %s\n"
-		    "pmNewContext failed: %s\n", pmProgname,
-		    findsource(dfltHost), pmErrStr(handle));
+		    "pmNewContext failed: %s\n", pmGetProgname(),
+		    dfltHostConn, pmErrStr(handle));
 	else if ((sts = pmNewContextZone()) < 0)
 	    fprintf(stderr, "%s: cannot set timezone from %s\n"
-		    "pmNewContextZone failed: %s\n", pmProgname,
-		    findsource(dfltHost), pmErrStr(sts));
+		    "pmNewContextZone failed: %s\n", pmGetProgname(),
+		    dfltHostConn, pmErrStr(sts));
 	else
-	    fprintf(stdout, "%s: timezone set to local timezone of host %s\n",
-		    pmProgname, dfltHost);
-	if (handle >= 0)
-	    pmDestroyContext(handle);
-    }
-    else if (hostZone) {		/* TZ from an archive */
-	a = archives;
-	while (a) {
-	    if (strcmp(dfltHost, a->hname) == 0)
-		break;
-	    a = a->next;
-	}
-	if (! a)
-	    fprintf(stderr, "%s: no archive supplied for host %s\n",
-		    pmProgname, dfltHost);
-	else if ((handle = pmNewContext(PM_CONTEXT_ARCHIVE, a->fname)) < 0)
-	    fprintf(stderr, "%s: cannot set timezone from %s\npmNewContext failed: %s\n",
-		    pmProgname, findsource(dfltHost), pmErrStr(handle));
-	else if ((sts = pmNewContextZone()) < 0)
-	    fprintf(stderr, "%s: cannot set timezone from %s\n"
-		    "pmNewContextZone failed: %s\n",
-		    pmProgname, findsource(dfltHost), pmErrStr(sts));
-	else
-	    fprintf(stdout, "%s: timezone set to local timezone of host %s\n",
-		    pmProgname, dfltHost);
+	    fprintf(stdout, "%s: timezone set to local timezone from %s\n",
+		    pmGetProgname(), dfltHostConn);
 	if (handle >= 0)
 	    pmDestroyContext(handle);
     }
@@ -648,7 +626,8 @@ scale(pmUnits in)
 int      /* 1: ok, 0: try again later, -1: fail */
 initMetric(Metric *m)
 {
-    char	*hname = symName(m->hname);
+    char	*hname;
+    char	*hconn = symName(m->hconn);
     char	*mname = symName(m->mname);
     char	**inames;
     int		*iids;
@@ -658,15 +637,16 @@ initMetric(Metric *m)
     int		i, j;
 
     /* set up temporary context */
-    if ((handle = newContext(hname)) < 0)
+    if ((handle = newContext(&m->hname, hconn, 1)) < 0)
 	return 0;
+    hname = symName(m->hname);
 
-    host_state_changed(hname, STATE_RECONN);
+    host_state_changed(hconn, STATE_RECONN);
 
     if ((sts = pmLookupName(1, &mname, &m->desc.pmid)) < 0) {
 	fprintf(stderr, "%s: metric %s not in namespace for %s\n"
 		"pmLookupName failed: %s\n",
-		pmProgname, mname, findsource(hname), pmErrStr(sts));
+		pmGetProgname(), mname, findsource(hname, hconn), pmErrStr(sts));
 	ret = 0;
 	goto end;
     }
@@ -675,22 +655,22 @@ initMetric(Metric *m)
     if ((sts = pmLookupDesc(m->desc.pmid, &m->desc)) < 0) {
 	fprintf(stderr, "%s: metric %s not currently available from %s\n"
 		"pmLookupDesc failed: %s\n",
-		pmProgname, mname, findsource(hname), pmErrStr(sts));
+		pmGetProgname(), mname, findsource(hname, hconn), pmErrStr(sts));
 	ret = 0;
 	goto end;
     }
 
-    if (m->desc.type == PM_TYPE_STRING ||
-	m->desc.type == PM_TYPE_AGGREGATE ||
+    if (m->desc.type == PM_TYPE_AGGREGATE ||
 	m->desc.type == PM_TYPE_AGGREGATE_STATIC ||
 	m->desc.type == PM_TYPE_EVENT ||
+	m->desc.type == PM_TYPE_HIGHRES_EVENT ||
 	m->desc.type == PM_TYPE_UNKNOWN) {
-	fprintf(stderr, "%s: metric %s has non-numeric type\n", pmProgname, mname);
+	fprintf(stderr, "%s: metric %s has inappropriate type\n", pmGetProgname(), mname);
 	ret = -1;
     }
     else if (m->desc.indom == PM_INDOM_NULL) {
 	if (m->specinst != 0) {
-	    fprintf(stderr, "%s: metric %s has no instances\n", pmProgname, mname);
+	    fprintf(stderr, "%s: metric %s has no instances\n", pmGetProgname(), mname);
 	    ret = -1;
 	}
 	else
@@ -702,13 +682,13 @@ initMetric(Metric *m)
 	    if ((sts = pmGetInDomArchive(m->desc.indom, &iids, &inames)) < 0) {
 		fprintf(stderr, "Metric %s from %s - instance domain not "
 			"available in archive\npmGetInDomArchive failed: %s\n",
-			mname, findsource(hname), pmErrStr(sts));
+			mname, findsource(hname, hconn), pmErrStr(sts));
 		ret = -1;
 	    }
 	}
 	else if ((sts = pmGetInDom(m->desc.indom, &iids, &inames)) < 0) {
 	    fprintf(stderr, "Instance domain for metric %s from %s not (currently) available\n"
-		    "pmGetIndom failed: %s\n", mname, findsource(hname), pmErrStr(sts));
+		    "pmGetIndom failed: %s\n", mname, findsource(hname, hconn), pmErrStr(sts));
 	    ret = 0;
 	}
 
@@ -735,9 +715,9 @@ initMetric(Metric *m)
 			}
 		    }
 		    if (j == sts) {
-			__pmNotifyErr(LOG_ERR, "metric %s from %s does not "
+			pmNotifyErr(LOG_ERR, "metric %s from %s does not "
 				"(currently) have instance \"%s\"\n",
-				mname, findsource(hname), m->inames[i]);
+                                      mname, findsource(hname, hconn), m->inames[i]);
 			m->iids[i] = PM_IN_NULL;
 			ret = 0;
 		    }
@@ -772,8 +752,7 @@ initMetric(Metric *m)
 		}
 	    }
 
-#if PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL1) {
+	    if (pmDebugOptions.appl1) {
 		int	numinst;
 		fprintf(stderr, "initMetric: %s from %s: instance domain specinst=%d\n",
 			mname, hname, m->specinst);
@@ -788,7 +767,6 @@ initMetric(Metric *m)
 		    fprintf(stderr, " \"%s\"\n", m->inames[i]);
 		}
 	    }
-#endif
 	    if (sts > 0) {
 		/*
 		 * pmGetInDom or pmGetInDomArchive returned some instances
@@ -828,7 +806,8 @@ end:
 int      /* 1: ok, 0: try again later, -1: fail */
 reinitMetric(Metric *m)
 {
-    char	*hname = symName(m->hname);
+    char	*hname;
+    char	*hconn = symName(m->hconn);
     char	*mname = symName(m->mname);
     char	**inames;
     int		*iids;
@@ -838,10 +817,11 @@ reinitMetric(Metric *m)
     int		i, j;
 
     /* set up temporary context */
-    if ((handle = newContext(hname)) < 0)
+    if ((handle = newContext(&m->hname, hconn, 1)) < 0)
 	return 0;
+    hname = symName(m->hname);
 
-    host_state_changed(hname, STATE_RECONN);
+    host_state_changed(hconn, STATE_RECONN);
 
     if ((sts = pmLookupName(1, &mname, &m->desc.pmid)) < 0) {
 	ret = 0;
@@ -854,17 +834,17 @@ reinitMetric(Metric *m)
         goto end;
     }
 
-    if (m->desc.type == PM_TYPE_STRING ||
-	m->desc.type == PM_TYPE_AGGREGATE ||
+    if (m->desc.type == PM_TYPE_AGGREGATE ||
 	m->desc.type == PM_TYPE_AGGREGATE_STATIC ||
 	m->desc.type == PM_TYPE_EVENT ||
+	m->desc.type == PM_TYPE_HIGHRES_EVENT ||
 	m->desc.type == PM_TYPE_UNKNOWN) {
-	fprintf(stderr, "%s: metric %s has non-numeric type\n", pmProgname, mname);
+	fprintf(stderr, "%s: metric %s has inappropriate type\n", pmGetProgname(), mname);
 	ret = -1;
     }
     else if (m->desc.indom == PM_INDOM_NULL) {
 	if (m->specinst != 0) {
-	    fprintf(stderr, "%s: metric %s has no instances\n", pmProgname, mname);
+	    fprintf(stderr, "%s: metric %s has no instances\n", pmGetProgname(), mname);
 	    ret = -1;
 	}
 	else
@@ -931,8 +911,7 @@ reinitMetric(Metric *m)
 		}
 	    }
 
-#if PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL1) {
+	    if (pmDebugOptions.appl1) {
 		int	numinst;
 		fprintf(stderr, "reinitMetric: %s from %s: instance domain specinst=%d\n",
 			mname, hname, m->specinst);
@@ -948,7 +927,6 @@ reinitMetric(Metric *m)
 		    fprintf(stderr, " \"%s\"\n", m->inames[i]);
 		}
 	    }
-#endif
 	    if (sts > 0) {
 		/*
 		 * pmGetInDom or pmGetInDomArchive returned some instances
@@ -998,16 +976,14 @@ reinitMetric(Metric *m)
 		x->op == CND_LT || x->op == CND_LTE ||
 		x->op == CND_GT || x->op == CND_GTE ||
 		x->op == CND_NOT || x->op == CND_AND || x->op == CND_OR ||
-		x->op == CND_RISE || x->op == CND_FALL ||
+		x->op == CND_RISE || x->op == CND_FALL || x->op == CND_INSTANT ||
 		x->op == CND_MATCH || x->op == CND_NOMATCH) {
 		instFetchExpr(x);
 		findEval(x);
-#if PCP_DEBUG
-		if (pmDebug & DBG_TRACE_APPL1) {
+		if (pmDebugOptions.appl1) {
 		    fprintf(stderr, "reinitMetric: re-shaped ...\n");
 		    dumpExpr(x);
 		}
-#endif
 	    }
 	    if (x->parent) {
 		x = x->parent;
@@ -1127,10 +1103,17 @@ taskFetch(Task *t)
 	    if (! h->down) {
 		pmUseContext(f->handle);
 		if ((sts = pmFetch(f->npmids, f->pmids, &f->result)) < 0) {
-		    if (! archives) {
-			__pmNotifyErr(LOG_ERR, "pmFetch from %s failed: %s\n",
+		    if (archives) {
+			if (sts == PM_ERR_LOGREC) {
+			    fprintf(stderr, "%s: pmFetch failed: %s\n", pmGetProgname(),
+				    pmErrStr(sts));
+			    exit(1);
+			}
+		    }
+		    else {
+			pmNotifyErr(LOG_ERR, "pmFetch from %s failed: %s\n",
 				symName(f->host->name), pmErrStr(sts));
-			host_state_changed(symName(f->host->name), STATE_LOSTCONN);
+			host_state_changed(symName(f->host->conn), STATE_LOSTCONN);
 			h->down = 1;
 			mark_all(h);
 		    }
@@ -1169,7 +1152,7 @@ taskFetch(Task *t)
 			    if (m->desc.pmid == r->vset[i]->pmid) {
 				if (r->vset[i]->numval > 0) {
 				    m->vset = r->vset[i];
-				    m->stamp = realize(r->timestamp);
+				    m->stamp = pmtimevalToReal(&r->timestamp);
 				}
 				break;
 			    }
