@@ -14,7 +14,7 @@
 
 #include <ctype.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 
 static int stomp_connect(const char *hostname, int port);
 static void stomp_disconnect(void);
@@ -36,36 +36,80 @@ static char buffer[4096];
 
 static int stomp_connect(const char *hostname, int port)
 {
-    int sts, nodelay = 1;
-    struct linger nolinger = { 1, 0 };
-    struct sockaddr_in myaddr;
-    struct hostent *servinfo;
+    __pmSockAddr *myaddr;
+    __pmHostEnt *servinfo;
+    void *enumIx;
+    struct timeval tv;
+    struct timeval *ptv;
+    __pmFdSet wfds;
+    int ret;
+    int flags = 0;
 
-    if ((servinfo = gethostbyname(hostname)) == NULL)
+    if ((servinfo = __pmGetAddrInfo(hostname)) == NULL)
 	return -1;
 
-    /* socket setup */
-    if ((fd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
-	return -2;
-    if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, /* avoid 200 ms delay */
-		   (char *)&nodelay, (socklen_t)sizeof(nodelay)) < 0) {
-	stomp_disconnect();
-	return -3;
-    }
-    if (setsockopt(fd, SOL_SOCKET, SO_LINGER, /* don't linger on close */
-		   (char *)&nolinger, (socklen_t)sizeof(nolinger)) < 0) {
-	stomp_disconnect();
-	return -4;
-    }
+    fd = -1;
+    enumIx = NULL;
+    for (myaddr = __pmHostEntGetSockAddr(servinfo, &enumIx);
+	 myaddr != NULL;
+	 myaddr = __pmHostEntGetSockAddr(servinfo, &enumIx)) {
+	/* Create a socket */
+	if (__pmSockAddrIsInet(myaddr))
+	    fd = __pmCreateSocket();
+	else if (__pmSockAddrIsIPv6(myaddr))
+	    fd = __pmCreateIPv6Socket();
+	else
+	    fd = -1;
 
-    memset(&myaddr, 0, sizeof(myaddr));
-    myaddr.sin_family = AF_INET;
-    memcpy(&myaddr.sin_addr, servinfo->h_addr, servinfo->h_length);
-    myaddr.sin_port = htons(port);
-    if ((sts = connect(fd, (struct sockaddr *)&myaddr, sizeof(myaddr))) < 0) {
-	stomp_disconnect();
+	if (fd < 0) {
+	    __pmSockAddrFree(myaddr);
+	    continue; /* Try the next address */
+	}
+
+	/* Attempt to connect */
+	flags = __pmConnectTo(fd, myaddr, port);
+	__pmSockAddrFree(myaddr);
+
+	if (flags < 0) {
+	    /*
+	     * Mark failure in case we fall out the end of the loop
+	     * and try next address. fd has been closed in __pmConnectTo().
+	     */
+	    fd = -1;
+	    continue;
+	}
+
+	/* FNDELAY and we're in progress - wait on select */
+	tv.tv_sec = timeout;
+	tv.tv_usec = 0;
+	ptv = (tv.tv_sec || tv.tv_usec) ? &tv : NULL;
+	__pmFD_ZERO(&wfds);
+	__pmFD_SET(fd, &wfds);
+	ret = __pmSelectWrite(fd+1, &wfds, ptv);
+
+	/* Was the connection successful? */
+	if (ret <= 0) {
+	    if (oserror() == EINTR)
+		return -2;
+	    continue;
+	}
+	ret = __pmConnectCheckError(fd);
+	if (ret == 0)
+	    break;
+
+	/* Unsuccessful connection. */
+	__pmCloseSocket(fd);
+	fd = -1;
+    } /* loop over addresses */
+
+    __pmHostEntFree(servinfo);
+
+    if(fd == -1)
+	return -4;
+
+    fd = __pmConnectRestoreFlags(fd, flags);
+    if(fd < 0)
 	return -5;
-    }
 
     return fd;
 }
@@ -84,10 +128,10 @@ static int stomp_read_ack(void)
     nready = select(fd + 1, &readyfds, NULL, NULL, &tv);
     if (nready <= 0) {
 	if (nready == 0)
-	    __pmNotifyErr(LOG_ERR, "Timed out waiting for server %s:%d - %s",
+	    pmNotifyErr(LOG_ERR, "Timed out waiting for server %s:%d - %s",
 				hostname, port, netstrerror());
 	else
-	    __pmNotifyErr(LOG_ERR, "Error waiting for server %s:%d - %s",
+	    pmNotifyErr(LOG_ERR, "Error waiting for server %s:%d - %s",
 				hostname, port, netstrerror());
 	stomp_disconnect();
 	return -1;
@@ -96,7 +140,7 @@ static int stomp_read_ack(void)
     do {
 	sts = recv(fd, buffer, sizeof(buffer), 0);
 	if (sts < 0) {
-	    __pmNotifyErr(LOG_ERR, "Error recving from server %s:%d - %s",
+	    pmNotifyErr(LOG_ERR, "Error recving from server %s:%d - %s",
 				hostname, port, netstrerror());
 	    stomp_disconnect();
 	    return -1;
@@ -116,7 +160,7 @@ static int stomp_write(const char *buffer, int length)
     do {
 	sts = send(fd, buffer, length, 0);
 	if (sts < 0) {
-	    __pmNotifyErr(LOG_ERR, "Write error to JMS server %s:%d - %s",
+	    pmNotifyErr(LOG_ERR, "Write error to JMS server %s:%d - %s",
 			hostname, port, netstrerror());
 	    stomp_disconnect();
 	    return -1;
@@ -135,7 +179,7 @@ static int stomp_authenticate(void)
 
     if (fd < 0)
 	return -1;
-    len = snprintf(buffer, sizeof(buffer),
+    len = pmsprintf(buffer, sizeof(buffer),
 		   "CONNECT\nlogin:%s\npasscode:%s\n\n", username, passcode);
     if (stomp_write(buffer, len) < 0)
 	return -1;
@@ -150,7 +194,7 @@ static int stomp_destination(void)
 
     if (fd < 0)
 	return -1;
-    len = snprintf(buffer, sizeof(buffer),
+    len = pmsprintf(buffer, sizeof(buffer),
 		   "SUB\ndestination:/topic/%s\n\n", topic);
     if (stomp_write(buffer, len) < 0)
 	return -1;
@@ -165,7 +209,7 @@ static int stomp_hello(void)
 
     if (fd < 0)
 	return -1;
-    len = snprintf(buffer, sizeof(buffer), "SEND\ndestination:/topic/%s\n\n"
+    len = pmsprintf(buffer, sizeof(buffer), "SEND\ndestination:/topic/%s\n\n"
 		   "INFO: PMIE: Established initial connection", topic);
     if (stomp_write(buffer, len) < 0)
 	return -1;
@@ -185,7 +229,7 @@ static char *isspace_terminate(char *string)
 {
     int i = 0;
 
-    while (!isspace(string[i++])) /* do nothing */ ;
+    while (!isspace((int)string[i++])) /* do nothing */ ;
     if (i)
 	string[i-1] = '\0';
     return string;
@@ -204,16 +248,16 @@ static void stomp_parse(void)
 {
     char config[MAXPATHLEN+1];
     FILE *f;
-    int sep = __pmPathSeparator();
+    int sep = pmPathSeparator();
 
     if (stompfile)
 	strncat(config, stompfile, sizeof(config)-1);
     else
-	snprintf(config, sizeof(config),
+	pmsprintf(config, sizeof(config),
 		"%s%c" "config" "%c" "pmie" "%c" "stomp",
 		 pmGetConfig("PCP_VAR_DIR"), sep, sep, sep);
     if ((f = fopen(config, "r")) == NULL) {
-	__pmNotifyErr(LOG_ERR, "Cannot open STOMP configuration file %s: %s",
+	pmNotifyErr(LOG_ERR, "Cannot open STOMP configuration file %s: %s",
 			config, osstrerror());
 	exit(1);
     }
@@ -240,13 +284,13 @@ static void stomp_parse(void)
     fclose(f);
 
     if (!hostname)
-	__pmNotifyErr(LOG_ERR, "No host in STOMP config file %s", config);
+	pmNotifyErr(LOG_ERR, "No host in STOMP config file %s", config);
     if (port == -1)
-	__pmNotifyErr(LOG_ERR, "No port in STOMP config file %s", config);
+	pmNotifyErr(LOG_ERR, "No port in STOMP config file %s", config);
     if (!username)
-	__pmNotifyErr(LOG_ERR, "No username in STOMP config file %s", config);
+	pmNotifyErr(LOG_ERR, "No username in STOMP config file %s", config);
     if (!passcode)
-	__pmNotifyErr(LOG_ERR, "No passcode in STOMP config file %s", config);
+	pmNotifyErr(LOG_ERR, "No passcode in STOMP config file %s", config);
     if (port == -1 || !hostname || !username || !passcode)
 	exit(1);
 }
@@ -276,48 +320,48 @@ void stompInit(void)
     }
 
     if (verbose)
-	__pmNotifyErr(LOG_INFO, "Connecting to %s, port %d", hostname, port);
+	pmNotifyErr(LOG_INFO, "Connecting to %s, port %d", hostname, port);
     if (stomp_connect(hostname, port) < 0) {
-	__pmNotifyErr(LOG_ERR, "Could not connect to the message server");
+	pmNotifyErr(LOG_ERR, "Could not connect to the message server");
 	goto disconnect;
     }
 
     if (verbose)
-	__pmNotifyErr(LOG_INFO, "Connected; sending stomp connect message");
+	pmNotifyErr(LOG_INFO, "Connected; sending stomp connect message");
     if (stomp_authenticate() < 0) {
-	__pmNotifyErr(LOG_ERR, "Could not sent STOMP CONNECT frame to server");
+	pmNotifyErr(LOG_ERR, "Could not sent STOMP CONNECT frame to server");
 	goto disconnect;
     }
 
     if (verbose)
-	__pmNotifyErr(LOG_INFO, "Sent; waiting for server ACK");
+	pmNotifyErr(LOG_INFO, "Sent; waiting for server ACK");
     if (stomp_read_ack() < 0) {
-	__pmNotifyErr(LOG_ERR, "Could not read STOMP ACK frame.");
+	pmNotifyErr(LOG_ERR, "Could not read STOMP ACK frame.");
 	goto disconnect;
     }
 
     if (verbose)
-	__pmNotifyErr(LOG_INFO, "ACK; sending initial PMIE topic and hello");
+	pmNotifyErr(LOG_INFO, "ACK; sending initial PMIE topic and hello");
     if (stomp_destination() < 0) {
-	__pmNotifyErr(LOG_ERR, "Could not read TOPIC frame.");
+	pmNotifyErr(LOG_ERR, "Could not read TOPIC frame.");
 	goto disconnect;
     }
     if (stomp_hello() < 0) {
-	__pmNotifyErr(LOG_ERR, "Could not send HELLO frame.");
+	pmNotifyErr(LOG_ERR, "Could not send HELLO frame.");
 	goto disconnect;
     }
 
     if (verbose)
-	__pmNotifyErr(LOG_INFO, "Sent; waiting for server ACK");
+	pmNotifyErr(LOG_INFO, "Sent; waiting for server ACK");
     if (stomp_read_ack() < 0) {
-	 __pmNotifyErr(LOG_ERR, "Could not read STOMP ACK frame");
+	 pmNotifyErr(LOG_ERR, "Could not read STOMP ACK frame");
 	goto disconnect;
     }
 
     if (!firsttime)
-	__pmNotifyErr(LOG_INFO, "Reconnected to STOMP protocol server");
+	pmNotifyErr(LOG_INFO, "Reconnected to STOMP protocol server");
     else if (verbose)
-	__pmNotifyErr(LOG_INFO, "Initial STOMP protocol setup complete");
+	pmNotifyErr(LOG_INFO, "Initial STOMP protocol setup complete");
     firsttime = 0;
     goto finished;
 
@@ -340,7 +384,7 @@ int stompSend(const char *msg)
     if (fd < 0) stompInit();	/* reconnect */
     if (fd < -1) return -1;
 
-    len = snprintf(buffer, sizeof(buffer),
+    len = pmsprintf(buffer, sizeof(buffer),
 		   "SEND\ndestination:/topic/%s\n\n", topic);
     if (stomp_write(buffer, len) < 0)
 	return -1;
