@@ -4,6 +4,7 @@
  ***********************************************************************
  *
  * Copyright (c) 1995 Silicon Graphics, Inc.  All Rights Reserved.
+ * Copyright (c) 2015 Red Hat
  * 
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -17,7 +18,7 @@
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program; if not, write to the Free Software Foundation, Inc.,
- * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 #ifndef SYMBOL_H
 #define SYMBOL_H
@@ -78,10 +79,10 @@ void symSetTable(SymbolTable *);
 void symClearTable(SymbolTable *);
 
 /* convert string to symbol */
-Symbol symIntern(SymbolTable *, char *);
+Symbol symIntern(SymbolTable *, const char *);
 
 /* lookup symbol by name */
-Symbol symLookup(SymbolTable *, char *);
+Symbol symLookup(SymbolTable *, const char *);
 
 /* copy symbol */
 Symbol symCopy(Symbol);
