@@ -11,6 +11,7 @@
 #		when the group was added to the configuration file
 #	delta	delta argument for pmlogger "logging ... on delta" clause
 #
+# Copyright (c) 2014,2016,2017 Red Hat.
 # Copyright (c) 1998,2003 Silicon Graphics, Inc.  All Rights Reserved.
 # 
 # This program is free software; you can redistribute it and/or modify it
@@ -23,83 +24,130 @@
 # or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 # for more details.
 # 
-# You should have received a copy of the GNU General Public License along
-# with this program; if not, write to the Free Software Foundation, Inc.,
-# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 
 # Get standard environment
 . $PCP_DIR/etc/pcp.env
 
+# Clear this part to ensure many short-lived pmprobe children
+# don't waste time analyzing derived metrics.
+PCP_DERIVED_CONFIG=
+export PCP_DERIVED_CONFIG
+
+status=1
+tmp=`mktemp -d /tmp/pcp.XXXXXXXXX` || exit 1
+trap "rm -rf $tmp; exit \$status" 0 1 2 3 15
+#debug# tmp=`pwd`/tmp
 prog=`basename $0`
 
+cat > $tmp/usage << EOF
+# Usage: [options] configfile
+
+Options:
+    -c                    add message and timestamp (not for interactive use)
+    -d=DIR,--groups=DIR   specify path to the pmlogconf groups directory
+    --host
+    -q,--quiet            quiet, suppress logging interval dialog
+    -r,--reprobe          every group reconsidered for inclusion in configfile
+    -v,--verbose          increase diagnostic verbosity
+    --help
+EOF
+
 _usage()
 {
-    echo "Usage: $prog [-qrv] [-d groupsdir] [-h hostname] configfile"
-    sts=1
+    pmgetopt --progname=$prog --config=$tmp/usage --usage
     exit
 }
 
+# Setup the $tmp/pmprobe.out file for use
+_setup()
+{
+    rm -f $tmp/pmprobe.out $tmp/pmprobe.in $tmp/pmprobe
+    find $BASE -type f \
+    | sed \
+	-e '/\/v1.0\//d' \
+    | LC_COLLATE=POSIX sort \
+    | while read tag
+    do
+	if sed 1q <"$tag" | grep '^#pmlogconf-setup 2.0' >/dev/null
+	then
+	    :
+	else
+	    # not one of our group files, skip it ...
+	    continue
+	fi
+	grep "^probe" $tag | $PCP_AWK_PROG '{ print $2 }' >>$tmp/pmprobe.in
+    done
+    sort -u $tmp/pmprobe.in > $tmp/pmprobe
+    [ -z "$HOST" ] && HOST=local:
+    pmprobe -F -h $HOST -v `cat $tmp/pmprobe` >> $tmp/pmprobe.out
+}
+
 quick=false
 pat=''
 prompt=true
 reprobe=false
+autocreate=false
 BASE=''
 HOST=''
 verbose=false
 setupflags=''
 
-while getopts d:h:qrv? c
+ARGS=`pmgetopt --progname=$prog --config=$tmp/usage -- "$@"`
+[ $? != 0 ] && exit 1
+
+eval set -- "$ARGS"
+while [ $# -gt 0 ]
 do
-    case $c
+    case "$1"
     in
-	d)	# base directory for the group files
-		BASE="$OPTARG"
+	-c)	# automated, non-interactive file creation
+		autocreate=true
+		prompt=false
 		;;
 
-	h)	# host to contact for "probe" tests
-		HOST=$OPTARG
+	-d)	# base directory for the group files
+		BASE="$2"
+		shift
 		;;
 
-	q)	# "quick" mode, don't change logging intervals
+	-h)	# host to contact for "probe" tests
+		HOST="$2"
+		shift
+		;;
+
+	-q)	# "quick" mode, don't change logging intervals
 		quick=true
 		;;
 
-	r)	# reprobe
+	-r)	# reprobe
 		reprobe=true
 		;;
 
-	v)	# verbose
+	-v)	# verbose
 		verbose=true
 		setupflags="$setupflags -v"
 		;;
 
-	?)	# eh?
-		_usage
-		# NOTREACHED
+	--)	# end options
+		shift
+		break
+		;;
+
+	-\?)	_usage
 		;;
     esac
+    shift
 done
-shift `expr $OPTIND - 1`
 
-if [ $# -ne 1 ]
-then
-    _usage
-    # NOTREACHED
-fi
+[ $# -eq 1 ] || _usage
 
-if [ ! -d $BASE ]
+if [ -n "$BASE" -a ! -d "$BASE" ]
 then
     echo "$prog: Error: base directory ($BASE) for group files does not exist"
-    sts=1
     exit
 fi
 
-config=$1
-
-tmp=`mktemp -d /var/tmp/pcp.XXXXXXXXX` || exit 1
-sts=0
-trap "rm -rf $tmp; exit \$sts" 0 1 2 3 15
-#debug# tmp=`pwd`/tmp
+config="$1"
 
 # split $tmp/ctl at the line containing the unprocessed tag to
 # produce
@@ -166,18 +214,18 @@ BEGIN						{ out = "'"$tmp/pre"'" }
 /DO NOT UPDATE THE FILE ABOVE/			{ out = "'"$tmp/post"'" }
 						{ print >out }'
 	    mv $tmp/pre $tmp/ctl
-	    [ -z "$HOST" ] && HOST=localhost
-	    if $PCP_BINADM_DIR/pmlogconf-setup -h $HOST $setupflags $BASE/"$tag" 2>$tmp/err >$tmp/out
+	    [ -z "$HOST" ] && HOST=local:
+	    if $PCP_BINADM_DIR/pmlogconf-setup -h $HOST -t $tmp/pmprobe.out $setupflags $BASE/"$tag" 2>$tmp/err >$tmp/out
 	    then
 		:
 	    else
-		echo "$prog: Warning: $BASE/$tag: pmlogconf-setup failed"
+		echo >&2 "$prog: Warning: $BASE/$tag: pmlogconf-setup failed"
 		sts=1
 	    fi
 	    sed -e "s;$BASE/;;" <$tmp/out >$tmp/tmp
 	    [ -s $tmp/err ] && cat $tmp/err
 	    sed -e '/^#+/s/+/?/' <$tmp/tmp >>$tmp/ctl
-	    cat $tmp/post >>$tmp/ctl
+	    [ -s $tmp/post ] && cat $tmp/post >>$tmp/ctl
 	fi
     done
 
@@ -186,16 +234,29 @@ BEGIN						{ out = "'"$tmp/pre"'" }
 	_split
 	[ ! -s $tmp/tag ] && break
 	eval `sed <$tmp/tag -e 's/^#? /tag="/' -e 's/:/" onoff="/' -e 's/:/" delta="/' -e 's/:.*/"/'`
+
+	if [ ! -f $BASE/"$tag" ]
+	then
+	    # the tag file has gone away ...
+	    #
+	    if $autocreate
+	    then
+		echo >&2 "$prog: Warning: cannot find group file ($tag): deleting obsolete group"
+		cat $tmp/head $tmp/tail >$tmp/ctl
+		continue
+	    fi
+	fi
+
 	[ -z "$delta" ] && delta=default
 
 	if $reprobe
 	then
-	    [ -z "$HOST" ] && HOST=localhost
-	    if $PCP_BINADM_DIR/pmlogconf-setup -h $HOST $setupflags $BASE/"$tag" 2>$tmp/err >$tmp/out
+	    [ -z "$HOST" ] && HOST=local:
+	    if $PCP_BINADM_DIR/pmlogconf-setup -h $HOST -t $tmp/pmprobe.out $setupflags $BASE/"$tag" 2>$tmp/err >$tmp/out
 	    then
 		:
 	    else
-		echo "$prog: Warning: $BASE/$tag: pmlogconf-setup failed"
+		echo >&2 "$prog: Warning: $BASE/$tag: pmlogconf-setup failed"
 		sts=1
 	    fi
 	    sed -e "s;$BASE/;;" <$tmp/out >$tmp/tmp
@@ -206,7 +267,7 @@ BEGIN						{ out = "'"$tmp/pre"'" }
 		[ -z "$delta_r" ] && delta_r=default
 		if [ "$tag" != "$tag_r" ]
 		then
-		    echo "Botch: reprobe for $tag found new tag ${tag_r}, no change"
+		    echo >&2 "Botch: reprobe for $tag found new tag ${tag_r}, no change"
 		    cat $tmp/tmp
 		else
 		    if [ "$onoff" = y ]
@@ -214,7 +275,7 @@ BEGIN						{ out = "'"$tmp/pre"'" }
 			# existing y takes precedence
 			if [ "$onoff_r" = x ]
 			then
-			    echo "Warning: reprobe for $tag suggests exclude, keeping current include status"
+			    echo >&2 "Warning: reprobe for $tag suggests exclude, keeping current include status"
 			fi
 		    else
 			onoff=$onoff_r
@@ -234,7 +295,7 @@ BEGIN						{ out = "'"$tmp/pre"'" }
 		    cat $tmp/tail >>$tmp/ctl
 		    continue
 		    ;;
-	    *)	echo "Warning: tag=$tag onoff is illegal ($onoff) ... setting to \"n\""
+	    *)	echo >&2 "Warning: tag=$tag onoff is illegal ($onoff) ... setting to \"n\""
 		    onoff=n
 		    ;;
 	esac
@@ -263,7 +324,7 @@ END		{ printf "desc='"'"'%s'"'"'\n",desc }'`
 			# from migration, silently do nothing
 			;;
 		*)
-			echo "Warning: cannot find group file ($tag) ... no change is possible"
+			echo >&2 "Warning: cannot find group file ($tag): no change is possible"
 			;;
 	    esac
 	    $PCP_AWK_PROG <"$config" >>$tmp/head '
@@ -294,7 +355,14 @@ want == 1 && /^#----/	{ exit }'
 	    #
 	    echo
 	    was_onoff=$onoff
-	    echo "Group: $desc" | fmt -74 | sed -e '1!s/^/       /'
+	    echo "Group: $desc" \
+	    | if [ "$PCP_PLATFORM" = netbsd ]
+	    then
+		fmt -g 74 -m 75
+	    else
+		fmt -w 74
+	    fi \
+	    | sed -e '1!s/^/       /'
 	    while true
 	    do
 		$PCP_ECHO_PROG $PCP_ECHO_N "Log this group? [$onoff] ""$PCP_ECHO_C"
@@ -414,7 +482,24 @@ y         log this group
     done
 }
 
-if [ ! -f "$config" ]
+if $autocreate || $reprobe
+then
+    # Once-off check for pmcd connectivity, to avoid subsequent repeated
+    # failures in pmlogconf-setup (which may take awhile, especially when
+    # the environment is setup with slow/faraway PMCD timeout values).
+    #
+    PMCD="$HOST"
+    [ -z "$PMCD" ] && PMCD=local:
+    if $PCP_BINADM_DIR/pmcd_wait -h "$PMCD" -v 2>$tmp/err
+    then
+	:
+    else
+	sed -e "s/pmcd_wait/$prog/g" < $tmp/err
+	exit
+    fi
+fi
+
+if [ ! -s "$config" ]
 then
     # create a new config file
     #
@@ -422,20 +507,23 @@ then
     if [ ! -f "$config" ]
     then
 	echo "$prog: Error: config file \"$config\" does not exist and cannot be created"
-	sts=1
 	exit
     fi
 
     $PCP_ECHO_PROG "Creating config file \"$config\" using default settings ..."
     prompt=false
     new=true
-    [ -z "$HOST" ] && HOST=localhost
+    [ -z "$HOST" ] && HOST=local:
     [ -z "$BASE" ] && BASE=$PCP_VAR_DIR/config/pmlogconf
+    _setup
 
     cat <<End-of-File >$tmp/in
 #pmlogconf 2.0
 #
 # pmlogger(1) config file created and updated by pmlogconf
+End-of-File
+    $autocreate && echo "# Auto-generated by pmlogconf on:  "`date` >>$tmp/in
+    cat <<End-of-File >>$tmp/in
 #
 # DO NOT UPDATE THE INITIAL SECTION OF THIS FILE.
 # Any changes may be lost the next time pmlogconf is used
@@ -458,11 +546,11 @@ End-of-File
 	    # not one of our group files, skip it ...
 	    continue
 	fi
-	if $PCP_BINADM_DIR/pmlogconf-setup -h $HOST $setupflags "$tag" 2>$tmp/err >$tmp/out
+	if $PCP_BINADM_DIR/pmlogconf-setup -h $HOST -t $tmp/pmprobe.out $setupflags "$tag" 2>$tmp/err >$tmp/out
 	then
 	    :
 	else
-	    echo "$prog: Warning: $BASE/$tag: pmlogconf-setup failed"
+	    echo >&2 "$prog: Warning: $BASE/$tag: pmlogconf-setup failed"
 	    [ -s $tmp/err ] && cat $tmp/err
 	    sts=1
 	fi
@@ -479,6 +567,10 @@ End-of-File
 # It is safe to make additions from here on ...
 #
 
+[access]
+disallow .* : all;
+disallow :* : all;
+allow local:* : enquire;
 End-of-File
 
 else
@@ -508,7 +600,7 @@ s; D1:; disk/percontroller:;
 s; D2:; disk/perdisk:;
 s; D3:; v1.0/D3:;
 s; F0:; filesystem/all:;
-s; F1:; filesystem/xfs-io-irix:;
+s; F1:; filesystem/xfs-io-linux:;
 s; F2:; filesystem/xfs-all:;
 s; F3:; sgi/xlv-activity:;
 s; F4:; sgi/xlv-stripe-io:;
@@ -563,22 +655,19 @@ s; S2:; networking/rpc:;
 	    cp "$config" $tmp/in
 	else
 	    echo "$prog: Error: existing config file \"$config\" is wrong version ($version)"
-	    sts=1
 	    exit
 	fi
     else
 	echo "$prog: Error: existing \"$config\" is not a $prog control file"
-	sts=1
 	exit
     fi
     if [ ! -w "$config" ]
     then
 	echo "$prog: Error: existing config file \"$config\" is not writeable"
-	sts=1
 	exit
     fi
 
-    [ -n "$HOST" ] && echo "$prog: Warning: existing config file, -h $HOST will be ignored"
+    [ -n "$HOST" -a ! $reprobe ] && echo >&2 "$prog: Warning: existing config file, -h $HOST will be ignored"
 
     CBASE=`sed -n -e '/^#+ groupdir /s///p' <$tmp/in`
     if [ -z "$BASE" ]
@@ -587,7 +676,7 @@ s; S2:; networking/rpc:;
     else
 	if [ "$BASE" != "$CBASE" ]
 	then
-	    echo "$prog: Warning: using base directory for group files from command line ($BASE) which is different from that in $config ($CBASE)"
+	    echo >&2 "$prog: Warning: using base directory for group files from command line ($BASE) which is different from that in $config ($CBASE)"
 	fi
     fi
 fi
@@ -641,4 +730,5 @@ else
     fi
 fi
 
-exit 0
+status=0
+exit
