@@ -1,6 +1,7 @@
 /*
  * utils for pmlogextract
  *
+ * Copyright (c) 2017 Red Hat.
  * Copyright (c) 1997-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -16,59 +17,54 @@
 
 #include <assert.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 
 /*
  * raw read of next log record - largely stolen from __pmLogRead in libpcp
  */
 int
-_pmLogGet(__pmLogCtl *lcp, int vol, __pmPDU **pb)
+_pmLogGet(__pmArchCtl *acp, int vol, __pmPDU **pb)
 {
+    __pmLogCtl	*lcp = acp->ac_log;
     int		head;
     int		tail;
     int		sts;
     long	offset;
     char	*p;
     __pmPDU	*lpb;
-    FILE	*f;
+    __pmFILE	*f;
 
     if (vol == PM_LOG_VOL_META)
 	f = lcp->l_mdfp;
     else
-	f = lcp->l_mfp;
+	f = acp->ac_mfp;
 
-    offset = ftell(f);
+    offset = __pmFtell(f);
     assert(offset >= 0);
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_LOG) {
+    if (pmDebugOptions.log) {
 	fprintf(stderr, "_pmLogGet: fd=%d vol=%d posn=%ld ",
-	    fileno(f), vol, offset);
+	    __pmFileno(f), vol, offset);
     }
-#endif
 
 again:
-    sts = (int)fread(&head, 1, sizeof(head), f);
+    sts = (int)__pmFread(&head, 1, sizeof(head), f);
     if (sts != sizeof(head)) {
 	if (sts == 0) {
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_LOG)
+	    if (pmDebugOptions.log)
 		fprintf(stderr, "AFTER end\n");
-#endif
-	    fseek(f, offset, SEEK_SET);
+	    __pmFseek(f, offset, SEEK_SET);
 	    if (vol != PM_LOG_VOL_META) {
-		if (lcp->l_curvol < lcp->l_maxvol) {
-		    if (__pmLogChangeVol(lcp, lcp->l_curvol+1) == 0) {
-			f = lcp->l_mfp;
+		if (acp->ac_curvol < lcp->l_maxvol) {
+		    if (__pmLogChangeVol(acp, acp->ac_curvol+1) == 0) {
+			f = acp->ac_mfp;
 			goto again;
 		    }
 		}
 	    }
 	    return PM_ERR_EOL;
 	}
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LOG)
+	if (pmDebugOptions.log)
 	    fprintf(stderr, "Error: hdr fread=%d %s\n", sts, osstrerror());
-#endif
 	if (sts > 0)
 	    return PM_ERR_LOGREC;
 	else
@@ -76,24 +72,20 @@ again:
     }
 
     if ((lpb = (__pmPDU *)malloc(ntohl(head))) == NULL) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LOG)
+	if (pmDebugOptions.log)
 	    fprintf(stderr, "Error: __pmFindPDUBuf(%d) %s\n",
 		(int)ntohl(head), osstrerror());
-#endif
-	fseek(f, offset, SEEK_SET);
+	__pmFseek(f, offset, SEEK_SET);
 	return -oserror();
     }
 
     lpb[0] = head;
-    if ((sts = (int)fread(&lpb[1], 1, ntohl(head) - sizeof(head), f)) != ntohl(head) - sizeof(head)) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LOG)
+    if ((sts = (int)__pmFread(&lpb[1], 1, ntohl(head) - sizeof(head), f)) != ntohl(head) - sizeof(head)) {
+	if (pmDebugOptions.log)
 	    fprintf(stderr, "Error: data fread=%d %s\n", sts, osstrerror());
-#endif
 	free(lpb);
 	if (sts == 0) {
-	    fseek(f, offset, SEEK_SET);
+	    __pmFseek(f, offset, SEEK_SET);
 	    return PM_ERR_EOL;
 	}
 	else if (sts > 0)
@@ -106,44 +98,39 @@ again:
     p = (char *)lpb;
     memcpy(&tail, &p[ntohl(head) - sizeof(head)], sizeof(head));
     if (head != tail) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_LOG)
+	if (pmDebugOptions.log)
 	    fprintf(stderr, "Error: head-tail mismatch (%d-%d)\n",
 		(int)ntohl(head), (int)ntohl(tail));
-#endif
 	free(lpb);
 	return PM_ERR_LOGREC;
     }
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_LOG) {
+    if (pmDebugOptions.log) {
 	if (vol != PM_LOG_VOL_META || ntohl(lpb[1]) == TYPE_INDOM) {
 	    fprintf(stderr, "@");
 	    if (sts >= 0) {
 		struct timeval	stamp;
-		__pmTimeval		*tvp = (__pmTimeval *)&lpb[vol == PM_LOG_VOL_META ? 2 : 1];
+		pmTimeval		*tvp = (pmTimeval *)&lpb[vol == PM_LOG_VOL_META ? 2 : 1];
 		stamp.tv_sec = ntohl(tvp->tv_sec);
 		stamp.tv_usec = ntohl(tvp->tv_usec);
-		__pmPrintStamp(stderr, &stamp);
+		pmPrintStamp(stderr, &stamp);
 	    }
 	    else
 		fprintf(stderr, "unknown time");
 	}
 	fprintf(stderr, " len=%d (incl head+tail)\n", (int)ntohl(head));
     }
-#endif
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_PDU) {
+    if (pmDebugOptions.pdu) {
 	int		i, j;
 	struct timeval	stamp;
-	__pmTimeval	*tvp = (__pmTimeval *)&lpb[vol == PM_LOG_VOL_META ? 2 : 1];
+	pmTimeval	*tvp = (pmTimeval *)&lpb[vol == PM_LOG_VOL_META ? 2 : 1];
 	fprintf(stderr, "_pmLogGet");
 	if (vol != PM_LOG_VOL_META || ntohl(lpb[1]) == TYPE_INDOM) {
 	    fprintf(stderr, " timestamp=");
 	    stamp.tv_sec = ntohl(tvp->tv_sec);
 	    stamp.tv_usec = ntohl(tvp->tv_usec);
-	    __pmPrintStamp(stderr, &stamp);
+	    pmPrintStamp(stderr, &stamp);
 	}
 	fprintf(stderr, " " PRINTF_P_PFX "%p ... " PRINTF_P_PFX "%p", lpb, &lpb[ntohl(head)/sizeof(__pmPDU) - 1]);
 	fputc('\n', stderr);
@@ -158,8 +145,17 @@ again:
 	}
 	fputc('\n', stderr);
     }
-#endif
 
     *pb = lpb;
     return 0;
 }
+
+pmUnits
+ntoh_pmUnits(pmUnits units)
+{
+    unsigned int x;
+
+    x = ntohl(*(unsigned int *)&units);
+    units = *(pmUnits *)&x;
+    return units;
+}
