@@ -1,6 +1,7 @@
 /*
  * Metric metadata support for pmlogrewrite
  *
+ * Copyright (c) 2017 Red Hat.
  * Copyright (c) 2011 Ken McDonell.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -15,7 +16,7 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "logger.h"
 #include <assert.h>
 
@@ -28,17 +29,13 @@ start_metric(pmID pmid)
     metricspec_t	*mp;
     int			sts;
 
-#if PCP_DEBUG
-    if ((pmDebug & (DBG_TRACE_APPL0 | DBG_TRACE_APPL1)) == (DBG_TRACE_APPL0 | DBG_TRACE_APPL1))
+    if (pmDebugOptions.appl0 && pmDebugOptions.appl1)
 	fprintf(stderr, "start_metric(%s)", pmIDStr(pmid));
-#endif
 
     for (mp = metric_root; mp != NULL; mp = mp->m_next) {
 	if (pmid == mp->old_desc.pmid) {
-#if PCP_DEBUG
-	    if ((pmDebug & (DBG_TRACE_APPL0 | DBG_TRACE_APPL1)) == (DBG_TRACE_APPL0 | DBG_TRACE_APPL1))
+	    if (pmDebugOptions.appl0 && pmDebugOptions.appl1)
 		fprintf(stderr, " -> %s\n", mp->old_name);
-#endif
 	    break;
 	}
     }
@@ -49,7 +46,7 @@ start_metric(pmID pmid)
 	sts = pmNameID(pmid, &name);
 	if (sts < 0) {
 	    if (wflag) {
-		snprintf(mess, sizeof(mess), "Metric %s pmNameID: %s", pmIDStr(pmid), pmErrStr(sts));
+		pmsprintf(mess, sizeof(mess), "Metric %s pmNameID: %s", pmIDStr(pmid), pmErrStr(sts));
 		yywarn(mess);
 	    }
 	    return NULL;
@@ -57,9 +54,10 @@ start_metric(pmID pmid)
 	sts = pmLookupDesc(pmid, &desc);
 	if (sts < 0) {
 	    if (wflag) {
-		snprintf(mess, sizeof(mess), "Metric %s: pmLookupDesc: %s", pmIDStr(pmid), pmErrStr(sts));
+		pmsprintf(mess, sizeof(mess), "Metric %s: pmLookupDesc: %s", pmIDStr(pmid), pmErrStr(sts));
 		yywarn(mess);
 	    }
+	    free(name);
 	    return NULL;
 	}
 
@@ -79,10 +77,8 @@ start_metric(pmID pmid)
 	mp->new_desc = mp->old_desc;
 	mp->flags = 0;
 	mp->ip = NULL;
-#if PCP_DEBUG
-	if ((pmDebug & (DBG_TRACE_APPL0 | DBG_TRACE_APPL1)) == (DBG_TRACE_APPL0 | DBG_TRACE_APPL1))
+	if (pmDebugOptions.appl0 && pmDebugOptions.appl1)
 	    fprintf(stderr, " -> %s [new entry]\n", mp->old_name);
-#endif
     }
 
     return mp;
@@ -109,9 +105,9 @@ _pmUnpackDesc(__pmPDU *pdubuf, pmDesc *desc, int *numnames, char ***names)
     pp = (desc_t *)pdubuf;
     desc->type = ntohl(pp->desc.type);
     desc->sem = ntohl(pp->desc.sem);
-    desc->indom = __ntohpmInDom(pp->desc.indom);
-    desc->units = __ntohpmUnits(pp->desc.units);
-    desc->pmid = __ntohpmID(pp->desc.pmid);
+    desc->indom = ntoh_pmInDom(pp->desc.indom);
+    desc->units = ntoh_pmUnits(pp->desc.units);
+    desc->pmid = ntoh_pmID(pp->desc.pmid);
     *numnames = ntohl(pp->numnames);
     *names = (char **)malloc(*numnames * sizeof(*names[1]));
     if (*names == NULL) {
@@ -156,23 +152,19 @@ do_desc(void)
     char		**names;
     long		out_offset;
 
-    out_offset = ftell(outarch.logctl.l_mdfp);
+    out_offset = __pmFtell(outarch.logctl.l_mdfp);
     _pmUnpackDesc(inarch.metarec, &desc, &numnames, &names);
 
     for (mp = metric_root; mp != NULL; mp = mp->m_next) {
 	if (desc.pmid != mp->old_desc.pmid || mp->flags == 0)
 	    continue;
 	if (mp->flags & METRIC_DELETE) {
-#if PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL1)
+	    if (pmDebugOptions.appl1)
 		fprintf(stderr, "Delete: pmDesc for %s\n", pmIDStr(desc.pmid));
-#endif
 	    goto done;
 	}
-#if PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL1)
+	if (pmDebugOptions.appl1)
 	    fprintf(stderr, "Rewrite: pmDesc for %s\n", pmIDStr(desc.pmid));
-#endif
 	if (mp->flags & METRIC_CHANGE_PMID)
 	    desc.pmid = mp->new_desc.pmid;
 	if (mp->flags & METRIC_CHANGE_NAME) {
@@ -190,7 +182,7 @@ do_desc(void)
 	    }
 	    if (i == numnames) {
 		fprintf(stderr, "%s: Botch: old name %s not found in list of %d names for pmid %s ...",
-			pmProgname, mp->old_name, numnames, pmIDStr(mp->old_desc.pmid));
+			pmGetProgname(), mp->old_name, numnames, pmIDStr(mp->old_desc.pmid));
 		for (i = 0; i < numnames; i++) {
 		    if (i > 0) fputc(',', stderr);
 		    fprintf(stderr, " %s", names[i]);
@@ -210,16 +202,14 @@ do_desc(void)
 	    desc.units = mp->new_desc.units;	/* struct assignment */
 	break;
     }
-    if ((sts = __pmLogPutDesc(&outarch.logctl, &desc, numnames, names)) < 0) {
+    if ((sts = __pmLogPutDesc(&outarch.archctl, &desc, numnames, names)) < 0) {
 	fprintf(stderr, "%s: Error: __pmLogPutDesc: %s (%s): %s\n",
-		pmProgname, names[0], pmIDStr(desc.pmid), pmErrStr(sts));
+		pmGetProgname(), names[0], pmIDStr(desc.pmid), pmErrStr(sts));
 	abandon();
 	/*NOTREACHED*/
     }
-#if PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
+    if (pmDebugOptions.appl0)
 	fprintf(stderr, "Metadata: write PMID %s @ offset=%ld\n", pmIDStr(desc.pmid), out_offset);
-#endif
 
 done:
     for (i = 0; i < numnames; i++)
