@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2014,2016 Red Hat.
  * Copyright (c) 1995-2001,2003 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -10,48 +11,51 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include <math.h>
 #include <stdarg.h>
 #include <limits.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 
-static void
-usage(void)
-{
-    fprintf(stderr,
-"Usage: %s [options] archive [metricname ...]\n\n"
-"Options:\n"
-"  -a            print all information (equivalent to -blmMy)\n"
-"  -b            print both stochastic and time averages for counter metrics\n"
-"  -B nbins      print value distribution across a number of bins\n"
-"  -f            print using \"spreadsheet\" format (tab delimited fields)\n"
-"  -F            print using \"spreadsheet\" format (comma separated values)\n"
-"  -i            also print timestamp for minimum value\n"
-"  -I            also print timestamp for maximum value\n"
-"  -l            also print the archive label\n"
-"  -m            also print minimum values\n"
-"  -M            also print maximum values\n"
-"  -n pmnsfile   use an alternative PMNS\n"
-"  -N            suppress warnings from individual archive fetches [default]\n"
-"  -p precision  number of digits to display after the decimal point\n"
-"  -S starttime  start of the time window\n"
-"  -T endtime    end of the time window\n"
-"  -v            verbose, enable warnings from individual archive fetches\n"
-"  -x            print only stochastic averages for counter metrics\n"
-"  -y            print sample count for each metric\n"
-"  -z            set reporting timezone to local time of metrics source\n"
-"  -Z timezone   set reporting timezone\n",
-	pmProgname);
-    exit(1);
-}
+static pmLongOptions longopts[] = {
+    PMAPI_OPTIONS_HEADER("Options"),
+    PMOPT_DEBUG,
+    { "all", 0, 'a', 0, "print all information (equivalent to -blmMy)" },
+    { "", 0, 'b', 0, "print both stochastic and time averages for counter metrics" },
+    { "bins", 1, 'B', "N", "print value distribution across a number of bins" },
+    { "", 0, 'f', 0, "print using \"spreadsheet\" format (tab delimited fields)" },
+    { "", 0, 'F', 0, "print using \"spreadsheet\" format (comma separated values)" },
+    { "header", 0, 'H', 0, "print one-line header at start showing each column" },
+    { "mintime", 0, 'i', 0, "also print timestamp for minimum value" },
+    { "maxtime", 0, 'I', 0, "also print timestamp for maximum value" },
+    { "label", 0, 'l', 0, "also print the archive label and time window" },
+    { "minimum", 0, 'm', 0, "also print minimum value" },
+    { "maximum", 0, 'M', 0, "also print maximum value" },
+    PMOPT_NAMESPACE,
+    { "", 0, 'N', 0, "suppress warnings from individual archive fetches (default)" },
+    { "precision", 0, 'p', 0, "number of digits to display after the decimal point" },
+    PMOPT_START,
+    PMOPT_FINISH,
+    { "verbose", 0, 'v', 0, "verbose, enable warnings from individual archive fetches" },
+    { "", 0, 'x', 0, "print only stochastic averages for counter metrics" },
+    { "samples", 0, 'y', "print sample count for each metric" },
+    PMOPT_TIMEZONE,
+    PMOPT_HOSTZONE,
+    PMOPT_VERSION,
+    PMOPT_HELP,
+    PMAPI_OPTIONS_END
+};
 
+static int override(int, pmOptions *);
+static pmOptions opts = {
+    .flags = PM_OPTFLAG_DONE | PM_OPTFLAG_BOUNDARIES | PM_OPTFLAG_STDOUT_TZ,
+    .short_options = "abB:D:fFHiIlmMNn:p:rsS:T:vVxyzZ:?",
+    .long_options = longopts,
+    .short_usage = "[options] archive [metricname ...]",
+    .override = override,
+};
 
 typedef struct {
     int			inst;
@@ -100,13 +104,7 @@ static unsigned int	nbins;		/* number of distribution bins */
 static unsigned int	precision = 3;	/* number of digits after "." */
 
 /* time window stuff */
-static int		sflag;
-static int		tflag;
 static int		dayflag;
-static struct timeval	logstart;
-static struct timeval   logend;
-static struct timeval	windowstart;
-static struct timeval   windowend;
 static char		timebuf[32];		/* for pmCtime result + .xxx */
 
 /* duration of log */
@@ -121,12 +119,7 @@ tsub(struct timeval *a, struct timeval *b)
 {
     if ((a == NULL) || (b == NULL))
 	return -1;
-    a->tv_usec -= b->tv_usec;
-    if (a->tv_usec < 0) {
-	a->tv_usec += 1000000;
-	a->tv_sec--;
-    }
-    a->tv_sec -= b->tv_sec;
+    pmtimevalDec(a, b);
     if (a->tv_sec < 0) {
 	/* clip negative values at zero */
 	a->tv_sec = 0;
@@ -140,37 +133,27 @@ tadd(struct timeval *a, struct timeval *b)
 {
     if ((a == NULL) || (b == NULL))
 	return -1;
-    a->tv_usec += b->tv_usec;
-    if (a->tv_usec > 1000000) {
-	a->tv_usec -= 1000000;
-	a->tv_sec++;
-    }
-    a->tv_sec += b->tv_sec;
+    pmtimevalInc(a, b);
     return 0;
 }
 
-static double
-tosec(struct timeval t)
-{
-    return t.tv_sec + (t.tv_usec / 1000000.0);
-}
-
 static void
 pmiderr(pmID pmid, const char *msg, ...)
 {
     if (warnflag && __pmHashSearch(pmid, &errlist) == NULL) {
 	va_list	arg;
-	char	*mname = NULL;
-	static char *unknown = "(cannot find metric name) ";
+	int	numnames;
+	char	**names;
 
-	pmNameID(pmid, &mname);
-	if (!mname) mname = unknown;
-	fprintf(stderr, "%s: %s(%s) - ", pmProgname, mname, pmIDStr(pmid));
+	numnames = pmNameAll(pmid, &names);
+	fprintf(stderr, "%s: ", pmGetProgname());
+	__pmPrintMetricNames(stderr, numnames, names, " or ");
+	fprintf(stderr, "(%s) - ", pmIDStr(pmid));
 	va_start(arg, msg);
 	vfprintf(stderr, msg, arg);
 	va_end(arg);
 	__pmHashAdd(pmid, NULL, &errlist);
-	if (mname && mname != unknown) free(mname);
+	if (numnames > 0) free(names);
     }
 }
 
@@ -181,17 +164,55 @@ printstamp(struct timeval *stamp, int delimiter)
 	char	*ddmm;
 	char	*yr;
 
-	ddmm = pmCtime(&stamp->tv_sec, timebuf);
+	ddmm = pmCtime((const time_t *)&stamp->tv_sec, timebuf);
 	ddmm[10] = ' ';
 	ddmm[11] = '\0';
 	yr = &ddmm[20];
 	printf("%c'%s", delimiter, ddmm);
-	__pmPrintStamp(stdout, stamp);
+	pmPrintStamp(stdout, stamp);
 	printf(" %4.4s\'", yr);
     }
     else {
 	printf("%c", delimiter);
-	__pmPrintStamp(stdout, stamp);
+	pmPrintStamp(stdout, stamp);
+    }
+}
+
+static void
+printlabel(void)
+{
+    pmLogLabel  label;
+    char        *ddmm;
+    char        *yr;
+    int         sts;
+
+    if ((sts = pmGetArchiveLabel(&label)) < 0) {
+	fprintf(stderr, "%s: Cannot get archive label record: %s\n",
+		pmGetProgname(), pmErrStr(sts));
+	exit(1);
+    }
+
+    printf("Log Label (Log Format Version %d)\n", label.ll_magic & 0xff);
+    printf("Performance metrics from host %s\n", label.ll_hostname);
+
+    ddmm = pmCtime((const time_t *)&opts.start.tv_sec, timebuf);
+    ddmm[10] = '\0';
+    yr = &ddmm[20];
+    printf("  commencing %s ", ddmm);
+    pmPrintStamp(stdout, &opts.start);
+    printf(" %4.4s\n", yr);
+
+    if (opts.finish.tv_sec == INT_MAX) {
+	/* pmGetArchiveEnd() failed! */
+	printf("  ending     UNKNOWN\n");
+    }
+    else {
+	ddmm = pmCtime((const time_t *)&opts.finish.tv_sec, timebuf);
+	ddmm[10] = '\0';
+	yr = &ddmm[20];
+	printf("  ending     %s ", ddmm);
+	pmPrintStamp(stdout, &opts.finish);
+	printf(" %4.4s\n", yr);
     }
 }
 
@@ -238,11 +259,11 @@ printsummary(const char *name)
     /* cast away const, pmLookupName should never modify name */
     if ((sts = pmLookupName(1, (char **)&name, &pmid)) < 0) {
 	fprintf(stderr, "%s: failed to lookup metric name (pmid=%s): %s\n",
-		pmProgname, pmIDStr(pmid), pmErrStr(sts));
+		pmGetProgname(), name, pmErrStr(sts));
 	return;
     }
 
-    /* lookup id, & print values according to set flags */
+    /* lookup using pmid, print values according to set flags */
     if ((hptr = __pmHashSearch(pmid, &hashlist)) != NULL) {
 	avedata = (aveData*)hptr->data;
 	for (i = 0; i < avedata->listsize; i++) {
@@ -253,17 +274,17 @@ printsummary(const char *name)
 		struct timeval	timediff;
 
 		/* extend discrete metrics to the archive end */
-		timediff = logend;
+		timediff = opts.finish;
 		tsub(&timediff, &instdata->lasttime);
 		val = instdata->lastval;
 		instdata->stocave += val;
-		instdata->timeave += val*tosec(timediff);
-		instdata->lasttime = logend;
+		instdata->timeave += val*pmtimevalToReal(&timediff);
+		instdata->lasttime = opts.finish;
 		instdata->count++;
 	    }
 	    metrictimespan = instdata->lasttime;
 	    tsub(&metrictimespan, &instdata->firsttime);
-	    metricspan = tosec(metrictimespan);
+	    metricspan = pmtimevalToReal(&metrictimespan);
 	    /* counter metric doesn't cover 90% of log */
 	    star = (avedata->desc.sem == PM_SEM_COUNTER && metricspan / logspan <= 0.1);
 
@@ -393,24 +414,24 @@ newHashInst(pmValue *vp,
 
     if ((sts = pmExtractValue(valfmt, vp, avedata->desc.type, &av, PM_TYPE_DOUBLE)) < 0) {
 	pmiderr(avedata->desc.pmid, "failed to extract value: %s\n", pmErrStr(sts));
-	fprintf(stderr, "%s: possibly corrupt archive?\n", pmProgname);
+	fprintf(stderr, "%s: possibly corrupt archive?\n", pmGetProgname());
 	exit(1);
     }
     size = (pos+1) * sizeof(instData *);
     avedata->instlist = (instData **) realloc(avedata->instlist, size);
     if (avedata->instlist == NULL)
-	__pmNoMem("newHashInst.instlist", size, PM_FATAL_ERR);
+	pmNoMem("newHashInst.instlist", size, PM_FATAL_ERR);
     size = sizeof(instData);
     avedata->instlist[pos] = instdata = (instData *) malloc(size);
     if (instdata == NULL)
-	__pmNoMem("newHashInst.instlist[inst]", size, PM_FATAL_ERR);
+	pmNoMem("newHashInst.instlist[inst]", size, PM_FATAL_ERR);
     if (nbins == 0)
 	instdata->bin = NULL;
     else {	/* we are doing binning ... make space for the bins */
 	size = nbins * sizeof(unsigned int);
 	instdata->bin = (unsigned int *)malloc(size);
 	if (instdata->bin == NULL)
-	    __pmNoMem("newHashInst.instlist[inst].bin", size, PM_FATAL_ERR);
+	    pmNoMem("newHashInst.instlist[inst].bin", size, PM_FATAL_ERR);
 	memset(instdata->bin, 0, size);
     }
     instdata->inst = vp->inst;
@@ -441,15 +462,15 @@ newHashInst(pmValue *vp,
     instdata->firsttime = *timestamp;
     instdata->lasttime = *timestamp;
     avedata->listsize++;
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0) {
-	char	*name = NULL;
-	pmNameID(avedata->desc.pmid, &name);
-	fprintf(stderr, "%s Initially - min/max=%f/%f\n", name,
+    if (pmDebugOptions.appl0) {
+	int	numnames;
+	char	**names;
+	numnames = pmNameAll(avedata->desc.pmid, &names);
+	__pmPrintMetricNames(stderr, numnames, names, " or ");
+	fprintf(stderr, " Initially - min/max=%f/%f\n",
 		instdata->min, instdata->max);
-	if (name) free(name);
+	if (numnames > 0) free(names);
     }
-#endif
 }
 
 static void
@@ -505,17 +526,17 @@ findbin(pmID pmid, double val, double min, double max)
 	bound = next;
     }
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0) {
-	char *mname = NULL;
-	pmNameID(pmid, &mname);
-	fprintf(stderr, "%s selected bin %u/%u (val=%.*f, min=%.*f, max=%.*f)\n",
-		mname, index, nbins, (int)precision, val, (int)precision,
+    if (pmDebugOptions.appl0) {
+	int	numnames;
+	char	**names;
+	numnames = pmNameAll(pmid, &names);
+	__pmPrintMetricNames(stderr, numnames, names, " or ");
+	fprintf(stderr, " selected bin %u/%u (val=%.*f, min=%.*f, max=%.*f)\n",
+		index, nbins, (int)precision, val, (int)precision,
 		min, (int)precision, max);
-	if (mname) free(mname);
+	if (numnames > 0) free(names);
 	if (index >= nbins) exit(1);
     }
-#endif
     return index;
 }
 
@@ -533,12 +554,10 @@ markrecord(pmResult *result)
     double		val;
     struct timeval	timediff;
 
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0) {
+    if (pmDebugOptions.appl0) {
 	printstamp(&result->timestamp, '\n');
 	printf(" - mark record\n\n");
     }
-#endif
     for (i = 0; i < hashlist.hsize; i++) {
 	for (hptr = hashlist.hash[i]; hptr != NULL; hptr = hptr->next) {
 	    avedata = (aveData *)hptr->data;
@@ -550,7 +569,7 @@ markrecord(pmResult *result)
 		    tsub(&timediff, &instdata->lasttime);
 		    val = instdata->lastval;
 		    instdata->stocave += val;
-		    instdata->timeave += val*tosec(timediff);
+		    instdata->timeave += val*pmtimevalToReal(&timediff);
 		    instdata->lasttime = result->timestamp;
 		    instdata->count++;
 		}
@@ -592,6 +611,7 @@ calcbinning(pmResult *result)
 	if ((hptr = __pmHashSearch(vsp->pmid, &hashlist)) != NULL) {
 	    avedata = (aveData *)hptr->data;
 	    for (j = 0; j < vsp->numval; j++) {	/* iterate thro result values */
+		int	fp_bad;
 		vp = &vsp->vlist[j];
 		k = j;	/* index into stored inst list, result may differ */
 		if ((vsp->numval > 1) || (avedata->desc.indom != PM_INDOM_NULL)) {
@@ -617,7 +637,15 @@ calcbinning(pmResult *result)
 		    pmiderr(avedata->desc.pmid, "failed to extract value: %s\n", pmErrStr(sts));
 		    continue;
 		}
-		if (isnan(av.d))
+		fp_bad = 0;
+#ifdef HAVE_FPCLASSIFY
+		fp_bad = fpclassify(av.d) == FP_NAN;
+#else
+#ifdef HAVE_ISNAN
+		fp_bad = isnan(av.d);
+#endif
+#endif
+		if (fp_bad)
 		    continue;
 
 		/* reset values from first pass needed in this second pass */
@@ -639,7 +667,7 @@ calcbinning(pmResult *result)
 
 		timediff = result->timestamp;
 		tsub(&timediff, &instdata->lasttime);
-		diff = tosec(timediff);
+		diff = pmtimevalToReal(&timediff);
 		wrap = 0;
 		if (avedata->desc.sem == PM_SEM_COUNTER) {
 		    diff *= avedata->scale;
@@ -725,7 +753,7 @@ calcaverage(pmResult *result)
 	    avedata = (aveData*) malloc(sizeof(aveData));
 	    newHashItem(vsp, &desc, avedata, &result->timestamp);
 	    if (__pmHashAdd(avedata->desc.pmid, (void*)avedata, &hashlist) < 0) {
-		pmiderr(avedata->desc.pmid, "failed %s hash table insertion\n", pmProgname);
+		pmiderr(avedata->desc.pmid, "failed %s hash table insertion\n", pmGetProgname());
 		/* free memory allocated above on insert failure */
 		for (j = 0; j < vsp->numval; j++)
 		    if (avedata->instlist[j]) free(avedata->instlist[j]);
@@ -736,6 +764,7 @@ calcaverage(pmResult *result)
 	else {	/* pmid exists - update statistics */
 	    avedata = (aveData*)hptr->data;
 	    for (j = 0; j < vsp->numval; j++) {	/* iterate thro result values */
+		int	fp_bad;
 		vp = &vsp->vlist[j];
 		k = j;	/* index into stored inst list, result may differ */
 		if ((vsp->numval > 1) || (avedata->desc.indom != PM_INDOM_NULL)) {
@@ -763,11 +792,19 @@ calcaverage(pmResult *result)
 		    pmiderr(avedata->desc.pmid, "failed to extract value: %s\n", pmErrStr(sts));
 		    continue;
 		}
-		if (isnan(av.d))
+		fp_bad = 0;
+#ifdef HAVE_FPCLASSIFY
+		fp_bad = fpclassify(av.d) == FP_NAN;
+#else
+#ifdef HAVE_ISNAN
+		fp_bad = isnan(av.d);
+#endif
+#endif
+		if (fp_bad)
 		    continue;
 		timediff = result->timestamp;
 		tsub(&timediff, &instdata->lasttime);
-		diff = tosec(timediff);
+		diff = pmtimevalToReal(&timediff);
 		wrap = 0;
 		if (avedata->desc.sem == PM_SEM_COUNTER) {
 		    diff *= avedata->scale;
@@ -776,26 +813,26 @@ calcaverage(pmResult *result)
 			val = av.d;
 		    else
 			val = unwrap(av.d, instdata->lastval, avedata->desc.type);
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_APPL0) {
-			char	*name = NULL;
-			pmNameID(avedata->desc.pmid, &name);
-			fprintf(stderr, "%s base value is %f, count %d\n",
-				name, val, instdata->count+1);
-			if (name) free(name);
+		    if (pmDebugOptions.appl0) {
+			int	numnames;
+			char	**names;
+			numnames = pmNameAll(avedata->desc.pmid, &names);
+			__pmPrintMetricNames(stderr, numnames, names, " or ");
+			fprintf(stderr, " base value is %f, count %d\n",
+				val, instdata->count+1);
+			if (numnames > 0) free(names);
 		    }
-#endif
 		    if (instdata->marked || val < instdata->lastval) {
 			/* either previous record was a "mark", or this is not */
 			/* the first one, and counter not monotonic increasing */
-#ifdef PCP_DEBUG
-			if (pmDebug & DBG_TRACE_APPL1) {
-			    char	*name = NULL;
-			    pmNameID(avedata->desc.pmid, &name);
-			    fprintf(stderr, "%s counter wrapped or <mark>\n", name);
-			    if (name) free(name);
+			if (pmDebugOptions.appl1) {
+			    int	numnames;
+			    char	**names;
+			    numnames = pmNameAll(avedata->desc.pmid, &names);
+			    __pmPrintMetricNames(stderr, numnames, names, " or ");
+			    fprintf(stderr, " counter wrapped or <mark>\n");
+			    if (numnames > 0) free(names);
 			}
-#endif
 			wrap = 1;
 			instdata->marked = 0;
 			tadd(&instdata->firsttime, &result->timestamp);
@@ -817,33 +854,34 @@ calcaverage(pmResult *result)
 			    instdata->sum = (val - instdata->lastval);
 			}
 			else {
-#ifdef PCP_DEBUG
-			    if (pmDebug & DBG_TRACE_APPL2) {
-				char	*name = NULL;
+			    if (pmDebugOptions.appl2) {
+				int	numnames;
+				char	**names;
 				char	*istr = NULL;
 
-				pmNameID(avedata->desc.pmid, &name);
+				numnames = pmNameAll(avedata->desc.pmid, &names);
 				if (pmNameInDom(avedata->desc.indom,
 				    instdata->inst, &istr) < 0)
 				    istr = NULL;
 				if (rate < instdata->min) {
-				    fprintf(stderr, "new min value for %s "
-						    "(inst[%s]: %f) at ",
-					name, (istr == NULL ? "":istr), rate);
-				    __pmPrintStamp(stderr, &result->timestamp);
+				    fprintf(stderr, "new min value for ");
+				    __pmPrintMetricNames(stderr, numnames, names, " or ");
+				    fprintf(stderr, " (inst[%s]: %f) at ",
+					(istr == NULL ? "":istr), rate);
+				    pmPrintStamp(stderr, &result->timestamp);
 				    fprintf(stderr, "\n");
 				}
 				if (rate > instdata->max) {
-				    fprintf(stderr, "new max value for %s "
-						    "(inst[%s]: %f) at ",
-					name, (istr == NULL ? "":istr), rate);
-				    __pmPrintStamp(stderr, &result->timestamp);
+				    fprintf(stderr, "new max value for ");
+				    __pmPrintMetricNames(stderr, numnames, names, " or ");
+				    fprintf(stderr, " (inst[%s]: %f) at ",
+					(istr == NULL ? "":istr), rate);
+				    pmPrintStamp(stderr, &result->timestamp);
 				    fprintf(stderr, "\n");
 				}
-				if (name) free(name);
+				if (numnames > 0) free(names);
 				if (istr) free(istr);
 			    }
-#endif
 			    if (rate < instdata->min) {
 				instdata->min = rate;
 				instdata->mintime = result->timestamp;
@@ -879,42 +917,42 @@ calcaverage(pmResult *result)
 		}
 		if (!wrap) {
 		    instdata->count++;
-#ifdef PCP_DEBUG
-		    if ((pmDebug & DBG_TRACE_APPL1) &&
+		    if (pmDebugOptions.appl1 &&
 			(avedata->desc.sem != PM_SEM_COUNTER || instdata->count > 0)) {
-			char	*name = NULL;
+			int	numnames;
+			char	**names;
 			double	metricspan = 0.0;
 			struct timeval	metrictimespan;
 
 			metrictimespan = result->timestamp;
 			tsub(&metrictimespan, &instdata->firsttime);
-			metricspan = tosec(metrictimespan);
-			pmNameID(avedata->desc.pmid, &name);
+			metricspan = pmtimevalToReal(&metrictimespan);
+			numnames = pmNameAll(avedata->desc.pmid, &names);
+			fprintf(stderr, "++ ");
+			__pmPrintMetricNames(stderr, numnames, names, " or ");
 
 			if (avedata->desc.sem == PM_SEM_COUNTER) {
-			    fprintf(stderr, "++ %s timedelta=%f count=%d\n"
+			    fprintf(stderr, " timedelta=%f count=%d\n"
 					    "sum=%f min=%f max=%f stocsum=%f\n"
 					    "rate=%f timesum=%f (+%f) timespan=%f\n",
-				    name, diff, instdata->count, instdata->sum,
+				    diff, instdata->count, instdata->sum,
 				    instdata->min, instdata->max,
 				    instdata->stocave, rate, instdata->timeave,
 				    diff * (val - instdata->lastval) / 2,
 				    metricspan);
 			}
 			else {	/* non-counters */
-			    fprintf(stderr, "++ %s timedelta=%f count=%d\n"
+			    fprintf(stderr, " timedelta=%f count=%d\n"
 					    "sum=%f min=%f max=%f stocsum=%f\n"
 					    "lastval=%f timesum=%f (+%f) timespan=%f\n",
-				    name, diff, instdata->count, instdata->sum,
+				    diff, instdata->count, instdata->sum,
 				    instdata->min, instdata->max,
 				    instdata->stocave, instdata->lastval,
 				    instdata->timeave, instdata->lastval*diff,
 				    metricspan);
 			}
-			if (name)
-			    free(name);
+			if (numnames > 0) free(names);
 		    }
-#endif
 		}
 		instdata->lastval = av.d;
 		instdata->lasttime = result->timestamp;
@@ -923,30 +961,26 @@ calcaverage(pmResult *result)
     }
 }
 
+static int
+override(int opt, pmOptions *opts)
+{
+    if (opt == 'a' || opt == 'H' || opt == 'N' || opt == 'p' || opt == 's')
+	return 1;
+    return 0;
+}
 
 int
 main(int argc, char *argv[])
 {
-    int			c, i, sts, trip, exitstatus=0;
-    char		*pmnsfile = PM_NS_DEFAULT;
-    char		*startstr = NULL;
-    char		*endstr = NULL;
-    char		*endnum = NULL;
-    char		*msg = NULL;
-    int			errflag = 0;
+    int			c, i, sts, trip, exitstatus = 0;
     int			lflag = 0;		/* no label by default */
     int			Hflag = 0;		/* no header by default */
     pmResult		*result;
-    pmLogLabel		label;			/* get hostname for archives */
     struct timeval 	timespan = {0, 0};
-    int			zflag = 0;		/* for -z */
-    char 		*tz = NULL;		/* for -Z timezone */
+    char		*endnum;
     char		*archive;
-    char		timebuf[32];		/* for pmCtime result + .xxx */
 
-    __pmSetProgname(argv[0]);
-
-    while ((c = getopt(argc, argv, "abB:D:fFHiIlmMNn:p:rsS:T:vxyzZ:?")) != EOF) {
+    while ((c = pmGetOptions(argc, argv, &opts)) != EOF) {
 	switch (c) {
 
 	case 'a':	/* provide all information */
@@ -961,26 +995,16 @@ main(int argc, char *argv[])
 	    break;
 
 	case 'B':	/* number of distribution bins */
-	    sts = (int)strtol(optarg, &endnum, 10);
+	    sts = (int)strtol(opts.optarg, &endnum, 10);
 	    if (*endnum != '\0' || sts < 0) {
-		fprintf(stderr, "%s: -B requires positive numeric argument\n", pmProgname);
-		errflag++;
+		pmprintf("%s: -B requires positive numeric argument\n",
+			pmGetProgname());
+		opts.errors++;
 	    }
 	    else
 		nbins = (unsigned int)sts;
 	    break;
 
-	case 'D':	/* debug flag */
-	    sts = __pmParseDebug(optarg);
-	    if (sts < 0) {
-		fprintf(stderr, "%s: unrecognized debug flag specification (%s)\n",
-		    pmProgname, optarg);
-		errflag++;
-	    }
-	    else
-		pmDebug |= sts;
-	    break;
-
 	case 'f':	/* spreadsheet format - use tab delimiters */
 	    delimiter = '\t';
 	    break;
@@ -1017,15 +1041,11 @@ main(int argc, char *argv[])
 	    warnflag = 0;
 	    break;
 
-	case 'n':	/* alternative name space file */
-	    pmnsfile = optarg;
-	    break;
-
 	case 'p':	/* number of digits after decimal point */
-	    precision = (unsigned int)strtol(optarg, &endnum, 10);
+	    precision = (unsigned int)strtol(opts.optarg, &endnum, 10);
 	    if (*endnum != '\0') {
-		fprintf(stderr, "%s: -p requires numeric argument\n", pmProgname);
-		errflag++;
+		pmprintf("%s: -p requires numeric argument\n", pmGetProgname());
+		opts.errors++;
 	    }
 	    break;
 
@@ -1034,16 +1054,6 @@ main(int argc, char *argv[])
 	    sumflag = 1;
 	    break;
 
-	case 'S':
-	    startstr = optarg;
-	    sflag = 1;
-	    break;
-
-	case 'T':
-	    endstr = optarg;
-	    tflag = 1;
-	    break;
-
 	case 'v':	/* verbose "fetch" warnings reported */
 	    warnflag = 1;
 	    break;
@@ -1057,144 +1067,54 @@ main(int argc, char *argv[])
 	case 'y':	/* print sample count */
 	    countflag = 1;
 	    break;
-
-	case 'z':	/* timezone from host */
-	    if (tz != NULL) {
-		fprintf(stderr, "%s: at most one of -Z and/or -z allowed\n", pmProgname);
-		errflag++;
-	    }
-	    zflag++;
-	    break;
-
-	case 'Z':	/* $TZ timezone */
-	    if (zflag) {
-		fprintf(stderr, "%s: at most one of -Z and/or -z allowed\n", pmProgname);
-		errflag++;
-	    }
-	    tz = optarg;
-	    break;
-
-	case '?':
-	default:
-	    errflag++;
-	    break;
 	}
     }
 
-    if (errflag || optind >= argc)
-	usage();
-
-    archive = argv[optind];
-    if ((sts = pmNewContext(PM_CONTEXT_ARCHIVE, archive)) < 0) {
-	fprintf(stderr, "%s: Cannot open archive \"%s\": %s\n", pmProgname, archive, pmErrStr(sts));
-	exit(1);
-    }
-    optind++;
-
-    if (pmnsfile != PM_NS_DEFAULT) {
-	if ((sts = pmLoadNameSpace(pmnsfile)) < 0) {
-	    fprintf(stderr, "%s: Cannot load namespace from \"%s\": %s\n", pmProgname, pmnsfile, pmErrStr(sts));
-	    exit(1);
-	}
+    if (!opts.errors && !(opts.flags & PM_OPTFLAG_EXIT) &&
+	opts.optind >= argc && !opts.archives) {
+	pmprintf("Error: no archive specified\n\n");
+	opts.errors++;
     }
 
-    if ((sts = pmTrimNameSpace()) < 0) {
-	fprintf(stderr, "%s: pmTrimNamespace failed: %s\n", pmProgname, pmErrStr(sts));
+    if (opts.errors || (opts.flags & PM_OPTFLAG_EXIT)) {
+	exitstatus = !(opts.flags & PM_OPTFLAG_EXIT);
+	pmUsageMessage(&opts);
+	exit(exitstatus);
+    }
+
+    if (opts.narchives > 0) {
+	archive = opts.archives[0];
+    } else {
+	archive = argv[opts.optind++];
+	__pmAddOptArchive(&opts, archive);
+    }
+    opts.flags &= ~PM_OPTFLAG_DONE;
+    __pmEndOptions(&opts);
+
+    if ((sts = c = pmNewContext(PM_CONTEXT_ARCHIVE, archive)) < 0) {
+	fprintf(stderr, "%s: Cannot open archive \"%s\": %s\n",
+		pmGetProgname(), archive, pmErrStr(sts));
 	exit(1);
     }
 
-    if ((sts = pmGetArchiveLabel(&label)) < 0) {
-	fprintf(stderr, "%s: Cannot get archive label record: %s\n",
-	    pmProgname, pmErrStr(sts));
-	exit(1);
+    if (pmGetContextOptions(c, &opts) < 0) {
+	pmflush();	/* runtime errors only at this stage */
+	exit(EXIT_FAILURE);
     }
-    else
-	logstart = label.ll_start;
-
-    if (zflag) {
-	if ((sts = pmNewContextZone()) < 0) {
-	    fprintf(stderr, "%s: Cannot set context timezone: %s\n",
-		pmProgname, pmErrStr(sts));
-	    exit(1);
-	}
-	printf("Note: timezone set to local timezone of host \"%s\" from archive\n\n",
-	    label.ll_hostname);
-    }
-    else if (tz != NULL) {
-	if ((sts = pmNewZone(tz)) < 0) {
-	    fprintf(stderr, "%s: Cannot set timezone to \"%s\": %s\n",
-		pmProgname, tz, pmErrStr(sts));
-	    exit(1);
-	}
-	printf("Note: timezone set to \"TZ=%s\"\n\n", tz);
-    }
-
-    if ((sts = pmGetArchiveEnd(&logend)) < 0) {
-	logend.tv_sec = INT_MAX;
-	logend.tv_usec = 0;
-	fflush(stdout);
-	fprintf(stderr, "%s: Cannot locate end of archive: %s\n",
-	    pmProgname, pmErrStr(sts));
-	fprintf(stderr, "\nWARNING: This archive is sufficiently damaged that it may not be possible to\n");
-	fprintf(stderr, "         produce complete information.  Continuing and hoping for the best.\n\n");
-	fflush(stderr);
-    }
-
-    if (tflag || sflag) {
-	if (pmParseTimeWindow(startstr, endstr, NULL, NULL, &logstart, &logend,
-		&windowstart, &windowend, &timespan, &msg) < 0) {
-	    fprintf(stderr, "%s: Invalid time window specified:\n%s\n", pmProgname, msg);
-	    exit(1);
-	}
-#ifdef PCP_DEBUG
-	else if (pmDebug & DBG_TRACE_APPL2) {
-	    fprintf(stderr, "parseTimeWindow: \n "
-		"Window start=%.0f end=%.0f sec\nArchive start=%.0f end=%.0f sec\n",
-		tosec(windowstart), tosec(windowend), tosec(logstart), tosec(logend));
-	    fprintf(stderr, "\nWindow Start: ");
-	    __pmPrintStamp(stderr, &windowstart);
-	    fprintf(stderr, "\nWindow End:   ");
-	    __pmPrintStamp(stderr, &windowend);
-	    fprintf(stderr, "\n\n");
-	}
-#endif
-    }
-    else {	/* time window covers whole log */
-	windowstart = logstart;
-	windowend = logend;
-    }
-    logspan = tosec(windowend) - tosec(windowstart);
-
-    if ((sts = pmSetMode(PM_MODE_FORW, &windowstart, 0)) < 0) {
-	fprintf(stderr, "%s: pmSetMode failed: %s\n", pmProgname, pmErrStr(sts));
+    
+    if ((sts = pmSetMode(PM_MODE_FORW, &opts.start, 0)) < 0) {
+	fprintf(stderr, "%s: pmSetMode failed: %s\n", pmGetProgname(), pmErrStr(sts));
 	exit(1);
     }
 
-    if (lflag == 1) {
-	char	       *ddmm;
-	char	       *yr;
+    if (lflag)
+	printlabel();
 
-	printf("Log Label (Log Format Version %d)\n", label.ll_magic & 0xff);
-	printf("Performance metrics from host %s\n", label.ll_hostname);
-
-	ddmm = pmCtime(&windowstart.tv_sec, timebuf);
-	ddmm[10] = '\0';
-	yr = &ddmm[20];
-	printf("  commencing %s ", ddmm);
-	__pmPrintStamp(stdout, &windowstart);
-	printf(" %4.4s\n", yr);
-
-	ddmm = pmCtime(&windowend.tv_sec, timebuf);
-	ddmm[10] = '\0';
-	yr = &ddmm[20];
-	printf("  ending     %s ", ddmm);
-	__pmPrintStamp(stdout, &windowend);
-	printf(" %4.4s\n", yr);
-    }
+    logspan = pmtimevalToReal(&opts.finish) - pmtimevalToReal(&opts.start);
 
     /* check which timestamp print format we should be using */
-    timespan = windowend;
-    tsub(&timespan, &windowstart);
+    timespan = opts.finish;
+    tsub(&timespan, &opts.start);
     if (timespan.tv_sec > 86400) /* seconds per day: 60*60*24 */
 	dayflag = 1;
 
@@ -1203,9 +1123,9 @@ main(int argc, char *argv[])
 	    if ((sts = pmFetchArchive(&result)) < 0)
 		break;
 
-	    if (windowend.tv_sec > result->timestamp.tv_sec ||
-		(windowend.tv_sec == result->timestamp.tv_sec &&
-		 windowend.tv_usec >= result->timestamp.tv_usec)) {
+	    if (opts.finish.tv_sec > result->timestamp.tv_sec ||
+		(opts.finish.tv_sec == result->timestamp.tv_sec &&
+		 opts.finish.tv_usec >= result->timestamp.tv_usec)) {
 		if (trip == 0)
 		    calcaverage(result);
 		else
@@ -1220,13 +1140,11 @@ main(int argc, char *argv[])
 	}
 
 	if (trip == 0 && nbins > 0) {	/* distribute values into bins */
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL0)
+	    if (pmDebugOptions.appl0)
 		fprintf(stderr, "resetting for second iteration\n");
-#endif
-	    if ((sts = pmSetMode(PM_MODE_FORW, &windowstart, 0)) < 0) {
+	    if ((sts = pmSetMode(PM_MODE_FORW, &opts.start, 0)) < 0) {
 		fprintf(stderr, "%s: pmSetMode reset failed: %s\n",
-		    pmProgname, pmErrStr(sts));
+		    pmGetProgname(), pmErrStr(sts));
 		exit(1);
 	    }
 	}
@@ -1235,21 +1153,21 @@ main(int argc, char *argv[])
     }
 
     if (sts != PM_ERR_EOL) {
-	fprintf(stderr, "%s: fetch failed: %s\n", pmProgname, pmErrStr(sts));
+	fprintf(stderr, "%s: fetch failed: %s\n", pmGetProgname(), pmErrStr(sts));
 	exitstatus = 1;
     }
 
     if (Hflag)
 	printheaders();
 
-    if (optind >= argc) {	/* print all results */
+    if (opts.optind >= argc) {	/* print all results */
 	if ((sts = pmTraversePMNS("", printsummary)) < 0) {
-	    fprintf(stderr, "%s: PMNS traversal failed: %s\n", pmProgname, pmErrStr(sts));
+	    fprintf(stderr, "%s: PMNS traversal failed: %s\n", pmGetProgname(), pmErrStr(sts));
 	    exit(1);
 	}
     }
     else {		/* print only selected results */
-	for (i = optind; i < argc; i++) {
+	for (i = opts.optind; i < argc; i++) {
 	    char *msg;
 
 	    if (pmParseMetricSpec(argv[i], 1, archive, &msp, &msg) < 0) {
@@ -1259,7 +1177,7 @@ main(int argc, char *argv[])
 	    }
 	    if ((sts = pmTraversePMNS(msp->metric, printsummary)) < 0)
 		fprintf(stderr, "%s: PMNS traversal failed for %s: %s\n",
-			pmProgname, msp->metric, pmErrStr(sts));
+			pmGetProgname(), msp->metric, pmErrStr(sts));
 	    pmFreeMetricSpec(msp);
 	}
     }
