@@ -17,7 +17,7 @@
  */
 
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 #include "pmnsutil.h"
 
 static FILE	*outf;
@@ -72,7 +72,7 @@ output(__pmnsNode *p, int depth, char *path)
 	return;
     }
     else if (depth < 0) {
-	if (lastdepth > 0 || (lastdepth == 1 && depth == -2))
+	if (lastdepth > 0)
 	    fprintf(outf, "}\n");
 	lastdepth = -1;
 	return;
@@ -83,10 +83,10 @@ output(__pmnsNode *p, int depth, char *path)
     if (p->first != NULL)
 	fprintf(outf, "\t%s\n", p->name);
     else {
-	if (pmid_domain(p->pmid) == DYNAMIC_PMID && pmid_item(p->pmid) == 0)
-	    fprintf(outf, "\t%s\t%d:*:*\n", p->name, pmid_cluster(p->pmid));
+	if (IS_DYNAMIC_ROOT(p->pmid))
+	    fprintf(outf, "\t%s\t%d:*:*\n", p->name, pmID_cluster(p->pmid));
 	else
-	    fprintf(outf, "\t%s\t%d:%d:%d\n", p->name, pmid_domain(p->pmid), pmid_cluster(p->pmid), pmid_item(p->pmid));
+	    fprintf(outf, "\t%s\t%d:%d:%d\n", p->name, pmID_domain(p->pmid), pmID_cluster(p->pmid), pmID_item(p->pmid));
     }
 }
 
