@@ -1,6 +1,7 @@
 /*
  * pmprobe - light-weight pminfo for configuring monitor apps
  *
+ * Copyright (c) 2013-2017 Red Hat.
  * Copyright (c) 2000-2001 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -12,15 +13,11 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include <unistd.h>
 #include "pmapi.h"
-#include "impl.h"
+#include "libpcp.h"
 
 static char	**namelist;
 static pmID	*pmidlist;
@@ -28,25 +25,66 @@ static int	listsize;
 static int	numpmid;
 static int	*instlist;
 static char	**instnamelist;
-static int	type = 0;
+
+static pmLongOptions longopts[] = {
+    PMAPI_OPTIONS_HEADER("General options"),
+    PMOPT_ARCHIVE,
+    PMOPT_DEBUG,
+    PMOPT_HOST,
+    PMOPT_CONTAINER,
+    PMOPT_DERIVED,
+    PMOPT_LOCALPMDA,
+    PMOPT_SPECLOCAL,
+    PMOPT_NAMESPACE,
+    PMOPT_ORIGIN,
+    PMOPT_TIMEZONE,
+    PMOPT_HOSTZONE,
+    { "version", 0, 'd', 0, "display version number and exit" },
+    PMOPT_HELP,
+    PMAPI_OPTIONS_HEADER("Reporting options"),
+    { "force", 0, 'f', 0, "report all pmGetIndom or pmGetInDomArchive instances" },
+    { "faster", 0, 'F', 0, "assume given metric names are PMNS leaf nodes" },
+    { "external", 0, 'I', 0, "list external instance names" },
+    { "internal", 0, 'i', 0, "list internal instance numbers" },
+    { "verbose", 0, 'V', 0, "report PDU operations (verbose)" },
+    { "values", 0, 'v', 0, "list metric values" },
+    PMAPI_OPTIONS_END
+};
+
+static int
+overrides(int opt, pmOptions *opts)
+{
+    if (opt == 'V')
+	return 1;	/* we've claimed this, inform pmGetOptions */
+    return 0;
+}
+
+static pmOptions opts = {
+    .flags = PM_OPTFLAG_STDOUT_TZ,
+    .short_options = "a:D:efh:IiK:Ln:FO:VvZ:z?",
+    .long_options = longopts,
+    .short_usage = "[options] [metricname ...]",
+    .override = overrides,
+};
 
 static void
 dometric(const char *name)
 {
     if (numpmid >= listsize) {
+	size_t size;
+
 	listsize = listsize == 0 ? 16 : listsize * 2;
-	if ((pmidlist = (pmID *)realloc(pmidlist, listsize * sizeof(pmidlist[0]))) == NULL) {
-	    __pmNoMem("realloc pmidlist", listsize * sizeof(pmidlist[0]), PM_FATAL_ERR);
-	}
-	if ((namelist = (char **)realloc(namelist, listsize * sizeof(namelist[0]))) == NULL) {
-	    __pmNoMem("realloc namelist", listsize * sizeof(namelist[0]), PM_FATAL_ERR);
-	}
+	size = listsize * sizeof(pmidlist[0]);
+	if ((pmidlist = (pmID *)realloc(pmidlist, size)) == NULL)
+	    pmNoMem("realloc pmidlist", size, PM_FATAL_ERR);
+	size = listsize * sizeof(namelist[0]);
+	if ((namelist = (char **)realloc(namelist, size)) == NULL)
+	    pmNoMem("realloc namelist", size, PM_FATAL_ERR);
     }
 
     namelist[numpmid]= strdup(name);
-    if (namelist[numpmid] == NULL) {
-	__pmNoMem("strdup name", strlen(name), PM_FATAL_ERR);
-    }
+    if (namelist[numpmid] == NULL)
+	pmNoMem("strdup name", strlen(name), PM_FATAL_ERR);
 
     numpmid++;
 }
@@ -62,7 +100,7 @@ lookup(pmInDom indom)
 	    free(instlist);
 	    free(instnamelist);
 	}
-	if (type == PM_CONTEXT_ARCHIVE)
+	if (opts.context == PM_CONTEXT_ARCHIVE)
 	    numinst = pmGetInDomArchive(indom, &instlist, &instnamelist);
 	else
 	    numinst = pmGetInDom(indom, &instlist, &instnamelist);
@@ -71,115 +109,54 @@ lookup(pmInDom indom)
     return numinst;
 }
 
-extern int do_local_spec(const char *);
-
 int
 main(int argc, char **argv)
 {
-    int		c;
-    int		sts;
+    int		c, i, j, sts;
     int		fetch_sts;
+    int		fetched;
     int		numinst;
-    int		fflag = 0;		/* -f for all instances from */
-					/*    pmGetIndom or pmGetIndomArchive */
+    int		fflag = 0;		/* -f pmGetIndom or pmGetIndomArchive for instances */
+    int		Fflag = 0;		/* -F for fast leaf names access */
     int		iflag = 0;		/* -i for instance numbers */
     int		Iflag = 0;		/* -I for instance names */
     int		vflag = 0;		/* -v for values */
     int		Vflag = 0;		/* -V for verbose */
-    int		zflag = 0;		/* -z for archive timezone */
-    int		errflag = 0;
-    char	*host = NULL;
-    char	*msg;
-    pmLogLabel	label;			/* get hostname for archives */
-    struct timeval	start;		/* start of sample window */
-    struct timeval	first;		/* initial sample time */
-    struct timeval	last;		/* final sample time */
-    char	local[MAXHOSTNAMELEN];
-    char	*pmnsfile = PM_NS_DEFAULT;
-    char	*Oflag = NULL;		/* -O for archive offset */
-    char	*tz = NULL;
-    int		i;
+    char	*source;
     pmResult	*result;
     pmValueSet	*vsp;
     pmDesc	desc;
-    char        *opts = "a:D:fh:IiK:Ln:O:VvZ:z?";
 
-    __pmSetProgname(argv[0]);
-
-    while ((c = getopt(argc, argv, opts)) != EOF) {
+    while ((c = pmGetOptions(argc, argv, &opts)) != EOF) {
 	switch (c) {
 
-	case 'a':	/* archive name */
-	    if (type != 0) {
-		fprintf(stderr, "%s: at most one of -a, -h and -L allowed\n", pmProgname);
-		errflag++;
-	    }
-	    type = PM_CONTEXT_ARCHIVE;
-	    host = optarg;
-	    break;
-
-	case 'D':	/* debug flag */
-	    sts = __pmParseDebug(optarg);
-	    if (sts < 0) {
-		fprintf(stderr, "%s: unrecognized debug flag specification (%s)\n",
-		    pmProgname, optarg);
-		errflag++;
-	    }
-	    else
-		pmDebug |= sts;
-	    break;
-
 	case 'f':	/* pmGetIndom or pmGetInDomArchive for instances with -i or -I */
 	    fflag++;
 	    break;
 
-	case 'h':	/* contact PMCD on this hostname */
-	    if (type != 0) {
-		fprintf(stderr, "%s: at most one of -a, -h and -L allowed\n", pmProgname);
-		errflag++;
-	    }
-	    host = optarg;
-	    type = PM_CONTEXT_HOST;
-	    break;
-
 	case 'i':	/* report internal instance numbers */
 	    if (vflag) {
-		fprintf(stderr, "%s: at most one of -i and -v allowed\n", pmProgname);
-		errflag++;
+		pmprintf("%s: at most one of -i and -v allowed\n", pmGetProgname());
+		opts.errors++;
 	    }
 	    iflag++;
 	    break;
 
 	case 'I':	/* report external instance names */
 	    if (vflag) {
-		fprintf(stderr, "%s: at most one of -I and -v allowed\n", pmProgname);
-		errflag++;
+		pmprintf("%s: at most one of -I and -v allowed\n", pmGetProgname());
+		opts.errors++;
 	    }
 	    Iflag++;
 	    break;
 
-	case 'K':	/* update local PMDA table */
-	    if ((msg = __pmSpecLocalPMDA(optarg)) != NULL) {
-		fprintf(stderr, "%s: __pmSpecLocalPMDA failed\n%s\n", pmProgname, msg);
-		errflag++;
-	    }
+	case 'F':	/* optimised access to leaf names */
+	    Fflag++;
 	    break;
 
-	case 'L':	/* local PMDA connection, no PMCD */
-	    if (type != 0) {
-		fprintf(stderr, "%s: at most one of -a, -h and -L allowed\n", pmProgname);
-		errflag++;
-	    }
-	    host = NULL;
-	    type = PM_CONTEXT_LOCAL;
-	    break;
-
-	case 'n':	/* alternative name space file */
-	    pmnsfile = optarg;
-	    break;
-
-	case 'O':	/* sample origin */
-	    Oflag = optarg;
+	case 'd':	/* version (d'oh - 'V' already in use) */
+	    pmprintf("%s version %s\n", pmGetProgname(), PCP_VERSION);
+	    opts.flags |= PM_OPTFLAG_EXIT;
 	    break;
 
 	case 'V':	/* verbose */
@@ -188,142 +165,67 @@ main(int argc, char **argv)
 
 	case 'v':	/* cheap values */
 	    if (iflag || Iflag) {
-		fprintf(stderr, "%s: at most one of -v and (-i or -I) allowed\n", pmProgname);
-		errflag++;
+		pmprintf("%s: at most one of -v and (-i or -I) allowed\n", pmGetProgname());
+		opts.errors++;
 	    }
 	    vflag++;
 	    break;
 
-	case 'z':	/* timezone from host */
-	    if (tz != NULL) {
-		fprintf(stderr, "%s: at most one of -Z and/or -z allowed\n", pmProgname);
-		errflag++;
-	    }
-	    zflag++;
-	    break;
-
-	case 'Z':	/* $TZ timezone */
-	    if (zflag) {
-		fprintf(stderr, "%s: at most one of -Z and/or -z allowed\n", pmProgname);
-		errflag++;
-	    }
-	    tz = optarg;
-	    break;
-
-	case '?':
 	default:
-	    errflag++;
+	    opts.errors++;
 	    break;
 	}
     }
 
-    if (type != PM_CONTEXT_ARCHIVE && zflag) {
-	fprintf(stderr, "%s: -z requires an archive source\n", pmProgname);
-	errflag++;
+    if (opts.errors || (opts.flags & PM_OPTFLAG_EXIT)) {
+	sts = !(opts.flags & PM_OPTFLAG_EXIT);
+	pmUsageMessage(&opts);
+	exit(sts);
     }
 
-    if (type != PM_CONTEXT_ARCHIVE && Oflag != NULL) {
-	fprintf(stderr, "%s: -O requires an archive source\n", pmProgname);
-	errflag++;
+    if (opts.context == PM_CONTEXT_ARCHIVE)
+	source = opts.archives[0];
+    else if (opts.context == PM_CONTEXT_HOST)
+	source = opts.hosts[0];
+    else if (opts.context == PM_CONTEXT_LOCAL)
+	source = NULL;
+    else {
+	opts.context = PM_CONTEXT_HOST;
+	source = "local:";
     }
 
-
-    if (errflag) {
-	fprintf(stderr,
-"Usage: %s [options] [metricname ...]\n\
-\n\
-Options:\n\
-  -a archive   metrics source is a PCP log archive\n\
-  -f           report all instances from pmGetIndom or pmGetInDomArchive\n\
-	       (with -i or -I)\n\
-               [default is to report instances from pmFetch]\n\
-  -h host      metrics source is PMCD on host\n\
-  -I           list external instance names\n\
-  -i           list internal instance numbers\n\
-  -K spec      optional additional PMDA spec for local connection\n\
-               spec is of the form op,domain,dso-path,init-routine\n\
-  -L           metrics source is local connection to PMDA, no PMCD\n\
-  -n pmnsfile  use an alternative PMNS\n\
-  -O time      origin for a fetch from the archive\n\
-  -V           report PDU operations (verbose)\n\
-  -v           list metric values\n\
-  -Z timezone  set timezone for -O\n\
-  -z           set timezone for -O to local time for host from -a\n",
-		pmProgname);
-	exit(1);
-    }
-
-    if (pmnsfile != PM_NS_DEFAULT) {
-	if ((sts = pmLoadNameSpace(pmnsfile)) < 0) {
-	    printf("%s: Cannot load namespace from \"%s\": %s\n", pmProgname, pmnsfile, pmErrStr(sts));
-	    exit(1);
-	}
-    }
-
-    if (type == 0) {
-	type = PM_CONTEXT_HOST;
-	(void)gethostname(local, MAXHOSTNAMELEN);
-	local[MAXHOSTNAMELEN-1] = '\0';
-	host = local;
-    }
-    if ((sts = pmNewContext(type, host)) < 0) {
-	if (type == PM_CONTEXT_HOST)
+    if ((sts = c = pmNewContext(opts.context, source)) < 0) {
+	if (opts.context == PM_CONTEXT_HOST)
 	    fprintf(stderr, "%s: Cannot connect to PMCD on host \"%s\": %s\n",
-		pmProgname, host, pmErrStr(sts));
-	else if (type == PM_CONTEXT_LOCAL)
+		pmGetProgname(), source, pmErrStr(sts));
+	else if (opts.context == PM_CONTEXT_LOCAL)
 	    fprintf(stderr, "%s: Cannot make standalone connection on localhost: %s\n",
-		    pmProgname, pmErrStr(sts));
+		    pmGetProgname(), pmErrStr(sts));
 	else
 	    fprintf(stderr, "%s: Cannot open archive \"%s\": %s\n",
-		pmProgname, host, pmErrStr(sts));
+		pmGetProgname(), source, pmErrStr(sts));
 	exit(1);
     }
 
-    if (type == PM_CONTEXT_ARCHIVE) {
-	if ((sts = pmGetArchiveLabel(&label)) < 0) {
-	    fprintf(stderr, "%s: Cannot get archive label record: %s\n",
-		pmProgname, pmErrStr(sts));
-	    exit(1);
-	}
-	first = label.ll_start;
-	if ((sts = pmGetArchiveEnd(&last)) < 0) {
-	    last.tv_sec = INT_MAX;
-	    last.tv_usec = 0;
-	}
-	if (zflag) {
-	    if ((sts = pmNewContextZone()) < 0) {
-		fprintf(stderr, "%s: Cannot set context timezone: %s\n",
-			pmProgname, pmErrStr(sts));
-		exit(1);
-	    }
-	}
-	else if (tz != NULL) {
-	    if ((sts = pmNewZone(tz)) < 0) {
-		fprintf(stderr, "%s: Cannot set timezone to \"%s\": %s\n",
-			pmProgname, tz, pmErrStr(sts));
-		exit(1);
-	    }
-	}
-	else
-	    pmNewContextZone();
-
-	if (pmParseTimeWindow(NULL, NULL, NULL, Oflag,
-				   &first, &last,
-				   &last, &first, &start, &msg) < 0) {
-	    fprintf(stderr, "%s: %s", pmProgname, msg);
+    /* complete TZ and time window option (origin) setup */
+    if (opts.context == PM_CONTEXT_ARCHIVE) {
+	if (pmGetContextOptions(c, &opts)) {
+	    pmflush();
 	    exit(1);
 	}
     }
 
-    if (optind >= argc) {
+    if (opts.optind >= argc)
 	pmTraversePMNS("", dometric);
+    else if (Fflag) {
+	for (i = opts.optind; i < argc; i++)
+	    dometric(argv[i]);
     }
     else {
-	int	a;
-	for (a = optind; a < argc; a++) {
-	    sts = pmTraversePMNS(argv[a], dometric);
+	for (i = opts.optind; i < argc; i++) {
+	    sts = pmTraversePMNS(argv[i], dometric);
 	    if (sts < 0)
-		printf("%s %d %s\n", argv[a], sts, pmErrStr(sts));
+		printf("%s %d %s\n", argv[i], sts, pmErrStr(sts));
 	}
     }
 
@@ -331,9 +233,8 @@ Options:\n\
      * Cull out names that were unsuccessfully looked up.
      * However, it is unlikely to fail because names come from a traverse PMNS.
      */
-    if ((sts = pmLookupName(numpmid, namelist, pmidlist)) < 0) {
-        int i, j = 0;
-	for (i = 0; i < numpmid; i++) {
+    if (numpmid > 0 && (sts = pmLookupName(numpmid, namelist, pmidlist)) < 0) {
+	for (i = j = 0; i < numpmid; i++) {
 	    if (pmidlist[i] == PM_ID_NULL) {
 		printf("%s %d %s\n", namelist[i], sts, pmErrStr(sts));
 		free(namelist[i]);
@@ -348,7 +249,7 @@ Options:\n\
 	numpmid = j;
     }
 
-    fetch_sts = 0;
+    fetch_sts = fetched = 0;
     for (i = 0; i < numpmid; i++) {
 	printf("%s ", namelist[i]);
 
@@ -389,27 +290,28 @@ Options:\n\
 	    continue;
 	}
 
-	if (type == PM_CONTEXT_ARCHIVE) {
+	if (opts.context == PM_CONTEXT_ARCHIVE) {
 	    /*
 	     * merics from archives are fetched one at a time, otherwise
 	     * get them all at once
 	     */
-	    if ((sts = pmSetMode(PM_MODE_FORW, &start, 0)) < 0) {
+	    if ((sts = pmSetMode(PM_MODE_FORW, &opts.origin, 0)) < 0) {
 		printf("%d %s (pmSetMode)\n", sts, pmErrStr(sts));
 		continue;
 	    }
 	    fetch_sts = pmFetch(1, &pmidlist[i], &result);
 	}
 	else {
-	    if (i == 0)
+	    if (!fetched)
 		fetch_sts = pmFetch(numpmid, pmidlist, &result);
+	    fetched = 1;
 	}
 
 	if (fetch_sts < 0) {
 	    printf("%d %s (pmFetch)", fetch_sts, pmErrStr(fetch_sts));
 	}
 	else {
-	    if (type == PM_CONTEXT_ARCHIVE)
+	    if (opts.context == PM_CONTEXT_ARCHIVE)
 	    	vsp = result->vset[0];
 	    else
 	    	vsp = result->vset[i];
@@ -424,7 +326,6 @@ Options:\n\
 	    else if (vsp->numval > 0) {
 		printf("%d", vsp->numval);
 		if (vflag) {
-		    int		j;
 		    for (j = 0; j < vsp->numval; j++) {
 			pmValue	*vp = &vsp->vlist[j];
 			putchar(' ');
@@ -439,7 +340,6 @@ Options:\n\
 		     * pmResult, so each instance in the pmResult should be
 		     * found by pmGetInDom or pmGetInDomArchive
 		     */
-		    int		j;
 		    if (desc.indom == PM_INDOM_NULL) {
 			printf(" PM_IN_NULL");
 			if ( iflag && Iflag )
@@ -473,21 +373,16 @@ Options:\n\
 	putchar('\n');
     }
 
-   if (Vflag) {
-	int j;
-
-	/* Warning: 16 is magic and from libpcp/src/pdu.c */
+    if (Vflag) {
 	printf("PDUs send");
-	j = 0;
-	for (i = 0; i < 16; i++) {
+	for (i = j = 0; i < PDU_MAX; i++) {
 	    printf(" %3d", __pmPDUCntOut[i]);
 	    j += __pmPDUCntOut[i];
 	}
 	printf("\nTotal: %d\n", j);
 
 	printf("PDUs recv");
-	j = 0;
-	for (i = 0; i < 16; i++) {
+	for (i = j = 0; i < PDU_MAX; i++) {
 	    printf(" %3d",__pmPDUCntIn[i]);
 	    j += __pmPDUCntIn[i];
 	}
