@@ -10,18 +10,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
  */
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <errno.h>
-#include <string.h>
-#include "platform_defs.h"
+#include "pmapi.h"
 #include "trace.h"
 
 typedef enum { UNKNOWN, COUNTER, OBSERVE, TRANSACT } trace_t;
@@ -46,8 +37,8 @@ main(int argc, char **argv)
 	} arg;
     } api;
 
-    for (p = me = argv[0]; *p; p++)	/* trim command name */
-	if (*p == '/') me = p+1;
+    pmSetProgname(argv[0]);
+    me = pmGetProgname();
 
     memset(&api, 0, sizeof(api));
 
@@ -139,22 +130,23 @@ Options:\n\
     }
 
     pmtracestate(api.state | PMTRACE_STATE_ASYNC);
-    c = 0;	/* reuse as the exit status */
 
     if (api.host != NULL) {
-	if ((p = (char *)malloc(strlen(api.host) + 20)) == NULL) {
+	c = strlen(api.host) + 20;
+	if ((p = (char *)malloc(c)) == NULL) {
 	    fprintf(stderr, "%s: malloc failed: %s\n",
-		    me, pmtraceerrstr(-errno));
+		    me, pmtraceerrstr(-oserror()));
 	    exit(0);
 	}
-	sprintf(p, "PCP_TRACE_HOST=%s", api.host);
+	pmsprintf(p, c, "PCP_TRACE_HOST=%s", api.host);
 	if (putenv(p) < 0) {
 	    fprintf(stderr, "%s: putenv failed: %s\n",
-		    me, pmtraceerrstr(-errno));
+		    me, pmtraceerrstr(-oserror()));
 	    exit(0);
 	}
     }
 
+    c = 0;	/* reuse as the exit status */
     switch (api.type) {
     case COUNTER:
 	if ((err = pmtracecounter(api.tag, api.arg.value)) < 0) {
@@ -189,7 +181,7 @@ Options:\n\
 
 	if ((c = system(api.arg.command)) < 0) {
 	    fprintf(stderr, "%s: system error running '%s': %s\n",
-		    me, api.arg.command, pmtraceerrstr(-errno));
+		    me, api.arg.command, pmtraceerrstr(-oserror()));
 	    exit(0);
 	}
 	if (err < 0)
